//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>
#include <sstream>

#include "g2pcerad.hh"
#include "g2pcera_trace.hh"
#include "g2pcera_config.hh"
#include "g2pcera_dumper.hh"
#include "g2pcera_dm.hh"
#include "g2pcera_algovis.hh"


////////////////////////////
// Class GnsCall
////////////////////////////
PCERA::GnsCall::GnsCall(G2PCERA *			parent,
			const call_ident_t &		ident,
			const net_res_spec_t &		src,
			const res_spec_t &		dst,
			const call_info_t &		callInfo,
			const recovery_info_t &		recInfo,
			const lsp_info_t &		lspInfo) :
	Ancestor<true, G2PCERA>(parent)
{
	METIN();

	ident_            = ident;
	src_              = src;
	dst_              = dst;
	callInfo_         = callInfo;
	recInfo_          = recInfo;
	lspInfo_          = lspInfo;

	// consistency hack
	lspInfo_.mask_.times = callInfo_.mask_.times;
	lspInfo_.times = callInfo_.times;


	disjoinnessLevel_ = DISJ_LEVEL_UNKNOWN;
	isConfirmed_      = false;

	memset(&srcHop_, 0, sizeof(srcHop_));
	memset(&dstHop_, 0, sizeof(dstHop_));

	gnsCallEro_.clear();

	memset(&eTnaRes_, 0, sizeof(eTnaRes_));

	METOUT();
}

PCERA::GnsCall::GnsCall(G2PCERA *			parent,
			const call_ident_t &		ident,
			const ero_hop_t &		src,
			const ero_hop_t &		dst,
			const lsp_info_t &		lspInfo) :
	Ancestor<true, G2PCERA>(parent)
{
	METIN();

	ident_   = ident;
	srcHop_  = src;
	dstHop_  = dst;
	lspInfo_ = lspInfo;

	disjoinnessLevel_ = DISJ_LEVEL_UNKNOWN;
	isConfirmed_      = false;

	memset(&src_, 0, sizeof(src_));
	memset(&dst_, 0, sizeof(dst_));
	memset(&callInfo_, 0, sizeof(callInfo_));
	memset(&recInfo_, 0, sizeof(recInfo_));

	gnsCallEro_.clear();

	memset(&eTnaRes_, 0, sizeof(eTnaRes_));

	METOUT();
}

PCERA::GnsCall::~GnsCall(void)
{
	METIN();

	gnsCallEro_.clear();

	METOUT();
}

call_ident_t
PCERA::GnsCall::ident(void)
{
	METIN();

	METOUT();
	return ident_;
}


void
PCERA::GnsCall::dump(const std::string &	prefix,
		     bool			recursive)
{
	G2PCERA_DEBUG("%sGnsCall", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0, ident_);


	G2PCERA_DEBUG("%s- src",
		      prefix0.c_str());
	logDump(prefix1, src_);

	G2PCERA_DEBUG("%s- dst",
		      prefix0.c_str());
	logDump(prefix1, dst_);

	G2PCERA_DEBUG("%s- srcHop",
		      prefix0.c_str());
	logDump(prefix1, srcHop_);

	G2PCERA_DEBUG("%s- dstHop",
		      prefix0.c_str());
	logDump(prefix1, dstHop_);

	logDump(prefix0, callInfo_);
	logDump(prefix0, recInfo_);
	logDump(prefix0, lspInfo_);

	G2PCERA_DEBUG("%s- isConfirmed         : %s",
		      prefix0.c_str(), isConfirmed_ ? "YES" : "NO") ;

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());

	G2PCERA_DEBUG("%s- Primary connection",
		      prefix0.c_str());
	primary_.dump(prefix1, recursive);

	G2PCERA_DEBUG("%s- Secondary connection",
		      prefix0.c_str());
	secondary_.dump(prefix1, recursive);

	G2PCERA_DEBUG("%s- disjoinnessLevel    : %s",
		      prefix0.c_str(),
		      DISJOINTNESS_LVL_TO_STRING(disjoinnessLevel_));

	G2PCERA_DEBUG("%s- gnsCall ERO",
		      prefix0.c_str());

	std::list<ero_hop_t>::iterator iter;
	for (iter = gnsCallEro_.begin(); iter != gnsCallEro_.end(); iter++) {
		logDump(prefix1, *iter);
	}

	if (gnsCallEro_.empty()) {
		G2PCERA_DEBUG("%s***EMPTY***", prefix1.c_str());
	}

	G2PCERA_DEBUG("%s- eTnaRes",
		      prefix0.c_str());
	logDump(prefix1, eTnaRes_);

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
}

bool
PCERA::GnsCall::route(void)
{
	METIN();

	std::list<ero_hop_t> ero;

	if (!resolveSrcHopFromSrcInfo()) {
		G2PCERA_ERROR("Cannot resolve SRC info "
			      "while routing GnsCall %s",
			      callIdent2string(ident()).c_str());
		METOUT();
		return false;
	}
	//logDump("[SRC-HOP]", srcHop_);

	if (!resolveEgressTnaFromDstInfo()) {
		G2PCERA_ERROR("Cannot resolve egress TNA "
			      "while routing GnsCall %s",
			      callIdent2string(ident()).c_str());
		METOUT();
		return false;
	}
	//logDump("[E-TNA]", eTnaRes_);

	if (!computeOneRoute(ero)) {
		G2PCERA_ERROR("Cannot compute one route for GnsCall %s",
			      callIdent2string(ident()).c_str());
		METOUT();
		return false;
	}

	// XXX-FIXME: add other cases

	primary_.flushEro();
	secondary_.flushEro();

	if (!ero.size()) {
		G2PCERA_WARNING("No route exists for GnsCall %s",
				callIdent2string(ident()).c_str());
		METOUT();
		return true;
	}

	std::list<ero_hop_t>::iterator iter;

	for (iter = ero.begin(); iter != ero.end(); iter++) {
		if (!addEroHopToConnection(LSP_ROLE_PRIMARY,
					   (*iter),
					   false /*onTop*/)) {
			G2PCERA_ERROR("Problems while inserting ero hop into "
				      "connection for GnsCall %s",
				      callIdent2string(ident()).c_str());
			primary_.flushEro();
			METOUT();
			return false;
		}
	}

	if (!resolveGnsCallEro()) {
		G2PCERA_ERROR("Problems while retrieving gnsCallEro "
			      "from connections in GnsCall %s",
			      callIdent2string(ident()).c_str());
		primary_.flushEro();
		METOUT();
		return false;
	}

	dump("[GnsCall::route]", true);

	METOUT();
	return true;
}

bool
PCERA::GnsCall::resolveGnsCallEro(void)
{
	METIN();

	std::list<ero_hop_t> ero;

	gnsCallEro_.clear();
	ero = getConnectionEro(LSP_ROLE_PRIMARY);

	if (ero.empty()) {
		G2PCERA_ERROR("Cannot resolve gnsCallEro. "
			      "PRIMARY ERO for GnsCall %s is empty",
			      callIdent2string(ident()).c_str());
		METOUT();
		return false;
	}

#ifdef G2PCERA_ALGO_STUB
	gnsCallEro_ = stubGnsEroComp(parent()->rootNodeId(), eTnaRes_.tna,ero);
#else
	std::list<ero_hop_t>::iterator iter, jiter;

	memset(&dstHop_, 0, sizeof(dstHop_));

	for (iter = ero.begin(); iter != ero.end(); iter++) {

		jiter = iter;

		if (!(*iter).idmn) continue;

		if (jiter != ero.begin() && gnsCallEro_.empty()) {
			ero_hop_t hop;
			hop = *(--jiter);
			gnsCallEro_.push_front(hop);
			dstHop_.node = hop.node;
		}

		gnsCallEro_.push_back(*iter);
	}

	if (gnsCallEro_.empty() && !ero.empty()) {
		gnsCallEro_.push_front(ero.back());
		dstHop_.node = ero.back().node;
	}
#endif

	if (gnsCallEro_.empty()) {
		G2PCERA_ERROR("Problems while resolving GnsCallEro");
		METOUT();
		return false;
	}
	METOUT();
	return true;
}
bool
PCERA::GnsCall::resolveSrcHopFromSrcInfo(void)
{
	METIN();

	if(!srcHop_.is_null()) {
		G2PCERA_DEBUG("SrcHop just declared!"
			      "No src info to check for GnsCall %s... ",
			      callIdent2string(ident()).c_str());
		METOUT();
		return true;
	}

	if (!BITMASK_BITTEST(src_.mask_, tna)) {
		G2PCERA_ERROR("No src TNA declared for GnsCall %s... ",
			      callIdent2string(ident()).c_str());
		METOUT();
		return false;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	ptr;
	node_ident_t		nid;
	tna_ident_t		tna;

	nid.id   = parent()->rootNodeId();
	nid.type = TOPO_NODE_NETWORK;

	tna.rc_id = 0;
	tna.rid   = nid.id;
	tna.id    = src_.tna;

	tmp = parent()->getTopology().getNode(nid);
	ptr = static_cast<PCERA::NetNode *>(tmp);
	if (!ptr) {
		G2PCERA_ERROR("Unrecoverable error: "
			      "rootNodeId %s does not exist in Topology..."
			      "asserting",
			      nodeIdent2string(nid).c_str());
		assert(0);
	}

	if (!ptr->getTna(tna)) {
		G2PCERA_ERROR("Ingress TNA %s is not present in "
			      "rootNodeId %s",
			      tnaIdent2string(tna).c_str(),
			      nodeIdent2string(nid).c_str());
		METOUT();
		return false;
	}

	srcHop_.node = nid.id;
	memset(&(srcHop_.te_link),
	       0,
	       sizeof(srcHop_.te_link));
	srcHop_.upstream_datalink = src_.data_link;
	memset(&(srcHop_.downstream_datalink),
	       0,
	       sizeof(srcHop_.downstream_datalink));
	srcHop_.upstream_label = src_.label;
	srcHop_.downstream_label = src_.label; // or 0??
	srcHop_.loose = 0;

	G2PCERA_DEBUG("srcHop set to for GnsCall %s",
		      callIdent2string(ident()).c_str());
	METOUT();
	return true;
}

bool
PCERA::GnsCall::resolveEgressTnaFromDstInfo(void)
{
	METIN();

	memset(&eTnaRes_, 0, sizeof(eTnaRes_));

	if (!BITMASK_BITTEST(dst_.mask_, net) &&
	    !BITMASK_BITTEST(dst_.mask_, grid)) {
		G2PCERA_WARNING("No dst info for GnsCall %s... "
				"probably a dstHop",
				callIdent2string(ident()).c_str());
		METOUT();
		return true;
	}

	if (BITMASK_BITTEST(dst_.mask_, net)) {
		eTnaRes_.tna = dst_.net.tna;
		BITMASK_BITSET(eTnaRes_.mask_, tna);
		METOUT();
		return true;
	} else if (BITMASK_BITTEST(dst_.mask_, grid)) {

		if (BITMASK_BITTEST(dst_.grid.mask_, tot_cpu_count)) {
			uint32_t reqCPUCount;

			reqCPUCount = dst_.grid.tot_cpu_count.lower_bound +
				(dst_.grid.tot_cpu_count.lb_included ? 0 : 1);

			if (dst_.grid.tot_cpu_count.upper_bound >
			    dst_.grid.tot_cpu_count.lower_bound) {
				reqCPUCount =
					dst_.grid.tot_cpu_count.upper_bound +
					(dst_.grid.tot_cpu_count.ub_included ?
					 0 : 1);
			}

			eTnaRes_.tna =
				stubTnaFromGridCE(parent()->getTopology(),
						  reqCPUCount);
		} else {
			eTnaRes_.tna =
				stubTnaFromGridSE(parent()->getTopology());
		}

		if (is_addr_null(eTnaRes_.tna) != 0) {
			G2PCERA_ERROR("Cannot resolve egress TNA "
				      "from Grid dst info");
			METOUT();
			return false;
		}

		BITMASK_BITSET(eTnaRes_.mask_, tna);
		METOUT();
		return true;
	}

	G2PCERA_ERROR("Corrupted dst info ! "
		      "Cannot resolve egress TNA from dst info");
	METOUT();
	return false;
}

bool
PCERA::GnsCall::computeOneRoute(std::list<ero_hop_t> & pEro)
{
	pEro.clear();

	//algoStubVisitor vis;
	//parent()->getTopology().accept(vis);

#ifdef G2PCERA_ALGO_STUB
	pEro = stubEroComp(parent()->rootNodeId(), eTnaRes_.tna, dstHop_.node);
#else
	if (parent()->spfSelector() == spfDijkstra) {
		NetNode * dst;
		uint32_t eRid;

		dst = dijkstraRun(eRid);
		if (!assembleEro(dst, eRid, pEro)) {
			G2PCERA_ERROR("Problems in computeOneRoute. "
				      "Cannot assemble ERO!");
			METOUT();
			return false;
		}
	}
#endif // G2PCERA_ALGO_STUB

	if (pEro.empty()) {
		G2PCERA_ERROR("Problems in computeOneRoute. No route found!");
		METOUT();
		return false;
	}

	std::list<ero_hop_t>::iterator iter;
	for (iter = pEro.begin(); iter != pEro.end(); iter++) {
		logDump("[computeOneRoute]", *iter);
	}

	METOUT();
	return true;
}

void
addNodeSorted(std::list<PCERA::NetNode *>  & nList, PCERA::NetNode * nItem)
{
	if (!nItem) {
		return;
	}

	std::list<PCERA::NetNode *>::iterator iter;
	for (iter = nList.begin(); iter != nList.end(); iter++) {
		assert(*iter);
		if ((*iter)->rootCost() > nItem->rootCost()) {
			nList.insert(iter, nItem);
			return;
		}
	}

	nList.push_back(nItem);
}

void
moveNodeSorted(std::list<PCERA::NetNode *>  & nList, PCERA::NetNode * nItem)
{
	if (!nItem || nList.empty()) {
		return;
	}

	std::list<PCERA::NetNode *>::iterator iter;
	for (iter = nList.begin(); iter != nList.end(); iter++) {

		assert(*iter);

		if ((*iter) == nItem) {
			nList.erase(iter);
			break;
		}
	}

	addNodeSorted(nList, nItem);
}

PCERA::NetNode *
PCERA::GnsCall::dijkstraRun(uint32_t & eRid)
{
	METIN();

	std::list<NetNode *>	candidateNodes;
	std::list<TeLink *>	currentLinks;
	PCERA::Node *		tmp;
	NetNode *		rootNode;
	NetNode *		currentNode;
	NetNode *		previousNode;
	NetNode *		nextNode;
	TeLink  *		currentLink;
	TeLink  *		reverseLink;
	uint32_t		dist;
	bool			s4tna;
	bool			s4rid;
	bool			s4tel;
	uint32_t		rev;
	nodeFlags_t		flags;
	node_ident_t		nid;
	std::string		prefix;

	eRid = 0;
	candidateNodes.clear();
	currentLinks.clear();
	currentNode  = 0;
	previousNode = 0;
	nextNode     = 0;
	currentLink  = 0;
	reverseLink  = 0;
	dist         = 0;
	s4tna        = false;
	s4rid        = false;
	s4tel        = false;

	prefix = "[Dijkstra init  ]";

	nid.id   = parent()->rootNodeId();
	nid.type = TOPO_NODE_NETWORK;

	tmp = parent()->getTopology().getNode(nid);
	rootNode = static_cast<PCERA::NetNode *>(tmp);
	if (!rootNode) {
		G2PCERA_ERROR("%s Unrecoverable error: "
			      "rootNodeId %s does not exist in Topology!"
			      "asserting....",
			      prefix.c_str(), nodeIdent2string(nid).c_str());
		METOUT();
		assert(0);
	}

	if (!rootNode->isUp()) {
		G2PCERA_WARNING("%s G2PCERA ROOT node %s is DOWN !",
				prefix.c_str(), nodeIdent2string(nid).c_str());
		METOUT();
		return 0;
	}

	if (!rootNode->isEnabled()) {
		G2PCERA_WARNING("%s G2PCERA ROOT node %s is DISABLED !\n",
				prefix.c_str(), nodeIdent2string(nid).c_str());
		METOUT();
		return 0;
	}

	G2PCERA_DEBUG(DUMP_BAR1 DUMP_BAR1);

	std::string err;
	if (isValid(dstHop_, err)) {
		if (is_addr_null(dstHop_.te_link) != 0) {
			G2PCERA_DEBUG("%s Searching route to NODE %s",
				      prefix.c_str(),
				      numb_htoa(dstHop_.node));
			s4rid = true;
		} else {
			G2PCERA_DEBUG("%s Searching route to TE-link %s",
				      prefix.c_str(),
				      addr_ntoa(dstHop_.te_link));
			s4tel = true;
		}
	} else if (isValid(eTnaRes_, err) && !(s4rid ^ s4tel)) {
		G2PCERA_DEBUG("%s Searching route to a TNA %s",
			      prefix.c_str(),
			      addr_ntoa(eTnaRes_.tna));
		s4tna = true;
	}

	if (!s4tna && !s4rid && !s4tel) {
		G2PCERA_WARNING("%s Invalid or unspecified "
				"destination options",
				prefix.c_str());
		dump("", false);
		METOUT();
		return 0;
	}

	prefix = "[Dijkstra step 1]";
	rev   = parent()->getTopology().spfRevision() + 1;
	parent()->getTopology().spfRevision(rev);

	rootNode->spfRevision(rev);
	rootNode->rootCost(0);
	addNodeSorted(candidateNodes, rootNode);
	flags = rootNode->nodeFlags();
	flags.inCandidateList = true;
	flags.inTree = false;
	rootNode->nodeFlags(flags);

	G2PCERA_DEBUG(DUMP_BAR2 DUMP_BAR2);
	G2PCERA_DEBUG("%s     +-- root node    --> %s",
		      prefix.c_str(),
		      nodeIdent2string(rootNode->ident()).c_str());
	G2PCERA_DEBUG("%s     +-- revision     --> %d",
		      prefix.c_str(),
		      rootNode->spfRevision());
	G2PCERA_DEBUG("%s     +-- flags        --> %s",
		      prefix.c_str(),
		      nodeFlags2string(rootNode->nodeFlags()).
		      c_str());

	while (1) {
		prefix = "[Dijkstra step 2]";
		G2PCERA_DEBUG(DUMP_BAR2 DUMP_BAR2);
		G2PCERA_DEBUG("%s Candidate nodes are %d",
			      prefix.c_str(), candidateNodes.size());

		if (candidateNodes.empty()) {
			G2PCERA_DEBUG("%s Algorithm ends !",
				      prefix.c_str());
			G2PCERA_DEBUG(DUMP_BAR1 DUMP_BAR1);
			METOUT();
			return 0;
		}

		currentNode = candidateNodes.front();
		assert(currentNode);
		candidateNodes.pop_front();

		std::string nname;
		nname = nodeIdent2string(currentNode->ident());

		flags = currentNode->nodeFlags();
		flags.inCandidateList = false;
		flags.inTree = true;

		currentNode->nodeFlags(flags);
		currentNode->spfRevision(rev);

		G2PCERA_DEBUG("%s     +-- current node --> %s",
			      prefix.c_str(), nname.c_str());
		G2PCERA_DEBUG("%s     +-- revision     --> %d",
			      prefix.c_str(),
			      currentNode->spfRevision());
		G2PCERA_DEBUG("%s     +-- flags        --> %s",
			      prefix.c_str(),
			      nodeFlags2string(currentNode->nodeFlags()).
			      c_str());


		if (s4rid && (currentNode->ident().id == dstHop_.node)) {
			G2PCERA_DEBUG("%s Reached the destination node",
				      prefix.c_str());
			G2PCERA_DEBUG(DUMP_BAR1 DUMP_BAR1);
			METOUT();
			return currentNode;
		}

		if (s4tna) {
			Tna * eTna;
			eTna = currentNode->getTna(eTnaRes_.tna);
			if (eTna) {
				eRid = eTna->ident().rid;
				G2PCERA_DEBUG("%s Reached the "
					      "destination TNA %s",
					      prefix.c_str(),
					      addr_ntoa(eTnaRes_.tna));
				G2PCERA_DEBUG(DUMP_BAR1 DUMP_BAR1);
				METOUT();
				return currentNode;
			}
		}

		currentLinks = currentNode->getOutLinks();

		std::list<TeLink *>::iterator iter;

		for (iter  = currentLinks.begin();
		     iter != currentLinks.end();
		     iter++) {
			assert(*iter);

			currentLink = (*iter);
			nextNode    = currentLink->getRemNode();
			reverseLink = currentLink->getRemLink();

			std::string lname;
			lname = teLinkIdent2string(currentLink->ident());

			G2PCERA_DEBUG("%s        +-- current link --> %s",
				      prefix.c_str(), lname.c_str());

			if (!nextNode || !reverseLink) {
				G2PCERA_DEBUG("%s Skipping uncorrelated "
					      "link %s",
					      prefix.c_str(), lname.c_str());
				continue;
			}

			if ((currentNode == rootNode) &&
			    addr_equal(currentLink->ident().lcl_id,
				       srcHop_.te_link) != 0) {
				G2PCERA_DEBUG("%s Skipping the specified "
					      "srcHop link %s: ",
					      prefix.c_str(), lname.c_str());
				continue;
			}

			if (nextNode == previousNode) {
				G2PCERA_DEBUG("%s Possible loop condition, "
					      "skipping link %s",
					      prefix.c_str(), lname.c_str());
				continue;
			}


			if (!currentLink->fitInConstraints(lspInfo_)) {
				G2PCERA_DEBUG("%s No fit in TE constraints, "
					      "skipping link %s",
					      prefix.c_str(), lname.c_str());
				continue;

			}

			if (!reverseLink->fitInConstraints(lspInfo_)) {
				G2PCERA_DEBUG("%s No reverse fit in "
					      "TE constraints, "
					      "skipping link %s",
					      prefix.c_str(), lname.c_str());
				continue;

			}

			int64_t cost;

			cost = currentLink->linkCost() ?
				currentLink->linkCost() :
				currentLink->generateCost(lspInfo_);
			currentLink->linkCost(cost);

			cost = reverseLink->linkCost() ?
				reverseLink->linkCost():
				reverseLink->generateCost(lspInfo_);
			reverseLink->linkCost(cost);

			if (cost > 0                     &&
			    cost < G2PCERA_MAX_LINK_COST &&
			    (cost >
			     (int64_t) parent()->getTopology().maxLinkCost())){
				parent()->getTopology().
					maxLinkCost((uint32_t)cost);
			}

			dist = (uint32_t)((int64_t)currentNode->rootCost() +
					  cost);

			G2PCERA_DEBUG("%s           +-- cost --> %d",
				      prefix.c_str(), currentLink->linkCost());
			G2PCERA_DEBUG("%s           +-- dist --> %d",
				      prefix.c_str(), dist);

			prefix = "[Dijkstra step 3]";
			G2PCERA_DEBUG(DUMP_BAR2 DUMP_BAR2);

			std::string nnname;
			nnname = nodeIdent2string(nextNode->ident());

			flags = nextNode->nodeFlags();

			if (nextNode->spfRevision() != rev) {
				nextNode->spfRevision(rev);
				nextNode->rootCost(dist);
				addNodeSorted(candidateNodes, nextNode);
				flags.inCandidateList = true;
				nextNode->nodeFlags(flags);

				G2PCERA_DEBUG("%s Adding node %s to the "
					      "candidate list and "
					      "updating its fwd_links list",
					      prefix.c_str(), nnname.c_str());

				nextNode->clearFwdLinks();
				nextNode->attachFwdLink(currentLink);

				if (s4tel &&
				    (addr_equal(reverseLink->ident().lcl_id,
						dstHop_.te_link) != 0)) {
					// Computation ended
					G2PCERA_DEBUG("%s Reached the dst "
						      "TE LINK",
						      prefix.c_str());
					break;
				}
			} else  if (flags.inCandidateList) {

				if (dist < nextNode->rootCost()) {
					// modify its position in the
					// candidate list,
					G2PCERA_DEBUG("%s Relaxation on "
						      "node %s",
						      prefix.c_str(),
						      nnname.c_str());

					nextNode->clearFwdLinks();
					nextNode->rootCost(dist);
					nextNode->attachFwdLink(currentLink);

					if (s4tel &&
					    (addr_equal(reverseLink->
							ident().lcl_id,
							dstHop_.te_link)!= 0)){
						// Computation ended
						G2PCERA_DEBUG("%s Reached "
							      "the dst "
							      "TE LINK",
							      prefix.c_str());
						G2PCERA_DEBUG(DUMP_BAR1
							      DUMP_BAR1);
						METOUT();
						return currentNode;
					}

					moveNodeSorted(candidateNodes,
						       nextNode);
				} else {
					if (s4tel &&
					    (addr_equal(reverseLink->
							ident().lcl_id,
							dstHop_.te_link)!= 0)){
						// Computation ended
						G2PCERA_DEBUG("%s Reached "
							      "the dst "
							      "TE LINK",
							      prefix.c_str());
						nextNode->clearFwdLinks();
						nextNode->rootCost(dist);
						nextNode->attachFwdLink(currentLink);
						G2PCERA_DEBUG(DUMP_BAR1
							      DUMP_BAR1);
						METOUT();
						return currentNode;
					}

				}
			}
		}
		previousNode = currentNode;
	}

	G2PCERA_DEBUG(DUMP_BAR1 DUMP_BAR1);
	METOUT();
	return 0;
}

bool
PCERA::GnsCall::assembleEro(PCERA::NetNode *		dst,
			    uint32_t			eRid,
			    std::list<ero_hop_t> &	ero)
{
	METIN();

	ero.clear();

	if (!dst) {
		G2PCERA_DEBUG("Cannot assemble ERO. Null destination node!");
		METOUT();
		return false;
	}

	NetNode *	currentNode;
	node_ident_t	nid;

	currentNode = dst;
	nid.id      = parent()->rootNodeId();
	nid.type    = TOPO_NODE_NETWORK;

	ero_hop_t   hop;

	if (eRid && dst->ident().id != eRid) {
		memset(&hop, 0, sizeof(hop));
		hop.idmn = true;
		hop.node = eRid;
		ero.push_front(hop);
	}

	while (1) {
		if (currentNode->ident() == nid) {
			G2PCERA_DEBUG("ERO assembling complete from "
				      "src node %s to dst node %s!",
				      nodeIdent2string(nid).c_str(),
				      nodeIdent2string(dst->ident()).c_str());
			METOUT();
			return true;
		}

		std::list<TeLink *>	fwdLinks;
		TeLink *		currentLink;

		fwdLinks = currentNode->getFwdLinks();
		currentLink = fwdLinks.front();

		ero_hop_t   hop;

		memset(&hop, 0, sizeof(hop));

		hop.idmn    = currentNode->isDomain() ? 1 : 0;
		hop.node    = currentLink->ident().rem_rid;
		hop.te_link = currentLink->ident().rem_id;

		std::list<isc_t> iscs;
		if (currentLink->getIscs(iscs)) {
			std::list<isc_t>::const_iterator iter;

			for (iter = iscs.begin(); iter != iscs.end(); iter++) {
				if ((*iter).switching == SWCAP_LSC) {

					label_id_t  label;
					memset(&label, 0, sizeof(label));

					if (BITMASK_BITTEST(src_.mask_,
							    label)) {
						label = src_.label;
					} else {
						label.type = LABEL_32BIT;
						label.value.label32.id =
							0x28000019;
					}

					//hop.upstream_label   = label;
					hop.downstream_label = label;
				}
			}
		}

		ero.push_front(hop);
		currentNode = currentLink->parent();
	}

	ero.clear();
	G2PCERA_DEBUG("An error occurrend while assembling ERO!");
	METOUT();
	return false;
}

bool
PCERA::GnsCall::addEroHopToConnection(lsp_role_t	role,
				      const ero_hop_t &	hop,
				      bool		onTop)
{
	METIN();

	switch (role) {
		case LSP_ROLE_PRIMARY: {
			primary_.addEroHop(onTop, hop);
		}
			break;
		case LSP_ROLE_SECONDARY: {
			secondary_.addEroHop(onTop, hop);
		}
			break;
		default:
			BUG();
			assert(0);
	}

	METOUT();
	return true;
}

bool
PCERA::GnsCall::delEroHopFromConnection(lsp_role_t	role,
					ero_hop_t &	hop)
{
	METIN();

	switch (role) {
		case LSP_ROLE_PRIMARY: {
			primary_.delEroHop(hop);
		}
			break;
		case LSP_ROLE_SECONDARY: {
			secondary_.delEroHop(hop);
		}
			break;
		default:
			BUG();
			assert(0);
	}

	METOUT();
	return true;
}


std::list<ero_hop_t>
PCERA::GnsCall::getConnectionEro(lsp_role_t role,
				 bool       justMyDmn)
{
	METIN();

	std::list<ero_hop_t>::iterator iter;
	std::list<ero_hop_t> tmp, res;

	tmp.clear();
	res.clear();

	switch (role) {
		case LSP_ROLE_PRIMARY: {
			tmp = primary_.getEro();
		}
			break;
		case LSP_ROLE_SECONDARY: {
			tmp = secondary_.getEro();
		}
			break;
		default:
			BUG();
			assert(0);
	}


	for (iter  = tmp.begin(); iter != tmp.end(); iter++) {
		if (justMyDmn && (*iter).idmn) {
			// stop at the first domain transition
			break;
		}
		res.push_back(*iter);
	}


	METOUT();
	return res;
}

std::list<ero_hop_t>
PCERA::GnsCall::getGnsCallEro(void)
{
	METIN();

	std::list<ero_hop_t> tmp;

	tmp.clear();

	std::list<ero_hop_t>::const_iterator iter;
	for (iter = gnsCallEro_.begin(); iter != gnsCallEro_.end(); iter++) {
		tmp.push_back(*iter);
	}

	METOUT();
	return tmp;
}

net_res_spec_t
PCERA::GnsCall::getGnsCallEgressTnaRes(void)
{
	METIN();

	METOUT();
	return eTnaRes_;
}


disjointness_level_t
PCERA::GnsCall::getDisjoinnessLevel(void)
{
	METIN();
	METOUT();

	return disjoinnessLevel_;
}

bool
PCERA::GnsCall::confirm(void)
{
	METIN();
	METOUT();

	return isConfirmed_;
}

void
PCERA::GnsCall::confirm(bool flag)
{
	METIN();

	isConfirmed_ = flag;

	METOUT();
}


bool
PCERA::GnsCall::match(const net_res_spec_t &	src,
		      const res_spec_t &	dst,
		      const call_info_t &	cInfo,
		      const recovery_info_t &	rInfo,
		      const lsp_info_t &	lInfo)
{
	METIN();

	if (src != src_) {
		G2PCERA_DEBUG("GnsCall sources are different");
		logDump("[GnsCall::src_]", src_);
		logDump("[other::src]", src);
		METOUT();
		return false;
	}

	if (dst != dst_) {
		G2PCERA_DEBUG("GnsCall destinations are different");
		logDump("[GnsCall::dst_]", dst_);
		logDump("[other::dst]", dst);
		METOUT();
		return false;
	}

	if (cInfo != callInfo_) {
		G2PCERA_DEBUG("GnsCall call info are different");
		logDump("[GnsCall::callInfo_]", callInfo_);
		logDump("[other::cInfo]", cInfo);
		METOUT();
		return false;
	}

	if (rInfo != recInfo_) {
		G2PCERA_DEBUG("GnsCall recovery info are different");
		logDump("[GnsCall::recInfo_]", recInfo_);
		logDump("[other::rInfo]", rInfo);
		METOUT();
		return false;
	}

	if (lInfo != lspInfo_) {
		G2PCERA_DEBUG("GnsCall lspInfo are different");
		logDump("[GnsCall::lspInfo_]", lspInfo_);
		logDump("[other::lInfo]", lInfo);
		METOUT();
		return false;
	}

	METOUT();
	return true;
}

bool
PCERA::GnsCall::match(const ero_hop_t &		srcHop,
		      const ero_hop_t &		dstHop,
		      const lsp_info_t &	lInfo)
{
	METIN();

	if (!srcHop.is_null() && (srcHop != srcHop_)) {
		G2PCERA_DEBUG("GnsCall source hops are different");
		logDump("[GnsCall::srcHop_]", srcHop_);
		logDump("[other::srcHop]", srcHop);
		METOUT();
		return false;
	}

	if (!dstHop.is_null() && (dstHop != dstHop_)) {
		G2PCERA_DEBUG("GnsCall destination hops are different");
		logDump("[GnsCall::dstHop_]", dstHop_);
		logDump("[other::dstHop]", dstHop);
		METOUT();
		return false;
	}

	if (lInfo != lspInfo_) {
		G2PCERA_DEBUG("GnsCall lspInfo are different");
		logDump("[GnsCall::lspInfo_]", lspInfo_);
		logDump("[other::lInfo]", lInfo);
		METOUT();
		return false;
	}

	METOUT();
	return true;
}

////////////////////////////
// Class Connection
////////////////////////////
PCERA::Connection::Connection(void)
{
	METIN();

	ero_.clear();

	METOUT();
}


PCERA::Connection::~Connection(void)
{
	METIN();

	ero_.clear();

	METOUT();
}

void
PCERA::Connection::dump(const std::string &	prefix,
			bool			recursive)
{
	METIN();

	G2PCERA_DEBUG("%sConnection",  prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	std::list<ero_hop_t>::const_iterator iter;

	G2PCERA_DEBUG("%s- ERO Hops",
		      prefix0.c_str());
	for (iter = ero_.begin(); iter != ero_.end(); iter++) {
		logDump(prefix1, *iter);
		G2PCERA_DEBUG("%s-------------------------------",
			      prefix1.c_str());
	}

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

bool
PCERA::Connection::addEroHop(bool		onTop,
			     const ero_hop_t &	hop)
{
	METIN();

	std::list<ero_hop_t>::const_iterator iter;
	for (iter = ero_.begin(); iter != ero_.end(); iter++) {
		if (*iter == hop) {
			G2PCERA_ERROR("Duplicated ero hop to add");
			METOUT();
			return false;
		}
	}

	if (onTop) {
		ero_.push_front(hop);
	} else {
		ero_.push_back(hop);
	}

	METOUT();
	return true;
}

bool
PCERA::Connection::delEroHop(ero_hop_t &	hop)
{
	METIN();

	std::list<ero_hop_t>::iterator iter;
	for (iter = ero_.begin(); iter != ero_.end(); iter++) {
		if (*iter == hop) {
			ero_.erase(iter);
			METOUT();
			return true;
		}
	}

	G2PCERA_ERROR("hop to delete not found in ero");

	METOUT();
	return false;
}

void
PCERA::Connection::flushEro(void)
{
	METIN();

	ero_.clear();

	METOUT();
}

std::list<ero_hop_t>
PCERA::Connection::getEro(void)
{
	METIN();

	std::list<ero_hop_t> tmp;

	tmp.clear();

	std::list<ero_hop_t>::const_iterator iter;
	for (iter = ero_.begin(); iter != ero_.end(); iter++) {
		tmp.push_back(*iter);
	}

	METOUT();
	return tmp;
}

bool
PCERA::Connection::checkDisjoinness(Connection &		other,
				    disjointness_t &		type,
				    disjointness_level_t &	level)
{
	METIN();

	type  = DISJOINTNESS_NONE;
	level = DISJ_LEVEL_UNKNOWN;

	std::list<ero_hop_t> cmpEro;
	cmpEro = other.getEro();

	if (ero_.empty() || cmpEro.empty()) {
		G2PCERA_ERROR("One of the two EROs t compare is empty");
		METOUT();
		return false;
	}

	// XXX Add code here

	METOUT();
	return true;
}



////////////////////////////
// nodeKey DM utils
////////////////////////////
std::string
PCERA::nodeKey2string(const nodeKey_t & key)
{
	METIN();
	std::string        tmp;
	std::ostringstream localId;

	tmp = std::string("");

	localId << key << "(" << numb_htoa(key) << ")";

	tmp += localId.str();

	METOUT();
	return tmp;
}

PCERA::nodeKey_t
PCERA::nodeKeyMangle(const uint32_t & rid)
{
	METIN();

	nodeKey_t tmp;

	tmp = rid;

	METOUT();
	return tmp;
}

uint32_t
PCERA::nodeKeyDemangle(const nodeKey_t & key)
{
	METIN();

	uint32_t tmp;

	tmp = key;

	METOUT();
	return tmp;
}


////////////////////////////
// Class Topology
////////////////////////////
PCERA::Topology::Topology(G2PCERA * parent) :
	Ancestor<true, G2PCERA>(parent)
{
	METIN();

	totNodes_    = 0;
	totLinks_    = 0;
	totTnas_     = 0;
	maxLinkCost_ = 0;
	spfRevision_ = 0;

	nodes_.clear();

	METOUT();
}

PCERA::Topology::~Topology(void)
{
	METIN();

	if (nodes_.size()) {
		G2PCERA_ERROR("There are %d nodes on the way",
			      nodes_.size());
		BUG();
		std::map<nodeKey_t, Node *>::iterator	iter;
		for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
		METOUT();
		return;
	}
	nodes_.clear();

	// a redundant consistency check
	assert(nodes_.size() == 0);

	METOUT();
}


bool
PCERA::Topology::attachNode(Node *  ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed node is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;

	key = nodeKeyMangle(ptr->ident().id);

	if (nodes_.find(key) != nodes_.end()) {
		G2PCERA_ERROR("An instance of node with the same id %s is "
			      "just present in topology... "
			      "detach it first",
			      nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (ptr->type() == TOPO_NODE_NETWORK) {

		NetNode * tmp;
		tmp = static_cast<NetNode *>(ptr);

		assert(tmp);

		if (!modTotLinksCount(tmp->getOutLinksCount(),
				      true /* add */)) {
			G2PCERA_ERROR("Unrecoverable problems on links count "
				      "in topology... "
				      "cannot attach node with id %s",
				      nodeKey2string(key).c_str());
			METOUT();
			return false;
		}

		if (!modTotTnasCount(tmp->getTnasCount(),
				     true /* add */)) {
			G2PCERA_ERROR("Unrecoverable problems on TNAs count "
				      "in topology... "
				      "cannot attach node with id %s",
				      nodeKey2string(key).c_str());
			METOUT();
			return false;
		}
	}

	if (!modTotNodesCount(1, true /* add */)) {
		G2PCERA_ERROR("Unrecoverable problems on node count "
			      "in topology... "
			      "cannot attach node with id %s",
			      nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	nodes_[key] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::Topology::detachNode(Node *  ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed node is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;

	key = nodeKeyMangle(ptr->ident().id);

	if (nodes_.find(key) == nodes_.end()) {
		G2PCERA_ERROR("No instance of node with id %s "
			      "is present in topology...",
			      nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (nodes_[key] != ptr) {
		G2PCERA_ERROR("A different instance of node with id %s "
			      "is present in topology...cannot detach",
			      nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (ptr->type() == TOPO_NODE_NETWORK) {

		NetNode * tmp;
		tmp = static_cast<NetNode *>(ptr);

		assert(tmp);

		if (!modTotLinksCount(tmp->getOutLinksCount(),
				      false /* del */)) {
			G2PCERA_ERROR("Unrecoverable problems on links count "
				      "in topology... "
				      "cannot detach node with id %s",
				      nodeKey2string(key).c_str());
			METOUT();
			return false;
		}

		if (!modTotTnasCount(tmp->getTnasCount(),
				     false /* del */)) {
			G2PCERA_ERROR("Unrecoverable problems on TNAs count "
				      "in topology... "
				      "cannot detach node with id %s",
				      nodeKey2string(key).c_str());
			METOUT();
			return false;
		}
	} else if (ptr->type() == TOPO_NODE_GRID) {
		uint32_t		peRouterId;
		grid_site_data_t	data;
		GridNode *		tmp1;

		tmp1 = static_cast<GridNode *>(ptr);

		assert(tmp1);

		if (!tmp1->getData(data)) {
			G2PCERA_ERROR("Unrecoverable problems on Grid Node "
				      "in topology... "
				      "cannot detach node with id %s",
				      nodeKey2string(key).c_str());
			METOUT();
			return false;
		}

		peRouterId = data.pe_router_id;

		// uncorrelate pe_router_id's TNAs, if any
		Node *			tmp;
		PCERA::NetNode *	lclNode;
		node_ident_t		nodeIdent;

		nodeIdent.id   = peRouterId;
		nodeIdent.type = TOPO_NODE_NETWORK;

		tmp = getNode(nodeIdent);
		lclNode = static_cast<PCERA::NetNode *>(tmp);

		if (lclNode) {
			lclNode->correlateAllTnas(peRouterId, 0);
		} else {
			// peRouter could be under a domain, get TNAs directly
			std::list<node_ident_t>			nodeList;
			std::list<node_ident_t>::iterator	iter;

			nodeList = getNodesIdents();

			for (iter  = nodeList.begin();
			     iter != nodeList.end();
			     iter++) {
				node_ident_t	id;

				id = (*iter);

				if (id.type != TOPO_NODE_NETWORK) {
					continue;
				}

				tmp = getNode(id);
				lclNode = static_cast<PCERA::NetNode *>(tmp);

				if (lclNode && lclNode->isDomain()) {
					lclNode->correlateAllTnas(peRouterId,
								  0);
				}
			}
		}

	}

	if (!modTotNodesCount(1, false /* del */)) {
		G2PCERA_ERROR("Unrecoverable problems on node count "
			      "in topology... "
			      "cannot detach node with id %s",
			      nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	nodes_.erase(key);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::Node *
PCERA::Topology::getNode(node_ident_t	ident)
{
	METIN();

	nodeKey_t					key;
	Node *						tmp;
	std::map<nodeKey_t, Node *>::const_iterator	iter;

	key = nodeKeyMangle(ident.id);
	tmp  = 0;
	iter = nodes_.find(key);

	if (iter != nodes_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}

PCERA::GridNode *
PCERA::Topology::getGridNode(uint32_t	 peRouterId)
{
	METIN();

	GridNode *					res;
	std::map<nodeKey_t, Node *>::const_iterator	iter;

	res  = 0;

	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		assert((*iter).second);
		Node *      item;
		GridNode *  tmp;

		item = (*iter).second;

		if (item->type() != TOPO_NODE_GRID) {
			continue;
		}

		tmp = static_cast<GridNode *>(item);

		if (tmp->getPeRouterId() == peRouterId) {
			res = tmp;
			break;
		}
	}

	METOUT();
	return res;
}

PCERA::NetNode *
PCERA::Topology::getAncestorNode(uint32_t routerId)
{
	METIN();

	NetNode *					res;
	std::map<nodeKey_t, Node *>::const_iterator	iter;

	res  = 0;

	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		assert((*iter).second);
		Node *     item;
		NetNode *  tmp;

		item = (*iter).second;

		if (item->type() != TOPO_NODE_NETWORK) {
			continue;
		}

		tmp = static_cast<NetNode *>(item);

		if (!tmp->isDomain()) {
			continue;
		}

		std::list<telink_ident_t>                 tels;
		std::list<telink_ident_t>::const_iterator telsIter;

		tels = tmp->getOutLinksIdents();
		for (telsIter  = tels.begin();
		     telsIter != tels.end();
		     telsIter++) {
			telink_ident_t id;

			id = *telsIter;

			if (id.lcl_rid == routerId) {
				res = tmp;
				METOUT();
				return res;
			}
		}
	}

	METOUT();
	return res;
}

std::list<PCERA::Node *>
PCERA::Topology::getNodes(void)
{
	METIN();

	std::list<Node *>				tmp;
	std::map<nodeKey_t, Node *>::const_iterator	iter;

	tmp.clear();

	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		assert((*iter).second);
		tmp.push_front((*iter).second);
	}

	METOUT();
	return tmp;
}

std::list<node_ident_t>
PCERA::Topology::getNodesIdents(void)
{
	METIN();

	std::list<node_ident_t>				tmp;
	std::map<nodeKey_t, Node *>::const_iterator	iter;

	tmp.clear();

	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		node_ident_t	id;

		assert((*iter).second);

		id = ((*iter).second)->ident();

		tmp.push_front(id);
	}

	METOUT();
	return tmp;
}

bool
PCERA::Topology::modTotNodesCount(uint32_t	howMany,
				  bool		add)
{
	METIN();

	int64_t newVal;

	newVal = totNodes_ + (add ? (+1) : (-1)) * howMany;

	if (newVal  < 0) {
		G2PCERA_ERROR("Underflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	if (add && (uint32_t)newVal  < totNodes_) {
		G2PCERA_ERROR("Overflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	totNodes_ = (uint32_t)newVal;

	METOUT();
	return true;
}

bool
PCERA::Topology::modTotLinksCount(uint32_t	howMany,
				  bool		add)
{
	METIN();

	int64_t newVal;

	newVal = totLinks_ + (add ? (+1) : (-1)) * howMany;

	if (newVal  < 0) {
		G2PCERA_ERROR("Underflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	if (add && (uint32_t)newVal  < totLinks_) {
		G2PCERA_ERROR("Overflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	totLinks_ = (uint32_t)newVal;

	METOUT();
	return true;
}

bool
PCERA::Topology::modTotTnasCount(uint32_t	howMany,
				 bool		add)
{
	METIN();

	int64_t newVal;

	newVal = totTnas_ + (add ? (+1) : (-1)) * howMany;

	if (newVal  < 0) {
		G2PCERA_ERROR("Underflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	if (add && (uint32_t)newVal  < totTnas_) {
		G2PCERA_ERROR("Overflow problem on %s...",
			      __PRETTY_FUNCTION__);
		METOUT();
		return false;
	}

	totTnas_ = (uint32_t)newVal;

	METOUT();
	return true;
}

void
PCERA::Topology::dump(const std::string &	prefix,
		      bool			recursive)
{
	METIN();

	G2PCERA_DEBUG("%sTopology", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	topology_summary_data_t data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0, data);

	std::map<nodeKey_t, Node *>::const_iterator iter;

	G2PCERA_DEBUG("%s- Topology Nodes",
		      prefix0.c_str());
	G2PCERA_DEBUG("");

	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		Node * ptr;
		topo_node_type_t nt;

		ptr = (*iter).second;
		assert(ptr);
		nt = ptr->type();

		switch (nt) {
			case TOPO_NODE_UNKNOWN: {
				ptr->dump(prefix1);
			}
				break;
			case TOPO_NODE_NETWORK: {
				NetNode * tmp;
				tmp = static_cast<NetNode *>(ptr);

				assert(tmp);
				tmp->dump(prefix1, recursive);
			}
				break;
			case TOPO_NODE_GRID: {
				GridNode * tmp;
				tmp = static_cast<GridNode *>(ptr);

				assert(tmp);
				tmp->dump(prefix1, recursive);
			}
				break;
			default:
				G2PCERA_ERROR("Unrecognized node type (%d)...",
					      nt);
				assert(0);
				break;
		}
	}

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

bool
PCERA::Topology::getData(topology_summary_data_t &	data)
{
	METIN();

	memset(&data, 0, sizeof(data));

	data.root_rid       = parent()->rootNodeId();
	data.tot_nodes      = totNodes_;
	data.tot_links      = totLinks_;
	data.tot_tnas       = totTnas_;
	data.tot_gns_calls  = parent()->getGnsCallsCount();
	data.max_link_cost  = maxLinkCost_;
	data.spf_selector   = parent()->spfSelector();
	data.spf_revision   = spfRevision_;

	METOUT();
	return true;
}

uint32_t
PCERA::Topology::spfRevision(void)
{
	METIN();
	METOUT();
	return spfRevision_;
}

void
PCERA::Topology::spfRevision(uint32_t  newRev)
{
	METIN();

	spfRevision_ = newRev;

	METOUT();
}

uint32_t
PCERA::Topology::maxLinkCost(void)
{
	METIN();
	METOUT();
	return maxLinkCost_;
}

void
PCERA::Topology::maxLinkCost(uint32_t newCost)
{
	METIN();

	maxLinkCost_ = newCost;

	METOUT();
}

void
PCERA::Topology::accept(topologyVisitor &	vis)
{
	METIN();

	std::map<nodeKey_t, Node *>::const_iterator	iter;
	for (iter = nodes_.begin(); iter != nodes_.end(); iter++) {
		Node * ptr;
		topo_node_type_t nt;

		ptr = (*iter).second;
		assert(ptr);
		nt = ptr->type();

		switch (nt) {
			case TOPO_NODE_UNKNOWN: {
				ptr->accept(vis);
			}
				break;
			case TOPO_NODE_NETWORK: {
				NetNode * tmp;
				tmp = static_cast<NetNode *>(ptr);

				assert(tmp);
				tmp->accept(vis);
			}
				break;
			case TOPO_NODE_GRID: {
				GridNode * tmp;
				tmp = static_cast<GridNode *>(ptr);

				assert(tmp);
				tmp->accept(vis);
			}
				break;
			default:
				G2PCERA_ERROR("Unrecognized node type (%d)...",
					      nt);
				assert(0);
				break;
		}
	}

	vis.visitTopology(this);

	METOUT();
}

bool
PCERA::Topology::transform(const std::list< std::list<ero_hop_t> > & multiXro,
			   bool forwardOrReverse)
{
	METOUT();

	std::list< std::list<ero_hop_t> >::const_iterator iter;
	std::string err;
	size_t i;

	for (i=0, iter=multiXro.begin(); iter!=multiXro.end(); iter++, i++) {
		G2PCERA_DEBUG("Transforming topology for XRO component %d", i);

		if (!transform(*iter, forwardOrReverse, err)) {
			G2PCERA_ERROR("multiXro elem #%d is not valid (%s)",
				      i, err.c_str());
			METOUT();
			return false;
		}
	}

	METOUT();
	return true;
}

bool
PCERA::Topology::transform(const std::list<ero_hop_t> &	xro,
			   bool				forwardOrReverse,
			   std::string &		err)
{
	METOUT();

	if (xro.empty()) {
		G2PCERA_DEBUG("No XRO to process");
		METOUT();
		return true;
	}

	std::string prefix0 = "[PCERA::Topology::transform] "DUMP_TAB;
	std::string prefix1 = prefix0 + DUMP_TAB;

	G2PCERA_DEBUG("%s- Processing XRO hops ...",
		      prefix0.c_str());

	std::list<ero_hop_t>::const_iterator iter;
	for (iter = xro.begin(); iter != xro.end(); iter++) {

		if (!isValid(*iter, err)) {
			err = "Invalid XRO hop to process (" + err + ")";
			METOUT();
			return false;
		}

		logDump(prefix1, *iter);
		G2PCERA_DEBUG("%s-------------------------------",
			      prefix1.c_str());

		PCERA::Node * tmp;
		PCERA::NetNode * lclNode;
		PCERA::NetNode * remNode;
		node_ident_t nodeIdent;
		PCERA::TeLink * lclLink;
		PCERA::TeLink * remLink;
		telink_ident_t telIdent;

		memset(&nodeIdent, 0, sizeof(nodeIdent));
		memset(&telIdent, 0, sizeof(telIdent));

		nodeIdent.id   = (*iter).node;
		nodeIdent.type = TOPO_NODE_NETWORK;

		tmp = getNode(nodeIdent);
		lclNode = static_cast<PCERA::NetNode *>(tmp);

		if (!lclNode) {
			err  = "Cannot get local node ";
			err += nodeIdent2string(nodeIdent);
			err += " described in XRO item";
			METOUT();
			return false;
		}


		lclLink = lclNode->getOutLink((*iter).te_link);
		if (!lclLink) {
			err  = "TELink ";
			err += addr_ntoa((*iter).te_link);
			err += " not present in node ";
			err += nodeIdent2string(nodeIdent);
			METOUT();
			return false;
		}
		telIdent = lclLink->ident();

		G2PCERA_DEBUG("%s Found TELINK %s by inspecting XRO",
			      prefix0.c_str(),
			      teLinkIdent2string(telIdent).c_str());


		nodeFlags_t flags;
		int64_t     oldCost, newCost;

		flags = lclNode->nodeFlags();

		oldCost = lclLink->linkCost();
		if (forwardOrReverse) {
			flags.inTopoTransform = true;
			newCost = oldCost + G2PCERA_EXTRA_LINK_COST;
		} else {
			flags.inTopoTransform = false;
			newCost = oldCost % G2PCERA_EXTRA_LINK_COST;
		}
		lclNode->nodeFlags(flags);

		G2PCERA_DEBUG("%s Changing (%s) cost (%lld --> %lld) for "
			      "TELINK %s",
			      prefix0.c_str(),
			      forwardOrReverse ? "FWD" : "RWD",
			      oldCost, newCost,
			      teLinkIdent2string(telIdent).c_str());

		lclLink->linkCost(newCost);

		remNode = lclLink->getRemNode();
		remLink = lclLink->getRemLink();

		if (!remNode && !remLink) {
			continue;
		}

		flags = remNode->nodeFlags();

		if (!flags.inTopoTransform) {
			flags.inTopoTransform = true;
			remNode->nodeFlags(flags);
		}

		oldCost = remLink->linkCost();
		if (forwardOrReverse) {
			newCost = oldCost + G2PCERA_EXTRA_LINK_COST;
		} else {
			newCost = oldCost % G2PCERA_EXTRA_LINK_COST;
		}

		G2PCERA_DEBUG("%s Changing (%s) cost (%lld --> %lld) for "
			      "remote TELINK %s",
			      prefix0.c_str(),
			      forwardOrReverse ? "FWD" : "RWD",
			      oldCost, newCost,
			      teLinkIdent2string(remLink->ident()).c_str());

		remLink->linkCost(newCost);

	}

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix0.c_str());

	METOUT();
	return true;
}

////////////////////////////
// Class Node
///////////////////////////
PCERA::Node::Node(topo_node_type_t	type,
		  uint32_t		id) :
	Ancestor<true, Topology>(0)
{
	METIN();

	type_ = type;
	id_   = id;

	METOUT();
}

PCERA::Node::~Node(void)
{
	METIN();

	METOUT();
}

topo_node_type_t
PCERA::Node::type(void)
{
	METIN();
	METOUT();
	return type_;
}

node_ident_t
PCERA::Node::ident(void)
{
	METIN();

	node_ident_t tmp;

	tmp.type = type_;
	tmp.id   = id_;

	METOUT();

	return tmp;
}

void
PCERA::Node::dump(const std::string &	prefix)
{
	METIN();

	G2PCERA_DEBUG("%sNode", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- type                : %s",
		      prefix0.c_str(),
		      NODE_TYPE_TO_STRING(type_));

	G2PCERA_DEBUG("%s- id                  : %s",
		      prefix0.c_str(),
		      rid2string(id_).c_str());

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

void
PCERA::Node::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitNode(this);

	METOUT();
}


////////////////////////////
// Class NetNode
///////////////////////////
PCERA::NetNode::NetNode(uint32_t		id) :
	Node(TOPO_NODE_NETWORK, id)
{
	METIN();

	isDomain_ = false;
	opState_  = DOWN;
	admState_ = DISABLED;
	teColors_ = 0;

	areas_.clear();

	tnas_.clear();
	links_.clear();

	// Path Computation related members
	rootCost_    = 0;
	spfRevision_ = 0;
	memset(&nodeFlags_, 0, sizeof(nodeFlags_));
	candidateElems_.clear();
	fwdLinks_.clear();

	METOUT();
}

PCERA::NetNode::~NetNode(void)
{
	METIN();

	areas_.clear();

	if (tnas_.size()) {
		G2PCERA_ERROR("There are %d tnas on the way",
			      tnas_.size());
		BUG();
		std::list<Tna *>::iterator iter;
		for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
			delete (*iter);
		}
		METOUT();
		return;
	}
	tnas_.clear();

	if (links_.size()) {
		G2PCERA_ERROR("There are %d links on the way",
			      links_.size());
		BUG();
		std::list<TeLink *>::iterator iter;
		for (iter = links_.begin(); iter != links_.end(); iter++) {
			delete (*iter);
		}
		METOUT();
		return;
	}
	links_.clear();

	// fwdLinks_ and candidateElems_ are pointers to candidate
	// that will be removed elsewhere
	candidateElems_.clear();
	fwdLinks_.clear();

	// a redundant consistency check
	assert(tnas_.size() == 0);
	assert(links_.size() == 0);
	assert(candidateElems_.size() == 0);
	assert(fwdLinks_.size() == 0);

	METOUT();
}

void
PCERA::NetNode::dump(const std::string &	prefix,
		     bool			recursive)
{
	METIN();

	G2PCERA_DEBUG("%sNetNode", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0, ident());

	net_node_data_t data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}

	logDump(prefix0, data);

	G2PCERA_DEBUG("%s- TNAs count          : %d",
		      prefix0.c_str(),
		      tnas_.size());

	if (recursive) {
		std::list<Tna *>::const_iterator tnaIter;
		for (tnaIter  = tnas_.begin();
		     tnaIter != tnas_.end();
		     tnaIter++) {
			assert(*tnaIter);
			(*tnaIter)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s- Out links count     : %d",
		      prefix0.c_str(),
		      links_.size());

	if (recursive) {
		std::list<TeLink *>::const_iterator linkIter;
		for (linkIter  = links_.begin();
		     linkIter != links_.end();
		     linkIter++) {
			assert(*linkIter);
			(*linkIter)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s- NetNode Path Computation info",
		      prefix0.c_str());
	G2PCERA_DEBUG("%s- root cost        : %d",
		      prefix1.c_str(),
		      rootCost_);
	G2PCERA_DEBUG("%s- spf revision     : %d",
		      prefix1.c_str(),
		      spfRevision_);
	G2PCERA_DEBUG("%s- node flags       : 0x%02x",
		      prefix1.c_str(),
		      nodeFlags_);

	G2PCERA_DEBUG("%s- candidate nodes  : %d",
		      prefix1.c_str(),
		      candidateElems_.size());

	G2PCERA_DEBUG("%s- forwarding links : %d",
		      prefix1.c_str(),
		      fwdLinks_.size());

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

bool
PCERA::NetNode::setData(const net_node_data_t &	data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, is_domain)) {
		isDomain_ = data.is_domain;
	}
	if (BITMASK_BITTEST(data.mask_, op_state)) {
		opState_ = data.op_state;
	}
	if (BITMASK_BITTEST(data.mask_, adm_state)) {
		admState_ = data.adm_state;
	}
	if (BITMASK_BITTEST(data.mask_, te_colors)) {
		teColors_ = data.te_colors;
	}

	if (BITMASK_BITTEST(data.mask_, areas)) {
		std::list<uint16_t>::const_iterator iter;

		for (iter  = data.areas.begin();
		     iter != data.areas.end();
		     iter++) {
			// ignore duplications
			addQ(areas_, *iter, true);
		}
	}

	METOUT();
	return true;
}

bool
PCERA::NetNode::getData(net_node_data_t &	data)
{
	METIN();

	data.is_domain = isDomain_;
	data.op_state  = opState_;
	data.adm_state = admState_;
	data.te_colors = teColors_;

	BITMASK_BITSET(data.mask_, is_domain);
	BITMASK_BITSET(data.mask_, op_state);
	BITMASK_BITSET(data.mask_, adm_state);
	BITMASK_BITSET(data.mask_, te_colors);

	data.areas.clear();

	if (areas_.size()) {
		std::list<uint16_t>::const_iterator iter;

		for (iter  = areas_.begin();
		     iter != areas_.end();
		     iter++) {
			addQ(data.areas, *iter, true);
		}
		BITMASK_BITSET(data.mask_, areas);
	}

	METOUT();
	return true;
}


bool
PCERA::NetNode::attachOutLink(TeLink *	ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed TeLink is NIL");
		METOUT();
		return false;
	}

	ptr->reparent(this);

	telink_ident_t				linkIdent;
	uint32_t				rid;
	std::list<TeLink *>::const_iterator	iter;

	linkIdent = ptr->ident();
	rid       = ident().id;

	G2PCERA_DEBUG("Attaching TeLink %s to node %s",
		      teLinkIdent2string(linkIdent).c_str(),
		      rid2string(rid).c_str());

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() == linkIdent) {
			G2PCERA_ERROR("An instance of TeLink with the same id "
				      "%s is just attached under node %s... "
				      "detach it first",
				      teLinkIdent2string(linkIdent).c_str(),
				      rid2string(rid).c_str());
			METOUT();
			return false;
		}
	}

	// correlate teLink with its peering NetNode/reverseTELink
	if (!ptr->correlate()) {
		G2PCERA_ERROR("TE-Link correlation failed... "
			      "cannot attach TE-link %s to node %s",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (!parent()->modTotLinksCount(1, true /* add */)) {
		G2PCERA_ERROR("Unrecoverable problems on links count "
			      "in topology... "
			      "cannot attach TE-link %s to node %s",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	links_.push_front(ptr);

	if (ptr->getRemLink() &&
	    !ptr->getRemLink()->getRemLink()) {
		G2PCERA_WARNING("Recorrelating remote TE-Link %s "
				"after correlation of TE-Link %s",
				teLinkIdent2string(ptr->getRemLink()->ident()).
				c_str(),
				teLinkIdent2string(linkIdent).c_str());

		ptr->getRemLink()->correlate();
	}

	METOUT();
	return true;
}

bool
PCERA::NetNode::detachOutLink(TeLink *	ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed TeLink is NIL");
		METOUT();
		return false;
	}

	telink_ident_t			linkIdent;
	uint32_t			rid;
	std::list<TeLink *>::iterator	iter;

	linkIdent = ptr->ident();
	rid       = ident().id;

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() == linkIdent) {
			break;
		}
	}

	if (iter == links_.end()) {
		G2PCERA_ERROR("No instance of TeLink with id "
			      "%s found under node %s... ",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (*iter != ptr) {
		G2PCERA_ERROR("A different instance of TeLink with id "
			      "%s found under node %s... ",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	// uncorrelate teLink with its peering NetNode/reverseTELink
	if (!ptr->uncorrelate()) {
		G2PCERA_ERROR("TE-Link uncorrelation failed... "
			      "cannot detach TE-link %s from node %s",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (!parent()->modTotLinksCount(1, false /* del */)) {
		G2PCERA_ERROR("Unrecoverable problems on links count "
			      "in topology... "
			      "cannot detach TE-link %s deom node %s",
			      teLinkIdent2string(linkIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	links_.erase(iter);
	ptr->reparent(0);

	METOUT();
	return true;
}

std::list<PCERA::TeLink *>
PCERA::NetNode::getOutLinks(void)
{
	METIN();

	std::list<TeLink *>			tmp;
	std::list<TeLink *>::const_iterator	iter;

	tmp.clear();

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);
		tmp.push_front(*iter);
	}

	METOUT();
	return tmp;
}

std::list<telink_ident_t>
PCERA::NetNode::getOutLinksIdents(void)
{
	METIN();

	std::list<telink_ident_t>		tmp;
	std::list<TeLink *>::const_iterator	iter;

	tmp.clear();

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);
		tmp.push_front((*iter)->ident());
	}

	METOUT();
	return tmp;
}


size_t
PCERA::NetNode::getOutLinksCount(void)
{
	METIN();

	size_t tmp;

	tmp = links_.size();

	METOUT();
	return tmp;
}

PCERA::TeLink *
PCERA::NetNode::getOutLink(const telink_ident_t & id)
{
	METIN();

	std::list<TeLink *>::const_iterator	iter;
	TeLink *				tmp;

	tmp = 0;

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() == id) {
			tmp = *iter;
			break;
		}
	}

	METOUT();
	return tmp;
}

PCERA::TeLink *
PCERA::NetNode::getOutLink(const g2mpls_addr_t & id)
{
	METIN();

	std::list<TeLink *>::const_iterator	iter;
	TeLink *				tmp;

	tmp = 0;

	for (iter = links_.begin(); iter != links_.end(); iter++) {
		assert(*iter);

		if (addr_equal((*iter)->ident().lcl_id, id) != 0) {
			tmp = *iter;
			break;
		}
	}

	METOUT();
	return tmp;
}

bool
PCERA::NetNode::attachTna(Tna *	ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed TNA is NIL");
		METOUT();
		return false;
	}

	ptr->reparent(this);

	tna_ident_t				tnaIdent;
	uint32_t				rid;
	std::list<Tna *>::const_iterator	iter;

	tnaIdent = ptr->ident();
	rid      = ident().id;

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() == tnaIdent) {
			G2PCERA_ERROR("An instance of TNA with the same id "
				      "%s is just attached under node %s... "
				      "detach it first",
				      tnaIdent2string(tnaIdent).c_str(),
				      rid2string(rid).c_str());
			METOUT();
			return false;
		}
	}

	// correlate TNA with its peering GridNode
	if (!ptr->correlate()) {
		G2PCERA_ERROR("Cannot correlate TNA with its "
			      "peering GridNode... "
			      "cannot attach TNA %s to node %s",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (!parent()->modTotTnasCount(1, true /* add */)) {
		G2PCERA_ERROR("Unrecoverable problems on TNA count "
			      "in topology... "
			      "cannot attach TNA %s to node %s",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	tnas_.push_front(ptr);

	METOUT();
	return true;
}

bool
PCERA::NetNode::detachTna(Tna *	ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed TNA is NIL");
		METOUT();
		return false;
	}

	tna_ident_t			tnaIdent;
	uint32_t			rid;
	std::list<Tna *>::iterator	iter;

	tnaIdent = ptr->ident();
	rid      = ident().id;

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() == tnaIdent) {
			break;
		}
	}

	if (iter == tnas_.end()) {
		G2PCERA_ERROR("No instance of TNA with id "
			      "%s found under node %s... ",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (*iter != ptr) {
		G2PCERA_ERROR("A different instance of TNA with id "
			      "%s found under node %s... ",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	// uncorrelate TNA from its peering GridNode
	if (!ptr->uncorrelate()) {
		G2PCERA_ERROR("Cannot uncorrelate TNA with its "
			      "peering GridNode... "
			      "cannot detach TNA %s to node %s",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	if (!parent()->modTotTnasCount(1, false /* del */)) {
		G2PCERA_ERROR("Unrecoverable problems on TNA count "
			      "in topology... "
			      "cannot detach TNA %s to node %s",
			      tnaIdent2string(tnaIdent).c_str(),
			      rid2string(rid).c_str());
		METOUT();
		return false;
	}

	tnas_.erase(iter);
	ptr->reparent(0);

	METOUT();
	return true;
}

std::list<tna_ident_t>
PCERA::NetNode::getTnasIdents(uint32_t	borderNodeId)
{
	METIN();

	std::list<tna_ident_t>			tmp;
	std::list<Tna *>::const_iterator	iter;

	tmp.clear();

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		tna_ident_t  id;

		assert(*iter);
		id = (*iter)->ident();

		if (borderNodeId && (id.rid != borderNodeId)) {
			continue;
		}

		tmp.push_front(id);
	}

	METOUT();
	return tmp;
}

size_t
PCERA::NetNode::getTnasCount(void)
{
	METIN();

	size_t tmp;

	tmp = tnas_.size();

	METOUT();
	return tmp;
}

PCERA::Tna *
PCERA::NetNode::getTna(const tna_ident_t & id)
{
	METIN();

	std::list<Tna *>::const_iterator	iter;
	Tna *					tmp;

	tmp = 0;

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident() >= id) {
			tmp = *iter;
			break;
		}
	}

	METOUT();
	return tmp;
}

PCERA::Tna *
PCERA::NetNode::getTna(const g2mpls_addr_t & id)
{
	METIN();

	std::list<Tna *>::const_iterator	iter;
	Tna *					tmp;

	tmp = 0;

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		assert(*iter);

		if (addr_in_net((*iter)->ident().id, id) != 0) {
			tmp = *iter;
			break;
		}
	}

	METOUT();
	return tmp;
}

bool
PCERA::NetNode::correlateAllTnas(uint32_t		peRouterId,
				 PCERA::GridNode *	ptr)
{
	METIN();
	std::list<Tna *>		tmp;
	std::list<Tna *>::iterator	iter;

	for (iter = tnas_.begin(); iter != tnas_.end(); iter++) {
		assert(*iter);

		if ((*iter)->ident().rid == peRouterId) {
			(*iter)->setRemNode(ptr);
		}
	}

	METOUT();
	return true;
}

bool
PCERA::NetNode::attachCandElems(NetNode * ptr)
{
	METIN();

	assert(ptr);

	METOUT();
	return addQ(candidateElems_, ptr, true /* on top */);
}

bool
PCERA::NetNode::detachCandElems(NetNode * ptr)
{
	METIN();

	assert(ptr);

	METOUT();
	return delQ(candidateElems_, ptr);
}

std::list<PCERA::NetNode *>
PCERA::NetNode::getCandElems(void)
{
	METIN();

	std::list<NetNode *>			tmp;
	std::list<NetNode *>::const_iterator	iter;

	tmp.clear();

	for (iter  = candidateElems_.begin();
	     iter != candidateElems_.end();
	     iter++) {
		assert(*iter);
		tmp.push_front(*iter);
	}

	METOUT();
	return tmp;
}

uint32_t
PCERA::NetNode::rootCost(void)
{
	METIN();
	METOUT();
	return rootCost_;
}

void
PCERA::NetNode::rootCost(uint32_t  newCost)
{
	METIN();
	rootCost_ = newCost;
	METOUT();
}

uint32_t
PCERA::NetNode::spfRevision(void)
{
	METIN();
	METOUT();
	return spfRevision_;
}

void
PCERA::NetNode::spfRevision(uint32_t  newRev)
{
	METIN();

	spfRevision_ = newRev;

	METOUT();
}


nodeFlags_t
PCERA::NetNode::nodeFlags(void)
{
	METIN();
	METOUT();
	return nodeFlags_;
}

void
PCERA::NetNode::nodeFlags(nodeFlags_t & newFlags)
{
	METIN();

	nodeFlags_ = newFlags;

	METOUT();
}

bool
PCERA::NetNode::attachFwdLink(TeLink *  ptr)
{
	METIN();

	assert(ptr);

	METOUT();
	return addQ(fwdLinks_, ptr, true /* on top */);
}

bool
PCERA::NetNode::detachFwdLink(TeLink *  ptr)
{
	METIN();

	assert(ptr);

	METOUT();
	return delQ(fwdLinks_, ptr);
}

std::list<PCERA::TeLink *>
PCERA::NetNode::getFwdLinks(void)
{
	METIN();

	std::list<TeLink *>			tmp;
	std::list<TeLink *>::const_iterator	iter;

	tmp.clear();

	for (iter  = fwdLinks_.begin();
	     iter != fwdLinks_.end();
	     iter++) {
		assert(*iter);
		tmp.push_front(*iter);
	}

	METOUT();
	return tmp;
}

void
PCERA::NetNode::clearFwdLinks(void)
{
	METIN();

	fwdLinks_.clear();

	METOUT();
}

bool
PCERA::NetNode::isDomain(void)
{
	METIN();
	METOUT();
	return isDomain_;
}

bool
PCERA::NetNode::isUp(void)
{
	METIN();
	METOUT();
	return (opState_ == UP);
}

bool
PCERA::NetNode::isEnabled(void)
{
	METIN();
	METOUT();
	return (admState_ == ENABLED);
}

bool
PCERA::NetNode::fitInConstraints(uint32_t	colors,
				 uint16_t	area)
{
	METIN();
	uint32_t	rid;

	rid       = ident().id;

	if (colors &&
	    !(colors & teColors_)) {
		METOUT();
		return false;
	}

	if (area) {
		std::list<uint16_t>::const_iterator iter;
		for (iter = areas_.begin(); iter != areas_.end(); iter++) {
			if ((*iter) != area) {
				G2PCERA_DEBUG("Node %s does not belong to "
					      "area %d",
					      rid2string(rid).c_str(),
					      area);
				METOUT();
				return false;
			}
		}
	}

	METOUT();
	return true;
}

void
PCERA::NetNode::accept(topologyVisitor & vis)
{
	METIN();

	std::list<Tna *>::const_iterator	tnaIter;

	for (tnaIter = tnas_.begin(); tnaIter != tnas_.end(); tnaIter++) {
		assert(*tnaIter);
		(*tnaIter)->accept(vis);
	}

	std::list<TeLink *>::const_iterator	teIter;

	for (teIter = links_.begin(); teIter != links_.end(); teIter++) {
		assert(*teIter);
		(*teIter)->accept(vis);
	}

	vis.visitNetNode(this);

	METOUT();
}

#ifdef G2MPLS
////////////////////////////
// Class GridNode
///////////////////////////
PCERA::GridNode::GridNode(uint32_t		id) :
	Node(TOPO_NODE_GRID, id)
{
	METIN();

	name_ = 0;
	memset(&location_, 0, sizeof(location_));
	peRouterId_ = 0;

	gridServices_.clear();
	gridCompElems_.clear();
	gridSubClusters_.clear();
	gridStorageElems_.clear();

	METOUT();
}

PCERA::GridNode::~GridNode(void)
{
	METIN();

	if (gridServices_.size()) {
		G2PCERA_ERROR("There are %d gridServices on the way",
			      gridServices_.size());
		BUG();
		std::map<uint32_t, GridService *>::iterator	iter;
		for (iter  = gridServices_.begin();
		     iter != gridServices_.end();
		     iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
		METOUT();
		return;
	}
	gridServices_.clear();

	if (gridCompElems_.size()) {
		G2PCERA_ERROR("There are %d gridCompElems on the way",
			      gridCompElems_.size());
		BUG();
		std::map<uint32_t, GridCompElem *>::iterator	iter;
		for (iter  = gridCompElems_.begin();
		     iter != gridCompElems_.end();
		     iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
		METOUT();
		return;
	}
	gridCompElems_.clear();

	if (gridSubClusters_.size()) {
		G2PCERA_ERROR("There are %d gridSubClusters on the way",
			      gridSubClusters_.size());
		BUG();
		std::map<uint32_t, GridSubCluster *>::iterator	iter;
		for (iter  = gridSubClusters_.begin();
		     iter != gridSubClusters_.end();
		     iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
		METOUT();
		return;
	}
	gridSubClusters_.clear();

	if (gridStorageElems_.size()) {
		G2PCERA_ERROR("There are %d gridStorageElems on the way",
			      gridStorageElems_.size());
		BUG();
		std::map<uint32_t, GridStorageElem *>::iterator	iter;
		for (iter  = gridStorageElems_.begin();
		     iter != gridStorageElems_.end();
		     iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
		METOUT();
		return;
	}
	gridStorageElems_.clear();

	delete name_;

	METOUT();
}

void
PCERA::GridNode::dump(const std::string &	prefix,
		      bool			recursive)
{
	METIN();

	G2PCERA_DEBUG("%sGridNode", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0,ident());

	grid_site_data_t	data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}

	logDump(prefix0, data);

	G2PCERA_DEBUG("%s- GridService count   : %d",
		      prefix0.c_str(),
		      gridServices_.size());

	if (recursive) {
		std::map<uint32_t, GridService *>::const_iterator iter;
		for (iter  = gridServices_.begin();
		     iter != gridServices_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s- GridCompElems count : %d",
		      prefix0.c_str(),
		      gridCompElems_.size());

	if (recursive) {
		std::map<uint32_t, GridCompElem *>::const_iterator iter;
		for (iter  = gridCompElems_.begin();
		     iter != gridCompElems_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s- GridSubCluster count: %d",
		      prefix0.c_str(),
		      gridSubClusters_.size());

	if (recursive) {
		std::map<uint32_t, GridSubCluster *>::const_iterator iter;
		for (iter  = gridSubClusters_.begin();
		     iter != gridSubClusters_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s- GridStorageElems cnt: %d",
		      prefix0.c_str(),
		      gridStorageElems_.size());

	if (recursive) {
		std::map<uint32_t, GridStorageElem *>::const_iterator iter;
		for (iter  = gridStorageElems_.begin();
		     iter != gridStorageElems_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->dump(prefix1);
		}
	}

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

bool
PCERA::GridNode::setData(const grid_site_data_t & data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, name)) {
		name_ = new std::string(*(data.name));
	}
	if (BITMASK_BITTEST(data.mask_, location)) {
		location_ = data.location;
	}
	if (BITMASK_BITTEST(data.mask_, pe_router_id)) {
		peRouterId_ = data.pe_router_id;

		// correlate pe_router_id's TNAs, if any
		Node *			tmp;
		PCERA::NetNode *	lclNode;
		node_ident_t		nodeIdent;

		nodeIdent.id   = peRouterId_;
		nodeIdent.type = TOPO_NODE_NETWORK;

		tmp = parent()->getNode(nodeIdent);

		if (tmp && tmp->type() != TOPO_NODE_NETWORK) {
			G2PCERA_ERROR("Got a node with clashing identity %s",
				      nodeIdent2string(nodeIdent).c_str());
			METOUT();
			return false;
		}

		lclNode = static_cast<PCERA::NetNode *>(tmp);

		if (lclNode) {
			lclNode->correlateAllTnas(peRouterId_, this);
		} else {
			// peRouter could be under a domain, get TNAs directly
			std::list<node_ident_t>			nodeList;
			std::list<node_ident_t>::iterator	iter;

			nodeList = parent()->getNodesIdents();

			for (iter  = nodeList.begin();
			     iter != nodeList.end();
			     iter++) {
				node_ident_t	id;

				id = (*iter);

				if (id.type != TOPO_NODE_NETWORK) {
					continue;
				}

				tmp = parent()->getNode(id);
				lclNode = static_cast<PCERA::NetNode *>(tmp);

				if (lclNode && lclNode->isDomain()) {
					lclNode->correlateAllTnas(peRouterId_,
								  this);
				}
			}
		}
	}

	METOUT();
	return true ;
}


bool
PCERA::GridNode::getData(grid_site_data_t &	data)
{
	METIN();

	memset(&data, 0, sizeof(data));

	if (name_ && !name_->empty()) {
		data.name = new std::string(*name_);
		BITMASK_BITSET(data.mask_, name);
	}

	data.location = location_;
	BITMASK_BITSET(data.mask_, location);

	if (peRouterId_) {
		data.pe_router_id = peRouterId_;
		BITMASK_BITSET(data.mask_, pe_router_id);
	}

	METOUT();
	return true;
}

uint32_t
PCERA::GridNode::getPeRouterId(void)
{
	METIN();
	METOUT();
	return peRouterId_;
}

grid_subnodes_t
PCERA::GridNode::getSubnodesIdents(void)
{
	METIN();

	grid_subnodes_t data;

	{
		data.services.clear();
		std::map<uint32_t, GridService *>::const_iterator	iter;
		for (iter  = gridServices_.begin();
		     iter != gridServices_.end();
		     iter++) {
			assert((*iter).second);
			data.services.push_front((*iter).second->ident());
		}
	}
	{
		data.comp_elems.clear();
		std::map<uint32_t, GridCompElem *>::const_iterator	iter;
		for (iter  = gridCompElems_.begin();
		     iter != gridCompElems_.end();
		     iter++) {
			assert((*iter).second);
			data.comp_elems.push_front((*iter).second->ident());
		}
	}
	{
		data.sub_clusters.clear();
		std::map<uint32_t, GridSubCluster *>::const_iterator	iter;
		for (iter  = gridSubClusters_.begin();
		     iter != gridSubClusters_.end();
		     iter++) {
			assert((*iter).second);
			data.sub_clusters.push_front((*iter).second->ident());
		}
	}
	{
		data.storage_elems.clear();
		std::map<uint32_t, GridStorageElem *>::const_iterator	iter;
		for (iter  = gridStorageElems_.begin();
		     iter != gridStorageElems_.end();
		     iter++) {
			assert((*iter).second);
			data.storage_elems.push_front((*iter).second->ident());
		}
	}

	METOUT();
	return data;
}

bool
PCERA::GridNode::attachGridService(GridService *  ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridService is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridServices_.find(snkey) != gridServices_.end()) {
		G2PCERA_ERROR("An instance of gridServices with the same "
			      "id 0x%08x is just present "
			      "under Grid site %s... "
			      "detach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridServices_[snkey] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::GridNode::detachGridService(GridService *  ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridService is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridServices_.find(snkey) == gridServices_.end()) {
		G2PCERA_ERROR("No instance of gridServices with id 0x%08x "
			      "is present under Grid site %s... "
			      "attach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (gridServices_[snkey] != ptr) {
		G2PCERA_ERROR("A different instance of gridServices "
			      "with id 0x%08x under Grid site %s..."
			      "cannot detach",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridServices_.erase(snkey);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::GridService *
PCERA::GridNode::getGridService(void /* policy*/)
{
	METIN();

	GridService * tmp;

	tmp = 0;

	std::map<uint32_t, GridService *>::const_iterator iter;

	for (iter  = gridServices_.begin();
	     iter != gridServices_.end();
	     iter++) {
		assert((*iter).second);

		// XXX FIXME policy... now get the first
		tmp = (*iter).second;
		break;
	}

	METOUT();
	return tmp;
}


PCERA::GridService *
PCERA::GridNode::getGridService(uint32_t  id)
{
	METIN();

	GridService * tmp;

	tmp = 0;

	std::map<uint32_t, GridService *>::const_iterator iter;

	iter = gridServices_.find(id);

	if (iter != gridServices_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}

bool
PCERA::GridNode::attachGridCompElem(GridCompElem * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridCompElem is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridCompElems_.find(snkey) != gridCompElems_.end()) {
		G2PCERA_ERROR("An instance of GridCompElem with the same "
			      "id 0x%08x is just present "
			      "under Grid site %s... "
			      "detach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridCompElems_[snkey] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::GridNode::detachGridCompElem(GridCompElem * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridCompElem is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridCompElems_.find(snkey) == gridCompElems_.end()) {
		G2PCERA_ERROR("No instance of GridCompElem with id 0x%08x "
			      "is present under Grid site %s... "
			      "attach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (gridCompElems_[snkey] != ptr) {
		G2PCERA_ERROR("A different instance of GridCompElem "
			      "with id 0x%08x under Grid site %s..."
			      "cannot detach",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridCompElems_.erase(snkey);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::GridCompElem *
PCERA::GridNode::getGridCompElem(void /* policy*/)
{
	METIN();

	GridCompElem * tmp;

	tmp = 0;

	std::map<uint32_t, GridCompElem *>::const_iterator iter;

	for (iter  = gridCompElems_.begin();
	     iter != gridCompElems_.end();
	     iter++) {
		assert((*iter).second);

		// XXX FIXME policy... now get the first
		tmp = (*iter).second;
		break;
	}

	METOUT();
	return tmp;
}

PCERA::GridCompElem *
PCERA::GridNode::getGridCompElem(uint32_t  id)
{
	METIN();

	GridCompElem * tmp;

	tmp = 0;

	std::map<uint32_t, GridCompElem *>::const_iterator iter;

	iter = gridCompElems_.find(id);

	if (iter != gridCompElems_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}


bool
PCERA::GridNode::attachGridSubCluster(GridSubCluster * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridSubCluster is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridSubClusters_.find(snkey) != gridSubClusters_.end()) {
		G2PCERA_ERROR("An instance of GridSubCluster with the same "
			      "id 0x%08x is just present "
			      "under Grid site %s... "
			      "detach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridSubClusters_[snkey] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::GridNode::detachGridSubCluster(GridSubCluster * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridSubCluster is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridSubClusters_.find(snkey) == gridSubClusters_.end()) {
		G2PCERA_ERROR("No instance of GridSubCluster with id 0x%08x "
			      "is present under Grid site %s... "
			      "attach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (gridSubClusters_[snkey] != ptr) {
		G2PCERA_ERROR("A different instance of GridSubCluster "
			      "with id 0x%08x under Grid site %s..."
			      "cannot detach",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridSubClusters_.erase(snkey);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::GridSubCluster *
PCERA::GridNode::getGridSubCluster(void /* policy*/)
{
	METIN();

	GridSubCluster * tmp;

	tmp = 0;

	std::map<uint32_t, GridSubCluster *>::const_iterator iter;

	for (iter  = gridSubClusters_.begin();
	     iter != gridSubClusters_.end();
	     iter++) {
		assert((*iter).second);

		// XXX FIXME policy... now get the first
		tmp = (*iter).second;
		break;
	}

	METOUT();
	return tmp;
}

PCERA::GridSubCluster *
PCERA::GridNode::getGridSubCluster(uint32_t  id)
{
	METIN();

	GridSubCluster * tmp;

	tmp = 0;

	std::map<uint32_t, GridSubCluster *>::const_iterator iter;

	iter = gridSubClusters_.find(id);

	if (iter != gridSubClusters_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}

bool
PCERA::GridNode::attachGridStorageElem(GridStorageElem * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridStorageElem is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridStorageElems_.find(snkey) != gridStorageElems_.end()) {
		G2PCERA_ERROR("An instance of GridStorageElem with the same "
			      "id 0x%08x is just present "
			      "under Grid site %s... "
			      "detach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridStorageElems_[snkey] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::GridNode::detachGridStorageElem(GridStorageElem * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GridStorageElem is NIL");
		METOUT();
		return false;
	}

	nodeKey_t key;
	uint32_t  snkey;

	key   = nodeKeyMangle(ident().id);
	snkey = ptr->ident().id;

	if (gridStorageElems_.find(snkey) == gridStorageElems_.end()) {
		G2PCERA_ERROR("No instance of GridStorageElem with id 0x%08x "
			      "is present under Grid site %s... "
			      "attach it first",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	if (gridStorageElems_[snkey] != ptr) {
		G2PCERA_ERROR("A different instance of GridStorageElem "
			      "with id 0x%08x under Grid site %s..."
			      "cannot detach",
			      snkey, nodeKey2string(key).c_str());
		METOUT();
		return false;
	}

	gridStorageElems_.erase(snkey);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::GridStorageElem *
PCERA::GridNode::getGridStorageElem(void /* policy*/)
{
	METIN();

	GridStorageElem * tmp;

	tmp = 0;

	std::map<uint32_t, GridStorageElem *>::const_iterator iter;

	for (iter  = gridStorageElems_.begin();
	     iter != gridStorageElems_.end();
	     iter++) {
		assert((*iter).second);

		// XXX FIXME policy... now get the first
		tmp = (*iter).second;
		break;
	}

	METOUT();
	return tmp;
}

PCERA::GridStorageElem *
PCERA::GridNode::getGridStorageElem(uint32_t  id)
{
	METIN();

	GridStorageElem * tmp;

	tmp = 0;

	std::map<uint32_t, GridStorageElem *>::const_iterator iter;

	iter = gridStorageElems_.find(id);

	if (iter != gridStorageElems_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}

uint32_t
PCERA::GridNode::match(const grid_res_spec_t &	gnsRes)
{
	METIN();
	uint32_t rank;

	rank = 0;

	// XXX Add code here

	METOUT();
	return rank;
}

void
PCERA::GridNode::accept(topologyVisitor &	vis)
{
	METIN();

	{
		std::map<uint32_t, GridService *>::const_iterator	iter;
		for (iter  = gridServices_.begin();
		     iter != gridServices_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->accept(vis);
		}
	}
	{
		std::map<uint32_t, GridCompElem *>::const_iterator	iter;
		for (iter  = gridCompElems_.begin();
		     iter != gridCompElems_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->accept(vis);
		}
	}
	{
		std::map<uint32_t, GridSubCluster *>::const_iterator	iter;
		for (iter  = gridSubClusters_.begin();
		     iter != gridSubClusters_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->accept(vis);
		}
	}
	{
		std::map<uint32_t, GridStorageElem *>::const_iterator	iter;
		for (iter  = gridStorageElems_.begin();
		     iter != gridStorageElems_.end();
		     iter++) {
			assert((*iter).second);
			((*iter).second)->accept(vis);
		}
	}

	vis.visitGridNode(this);

	METOUT();
}

void
PCERA::GridService::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitGridService(this);

	METOUT();
}

void
PCERA::GridCompElem::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitGridCompElem(this);

	METOUT();
}

void
PCERA::GridSubCluster::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitGridSubCluster(this);

	METOUT();
}

void
PCERA::GridStorageElem::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitGridStorageElem(this);

	METOUT();
}
#endif // G2MPLS

////////////////////////////
// Class Tna
///////////////////////////
PCERA::Tna::Tna(const tna_ident_t &	id) :
	Ancestor<true, NetNode> (0)
{
	METIN();

	ident_   = id;
	// tna attach will set the rem[Grid]Node via lookup into topology
	remNode_ = 0;


	METOUT();

}

PCERA::Tna::~Tna(void)
{
	METIN();

	remNode_ = 0;

	METOUT();
}

tna_ident_t
PCERA::Tna::ident(void)
{
	METIN();
	METOUT();
	return ident_;
}

void
PCERA::Tna::setRemNode(GridNode * ptr)
{
	METIN();

	remNode_ = ptr;

	METOUT();
}

bool
PCERA::Tna::getRemNodeIdent(node_ident_t & ident)
{
	METIN();

	memset(&ident, 0, sizeof(ident));

	if (!remNode_) {
		METOUT();
		return false;
	}

	ident = (remNode_)->ident();

	METOUT();
	return true;
}

bool
PCERA::Tna::correlate(void)
{
	METIN();

	NetNode * ptr;

	ptr = parent();

	if (!ptr) {
		G2PCERA_ERROR("TNA has no parent NetNode yet");
		METOUT();
		return false;
	}

	uint32_t	rid;
	GridNode *	gn;

	//rid = ptr->ident().id;
	rid = ident_.rid;
	gn  = ptr->parent()->getGridNode(rid);
	if (!gn) {
		G2PCERA_WARNING("Cannot get a GridNode peering with "
				"NetNode %s... "
				"cannot correlate TNA %s",
				rid2string(rid).c_str(),
				tnaIdent2string(ident()).c_str());
		//METOUT();
		//return false;
	}

	setRemNode(gn);

	METOUT();
	return true;
}

bool
PCERA::Tna::uncorrelate(void)
{

	METIN();

	setRemNode(0);

	METOUT();
	return true;
}

void
PCERA::Tna::dump(const std::string &	prefix)
{
	METIN();

	G2PCERA_DEBUG("%sTna", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0, ident());

	node_ident_t	remGNid;
	bool		correlated;

	correlated = getRemNodeIdent(remGNid);

	G2PCERA_DEBUG("%s- Remote Grid Node       : %s",
		      prefix0.c_str(),
		      correlated ? "" : "not present yet");

	if (correlated) {
		logDump(prefix1, remGNid);
	}

	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

void
PCERA::Tna::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitTna(this);

	METOUT();
}

////////////////////////////
// Class TeLink
///////////////////////////
PCERA::TeLink::TeLink(telink_ident_t	id) :
	Ancestor<true, NetNode> (0)
{
	METIN();

	ident_                = id;
	adminMetric_          = G2PCERA_MAX_LINK_COST;
	teMetric_             = G2PCERA_MAX_LINK_COST;
	teColorMask_          = 0;
	teProtectionTypeMask_ = 0;
	teMaxBw_              = 0;
	teMaxResvBw_          = 0;
	opState_              = DOWN;
	admState_             = DISABLED;

	memset(&teAvailBw_, 0, sizeof(teAvailBw_));

	teSrlgs_.clear();
	teLinkCalendar_.clear();

	teIscsInfo_.clear();

	// used for algorithm purposes
	reverseLink_          = 0;
	linkCost_             = 0;//or G2PCERA_MAX_LINK_COST;

	METOUT();
}


PCERA::TeLink::~TeLink(void)
{
	METIN();


	teSrlgs_.clear();
	teLinkCalendar_.clear();
	reverseLink_ = 0;

	std::list<IscInfo *>::iterator	iter;

	for (iter  = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);
		delete (*iter);
	}

	teIscsInfo_.clear();

	METOUT();
}

telink_ident_t
PCERA::TeLink::ident(void)
{
	METIN();

	METOUT();
	return ident_;
}

void
PCERA::TeLink::dump(const std::string &	prefix)
{
	METIN();

	size_t sz;
	sz = 0;

	G2PCERA_DEBUG("%sTeLink", prefix.c_str());
	G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

	std::string prefix0 = prefix + DUMP_TAB;
	std::string prefix1 = prefix + DUMP_TAB + DUMP_TAB;

	logDump(prefix0, ident());

	telink_com_data_t data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}
	logDump(prefix0, data);

	opstate_t	op;
	admstate_t	adm;
	if (!getStates(op, adm)) {
		G2PCERA_ERROR("Cannot get states from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}
	logDump(prefix0, op, adm);

	sz = teIscsInfo_.size();
	G2PCERA_DEBUG("%s- ISC list size          : %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		std::list<IscInfo *>::iterator	iter;
		for (iter  = teIscsInfo_.begin();
		     iter != teIscsInfo_.end();
		     iter++) {
			assert(*iter);
			sw_cap_t	switching;
			enc_type_t	encoding;

			(*iter)->getIscTypes(switching, encoding);
			G2PCERA_DEBUG("%s-----------------------"
				      "---------------------",
				      prefix1.c_str());

			switch (switching) {
				case SWCAP_PSC_1:
				case SWCAP_PSC_2:
				case SWCAP_PSC_3:
				case SWCAP_PSC_4: {
					if (encoding == ENCT_PACKET) {
						IscPscPacketInfo * ptr;
						ptr = static_cast
							<IscPscPacketInfo *>
							(*iter);
						if (!ptr) {
							BUG();
							assert(0);
						}
						ptr->dump(prefix1);
					} else {
						(*iter)->dump(prefix1);
					}
				}
					break;
				case SWCAP_TDM: {
					if (encoding == ENCT_SDH_SONET) {
						IscTdmSdhSonetInfo * ptr;
						ptr = static_cast<
							IscTdmSdhSonetInfo *>
							(*iter);
						if (!ptr) {
							BUG();
							assert(0);
						}
						ptr->dump(prefix1);
					} else {
						(*iter)->dump(prefix1);
					}
				}
					break;
				case SWCAP_LSC: {
					if ((encoding == ENCT_G709_ODU) ||
					    (encoding == ENCT_G709_OC)) {
						IscLscG709Info * ptr;
						ptr = static_cast<
							IscLscG709Info *>
							(*iter);
						if (!ptr) {
							BUG();
							assert(0);
						}
						ptr->dump(prefix1);
					} else if (encoding == ENCT_LAMBDA) {
						IscLscLambdaInfo * ptr;
						ptr = static_cast<
							IscLscLambdaInfo *>
							(*iter);
						if (!ptr) {
							BUG();
							assert(0);
						}
						ptr->dump(prefix1);
					} else {
						(*iter)->dump(prefix1);
					}
				}
					break;
				default:
					(*iter)->dump(prefix1);
			}
		}
		G2PCERA_DEBUG("%s--------------------------------------------",
			      prefix1.c_str());
	}

	bw_per_prio_t	  bw;
	if (!getGenAvailBw(bw)) {
		G2PCERA_ERROR("Cannot get avail bw from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}
	logDump(prefix0, bw);

	std::list<uint32_t> srlgs;
	if (!getSrlgs(srlgs)) {
		G2PCERA_ERROR("Cannot get srlgs from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}
	sz = srlgs.size();
	G2PCERA_DEBUG("%s- SRLG list size         : %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		logDump(prefix0, srlgs, false, false, false);
	}

	std::map<uint32_t, bw_per_prio_t>	cal;

	if (!getCalEvents(cal)) {
		G2PCERA_ERROR("Cannot get calendar from object in %s",
			      __PRETTY_FUNCTION__);
		METOUT();
		return;
	}
	sz = cal.size();

	G2PCERA_DEBUG("%s- Calendar size          : %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		logDump(prefix0, cal);
	}

	G2PCERA_DEBUG("%s- TE-Link Path Computation info",
		      prefix0.c_str());
	G2PCERA_DEBUG("%s- link cost           : %d",
		      prefix1.c_str(),
		      linkCost_);
	G2PCERA_DEBUG("%s- Correlation info",
		      prefix0.c_str());
	G2PCERA_DEBUG("%s- reverse te-link     : %s",
		      prefix1.c_str(),
		      (reverseLink_ ?
		       teLinkIdent2string(reverseLink_->ident()).c_str() :
		       "--------"));


	G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
	METOUT();
}

bool
PCERA::TeLink::setData(const telink_com_data_t &	data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, admin_metric)) {
		adminMetric_ = data.admin_metric;
	}
	if (BITMASK_BITTEST(data.mask_, te_metric)) {
		teMetric_ = data.te_metric;
	}
	if (BITMASK_BITTEST(data.mask_, te_color_mask)) {
		teColorMask_ = data.te_color_mask;
	}
	if (BITMASK_BITTEST(data.mask_, te_prot_type_mask)) {
		teProtectionTypeMask_ = data.te_prot_type_mask;
	}
	if (BITMASK_BITTEST(data.mask_, te_max_bw)) {
		teMaxBw_ = data.te_max_bw;
	}
	if (BITMASK_BITTEST(data.mask_, te_max_resv_bw)) {
		teMaxResvBw_ = data.te_max_resv_bw;
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::getData(telink_com_data_t &	data)
{
	METIN();

	memset(&data, 0, sizeof(data));

	data.admin_metric      = adminMetric_;
	data.te_metric         = teMetric_;
	data.te_color_mask     = teColorMask_;
	data.te_prot_type_mask = teProtectionTypeMask_;
	data.te_max_bw         = teMaxBw_;
	data.te_max_resv_bw    = teMaxResvBw_;

	BITMASK_BITSET(data.mask_, admin_metric);
	BITMASK_BITSET(data.mask_, te_metric);
	BITMASK_BITSET(data.mask_, te_color_mask);
	BITMASK_BITSET(data.mask_, te_prot_type_mask);
	BITMASK_BITSET(data.mask_, te_max_bw);
	BITMASK_BITSET(data.mask_, te_max_resv_bw);

	METOUT();
	return true;
}

PCERA::NetNode *
PCERA::TeLink::getRemNode(void)
{
	METIN();

	Node *		nd;
	NetNode *	nn;
	NetNode *	rn;
	node_ident_t	remNodeIdent;
	telink_ident_t	linkIdent;

	linkIdent = ident();
	nn        = parent();

	remNodeIdent.type = TOPO_NODE_NETWORK;


	switch (ident_.mode) {
		case LINK_MODE_ENNI_INTERDOMAIN: {
			remNodeIdent.id = linkIdent.rem_rc_id;

			if ((!linkIdent.lcl_rc_id && !linkIdent.rem_rc_id) ||
			    (linkIdent.lcl_rc_id &&
			     (linkIdent.rem_rc_id == linkIdent.lcl_rc_id))) {
				G2PCERA_ERROR("Incoherent identity for "
					      "LINK_MODE_ENNI_INTERDOMAIN %s",
					      teLinkIdent2string(linkIdent).
					      c_str());
				METOUT();
				return NULL;
			}

			if (!linkIdent.rem_rc_id && (nn && nn->isDomain())) {
				G2PCERA_DEBUG("Remote node for TElink %s "
					      "should be in my native domain!",
					      teLinkIdent2string(linkIdent).
					      c_str());
				// remote node for these links is in my domain
				remNodeIdent.id = linkIdent.rem_rid;
			}
		}
			break;
		case LINK_MODE_ENNI_INTRADOMAIN: {
			remNodeIdent.id = linkIdent.rem_rc_id;
			if (linkIdent.rem_rc_id != linkIdent.lcl_rc_id) {
				G2PCERA_ERROR("Incoherent identity for "
					      "LINK_MODE_ENNI_INTRADOMAIN %s",
					      teLinkIdent2string(linkIdent).
					      c_str());
				METOUT();
				return NULL;
			}
		}
			break;
		default:
			remNodeIdent.id = linkIdent.rem_rid;
			break;
	}

	nd = nn ? nn->parent()->getNode(remNodeIdent) : 0;
	rn = static_cast<PCERA::NetNode *>(nd);
	if (!rn) {
		G2PCERA_WARNING("Cannot get remote NetNode %s peering with "
				"node %s through TE-link %s... "
				"correlation incomplete",
				rid2string(linkIdent.rem_rid).c_str(),
				rid2string(linkIdent.lcl_rid).c_str(),
				teLinkIdent2string(linkIdent).c_str());
	}

	METOUT();
	return rn;
}

void
PCERA::TeLink::setRemLink(TeLink * ptr)
{
	METIN();

	reverseLink_ = ptr;

	METOUT();
}

PCERA::TeLink *
PCERA::TeLink::getRemLink(void)
{
	METIN();
	METOUT();
	return reverseLink_;
}

bool
PCERA::TeLink::correlate(void)
{
	METIN();

	// correlate teLink with its remote NetNode
	NetNode *	rn;
	telink_ident_t	linkIdent;

	linkIdent = ident();

	rn = getRemNode();

	if (rn) {
		// correlate to reverse TE-link if any
		TeLink *	remLinkPtr;
		telink_ident_t	remLinkIdent;

		remLinkIdent.lcl_rid   = linkIdent.rem_rid;
		remLinkIdent.lcl_id    = linkIdent.rem_id;
		remLinkIdent.rem_rid   = linkIdent.lcl_rid;
		remLinkIdent.rem_id    = linkIdent.lcl_id;
		remLinkIdent.mode      = linkIdent.mode;
		remLinkIdent.lcl_rc_id = linkIdent.rem_rc_id;
		remLinkIdent.rem_rc_id = linkIdent.lcl_rc_id;

		remLinkPtr = rn->getOutLink(remLinkIdent);

		setRemLink(remLinkPtr);

		if (!remLinkPtr) {
			G2PCERA_WARNING("Cannot get remote TE-Link %s "
					"into remote node %s to "
					"correlate to TE-link %s",
					teLinkIdent2string(remLinkIdent).
					c_str(),
					rid2string(linkIdent.rem_rid).c_str(),
					teLinkIdent2string(linkIdent).c_str());
		}
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::uncorrelate(void)
{
	METIN();

	if (reverseLink_) {
		// reset the pointer to me from my remote brother object
		reverseLink_->setRemLink(0);
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::setStates(const opstate_t &	op,
			 const admstate_t &	adm)
{
	METIN();

	opState_  = op;
	admState_ = adm;

	METOUT();
	return true;
}

bool
PCERA::TeLink::getStates(opstate_t &		op,
			 admstate_t &		adm)
{
	METIN();

	op  = opState_;
	adm = admState_;

	METOUT();
	return true;
}

bool
PCERA::TeLink::attachIscInfo(IscInfo *	ptr)
{
	METIN();
	std::list<IscInfo *>::const_iterator	iter;

	if (!ptr) {
		G2PCERA_ERROR("Passed IscInfo is NIL");
		METOUT();
		return false;
	}

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);

		if ((*(*iter)) == *ptr) {
			G2PCERA_ERROR("Duplicated ISC to append "
				      "under TeLink %s",
				      teLinkIdent2string(ident_).c_str());
			METOUT();
			return false;
		}
	}

	teIscsInfo_.push_front(ptr);

	METOUT();
	return true;
}

bool
PCERA::TeLink::detachIscInfo(IscInfo *	ptr)
{
	METIN();
	std::list<IscInfo *>::iterator	iter;

	if (!ptr) {
		G2PCERA_ERROR("Passed IscInfo is NIL");
		METOUT();
		return false;
	}

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);

		if ((*(*iter)) == *ptr) {
			break;
		}
	}

	if (iter == teIscsInfo_.end()) {
		G2PCERA_ERROR("Missing ISC to remove "
			      "under TeLink %s",
			      teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (*iter != ptr) {
		G2PCERA_ERROR("A different instance of the ISC "
			      "found under TeLink %s... ",
			      teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	teIscsInfo_.erase(iter);

	METOUT();
	return true;
}

bool
PCERA::TeLink::getIscs(std::list<isc_t> &	iscs)
{
	METIN();

	std::list<IscInfo *>::const_iterator	iter;

	iscs.clear();

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);

		isc_t	tmp;

		memset(&tmp, 0, sizeof(tmp));

		(*iter)->getIscTypes(tmp.switching, tmp.encoding);
		(*iter)->getMaxLspBw(tmp.max_lsp_bw);

		if ((tmp.encoding == ENCT_PACKET) &&
		    ((tmp.switching == SWCAP_PSC_1)  ||
		     (tmp.switching == SWCAP_PSC_2)  ||
		     (tmp.switching == SWCAP_PSC_3)  ||
		     (tmp.switching == SWCAP_PSC_4))) {

			telink_psc_data_t	data;
			IscPscPacketInfo *	tmp1;

			tmp1 = static_cast<IscPscPacketInfo *>(*iter);
			if (tmp1) {
				tmp1->getData(data);

				tmp.spec.psc.min_lsp_bw = data.min_lsp_bw;
				tmp.spec.psc.if_mtu     = data.if_mtu;
			}
		} else if ((tmp.encoding  == ENCT_SDH_SONET) &&
			   (tmp.switching == SWCAP_TDM)) {
			telink_tdm_data_t data;

			IscTdmSdhSonetInfo *	tmp1;

			tmp1 = static_cast<IscTdmSdhSonetInfo *>(*iter);
			if (tmp1) {
				tmp1->getData(data);

				tmp.spec.tdm.min_lsp_bw = data.min_lsp_bw;
				tmp.spec.tdm.indication =
					(uint8_t) data.stdarb_cc_indication;
			}
		}

		iscs.push_front(tmp);
	}

	METOUT();
	return true;
}

PCERA::IscPscPacketInfo *
PCERA::TeLink::getIscPscPacket(void)
{
	IscPscPacketInfo *			tmp;
	std::list<IscInfo *>::const_iterator	iter;

	METIN();

	tmp = 0;

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);
		sw_cap_t	sw;
		enc_type_t	enc;

		(*iter)->getIscTypes(sw, enc);

		if ((sw  == SWCAP_PSC_1 ||
		     sw  == SWCAP_PSC_2 ||
		     sw  == SWCAP_PSC_3 ||
		     sw  == SWCAP_PSC_4) &&
		    (enc == ENCT_PACKET))  {
			tmp = static_cast<IscPscPacketInfo *> (*iter);
			break;
		}
	}

	METOUT();
	return tmp;
}

PCERA::IscTdmSdhSonetInfo *
PCERA::TeLink::getIscTdmSdhSonet(void)
{
	IscTdmSdhSonetInfo *			tmp;
	std::list<IscInfo *>::const_iterator	iter;

	METIN();

	tmp = 0;

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);
		sw_cap_t	sw;
		enc_type_t	enc;

		(*iter)->getIscTypes(sw, enc);

		if ((sw  == SWCAP_TDM) &&
		    (enc == ENCT_SDH_SONET))  {
			tmp = static_cast<IscTdmSdhSonetInfo *> (*iter);
			break;
		}
	}

	METOUT();
	return tmp;
}


PCERA::IscLscG709Info *
PCERA::TeLink::getIscLscG709(void)
{
	IscLscG709Info *			tmp;
	std::list<IscInfo *>::const_iterator	iter;

	METIN();

	tmp = 0;

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);
		sw_cap_t	sw;
		enc_type_t	enc;

		(*iter)->getIscTypes(sw, enc);

		if ((sw  == SWCAP_LSC) &&
		    (enc == ENCT_G709_ODU ||
		     enc == ENCT_G709_OC))  {
			tmp = static_cast<IscLscG709Info *> (*iter);
			break;
		}
	}

	METOUT();
	return tmp;
}


PCERA::IscLscLambdaInfo *
PCERA::TeLink::getIscLscLambda(void)
{
	IscLscLambdaInfo *			tmp;
	std::list<IscInfo *>::const_iterator	iter;

	METIN();

	tmp = 0;

	for (iter = teIscsInfo_.begin(); iter != teIscsInfo_.end(); iter++) {
		assert(*iter);
		sw_cap_t	sw;
		enc_type_t	enc;

		(*iter)->getIscTypes(sw, enc);

		if ((sw  == SWCAP_LSC) &&
		    (enc == ENCT_LAMBDA))  {
			tmp = static_cast<IscLscLambdaInfo *> (*iter);
			break;
		}
	}

	METOUT();
	return tmp;
}


bool
PCERA::TeLink::setGenAvailBw(const bw_per_prio_t & bw)
{
	METIN();

	teAvailBw_ = bw;

	METOUT();
	return true;
}

bool
PCERA::TeLink::getGenAvailBw(bw_per_prio_t &	  bw)
{
	METIN();

	bw = teAvailBw_;

	METOUT();
	return true;
}

bool
PCERA::TeLink::appendSrlgs(const std::list<uint32_t> &	srlgs)
{
	METIN();

	std::list<uint32_t>::const_iterator iter;

	for (iter = srlgs.begin(); iter != srlgs.end(); iter++) {
		if (!addQ(teSrlgs_, *iter, true)) {
			G2PCERA_WARNING("Duplicated SRLG (0x%x) to append "
					"under TeLink %s",
					*iter,
					teLinkIdent2string(ident_).c_str());
			//METOUT();
			//return false;
		}
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::removeSrlgs(std::list<uint32_t> &	srlgs)
{
	METIN();

	std::list<uint32_t>::iterator iter;

	for (iter = srlgs.begin(); iter != srlgs.end(); iter++) {
		if (!delQ(teSrlgs_, *iter)) {
			G2PCERA_ERROR("Missing SRLG (0x%x) to remove "
				      "under TeLink %s",
				      *iter,
				      teLinkIdent2string(ident_).c_str());
			METOUT();
			return false;
		}
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::getSrlgs(std::list<uint32_t> &		srlgs)
{
	METIN();

	std::list<uint32_t>::const_iterator	iter;

	srlgs.clear();

	for (iter = teSrlgs_.begin(); iter != teSrlgs_.end(); iter++) {
		srlgs.push_front(*iter);
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::appendCalEvents(const std::map<uint32_t, bw_per_prio_t> & cal)
{
	METIN();

	std::map<uint32_t, bw_per_prio_t>::const_iterator	iter;

	for (iter = cal.begin(); iter != cal.end(); iter++) {

		uint32_t key;
		key = (*iter).first;
		teLinkCalendar_[key] = (*iter).second;
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::removeCalEvents(std::map<uint32_t, bw_per_prio_t> & cal)
{
	METIN();

	std::map<uint32_t, bw_per_prio_t>::iterator	iter;

	for (iter = cal.begin(); iter != cal.end(); iter++) {

		uint32_t key;
		key = (*iter).first;

		if (teLinkCalendar_.find(key) == teLinkCalendar_.end()) {
			G2PCERA_ERROR("No calendar event found at time %s "
				      "under TeLink %s",
				      unixTime2string(key).c_str(),
				      teLinkIdent2string(ident_).c_str());
			METOUT();
			return false;
		}

		teLinkCalendar_.erase(key);
	}

	METOUT();
	return true;
}

bool
PCERA::TeLink::getCalEvents(std::map<uint32_t, bw_per_prio_t> & cal)
{
	METIN();

	std::map<uint32_t, bw_per_prio_t>::iterator	iter;

	cal.clear();

	cal = teLinkCalendar_;

	METOUT();
	return true;
}


bool
PCERA::TeLink::fitInConstraints(const cspf_constr_t & data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::TeLink::fitInConstraints(const lsp_info_t & data)
{
	METIN();

	if (!isUp()) {
		G2PCERA_WARNING("Operative state is DOWN for link %s",
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (!isEnabled()) {
		G2PCERA_WARNING("Administrative state is DISABLED fo link %s",
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}


	if (data.bw > teMaxBw_) {
		G2PCERA_WARNING("Requested bw (%.3fMbps) is higher than "
				"TE MAX LINK BW (%.3fMbps) for link %s",
				BW_HEX2MBPS(data.bw),
				BW_HEX2MBPS(teMaxBw_),
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.bw > teMaxResvBw_) {
		G2PCERA_WARNING("Requested bw (%.3fMbps) is higher than "
				"TE MAX RESV BW (%.3fMbps) for link %s",
				BW_HEX2MBPS(data.bw),
				BW_HEX2MBPS(teMaxResvBw_),
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.bw > teAvailBw_.bw[data.setup_prio]) {
		G2PCERA_WARNING("Requested bw (%.3fMbps) is higher than "
				"TE AVAIL BW at prio %d (%.3fMbps) "
				"for link %s",
				BW_HEX2MBPS(data.bw),
				data.setup_prio,
				BW_HEX2MBPS(teAvailBw_.bw[data.setup_prio]),
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	std::list<isc_t> iscs;
	bool iscOk;

	iscOk = false;

	if (getIscs(iscs)) {
		std::list<isc_t>::const_iterator iter;

		for (iter = iscs.begin(); iter != iscs.end(); iter++) {
			if ((uint32_t)(data.sw_cap) >
			    (uint32_t)((*iter).switching)) {
				continue;
			}
			if ((uint32_t)(data.enc_type) >
			    (uint32_t)((*iter).encoding)) {
				continue;
			}

			if (data.bw <= (*iter).max_lsp_bw.bw[data.setup_prio]){
				iscOk = true;
				break;
			}

		}
	}

	if (!iscOk) {
		G2PCERA_WARNING("No ISC match for link %s",
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.link_prot_mask &&
	    !(data.link_prot_mask & teProtectionTypeMask_)) {
		G2PCERA_WARNING("Mismatch in  TE PROTECTION TYPE MASKs "
				"(req. 0x%02x | link 0x%02x) for link %s",
				data.link_prot_mask,
				teProtectionTypeMask_,
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.include_any &&
	    !(data.include_any & teColorMask_)) {
		G2PCERA_WARNING("Mismatch in INCLUDE ANY COLORs "
				"(req. 0x%08x | link 0x%08x) for link %s",
				data.include_any,
				teColorMask_,
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.include_all &&
	    (data.include_all != (data.include_all & teColorMask_))) {
		G2PCERA_WARNING("Mismatch in INCLUDE ALL COLORs "
				"(req. 0x%08x | link 0x%08x) for link %s",
				data.include_all,
				teColorMask_,
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.exclude_any & teColorMask_) {
		G2PCERA_WARNING("Mismatch in EXCLUDE ANY COLORs "
				"(req. 0x%08x | link 0x%08x) for link %s",
				data.exclude_any,
				teColorMask_,
				teLinkIdent2string(ident_).c_str());
		METOUT();
		return false;
	}

	if (data.mask_.times && !teLinkCalendar_.empty()) {
		std::map<uint32_t, bw_per_prio_t>::const_iterator  iter;
		std::map<uint32_t, bw_per_prio_t>::const_iterator  hiIter;
		std::map<uint32_t, bw_per_prio_t>::const_iterator  lowIter;

		lowIter = teLinkCalendar_.upper_bound(data.times.start_time);
		hiIter  = teLinkCalendar_.lower_bound(data.times.end_time);

		for (iter = --lowIter; iter != hiIter; iter++) {
			if (data.bw > (*iter).second.bw[data.setup_prio]) {
				G2PCERA_WARNING("Requested bw in advance "
						"(%.3fMbps) is "
						"higher than the available "
						"TE AVAIL BW at prio %d "
						"(%.3fMbps) "
						"at time %d "
						"for link %s",
						BW_HEX2MBPS(data.bw),
						data.setup_prio,
						BW_HEX2MBPS((*iter).second.bw[data.setup_prio]),
						(*iter).first,
						teLinkIdent2string(ident_).c_str());
				METOUT();
				return false;
			}
		}
	}

	METOUT();
	return true;
}

int64_t
PCERA::TeLink::linkCost(void)
{
	METIN();
	METOUT();
	return linkCost_;
}

int64_t
PCERA::TeLink::generateCost(const lsp_info_t & data)
{
	METIN();

	int64_t res;

	res = teMetric_;

	if (data.mask_.times && teLinkCalendar_.empty()) {
		res+= G2PCERA_ADVRES_EXTRA_LINK_COST;
	}

	METOUT();
	return res;
}

void
PCERA::TeLink::linkCost(int64_t newCost)
{
	METIN();

	linkCost_ = newCost;

	METOUT();
}

bool
PCERA::TeLink::isUp(void)
{
	METIN();
	METOUT();
	return (opState_ == UP);
}

bool
PCERA::TeLink::isEnabled(void)
{
	METIN();
	METOUT();
	return (admState_ == ENABLED);
}

void
PCERA::TeLink::accept(topologyVisitor &	vis)
{
	METIN();

	vis.visitTeLink(this);

	METOUT();
}

////////////////////////////
// Class IscInfo
///////////////////////////
PCERA::IscInfo::IscInfo(sw_cap_t			switching,
			enc_type_t			encoding,
			const bw_per_prio_t &		maxLspBw)
{
	METIN();

	switching_ = switching;
	encoding_  = encoding;
	maxLspBw_  = maxLspBw;

	METOUT();
}

PCERA::IscInfo::~IscInfo()
{
	METIN();
	METOUT();
}

void
PCERA::IscInfo::dump(const std::string &	prefix)

{
	size_t		i;
	std::string	prefix0 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- swcap            : %s",
		      prefix0.c_str(),
		      SHOW_SW_CAP(switching_));
	G2PCERA_DEBUG("%s- enctype          : %s",
		      prefix0.c_str(),
		      SHOW_ENC_TYPE(encoding_));
	for (i = 0; i < GMPLS_BW_PRIORITIES; i++) {
		G2PCERA_DEBUG("%s- te max LSP bw[%d] : %.3fMbps",
			      prefix0.c_str(),
			      i,
			      BW_HEX2MBPS(maxLspBw_.bw[i]));
	}
}

void
PCERA::IscInfo::getIscTypes(sw_cap_t &		switching,
			    enc_type_t &	encoding) const
{
	METIN();

	switching = switching_;
	encoding  = encoding_;

	METOUT();
}

bool
PCERA::IscInfo::getMaxLspBw(bw_per_prio_t &	bw) const
{
	METIN();

	bw = maxLspBw_;

	METOUT();
}

bool
PCERA::IscInfo::fitInConstraints(const cspf_constr_t &	data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::IscInfo::operator==(const IscInfo & other) const
{
	size_t i;

	if (switching_ != other.switching_) {
		return false;
	}
	if (encoding_  != other.encoding_ ) {
		return false;
	}
	for (i = 0; i < MAX_BW_PRIORITIES; i++) {
		if (maxLspBw_.bw[i] != other.maxLspBw_.bw[i]) {
			return false;
		}
	}
	return true;
}

bool
PCERA::IscInfo::operator!=(const IscInfo & other) const
{
	return !(*this == other);
}



////////////////////////////
// Class IscPscPacketInfo
///////////////////////////
PCERA::IscPscPacketInfo::IscPscPacketInfo(sw_cap_t		switching,
					  enc_type_t		encoding,
					  const bw_per_prio_t &	maxLspBw,
					  uint32_t		minLspBw,
					  uint16_t		ifMtu) :
	IscInfo(switching, encoding, maxLspBw)
{
	METIN();

	minLspBw_ = minLspBw;
	ifMtu_    = ifMtu;

	METOUT();
}

PCERA::IscPscPacketInfo::~IscPscPacketInfo()
{
	METIN();
	METOUT();
}

void
PCERA::IscPscPacketInfo::dump(const std::string &	prefix)

{
	IscInfo::dump(prefix);

	std::string prefix0 = prefix + DUMP_TAB;

	telink_psc_data_t	data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	logDump(prefix0, data);
}


bool
PCERA::IscPscPacketInfo::getData(telink_psc_data_t &		data)
{
	METIN();

	data.min_lsp_bw = minLspBw_;
	data.if_mtu     = ifMtu_;

	BITMASK_BITSET(data.mask_, min_lsp_bw);
	BITMASK_BITSET(data.mask_, if_mtu);

	METOUT();
}

bool
PCERA::IscPscPacketInfo::fitInConstraints(const cspf_constr_t &	data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::IscPscPacketInfo::operator==(const IscPscPacketInfo & other) const
{
	size_t		i;
	sw_cap_t	sw1, sw2;
	enc_type_t	enc1, enc2;
	bw_per_prio_t	bw1, bw2;

	getIscTypes(sw1, enc1);
	getMaxLspBw(bw1);

	other.getIscTypes(sw2, enc2);
	other.getMaxLspBw(bw2);

	if (sw1 != sw2) {
		return false;
	}
	if (enc1 != enc2) {
		return false;
	}
	for (i = 0; i < MAX_BW_PRIORITIES; i++) {
		if (bw1.bw[i] != bw2.bw[i]) {
			return false;
		}
	}
	if (minLspBw_ != other.minLspBw_) {
		return false;
	}
	if (ifMtu_  != other.ifMtu_ ) {
		return false;
	}
	return true;
}

bool
PCERA::IscPscPacketInfo::operator!=(const IscPscPacketInfo & other) const
{
	return !(*this == other);
}


////////////////////////////
// Class IscTdmSdhSonetInfo
///////////////////////////
PCERA::IscTdmSdhSonetInfo::IscTdmSdhSonetInfo(sw_cap_t		    switching,
					      enc_type_t	    encoding,
					      const bw_per_prio_t & maxLspBw,
					      uint32_t		    minLspBw,
					      gmpls_sdhsonet_stdarbcap_t sACC):
	IscInfo(switching, encoding, maxLspBw)
{
	METIN();

	minLspBw_         = minLspBw;
	stdArbConc_       = sACC;

	hoMuxCapMask_     = 0;
	loMuxCapMask_     = 0;
	transparencyMask_ = 0;
	blsrRingId_       = 0;

	freeTimeslots_.clear();

	METOUT();
}

PCERA::IscTdmSdhSonetInfo::~IscTdmSdhSonetInfo()
{
	METIN();

	freeTimeslots_.clear();

	METOUT();
}

void
PCERA::IscTdmSdhSonetInfo::dump(const std::string &	prefix)

{
	IscInfo::dump(prefix);

	std::string prefix0 = prefix + DUMP_TAB;

	telink_tdm_data_t	data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	logDump(prefix0, data);

	size_t			sz;
	std::list<uint32_t>	fts;

	if (!getTdmAvailBw(fts)) {
		G2PCERA_ERROR("Cannot get free timeslots from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	sz = fts.size();
	G2PCERA_DEBUG("%s- SDH/Sonet free TS: %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		logDump(prefix0, fts, true, false, false);
	}
}

bool
PCERA::IscTdmSdhSonetInfo::setData(const telink_tdm_data_t &	data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, min_lsp_bw)) {
		minLspBw_ = data.min_lsp_bw;
	}
	if (BITMASK_BITTEST(data.mask_, stdarb_cc_indication)) {
		stdArbConc_ = data.stdarb_cc_indication;
	}
	if (BITMASK_BITTEST(data.mask_, ho_muxcap_mask)) {
		hoMuxCapMask_ = data.ho_muxcap_mask;
	}
	if (BITMASK_BITTEST(data.mask_, lo_muxcap_mask)) {
		loMuxCapMask_ = data.lo_muxcap_mask;
	}
	if (BITMASK_BITTEST(data.mask_, transparency_mask)) {
		transparencyMask_ = data.transparency_mask;
	}
	if (BITMASK_BITTEST(data.mask_, blsr_ring_id)) {
		blsrRingId_ = data.blsr_ring_id;
	}

	METOUT();

	return true;
}

bool
PCERA::IscTdmSdhSonetInfo::getData(telink_tdm_data_t &		data)
{
	METIN();

	data.min_lsp_bw           = minLspBw_;
	data.stdarb_cc_indication = stdArbConc_;

	data.ho_muxcap_mask       = hoMuxCapMask_;
	data.lo_muxcap_mask       = loMuxCapMask_;
	data.transparency_mask    = transparencyMask_;
	data.blsr_ring_id         = blsrRingId_;

	BITMASK_BITSET(data.mask_, min_lsp_bw);
	BITMASK_BITSET(data.mask_, stdarb_cc_indication);
	BITMASK_BITSET(data.mask_, ho_muxcap_mask);
	BITMASK_BITSET(data.mask_, lo_muxcap_mask);
	BITMASK_BITSET(data.mask_, transparency_mask);
	BITMASK_BITSET(data.mask_, blsr_ring_id);

	METOUT();
	return true;
}

bool
PCERA::IscTdmSdhSonetInfo::setTdmAvailBw(const std::list<uint32_t> & fts)
{
	METIN();

	freeTimeslots_ = fts;

	METOUT();
	return true;
}

bool
PCERA::IscTdmSdhSonetInfo::getTdmAvailBw(std::list<uint32_t> &	  fts)
{
	METIN();

	fts = freeTimeslots_;

	METOUT();
	return true;
}

bool
PCERA::IscTdmSdhSonetInfo::fitInConstraints(const cspf_constr_t &	data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::IscTdmSdhSonetInfo::operator==(const IscTdmSdhSonetInfo & other) const
{
	size_t		i;
	sw_cap_t	sw1, sw2;
	enc_type_t	enc1, enc2;
	bw_per_prio_t	bw1, bw2;

	getIscTypes(sw1, enc1);
	getMaxLspBw(bw1);

	other.getIscTypes(sw2, enc2);
	other.getMaxLspBw(bw2);

	if (sw1 != sw2) {
		return false;
	}
	if (enc1 != enc2) {
		return false;
	}
	for (i = 0; i < MAX_BW_PRIORITIES; i++) {
		if (bw1.bw[i] != bw2.bw[i]) {
			return false;
		}
	}
	if (minLspBw_ != other.minLspBw_) {
		return false;
	}
	if (stdArbConc_  != other.stdArbConc_) {
		return false;
	}
	return true;
}

bool
PCERA::IscTdmSdhSonetInfo::operator!=(const IscTdmSdhSonetInfo & other) const
{
	return !(*this == other);
}


////////////////////////////
// Class IscLscG709Info
///////////////////////////
PCERA::IscLscG709Info::IscLscG709Info(sw_cap_t		    switching,
				      enc_type_t	    encoding,
				      const bw_per_prio_t & maxLspBw):
	IscInfo(switching, encoding, maxLspBw)
{
	METIN();

	odukMuxCapMask_ = 0;

	unallocODUk_.clear();
	unallocOCh_.clear();

	METOUT();
}

PCERA::IscLscG709Info::~IscLscG709Info()
{
	METIN();

	unallocODUk_.clear();
	unallocOCh_.clear();

	METOUT();
}

void
PCERA::IscLscG709Info::dump(const std::string &	prefix)

{
	IscInfo::dump(prefix);

	std::string prefix0 = prefix + DUMP_TAB;

	telink_lscg709_data_t	data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	logDump(prefix0, data);

	size_t			sz;
	std::list<uint32_t>	foduk;
	std::list<uint32_t>	foch;

	if (!getLscG709AvailBw(foduk, foch)) {
		G2PCERA_ERROR("Cannot get free oduk/och from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	sz = foduk.size();
	G2PCERA_DEBUG("%s- G.709 free ODUk  : %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		logDump(prefix0, foduk, false, true, false);
	}

	sz = foch.size();
	G2PCERA_DEBUG("%s- G.709 free OCh   : %d items",
		      prefix0.c_str(),
		      sz);
	if (sz) {
		logDump(prefix0, foch, false, false, true);
	}
}

bool
PCERA::IscLscG709Info::setData(const telink_lscg709_data_t &	data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, oduk_muxcap_mask)) {
		odukMuxCapMask_ = data.oduk_muxcap_mask;
	}

	METOUT();

	return true;
}

bool
PCERA::IscLscG709Info::getData(telink_lscg709_data_t &		data)
{
	METIN();

	data.oduk_muxcap_mask = odukMuxCapMask_;

	BITMASK_BITSET(data.mask_, oduk_muxcap_mask);

	METOUT();
	return true;
}

bool
PCERA::IscLscG709Info::setLscG709AvailBw(const std::list<uint32_t> &	foduk,
					 const std::list<uint32_t> &	foch)
{
	METIN();

	unallocODUk_ = foduk;
	unallocOCh_  = foch;

	METOUT();
	return true;
}

bool
PCERA::IscLscG709Info::getLscG709AvailBw(std::list<uint32_t> &	foduk,
					 std::list<uint32_t> &	foch)
{
	METIN();

	foduk = unallocODUk_;
	foch  = unallocOCh_;

	METOUT();
	return true;
}

bool
PCERA::IscLscG709Info::fitInConstraints(const cspf_constr_t &	data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::IscLscG709Info::operator==(const IscLscG709Info & other) const
{
	size_t		i;
	sw_cap_t	sw1, sw2;
	enc_type_t	enc1, enc2;
	bw_per_prio_t	bw1, bw2;

	getIscTypes(sw1, enc1);
	getMaxLspBw(bw1);

	other.getIscTypes(sw2, enc2);
	other.getMaxLspBw(bw2);

	if (sw1 != sw2) {
		return false;
	}
	if (enc1 != enc2) {
		return false;
	}
	for (i = 0; i < MAX_BW_PRIORITIES; i++) {
		if (bw1.bw[i] != bw2.bw[i]) {
			return false;
		}
	}
	return true;
}

bool
PCERA::IscLscG709Info::operator!=(const IscLscG709Info & other) const
{
	return !(*this == other);
}


////////////////////////////
// Class IscLscLambdaInfo
///////////////////////////
PCERA::IscLscLambdaInfo::IscLscLambdaInfo(sw_cap_t		switching,
					  enc_type_t		encoding,
					  const bw_per_prio_t &	maxLspBw):
	IscInfo(switching, encoding, maxLspBw)
{
	METIN();

	dispersionPMD_ = 0;
	spanLength_    = 0;

	amplifiers_.clear();

	memset(&lambdasBitmap_, 0, sizeof(lambdasBitmap_));

	METOUT();
}

PCERA::IscLscLambdaInfo::~IscLscLambdaInfo()
{
	METIN();

	amplifiers_.clear();
	delete [] lambdasBitmap_.bitmap_word;

	METOUT();
}

void
PCERA::IscLscLambdaInfo::dump(const std::string &	prefix)

{
	IscInfo::dump(prefix);

	std::string prefix0 = prefix + DUMP_TAB;

	telink_lscwdm_data_t	data;
	if (!getData(data)) {
		G2PCERA_ERROR("Cannot get data from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	logDump(prefix0, data);

	wdm_link_lambdas_bitmap_t bm;

	if (!getLscWdmAvailBw(bm)) {
		G2PCERA_ERROR("Cannot get lambda bitmap from object in %s",
			      __PRETTY_FUNCTION__);
		return;
	}

	logDump(prefix0, bm);
}

bool
PCERA::IscLscLambdaInfo::setData(const telink_lscwdm_data_t &	data)
{
	METIN();

	if (BITMASK_BITTEST(data.mask_, dispersion_PMD)) {
		dispersionPMD_ = data.dispersion_PMD;
	}

	if (BITMASK_BITTEST(data.mask_, span_length)) {
		spanLength_ = data.span_length;
	}

	if (BITMASK_BITTEST(data.mask_, amplifiers)) {
		amplifiers_ = data.amplifiers;
	}

	METOUT();

	return true;
}

bool
PCERA::IscLscLambdaInfo::getData(telink_lscwdm_data_t &		data)
{
	METIN();

	data.dispersion_PMD = dispersionPMD_;
	data.span_length    = spanLength_;
	data.amplifiers     = amplifiers_;

	BITMASK_BITSET(data.mask_, dispersion_PMD);
	BITMASK_BITSET(data.mask_, span_length);
	BITMASK_BITSET(data.mask_, amplifiers);

	METOUT();
	return true;
}

bool
PCERA::IscLscLambdaInfo::setLscWdmAvailBw(const wdm_link_lambdas_bitmap_t & bm)
{
	METIN();

	delete [] lambdasBitmap_.bitmap_word;

	lambdasBitmap_.base_lambda_label = bm.base_lambda_label;
	lambdasBitmap_.num_wavelengths   = bm.num_wavelengths;
	lambdasBitmap_.bitmap_size       = bm.bitmap_size;
	lambdasBitmap_.bitmap_word       = 0;

	if (lambdasBitmap_.bitmap_size) {
		lambdasBitmap_.bitmap_word = new uint32_t[bm.bitmap_size];

		if (!lambdasBitmap_.bitmap_word) {
			G2PCERA_ERROR("Insufficient memory while allocating a"
				      "lambda bitmap");
			METOUT();
			BUG();
			assert(0);
			return false;
		}

		memcpy(lambdasBitmap_.bitmap_word,
		       bm.bitmap_word,
		       (bm.bitmap_size * 4));
	}

	METOUT();
	return true;
}

bool
PCERA::IscLscLambdaInfo::getLscWdmAvailBw(wdm_link_lambdas_bitmap_t &	   bm)
{
	METIN();

	memset(&bm, 0, sizeof(bm));

	bm.base_lambda_label = lambdasBitmap_.base_lambda_label;
	bm.num_wavelengths   = lambdasBitmap_.num_wavelengths;
	bm.bitmap_size       = lambdasBitmap_.bitmap_size;
	bm.bitmap_word       = 0;

	if (bm.bitmap_size) {
		bm.bitmap_word = new uint32_t[bm.bitmap_size];

		if (!bm.bitmap_word) {
			G2PCERA_ERROR("Insufficient memory while allocating a"
				      "lambda bitmap");
			METOUT();
			BUG();
			assert(0);
			return false;
		}

		memcpy(bm.bitmap_word,
		       lambdasBitmap_.bitmap_word,
		       (bm.bitmap_size * 4));
	}

	METOUT();
	return true;
}

bool
PCERA::IscLscLambdaInfo::fitInConstraints(const cspf_constr_t &	data)
{
	METIN();

	// XXX Add code here

	METOUT();
	return false;
}

bool
PCERA::IscLscLambdaInfo::operator==(const IscLscLambdaInfo & other) const
{
	size_t		i;
	sw_cap_t	sw1, sw2;
	enc_type_t	enc1, enc2;
	bw_per_prio_t	bw1, bw2;

	getIscTypes(sw1, enc1);
	getMaxLspBw(bw1);

	other.getIscTypes(sw2, enc2);
	other.getMaxLspBw(bw2);

	if (sw1 != sw2) {
		return false;
	}
	if (enc1 != enc2) {
		return false;
	}
	for (i = 0; i < MAX_BW_PRIORITIES; i++) {
		if (bw1.bw[i] != bw2.bw[i]) {
			return false;
		}
	}
	return true;
}

bool
PCERA::IscLscLambdaInfo::operator!=(const IscLscLambdaInfo & other) const
{
	return !(*this == other);
}














////////////////////////////
// Class G2PCERA
///////////////////////////
PCERA::G2PCERA::G2PCERA(void) :
	topology_(this)
{
	METIN();

	G2PCERA_DEBUG("Called G2PCERA constructor");

	spfSelector_ = spfDijkstra;
	rootNodeId_  = 0xFAACACAA;
	gnsCalls_.clear();

	METOUT();
}

PCERA::G2PCERA::G2PCERA(uint32_t		rootNodeId,
			spfType_t		spfSelector) :
	topology_(this)
{
	METIN();

	G2PCERA_DEBUG("Called G2PCERA constructor");

	assert(rootNodeId != 0);
	assert(spfSelector == spfDijkstra);

	spfSelector_ = spfSelector;
	rootNodeId_  = rootNodeId;
	gnsCalls_.clear();

	METOUT();
}

PCERA::G2PCERA::~G2PCERA(void)
{
	METIN();

	if (gnsCalls_.size()) {
		G2PCERA_DEBUG("Deleting gnsCalls...");
		std::map<gnsCallIdent_t, GnsCall *>::iterator iter;

		for (iter  = gnsCalls_.begin();
		     iter != gnsCalls_.end();
		     iter++) {
			assert((*iter).second);
			delete (*iter).second;
		}
	}

	METOUT();
}

uint32_t
PCERA::G2PCERA::rootNodeId(void)
{
	METIN();

	METOUT();
	return rootNodeId_;
}

spfType_t
PCERA::G2PCERA::spfSelector(void)
{
	METIN();

	METOUT();
	return spfSelector_;
}

PCERA::Topology &
PCERA::G2PCERA::getTopology(void)
{
	METIN();

	METOUT();
	return topology_;
}

bool
PCERA::G2PCERA::attachGnsCall(GnsCall * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GnsCall is NIL");
		METOUT();
		return false;
	}

	gnsCallIdent_t key;

	key = callIdMangle(ptr->ident());

	if (gnsCalls_.find(key) != gnsCalls_.end()) {
		G2PCERA_ERROR("An instance of gnsCall with the same id %s is "
			      "just present in G2PCERA... "
			      "detach it first",
			      key.c_str());
		METOUT();
		return false;
	}

	gnsCalls_[key] = ptr;
	ptr->reparent(this);

	METOUT();
	return true;
}

bool
PCERA::G2PCERA::detachGnsCall(GnsCall * ptr)
{
	METIN();

	if (!ptr) {
		G2PCERA_ERROR("Passed GnsCall is NIL");
		METOUT();
		return false;
	}

	gnsCallIdent_t key;

	key = callIdMangle(ptr->ident());

	if (gnsCalls_.find(key) == gnsCalls_.end()) {
		G2PCERA_ERROR("No instance of gnsCall with id %s is "
			      "present in G2PCERA...",
			      key.c_str());
		METOUT();
		return false;
	}

	if (gnsCalls_[key] != ptr) {
		G2PCERA_ERROR("A different instance of gnsCall with id %s "
			      "is present in G2PCERA...cannot detach",
			      key.c_str());
		METOUT();
		return false;
	}

	// XXX-FIXME: Act on topology to reset reservations

	gnsCalls_.erase(key);
	ptr->reparent(0);

	METOUT();
	return true;
}

PCERA::GnsCall *
PCERA::G2PCERA::getGnsCall(const call_ident_t & id)
{
	METIN();

	GnsCall *						tmp;
	gnsCallIdent_t						key;
	std::map<gnsCallIdent_t, GnsCall *>::const_iterator	iter;

	tmp  = 0;
	key  = callIdMangle(id);
	iter = gnsCalls_.find(key);

	if (iter != gnsCalls_.end()) {
		assert((*iter).second);
		tmp = (*iter).second;
	}

	METOUT();
	return tmp;
}

std::list<PCERA::GnsCall *>
PCERA::G2PCERA::getGnsCalls(void)
{
	METIN();

	std::list<GnsCall *> tmp;

	tmp.clear();

	std::map<gnsCallIdent_t, GnsCall *>::const_iterator iter;

	for (iter = gnsCalls_.begin(); iter != gnsCalls_.end(); iter++) {
		tmp.push_front((*iter).second);
	}

	METOUT();
	return tmp;
}

size_t
PCERA::G2PCERA::getGnsCallsCount(void)
{
	METIN();
	METOUT();
	return gnsCalls_.size();
}

PCERA::gnsCallIdent_t
PCERA::G2PCERA::callIdMangle(const call_ident_t & id)
{
	METIN();

	gnsCallIdent_t		tmp; // this is a string
	std::ostringstream	tmpStr;
	tmp.clear();


	tmpStr << "SRC:"  << addr_ntoa(id.src_addr)
	       << "-CID:" << key2string(id.local_id).c_str();

	tmp += tmpStr.str();

	METOUT();
	return tmp;
}
