//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#ifndef __G2PCERA_DM_HH__
#define __G2PCERA_DM_HH__

#include <list>
#include <map>
#include <iterator>
#include <string>
#include <assert.h>

#include "g2pcera_trace.hh"
#include "g2pcera_common.hh"
#include "g2pcera_dumper.hh"

//
// common utilities
//

namespace PCERA {
	// Forward declarations
	class G2PCERA;
	class Topology;
	class Node;
	class NetNode;
	class GridNode;
	class GridService;
	class GridCompElem;
	class GridSubCluster;
	class GridStorageElem;
	class Tna;
	class TeLink;
	class IscInfo;
	class IscPscPacketInfo;
	class IscTdmSdhSonetInfo;
	class IscLscG709Info;
	class IscLscLambdaInfo;
	class topologyVisitor;

	template <bool REP, class P> class Ancestor {
	public:
		Ancestor(P * p) {
			//assert(p);

			parent_      = p;
			reparenting_ = REP;
		}
		~Ancestor(void) {
			// mindless
			parent_ = 0;
		}

		// returns the parent
		P * parent(void) {
			return parent_;
		}

		// assign the new parent and returns the previous one
		P * reparent(P * p) {
			//assert(p);

			P * tmp;

			if (!reparenting_) {
				G2PCERA_ERROR("Reparenting not allowed!");
				return 0;
			}

			tmp = parent_;
			parent_ = p;

			return tmp;
		}

	private:
		P *	parent_;
		bool	reparenting_;
	};

	template <class DATA>
	class GridSubNode : public Ancestor<true, GridNode> {
	public:
		GridSubNode(grid_subnode_ident_t &	ident,
			    const DATA &		data) :
			Ancestor<true, GridNode> (0) {
			//assert(p);

			ident_ = ident;
			data_  = data;
		}

		~GridSubNode(void) {
			// mindless
		}

		grid_subnode_ident_t	ident(void) {
			return ident_;
		}

		bool			setData(const DATA & data) {
			data_ = data;
			return true;
		}

		bool			getData(DATA & data) {
			data = data_;
			return true;
		}

		void			dump(const std::string & prefix) {
			G2PCERA_DEBUG("%s"DUMP_BAR1, prefix.c_str());

			std::string prefix0 = prefix + DUMP_TAB;

			logDump(prefix0, ident());

			DATA data;
			if (!getData(data)) {
				G2PCERA_ERROR("Cannot get data "
					      "from object in %s",
					      __PRETTY_FUNCTION__);
				return;
			}

			logDump(prefix0, data);

			G2PCERA_DEBUG("%s"DUMP_BAR2, prefix.c_str());
		}


	private:
		grid_subnode_ident_t	ident_;
		DATA			data_;
	};

	template <class T>
	bool addQ(std::list<T> & q, T t, bool onTop)
	{
		typename std::list<T>::const_iterator iter;
		for (iter = q.begin(); iter != q.end(); iter++) {
			if (*iter == t) {
				return false;
			}
		}

		if (onTop) {
			q.push_front(t);
		} else {
			q.push_back(t);
		}

		return true;
	};

	template <class T>
	bool delQ(std::list<T> & q, T t)
	{
		typename std::list<T>::iterator iter;
		for (iter = q.begin(); iter != q.end(); iter++) {
			if (*iter == t) {
				break;
			}
		}

		if (iter == q.end()) {
			return false;
		}

		q.erase(iter);
		return true;
	};
}

//
//  GNS Call and Connection related data model elements
//  in the PCERA namespace
//
namespace PCERA {
	// map keys
	typedef std::string	gnsCallIdent_t;  // key

	class Connection {
	public:
		Connection(void);
		~Connection(void);

		void		dump(const std::string &	prefix,
				     bool			recursive);

		bool			addEroHop(bool			onTop,
						  const ero_hop_t &	hop);
		bool			delEroHop(ero_hop_t &		hop);

		void			flushEro(void);

		std::list<ero_hop_t>	getEro(void);

		bool		checkDisjoinness(Connection &	other,
						 disjointness_t	&	type,
						 disjointness_level_t &	level);

	private:
		std::list<ero_hop_t>	ero_;
	};

	class GnsCall: public Ancestor<true, G2PCERA> {
	public:
		GnsCall(G2PCERA *			parent,
			const call_ident_t &		ident,
			const net_res_spec_t &		src,
			const res_spec_t &		dst,
			const call_info_t &		callInfo,
			const recovery_info_t &		recInfo,
			const lsp_info_t &		lspInfo);
		GnsCall(G2PCERA *			parent,
			const call_ident_t &		ident,
			const ero_hop_t &		src,
			const ero_hop_t &		dst,
			const lsp_info_t &		lspInfo);
		~GnsCall(void);

		call_ident_t		ident(void);

		void			dump(const std::string &  prefix,
					     bool		  recursive);

		bool			route(void);
		bool			confirm(void);
		void			confirm(bool	flag);

		bool		addEroHopToConnection(lsp_role_t	role,
						      const ero_hop_t &	hop,
						      bool		onTop);
		bool		delEroHopFromConnection(lsp_role_t	role,
							ero_hop_t &	hop);

		std::list<ero_hop_t>	getConnectionEro(lsp_role_t role,
							 bool justMyDmn=false);
		std::list<ero_hop_t>	getGnsCallEro(void);

		net_res_spec_t		getGnsCallEgressTnaRes(void);

		disjointness_level_t	getDisjoinnessLevel(void);


		bool		match(const net_res_spec_t &		src,
				      const res_spec_t &		dst,
				      const call_info_t &		cInfo,
				      const recovery_info_t &		rInfo,
				      const lsp_info_t &		lInfo);
		bool		match(const ero_hop_t &			srcHop,
				      const ero_hop_t &			dstHop,
				      const lsp_info_t &		lInfo);
	private:
		bool		resolveSrcHopFromSrcInfo(void);
		bool		resolveEgressTnaFromDstInfo(void);
		bool		computeOneRoute(std::list<ero_hop_t> &	pEro);
		bool		resolveGnsCallEro(void);
		NetNode *	dijkstraRun(uint32_t &			eRid);
		bool		assembleEro(NetNode *			dst,
					    uint32_t			eRid,
					    std::list<ero_hop_t> &	ero);

		call_ident_t			ident_;
		net_res_spec_t			src_;
		res_spec_t			dst_;
		call_info_t			callInfo_;
		recovery_info_t			recInfo_;
		lsp_info_t			lspInfo_;

		Connection			primary_;
		Connection			secondary_;

		disjointness_level_t		disjoinnessLevel_;
		ero_hop_t			srcHop_;
		ero_hop_t			dstHop_;

		bool				isConfirmed_;

		std::list<ero_hop_t>		gnsCallEro_;
		net_res_spec_t			eTnaRes_;
	};


}

//
//  GNS topology related data model elements
//  in the PCERA namespace
//
namespace PCERA {

	// map keys
	typedef uint32_t	nodeKey_t;

	std::string		nodeKey2string(const nodeKey_t & key);
	nodeKey_t		nodeKeyMangle(const uint32_t & rid);
	uint32_t		nodeKeyDemangle(const nodeKey_t & key);

	class Topology: public Ancestor<true, G2PCERA> {
	public:
		Topology(G2PCERA *	parent);
		~Topology(void);

		bool			attachNode(Node *  ptr);
		bool			detachNode(Node *  ptr);
		Node *			getNode(node_ident_t	ident);
		GridNode *		getGridNode(uint32_t peRouterId);
		NetNode *		getAncestorNode(uint32_t routerId);

		std::list<Node *>	getNodes(void);
		std::list<node_ident_t>	getNodesIdents(void);
		bool			modTotNodesCount(uint32_t  howMany,
							 bool		add);
		bool			modTotLinksCount(uint32_t  howMany,
							 bool		add);
		bool			modTotTnasCount(uint32_t   howMany,
							bool		add);

		void		dump(const std::string &	prefix,
				     bool			recursive);

		bool		getData(topology_summary_data_t & data);

		uint32_t	spfRevision(void);
		void		spfRevision(uint32_t newRev);

		uint32_t	maxLinkCost(void);
		void		maxLinkCost(uint32_t newCost);

		void		accept(topologyVisitor &	vis);

		bool	transform(const std::list< std::list<ero_hop_t> > & multiXro,
				  bool forwardOrReverse = false);

		bool	transform(const std::list<ero_hop_t> &	xro,
				  bool				fwdOrRev,
				  std::string &			err);

	private:
		uint32_t				totNodes_;
		uint32_t				totLinks_;
		uint32_t				totTnas_;

		// Maximum link	cost in the topology
		uint32_t				maxLinkCost_;
		// Global SPF revision number
		uint32_t				spfRevision_;

		std::map<nodeKey_t, Node *>		nodes_;
	};

	class Node: public Ancestor<true, Topology>  {
	public:
		Node(topo_node_type_t	type,
		     uint32_t		id);
		~Node(void);

		topo_node_type_t	type(void);
		node_ident_t		ident(void);

		void			dump(const std::string & prefix);
		void			accept(topologyVisitor & vis);
	private:
		topo_node_type_t	type_;
		uint32_t		id_;
	};

	class NetNode: public Node {
	public:
		NetNode(uint32_t		id);
		~NetNode(void);

		void		dump(const std::string &	prefix,
				     bool			recursive);

		bool			setData(const net_node_data_t &	data);
		bool			getData(net_node_data_t & data);

		bool			attachOutLink(TeLink *	ptr);
		bool			detachOutLink(TeLink *	ptr);
		std::list<TeLink *>	getOutLinks(void);
		std::list<telink_ident_t>	getOutLinksIdents(void);
		size_t			getOutLinksCount(void);
		TeLink *	getOutLink(const telink_ident_t & id);
		TeLink *	getOutLink(const g2mpls_addr_t & id);

		bool			attachTna(Tna *	ptr);
		bool			detachTna(Tna *	ptr);
		std::list<tna_ident_t>	getTnasIdents(uint32_t	borderNodeId);
		size_t			getTnasCount(void);
		Tna *			getTna(const tna_ident_t & id);
		Tna *			getTna(const g2mpls_addr_t & ixd);
		bool		correlateAllTnas(uint32_t	   peRouterId,
						 PCERA::GridNode * ptr);
		bool			attachCandElems(NetNode * ptr);
		bool			detachCandElems(NetNode * ptr);
		std::list<NetNode *>	getCandElems(void);

		uint32_t		rootCost(void);
		void			rootCost(uint32_t	newCost);

		uint32_t		spfRevision(void);
		void			spfRevision(uint32_t	newRev);

		nodeFlags_t		nodeFlags(void);
		void			nodeFlags(nodeFlags_t & newFlags);

		bool			attachFwdLink(TeLink *  ptr);
		bool			detachFwdLink(TeLink *  ptr);
		std::list<TeLink *>	getFwdLinks(void);
		void			clearFwdLinks(void);

		bool			isDomain(void);
		bool			isUp(void);
		bool			isEnabled(void);

		bool			fitInConstraints(uint32_t	colors,
							 uint16_t	area);

		void			accept(topologyVisitor &	vis);
	private:
		bool				isDomain_;
		opstate_t			opState_;
		admstate_t			admState_;
		uint32_t			teColors_;
		std::list<uint16_t>		areas_;

		std::list<Tna *>		tnas_;
		std::list<TeLink *>		links_; // outgoing links

		// Path Computation related members
		std::list<NetNode *>		candidateElems_;
		uint32_t			rootCost_;
		uint32_t			spfRevision_;
		nodeFlags_t			nodeFlags_;	// bitmask
		std::list<TeLink *>		fwdLinks_;	// cand. links

	};

#ifdef G2MPLS
	class GridNode: public Node {
	public:
		GridNode(uint32_t		id);
		~GridNode(void);

		void		dump(const std::string &	prefix,
				     bool			recursive);

		bool			setData(const grid_site_data_t & data);
		bool			getData(grid_site_data_t & data);
		uint32_t		getPeRouterId(void);
		grid_subnodes_t		getSubnodesIdents(void);

		bool			attachGridService(GridService *  ptr);
		bool			detachGridService(GridService *  ptr);
		GridService *		getGridService(void /* policy*/);
		GridService *		getGridService(uint32_t  id);

		bool			attachGridCompElem(GridCompElem * ptr);
		bool			detachGridCompElem(GridCompElem * ptr);
		GridCompElem *		getGridCompElem(void /* pol.*/);
		GridCompElem *		getGridCompElem(uint32_t  id);


		bool		attachGridSubCluster(GridSubCluster * ptr);
		bool		detachGridSubCluster(GridSubCluster * ptr);
		GridSubCluster *	getGridSubCluster(void /* po.*/);
		GridSubCluster *	getGridSubCluster(uint32_t  id);

		bool		attachGridStorageElem(GridStorageElem * ptr);
		bool		detachGridStorageElem(GridStorageElem * ptr);
		GridStorageElem *	getGridStorageElem(void /* po*/);
		GridStorageElem *	getGridStorageElem(uint32_t	id);

		uint32_t		match(const grid_res_spec_t & gnsRes);

		void			accept(topologyVisitor &	vis);
	private:
		std::string *				name_;
		geo_coords_t				location_;
		uint32_t				peRouterId_;

		std::map<uint32_t, GridService *>	gridServices_;
		std::map<uint32_t, GridCompElem *>	gridCompElems_;
		std::map<uint32_t, GridSubCluster *>	gridSubClusters_;
		std::map<uint32_t, GridStorageElem *>	gridStorageElems_;
	};


	class GridService: public GridSubNode<grid_service_data_t> {
	public:
		GridService(grid_subnode_ident_t &	subNodeId,
			    const grid_service_data_t &	data) :
			GridSubNode<grid_service_data_t>(subNodeId, data) {};
		~GridService(void) {};

		void	accept(topologyVisitor &	vis);
	};

	class GridCompElem: public GridSubNode<grid_ce_data_t>  {
	public:
		GridCompElem(grid_subnode_ident_t &	subNodeId,
			     const grid_ce_data_t &	data) :
			GridSubNode<grid_ce_data_t>(subNodeId, data) {};
		~GridCompElem(void) {};

		void	accept(topologyVisitor &	vis);
	};

	class GridSubCluster: public GridSubNode<grid_subcluster_data_t>  {
	public:
		GridSubCluster(grid_subnode_ident_t &	subNodeId,
			       const grid_subcluster_data_t &	data) :
			GridSubNode<grid_subcluster_data_t>(subNodeId, data){};
		~GridSubCluster(void) {};

		void	accept(topologyVisitor &	vis);
	};

	class GridStorageElem: public GridSubNode<grid_se_data_t>  {
	public:
		GridStorageElem(grid_subnode_ident_t &	subNodeId,
				const grid_se_data_t &	data) :
			GridSubNode<grid_se_data_t>(subNodeId, data) {};
		~GridStorageElem(void) {};

		void	accept(topologyVisitor &	vis);
	};
#endif // G2MPLS

	class Tna: public Ancestor<true, NetNode> {
	public:
		Tna(const tna_ident_t &	ident);
		~Tna(void);

		 tna_ident_t	ident(void);
		void		setRemNode(GridNode *		ptr);
		bool		getRemNodeIdent(node_ident_t &	ident);
		bool		correlate(void);
		bool		uncorrelate(void);
		void		dump(const std::string &	prefix);

		void		accept(topologyVisitor &	vis);
	private:
		tna_ident_t				ident_;
		// used for algorithm purposes
		GridNode *				remNode_;
	};

	class TeLink: public Ancestor<true, NetNode> {
	public:
		TeLink(telink_ident_t	id);
		~TeLink(void);

		telink_ident_t	ident(void);

		void		dump(const std::string &	prefix);

		bool		setData(const telink_com_data_t &	data);
		bool		getData(telink_com_data_t &	data);

		NetNode *	getRemNode(void);

		void		setRemLink(TeLink * ptr);
		TeLink *	getRemLink(void);

		bool		correlate(void);
		bool		uncorrelate(void);

		bool		setStates(const opstate_t &		op,
					  const admstate_t &		adm);
		bool		getStates(opstate_t &		op,
					  admstate_t &		adm);

		bool		attachIscInfo(IscInfo *		ptr);
		bool		detachIscInfo(IscInfo *		ptr);
		bool		getIscs(std::list<isc_t> &	iscs);

		IscPscPacketInfo *	getIscPscPacket(void);
		IscTdmSdhSonetInfo *	getIscTdmSdhSonet(void);
		IscLscG709Info *	getIscLscG709(void);
		IscLscLambdaInfo *	getIscLscLambda(void);

		bool		setGenAvailBw(const bw_per_prio_t & bw);
		bool		getGenAvailBw(bw_per_prio_t & bw);

		bool		appendSrlgs(const std::list<uint32_t> &	srlgs);
		bool		removeSrlgs(std::list<uint32_t> &	srlgs);
		bool		getSrlgs(std::list<uint32_t> &	srlgs);

		bool		appendCalEvents(const std::map<uint32_t,
						bw_per_prio_t> &	cal);
		bool		removeCalEvents(std::map<uint32_t,
						bw_per_prio_t> &	cal);
		bool		getCalEvents(std::map<uint32_t,
					     bw_per_prio_t> &		cal);

		bool		fitInConstraints(const cspf_constr_t & data);
		bool		fitInConstraints(const lsp_info_t & data);

		int64_t		linkCost(void);
		void		linkCost(int64_t		newCost);

		int64_t		generateCost(const lsp_info_t & data);

		bool		isUp(void);
		bool		isEnabled(void);

		void		accept(topologyVisitor &	vis);
	private:
		telink_ident_t				ident_;

		// adminMetric_ is the base OSPF link metric
		uint32_t				adminMetric_;

		uint32_t				teMetric_;
		uint32_t				teColorMask_;
		uint8_t					teProtectionTypeMask_;
		uint32_t				teMaxBw_;
		uint32_t				teMaxResvBw_;

		opstate_t				opState_;
		admstate_t				admState_;

		bw_per_prio_t				teAvailBw_;
		std::list<uint32_t>			teSrlgs_;
		std::map<uint32_t, bw_per_prio_t>	teLinkCalendar_;

		std::list<IscInfo *>			teIscsInfo_;

		// used for algorithm purposes
		TeLink *				reverseLink_;
		int64_t					linkCost_;
	};

	class IscInfo {
	public:
		IscInfo(sw_cap_t			switching,
			enc_type_t			encoding,
			const bw_per_prio_t &		maxLspBw);
		~IscInfo();

		void		dump(const std::string &	prefix);

		void		getIscTypes(sw_cap_t &	 switching,
					    enc_type_t & encoding) const;

		bool		getMaxLspBw(bw_per_prio_t &	bw) const;

		bool		fitInConstraints(const cspf_constr_t &	data);

		bool		operator==(const IscInfo & other) const;
		bool		operator!=(const IscInfo & other) const;

	private:
		sw_cap_t		switching_;
		enc_type_t		encoding_;

		bw_per_prio_t		maxLspBw_;
	};

	class IscPscPacketInfo: public IscInfo {
	public:
		IscPscPacketInfo(sw_cap_t		switching,
				 enc_type_t		encoding,
				 const bw_per_prio_t &	maxLspBw,
				 uint32_t		minLspBw,
				 uint16_t		ifMtu);
		~IscPscPacketInfo(void);

		void		dump(const std::string &	prefix);

		bool		getData(telink_psc_data_t &		data);

		bool		fitInConstraints(const cspf_constr_t &	data);

		bool	operator==(const IscPscPacketInfo & other) const;
		bool	operator!=(const IscPscPacketInfo & other) const;

	private:
		uint32_t		minLspBw_;
		uint16_t		ifMtu_;
	};

	class IscTdmSdhSonetInfo: public IscInfo {
	public:
		IscTdmSdhSonetInfo(sw_cap_t			switching,
				   enc_type_t			encoding,
				   const bw_per_prio_t &	maxLspBw,
				   uint32_t			minLspBw,
				   gmpls_sdhsonet_stdarbcap_t	stdArbConc);

		~IscTdmSdhSonetInfo(void);

		void		dump(const std::string &	prefix);

		bool		setData(const telink_tdm_data_t &	data);
		bool		getData(telink_tdm_data_t &		data);

		bool		setTdmAvailBw(const std::list<uint32_t> & fts);
		bool		getTdmAvailBw(std::list<uint32_t> &	  fts);

		bool		fitInConstraints(const cspf_constr_t &	data);

		bool	operator==(const IscTdmSdhSonetInfo & other) const;
		bool	operator!=(const IscTdmSdhSonetInfo & other) const;

	private:
		uint32_t			minLspBw_;
		gmpls_sdhsonet_stdarbcap_t	stdArbConc_;

		uint8_t				hoMuxCapMask_;
		uint8_t				loMuxCapMask_;
		uint32_t			transparencyMask_;
		uint32_t			blsrRingId_;

		std::list<uint32_t>		freeTimeslots_;
	};

	class IscLscG709Info: public IscInfo {
	public:
		IscLscG709Info(sw_cap_t			switching,
			       enc_type_t		encoding,
			       const bw_per_prio_t &	maxLspBBw);
		~IscLscG709Info(void);

		void		dump(const std::string &	prefix);

		bool		setData(const telink_lscg709_data_t &	data);
		bool		getData(telink_lscg709_data_t &		data);

		bool	setLscG709AvailBw(const std::list<uint32_t> &	foduk,
					  const std::list<uint32_t> &	foch);
		bool		getLscG709AvailBw(std::list<uint32_t> &	foduk,
						  std::list<uint32_t> &	foch);
		bool		fitInConstraints(const cspf_constr_t &	data);

		bool	operator==(const IscLscG709Info & other) const;
		bool	operator!=(const IscLscG709Info & other) const;

	private:
		uint8_t				odukMuxCapMask_;

		std::list<uint32_t>		unallocODUk_;
		std::list<uint32_t>		unallocOCh_;
	};

	class IscLscLambdaInfo: public IscInfo {
	public:
		IscLscLambdaInfo(sw_cap_t			switching,
				 enc_type_t			encoding,
				 const bw_per_prio_t &		maxLspBBw);
		~IscLscLambdaInfo(void);

		void		dump(const std::string &	prefix);

		bool		setData(const telink_lscwdm_data_t &	data);
		bool		getData(telink_lscwdm_data_t &		data);

		bool	setLscWdmAvailBw(const wdm_link_lambdas_bitmap_t & bm);
		bool	getLscWdmAvailBw(wdm_link_lambdas_bitmap_t &	   bm);

		bool		fitInConstraints(const cspf_constr_t &	data);

		bool	operator==(const IscLscLambdaInfo & other) const;
		bool	operator!=(const IscLscLambdaInfo & other) const;

	private:
		uint32_t			dispersionPMD_;
		uint32_t			spanLength_;
		std::list<wdm_amplifier_data_t>	amplifiers_;
		wdm_link_lambdas_bitmap_t	lambdasBitmap_;
	};
}



//
//  Main G2PCERA instance in the PCERA namespace
//
namespace PCERA {
	class G2PCERA {
	public:
		G2PCERA(void);
		G2PCERA(uint32_t	rootNodeId,
			spfType_t	spfSelector = spfDijkstra);

		~G2PCERA(void);

		uint32_t		rootNodeId(void);
		spfType_t		spfSelector(void);

	        Topology &		getTopology(void);

		bool			attachGnsCall(GnsCall *		ptr);
		bool			detachGnsCall(GnsCall *		ptr);
		GnsCall *	getGnsCall(const call_ident_t &	id);
		std::list<GnsCall *>	getGnsCalls(void);
		size_t			getGnsCallsCount(void);

	private:
		gnsCallIdent_t		callIdMangle(const call_ident_t & id);

		spfType_t				spfSelector_;
		uint32_t				rootNodeId_;
		Topology				topology_;
		std::map<gnsCallIdent_t, GnsCall *>	gnsCalls_;
	};
}

#endif // __G2PCERA_DM_HH__
