//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"

#include "g2pcerad.hh"
#include "g2pcera_api.hh"
#include "g2pcera_vty_utils.hh"

#include <map>
#include <list>
#include <string>
#include <sstream>

using namespace std;

static bool get_telink_ident(struct vty *	vty,
			     const char **	argv,
			     const int		argc,
			     telink_ident_t &	ident)
{
	ipv4_t		ipAddr;
	uint32_t	preflen;

	memset(&ident, 0, sizeof(ident));

	ident.mode = LINK_MODE_P2P_NUMBERED;

	// local router id
	if (strchr(argv[0], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[0], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[0], VTY_NEWLINE);
			return false;
		}

		ident.lcl_rid = ntohl(ipAddr.s_addr);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[0], VTY_NEWLINE);
		return false;
	}

	// local router id + preflen
	if (strchr(argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return false;
		}
		ident.lcl_id.type       = IPv4;
		ident.lcl_id.value.ipv4 = ipAddr;
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[1], VTY_NEWLINE);
		return false;
	}

	preflen = strtoul(argv[2], NULL, 10);
	ident.lcl_id.preflen = ((preflen > 32) ? 32 : (uint8_t) preflen);

	// remote router id
	if (strchr(argv[3], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[3], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[3], VTY_NEWLINE);
			return false;
		}

		ident.rem_rid = ntohl(ipAddr.s_addr);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[3], VTY_NEWLINE);
		return false;
	}

	// remote id + preflen
	if (strchr(argv[4], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[4], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[4], VTY_NEWLINE);
			return false;
		}
		ident.rem_id.type       = IPv4;
		ident.rem_id.value.ipv4 = ipAddr;
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[4], VTY_NEWLINE);
		return false;
	}
	preflen = strtoul(argv[5], NULL, 10);
	ident.rem_id.preflen = ((preflen > 32) ? 32 : (uint8_t) preflen);

	return true;
}

//
//
//
//   VTY commands
//
//
//

static uint32_t g2pceraId = 0;

#define VTY_TAB		"    "

struct cmd_node g2pcera_node = {
	G2PCERA_NODE,
	"%s(config-g2pcera)# ",
	1
};


DEFUN(g2pcera_load_topology,
      g2pcera_load_topology_cmd,
      "g2pcera load topology-file WORD",
      "Load a G2PCERA instances and topologies by file\n")
{
	std::string		err;
	std::string		filename;
	pceraErrorCode_t	ret;

	filename = argv[0];

	ret = PCERA_API::g2pceraLoadFromFile(filename, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot load G2.PCE-RA instance and related topogogy "
			"from file %s "
			"(err: code=%s, reason='%s')%s",
			filename.c_str(),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"G2.PCE-RA instances and related topologies successfully "
		"loaded from %s !%s",
		filename.c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera,
      g2pcera_cmd,
      "g2pcera root-router-id A.B.C.D",
      "Enable a G2PCERA instance with a root-router-id\n")
{
	std::string		err;
	pceraErrorCode_t	ret;

	vty->node  = G2PCERA_NODE;
	vty->index = NULL;//rsvp_get();


	if (argc == 0) {
		ret = PCERA_API::g2pceraStart(err);
	} else {
		ipv4_t rootId;

		if (strchr(argv[0], '.') != NULL) {
			// matched "A.B.C.D".
			if (!inet_aton(argv[0], &(rootId))) {
				vty_out(vty,
					"Invalid IP address (%s) "
					"for root-router-id%s",
					argv[0], VTY_NEWLINE);
				return CMD_WARNING;
			}
		} else {
			vty_out(vty,
				"Invalid IP address (%s) for root-router-id%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}

		g2pceraId = ntohl(rootId.s_addr);

		ret = PCERA_API::g2pceraStart(g2pceraId, err);
	}

	if (ret == PCERA_ERROR_INSTANCE_ALREADY_EXISTS) {
		vty_out(vty,
			"Got existing G2.PCE-RA instance%s",
			VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot start G2.PCE-RA instance "
			"(err: code=%s, reason='%s')%s",
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!g2pceraId) {
		std::list<uint32_t> pces;

		pces = PCERA_API::g2pceraGetAll();

		g2pceraId = pces.front();
	}

	vty_out(vty,
		"G2.PCE-RA instance started !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(g2pcera,
      g2pcera_cmpt_cmd,
      "g2pcera",
      "Enable default G2PCERA instance\n");

DEFUN(no_g2pcera,
      no_g2pcera_cmd,
      "no g2pcera root-router-id A.B.C.D",
      "Stop/disable a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;

	vty->node  = G2PCERA_NODE;
	vty->index = NULL;

	if (argc == 0) {
		ret = PCERA_API::g2pceraStop(err);
	} else {
		ipv4_t rootId;

		if (strchr(argv[0], '.') != NULL) {
			// matched "A.B.C.D".
			if (!inet_aton(argv[0], &(rootId))) {
				vty_out(vty,
					"Invalid IP address (%s) "
					"for root-router-id%s",
					argv[0], VTY_NEWLINE);
				return CMD_WARNING;
			}
		} else {
			vty_out(vty,
				"Invalid IP address (%s) for root-router-id%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}

		g2pceraId = ntohl(rootId.s_addr);

		ret = PCERA_API::g2pceraStop(g2pceraId, err);
	}

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot stop G2.PCE-RA instance "
			"(err: code=%s, reason='%s')%s",
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"G2.PCE-RA instance stopped !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(no_g2pcera,
      no_g2pcera_cmpt_cmd,
      "no g2pcera",
      "Stop/disable default G2PCERA instance\n");

DEFUN(g2pcera_topo_show,
      g2pcera_topo_show_cmd,
      "show g2pcera root-router-id A.B.C.D (summary|details)",
      "Show configured G2.PCE-RA instance")
{

	std::string		err;
	pceraErrorCode_t	ret;
	topology_summary_data_t	data;
	bool			details;
	ipv4_t			rootId;

	details = false;

	if (strchr(argv[0], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[0], &(rootId))) {
			vty_out(vty,
				"Invalid IP address (%s) "
				"for root-router-id%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for root-router-id%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (argc > 1) {
		if (strcmp(argv[1], "summary") == 0) {
			details = false;
		} else if (strcmp(argv[1], "details") == 0) {
			details = true;
		} else {
			vty_out(vty,
				"Invalid detail argument (%s) %s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	g2pceraId = ntohl(rootId.s_addr);

	ret = PCERA_TOPO_API::topologyGet(g2pceraId, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot get topology data from G2.PCE-RA instance "
			"(err: code=%s, reason='%s')%s",
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "G2.PCE-RA INSTANCE%s", VTY_NEWLINE);
	vty_out(vty,
		"%s- root router id             : %s%s",
		VTY_TAB,
		numb_htoa(data.root_rid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total nodes                : %d%s",
		VTY_TAB,
		data.tot_nodes,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total TE links             : %d%s",
		VTY_TAB,
		data.tot_links,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total TNAs                 : %d%s",
		VTY_TAB,
		data.tot_tnas,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total GNS calls computed   : %d%s",
		VTY_TAB,
		data.tot_gns_calls,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- algorithm info%s",
		VTY_TAB,
		VTY_NEWLINE);
	vty_out(vty,
		"%s%s- SPF algorithm          : %s%s",
		VTY_TAB, VTY_TAB,
		SPF_TYPE_TO_STRING(data.spf_selector),
		VTY_NEWLINE);
	vty_out(vty,
		"%s%s- max TE link cost       : %d%s",
		VTY_TAB, VTY_TAB,
		data.max_link_cost,
		VTY_NEWLINE);
	vty_out(vty,
		"%s%s- SPF revision           : %d%s",
		VTY_TAB, VTY_TAB,
		data.spf_revision,
		VTY_NEWLINE);


	std::list<node_ident_t> tmp;

	tmp.clear();
	if (details) {
		tmp = PCERA_TOPO_API::nodeGetAll(g2pceraId, err);
	}

	if (tmp.size()) {
		vty_out(vty,
			"%s- nodes list%s",
			VTY_TAB,
			VTY_NEWLINE);

		std::list<node_ident_t>::const_iterator iter;
		for (iter = tmp.begin(); iter != tmp.end(); iter++) {
			vty_out(vty,
				"%s%s- %s%s",
				VTY_TAB, VTY_TAB,
				nodeIdent2string(*iter).c_str(),
				VTY_NEWLINE);
		}
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_topo_show,
      g2pcera_topo_show_cmpt_cmd,
      "show g2pcera root-router-id A.B.C.D",
      "Show configured G2.PCE-RA instance");

DEFUN(g2pcera_show,
      g2pcera_show_cmd,
      "show g2pcera",
      "Show all G2.PCE-RA instances")
{
	size_t					i;
	std::list<uint32_t>			pces;
	std::list<uint32_t>::const_iterator	iter;

	pces = PCERA_API::g2pceraGetAll();

	vty_out(vty, "G2.PCE-RA PROCESS%s", VTY_NEWLINE);
	vty_out(vty,
		"%s- instances (%d)%s",
		VTY_TAB,
		pces.size(),
		VTY_NEWLINE);

	for (i = 0, iter = pces.begin(); iter != pces.end(); iter++) {
		i++;
		vty_out(vty,
			"%s%s- [#%02d]: root router ID: %s%s",
			VTY_TAB,VTY_TAB,
			i, numb_htoa(*iter),
			VTY_NEWLINE);
	}

	if (!i) {
		vty_out(vty,
			"%s%s***EMPTY***%s",
			VTY_TAB,VTY_TAB,
			VTY_NEWLINE);
	}

	return CMD_SUCCESS;
}


DEFUN(g2pcera_node_add,
      g2pcera_node_add_cmd,
      "node add (network|grid-site) id (A.B.C.D|WORD)",
      "Add a node to a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	node_ident_t		ident;

	if (strcmp(argv[0], "network") == 0) {
		ident.type = TOPO_NODE_NETWORK;
	} else if (strcmp(argv[0], "grid-site") == 0) {
		ident.type = TOPO_NODE_GRID;
	} else {
		vty_out(vty,
			"Invalid node type (%s)%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr(argv[1], '.') != NULL) {
		ipv4_t ipAddr;

		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ident.id = ntohl(ipAddr.s_addr);
	} else {
		// match "<0-4294967295>".
		ident.id = strtoul(argv[1], NULL, 16);
	}

	ret = PCERA_TOPO_API::nodeAdd(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot add node into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_node_del,
      g2pcera_node_del_cmd,
      "node del (network|grid-site) id (A.B.C.D|WORD)",
      "Delete a node from a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	node_ident_t		ident;

	if (strcmp(argv[0], "network") == 0) {
		ident.type = TOPO_NODE_NETWORK;
	} else if (strcmp(argv[0], "grid-site") == 0) {
		ident.type = TOPO_NODE_GRID;
	} else {
		vty_out(vty,
			"Invalid node type (%s)%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr(argv[1], '.') != NULL) {
		ipv4_t ipAddr;

		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ident.id = ntohl(ipAddr.s_addr);
	} else {
		// match "<0-4294967295>".
		ident.id = strtoul(argv[1], NULL, 16);
	}

	ret = PCERA_TOPO_API::nodeDel(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot delete node from topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_node_netup,
      g2pcera_node_netup_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD WORD WORD WORD WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	uint32_t		id;
	net_node_data_t		data;

	memset(&data.mask_, 0, sizeof (data.mask_));
	data.areas.clear();

	if (strchr(argv[0], '.') != NULL) {
		ipv4_t ipAddr;

		// matched "A.B.C.D".
		if (!inet_aton(argv[0], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}

		id = ntohl(ipAddr.s_addr);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp(argv[1], "router") == 0) {
		data.is_domain = 0;
	} else if (strcmp(argv[1], "domain") == 0) {
		data.is_domain = 1;
	} else {
		vty_out(vty,
			"Invalid network node type (%s)%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(data.mask_, is_domain);

	if (strcmp(argv[2], "up") == 0) {
		data.op_state = UP;
	} else if (strcmp(argv[2], "down") == 0) {
		data.op_state = DOWN;
	} else {
		vty_out(vty,
			"Invalid network node op-state (%s)%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(data.mask_, op_state);

	if (strcmp(argv[3], "enabled") == 0) {
		data.adm_state = ENABLED;
	} else if (strcmp(argv[3], "disabled") == 0) {
		data.adm_state = DISABLED;
	} else {
		vty_out(vty,
			"Invalid network node adm-state (%s)%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(data.mask_, adm_state);

	if (argc > 4) {
		data.te_colors = strtoul(argv[4], NULL, 16);
		BITMASK_BITSET(data.mask_, te_colors);
	}

	if (argc > 5) {
		size_t i;

		BITMASK_BITSET(data.mask_, areas);

		for (i = 5; i < argc; i++) {
			data.areas.push_back(strtoul(argv[i], NULL, 10));
		}
	}

	ret = PCERA_TOPO_API::netNodeUpdate(g2pceraId, id, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update network node info into topology "
			"for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt1_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)",
      "Update network node info into G2PCERA instance\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt2_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD",
      "Update network node info into G2PCERA instance\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt3_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt4_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt5_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD WORD WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt6_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD WORD WORD WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n");
ALIAS(g2pcera_node_netup,
      g2pcera_node_netup_cmpt7_cmd,
      "node up network id A.B.C.D (router|domain) "
      "op-state (up|down) adm-state (enabled|disabled)"
      "te-colors WORD area WORD WORD WORD WORD WORD",
      "Update network node info into G2PCERA instance (up to 5 areas)\n");

DEFUN(g2pcera_node_gsup,
      g2pcera_node_gsup_cmd,
      "node up grid-site id WORD pe-router-id A.B.C.D "
      "name WORD latitude WORD longitude WORD",
      "Update grid-site node info into G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	uint32_t		id;
	grid_site_data_t	data;

	memset(&data.mask_, 0, sizeof (data.mask_));
	data.name = 0;

	id = strtoul(argv[0], NULL, 16);

	if (strchr(argv[1], '.') != NULL) {
		ipv4_t ipAddr;

		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}

		data.pe_router_id = ntohl(ipAddr.s_addr);
		BITMASK_BITSET(data.mask_, pe_router_id);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (argc > 2) {
		data.name = new std::string(argv[2]);
		BITMASK_BITSET(data.mask_, name);
	}

	if (argc > 3) {
		double val;
		val = strtod(argv[3], NULL);
		data.location.lat_resolution = 4;
		data.location.latitude       = (*((uint64_t *)(&(val))));

		val = strtod(argv[4], NULL);
		data.location.lon_resolution = 4;
		data.location.longitude      = (*((uint64_t *)(&(val))));
		BITMASK_BITSET(data.mask_, location);
	}

	ret = PCERA_TOPO_API::gridSiteUpdate(g2pceraId, id, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update grid-site node info into topology "
			"for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_node_gsup,
      g2pcera_node_gsup_cmpt1_cmd,
      "node up grid-site id WORD pe-router-id A.B.C.D ",
      "Update grid-site node info into G2PCERA instance\n");

ALIAS(g2pcera_node_gsup,
      g2pcera_node_gsup_cmpt2_cmd,
      "node up grid-site id WORD pe-router-id A.B.C.D "
      "name WORD",
      "Update grid-site node info into G2PCERA instance\n");

static void set_dataservice(grid_service_data_t & dd)
{
	dd.name        = new string("DDSS-Service");
	dd.data.type   = SERVICE_OTHER; //(grid_service_type_t) myrand(1, 24);

	dd.data.version.major_revision = 1; //myrand(0, 2);
	dd.data.version.minor_revision = 0; //myrand(0, 99);
	dd.data.version.build_fix      = 0; //myrand(1, 9999);

	dd.state                       = SERVICE_STATE_OK;
	//(grid_service_state_t) myrand(0, 3);

	dd.endpoint_addr.type          = IPv4;
	dd.endpoint_addr.preflen       = 32;
	inet_aton("192.168.150.115", &(dd.endpoint_addr.value.ipv4));

	BITMASK_BITSET(dd.mask_, name);
	BITMASK_BITSET(dd.mask_, data);
	BITMASK_BITSET(dd.mask_, state);
	BITMASK_BITSET(dd.mask_, endpoint_addr);
}

static void set_datacomputing(grid_ce_data_t & dd)
{
	dd.name           = new string("ph-CompElem");
	dd.lrms_info.type = (grid_lrms_t) myrand(1, 11);

	dd.lrms_info.version.major_revision = myrand(0, 2);
	dd.lrms_info.version.minor_revision = myrand(0, 99);
	dd.lrms_info.version.build_fix      = myrand(1, 9999);

	dd.host_addr.type              = IPv4;
	dd.host_addr.preflen           = 32;
	dd.host_addr.value.ipv4.s_addr = 0xCACACACA;

	dd.gatekeeper_port          = myrand(100, 999);
	dd.job_manager              = new string("phosphorus-mss");
	dd.data_dir                 = new string("/etc/init.d/phosphorus-mss");
	dd.default_storage_elem_id  = 0x666;

	dd.jobs_state.free_job_slots                     =  myrand(0,100);
	dd.jobs_state.state = (grid_cese_state_t) myrand(1,4);
	dd.jobs_stats.running_jobs                       = myrand(0,100);
	dd.jobs_stats.waiting_jobs                       = myrand(0,100);
	dd.jobs_stats.total_jobs                         = myrand(0,100);

	dd.jobs_timeperf.estimated_response_time         = myrand(1000,2000);
	dd.jobs_timeperf.worst_response_time             = myrand(1800,2000);

	dd.jobs_timepolicy.max_wallclock_time            = myrand(1900,1950);
	dd.jobs_timepolicy.max_obtainable_wallclock_time = myrand(1900,1950);
	dd.jobs_timepolicy.max_cpu_time                  = myrand(1900,1950);
	dd.jobs_timepolicy.max_obtainable_cpu_time       = myrand(1900,1950);

	dd.jobs_loadpolicy.max_total_jobs    = myrand(100,100);
	dd.jobs_loadpolicy.max_running_jobs  = myrand(75 , 80);
	dd.jobs_loadpolicy.max_waiting_jobs  = myrand(20 , 25);
	dd.jobs_loadpolicy.assigned_job_slots= myrand(3  , 5);
	dd.jobs_loadpolicy.max_slots_per_job = myrand(6  , 10);
	dd.jobs_loadpolicy.priority          = myrand(0, 255);
	dd.jobs_loadpolicy.preemption_flag   = myrand(0,1);

	size_t i;
	for (i = 0; i < 3; i++) {
		struct tm * bdTime;
		time_t      unixTime;

		time(&unixTime);
		bdTime = localtime(&unixTime);

		bdTime->tm_year  = 108;
		bdTime->tm_mon   = i;
		bdTime->tm_mday  = 1 + (i * 14);
		bdTime->tm_hour  = 0;
		bdTime->tm_min   = 0;
		bdTime->tm_sec   = 0;
		bdTime->tm_isdst = -1;

		unixTime = mktime(bdTime);

		dd.free_job_slots_calendar[unixTime] = i * 50;
	}

	BITMASK_BITSET(dd.mask_, name);
	BITMASK_BITSET(dd.mask_, lrms_info);
	BITMASK_BITSET(dd.mask_, host_addr);
	BITMASK_BITSET(dd.mask_, gatekeeper_port);
	BITMASK_BITSET(dd.mask_, job_manager);
	BITMASK_BITSET(dd.mask_, data_dir);
	BITMASK_BITSET(dd.mask_, default_storage_elem_id);
	BITMASK_BITSET(dd.mask_, jobs_state);
	BITMASK_BITSET(dd.mask_, jobs_stats);
	BITMASK_BITSET(dd.mask_, jobs_timeperf);
	BITMASK_BITSET(dd.mask_, jobs_timepolicy);
	BITMASK_BITSET(dd.mask_, jobs_loadpolicy);
	BITMASK_BITSET(dd.mask_, free_job_slots_calendar);
}

static void set_datasubcluster(grid_subcluster_data_t & dd)
{
	dd.name               = new string("ph-SubCluster");
	dd.cpu.count.physical = myrand(1, 50);
	dd.cpu.count.logical  = myrand(1, 100);
	dd.cpu.arch           = (grid_cpu_arch_t) myrand(1, 9);
	dd.os.type            = (grid_os_type_t) myrand(1, 67);
	dd.os.version.major_revision  = myrand(2, 10);
	dd.os.version.minor_revision  = myrand(0, 10);
	dd.os.version.build_fix       = myrand(1, 9999);
	dd.memory.ram_size            = myrand(1, 4) * 1000;
	dd.memory.virtual_memory_size = myrand(1, 2) * 1000;

	size_t i;
	for (i = 0; i < 3; i++) {

		grid_software_package_t * tmp;
		tmp = new grid_software_package_t;
		if (!tmp) {
			assert(0);
		}
		tmp->software.type = (grid_appname_t) myrand(1, 5);
		tmp->software.version.major_revision  = myrand(2, 10);
		tmp->software.version.minor_revision  = myrand(0, 10);
		tmp->software.version.build_fix       = myrand(1, 9999);
		tmp->software_env_setup = new string("/etc/conf.d/tapioca/");
		dd.software_packages.push_front(tmp);
	}

	for (i = 0; i < 3; i++) {
		struct tm * bdTime;
		time_t      unixTime;

		time(&unixTime);
		bdTime = localtime(&unixTime);

		bdTime->tm_year  = 108;
		bdTime->tm_mon   = i;
		bdTime->tm_mday  = 1 + (i * 14);
		bdTime->tm_hour  = 0;
		bdTime->tm_min   = 0;
		bdTime->tm_sec   = 0;
		bdTime->tm_isdst = -1;

		unixTime = mktime(bdTime);

		dd.subcluster_calendar[unixTime].physical = myrand(0, i*100);
		dd.subcluster_calendar[unixTime].logical  = myrand(0, i*100);
	}

	BITMASK_BITSET(dd.mask_, cpu);
	BITMASK_BITSET(dd.mask_, name);
	BITMASK_BITSET(dd.mask_, os);
	BITMASK_BITSET(dd.mask_, memory);
	BITMASK_BITSET(dd.mask_, software_packages);
	BITMASK_BITSET(dd.mask_, subcluster_calendar);
}

static void set_datastorage(grid_se_data_t & dd)
{
	dd.name               = new string("DDSS-StoElem");
	dd.storage_info.arch  = STORAGE_DISK;
	//(grid_storage_arch_t) myrand(1,3);
	dd.storage_info.state = CESE_STATE_PRODUCTION;
	//(grid_cese_state_t) myrand(0,4);
	dd.storage_info.access_protocols.nfs  = 0xFFFF;
	dd.storage_info.control_protocols.srm = 0xFFFF;
	dd.online_size.total   = myrand(1,3) * 1000;
	dd.online_size.used    = myrand(1,3) * 500;
	dd.nearline_size.total = 0; //myrand(1,6) * 1000;
	dd.nearline_size.used  = 0; //myrand(1,6) * 500;

	size_t i;

	for (i = 0; i < 1; i++) {
		grid_storage_area_t * tmp;
		tmp = new grid_storage_area_t;
		if (!tmp) {
			assert(0);
		}
		tmp->storage_area_name = new string("phosphorus-ddss-storage");
		tmp->storage_area_path = new string("/tmp/");
		tmp->storage_area_info.total_online_size =
			dd.online_size.total;
		//myrand(1,10) * 1000;
		tmp->storage_area_info.free_online_size =
			dd.online_size.total - dd.online_size.used;
		//myrand(1,10) * 1000 ;
		tmp->storage_area_info.reserved_total_online_size =
			0;//myrand(1,10) * 1000;
		tmp->storage_area_info.total_nearline_size =
			0; //myrand(1,10) * 1000;
		tmp->storage_area_info.free_nearline_size =
			0; //myrand(1,10) * 1000;
		tmp->storage_area_info.reserved_nearline_size =
			0; //myrand(1,10) * 1000;
		tmp->storage_area_info.retention_policy = RETPOL_CUSTODIAL;
		tmp->storage_area_info.access_latency = ACCLAT_ONLINE;
		tmp->storage_area_info.expiration_mode = EXP_NEVER_EXPIRE;

		dd.storage_areas.push_back(tmp);
	}

	for (i = 0; i < 0; i++) {
		struct tm * bdTime;
		time_t      unixTime;

		time(&unixTime);
		bdTime = localtime(&unixTime);

		bdTime->tm_year  = 108;
		bdTime->tm_mon   = i;
		bdTime->tm_mday  = 1 + (i * 14);
		bdTime->tm_hour  = 0;
		bdTime->tm_min   = 0;
		bdTime->tm_sec   = 0;
		bdTime->tm_isdst = -1;

		unixTime = mktime(bdTime);

		dd.se_calendar[unixTime].free_online_size = myrand(0, i*100);
		dd.se_calendar[unixTime].logical_cpus     = myrand(0, i*100);
	}

	BITMASK_BITSET(dd.mask_, name);
	BITMASK_BITSET(dd.mask_, storage_info);
	BITMASK_BITSET(dd.mask_, online_size);
	BITMASK_BITSET(dd.mask_, nearline_size);
	BITMASK_BITSET(dd.mask_, storage_areas);
	//BITMASK_BITSET(dd.mask_, se_calendar);
}

DEFUN(g2pcera_subnode_up,
      g2pcera_subnode_up_cmd,
      "subnode up (service|computing-elem|sub-cluster|storage-elem) id WORD "
      "grid-site-id WORD",
      "Add a subnode to a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	uint32_t		gnId;
	grid_subnode_ident_t	snIdent;

	if (strcmp(argv[0], "service") == 0) {
		snIdent.type = GRIDSUBNODE_SERVICE;
	} else if (strcmp(argv[0], "computing-elem") == 0) {
		snIdent.type = GRIDSUBNODE_COMPUTINGELEMENT;
	} else if (strcmp(argv[0], "sub-cluster") == 0) {
		snIdent.type = GRIDSUBNODE_SUBCLUSTER;
	} else if (strcmp(argv[0], "storage-elem") == 0) {
		snIdent.type = GRIDSUBNODE_STORAGEELEMENT;
	} else {
		vty_out(vty,
			"Invalid sub-node type (%s)%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	// match "<0-4294967295>".
	snIdent.id = strtoul(argv[1], NULL, 16);
	gnId	   = strtoul(argv[2], NULL, 16);

	switch (snIdent.type) {
		case GRIDSUBNODE_SERVICE: {
			grid_service_data_t	dataServ;
			set_dataservice(dataServ);

			ret = PCERA_TOPO_API::gridServiceUpdate(g2pceraId,
								gnId,
								snIdent.id,
								dataServ,
								err);
		}
			break;
		case GRIDSUBNODE_COMPUTINGELEMENT: {
			grid_ce_data_t		dataCE;
			set_datacomputing(dataCE);

			ret = PCERA_TOPO_API::gridCompElemUpdate(g2pceraId,
								 gnId,
								 snIdent.id,
								 dataCE,
								 err);
		}
			break;
		case GRIDSUBNODE_SUBCLUSTER: {
			grid_subcluster_data_t	dataSubC;
			set_datasubcluster(dataSubC);

			ret = PCERA_TOPO_API::gridSubClusterUpdate(g2pceraId,
								   gnId,
								   snIdent.id,
								   dataSubC,
								   err);
		}
			break;
		case GRIDSUBNODE_STORAGEELEMENT: {
			grid_se_data_t		dataSE;
			set_datastorage(dataSE);

			ret = PCERA_TOPO_API::gridStorageElemUpdate(g2pceraId,
								    gnId,
								    snIdent.id,
								    dataSE,
								    err);
		}
			break;
		default:
			assert(0);
	}

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update sub-node %s into grid-site %s "
			"in topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			subnodeIdent2string(snIdent).c_str(),
			numb_htoa(gnId),
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_subnode_del,
      g2pcera_subnode_del_cmd,
      "subnode del (service|computing-elem|sub-cluster|storage-elem) id WORD "
      "grid-site-id WORD",
      "Delete a subnode from a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	uint32_t		gnId;
	grid_subnode_ident_t	snIdent;

	if (strcmp(argv[0], "service") == 0) {
		snIdent.type = GRIDSUBNODE_SERVICE;
	} else if (strcmp(argv[0], "computing-elem") == 0) {
		snIdent.type = GRIDSUBNODE_COMPUTINGELEMENT;
	} else if (strcmp(argv[0], "sub-cluster") == 0) {
		snIdent.type = GRIDSUBNODE_SUBCLUSTER;
	} else if (strcmp(argv[0], "storage-elem") == 0) {
		snIdent.type = GRIDSUBNODE_STORAGEELEMENT;
	} else {
		vty_out(vty,
			"Invalid sub-node type (%s)%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	// match "<0-4294967295>".
	snIdent.id = strtoul(argv[1], NULL, 16);
	gnId	   = strtoul(argv[2], NULL, 16);

	ret = PCERA_TOPO_API::gridSubNodeDel(g2pceraId, gnId, snIdent, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot delete sub-node %s from grid-site %s "
			"in topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			subnodeIdent2string(snIdent).c_str(),
			numb_htoa(gnId),
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_tna_add,
      g2pcera_tna_add_cmd,
      "tna add rid A.B.C.D id-addr A.B.C.D id-pref WORD rc-id A.B.C.D",
      "Add a TNA to a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	tna_ident_t		ident;
	ipv4_t			ipAddr;
	uint32_t		preflen;

	memset(&ident, 0, sizeof(ident));

	// router id
	if (strchr(argv[0], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[0], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ident.rid = ntohl(ipAddr.s_addr);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	// address
	if (strchr(argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
		ident.id.type	    = IPv4;
		ident.id.value.ipv4 = ipAddr;
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	preflen = strtoul(argv[2], NULL, 10);
	ident.id.preflen = ((preflen > 32) ? 32 : (uint8_t) preflen);

	if (argc == 4) {
		// RC id
		if (strchr(argv[3], '.') != NULL) {
			// matched "A.B.C.D".
			if (!inet_aton(argv[3], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[3], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[3], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = PCERA_TOPO_API::tnaAdd(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot add TNA into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}
ALIAS(g2pcera_tna_add,
      g2pcera_tna_add_cmpt_cmd,
      "tna add rid A.B.C.D id-addr A.B.C.D id-pref WORD",
      "Add a TNA to a G2PCERA instance\n");

DEFUN(g2pcera_tna_del,
      g2pcera_tna_del_cmd,
      "tna del rid A.B.C.D id-addr A.B.C.D id-pref WORD rc-id A.B.C.D",
      "Delete a TNA from a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	tna_ident_t		ident;
	ipv4_t			ipAddr;
	uint32_t		preflen;

	memset(&ident, 0, sizeof(ident));

	// router id
	if (strchr(argv[0], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[0], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ident.rid = ntohl(ipAddr.s_addr);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	// address
	if (strchr(argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		if (!inet_aton(argv[1], &(ipAddr))) {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
		ident.id.type	    = IPv4;
		ident.id.value.ipv4 = ipAddr;
	} else {
		vty_out(vty,
			"Invalid IP address (%s) format%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	preflen = strtoul(argv[2], NULL, 10);
	ident.id.preflen = ((preflen > 32) ? 32 : (uint8_t) preflen);

	if (argc == 4) {
		// RC id
		if (strchr(argv[3], '.') != NULL) {
			// matched "A.B.C.D".
			if (!inet_aton(argv[3], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[3], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[3], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = PCERA_TOPO_API::tnaDel(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot delete TNA into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}
ALIAS(g2pcera_tna_del,
      g2pcera_tna_del_cmpt_cmd,
      "tna del rid A.B.C.D id-addr A.B.C.D id-pref WORD",
      "Delete a TNA from a G2PCERA instance\n");

DEFUN(g2pcera_telink_add,
      g2pcera_telink_add_cmd,
      "telink add "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "mode (p2p|enni-interdmn|enni-intradmn) "
      "ancestor-rc A.B.C.D neighbor-rc A.B.C.D",
      "Add a TE-link to a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	if (argc > 6) {
		if (strcmp(argv[6], "p2p") == 0) {
			ident.mode = LINK_MODE_P2P_NUMBERED;
		} else if (strcmp(argv[6], "enni-interdmn") == 0) {
			ident.mode = LINK_MODE_ENNI_INTERDOMAIN;
		} else if (strcmp(argv[6], "enni-intradmn") == 0) {
			ident.mode = LINK_MODE_ENNI_INTRADOMAIN;
		} else {
			vty_out(vty,
				"Invalid telink mode (%s)%s",
				argv[6], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	if (((ident.mode == LINK_MODE_ENNI_INTERDOMAIN) ||
	     (ident.mode == LINK_MODE_ENNI_INTRADOMAIN)) &&
	    (argc == 7)) {
		vty_out(vty,
			"Misssing RC id reference for ENNI links%s",
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (argc > 7) {
		// ancestor RCid
		if (strchr(argv[7], '.') != NULL) {
			// matched "A.B.C.D".
			ipv4_t ipAddr;

			if (!inet_aton(argv[7], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[7], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.lcl_rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}

		// neighbor RCid
		if (strchr(argv[8], '.') != NULL) {
			// matched "A.B.C.D".
			ipv4_t ipAddr;

			if (!inet_aton(argv[8], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[8], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.rem_rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[8], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = PCERA_TOPO_API::linkAdd(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot add telink into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}
ALIAS(g2pcera_telink_add,
      g2pcera_telink_add_cmpt1_cmd,
      "telink add "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "mode (p2p|enni-interdmn|enni-intradmn)",
      "Add a TE-link to a G2PCERA instance\n");
ALIAS(g2pcera_telink_add,
      g2pcera_telink_add_cmpt2_cmd,
      "telink add "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD",
      "Add a TE-link to a G2PCERA instance\n");

DEFUN(g2pcera_telink_del,
      g2pcera_telink_del_cmd,
      "telink del "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "mode (p2p|enni-interdmn|enni-intradmn) "
      "ancestor-rc A.B.C.D neighbor-rc A.B.C.D",
      "Delete a TE-link from a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	if (argc > 6) {
		if (strcmp(argv[6], "p2p") == 0) {
			ident.mode = LINK_MODE_P2P_NUMBERED;
		} else if (strcmp(argv[6], "enni-interdmn") == 0) {
			ident.mode = LINK_MODE_ENNI_INTERDOMAIN;
		} else if (strcmp(argv[6], "enni-intradmn") == 0) {
			ident.mode = LINK_MODE_ENNI_INTRADOMAIN;
		} else {
			vty_out(vty,
				"Invalid telink mode (%s)%s",
				argv[6], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	if (((ident.mode == LINK_MODE_ENNI_INTERDOMAIN) ||
	     (ident.mode == LINK_MODE_ENNI_INTRADOMAIN)) &&
	    (argc == 7)) {
		vty_out(vty,
			"Misssing RC id reference for ENNI links%s",
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (argc > 7) {
		// ancestor RCid
		if (strchr(argv[7], '.') != NULL) {
			// matched "A.B.C.D".
			ipv4_t ipAddr;

			if (!inet_aton(argv[7], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[7], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.lcl_rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}

		// neighbor RCid
		if (strchr(argv[8], '.') != NULL) {
			// matched "A.B.C.D".
			ipv4_t ipAddr;

			if (!inet_aton(argv[8], &(ipAddr))) {
				vty_out(vty,
					"Invalid IP address (%s) format%s",
					argv[8], VTY_NEWLINE);
				return CMD_WARNING;
			}

			ident.rem_rc_id = ntohl(ipAddr.s_addr);
		} else {
			vty_out(vty,
				"Invalid IP address (%s) format%s",
				argv[8], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = PCERA_TOPO_API::linkDel(g2pceraId, ident, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot delete telink into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}
ALIAS(g2pcera_telink_del,
      g2pcera_telink_del_cmpt1_cmd,
      "telink del "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "mode (p2p|enni-interdmn|enni-intradmn)",
      "Delete a TE-link from a G2PCERA instance\n");
ALIAS(g2pcera_telink_del,
      g2pcera_telink_del_cmpt2_cmd,
      "telink del "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD",
      "Delete a TE-link from a G2PCERA instance\n");

DEFUN(g2pcera_telink_upcom,
      g2pcera_telink_upcom_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "admin-metric WORD te-metric WORD color WORD prot-mask WORD "
      "max-bw WORD max-resv-bw WORD",
      "Update common TE-link params in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	telink_com_data_t	data;
	float			bw;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	memset(&data, 0, sizeof(data));

	data.admin_metric      = strtoul(argv[6], NULL, 10);
	data.te_metric	       = strtoul(argv[7], NULL, 10);
	data.te_color_mask     = strtoul(argv[8], NULL, 16);
	data.te_prot_type_mask = strtoul(argv[9], NULL, 16);
	bw		       = strtof(argv[10], NULL) / 8;
	data.te_max_bw	       = BW_BPS2HEX(bw);
	bw		       = strtof(argv[11], NULL) / 8;
	data.te_max_resv_bw    = BW_BPS2HEX(bw);

	BITMASK_BITSET(data.mask_, admin_metric);
	BITMASK_BITSET(data.mask_, te_metric);
	BITMASK_BITSET(data.mask_, te_color_mask);
	BITMASK_BITSET(data.mask_, te_prot_type_mask);
	BITMASK_BITSET(data.mask_, te_max_bw);
	BITMASK_BITSET(data.mask_, te_max_resv_bw);

	ret = PCERA_TOPO_API::teLinkUpdateCom(g2pceraId, ident, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upstates,
      g2pcera_telink_upstates_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "op-state (up|down) adm-state (enabled|disabled)",
      "Update TE-link state params in G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	opstate_t		opState;
	admstate_t		admState;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	if (strcmp(argv[6], "up") == 0) {
		opState = UP;
	} else if (strcmp(argv[6], "down") == 0) {
		opState = DOWN;
	} else {
		vty_out(vty,
			"Invalid operational state (%s)%s",
			argv[6], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp(argv[7], "enabled") == 0) {
		admState = ENABLED;
	} else if (strcmp(argv[7], "disabled") == 0) {
		admState = DISABLED;
	} else {
		vty_out(vty,
			"Invalid administrative state (%s)%s",
			argv[7], VTY_NEWLINE);
		return CMD_WARNING;
	}

	ret = PCERA_TOPO_API::teLinkUpdateStates(g2pceraId,
						 ident,
						 opState,
						 admState,
						 err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upabw,
      g2pcera_telink_upabw_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "avail-bw WORD at-prio <0-7>",
      "Update TE-link available bw params in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	bw_per_prio_t		abw;
	size_t			prio;
	float			bw;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	ret = PCERA_TOPO_API::teLinkGetGenBw(g2pceraId, ident, abw, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot get a telink available bw for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	bw   = strtof(argv[6], NULL) / 8;
	prio = strtoul(argv[7], NULL, 10);

	abw.bw[prio] = BW_BPS2HEX(bw);

	ret = PCERA_TOPO_API::teLinkUpdateGenBw(g2pceraId,
						ident,
						abw,
						err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink available bw "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upsrlg,
      g2pcera_telink_upsrlg_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "srlg-id WORD WORD WORD WORD WORD",
      "Update TE-link srlg list in a G2PCERA instance "
      "(up to 5 ids per command) \n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	std::list<uint32_t>	srlgs;
	size_t			id;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	srlgs.clear();

	for (id = 6; id < argc; id++) {
		srlgs.push_back(strtoul(argv[id], NULL, 10));
	}

	ret = PCERA_TOPO_API::teLinkAppendSrlgs(g2pceraId,
						ident,
						srlgs,
						err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink srlg list "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_telink_upsrlg,
      g2pcera_telink_upsrlg_cmpt1_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "srlg-id WORD WORD WORD WORD",
      "Update TE-link srlg list in a G2PCERA instance "
      "(up to 5 ids per command) \n");
ALIAS(g2pcera_telink_upsrlg,
      g2pcera_telink_upsrlg_cmpt2_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "srlg-id WORD WORD WORD",
      "Update TE-link srlg list in a G2PCERA instance "
      "(up to 5 ids per command) \n");
ALIAS(g2pcera_telink_upsrlg,
      g2pcera_telink_upsrlg_cmpt3_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "srlg-id WORD WORD",
      "Update TE-link srlg list in a G2PCERA instance "
      "(up to 5 ids per command) \n");
ALIAS(g2pcera_telink_upsrlg,
      g2pcera_telink_upsrlg_cmpt4_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "srlg-id WORD",
      "Update TE-link srlg list in a G2PCERA instance "
      "(up to 5 ids per command) \n");

DEFUN(g2pcera_telink_upcal,
      g2pcera_telink_upcal_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "year <2008-2037> "
      "mon <1-12> "
      "day <1-31> "
      "hour <0-23> "
      "min <0-59> "
      "sec <0-59> "
      "avail-bw WORD WORD WORD WORD WORD WORD WORD WORD",
      "Update TE-link calendar in a G2PCERA instance\n")
{
	std::string				err;
	pceraErrorCode_t			ret;
	telink_ident_t				ident;
	size_t					i;
	bw_per_prio_t				abw;
	struct tm *				bdTime;
	time_t					unixTime;
	std::map<uint32_t,bw_per_prio_t>	cal;

	cal.clear();

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	time(&unixTime);
	bdTime = localtime(&unixTime);

	bdTime->tm_year	 = strtoul(argv[6], NULL, 10) - 1900;
	bdTime->tm_mon	 = strtoul(argv[7], NULL, 10) - 1;
	bdTime->tm_mday	 = strtoul(argv[8], NULL, 10);
	bdTime->tm_hour	 = strtoul(argv[9], NULL, 10);
	bdTime->tm_min	 = strtoul(argv[10], NULL, 10);
	bdTime->tm_sec	 = strtoul(argv[11], NULL, 10);
	bdTime->tm_isdst = -1;	// Actively determines whether DST is in effect

	unixTime = mktime(bdTime);

	memset(&abw, 0, sizeof(abw));

	for (i = 12; i < argc; i++) {
		float bw;

		bw = strtof(argv[i], NULL) / 8;

		abw.bw[(i - 12)] = BW_BPS2HEX(bw);
	}

	cal[unixTime] = abw;

	ret = PCERA_TOPO_API::teLinkAppendCalendar(g2pceraId, ident, cal, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink calendar "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_telink_upcal,
      g2pcera_telink_upcal_cmpt1_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "year <2008-2037> "
      "mon <1-12> "
      "day <1-31> "
      "hour <0-23> "
      "min <0-59> "
      "sec <0-59> "
      "avail-bw WORD",
      "Update TE-link calendar in a G2PCERA instance\n");

DEFUN(g2pcera_telink_upisc,
      g2pcera_telink_upisc_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "sw-cap WORD enc-type WORD max-lsp-bw WORD "
      "min-lsp-bw WORD if-mtu WORD indication <0-2>",
      "Update TE-link ISC list in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	std::list<isc_t>	iscs;
	isc_t			isc;
	float			bw;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	iscs.clear();
	memset(&isc, 0, sizeof(isc));

	isc.switching	     = (sw_cap_t) strtoul(argv[6], NULL, 10);
	isc.encoding	     = (enc_type_t) strtoul(argv[7], NULL, 10);
	bw		     = strtof(argv[8], NULL) / 8;
	isc.max_lsp_bw.bw[0] = BW_BPS2HEX(bw);

	if (argc > 9) {
		uint16_t	mtu;
		uint8_t		ind;

		bw  = strtof(argv[9], NULL) / 8;
		mtu = (uint16_t) strtoul(argv[10], NULL, 10);
		ind = (uint8_t) strtoul(argv[11], NULL, 10);

		switch (isc.switching) {
			case SWCAP_PSC_1:
			case SWCAP_PSC_2:
			case SWCAP_PSC_3:
			case SWCAP_PSC_4: {
				isc.spec.psc.min_lsp_bw = BW_BPS2HEX(bw);
				isc.spec.psc.if_mtu	= mtu;
			}
				break;

			case SWCAP_TDM: {
				isc.spec.tdm.min_lsp_bw = BW_BPS2HEX(bw);
				isc.spec.tdm.indication = ind;
			}
				break;
			default:
				vty_out(vty,
					"Bad configuration params for "
					"isc %s%s",
					SHOW_SW_CAP(isc.switching),
					VTY_NEWLINE);
				return CMD_WARNING;

		}
	}

	iscs.push_front(isc);

	ret = PCERA_TOPO_API::teLinkAppendIsc(g2pceraId, ident, iscs, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink ISC list "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_telink_upisc,
      g2pcera_telink_upisc_cmpt1_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "sw-cap WORD enc-type WORD max-lsp-bw WORD",
      "Update TE-link ISC list in a G2PCERA instance\n");

DEFUN(g2pcera_telink_uptdm,
      g2pcera_telink_uptdm_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "tdm ho-mux WORD lo-mux WORD transp WORD ring-id WORD",
      "Update TDM TE-link data in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	telink_tdm_data_t	data;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	memset(&data, 0, sizeof(data));

	data.ho_muxcap_mask    = (uint8_t) strtoul(argv[6], NULL, 16);
	data.lo_muxcap_mask    = (uint8_t) strtoul(argv[7], NULL, 16);
	data.transparency_mask = strtoul(argv[8], NULL, 16);
	data.blsr_ring_id      = strtoul(argv[9], NULL, 10);

	BITMASK_BITSET(data.mask_, ho_muxcap_mask);
	BITMASK_BITSET(data.mask_, lo_muxcap_mask);
	BITMASK_BITSET(data.mask_, transparency_mask);
	BITMASK_BITSET(data.mask_, blsr_ring_id);

	ret = PCERA_TOPO_API::teLinkUpdateTdm(g2pceraId, ident, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink TDM data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_uptdmbw,
      g2pcera_telink_uptdmbw_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "tdm signal <1-19> free-timeslots WORD",
      "Update TDM TE-link free timeslots data in a G2PCERA instance\n")
{
	std::string			err;
	pceraErrorCode_t		ret;
	telink_ident_t			ident;
	std::list<uint32_t>		freeTS;
	std::list<uint32_t>::iterator	iter;
	uint32_t			ftsEntry;
	bool				added;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	freeTS.clear();
	ftsEntry = 0;

	ret = PCERA_TOPO_API::teLinkGetTdmBw(g2pceraId,
					     ident,
					     freeTS,
					     err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot get TDM available bw for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}


	ftsEntry |= ((strtoul(argv[6], NULL, 10) & 0x000000FF) << 24);
	ftsEntry |= ((strtoul(argv[7], NULL, 10) & 0x00FFFFFF) <<  0);

	for (added = false, iter = freeTS.begin();
	     iter != freeTS.end();
	     iter++) {
		if (GMPLS_SIGNAL_TYPE(*iter) == GMPLS_SIGNAL_TYPE(ftsEntry)) {
			(*iter) = ftsEntry;
			added = true;
			break;
		} else if (GMPLS_SIGNAL_TYPE(*iter) >
			   GMPLS_SIGNAL_TYPE(ftsEntry)) {
			freeTS.insert(iter, ftsEntry);
			added = true;
		}
	}

	if (!added) {
		freeTS.push_back(ftsEntry);
	}

	ret = PCERA_TOPO_API::teLinkUpdateTdmBw(g2pceraId, ident, freeTS, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink TDM bw data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upg709,
      g2pcera_telink_upg709_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "g709 oduk-mux WORD",
      "Update G.709 TE-link data in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	telink_lscg709_data_t	data;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	memset(&data, 0, sizeof(data));

	data.oduk_muxcap_mask = (uint8_t) strtoul(argv[6], NULL, 16);

	BITMASK_BITSET(data.mask_, oduk_muxcap_mask);

	ret = PCERA_TOPO_API::teLinkUpdateLscG709(g2pceraId, ident, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink G.709 data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upg709bw,
      g2pcera_telink_upg709bw_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "g709 (oduk|och) signal <1-8> free-slots WORD",
      "Update G709 TE-link free ODUk/OCh data in a G2PCERA instance\n")
{
	std::string			err;
	pceraErrorCode_t		ret;
	telink_ident_t			ident;
	std::list<uint32_t>		freeODU;
	std::list<uint32_t>		freeOC;
	std::list<uint32_t>::iterator	iter;
	uint32_t			ftsEntry;
	bool				added;
	bool				oduOrOc;
	gmpls_g709_sigtype_t		sig;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	freeODU.clear();
	freeOC.clear();
	ftsEntry = 0;

	ret = PCERA_TOPO_API::teLinkGetLscG709Bw(g2pceraId,
						 ident,
						 freeODU,
						 freeOC,
						 err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot get G709 available bw for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp(argv[6], "oduk") == 0) {
		oduOrOc = false;
	} else if (strcmp(argv[6], "och") == 0) {
		oduOrOc = true;
	} else {
		vty_out(vty,
			"Invalid signal argument (%s) %s",
			argv[6], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sig = (gmpls_g709_sigtype_t) strtoul(argv[7], NULL, 10);

	if (!oduOrOc) {
		// ODUk case
		if (!sig | sig > GMPLS_G709_SigType_ODU3_40_Gbps) {
			vty_out(vty,
				"Invalid ODUk signal argument (%s) %s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ftsEntry |= ((sig & 0x000000FF) << 24);
		ftsEntry |= ((strtoul(argv[8], NULL, 10) & 0x00FFFFFF) <<  0);

		for (added = false, iter = freeODU.begin();
		     iter != freeODU.end();
		     iter++) {
			if (GMPLS_SIGNAL_TYPE(*iter) ==
			    GMPLS_SIGNAL_TYPE(ftsEntry)) {
				(*iter) = ftsEntry;
				added = true;
				break;
			} else if (GMPLS_SIGNAL_TYPE(*iter) >
				   GMPLS_SIGNAL_TYPE(ftsEntry)) {
				freeODU.insert(iter, ftsEntry);
				added = true;
			}
		}

		if (!added) {
			freeODU.push_back(ftsEntry);
		}
	} else {
		// OCh case
		if (sig < GMPLS_G709_SigType_OCh_2_5_Gbps |
		    sig > GMPLS_G709_SigType_OCh_40_Gbps) {
			vty_out(vty,
				"Invalid OCh signal argument (%s) %s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}

		ftsEntry |= ((sig & 0x000000FF) << 24);
		ftsEntry |= ((strtoul(argv[8], NULL, 10) & 0x00EFFFFF) <<  0);

		for (added = false, iter = freeOC.begin();
		     iter != freeOC.end();
		     iter++) {
			if (GMPLS_SIGNAL_TYPE(*iter) ==
			    GMPLS_SIGNAL_TYPE(ftsEntry)) {
				(*iter) = ftsEntry;
				added = true;
				break;
			} else if (GMPLS_SIGNAL_TYPE(*iter) >
				   GMPLS_SIGNAL_TYPE(ftsEntry)) {
				freeOC.insert(iter, ftsEntry);
				added = true;
			}
		}

		if (!added) {
			freeOC.push_back(ftsEntry);
		}
	}

	ret = PCERA_TOPO_API::teLinkUpdateLscG709Bw(g2pceraId,
						    ident,
						    freeODU,
						    freeOC,
						    err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink G709 bw data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}

DEFUN(g2pcera_telink_upwdm,
      g2pcera_telink_upwdm_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "wdm d-pmd WORD span-length WORD "
      "amp-gain WORD amp-noise WORD amp-gain WORD amp-noise WORD",
      "Update WDM TE-link data in a G2PCERA instance\n")
{
	std::string		err;
	pceraErrorCode_t	ret;
	telink_ident_t		ident;
	telink_lscwdm_data_t	data;
	size_t			i;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	memset(&data.mask_, 0, sizeof(data.mask_));
	data.amplifiers.clear();

	data.dispersion_PMD = strtoul(argv[6], NULL, 10);
	data.span_length    = strtoul(argv[7], NULL, 10);

	for (i = 8; i < argc; i = i+2) {
		wdm_amplifier_data_t amp;

		amp.gain	 = strtoul(argv[i],   NULL, 10);
		amp.noise_figure = strtoul(argv[i+1], NULL, 10);
		data.amplifiers.push_back(amp);
	}

	BITMASK_BITSET(data.mask_, dispersion_PMD);
	BITMASK_BITSET(data.mask_, span_length);
	BITMASK_BITSET(data.mask_, amplifiers);

	ret = PCERA_TOPO_API::teLinkUpdateLscWdm(g2pceraId, ident, data, err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink WDM data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	return CMD_SUCCESS;
}
ALIAS(g2pcera_telink_upwdm,
      g2pcera_telink_upwdm_cmpt_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "wdm d-pmd WORD span-length WORD "
      "amp-gain WORD amp-noise WORD",
      "Update WDM TE-link data in a G2PCERA instance\n");

DEFUN(g2pcera_telink_upwdmbw,
      g2pcera_telink_upwdmbw_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "wdm w-count (1|8|16|32|40|100) w-base WORD busy-position <0-100>",
      "Update WDM TE-link lambda bitmap data in a G2PCERA instance\n")
{
	std::string			err;
	pceraErrorCode_t		ret;
	telink_ident_t			ident;
	wdm_link_lambdas_bitmap_t	bm;
	uint32_t			wcount;
	uint32_t			wbase;

	if (!get_telink_ident(vty, argv, argc, ident)) {
		return CMD_WARNING;
	}

	memset(&bm, 0, sizeof(bm));

	ret = PCERA_TOPO_API::teLinkGetLscWdmBw(g2pceraId,
						ident,
						bm,
						err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot get WDM lambda bitmap for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	wcount = strtoul(argv[6], NULL, 10);
	wbase  = strtoul(argv[7], NULL, 16);

	if (bm.num_wavelengths && bm.num_wavelengths != wcount) {
		if (bm.bitmap_word) {
			delete [] bm.bitmap_word;
		}

		vty_out(vty,
			"Inconsistent WDM lambda count in bitmap "
			"for G2.PCE-RA %s "
			"(passed: %d, existent:=%d)%s",
			numb_htoa(g2pceraId),
			wcount, bm.num_wavelengths,
			VTY_NEWLINE);
		return CMD_WARNING;
	}
	bm.num_wavelengths = wcount;

	if (bm.base_lambda_label && bm.base_lambda_label != wbase) {
		if (bm.bitmap_word) {
			delete [] bm.bitmap_word;
		}

		vty_out(vty,
			"Inconsistent WDM base lambda in bitmap "
			"for G2.PCE-RA %s "
			"(passed: 0x%08x, existent:=0x%08x)%s",
			numb_htoa(g2pceraId),
			wbase, bm.base_lambda_label,
			VTY_NEWLINE);
		return CMD_WARNING;
	}
	bm.base_lambda_label = wbase;

	if ((bm.bitmap_size  && !bm.bitmap_word) ||
	    (!bm.bitmap_size && bm.bitmap_word)) {
		if (bm.bitmap_word) {
			delete [] bm.bitmap_word;
		}

		vty_out(vty,
			"Inconsistent WDM lambda bitmap "
			"for G2.PCE-RA %s%s",
			numb_htoa(g2pceraId),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	uint32_t word_size;
	uint32_t word_mask;

	word_size = sizeof(uint32_t) * 8;

	if (!bm.bitmap_size  && !bm.bitmap_word) {

		bm.bitmap_size = ((wcount /  word_size) +
				  ((wcount % word_size) ? 1 : 0));

		bm.bitmap_word = new uint32_t[bm.bitmap_size];

		assert(bm.bitmap_word);

		memset(bm.bitmap_word, 0xFF, (bm.bitmap_size * 4));

		bm.bitmap_word[bm.bitmap_size - 1] = (0xFFFFFFFF <<
						      (word_size -
						       wcount % word_size));
	}

	if (argc > 8 ) {
		uint32_t bpos;

		bpos = strtoul(argv[8], NULL, 10);

		if (bpos >= wcount) {
			if (bm.bitmap_word) {
				delete [] bm.bitmap_word;
			}
			vty_out(vty,
				"Inconsistent WDM lambda busy-position with "
				"lambda count in bitmap "
				"for G2.PCE-RA %s "
				"(pos: %d, count: %d)%s",
				numb_htoa(g2pceraId),
				bpos,  wcount,
				VTY_NEWLINE);
			return CMD_WARNING;
		}

		word_mask = 1 << (word_size - 1 - (bpos % word_size));
		bm.bitmap_word[bpos / word_size] &= ~word_mask;
	}

	ret = PCERA_TOPO_API::teLinkUpdateLscWdmBw(g2pceraId,
						   ident,
						   bm,
						   err);

	if (ret != PCERA_ERROR_NONE) {
		vty_out(vty,
			"Cannot update telink G709 bw data "
			"into topology for G2.PCE-RA %s "
			"(err: code=%s, reason='%s')%s",
			numb_htoa(g2pceraId),
			PCERA_ERR_TO_STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (bm.bitmap_word) {
			delete [] bm.bitmap_word;
	}

	return CMD_SUCCESS;
}

ALIAS(g2pcera_telink_upwdmbw,
      g2pcera_telink_upwdmbw_cmpt_cmd,
      "telink update "
      "lcl-rid A.B.C.D lcl-id-addr A.B.C.D lcl-id-pref WORD "
      "rem-rid A.B.C.D rem-id-addr A.B.C.D rem-id-pref WORD "
      "wdm w-count (1|8|16|32|40|100) w-base WORD",
      "Update WDM TE-link lambda bitmap data in a G2PCERA instance\n");

DEFUN(g2pcera_gnscall_create,
      g2pcera_gnscall_create_cmd,
      "gnscall create id WORD src-lsr A.B.C.D "
      "itu-country WORD itu-carrier WORD unique-AP WORD",
      "Create Gns Call with given input parameters")
{
	mapKey_t	key;
	g2mpls_addr_t	lsr;
	uint32_t	ituCountryCode;	//  just 24 bits
	uint64_t	ituCarrierCode;	//  just 48 bits
	uint64_t	uniqueAP;	//  just 48 bits

	std::string	err;

	ituCountryCode = 0;
	ituCarrierCode = 0;
	uniqueAP       = 0;
	memset(&lsr, 0, sizeof(lsr));

	key = strtoull(argv[0], NULL, 16);

	if (argc < 2) {
		vty_out(vty,
			"Missing parameters in 'gnscall create'%s",
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		lsr.type    = IPv4;
		lsr.preflen = 32;
		if (!inet_aton(argv[1], &(lsr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src-LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for src-LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}


	if (argc > 2) {
		ituCountryCode  = strtoul(argv[2], NULL, 16);
		ituCountryCode &= 0x00FFFFFF;			// just 24 bits

		ituCarrierCode  = strtoull(argv[3], NULL, 16);
		ituCarrierCode &= 0x0000FFFFFFFFFFFFull;	// 48 bits

		uniqueAP        = strtoull(argv[4], NULL, 16);
		uniqueAP       &= 0x0000FFFFFFFFFFFFull;	// 48 bits
	}

	vty_out(vty,
		"Going to create a Gns Call with localID %s "
		"in G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key,
			   lsr,
			   ituCountryCode,
			   ituCarrierCode,
			   uniqueAP,
			   err)) {
		vty_out(vty,
			"Cannot update data into G2PCERA VTY for "
			"Gns Call with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Gns Call with localID %s successfully created "
		"in G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(g2pcera_gnscall_create,
      g2pcera_gnscall_create_cmpt_cmd,
      "gnscall create id WORD src-lsr A.B.C.D",
      "Create Gns Call with given input parameters");


DEFUN(g2pcera_gnscall_set_res60,
      g2pcera_gnscall_set_res60_cmd,
      "gnscall set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "data-link (WORD|A.B.C.D) "
      "label60 dest-mac WORD vlan-id <1-4094>",
      "Update info on tributary resources into Gns Call")
{
	mapKey_t	key;
	call_info_t	cinfo;
	net_res_spec_t	res;
	std::string	err;

	memset(&cinfo, 0, sizeof(cinfo));
	memset(&res, 0, sizeof(res));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "in-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, iTNA_res);
		BITMASK_BITSET(cinfo.iTNA_res.mask_, net);
	} else if (strcmp(argv[1], "eg-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, eTNA_res);
		BITMASK_BITSET(cinfo.eTNA_res.mask_, net);
	} else {
		vty_out(vty,
			"Invalid TNA resource specification (%s) "
			"for a Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		res.tna.type    = IPv4;
		res.tna.preflen = 32;
		if (!inet_aton(argv[2], &(res.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for TNA%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
		g2mpls_addr_t nullAddr;
		memset(&nullAddr, 0, sizeof(nullAddr));
		if (addr_equal(res.tna, nullAddr) == 1) {
			if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
				BITMASK_BITRESET(cinfo.iTNA_res.mask_, net);
			} else if (BITMASK_BITTEST(cinfo.mask_, eTNA_res)) {
				BITMASK_BITRESET(cinfo.eTNA_res.mask_, net);
			}
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for TNA%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(res.mask_, tna);

	if (argc > 3) {
		if (strchr (argv[3], '.') != NULL) {
			// matched "A.B.C.D".
			res.data_link.type    = IPv4;
			res.data_link.preflen = 32;
			if (!inet_aton(argv[3], &(res.data_link.value.ipv4))) {
				vty_out(vty,
					"Invalid IP address (%s) "
					"for data link%s",
					argv[2], VTY_NEWLINE);
				return CMD_WARNING;
			}
		} else {
			// match "<0-4294967295>".
			res.data_link.type         = UNNUMBERED;
			res.data_link.preflen      = 0;
			res.data_link.value.unnum  = strtoul(argv[3],NULL,16);
		}
		BITMASK_BITSET(res.mask_, data_link);

		res.label.type = LABEL_60BIT;
		std::string str;
		int   i;

		str = std::string(argv[4]);

		for (i = 0;  i< 6; i++) {
			std::string::size_type cursor;
			cursor = str.find(":");
			if (!str.empty()) {
				uint32_t num;
				num = strtoul(str.c_str(), NULL, 16);
				res.label.value.label60.mac[i] = (uint8_t) num;
			}

			str = (cursor == std::string::npos) ?
				"" :str.substr(cursor + 1);
		}

		res.label.value.label60.vlan_id = strtoul(argv[5], NULL, 10);
		BITMASK_BITSET(res.mask_, label);
	}

	if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
		memcpy(&(cinfo.iTNA_res.net), &res, sizeof(res));
		//defaultGridResSpec(cinfo.iTNA_res);
	} else {
		memcpy(&(cinfo.eTNA_res.net), &res, sizeof(res));
		defaultGridResSpec(cinfo.eTNA_res);
	}

	vty_out(vty,
		"Going to update info on tributary resources "
		"forGns Call with localID %s in G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update info in G2PCERA VTY on "
			"tributary resources for in Gns Call with "
			"localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Info on tributary resources successfully updated "
		"for Gns Call with localID %s in  G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_res32,
      g2pcera_gnscall_set_res32_cmd,
      "gnscall set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "data-link (WORD|A.B.C.D) "
      "label32 id WORD",
      "Update info on tributary resources into Gns Call")
{
	mapKey_t	key;
	call_info_t	cinfo;
	net_res_spec_t	res;
	std::string	err;

	memset(&cinfo, 0, sizeof(cinfo));
	memset(&res, 0, sizeof(res));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "in-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, iTNA_res);
		BITMASK_BITSET(cinfo.iTNA_res.mask_, net);
	} else if (strcmp(argv[1], "eg-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, eTNA_res);
		BITMASK_BITSET(cinfo.eTNA_res.mask_, net);
	} else {
		vty_out(vty,
			"Invalid TNA resource specification (%s) "
			"for a Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		res.tna.type    = IPv4;
		res.tna.preflen = 32;
		if (!inet_aton(argv[2], &(res.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for TNA%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
		g2mpls_addr_t nullAddr;
		memset(&nullAddr, 0, sizeof(nullAddr));
		if (addr_equal(res.tna, nullAddr) == 1) {
			if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
				BITMASK_BITRESET(cinfo.iTNA_res.mask_, net);
			} else if (BITMASK_BITTEST(cinfo.mask_, eTNA_res)) {
				BITMASK_BITRESET(cinfo.eTNA_res.mask_, net);
			}
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for TNA%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(res.mask_, tna);

	if (argc > 3) {
		if (strchr (argv[3], '.') != NULL) {
			// matched "A.B.C.D".
			res.data_link.type    = IPv4;
			res.data_link.preflen = 32;
			if (!inet_aton(argv[3], &(res.data_link.value.ipv4))) {
				vty_out(vty,
					"Invalid IP address (%s) "
					"for data link%s",
					argv[3], VTY_NEWLINE);
				return CMD_WARNING;
			}
		} else {
			// match "<0-4294967295>".
			res.data_link.type        = UNNUMBERED;
			res.data_link.preflen     = 0;
			res.data_link.value.unnum = strtoul(argv[3], NULL, 16);
		}
		BITMASK_BITSET(res.mask_, data_link);

		res.label.type = LABEL_32BIT;
		res.label.value.label32.id = strtoul(argv[4], NULL, 16);
		BITMASK_BITSET(res.mask_, label);
	}

	if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
		memcpy(&(cinfo.iTNA_res.net), &res, sizeof(res));
		//defaultGridResSpec(cinfo.iTNA_res);
	} else {
		memcpy(&(cinfo.eTNA_res.net), &res, sizeof(res));
		defaultGridResSpec(cinfo.eTNA_res);
	}

	vty_out(vty,
		"Going to update info on tributary resources "
		"forGns Call with localID %s in G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update info on tributary resources "
			"on VTY Gns Call with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Info on tributary resources successfully updated "
		"for Gns Call with localID %s in  G2PCERA VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(g2pcera_gnscall_set_res32,
      g2pcera_gnscall_set_res32_cmpt_cmd,
      "gnscall set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "dl (WORD|A.B.C.D) "
      "l32 WORD",
      "Update info on tributary resources into Gns Call");
ALIAS(g2pcera_gnscall_set_res32,
      g2pcera_gnscall_set_res32_cmpt1_cmd,
      "gnscall set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D",
      "Update info on tributary resources into Gns Call");

ALIAS(g2pcera_gnscall_set_res60,
      g2pcera_gnscall_set_res60_cmpt_cmd,
      "gnscall set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "dl (WORD|A.B.C.D) "
      "l60 dmac WORD vid <1-4094>)",
      "Update info on tributary resources into Gns Call");

DEFUN(g2pcera_gnscall_set_calltime,
      g2pcera_gnscall_set_calltime_cmd,
      "gnscall set-info id WORD gns-call-data (start-time|end-time) "
      "year <2008-2037> "
      "mon <1-12> "
      "day <1-31> "
      "hour <0-23> "
      "min <0-59> "
      "sec <0-59>",
      "Update GNS time info into Gns Call (in local timezone)")
{
	mapKey_t	key;
	call_info_t	cinfo;
	struct tm *	bdTime;
	time_t		unixTime;
	std::string	err;

	memset(&cinfo, 0, sizeof(cinfo));

	time(&unixTime);
	bdTime = localtime(&unixTime);

	key = strtoull(argv[0], NULL, 16);

	bdTime->tm_year  = atoi(argv[2]) - 1900;
	bdTime->tm_mon   = atoi(argv[3]) - 1;
	bdTime->tm_mday  = atoi(argv[4]);
	bdTime->tm_hour  = atoi(argv[5]);
	bdTime->tm_min   = atoi(argv[6]);
	bdTime->tm_sec   = atoi(argv[7]);
	bdTime->tm_isdst = -1;	// Actively determines whether DST is in effect

	unixTime = mktime(bdTime);

	if (strcmp(argv[1], "start-time") == 0) {
		BITMASK_BITSET(cinfo.mask_, times);
		cinfo.times.start_time = unixTime;
	} else if (strcmp(argv[1], "end-time") == 0) {
		BITMASK_BITSET(cinfo.mask_, times);
		cinfo.times.end_time = unixTime;
	} else {
		vty_out(vty,
			"Invalid timing specification (%s) for a Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to update GNS time info for "
		"Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update GNS time info into VTY Gns Call "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"GNS time info successfully updated "
		"in Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_callnames,
      g2pcera_gnscall_set_callnames_cmd,
      "gnscall set-info id WORD gns-call-data "
      "job-name WORD "
      "job-project WORD",
      "Update GNS naming info into Gns Call")
{
	mapKey_t	key;
	call_info_t	cinfo;
	std::string	err;

	memset(&cinfo, 0, sizeof(cinfo));

	key = strtoull(argv[0], NULL, 16);

	cinfo.job_name = new std::string(argv[1]);
	BITMASK_BITSET(cinfo.mask_, job_name);
	cinfo.job_project = new std::string(argv[2]);
	BITMASK_BITSET(cinfo.mask_, job_project);

	vty_out(vty,
		"Going to update GNS naming info "
		"for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update GNS naming info into VTY Gns Call "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"GNS naming info successfully updated "
		"in Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_recovery,
      g2pcera_gnscall_set_recovery_cmd,
      "gnscall set-info id WORD recovery-data "
      "type (none|protection|pre-planned|on-the-fly|revertive-on-the-fly) "
      "disjointness (none|link|node|srlg)",
      "Update info on recovery into Gns Call")
{
	mapKey_t	key;
	recovery_info_t	rinfo;
	std::string	err;

	memset(&rinfo, 0, sizeof(rinfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "none") == 0) {
		rinfo.rec_type = RECOVERY_NONE;
	} else if (strcmp(argv[1], "protection") == 0) {
		rinfo.rec_type = RECOVERY_PROTECTION;
	} else if (strcmp(argv[1], "pre-planned") == 0) {
		rinfo.rec_type = RECOVERY_PREPLANNED;
	} else if (strcmp(argv[1], "on-the-fly") == 0) {
		rinfo.rec_type = RECOVERY_ONTHEFLY;
	} else if (strcmp(argv[1], "revertive-on-the-fly") == 0) {
		rinfo.rec_type = RECOVERY_REVERTIVEONTHEFLY;
	} else {
		vty_out(vty,
			"Invalid recovery type (%s) for a Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	BITMASK_BITSET(rinfo.mask_, rec_type);

	if (strcmp(argv[2], "none") == 0) {
		rinfo.disj_type = DISJOINTNESS_NONE;
	} else if (strcmp(argv[2], "link") == 0) {
		rinfo.disj_type = DISJOINTNESS_LINK;
	} else if (strcmp(argv[2], "node") == 0) {
		rinfo.disj_type = DISJOINTNESS_NODE;
	} else if (strcmp(argv[2], "srlg") == 0) {
		rinfo.disj_type = DISJOINTNESS_SRLG;
	} else {
		vty_out(vty,
			"Invalid recovery type (%s) for a Gns Call%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	BITMASK_BITSET(rinfo.mask_, disj_type);

	vty_out(vty,
		"Going to update info on recovery "
		"for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyCallUpdate(key, rinfo, err)) {
		vty_out(vty,
			"Cannot update info on recovery LSR "
			"on VTY Gns Call with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Recovery info successfully updated "
		"in Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_lspbase,
      g2pcera_gnscall_set_lspbase_cmd,
      "gnscall set-info id WORD lsp-data "
      "sw-cap (l2sc|lsc|fsc) "
      "enc-type (ethernet|lambda|fiber) "
      "bw WORD "
      "gpid WORD "
      "action (xconnect|book) "
      "type (spc|pc)"
      "role (primary|secondary)",
      "Update LSP base info into Gns Call")
{
	mapKey_t	key;
	lsp_info_t	linfo;
	float		bw;
	std::string	err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "l2sc") == 0) {
		linfo.sw_cap = SWCAP_L2SC;
	} else if (strcmp(argv[1], "lsc") == 0) {
		linfo.sw_cap = SWCAP_LSC;
	} else if (strcmp(argv[1], "fsc") == 0) {
		linfo.sw_cap = SWCAP_FSC;
	} else {
		vty_out(vty,
			"Invalid switching cap. (%s) for a Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, sw_cap);

	if (strcmp(argv[2], "ethernet") == 0) {
		linfo.enc_type = ENCT_ETHERNET;
	} else if (strcmp(argv[2], "lambda") == 0) {
		linfo.enc_type = ENCT_LAMBDA;
	} else if (strcmp(argv[2], "fiber") == 0) {
		linfo.enc_type = ENCT_FIBER;
	} else {
		vty_out(vty,
			"Invalid encoding type (%s) for a Gns Call%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, enc_type);

	bw	   = strtof(argv[3], NULL);
	linfo.bw   = BW_BPS2HEX(bw);
	BITMASK_BITSET(linfo.mask_, bw);
	linfo.gpid = (gpid_t) strtoul(argv[4], NULL, 10);
	BITMASK_BITSET(linfo.mask_, gpid);

	if (argc > 5) {

		if (strcmp (argv[5], "xconnect") == 0) {
			linfo.action = RESOURCE_XCONNECT;
		} else if (strcmp (argv[5], "book") == 0) {
			linfo.action = RESOURCE_BOOK;
		} else {
			vty_out(vty,
				"Invalid resource behavior type (%s) "
				"for Gns Call%s",
				argv[5], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, action);

		if (strcmp (argv[6], "spc") == 0) {
			linfo.type = LSP_TYPE_SPC;
		} else if (strcmp (argv[6], "pc") == 0) {
			linfo.type = LSP_TYPE_PC;
		} else {
			vty_out(vty,
				"Invalid LSP type (%s) "
				"for Gns Call%s",
				argv[6], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, type);

		if (strcmp (argv[7], "primary") == 0) {
			linfo.role = LSP_ROLE_PRIMARY;
		} else if (strcmp (argv[7], "secondary") == 0) {
			linfo.role = LSP_ROLE_SECONDARY;
		} else {
			vty_out(vty,
				"Invalid LSP role (%s) "
				"for Gns Call%s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, role);
	}

	vty_out(vty,
		"Going to update LSP base info for "
		"Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update base info into VTY Gns Call "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP base info successfully updated "
		"in Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(g2pcera_gnscall_set_lspbase,
      g2pcera_gnscall_set_lspbase_cmpt_cmd,
      "gnscall set-info id WORD lsp-data "
      "sw-cap (l2sc|lsc|fsc) "
      "enc-type (ethernet|lambda|fiber) "
      "bw WORD "
      "gpid WORD",
      "Update LSP info data into Gns Call");

DEFUN(g2pcera_gnscall_set_lspconstr,
      g2pcera_gnscall_set_lspconstr_cmd,
      "gnscall set-info id WORD lsp-constraints "
      "s-prio <0-7> "
      "h-prio <0-7> "
      "exc-any WORD "
      "inc-any WORD "
      "inc-all WORD "
      "link-prot (none|extra|unprotected|shared|ded-1:1|ded-1+1|enhanced)",
      "Update LSP constraints info into Gns Call")
{
	mapKey_t	key;
	lsp_info_t	linfo;
	std::string	err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	linfo.setup_prio   = strtoul(argv[1], NULL, 10);
	linfo.holding_prio = strtoul(argv[2], NULL, 10);
	linfo.exclude_any  = strtoul(argv[3], NULL, 16);
	linfo.include_any  = strtoul(argv[4], NULL, 16);
	linfo.include_all  = strtoul(argv[5], NULL, 16);

	BITMASK_BITSET(linfo.mask_, setup_prio);
	BITMASK_BITSET(linfo.mask_, holding_prio);
	BITMASK_BITSET(linfo.mask_, exclude_any);
	BITMASK_BITSET(linfo.mask_, include_any);
	BITMASK_BITSET(linfo.mask_, include_all);

	if (strcmp (argv[6], "none") == 0) {
		linfo.link_prot_mask = PROTTYPE_NONE;
	} else if (strcmp (argv[6], "extra") == 0) {
		linfo.link_prot_mask = PROTTYPE_EXTRA;
	} else if (strcmp (argv[6], "unprotected") == 0) {
		linfo.link_prot_mask = PROTTYPE_UNPROTECTED;
	} else if (strcmp (argv[6], "shared") == 0) {
		linfo.link_prot_mask = PROTTYPE_SHARED;
	} else if (strcmp (argv[6], "ded-1:1") == 0) {
		linfo.link_prot_mask = PROTTYPE_DEDICATED_1TO1;
	} else if (strcmp (argv[6], "ded-1+1") == 0) {
		linfo.link_prot_mask = PROTTYPE_DEDICATED_1PLUS1;
	} else if (strcmp (argv[6], "enhanced") == 0) {
		linfo.link_prot_mask = PROTTYPE_ENHANCED;
	} else {
		vty_out(vty,
			"Invalid link protection type (%s) "
			"for Gns Call%s",
			argv[6], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, link_prot_mask);

	vty_out(vty,
		"Going to update LSP constraints "
		"for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update LSP constraints into VTY Gns Call "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP constraints successfully updated "
		"in Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_lspproto,
      g2pcera_gnscall_set_lspproto_cmd,
      "gnscall set-info id WORD lsp-proto "
      "cback-scope (none|e2e|boundary|segment) "
      "cback-max-retr-src <0-5> "
      "cback-max-retr-intmd <0-5> "
      "record-route (all|dl-detail|tel-detail|off)",
      "Update protocol related LSP info data into Gns Call")
{
	mapKey_t	key;
	lsp_info_t	linfo;
	std::string	err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp (argv[1], "none") == 0) {
		linfo.crankback = CB_NONE;
	} else if (strcmp (argv[1], "e2e") == 0) {
		linfo.crankback = CB_E2E;
	} else if (strcmp (argv[1], "boundary") == 0) {
		linfo.crankback = CB_BOUNDARY;
	} else if (strcmp (argv[1], "segment") == 0) {
		linfo.crankback = CB_SEGMENT;
	} else {
		vty_out(vty,
			"Invalid crakback scope (%s) "
			"for Gns Call%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, crankback);

	linfo.max_cback_retries_src    = strtoul(argv[2], NULL, 10);
	BITMASK_BITSET(linfo.mask_, max_cback_retries_src);

	linfo.max_cback_retries_intmd  = strtoul(argv[3], NULL, 10);
	BITMASK_BITSET(linfo.mask_, max_cback_retries_intmd);

	if (strcmp (argv[4], "all") == 0) {
		linfo.rro_mode = LSP_RRO_ALL;
	} else if (strcmp (argv[4], "dl-detail") == 0) {
		linfo.rro_mode = LSP_RRO_DL_DETAIL;
	} else if (strcmp (argv[4], "tel-detail") == 0) {
		linfo.rro_mode = LSP_RRO_TEL_DETAIL;
	} else if (strcmp (argv[4], "off") == 0) {
		linfo.rro_mode = LSP_RRO_OFF;
	} else {
		vty_out(vty,
			"Invalid RRO recording mode (%s) for Gns Call%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, rro_mode);

	vty_out(vty,
		"Going to update LSP data for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update data into VTY Gns Call with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP data successfully updated in "
		"Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_set_lsptimers,
      g2pcera_gnscall_set_lsptimers_cmd,
      "gnscall set-info id WORD lsp-proto-timers "
      "refresh-period <0-4294967295> "
      "ack-procedure (on|off) "
      "rapid-retransmission-interval <0-4294967295> "
      "rapid-retry-limit <1-5> "
      "increment-value-delta <1-10>",
      "Update protocol related LSP info data into Gns Call")
{
	mapKey_t	     key;
	lsp_info_t	     linfo;
	std::string	     err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	linfo.refresh_interval = strtoul(argv[1], NULL, 10);
	BITMASK_BITSET(linfo.mask_, refresh_interval);

	if (strcmp (argv[2], "on") == 0) {
		linfo.activate_ack = 1; //true
	} else if (strcmp (argv[2], "off") == 0) {
		linfo.activate_ack = 0; //false
	} else {
		vty_out(vty,
			"Invalid ACK mode (%s) for Gns Call%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, activate_ack);

	linfo.rapid_retransm_interval = strtoul(argv[3], NULL, 10);
	linfo.rapid_retry_limit       = strtoul(argv[4], NULL, 10);
	linfo.increment_value_delta   = strtoul(argv[5], NULL, 10);

	BITMASK_BITSET(linfo.mask_, rapid_retransm_interval);
	BITMASK_BITSET(linfo.mask_, rapid_retry_limit);
	BITMASK_BITSET(linfo.mask_, increment_value_delta);

	vty_out(vty,
		"Going to update LSP data for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update data into VTY Gns Call with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP data successfully updated in "
		"Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_route,
      g2pcera_gnscall_route_cmd,
      "gnscall route id WORD",
      "Route Gns Call with input parameters given before")
{
	mapKey_t			key;
	std::string			err;
	std::list<ero_hop_t>		ero;
	std::list<ero_hop_t>::iterator	iter;
	net_res_spec_t			eTna;

	key = strtoull(argv[0], NULL, 16);

	vty_out(vty,
		"Going to route a Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallRoute(g2pceraId, key, ero, eTna, err)) {
		vty_out(vty,
			"Cannot route a Gns Call from G2PCERA VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Gns Call with localID %s successfully routed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	vty_out(vty,
		"Computed gnscall ERO:%s",
		VTY_NEWLINE);
	for (iter = ero.begin(); iter != ero.end(); iter++) {
		vtyDump(vty, "   ", *iter);
	}

	if (ero.empty()) {
		vty_out(vty,
			"%s%s***EMPTY***%s",
			VTY_TAB,VTY_TAB,
			VTY_NEWLINE);
	}

	vty_out(vty,
		"Computed gnscall egress TNA%s",
		VTY_NEWLINE);
	vtyDump(vty, "   ", eTna);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_flush,
      g2pcera_gnscall_flush_cmd,
      "gnscall flush id WORD",
      "Flush Gns Call with input parameters given before")
{
	mapKey_t      key;
	std::string   err;

	key = strtoull(argv[0], NULL, 16);

	vty_out(vty,
		"Going to flush a Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallFlush(g2pceraId, key, err)) {
		vty_out(vty,
			"Cannot flush a Gns Call from G2PCERA VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Gns Call with localID %s successfully flushed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_confirm,
      g2pcera_gnscall_confirm_cmd,
      "gnscall confirm id WORD",
      "Confirm Gns Call with input parameters given before")
{
	mapKey_t      key;
	std::string   err;

	key = strtoull(argv[0], NULL, 16);

	vty_out(vty,
		"Going to confirm a Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyCallConfirm(g2pceraId, key, err)) {
		vty_out(vty,
			"Cannot confirm a Gns Call from G2PCERA VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Gns Call with localID %s successfully confirmed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(g2pcera_gnscall_connget,
      g2pcera_gnscall_connget_cmd,
      "gnscall connections-get id WORD",
      "Get connections for a Gns Call with input parameters given before")
{
	mapKey_t			key;
	std::string			err;
	std::list<ero_hop_t>		pEro;
	std::list<ero_hop_t>		sEro;
	std::list<ero_hop_t>::iterator	iter;

	key = strtoull(argv[0], NULL, 16);

	vty_out(vty,
		"Going to fetch connections for Gns Call with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyGetConnections(g2pceraId, key, pEro, sEro, err)) {
		vty_out(vty,
			"Cannot Get connections for a Gns Call "
			"from G2PCERA VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Connections for Gns Call with localID %s "
		"successfully fetched%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	vty_out(vty,
		"Computed primary ERO:%s",
		VTY_NEWLINE);
	for (iter = pEro.begin(); iter != pEro.end(); iter++) {
		vtyDump(vty, "   ", *iter);
	}
	if (pEro.empty()) {
		vty_out(vty,
			"%s%s***EMPTY***%s",
			VTY_TAB,VTY_TAB,
			VTY_NEWLINE);
	}


	vty_out(vty,
		"Computed secondary ERO:%s",
		VTY_NEWLINE);
	for (iter = sEro.begin(); iter != sEro.end(); iter++) {
		vtyDump(vty, "   ", *iter);
	}
	if (sEro.empty()) {
		vty_out(vty,
			"%s%s***EMPTY***%s",
			VTY_TAB,VTY_TAB,
			VTY_NEWLINE);
	}

	return CMD_SUCCESS;
}

void G2PCERA_Master::init_vty (void)
{
	// Install g2pcera top node
	install_node (&g2pcera_node,	NULL);

	install_element(CONFIG_NODE,	&g2pcera_cmd);
	install_element(CONFIG_NODE,	&no_g2pcera_cmd);
	install_element(CONFIG_NODE,	&g2pcera_cmpt_cmd);
	install_element(CONFIG_NODE,	&no_g2pcera_cmpt_cmd);
	install_element(CONFIG_NODE,	&g2pcera_load_topology_cmd);
	install_element(VIEW_NODE,	&g2pcera_load_topology_cmd);
	install_element(ENABLE_NODE,	&g2pcera_load_topology_cmd);

	install_default(G2PCERA_NODE);

	//instance commands
	install_element(G2PCERA_NODE,	&g2pcera_node_add_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_del_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt2_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt3_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt4_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt5_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt6_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_netup_cmpt7_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_gsup_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_gsup_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_node_gsup_cmpt2_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_subnode_up_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_subnode_del_cmd);

	install_element(G2PCERA_NODE,	&g2pcera_tna_add_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_tna_add_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_tna_del_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_tna_del_cmpt_cmd);

	install_element(G2PCERA_NODE,	&g2pcera_telink_add_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_add_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_add_cmpt2_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_del_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_del_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_del_cmpt2_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upcom_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upstates_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upabw_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upsrlg_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upsrlg_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upsrlg_cmpt2_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upsrlg_cmpt3_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upsrlg_cmpt4_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upcal_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upcal_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upisc_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upisc_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_uptdm_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_uptdmbw_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upg709_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upg709bw_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upwdm_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upwdm_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upwdmbw_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_telink_upwdmbw_cmpt_cmd);

	install_element(G2PCERA_NODE,	&g2pcera_gnscall_create_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_create_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_res60_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_res32_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_res32_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_res32_cmpt1_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_res60_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_calltime_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_callnames_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_recovery_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_lspbase_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_lspbase_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_lspconstr_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_lspproto_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_set_lsptimers_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_route_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_flush_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_confirm_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_gnscall_connget_cmd);

	//show commands
	install_element(VIEW_NODE,	&g2pcera_topo_show_cmd);
	install_element(ENABLE_NODE,	&g2pcera_topo_show_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_topo_show_cmd);
	install_element(VIEW_NODE,	&g2pcera_topo_show_cmpt_cmd);
	install_element(ENABLE_NODE,	&g2pcera_topo_show_cmpt_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_topo_show_cmpt_cmd);


	install_element(VIEW_NODE,	&g2pcera_show_cmd);
	install_element(ENABLE_NODE,	&g2pcera_show_cmd);
	install_element(G2PCERA_NODE,	&g2pcera_show_cmd);

}
