//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include "checksum.h"

#include "g2rsvpte_buffer.h"
#include "g2rsvpte_messages.h"

//
// Local Utilities (usefull ONLY here!!!)
//
#define CHECK_BUFFER_INDEX(START,STOP,CURR)		\
{							\
	assert(START <= STOP);				\
	/* CURR is stop-start+1 when buffer is full */	\
	assert(CURR  <= STOP - START + 1);		\
}

#define OPERATOR_WRITE(FUNC)						\
{									\
	assert(cursor.remainingSize() >= sizeof(data));			\
									\
	/* (cursor.current_, data) */					\
	cursor.buffer_->FUNC(cursor.start_ + cursor.current_, data);	\
	cursor.current_ += sizeof(data);				\
									\
	CHECK_BUFFER_INDEX(cursor.start_,				\
			   cursor.stop_,				\
			   cursor.current_);				\
}

#define OPERATOR_READ(FUNC)						\
{									\
	assert(cursor.remainingSize() >= sizeof(data));			\
									\
	/* (cursor.current_) */						\
	data = cursor.buffer_->FUNC(cursor.start_ + cursor.current_);	\
	cursor.current_ += sizeof(data);				\
									\
	CHECK_BUFFER_INDEX(cursor.start_,				\
			   cursor.stop_,				\
			   cursor.current_);				\
}

namespace RSVP {

	//
	// MemoryBuffer
	//
	MemoryBuffer::MemoryBuffer(size_t size)
	{
		assert(size > 0);
		buffer_ = new uint8_t[size];  // malloc(size);
		assert(buffer_);
		memset(buffer_, 0, size);
		size_ = size;
	};

	// MemoryBuffer initialized copying the given buffer
	MemoryBuffer::MemoryBuffer(uint8_t * buffer, size_t size)
	{
		assert(buffer);
		assert(size > 0);

		buffer_ = new uint8_t[size];  // malloc(size);
		assert(buffer_);
		size_   = size;
		memcpy(buffer_, buffer, size);
	};

	MemoryBuffer::~MemoryBuffer(void)
	{
		if (buffer_) {
			delete [] buffer_;       // free(buffer_)
			buffer_ = 0;
		}
	}

	// NOTE: buffer is in the network byte order form, so the checksum
	//       is already calculated in network byte order
	uint16_t MemoryBuffer::calculateRSVPChecksum(void) const
	{
		// Quagga lib (TMP ???)
		return in_cksum(buffer_, size_);
	}

	uint16_t MemoryBuffer::calculateChecksum(void) const
	{
		uint16_t checksum;
		checksum = calculateRSVPChecksum();

		G2RSVPTE_DBG("calculated CHECKSUM: 0x%x", checksum);

		return checksum;
	}

	void MemoryBuffer::writeChecksum(void)
	{
		// checksum is in the network byte order form,
		// so the htons(checksum) function is not needed
		setChecksum(calculateChecksum());
	}

	bool MemoryBuffer::isChecksumOk(void)
	{
		uint16_t checksum;
		// Store original checksum
		checksum = getChecksum();

		if (checksum != 0) {
			setChecksum(0);
			bool res;
			res = (checksum == calculateChecksum());
			// Restore original checksum (in any case)
			setChecksum(checksum);
			return res;
		}

		return true;
	}

	const uint8_t * MemoryBuffer::getData(void)
	{
		return buffer_;
	}

	MemoryCursor & MemoryBuffer::cursor(void)
	{
		return MemoryCursor(this).resize(0, size_ - 1);
	}

	std::ostream & operator << (std::ostream & os, const MemoryBuffer & mb)
	{
		mb.print(os, 0, mb.size() - 1);
		return os;
	}

	void MemoryBuffer::print(std::ostream & os,
				 size_t         start,
				 size_t         stop) const
	{
		const uint32_t ALIGN     = 16;
		const uint32_t SEPARATOR = 4;

		assert(start <= stop);
		assert(stop - start <= size_);

		//G2RSVPTE_DBG("buffer [%d, %d]", start, stop);

		os << std::hex;
		int i, len = stop - ALIGN;
		uint32_t j, k;
		for (i = start; i <= len; i += ALIGN) {
			for (j = 0; j < ALIGN; j += 1) {
				os << std::setw(2) << std::setfill('0')
				   << (uint16_t)(buffer_[i + j]) << " ";
				if ((j + 1) % SEPARATOR == 0) os << " ";
			}
			os << std::endl;
		}

		for (k = i; k <= stop; k++) {
			os << std::setw(2) << std::setfill('0')
			   << (uint16_t)(buffer_[k]) << " ";
			if ((k + 1) % SEPARATOR == 0) os << " ";
		}
		os << std::dec << std::endl;
	}

	//
	// MemoryCursor
	//
	MemoryCursor::MemoryCursor(MemoryBuffer * buffer,
				   size_t         start,
				   size_t         stop)
	{
		//G2RSVPTE_DBG("---> Cursor CONSTRUCT: [%d,%d] size %d",
		//	     start, stop, buffer->size());
		assert(buffer);
		assert(start <= stop);
		assert(stop < buffer->size());

		buffer_ = buffer;
		initIndex(start, stop, 0); //x, x, start_
	}

	MemoryCursor::MemoryCursor(MemoryBuffer * buffer)
	{
		assert(buffer);

		buffer_ = buffer;
		initIndex(0, buffer->size() - 1, 0); //x, x, start_
	}

	MemoryCursor::~MemoryCursor(void)
	{
	}

	void MemoryCursor::initIndex(size_t start, size_t stop, size_t curr)
	{
		start_   = start;
		stop_    = stop;
		current_ = curr;
	}

	MemoryCursor & MemoryCursor::resize(size_t start, size_t stop)
	{
		//G2RSVPTE_DBG("---> Cursor RESIZE: act [%d,%d] next [%d,%d]",
		//	     start_, stop_, start, stop);

		assert(start <= stop);
		assert(start >= start_);
		assert(stop <= stop_);

		initIndex(start, stop, 0); //x, x, start_
		return *this;
	}

	size_t MemoryCursor::remainingSize(void)
	{
		return stop_ - start_ - current_ + 1;
	}

	MemoryCursor MemoryCursor::neighbor(size_t len)
	{
		return MemoryCursor(buffer_, stop_ + 1, stop_ + len);
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, ipv4_t & data)
	{
		OPERATOR_READ(r32_addr);
		return cursor;
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, ipv6_t & data)
	{
		throw("Read IPv6 address not supported");
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, uint8_t & data)
	{
		OPERATOR_READ(r8);
		return cursor;
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, uint16_t & data)
	{
		OPERATOR_READ(r16);
		return cursor;
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, uint32_t & data)
	{
		OPERATOR_READ(r32);
		return cursor;
	}

	MemoryCursor & operator >> (MemoryCursor & cursor, uint64_t & data)
	{
		OPERATOR_READ(r64);
		return cursor;
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const ipv4_t data)
	{
		OPERATOR_WRITE(w32_addr);
		return cursor;
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const ipv6_t data)
	{
		throw("Write IPv6 address not supported");
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const uint8_t data)
	{
		OPERATOR_WRITE(w8);
		return cursor;
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const uint16_t data)
	{
		OPERATOR_WRITE(w16);
		return cursor;
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const uint32_t data)
	{
		OPERATOR_WRITE(w32);
		return cursor;
	}

	MemoryCursor & operator << (MemoryCursor & cursor, const uint64_t data)
	{
		OPERATOR_WRITE(w64);
		return cursor;
	}

	std::ostream & operator << (std::ostream & os, const MemoryCursor & mc)
	{
		//assert(mc.buffer_);
		if (mc.buffer_) {
			mc.buffer_->print(os, mc.start_, mc.stop_);
		}
		return os;
	}

	//
	// MemoryStream class
	//
	// For OMemoryStream
	MemoryStream::MemoryStream(void)
	{
		buffer_ = 0;
	}

	// For IMemoryStream
	MemoryStream::MemoryStream(uint8_t * buffer, size_t size)
	{
		buffer_ = new MemoryBuffer(buffer, size);
		assert(buffer_);
	}

	MemoryStream::~MemoryStream(void)
	{
		if (buffer_) {
			delete buffer_;
			buffer_ = 0;
		}
	}

	void MemoryStream::flushBuffer(void)
	{
		if (buffer_) {
			G2RSVPTE_DBG("Flushing the buffer of the MemoryStream");
			delete buffer_;
			buffer_ = 0;
		}
	}

	const uint8_t * MemoryStream::getBufferData(void)
	{
		if (buffer_)
			return buffer_->getData();
		else
			return 0;
	}

	size_t MemoryStream::getBufferDataSize(void) const
	{
		if (buffer_)
			return buffer_->size();
		else
			return 0;
	}

	MemoryStream & operator << (MemoryStream & ms, const Message & msg)
	{
		ms.flushBuffer();
		assert(ms.buffer_ == 0);

		ms.buffer_ = new MemoryBuffer(msg.getLength());
		assert(ms.buffer_);

		MemoryCursor mc(ms.buffer_->cursor());

		try {
			mc << msg;

		} catch (std::string e) {
			e = "exception on [MemStream << Msg]: " + e;
			throw(e);
		} catch (...) {
			std::string e = "UNKNOWN exception on "
				"[MemStream << Msg]";
			throw(e);
		}

		ms.buffer_->writeChecksum();

		return ms;
	}

	std::ostream & operator << (std::ostream & os, const MemoryStream & ms)
	{
		try {
			os << ms.buffer_->cursor();

		} catch (std::string e) {
			e = "exception on [OS << MemStream]: " + e;
			throw(e);
		} catch (...) {
			std::string e = "UNKNOWN exception on [OS<<MemStream]";
			throw(e);
		}
		return os;
	}

	MemoryStream & operator >> (MemoryStream & ms, Message * & msg)
	{
		if (!ms.buffer_) {
			throw(std::string("Impossible to build message if the "
					  "MemoryStream buffer is empty"));
		}

		if (!ms.buffer_->isChecksumOk()) {
			throw(std::string("Wrong Checksum!"));
		}

		MemoryCursor mc(ms.buffer_->cursor());

		try {
			mc >> msg;

		} catch (std::string e) {
			e = "exception on [MemStream >> Msg]: " + e;
			throw(e);
		} catch (...) {
			std::string e = "UNKNOWN exception on "
				"[MemStream >> Msg]";
			throw(e);
		}

		return ms;
	}
}
