/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#ifndef G2RSVPTE_BUFFER_H
#define G2RSVPTE_BUFFER_H

#include <assert.h>
#include <arpa/inet.h>

#include <stdint.h>
#include <iostream>
#include <iomanip>

#include "g2mpls_addr.h"
#include "g2rsvpte_utils.h"

namespace RSVP {

	class Message;
	class MemoryBuffer;
	class MemoryCursor;
	class MemoryStream;

//
// Local Utilities (usefull ONLY here!!!)
//
#define NTOHC
#define HTONC

#define WOP(NAME,TYPE,HTON)					\
	void NAME (size_t address, TYPE value) {		\
		assert(address < size_);			\
		assert(address + sizeof(TYPE) <= size_);	\
								\
		*((TYPE *)(buffer_ + address)) = HTON(value);	\
	}

#define ROP(NAME,TYPE,NTOH)					\
	TYPE NAME (size_t address) {				\
		assert(address < size_);			\
		assert(address + sizeof(TYPE) <= size_);	\
								\
		TYPE tmp;					\
		tmp = NTOH( *(TYPE *)(buffer_ + address) );	\
								\
		return tmp;					\
	}

	extern MemoryCursor & operator << (MemoryCursor &, const Message &);
	extern MemoryCursor & operator >> (MemoryCursor &, Message * &);

	class MemoryCursor {
		friend class MemoryBuffer;

		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   ipv4_t &       data);
		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   ipv6_t &       data);
		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   uint8_t &      data);
		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   uint16_t &     data);
		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   uint32_t &     data);
		friend MemoryCursor & operator >> (MemoryCursor & cursor,
						   uint64_t &     data);
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   Message * &    msg);

		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const ipv4_t   data);
		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const ipv6_t   data);
		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const uint8_t  data);
		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const uint16_t data);
		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const uint32_t data);
		friend MemoryCursor & operator << (MemoryCursor & cursor,
						   const uint64_t data);
		friend MemoryCursor & operator << (MemoryCursor &  cursor,
						   const Message & msg);

		friend std::ostream & operator << (std::ostream &       os,
						   const MemoryCursor & mb);
	public:
		// Add this method with the current of the new cursor = 0 ???
		//MemoryCursor(const MemoryCursor & cursor);

		MemoryCursor(MemoryBuffer * buffer,
			     size_t         start,
			     size_t         stop);
		~MemoryCursor(void);

		MemoryCursor & resize(size_t start, size_t stop);
		MemoryCursor neighbor(size_t len);
		size_t remainingSize(void);

		size_t start(void) { return start_; }
		size_t stop(void)  { return stop_;  }

	protected:
		MemoryCursor(void);

	private:
		MemoryCursor(MemoryBuffer * buffer);
		void initIndex(size_t start, size_t stop, size_t curr);

	private:
		// pointer to main buffer for this family of cursors
		MemoryBuffer * buffer_;

		// buffer index (range [0, size -1])
		size_t start_;
		size_t stop_;

		// Cursor index: range [0, stop_ - start_ + 1]
		// if (current == stop_ - start_ + 1)  =>  buffer is full
		size_t current_;
	};

	class MemoryBuffer {
		friend std::ostream & operator << (std::ostream &       os,
						   const MemoryBuffer & mb);
	public:
		// MemoryBuffer initialized copying the given buffer
		MemoryBuffer(uint8_t * buffer, size_t size);
		MemoryBuffer(size_t size);
		~MemoryBuffer(void);

		MemoryCursor & cursor(void);
		void print(std::ostream & os, size_t start, size_t stop) const;
		size_t size(void) const { return size_; }

		// Checksum utils
		uint16_t calculateChecksum(void) const;
		void     writeChecksum(void);
		bool     isChecksumOk(void);

		const uint8_t * getData(void);

		//
		// offset MUST have a range [0, size -1]
		//
		WOP(w8,  uint8_t,  HTONC);  // void  w8(size_t off, uint8_t d);
		WOP(w16, uint16_t, htons);
		WOP(w32, uint32_t, htonl);
		WOP(w64, uint64_t, htonll);
		WOP(w32_addr, ipv4_t, HTONC);

		ROP(r8,  uint8_t,  NTOHC);  // uint8_t r8(size_t off);
		ROP(r16, uint16_t, ntohs);
		ROP(r32, uint32_t, ntohl);
		ROP(r64, uint64_t, ntohll);
		ROP(r32_addr, ipv4_t, NTOHC);

	protected:
		MemoryBuffer(void);

	private:
		uint16_t calculateRSVPChecksum(void) const;

		uint16_t getChecksum(void) const {
			assert(size_ >= 4);
			return *(uint16_t *)(buffer_ + 2);
		}
		void setChecksum(uint16_t checksum) {
			assert(size_ >= 4);
			*(uint16_t *)(buffer_ + 2) = checksum;
		}

	private:
		uint8_t * buffer_;
		size_t    size_;
	};

	class MemoryStream {
		friend MemoryStream & operator >> (MemoryStream & ms,
						   Message * &    msg);
		friend MemoryStream & operator << (MemoryStream &  ms,
						   const Message & msg);

		friend std::ostream & operator << (std::ostream &       os,
						   const MemoryStream & ms);
	public:
		MemoryStream(void);
		MemoryStream(uint8_t * buffer, size_t size);
		~MemoryStream(void);

		void flushBuffer(void);

		const uint8_t * getBufferData(void);
		size_t          getBufferDataSize(void) const;

	protected:

	private:
		MemoryBuffer * buffer_;
	};
}
#endif // G2RSVPTE_BUFFER_H
