/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#ifndef G2RSVPTE_DM_H
#define G2RSVPTE_DM_H

#include <list>
#include <map>
#include <iterator>
#include <assert.h>
#include "thread.h"

#include <g2mpls_addr.h>
#include <g2mpls_types.h>

#include "g2rsvpted/g2rsvpte_config.h"
#include "g2rsvpted/g2rsvpte_messages.h"
#include "g2rsvpted/g2rsvpte_trace.h"
#include "g2rsvpted/g2rsvpte_scn_socket.h"
#include "g2rsvpted/g2rsvpte_lsp_gen.h"

#ifdef PERSISTENCY_ON
#include "g2rsvpted/g2rsvpte_persistency_py.h"
#endif // PERSISTENCY_ON

#define DEF_REFRESH_INTERVAL       10
#define DEF_RAPID_RETRANS_INTERVAL 10
#define DEF_RAPID_RETRY_LIMIT      10
#define DEF_EXPO_BACKOFF_DELTA     10

#define MAKE_SESS_KEY(DNID, TID)			\
	(((uint64_t)(DNID)) << 32 | ((uint64_t)(TID)))

#define MAKE_LSP_KEY(SNID, LSPID)			\
	(((uint64_t)(SNID)) << 32 | ((uint64_t)(LSPID)))


#define _DEFINE_MAP_ITERATOR(TD,T1,T2,P,V)		\
	typedef std::map<T1, T2>::iterator TD;		\
	TD begin##P(void) { return V.begin(); };	\
	TD end##P(void)   { return V.end();   };

#define _DEFINE_MAP_CONST_ITERATOR(TD,T1,T2,P,V)	\
	typedef std::map<T1, T2>::const_iterator TD;	\
	TD begin##P(void) const { return V.begin(); };	\
	TD end##P(void)   const { return V.end();   };

#define DEFINE_MAP_ITERATOR(N,T)					      \
   _DEFINE_MAP_ITERATOR(iterator_##N, T##Key_t, T *, _##N, N##_);	      \
   _DEFINE_MAP_CONST_ITERATOR(const_iterator_##N, T##Key_t, T *, _##N, N##_);

namespace RSVP {
	//
	// Forward decls
	//
	class Neighbour;
	class Peer;
	class G2RSVPTE;
	class LSP;
	class Session;

	//
	// Real decls
	//
	template <bool REP, class P> class Ancestor {
	public:
		Ancestor(P * p) {
			assert(p);

			parent_      = p;
			reparenting_ = REP;
		}
		~Ancestor(void) {
			// mindless
			parent_ = 0;
		}

		P * parent(P * p) {
			assert(reparenting_);
			assert(p);

			P * tmp;

			tmp = parent_;
			parent_ = p;

			return tmp;
		}

		P * parent(void) {
			return parent_;
		}
	private:
		P *  parent_;
		bool reparenting_;
	};

	class LSPElem : public Ancestor<true, Neighbour> {
	public:
		LSPElem(Neighbour * parent) :
			Ancestor<true, Neighbour>(parent) { };
		LSPElem(void);
	private:
	};

	typedef uint32_t LSPElemKey_t; // Fixme

	class Neighbour : public Ancestor<true, Peer> {
	public:
		Neighbour(Peer * parent) :
			Ancestor<true, Peer>(parent) { };
		~Neighbour(void);

		bool attach(LSPElemKey_t k , LSPElem * e);
		bool detach(LSPElemKey_t k);

		// Defines iterator_lsp_elems
		DEFINE_MAP_ITERATOR(lsp_elems, LSPElem);

	private:
		std::map<LSPElemKey_t, LSPElem *> lsp_elems_;
	};

	class Notify : public Ancestor<true, Peer>{
		Notify(Peer * parent) :
			Ancestor<true, Peer>(parent) { };
		~Notify(void);
	public:
	private:
	};

	class Peer : public Ancestor<true, G2RSVPTE> {
	public:
		Peer(G2RSVPTE * parent) :
			Ancestor<true, G2RSVPTE>(parent) {
			neighbour_ = 0;
			notify_    = 0;
		}
		~Peer(void);

		bool attach(Neighbour * n);
		bool detach(Neighbour * n);

		bool attach(Notify * n);
		bool detach(Notify * n);

	private:
		Neighbour * neighbour_;
		Notify *    notify_;
	};

	typedef uint32_t PeerKey_t; // Fixme

	class Interface : public Ancestor<true, G2RSVPTE> {
	public:
		Interface(G2RSVPTE * parent) :
			Ancestor<true, G2RSVPTE>(parent) {

			memset(&localId_, 0, sizeof(localId_));
			memset(&remoteId_, 0, sizeof(remoteId_));

			neighbourId_ = 0;
			op_state_    = DOWN;
			adm_state_   = DISABLED;
		};
		Interface(G2RSVPTE *    parent,
			  g2mpls_addr_t localId,
			  g2mpls_addr_t remoteId,
			  uint32_t      nId,
			  opstate_t     op_state,
			  admstate_t    adm_state);

		~Interface(void) { };

		g2mpls_addr_t remoteId(void) const        { return remoteId_; }
		void          remoteId(g2mpls_addr_t add) {
			addr_copy(&remoteId_, &add);
		}

		g2mpls_addr_t localId(void) const         { return localId_; }
		void          localId(g2mpls_addr_t add)  {
			addr_copy(&localId_, &add);
		}

		opstate_t     opState(void) const         { return op_state_;  }
		void          opState(opstate_t state)    { op_state_ = state; }

		admstate_t    admState(void) const        { return adm_state_; }
		void          admState(admstate_t state)  { adm_state_ = state;}

		uint32_t      neighbourId(void) const   { return neighbourId_; }
		void          neighbourId(uint32_t nId) { neighbourId_ = nId;  }

		void          dump(void) const;

		friend std::ostringstream & operator << (std::ostringstream &
							 os,
							 const Interface &
							 xsintf);

	private:
		g2mpls_addr_t localId_;
		g2mpls_addr_t remoteId_;

		uint32_t      neighbourId_;

		opstate_t     op_state_;
		admstate_t    adm_state_;
	};

	typedef uint32_t InterfaceKey_t; // Fixme

	class AckNack : public Ancestor<true, LSP> {
	public:
		AckNack(LSP * parent, bool useFsm) :
			Ancestor<true, LSP>(parent),
			useFsm_(useFsm) { }
		~AckNack(void) { }

	protected:
		bool      useFsm_;
		SCNSocket scngw_;

		// ACK/NACK FSM instance
	};

	class UpstreamAckNack : public AckNack {
	public:
		UpstreamAckNack(LSP * parent, bool useFsm = false) :
			AckNack(parent, useFsm) { }
		~UpstreamAckNack(void) { }

		bool sendMsg(Message * msg);
		bool errorHandle(int  flags,
				 int  code,
				 int  val,
				 bool addInfo = false,
				 bool xcErr = false);

	private:
	};

	class DownstreamAckNack : public AckNack {
	public:
		DownstreamAckNack(LSP * parent, bool useFsm = false) :
			AckNack(parent, useFsm) { }
		~DownstreamAckNack(void) { }

		bool sendMsg(Message * msg);
		bool errorHandle(int  flags,
				 int  code,
				 int  val,
				 bool addInfo = false,
				 bool xcErr = false);

	private:
	};

	class StateBlock {
	public:
		StateBlock(void) {
			intf_            = 0;
			nodeId_          = 0;
			lih_             = 0;
			tx_label_        = 0;
			rx_label_        = 0;
			refreshTimeout_  = 0;
			refreshInterval_ = 0;
			memset(&remoteDL_, 0, sizeof(remoteDL_));
			memset(&localDL_, 0, sizeof(localDL_));
		}
		~StateBlock(void) { }

		uint32_t lih(void)           { return lih_; }
		void     lih(uint32_t lih)   { lih_ = lih;   }

		uint32_t nodeId(void)        { return nodeId_; }
		void     nodeId(uint32_t id) { nodeId_ = id;   }

		Interface * interface(void)  { return intf_;   }
		void        interface(Interface * intf, bool check = true) {
			if (check) {
				assert(intf);
			}
			intf_ = intf;
		}

		g2mpls_addr_t remoteDL(void)              { return remoteDL_; }
		void          remoteDL(g2mpls_addr_t add) {
			addr_copy(&remoteDL_, &add);
		}

		g2mpls_addr_t localDL(void)               { return localDL_; }
		void          localDL(g2mpls_addr_t add)  {
			addr_copy(&localDL_, &add);
		}

		uint32_t txLabel(void)           { return tx_label_;  }
		void     txLabel(uint32_t label) { tx_label_ = label; }

		uint32_t rxLabel(void)           { return rx_label_;  }
		void     rxLabel(uint32_t label) { rx_label_ = label; }

		uint32_t refreshTimeout(void)    { return refreshTimeout_; }
		void     refreshTimeout(uint32_t rTimeout) {
			refreshTimeout_ = rTimeout;
		}

		uint32_t refreshInterval(void)   { return refreshInterval_; }
		void     refreshInterval(uint32_t rInterval) {
			refreshInterval_ = rInterval;
		}

	private:
		g2mpls_addr_t remoteDL_; // remote data link (used by
					 //  next/previous HOP)
		g2mpls_addr_t localDL_;  // upstream/downstream data link (used
					 //  to go to previous/next HOP)

		Interface * intf_;     // interface to next/previous HOP

		uint32_t    nodeId_;   // next/previous HOP node id
		uint32_t    lih_;      // next/previous logical interf handler


		// TMP label as uint32 ???
		uint32_t    tx_label_; // upstream/downstream label (used to
				       // tx to previous/next HOP)
		uint32_t    rx_label_; // upstream/downstream label (used to
				       // rx from previous/next HOP)

		uint32_t    refreshTimeout_;  //used Path/Resv refresh timeout
		uint32_t    refreshInterval_; //used Path/Resv refresh interval

		//refreshTimeoutHandle;//zebra Path/Resv refresh timeout handle
		//refreshTimerHandle;  //zebra Path/Resv refresh timer handle
	};

	class PSB : public StateBlock {
	public:
		PSB(void) {
			pathIn_   = 0;
			pathOut_  = 0;
			pathErr_  = 0;
			resvConf_ = 0;
			pathDown_ = 0;
			pathTear_ = 0;
		}
		~PSB(void) {
			DELETE_OBJ(pathIn_);
			DELETE_OBJ(pathOut_);
			DELETE_OBJ(pathErr_);
			DELETE_OBJ(resvConf_);
			DELETE_OBJ(pathDown_);
			DELETE_OBJ(pathTear_);
		}

		PathMessage * pathOut(void) {
			return pathOut_;
		}
		// TODO: change into attachPathOut ...
		void          pathOut(PathMessage * m) {
			//assert(m);
			pathOut_ = m;
		}
		PathMessage * pathIn(void) {
			return pathIn_;
		}
		void          pathIn(PathMessage * m) {
			//assert(m);
			pathIn_ = m;
		}

		// TMP
		PathErrMessage * pathErr(void) {
			return pathErr_;
		}
		void          pathErr(PathErrMessage * m) {
			//assert(m);
			pathErr_ = m;
		}

		ResvConfMessage * resvConf(void) {
			return resvConf_;
		}
		void          resvConf(ResvConfMessage * m) {
			//assert(m);
			resvConf_ = m;
		}

		PathMessage * pathDown(void) {
			return pathDown_;
		}
		void          pathDown(PathMessage * m) {
			//assert(m);
			pathDown_ = m;
		}

		PathTearMessage * pathTear(void) {
			return pathTear_;
		}
		void          pathTear(PathTearMessage * m) {
			//assert(m);
			pathTear_ = m;
		}

	private:
		PathMessage *     pathIn_;   // received Path msg
		PathMessage *     pathOut_;  // transmitted Path msg

		PathErrMessage *  pathErr_;  // TMP rx/tx Path Err msg
		ResvConfMessage * resvConf_; // TMP rx/tx ResvConf msg
		PathMessage *     pathDown_; // TMP rx/tx Path (D=1 R=1) msg
		PathTearMessage * pathTear_; // TMP rx/tx PathTear msg
	};


	class RSB : public StateBlock {
	public:
		RSB(void) {
			resvIn_   = 0;
			resvOut_  = 0;
			resvDown_ = 0;
			resvTear_ = 0;
			resvErr_  = 0;
		}
		~RSB(void) {
			DELETE_OBJ(resvIn_);
			DELETE_OBJ(resvOut_);
			DELETE_OBJ(resvDown_);
			DELETE_OBJ(resvTear_);
			DELETE_OBJ(resvErr_);
		}

		ResvMessage * resvOut(void) {
			return resvOut_;
		}
		// TODO: change into attachResvOut ...
		void          resvOut(ResvMessage * m) {
			//assert(m);
			resvOut_ = m;
		}
		ResvMessage * resvIn(void) {
			return resvIn_;
		}
		void          resvIn(ResvMessage * m) {
			//assert(m);
			resvIn_ = m;
		}

		//TMP
		ResvMessage * resvDown(void) {
			return resvDown_;
		}
		void          resvDown(ResvMessage * m) {
			//assert(m);
			resvDown_ = m;
		}

		ResvTearMessage * resvTear(void) {
			return resvTear_;
		}
		void          resvTear(ResvTearMessage * m) {
			//assert(m);
			resvTear_ = m;
		}

		ResvErrMessage * resvErr(void) {
			return resvErr_;
		}
		void          resvErr(ResvErrMessage * m) {
			//assert(m);
			resvErr_ = m;
		}

	private:
		ResvMessage *     resvIn_;   // received Resv msg
		ResvMessage *     resvOut_;  // transmitted Resv msg

		ResvMessage *     resvDown_; // TMP rx/tx Resv (D=1 R=1) msg
		ResvTearMessage * resvTear_; // TMP rx/tx ResvTear msg
		ResvErrMessage *  resvErr_;  // TMP rx/tx ResvTear msg
	};

	class LSPCtrl {
	public:
		LSPCtrl(void) { };
		~LSPCtrl(void) { };

	private:
	};

	class LSP : public Ancestor<true, Session> {
		friend std::ostringstream & operator<< (std::ostringstream & os,
							const LSP &          l);
	public:
		LSP(Session * parent);
		LSP(Session * parent, uint32_t nId, uint32_t lspId);
		~LSP(void);

		bool attach(LSPCtrl * ctrl);
		bool detach(LSPCtrl * ctrl);

		bool attach(UpstreamAckNack * u);
		bool detach(UpstreamAckNack * u);

		bool attach(DownstreamAckNack * d);
		bool detach(DownstreamAckNack * d);

		bool attach(PSB * psb);
		bool detach(PSB * psb);

		bool attach(RSB * rsb);
		bool detach(RSB * rsb);

		bool                isEnabled(void) const { return enabled_;   }
		uint32_t            nid(void) const       { return nid_;       }
		uint32_t            id(void) const        { return id_;        }
		net_res_spec_t      iRes(void) const      { return in_net_res_;}
		net_res_spec_t      eRes(void) const      { return eg_net_res_;}
#ifdef CRANKBACK_ON
		crankback_scope_t   cbackScope(void) const { return cbScope_;  }
	        uint32_t            maxCbackRetries(void) const {
			return maxCbackRetries_;
		}
		uint32_t            currentCbackAttempt(void) const { return currentCbackAtt_;}
		std::list<lsp_ero_sobj_t> getExcludeEro(void) const {
			return excludeEro_;
		}
		void                clearExcludeEro(void) { excludeEro_.clear();}
		bool                 cbackWaitReleaseRes(void) const {
			return cbackWaitRelRes_;
		}
#endif // CRANKBACK_ON
		call_type_t         callType(void) const  { return callType_;  }
		call_ident_t        callId(void) const    { return callId_;    }
		call_ident_t        callId_old(void) const { return callId_old_;}
		PSB *               psb(void)       { return psb_;        }
		RSB *               rsb(void)       { return rsb_;        }
		LSPCtrl *           lspCtrl(void)   { return lspCtrl_;    }
		UpstreamAckNack *   usAckNack(void) { return us_AckNack_; }
		DownstreamAckNack * dsAckNack(void) { return ds_AckNack_; }
		uint32_t            tnrcCookie(void) const {return tnrcCookie_;}

		//void     iTNA(g2mpls_addr_t addr) {
		//	addr_copy(&ingress_tna_, &addr);
		//}
		//void     eTNA(g2mpls_addr_t addr) {
		//	addr_copy(&egress_tna_, &addr);
		//}

		void iRes(net_res_spec_t res) { in_net_res_ = res; }
		void eRes(net_res_spec_t res) { eg_net_res_ = res; }

#ifdef CRANKBACK_ON
		void cbackScope(crankback_scope_t cb) { cbScope_ = cb; }
		void currentCbackAttemptIncr(void)    { currentCbackAtt_++; }
		void maxCbackRetries(uint32_t max)    { maxCbackRetries_ = max; }
		void attachExcludeEro(std::list<lsp_ero_sobj_t> exclEro);
		void cbackWaitReleaseRes(bool wait)   { cbackWaitRelRes_ = wait; }
#endif // CRANKBACK_ON

		void callType(call_type_t type)  { callType_ = type;  }
		void callId(call_ident_t Id)     { callId_ = Id;      }
		void callId_old(call_ident_t Id) { callId_old_ = Id;  }

		bool updateCallId(call_ident_t Id);
		bool getCallId(call_ident_t & Id);

		void tnrcCookie(uint32_t cookie) { tnrcCookie_ = cookie; }

		lsp_ident_t getIdent(void);
		bool getInfo(lsp_info_t & info);
		bool getCallInfo(call_info_t & info);
		std::string sessionName(void) const;

		// Time functions
		uint32_t refreshInterval(void)  { return refreshInterval_;  }
		uint32_t rapidRetryLimit(void)  { return rapidRetryLimit_;  }
		uint32_t expoBackoffDelta(void) { return expoBackoffDelta_; }
		uint32_t rapidRetransInterval(void) {
			return rapidRetransInterval_;
		}
		times_t times(void) const       { return times_;}

		void     refreshInterval(uint32_t time) {
			refreshInterval_ = time;
		}
		void     rapidRetransInterval(uint32_t time) {
			rapidRetransInterval_ = time;
		}
		void     rapidRetryLimit(uint32_t time) {
			rapidRetryLimit_ = time;
		}
		void     expoBackoffDelta(uint32_t time) {
			expoBackoffDelta_ = time;
		}
		void     times(times_t timesValue) {
			times_ = timesValue;
		}

		uint32_t context(void) const       { return context_; }

		// LSP methods
		bool eroProcess(bool recursive = true);
		bool loopDetect(void);
		bool setDsDL_checkDsTxLbl_fromTELink(g2mpls_addr_t teLink,
						     uint32_t      label);
		bool setDsDL_DsTxLbl_fromTELink(g2mpls_addr_t teLink);
		bool setUsDL_UsRxLbl_fromTELink(g2mpls_addr_t teLink);
		bool checkSet_UsTEL_UsDL_UsRxLbl(g2mpls_addr_t teLink,
						 g2mpls_addr_t dataLink,
						 uint32_t      lbl);
		bool checkSet_DsTEL_DsDL_DsTxLbl(g2mpls_addr_t teLink,
						 g2mpls_addr_t dataLink,
						 uint32_t      lbl);
		bool checkSet_DsTEL_DsDL_DsTxLbl_fromTna(g2mpls_addr_t tna,
							 g2mpls_addr_t dataLink,
							 uint32_t      lbl);
		bool setDsTEL_DsDL_DsTxLbl_fromTna(g2mpls_addr_t tna);
		// For APIs
		bool signalUpLSP(bool enqueueEvent = false);
		bool signalDownLSP(void);
		bool enableLSP(void);
		bool disableLSP(void);
		bool attachEroSubObj(EroSubObject * eroSubObj,
				     bool           insertTail = true);
		bool cleanEro(void);
		bool cleanRsvpHop(void);
		bool cleanLabelObjs(void);
		bool cleanSBs(void);
		bool setTna(const net_res_spec & tna,
			    bool                 ingress);
		bool updateTna(const net_res_spec & tna,
			       bool                 ingress);
		bool getTna(bool           ingress,
			    net_res_spec & tna);
		bool getTnResource(net_res_spec & tnRes);
		bool setGnsTna(const grid_res_spec_t & gnsTna,
			       bool                    ingress);
		bool getGnsTna(bool              ingress,
			       grid_res_spec_t & gnsTna);
		bool getPolicy(std::string & authz);
		bool getErrInfo(error_info_t &            error,
				std::list<lsp_ero_sobj_t> excludeEro,
				bool                      lookAtPath);

		bool sendPath(bool enqueueEvent = false);
		bool sendResv(bool enqueueEvent = false);
		bool sendDown(bool enqueueEvent = false);
		bool sendTear(bool enqueueEvent = false);
		bool sendConfirm(bool enqueueEvent = false);
		bool sendError(const error_info_t & error,
			       bool enqueueEvent = false);

		bool xConnCompleted(bool enqueueEvent = false);
		bool xConnFailed(bool enqueueEvent = false);

		// Utilities
		bool            isHead(void);
		bool            isTail(void);
		bool            checkGnsTna(bool ingress);
		bool            checkPolicy(void);
		bool            checkEro(void);
		G2RSVPTE *      getInstance(void);
		bool            toBeDestroyed(void) const      { return toBeDestroyed_; }
		void            toBeDestroyed(bool tbd)        { toBeDestroyed_ = tbd;  }
		bool            isErrored(void) const          { return isErrored_;     }
		void            isErrored(bool isErr)          { isErrored_ = isErr;    }
		g2mpls_addr_t   errTELink(void)                { return errTELink_;     }
		void            errTELink(g2mpls_addr_t add)   {
			addr_copy(&errTELink_, &add);
		}
		void            resetErr(void)                 {
			memset(&errTELink_, 0, sizeof(errTELink_));
			errLabel_ = 0;
		}
		uint32_t        errLabel(void)                 { return errLabel_;      }
		void            errLabel(uint32_t lbl)         { errLabel_ = lbl;       }
		struct thread * FSMtimerThr(void) const        { return timerThr_;      }
		void            FSMtimerThr(struct thread * t) { timerThr_ = t;         }

#ifdef CRANKBACK_ON
		bool processLSPAttrObj(void);
#endif // CRANKBACK_ON

		bool processNHOPobj(void);
		bool prepareNHOPobj(void);
		bool processLabelObj(void);
		bool prepareLabelObj(void);
		bool insertIntoRRO(RecordedRoute_Class * rro,
				   g2mpls_addr_t         addr);
		bool processPHOPobj(void);
		bool preparePHOPobj(void);
		bool processLabelSetObj(void);
		bool prepareLabelSetObj(void);
		bool processUpstreamLabelObj(void);
		bool prepareUpstreamLabelObj(void);
		bool processGenUniObj(void);
		//bool prepareGenUniObj(void);
		bool processTimesObj(void);
		bool processPolicyData(void);
		bool processCallIdObj(bool lookAtPath);
		bool matchCallId(CallId_Class * callObj);

	private:
		// Utils
		Interface * getInterfaceFromTELink(g2mpls_addr_t addr,
						   bool          checkRemote);
		uint32_t    getNbrId(void);
		bool        eroComplete(void);
		bool        initConf(void);

	private:
		LSPCtrl *           lspCtrl_;
		UpstreamAckNack *   us_AckNack_;
		DownstreamAckNack * ds_AckNack_;

		PSB *               psb_;
		RSB *               rsb_;

		uint32_t            tnrcCookie_;     // TNRC cookie for the LSP (>0)
		                                     // if tnrcCookie_ == 0 no action
		                                     // executed on TNRC

		uint32_t            nid_;            // source node id
		uint32_t            id_;             // LSP id
		net_res_spec_t      in_net_res_;
		net_res_spec_t      eg_net_res_;

#ifdef CRANKBACK_ON
		crankback_scope_t         cbScope_;
		uint32_t                  currentCbackAtt_;
		uint32_t                  maxCbackRetries_;
		std::list<lsp_ero_sobj_t> excludeEro_;
		bool                      cbackWaitRelRes_;
#endif // CRANKBACK_ON
		call_type_t         callType_;
		call_ident_t        callId_;
		call_ident_t        callId_old_;

		uint32_t            numHops_;        // ???

		// Flags
		bool                enabled_;
		bool                toBeDestroyed_;
		bool                isErrored_;

		g2mpls_addr_t       errTELink_;
		uint32_t            errLabel_;

		// FSM state timer thread
		struct thread *     timerThr_;
		// time intervals
		uint32_t            refreshInterval_;      // refresh interval
		uint32_t            rapidRetransInterval_; // retrans. interval
		uint32_t            rapidRetryLimit_;      // retry limit
		uint32_t            expoBackoffDelta_;     // incr value delta
		times_t             times_;                // start and end time
		uint32_t            context_;              // UID for the LSP
	public:
		// LSP FSM instance
		fsm::G2RSVPTE_LSP_FSM::virtFsm * fsmInst_;
	};

	typedef uint64_t LSPKey_t; // Fixme

	class Session : public Ancestor<true, G2RSVPTE> {
		friend std::ostringstream & operator<< (std::ostringstream & os,
							const Session &      s);
	public:
		Session(G2RSVPTE * parent);
		Session(G2RSVPTE * parent,
			uint32_t   nId,
			uint32_t   tunId,
			uint32_t   extTunId);
		~Session(void);

		bool     attach(LSPKey_t k, LSP * l);
		bool     detach(LSPKey_t k, LSP * l);

		// return the number of LSPs attached at this session
		uint32_t size(void)  { return lsps_.size();  }
		bool     empty(void) { return lsps_.empty(); }

		// Defines iterator_lsps and methods: begin_lsps/end_lsps
		DEFINE_MAP_ITERATOR(lsps, LSP);

		LSP *    findLSP(LSPKey_t key);
		LSP *    findLSP(uint32_t context);
		bool     checkLSP(LSP * lsp);//VERY TMP

		uint32_t nodeId(void) const    { return nodeId_; }
		void     nodeId(uint32_t id)   { nodeId_ = id;   }

		uint32_t tunnelId(void) const  { return tunnelId_; }
		void     tunnelId(uint32_t id) { tunnelId_ = id;   }

		uint32_t extTunId(void) const  { return extTunId_; }
		void     extTunId(uint32_t id) { extTunId_ = id;   }

	private:
		uint32_t nodeId_;      // Destination Node Id
		uint32_t tunnelId_;    // Tunnel Id
		uint32_t extTunId_;    // Extended Tunnel Id

		std::map<LSPKey_t, LSP *> lsps_;
	};

	typedef  uint64_t SessionKey_t; // Fixme

	class PendingMessage : public Ancestor<true, G2RSVPTE> {
		PendingMessage(G2RSVPTE * parent) :
			Ancestor<true, G2RSVPTE>(parent) { };
		~PendingMessage(void);
	public:
	private:
	};

	//
	// Data Model utils
	//
	InterfaceKey_t g2mpls_addr2key(g2mpls_addr_t addr);

	typedef uint32_t PendingMessageKey_t;


	class G2RSVPTE_InstanceType {
	public:
		G2RSVPTE_InstanceType(void);
		G2RSVPTE_InstanceType(adj_type_t adjType, bool tunneledPackets);
		virtual ~G2RSVPTE_InstanceType(void);

		struct thread * readThread(void);
		void            readThread(struct thread * thr);

		void        initScngw(void);
		adj_type_t  adjType(void)          { return adjType_;     }
		bool        tunneledPackets(void)  { return tunneledPackets_; }
		bool        isUniClient(void)      { return isUniClient_; }
		void        isUniClient(bool isCl) { isUniClient_ = isCl; }
		bool        isCore(void)           { return isCore_;      }
		void        isCore(bool isC)       { isCore_ = isC;       }
		int         scngwFd(void)          { return scngwFd_;     }
		void        scngwFd(int fd)        { scngwFd_ = fd;       }

	protected:
		struct thread * readThread_;      // read thread from SCNgw
		adj_type_t      adjType_;         // type of instance (adjacency)
		bool            tunneledPackets_; // SCNgw tunnel/no tunnel
		bool            isUniClient_;     // type of UNI instance  (c/n)
		bool            isCore_;          // type of INNI instance (border/core)
		int             scngwFd_;         // SCNgw file descriptor
	};

	class G2RSVPTE : public G2RSVPTE_InstanceType {
	public:
		G2RSVPTE(void);
		G2RSVPTE(uint32_t defaultRefreshInterval,
			 uint32_t defaultRapidRetransInterval,
			 uint32_t defaultRapidRetryLimit,
			 uint32_t defaultExpoBackoffDelta);
		G2RSVPTE(adj_type_t adjType, bool tunneledPackets);
		~G2RSVPTE(void);

		bool attach(PeerKey_t k, Peer * p);
		// bool detach(PeerKey_t k);

		bool attach(InterfaceKey_t l, Interface * e);
		// bool detach(InterfaceKey_t l);

		bool attach(SessionKey_t k, Session * e);
		bool detach(SessionKey_t k, Session * e);

		bool attach(PendingMessageKey_t k, PendingMessage * e);
		// bool detach(PendingMessageKey_t k);

		// Defines iterator_peers
		DEFINE_MAP_ITERATOR(peers,            Peer);

		// Defines iterator_interfaces
		DEFINE_MAP_ITERATOR(interfaces,       Interface);

		// Defines iterator_sessions
		DEFINE_MAP_ITERATOR(sessions,         Session);

		// Defines iterator_pending_messages
		DEFINE_MAP_ITERATOR(pending_messages, PendingMessage);

		uint32_t    nodeId(void)        { return nodeId_; }
		void        nodeId(uint32_t id) { nodeId_ = id;   }

		//
		// Session utils
		//
		Session * findSession(SessionKey_t key);

		//
		// Interface utils
		//
		Interface * findInterface(InterfaceKey_t key);
		Interface * findInterface(g2mpls_addr_t  addr,
					  bool           checkRemote);
		// returns the number of loaded interfaces
		int         loadInterfaces(void);

		//
		// LSP utils
		//
		LSP * findLSP(lsp_ident_t info);
		LSP * findLSP(uint32_t    context);
		bool  checkLSP(LSP * lsp);//VERY TMP

		LSP * createLSP(const lsp_ident_t & ident,
				Interface *         intf,
				Message *           msg);

		//TODO:
		// - Add callIdent parameters (callIdType, sourceId, localId)
		// - Add missing lspParams ()
		LSP * createLSP(const lsp_ident_t &      ident,
				const std::string &      sessionName,
				const call_ident_t &     callIdent,
				const net_res_spec_t &   iRes,
				const net_res_spec_t &   eRes,

				//LSP params:
				const lsp_info_t &       lspInfo,

				//LSP recovery params
				const recovery_info_t &  recoveryInfo
				);

		LSP * createLSP(const lsp_ident_t &      ident,
				const std::string &      sessionName,
				const call_ident_t &     callIdent,
				const call_info_t &      callInfo,
				const lsp_info_t &       lspInfo,
				const recovery_info_t &  recoveryInfo,
				const std::string &      aaaInfo);

		bool  destroyLSP(const lsp_ident_t & id, bool internal = true);

#ifdef PERSISTENCY_ON
		void          databaseSync(void);

		bool          retrieveLsp(const lsp_ident_t ident,
					  lsp_ident_t  &    nextIdent,
					  std::string &     resp);

		PersistencyPy persistency_;
#endif // PERSISTENCY_ON
	private:
		void  initConf(void);
		bool  removeAllInterfaces(void);

		//TODO change this methods according to the public interface!!!
		LSP * createLSP_(const lsp_ident_t &       id,
				 const net_res_spec_t *    iRes,
				 const net_res_spec_t *    eRes,
				 const std::string &       sessionName,
				 const call_ident_t *      callIdent,
				 const lsp_info_t &        lspInfo,
				 const recovery_info_t *   recoveryInfo,
				 bool                      internal,
				 bool                      isRebuilded = false);

		std::map<PeerKey_t,
			Peer *>              peers_;
		std::map<InterfaceKey_t,
			Interface *>         interfaces_;
		std::map<SessionKey_t,
			Session *>           sessions_;
		std::map<PendingMessageKey_t,
			PendingMessage *>    pending_messages_;

		uint32_t   nodeId_;                      // router id

		uint32_t   defaultRefreshInterval_;      // refresh interval
		uint32_t   defaultRapidRetransInterval_; // retrans. interval
		uint32_t   defaultRapidRetryLimit_;      // retry limit
		uint32_t   defaultExpoBackoffDelta_;     // incr value delta
	};
};
#endif /* G2RSVPTE_DM_H */
