//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"

#include <map>
#include <list>
#include <string>
#include <sstream>
//#include <iomanip>

#include "g2rsvpted.h"
#include "g2mpls_utils.h"
#include "g2rsvpte_apis.h"

typedef uint64_t mapKey_t;	// tunnelId + dstAddr

#define LSPMAP_FIND(MAP, TYPE, KEY, RES) {		\
	std::map<mapKey_t, TYPE >::iterator itmp;	\
							\
	(RES) = 0;					\
	itmp = (MAP).find(KEY);				\
	if (itmp != (MAP).end()) {			\
		(RES) = (*itmp).second;			\
	}						\
}

#define LSPMAP_COND_CREATE_AND_BIND(MAP, TYPE, KEY, RES, ERR) {		\
	if (!RES) {							\
		(RES) = new (TYPE);					\
		if (!RES) {						\
			(ERR) += "Cannot allocate a VTY LSP container";	\
			(MAP).erase(KEY);				\
			return false;					\
		}							\
									\
		memset((RES), 0, sizeof(*(RES)));			\
		(MAP)[KEY] = (RES);					\
	}								\
}


static bool mangleLspKey(uint16_t    tun_id,
			 uint32_t    dst_nid,
			 mapKey_t *  key)
{
	assert(key);

	*key = ((((mapKey_t) (tun_id)) << 32) |
		(((mapKey_t) (dst_nid)) <<  0));

	return true;
}

static bool demangleLspKey(mapKey_t   key,
			   uint16_t * tun_id,
			   uint32_t * dst_nid)
{
	assert(tun_id);
	assert(dst_nid);

	*tun_id  = ((uint32_t) (((key) >> 32) & 0x0000FFFF));
	*dst_nid = ((uint32_t) (((key) >>  0) & 0xFFFFFFFF));

	return true;
}

static std::string key2string(const mapKey_t & key)
{
	std::string        tmp;
	std::ostringstream tunnelId;
	std::ostringstream dstRouterId;
	uint16_t           tun_id;
	struct in_addr     dst_nid;

	tmp = std::string("");

	if (!demangleLspKey(key, &tun_id, &(dst_nid.s_addr))) {
		return tmp;
	}

	//tunnelId << "TUN-ID: 0x" << hex << setw(8) << setfill('0') << tun_id;
	tunnelId    << "TUN-ID: "  << tun_id;
	dstRouterId << "DST-NID: " << inet_ntoa(dst_nid);


	tmp += tunnelId.str();
	tmp += std::string(" / ");
	tmp += dstRouterId.str();

	return tmp;
}


//
//
//  LSP related containers
//
//
typedef struct lspData {
	uint32_t         ext_tun_id;
	uint16_t         lsp_id;
	std::string *    sessionName;
	net_res_spec_t   iRes;
	net_res_spec_t   eRes;
	sw_cap_t         swcap;
	enc_type_t       enctype;
	gmpls_bwenc_t    bw;
	gpid_t           gpid;
	uint32_t         setupPrio;
	uint32_t         holdingPrio;
	lsp_type_t       type;
	lsp_res_action_t action;
	lsp_rro_mode_t   rroMode;
	uint32_t         refresh;
	bool             activateAck;
	uint32_t         rapidRetransmissionInterval;
	uint32_t         rapidRetryLimit;
	uint32_t         incrementValueDelta;
	times_t          times;
} lspData_t;

std::map<mapKey_t, lspData_t *>  vtyLspContainer;

bool
vtyLspCommittable(const mapKey_t & lspKey)
{
	lspData_t *   data;
	g2mpls_addr_t nullAddr;

	memset(&nullAddr, 0, sizeof(nullAddr));

	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);

	return (data &&
		//(addr_equal(data->itna, nullAddr) != 1) &&
		//(addr_equal(data->etna, nullAddr) != 1) &&
		isNetResSpecValid(data->iRes) &&
		isNetResSpecValid(data->eRes) &&
		(data->swcap   != SWCAP_UNKNOWN) &&
		(data->enctype != ENCT_UNKNOWN) &&
		(data->bw      != BwEnc_Unknown));
}


void
logDump(const net_res_spec_t & net)
{
	zlog_debug("\t   - network data:");
	zlog_debug("\t       - tna                    : %s",
		   BITMASK_BITTEST(net.mask_, tna) ?
		   addr_ntoa(net.tna) : "undeclared");
	zlog_debug("\t       - data link              : %s",
		   BITMASK_BITTEST(net.mask_, data_link) ?
		   addr_ntoa(net.data_link) : "undeclared");
	zlog_debug("\t       - label                  : %s",
		   BITMASK_BITTEST(net.mask_, label) ?
		   label2string(net.label).c_str() : "undeclared");
}

void
vtyLspDump(const mapKey_t & lspKey)
{
	lspData_t *       data;

	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);

	if (!data) {
		zlog_debug("LSP with %s not present in VTY containers",
			   key2string(lspKey).c_str());
		vtyLspContainer.erase(lspKey);
		return;
	}

	zlog_debug("LSP identity:");
	zlog_debug("\t %s", key2string(lspKey).c_str());
	zlog_debug("\t   - committable                : %s",
		   vtyLspCommittable(lspKey) ? "YES" : "NO");
	zlog_debug("\t   - ext_tun_id                 : %d",
		   data->ext_tun_id);
	zlog_debug("\t   - lspid                      : %d",
		   data->lsp_id);
	zlog_debug("\t   - sessionName                : %s",
		   (data->sessionName) ? (data->sessionName)->c_str() : "");
	//zlog_debug("\t   - itna                       : %s",
	//	   addr_ntoa(data->itna));
	//zlog_debug("\t   - etna                       : %s",
	//	   addr_ntoa(data->etna));
	logDump(data->iRes);
	logDump(data->eRes);
	zlog_debug("\t   - swcap                      : %s",
		   SHOW_SW_CAP(data->swcap));
	zlog_debug("\t   - enctype                    : %s",
		   SHOW_ENC_TYPE(data->enctype));
	zlog_debug("\t   - bw                         : %s",
		   SHOW_GMPLS_BWENC(data->bw));
	zlog_debug("\t   - gpid                       : %s",
		   SHOW_GPID(data->gpid));
	if (data->times.start_time && data->times.end_time) {
		zlog_debug("\t   - start UNIX time            : %d",
			   data->times.start_time);
		zlog_debug("\t   - end   UNIX time            : %d",
			   data->times.end_time);
	}
	zlog_debug("\t   - setupPrio                  : %d",
		   data->setupPrio);
	zlog_debug("\t   - holdingPrio                : %d",
		   data->holdingPrio);
	zlog_debug("\t   - type                       : %s",
		   SHOW_LSP_TYPE(data->type));
	zlog_debug("\t   - action                     : %s",
		   SHOW_LSP_RES_ACTION(data->action));
	zlog_debug("\t   - rroMode                    : %s",
		   SHOW_LSP_RRO_MODE(data->rroMode));
	zlog_debug("\t   - refresh                    : %d ms",
		   data->refresh);
	zlog_debug("\t   - activateAck                : %s",
		   data->activateAck ? "YES" : "NO");
	zlog_debug("\t   - rapidRetransmissionInterval: %d ms",
		   data->rapidRetransmissionInterval);
	zlog_debug("\t   - rapidRetryLimit            : %d",
		   data->rapidRetryLimit);
	zlog_debug("\t   - incrementValueDelta        : %d",
		   data->incrementValueDelta);
}

bool
vtyLspUpdate(const mapKey_t     & lspKey,
	     const uint32_t     & ext_tun_id,
	     const uint16_t     & lsp_id,
	     const std::string  & sessionName,
	     std::string        & err)
{
	lspData_t * data;

	err = "";
	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);
	LSPMAP_COND_CREATE_AND_BIND(vtyLspContainer, lspData_t,
				    lspKey, data, err);

	data->ext_tun_id  = ext_tun_id;
	data->lsp_id      = lsp_id;
	data->sessionName = new std::string(sessionName);
	if (!data->sessionName) {
		err += "Cannot allocate sessionName for LSP in VTY";
		return false;
	}

	return true;
}

bool
vtyLspUpdate(const mapKey_t         & lspKey,
	     //const g2mpls_addr_t    & itna,
	     //const g2mpls_addr_t    & etna,
	     const net_res_spec_t   & iRes,
	     const net_res_spec_t   & eRes,
	     const sw_cap_t         & swcap,
	     const enc_type_t       & enctype,
	     const gmpls_bwenc_t    & bw,
	     const gpid_t           & gpid,
	     const times_t          & times,
	     const uint32_t         & setupPrio,
	     const uint32_t         & holdingPrio,
	     const lsp_type_t       & type,
	     const lsp_res_action_t & action,
	     std::string            & err)
{
	lspData_t * data;

	err = "";
	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);
	LSPMAP_COND_CREATE_AND_BIND(vtyLspContainer, lspData_t,
				    lspKey, data, err);

	//data->itna        = itna;
	//data->etna        = etna;

	data->iRes        = iRes;
	data->eRes        = eRes;

	data->swcap	  = swcap;
	data->enctype	  = enctype;
	data->bw	  = bw;
	data->gpid	  = gpid;
	data->times       = times;
	data->setupPrio	  = setupPrio;
	data->holdingPrio = holdingPrio;
	data->type	  = type;
	data->action      = action;

	return true;
}

bool
vtyLspUpdate(const mapKey_t       & lspKey,
	     const lsp_rro_mode_t & rroMode,
	     const uint32_t       & refresh,
	     const bool           & activateAck,
	     const uint32_t       & rapidRetransmissionInterval,
	     const uint32_t       & rapidRetryLimit,
	     const uint32_t       & incrementValueDelta,
	     std::string          & err)
{
	lspData_t * data;

	err = "";
	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);
	LSPMAP_COND_CREATE_AND_BIND(vtyLspContainer, lspData_t,
				    lspKey, data, err);

	data->rroMode                     = rroMode;
	data->refresh                     = refresh;
	data->activateAck                 = activateAck;
	data->rapidRetransmissionInterval = rapidRetransmissionInterval;
	data->rapidRetryLimit             = rapidRetryLimit;
	data->incrementValueDelta         = incrementValueDelta;

	return true;
}

bool
vtyLspCommit(const mapKey_t &  lspKey,
	     std::string &     err)
{
	lspData_t *       data;
	lsp_info_t        lspInfo;
	call_ident_t      callIdent;
	recovery_info_t   recoveryInfo;
	grapiErrorCode_t  ret;
	lsp_ident_t       apiLspKey;
	uint16_t          tun_id;
	struct in_addr    dst_nid;
	std::string       resp;

	err = "";

	LSPMAP_FIND(vtyLspContainer, lspData_t*, lspKey, data);

	vtyLspDump(lspKey);

	if (!data) {
		err += "LSP with ";
		err += key2string(lspKey);
		err += " not present in VTY containers";
		vtyLspContainer.erase(lspKey);
		return false;
	}

	if (!vtyLspCommittable(lspKey)) {
		err += "LSP with ";
		err += key2string(lspKey);
		err += " not yet committable to GRAPI";
		return false;
	}

	if (!demangleLspKey(lspKey, &tun_id, &(dst_nid.s_addr))) {
		err += "Cannot demangle an LSP key in VTY";
		return false;
	}

	memset(&apiLspKey, 0, sizeof(apiLspKey));
	apiLspKey.tun_id    = tun_id;
	apiLspKey.dst_nid   = ntohl(dst_nid.s_addr);
	apiLspKey.ext_tun_id = data->ext_tun_id;
	apiLspKey.lsp_id    = data->lsp_id;

	memset(&lspInfo, 0, sizeof(lspInfo));
	memset(&lspInfo.mask_, 1, sizeof(lspInfo.mask_));
	lspInfo.sw_cap                  = data->swcap;
	lspInfo.enc_type                = data->enctype;
	lspInfo.bw                      = data->bw;
	lspInfo.gpid                    = data->gpid;
	lspInfo.setup_prio              = data->setupPrio;
	lspInfo.holding_prio            = data->holdingPrio;
	lspInfo.type                    = data->type;
	lspInfo.action                  = data->action;
	lspInfo.rro_mode                = data->rroMode;
	lspInfo.refresh_interval        = data->refresh;
	lspInfo.activate_ack            = data->activateAck;
	lspInfo.rapid_retransm_interval = data->rapidRetransmissionInterval;
	lspInfo.rapid_retry_limit       = data->rapidRetryLimit;
	lspInfo.increment_value_delta   = data->incrementValueDelta;
	lspInfo.times                   = data->times;

	//TMP
	memset(&callIdent, 0, sizeof(callIdent));

	memset(&recoveryInfo, 0, sizeof(recoveryInfo));
	memset(&recoveryInfo.mask_, 1, sizeof(recoveryInfo.mask_));
	//

	ret =  RSVP_G2RAPI::lspCreate(apiLspKey,
				      *(data->sessionName),
				      callIdent,
				      data->iRes,
				      data->eRes,
				      lspInfo,
				      recoveryInfo,
				      resp);

	if (ret != GRAPI_ERROR_NONE) {
		err += "Cannot create an LSP via GRAPI with ";
		err += key2string(lspKey);
		err += " (err: code=";
		err += GRAPIERR2STRING(ret);
		err += ", reason='";
		err += resp;
		err += "')";
		return false;
	}

	// remove item in map
	delete data->sessionName;
	delete data;
	vtyLspContainer.erase(lspKey);

	return true;
}


//
//
//  ERO related containers
//
//
std::map<mapKey_t, std::list<lsp_ero_sobj_t > * > vtyEroContainer;

void
vtyLspEroDump(const mapKey_t & lspKey)
{
	std::list<lsp_ero_sobj_t> *               ero;
	std::list<lsp_ero_sobj_t>::const_iterator iter;



	LSPMAP_FIND(vtyEroContainer, std::list<lsp_ero_sobj_t>*, lspKey, ero);

	if (!ero || !ero->size()) {
		zlog_debug("LSP with %s not present in VTY containers",
			   key2string(lspKey).c_str());
		vtyLspContainer.erase(lspKey);
		return;
	}

	zlog_debug("LSP %s", key2string(lspKey).c_str());
	zlog_debug("\t   - ERO (%d hops):",
		   ero->size());

	for (iter = ero->begin(); iter != ero->end(); iter++) {
		zlog_debug("\t      - HOP: %s",
			   hop2string(*iter).c_str());
	}
}

bool
vtyEroCommit(const mapKey_t &  lspKey,
	     uint16_t          lsp_id,
	     std::string &     err)
{
	std::list<lsp_ero_sobj_t> * ero;
	grapiErrorCode_t            ret;
	lsp_ident_t                 apiLspKey;
	uint16_t                    tun_id;
	struct in_addr              dst_nid;
	std::string                 resp;

	err = "";

	vtyLspEroDump(lspKey);

	LSPMAP_FIND(vtyEroContainer, std::list<lsp_ero_sobj_t>*, lspKey, ero);

	if (!ero || !ero->size()) {
		err += "ERO not present VTY containers for LSP ";
		err += key2string(lspKey);
		vtyEroContainer.erase(lspKey);
		return false;
	}

	if (!demangleLspKey(lspKey, &tun_id, &(dst_nid.s_addr))) {
		err += "Cannot demangle an LSP key in VTY";
		return false;
	}

	memset(&apiLspKey, 0, sizeof(apiLspKey));
	apiLspKey.dst_nid = ntohl(dst_nid.s_addr);
	apiLspKey.tun_id  = tun_id;
	apiLspKey.lsp_id  = lsp_id;

	ret =  RSVP_G2RAPI::lspEroAttach(apiLspKey, *ero, resp);

	if (ret != GRAPI_ERROR_NONE) {
		err += "Cannot commit an ERO object via GRAPI for LSP with ";
		err += key2string(lspKey);
		err += " (err: code=";
		err += GRAPIERR2STRING(ret);
		err += ", reason='";
		err += resp;
		err += "')";
		return false;
	}

	// remove item in map
	ero->clear();
	vtyEroContainer.erase(lspKey);

	return true;
}

bool
vtyEroSobjAttach(const mapKey_t &       lspKey,
		 const lsp_ero_sobj_t & data,
		 std::string &          err)
{
	std::list<lsp_ero_sobj_t> *               ero;
	std::list<lsp_ero_sobj_t>::const_iterator iter;

	err = "";

	LSPMAP_FIND(vtyEroContainer, std::list<lsp_ero_sobj_t>*, lspKey, ero);

	if (!ero) {
		ero = new std::list<lsp_ero_sobj_t>;

		if (!ero) {
			err += "Cannot allocate a VTY LSP container";
			vtyEroContainer.erase(lspKey);
			return false;
		}

		ero->clear();
		vtyEroContainer[lspKey] = ero;
	}

	for (iter = ero->begin(); iter != ero->end(); iter++) {
		if ((*iter) == data && data.type != ERO_SUBOBJ_LABEL) {
			err += "ERO subobject just present in LSP ";
			err += key2string(lspKey);
			return false;
		}
	}

	ero->push_back(data);

	zlog_debug("Attaching HOP: %s", hop2string(data).c_str());

	vtyLspEroDump(lspKey);

	return true;
}

bool
vtyEroSobjDetach(const mapKey_t &       lspKey,
		 const lsp_ero_sobj_t & data,
		 std::string &          err)
{
	std::list<lsp_ero_sobj_t> *         ero;
	std::list<lsp_ero_sobj_t>::iterator iter;

	err = "";

	LSPMAP_FIND(vtyEroContainer, std::list<lsp_ero_sobj_t>*, lspKey, ero);

	if (!ero || !ero->size()) {
		err += "ERO not present VTY containers for LSP ";
		err += key2string(lspKey);
		vtyEroContainer.erase(lspKey);
		return false;

	}

	for (iter  = ero->begin(); iter != ero->end(); iter++) {
		if ((*iter) == data) {

			ero->erase(iter);

			if (!ero->size()) {
				err += "Last hop in ERO for LSP ";
				err += key2string(lspKey);
				ero->clear();
				vtyEroContainer.erase(lspKey);
			}

			vtyLspEroDump(lspKey);

			return true;
		}
	}

	err += "ERO suboject not present for LSPId " +
	err += key2string(lspKey);

	return false;
}





//
//
//
//   VTY commands
//
//
//
struct cmd_node g2rsvpte_node = {
	G2RSVPTE_NODE,
	"%s(config-g2rsvpte)# ",
	1
};

#define MYVTY_DEFUN(funcname, cmdname, cmdstr, helpstr)			\
	DEFUN_CMD_FUNC_DECL(funcname)					\
	DEFUN_CMD_ELEMENT(funcname, cmdname, cmdstr, helpstr, 0, 0)	\
	DEFUN_CMD_FUNC_TEXT(funcname)


MYVTY_DEFUN(g2rsvpte,
	    g2rsvpte_cmd,
	    "g2rsvpte (border|core)",
	    "Enable a G2.RSVP-TE process\n"
	    "Start G2.RSVP-TE configuration\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;
	bool             isCore = false;

	vty->node  = G2RSVPTE_NODE;
	vty->index = NULL;//rsvp_get();

	vtyEroContainer.clear();

	adjType = INNI;

	if (argc == 1) {
		if (strcmp (argv[0], "border") == 0) {
		        isCore = false;
		} else if (strcmp (argv[0], "core") == 0) {
			isCore = true;
		} else {
			vty_out(vty,
				"Invalid G2.RSVP-TE instance type (%s)%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = RSVP_API::g2rsvpteStart(adjType, err, false, isCore);

	if (ret == GRAPI_ERROR_INSTANCE_ALREADY_EXISTS) {
		vty_out(vty,
			"Got existing G2.RSVP-TE instance%s",
			VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot start G2.RSVP-TE instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}


	vty_out(vty,
		"G2.RSVP-TE instance started !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(no_g2rsvpte,
	    no_g2rsvpte_cmd,
	    "no g2rsvpte",
	    NO_STR
	    "Disable a G2.RSVP-TE process\n"
	    "Stop G2.RSVP-TE operation\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;

	vtyEroContainer.clear();

	adjType = INNI;

	ret = RSVP_API::g2rsvpteStop(adjType, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot stop G2.RSVP-TE instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"G2.RSVP-TE instance stopped !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gunirsvp,
	    gunirsvp_cmd,
	    "gunirsvp (client|network)",
	    "Enable a G.UNI-RSVP instance\n"
	    "Start G.UNI-RSVP configuration\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;
	bool             isClient = false;

	vty->node  = G2RSVPTE_NODE;
	vty->index = NULL;//rsvp_get();

	vtyEroContainer.clear();

	adjType = UNI;

	if (argc == 1) {
		if (strcmp (argv[0], "client") == 0) {
		        isClient = true;
		} else if (strcmp (argv[0], "network") == 0) {
			isClient = false;
		} else {
			vty_out(vty,
				"Invalid G.UNI-RSVP instance type (%s)%s",
				argv[0], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	ret = RSVP_API::g2rsvpteStart(adjType, err, isClient);

	if (ret == GRAPI_ERROR_INSTANCE_ALREADY_EXISTS) {
		vty_out(vty,
			"Got existing G.UNI-RSVP instance%s",
			VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot start G.UNI-RSVP instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}


	vty_out(vty,
		"G.UNI-RSVP instance started !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(no_gunirsvp,
	    no_gunirsvp_cmd,
	    "no gunirsvp",
	    NO_STR
	    "Disable a G.UNI-RSVP instance\n"
	    "Stop G.UNI-RSVP operation\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;

	vtyEroContainer.clear();

	adjType = UNI;

	ret = RSVP_API::g2rsvpteStop(adjType, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot stop G.UNI-RSVP instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"G.UNI-RSVP instance stopped !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gennirsvp,
	    gennirsvp_cmd,
	    "gennirsvp",
	    "Enable a G.ENNI-RSVP instance\n"
	    "Start G.ENNI-RSVP configuration\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;

	vty->node  = G2RSVPTE_NODE;
	vty->index = NULL;//rsvp_get();

	vtyEroContainer.clear();

	adjType = ENNI;

	ret = RSVP_API::g2rsvpteStart(adjType, err);

	if (ret == GRAPI_ERROR_INSTANCE_ALREADY_EXISTS) {
		vty_out(vty,
			"Got existing G.ENNI-RSVP instance%s",
			VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot start G.ENNI-RSVP instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}


	vty_out(vty,
		"G.ENNI-RSVP instance started !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(no_gennirsvp,
	    no_gennirsvp_cmd,
	    "no gennirsvp",
	    NO_STR
	    "Disable a G.ENNI-RSVP process\n"
	    "Stop G.ENNI-RSVP operation\n")
{
	std::string      err;
	grapiErrorCode_t ret;
	adj_type_t       adjType;

	vtyEroContainer.clear();

	adjType = ENNI;

	ret = RSVP_API::g2rsvpteStop(adjType, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot stop G.ENNI-RSVP instance "
			"(err: code=%s, reason='%s')%s",
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"G.ENNI-RSVP instance stopped !%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}


MYVTY_DEFUN(g2rsvpte_show,
	    g2rsvpte_gen_show_cmd,
	    "show g2rsvpte general",
	    "Show configured protocol instance")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_sessions   sessionIter;
	RSVP::Session *                     session;
	RSVP::Session::iterator_lsps        lspIter;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	adj_type_t                          adjType;
	int                                 countIntf;
	int                                 countLsp;

	adjType = INNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	countIntf = 0;
	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		countIntf++;
	}

	countLsp = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			countLsp++;
		}
	}

	vty_out(vty, "G2.RSVP-TE PROTOCOL INSTANCE%s", VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	vty_out(vty, "   Number of interfaces: %d%s", countIntf, VTY_NEWLINE);
	vty_out(vty, "   Number of LSPs      : %d%s", countLsp, VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gunirsvp_show,
	    gunirsvp_gen_show_cmd,
	    "show gunirsvp general",
	    "Show configured protocol instance")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_sessions   sessionIter;
	RSVP::Session *                     session;
	RSVP::Session::iterator_lsps        lspIter;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	adj_type_t                          adjType;
	int                                 countIntf;
	int                                 countLsp;

	adjType = UNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	countIntf = 0;
	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		countIntf++;
	}

	countLsp = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			countLsp++;
		}
	}


	vty_out(vty, "G.UNI-RSVP PROTOCOL INSTANCE%s", VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	vty_out(vty, "   Number of interfaces: %d%s", countIntf, VTY_NEWLINE);
	vty_out(vty, "   Number of LSPs      : %d%s", countLsp, VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gennirsvp_show,
	    gennirsvp_gen_show_cmd,
	    "show gennirsvp general",
	    "Show configured protocol instance")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_sessions   sessionIter;
	RSVP::Session *                     session;
	RSVP::Session::iterator_lsps        lspIter;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	adj_type_t                          adjType;
	int                                 countIntf;
	int                                 countLsp;

	adjType = ENNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	countIntf = 0;
	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		countIntf++;
	}

	countLsp = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			countLsp++;
		}
	}

	vty_out(vty, "G.ENNI-RSVP PROTOCOL INSTANCE%s", VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	vty_out(vty, "   Number of interfaces: %d%s", countIntf, VTY_NEWLINE);
	vty_out(vty, "   Number of LSPs      : %d%s", countLsp, VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_intf_show,
	    g2rsvpte_intf_show_cmd,
	    "show g2rsvpte interfaces",
	    "Show configured interfaces")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	RSVP::Interface *                   intf;
	std::ostringstream                  os;
	adj_type_t                          adjType;

	adjType = INNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		intf = (*iter).second;
		assert(intf);
		os << *intf;
		os << VTY_NEWLINE;
	}
	vty_out(vty, "G2.RSVP-TE INTERFACES%s%s", VTY_NEWLINE, VTY_NEWLINE);
	vty_out(vty, "%s", os.str().c_str());

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gunirsvp_intf_show,
	    gunirsvp_intf_show_cmd,
	    "show gunirsvp interfaces",
	    "Show configured interfaces")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	RSVP::Interface *                   intf;
	std::ostringstream                  os;
	adj_type_t                          adjType;

	adjType = UNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		intf = (*iter).second;
		assert(intf);
		os << *intf;
		os << VTY_NEWLINE;
	}
	vty_out(vty, "G.UNI-RSVP INTERFACES%s%s", VTY_NEWLINE, VTY_NEWLINE);
	vty_out(vty, "%s", os.str().c_str());

	return CMD_SUCCESS;
}

MYVTY_DEFUN(gennirsvp_intf_show,
	    gennirsvp_intf_show_cmd,
	    "show gennnirsvp interfaces",
	    "Show configured interfaces")
{
	RSVP::G2RSVPTE *                    proto;
	RSVP::G2RSVPTE::iterator_interfaces iter;
	RSVP::Interface *                   intf;
	std::ostringstream                  os;
	adj_type_t                          adjType;

	adjType = ENNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	for (iter  = proto->begin_interfaces();
	     iter != proto->end_interfaces();
	     iter ++) {
		intf = (*iter).second;
		assert(intf);
		os << *intf;
		os << VTY_NEWLINE;
	}
	vty_out(vty, "G.ENNI-RSVP INTERFACES%s%s", VTY_NEWLINE, VTY_NEWLINE);
	vty_out(vty, "%s", os.str().c_str());

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_create,
	    g2rsvpte_lsp_create_cmd,
	    "lsp create tunnel-id WORD "
	    "dst-rid A.B.C.D "
	    "extended-tunnel-id WORD "
	    "lsp-id WORD "
	    "session-name WORD ",
	    "Create LSP with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	uint32_t               ext_tun_id;
	uint16_t               lsp_id;
	std::string            sessionName;
	mapKey_t               key;
	std::string            err;

	tun_id       = strtoul(argv[0], NULL, 10);
	lsp_id       = tun_id;
	ext_tun_id    = 0;
	sessionName = "";

	if (!inet_aton(argv[1], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (argc > 2) {
		ext_tun_id = strtoul(argv[2], NULL, 10);
		lsp_id    = strtoul(argv[3], NULL, 10);
		sessionName = std::string(argv[4]);
	}

	memset(&key, 0, sizeof(key));
	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to create a G2.RSVP-TE LSP with %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyLspUpdate(key, ext_tun_id, lsp_id, sessionName, err)) {
		vty_out(vty,
			"Cannot update data into VTY LSP with %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with %s successfully created in VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}


// Added in-TEL/DL/LBL and out-TEL/DL/LBL and removed:
//   - "in-TNA A.B.C.D "
//   - "out-TNA A.B.C.D "
MYVTY_DEFUN(g2rsvpte_lsp_update,
	    g2rsvpte_lsp_update_cmd,
	    "lsp update data tunnel-id WORD dst-rid A.B.C.D "
	    "in-te-link A.B.C.D "
	    "in-data-link (WORD|A.B.C.D) "
	    "in-label32-id WORD "
	    "out-te-link A.B.C.D "
	    "out-data-link (WORD|A.B.C.D) "
	    "out-label32-id WORD "
	    "sw-cap WORD "
	    "enc-type WORD "
	    "bw WORD "
	    "gpid WORD "
	    "start-time WORD "
	    "end-time WORD "
	    "setup-prio <0-7> "
	    "holding-prio <0-7> "
	    "type (spc|pc) "
	    "resources (xconnect|book)",
	    "Update data into LSP")
{
	uint16_t                    tun_id;
	struct in_addr              dst_nid;
	mapKey_t                    key;

	net_res_spec_t              iRes;
	net_res_spec_t              eRes;
	//ipv4_t                      itna_ip;
	//ipv4_t                      etna_ip;
	//g2mpls_addr_t               itna_id;
	//g2mpls_addr_t               etna_id;

	sw_cap_t                    swcap;
	enc_type_t                  enctype;
	gmpls_bwenc_t               bw;
	gpid_t                      gpid;
	uint32_t                    setupPrio;
	uint32_t                    holdingPrio;
	lsp_type_t                  type;
	lsp_res_action_t            action;
	times_t                     times;
	std::string                 err;

	//memset(&itna_id, 0, sizeof(itna_id));
	//memset(&etna_id, 0, sizeof(etna_id));
	memset(&iRes, 0, sizeof(net_res_spec_t));
	memset(&eRes, 0, sizeof(net_res_spec_t));
	memset(&times, 0, sizeof(times_t));

	setupPrio   = 0;
	holdingPrio = 0;
	type        = LSP_TYPE_SPC;
	action      = RESOURCE_XCONNECT;

	tun_id = strtoul(argv[0], NULL, 10);

	if (!inet_aton(argv[1], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	// Ingress NetRes
	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		iRes.tna.type       = IPv4;
		iRes.tna.preflen    = 32;
		if (!inet_aton(argv[2], &(iRes.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for ingress TE-Link%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for ingress TE-Link%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(iRes.mask_, tna);

	if (strchr (argv[3], '.') != NULL) {
		// matched "A.B.C.D".
		iRes.data_link.type       = IPv4;
		iRes.data_link.preflen    = 32;
		if (!inet_aton(argv[3], &(iRes.data_link.value.ipv4))) {
			vty_out(vty,
			      "Invalid IP address (%s) for ingress data link%s",
				argv[3], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		// match "<0-4294967295>".
		iRes.data_link.type                = UNNUMBERED;
		iRes.data_link.preflen             = 0;
		iRes.data_link.value.unnum         = strtoul(argv[3], NULL, 0);
	}
	BITMASK_BITSET(iRes.mask_, data_link);

	iRes.label.type = LABEL_32BIT;
	iRes.label.value.label32.id = strtoul(argv[4], NULL, 0);
	BITMASK_BITSET(iRes.mask_, label);

	// Egress NetRes
	if (strchr (argv[5], '.') != NULL) {
		// matched "A.B.C.D".
		eRes.tna.type       = IPv4;
		eRes.tna.preflen    = 32;
		if (!inet_aton(argv[5], &(eRes.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for egress TE-Link%s",
				argv[5], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for egress TE-Link%s",
			argv[5], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(eRes.mask_, tna);

	if (strchr (argv[6], '.') != NULL) {
		// matched "A.B.C.D".
		eRes.data_link.type       = IPv4;
		eRes.data_link.preflen    = 32;
		if (!inet_aton(argv[6], &(eRes.data_link.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for egress data link%s",
				argv[6], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		// match "<0-4294967295>".
		eRes.data_link.type                = UNNUMBERED;
		eRes.data_link.preflen             = 0;
		eRes.data_link.value.unnum         = strtoul(argv[6], NULL, 0);
	}
	BITMASK_BITSET(eRes.mask_, data_link);

	eRes.label.type = LABEL_32BIT;
	eRes.label.value.label32.id = strtoul(argv[7], NULL, 0);
	BITMASK_BITSET(eRes.mask_, label);

	//if (strchr(argv[2], '.') != NULL) {
	//	// matched "A.B.C.D".
	//	if (!inet_aton(argv[2], &(itna_ip))) {
	//		vty_out(vty,
	//			"Invalid IP address (%s) for itna-id%s",
	//			argv[2], VTY_NEWLINE);
	//		return CMD_WARNING;
	//	}
	//	itna_id.type       = IPv4;
	//	itna_id.preflen    = 32;
	//	itna_id.value.ipv4 = itna_ip;
	//} else {
	//	vty_out(vty,
	//		"Invalid IP address (%s) for itna-id%s",
	//		argv[2], VTY_NEWLINE);
	//		return CMD_WARNING;
	//}
	//
	//if (strchr(argv[3], '.') != NULL) {
	//	// matched "A.B.C.D".
	//	if (!inet_aton(argv[3], &(etna_ip))) {
	//		vty_out(vty,
	//			"Invalid IP address (%s) for etna-id%s",
	//			argv[3], VTY_NEWLINE);
	//		return CMD_WARNING;
	//	}
	//	etna_id.type       = IPv4;
	//	etna_id.preflen    = 32;
	//	etna_id.value.ipv4 = etna_ip;
	//} else {
	//	vty_out(vty,
	//		"Invalid IP address (%s) for etna-id%s",
	//		argv[3], VTY_NEWLINE);
	//	return CMD_WARNING;
	//}

	swcap            = (sw_cap_t) strtoul(argv[8], NULL, 10);
	enctype          = (enc_type_t) strtoul(argv[9], NULL, 10);
	bw               = (gmpls_bwenc_t) strtoul(argv[10], NULL, 16);
	gpid             = (gpid_t) strtoul(argv[11], NULL, 10);

	if (argc > 12) {
		times.start_time = strtoul(argv[12], NULL, 10);
		times.end_time   = strtoul(argv[13], NULL, 10);
	}
	if (argc > 14) {
		setupPrio   = strtoul(argv[14], NULL, 10);
		holdingPrio = strtoul(argv[15], NULL, 10);
		if (strcmp (argv[16], "spc") == 0) {
			type = LSP_TYPE_SPC;
		} else if (strcmp (argv[16], "pc") == 0) {
			type = LSP_TYPE_PC;
		} else {
			vty_out(vty,
				"Invalid resource behavior type (%s) "
				"for LSP%s",
				argv[16], VTY_NEWLINE);
			return CMD_WARNING;
		}
		if (strcmp (argv[17], "xconnect") == 0) {
			action = RESOURCE_XCONNECT;
		} else if (strcmp (argv[17], "book") == 0) {
			action = RESOURCE_BOOK;
		} else {
			vty_out(vty,
				"Invalid resource behavior type (%s) "
				"for LSP%s",
				argv[17], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}

	memset(&key, 0, sizeof(key));
	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to update data for G2.RSVP-TE LSP with %s%s",
		key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyLspUpdate(key,
			  iRes,
			  eRes,
			  //itna_id,
			  //etna_id,
			  swcap,
			  enctype,
			  bw,
			  gpid,
			  times,
			  setupPrio,
			  holdingPrio,
			  type,
			  action,
			  err)) {
		vty_out(vty,
			"Cannot update data into VTY LSP with %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with %s sucessfully updated%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_behavior,
	    g2rsvpte_lsp_behavior_cmd,
	    "lsp update behavior tunnel-id WORD dst-rid A.B.C.D "
	    "record-route (all|dl-detail|tel-detail|off) "
	    "refresh-period <0-4294967295> "
	    "ack-procedure (on|off) "
	    "rapid-retransmission-interval <0-4294967295> "
	    "rapid-retry-limit <1-5> "
	    "increment-value-delta <1-10>",
	    "Update behavior data for LSP")
{
	uint16_t                    tun_id;
	struct in_addr              dst_nid;
	mapKey_t                    key;
	lsp_rro_mode_t              rroMode;
	uint32_t                    refresh;
	bool                        activateAck;
	uint32_t                    rapidRetransmissionInterval;
	uint32_t                    rapidRetryLimit;
	uint32_t                    incrementValueDelta;
	std::string                 err;

	tun_id = strtoul(argv[0], NULL, 10);

	if (!inet_aton(argv[1], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[2], "all") == 0) {
		rroMode = LSP_RRO_ALL;
	} else if (strcmp (argv[2], "dl-detail") == 0) {
		rroMode = LSP_RRO_DL_DETAIL;
	} else if (strcmp (argv[2], "tel-detail") == 0) {
		rroMode = LSP_RRO_TEL_DETAIL;
	} else if (strcmp (argv[2], "off") == 0) {
		rroMode = LSP_RRO_OFF;
	} else {
		vty_out(vty,
			"Invalid RRO recording mode (%s) for LSP%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	refresh = strtoul(argv[3], NULL, 10);

	if (strcmp (argv[4], "on") == 0) {
		activateAck = true;
	} else if (strcmp (argv[4], "off") == 0) {
		activateAck = false;
	} else {
		vty_out(vty,
			"Invalid ACK mode (%s) for LSP%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}

	rapidRetransmissionInterval = strtoul(argv[3], NULL, 10);
	rapidRetryLimit             = strtoul(argv[3], NULL, 10);
	incrementValueDelta         = strtoul(argv[3], NULL, 10);

	memset(&key, 0, sizeof(key));
	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to update data for G2.RSVP-TE LSP with %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyLspUpdate(key,
			  rroMode,
			  refresh,
			  activateAck,
			  rapidRetransmissionInterval,
			  rapidRetryLimit,
			  incrementValueDelta,
			  err)) {
		vty_out(vty,
			"Cannot update data into VTY LSP with %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with %s sucessfully updated%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_commit,
	    g2rsvpte_lsp_commit_cmd,
	    "lsp commit tunnel-id WORD dst-rid A.B.C.D",
	    "Commit LSP with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	mapKey_t               key;
	std::string            err;

	tun_id = strtoul(argv[0], NULL, 10);

	if (!inet_aton(argv[1], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to commit LSP with %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyLspCommit(key, err)) {
		vty_out(vty,
			"Cannot commit an LSP from VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with %s successfully committed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}



MYVTY_DEFUN(g2rsvpte_lsp_ero_ipv4,
	    g2rsvpte_lsp_ero_ipv4_cmd,
	    "lsp ero (attach|detach) tunnel-id WORD dst-rid A.B.C.D "
	    "subobject (loose|strict) ipv4 A.B.C.D prefix <1-32>",
	    "Attach ERO subobject of type IPv4 into LSP "
	    "with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	bool                   attach;
	bool                   ret;
	mapKey_t               key;
	lsp_ero_sobj_t         sobj;
	std::string            err;

	memset(&sobj, 0, sizeof(sobj));

	if (strcmp (argv[0], "attach") == 0) {
		attach = true;
	} else	if (strcmp (argv[0], "detach") == 0) {
		attach = false;
	} else {
		vty_out(vty,
			"Invalid attach|detach action (%s) for an LSP%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	tun_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[3], "loose") == 0) {
		sobj.loose = ERO_SUBOBJ_LOOSE;
	} else	if (strcmp (argv[3], "strict") == 0) {
		sobj.loose = ERO_SUBOBJ_STRICT;
	} else {
		vty_out(vty,
			"Invalid Loose|Strict mode (%s) for LSP%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.type = ERO_SUBOBJ_IPV4;
	if (!inet_aton(argv[4], &sobj.hop.ipv4.addr)) {
		vty_out(vty,
			"Invalid IP address (%s) for dERO hop%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.hop.ipv4.prefix = (uint8_t) strtoul(argv[5], NULL, 10);

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to %s an IPv4 ERO subobject to LSP with %s%s",
		attach ? "attach" : "detach",
		key2string(key).c_str(),
		VTY_NEWLINE);

	ret = ((attach) ?
	       vtyEroSobjAttach(key, sobj, err) :
	       vtyEroSobjDetach(key, sobj, err));

	if (!ret) {
		vty_out(vty,
			"Cannot %s an IPv4 ERO subobject to LSP "
			"%s (err reason='%s')%s",
			attach ? "attach" : "detach",
			key2string(key).c_str(),
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"IPv4 ERO subobject for LSP with %s successfully %s%s",
		key2string(key).c_str(),
		attach ? "attached" : "detached",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_ero_ipv6,
	    g2rsvpte_lsp_ero_ipv6_cmd,
	    "lsp ero (attach|detach) tunnel-id WORD dst-rid A.B.C.D "
	    "subobject (loose|strict) "
	    "ipv6 WORD WORD WORD WORD prefix <1-128>",
	    "Attach ERO subobject of type IPv6 into LSP "
	    "with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	bool                   attach;
	bool                   ret;
	mapKey_t               key;
	lsp_ero_sobj_t         sobj;
	std::string            err;

	memset(&sobj, 0, sizeof(sobj));

	if (strcmp (argv[0], "attach") == 0) {
		attach = true;
	} else	if (strcmp (argv[0], "detach") == 0) {
		attach = false;
	} else {
		vty_out(vty,
			"Invalid attach|detach action (%s) for an LSP%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	tun_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[3], "loose") == 0) {
		sobj.loose = ERO_SUBOBJ_LOOSE;
	} else	if (strcmp (argv[3], "strict") == 0) {
		sobj.loose = ERO_SUBOBJ_STRICT;
	} else {
		vty_out(vty,
			"Invalid Loose|Strict mode (%s) for LSP%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.type = ERO_SUBOBJ_IPV6;
	for (int i = 0; i < 4; i++) {
		sobj.hop.ipv6.addr.s6_addr32[i] =
			strtoul(argv[i + 4], NULL, 16);
	}
	sobj.hop.ipv6.prefix = (uint8_t) strtoul(argv[8], NULL, 10);

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to %s an IPv6 ERO subobject to LSP with %s%s",
		attach ? "attach" : "detach",
		key2string(key).c_str(),
		VTY_NEWLINE);

	ret = ((attach) ?
	       vtyEroSobjAttach(key, sobj, err) :
	       vtyEroSobjDetach(key, sobj, err));

	if (!ret) {
		vty_out(vty,
			"Cannot %s an IPv6 ERO subobject to LSP "
			"%s (err reason='%s')%s",
			attach ? "attach" : "detach",
			key2string(key).c_str(),
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"IPv6 ERO subobject for LSP with %s successfully %s%s",
		key2string(key).c_str(),
		attach ? "attached" : "detached",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_ero_as,
	    g2rsvpte_lsp_ero_as_cmd,
	    "lsp ero (attach|detach) tunnel-id WORD dst-rid A.B.C.D "
	    "subobject (loose|strict) as WORD",
	    "Attach ERO subobject of type AS number into LSP "
	    "with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	bool                   attach;
	bool                   ret;
	mapKey_t               key;
	lsp_ero_sobj_t         sobj;
	std::string            err;

	memset(&sobj, 0, sizeof(sobj));

	if (strcmp (argv[0], "attach") == 0) {
		attach = true;
	} else	if (strcmp (argv[0], "detach") == 0) {
		attach = false;
	} else {
		vty_out(vty,
			"Invalid attach|detach action (%s) for an LSP%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	tun_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[3], "loose") == 0) {
		sobj.loose = ERO_SUBOBJ_LOOSE;
	} else	if (strcmp (argv[3], "strict") == 0) {
		sobj.loose = ERO_SUBOBJ_STRICT;
	} else {
		vty_out(vty,
			"Invalid Loose|Strict mode (%s) for LSP%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.type = ERO_SUBOBJ_AS;
	sobj.hop.as.number = (uint16_t) strtoul(argv[4], NULL, 16);

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to %s an AS-number ERO subobject to LSP with %s%s",
		attach ? "attach" : "detach",
		key2string(key).c_str(),
		VTY_NEWLINE);

	ret = ((attach) ?
	       vtyEroSobjAttach(key, sobj, err) :
	       vtyEroSobjDetach(key, sobj, err));

	if (!ret) {
		vty_out(vty,
			"Cannot %s an AS-number ERO subobject to LSP "
			"%s (err reason='%s')%s",
			attach ? "attach" : "detach",
			key2string(key).c_str(),
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"AS-number ERO subobject for LSP with %s successfully %s%s",
		key2string(key).c_str(),
		attach ? "attached" : "detached",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_ero_unnum,
	    g2rsvpte_lsp_ero_unnum_cmd,
	    "lsp ero (attach|detach) tunnel-id WORD dst-rid A.B.C.D "
	    "subobject (loose|strict) unnum routerId A.B.C.D intfId WORD",
	    "Attach ERO subobject of type UNNUM into LSP "
	    "with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	bool                   attach;
	struct in_addr         rtrId;
	bool                   ret;
	mapKey_t               key;
	lsp_ero_sobj_t         sobj;
	std::string            err;

	memset(&sobj, 0, sizeof(sobj));

	if (strcmp (argv[0], "attach") == 0) {
		attach = true;
	} else	if (strcmp (argv[0], "detach") == 0) {
		attach = false;
	} else {
		vty_out(vty,
			"Invalid attach|detach action (%s) for an LSP%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	tun_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[3], "loose") == 0) {
		sobj.loose = ERO_SUBOBJ_LOOSE;
	} else	if (strcmp (argv[3], "strict") == 0) {
		sobj.loose = ERO_SUBOBJ_STRICT;
	} else {
		vty_out(vty,
			"Invalid Loose|Strict mode (%s) for LSP%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.type = ERO_SUBOBJ_UNNUM;
	if (!inet_aton(argv[4], &rtrId)) {
		vty_out(vty,
			"Invalid IP address (%s) for hop router ID%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}
	sobj.hop.unnum.router_id = ntohl(rtrId.s_addr);
	sobj.hop.unnum.intf_id   = strtoul(argv[5], NULL, 16);

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to %s an Unnumbered ERO subobject to LSP with %s%s",
		attach ? "attach" : "detach",
		key2string(key).c_str(),
		VTY_NEWLINE);

	ret = ((attach) ?
	       vtyEroSobjAttach(key, sobj, err) :
	       vtyEroSobjDetach(key, sobj, err));

	if (!ret) {
		vty_out(vty,
			"Cannot %s an Unnumbered ERO subobject to LSP "
			"%s (err reason='%s')%s",
			attach ? "attach" : "detach",
			key2string(key).c_str(),
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Unnumbered ERO subobject for LSP with %s successfully %s%s",
		key2string(key).c_str(),
		attach ? "attached" : "detached",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_ero_label,
	    g2rsvpte_lsp_ero_label_cmd,
	    "lsp ero (attach|detach) tunnel-id WORD dst-rid A.B.C.D "
	    "subobject (loose|strict) label (upstream|downstream) value WORD",
	    "Attach ERO subobject of type LABEL into LSP "
	    "with given input parameters")
{
	uint16_t               tun_id;
	struct in_addr         dst_nid;
	bool                   attach;
	bool                   ret;
	mapKey_t               key;
	lsp_ero_sobj_t         sobj;
	std::string            err;

	memset(&sobj, 0, sizeof(sobj));

	if (strcmp (argv[0], "attach") == 0) {
		attach = true;
	} else	if (strcmp (argv[0], "detach") == 0) {
		attach = false;
	} else {
		vty_out(vty,
			"Invalid attach|detach action (%s) for an LSP%s",
			argv[0], VTY_NEWLINE);
		return CMD_WARNING;
	}

	tun_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strcmp (argv[3], "loose") == 0) {
		sobj.loose = ERO_SUBOBJ_LOOSE;
	} else	if (strcmp (argv[3], "strict") == 0) {
		sobj.loose = ERO_SUBOBJ_STRICT;
	} else {
		vty_out(vty,
			"Invalid Loose|Strict mode (%s) for LSP%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.type = ERO_SUBOBJ_LABEL;

	if (strcmp (argv[4], "upstream") == 0) {
		sobj.hop.label.upstream = LABEL_UPSTREAM;
	} else	if (strcmp (argv[4], "downstream") == 0) {
		sobj.hop.label.upstream = LABEL_DOWNSTREAM;
	} else {
		vty_out(vty,
			"Invalid upstream|downstream label mode (%s) "
			"for LSP%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}

	sobj.hop.label.value = strtoul(argv[5], NULL, 0);

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to %s Label ERO subobject to LSP with %s%s",
		attach ? "attach" : "detach",
		key2string(key).c_str(),
		VTY_NEWLINE);

	ret = ((attach) ?
	       vtyEroSobjAttach(key, sobj, err) :
	       vtyEroSobjDetach(key, sobj, err));

	if (!ret) {
		vty_out(vty,
			"Cannot %s a Label ERO subobject to LSP "
			"%s (err reason='%s')%s",
			attach ? "attach" : "detach",
			key2string(key).c_str(),
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Label ERO subobject for LSP with %s successfully %s%s",
		key2string(key).c_str(),
		attach ? "attached" : "detached",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_ero_commit,
	    g2rsvpte_lsp_ero_commit_cmd,
	    "lsp ero commit tunnel-id WORD lsp-id WORD dst-rid A.B.C.D",
	    "Commit ERO built for this LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         dst_nid;
	mapKey_t               key;
	std::string            err;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	memset(&key, 0, sizeof(key));

	if (!mangleLspKey(tun_id, dst_nid.s_addr, &key)) {
		vty_out(vty,
			"Cannot mangle an LSP key in VTY%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to commit ERO object to LSP with %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	if (!vtyEroCommit(key, lsp_id, err)) {
		vty_out(vty,
			"Cannot commit an ERO object: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"ERO object for LSP with %s successfully committed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_destroy,
	    g2rsvpte_lsp_destroy_cmd,
	    "lsp destroy tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Destroy LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to destroy a G2.RSVP-TE LSP with id %d "
		"to destination router %s%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid    = ntohl(src_nid.s_addr);
	key.dst_nid    = ntohl(dst_nid.s_addr);
	key.tun_id     = tun_id;
	key.lsp_id     = lsp_id;

	ret = RSVP_G2RAPI::lspDestroy(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot destroy G2.RSVP-TE LSP with id %d "
			"to destination router %s "
			"(err: code=%s, reason='%s')%s",
			tun_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with id %d to destination router %s "
		"successfully destroyed%s",
		tun_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_enable,
	    g2rsvpte_lsp_enable_cmd,
	    "lsp enable tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Enable LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to enable a G2.RSVP-TE LSP with id %d "
		"to destination router %s%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspEnable(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot enable G2.RSVP-TE LSP with id %d "
			"to destination router %s "
			"(err: code=%s, reason='%s')%s",
			lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with id %d to destination router %s "
		"successfully enabled%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_disable,
	    g2rsvpte_lsp_disable_cmd,
	    "lsp disable tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Disable LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to disable a G2.RSVP-TE LSP with id %d "
		"to destination router %s%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspDisable(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot disable G2.RSVP-TE LSP with id %d "
			"to destination router %s "
			"(err: code=%s, reason='%s')%s",
			lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP with id %d to destination router %s "
		"successfully disabled%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_forceup,
	    g2rsvpte_lsp_forceup_cmd,
	    "lsp force-up tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Force signalling up for LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to force up signalling for G2.RSVP-TE LSP with tun-id "
		"%d lsp-id %d to destination router %s%s",
		tun_id, lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspForceUp(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot force up signalling for G2.RSVP-TE LSP "
			"with tun-id %d lsp-id %d to destination router %s "
			"(err: code=%s, reason='%s')%s",
			tun_id, lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Signalling for LSP with tun-id %d lsp-id %d to destination "
		"router %s successfully forced up%s",
		tun_id, lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_forcedown,
	    g2rsvpte_lsp_forcedown_cmd,
	    "lsp force-down tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Force down signalling for LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to force down signalling for G2.RSVP-TE LSP with tun-id"
		" %d lsp-id %d to destination router %s%s",
		tun_id, lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspForceDown(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot force down signalling for G2.RSVP-TE LSP "
			"with tun-id %d lsp-id %d to destination router %s "
			"(err: code=%s, reason='%s')%s",
			tun_id, lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Signalling for LSP with tun-id %d lsp-id %d to destination "
		"router %s successfully forced down%s",
		tun_id, lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_send_resv,
	    g2rsvpte_lsp_send_resv_cmd,
	    "lsp send-resv tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Send Resv for LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to send_resv for G2.RSVP-TE LSP with id %d "
		"src router %s destination router %s%s",
		lsp_id, inet_ntoa(src_nid), inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspSendResv(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot sendResv for G2.RSVP-TE LSP with id %d "
			"to destination router %s "
			"(err: code=%s, reason='%s')%s",
			lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"SendResv for LSP with id %d to destination router %s "
		"successfully sent%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_send_confirm,
	    g2rsvpte_lsp_send_confirm_cmd,
	    "lsp send-confirm tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Send ResvConf for LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to send_confirm for G2.RSVP-TE LSP with id %d "
		"src router %s destination router %s%s",
		lsp_id, inet_ntoa(src_nid), inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspSendConfirm(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot sendConfirm for G2.RSVP-TE LSP with id %d "
			"to destination router %s "
			"(err: code=%s, reason='%s')%s",
			lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"SendConfirm for LSP with id %d to destination router %s "
		"successfully sent%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_post_xc_completed,
	    g2rsvpte_lsp_post_xc_completed_cmd,
	    "lsp post-xc-completed tunnel-id WORD lsp-id WORD "
	    "src-rid A.B.C.D dst-rid A.B.C.D",
	    "Post XC completed event to LSP with given input parameters")
{
	uint16_t               tun_id;
	uint16_t               lsp_id;
	struct in_addr         src_nid;
	struct in_addr         dst_nid;
	lsp_ident_t            key;
	std::string            err;
	grapiErrorCode_t       ret;

	tun_id = strtoul(argv[0], NULL, 10);
	lsp_id = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!inet_aton(argv[3], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[3], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Going to post xc_completed event to G2.RSVP-TE LSP with id %d "
		"src router %s destination router %s%s",
		lsp_id, inet_ntoa(src_nid), inet_ntoa(dst_nid), VTY_NEWLINE);

	memset(&key, 0, sizeof(key));
	key.src_nid = ntohl(src_nid.s_addr);
	key.dst_nid = ntohl(dst_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	ret = RSVP_G2RAPI::lspXConnCompleted(key, err);

	if (ret != GRAPI_ERROR_NONE) {
		vty_out(vty,
			"Cannot post xc_completed event to G2.RSVP-TE LSP with "
			"id %d to destination router %s "
			"(err: code=%s, reason='%s')%s",
			lsp_id, inet_ntoa(dst_nid),
			GRAPIERR2STRING(ret), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Post xc_completed event to LSP with id %d to destination "
		"router %s successfully sent%s",
		lsp_id, inet_ntoa(dst_nid), VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsp_show,
	    g2rsvpte_lsp_show_cmd,
	    "show g2rsvpte lsp tunnel-id WORD src-rid A.B.C.D dst-rid A.B.C.D "
	    "lsp-id WORD",
	    "Show details for specified LSP")
{
	RSVP::G2RSVPTE *   proto;
	RSVP::LSP *        lsp;
	uint16_t           tun_id;
	uint16_t           lsp_id;
	struct in_addr     src_nid;
	struct in_addr     dst_nid;
	lsp_ident_t        key;
	std::ostringstream os;
	adj_type_t         adjType;

	adjType = INNI;

	tun_id = strtoul(argv[0], NULL, 10);

	if (!inet_aton(argv[1], &src_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for source router ID%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	if (!inet_aton(argv[2], &dst_nid)) {
		vty_out(vty,
			"Invalid IP address (%s) for destination router ID%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	lsp_id = strtoul(argv[3], NULL, 10);

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	memset (&key, 0, sizeof(lsp_ident_t));
	key.dst_nid = ntohl(dst_nid.s_addr);;
	key.src_nid = ntohl(src_nid.s_addr);
	key.tun_id  = tun_id;
	key.lsp_id  = lsp_id;

	lsp = proto->findLSP(key);

	if (!lsp) {
		vty_out(vty, "ERROR: LSP not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	os << *lsp;

	vty_out(vty, "%s%s", os.str().c_str(), VTY_NEWLINE);

	//vty_out(vty, "G2.RSVP-TE LSP %d to destination router %s%s",
	//	tun_id, inet_ntoa(dst_nid), VTY_NEWLINE);
	//
	//vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);

	return CMD_SUCCESS;
}

MYVTY_DEFUN(g2rsvpte_lsps_show,
	    g2rsvpte_lsps_show_cmd,
	    "show g2rsvpte lsp",
	    "Show configured LSPs")
{
	RSVP::G2RSVPTE *                  proto;
	RSVP::G2RSVPTE::iterator_sessions sessionIter;
	RSVP::Session *                   session;
	RSVP::Session::iterator_lsps      lspIter;
	RSVP::LSP *                       lsp;
	std::ostringstream                os;
	int                               found;
	adj_type_t                        adjType;

	adjType = INNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	found = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			lsp = (*lspIter).second;
			assert(lsp);
			found = 1;
			os << *lsp;
			os << *session;
			os << VTY_NEWLINE;
		}
	}

	if (found) {
		vty_out(vty, "G2.RSVP-TE LSPs%s%s", VTY_NEWLINE, VTY_NEWLINE);
		vty_out(vty, "%s", os.str().c_str());
	} else {
		vty_out(vty, "No LSPs for this G2.RSVP-TE instance%s",
			VTY_NEWLINE);
	}
	return CMD_SUCCESS;
}

MYVTY_DEFUN(gunirsvp_lsps_show,
	    gunirsvp_lsps_show_cmd,
	    "show gunirsvp lsp",
	    "Show configured LSPs")
{
	RSVP::G2RSVPTE *                  proto;
	RSVP::G2RSVPTE::iterator_sessions sessionIter;
	RSVP::Session *                   session;
	RSVP::Session::iterator_lsps      lspIter;
	RSVP::LSP *                       lsp;
	std::ostringstream                os;
	int                               found;
	adj_type_t                        adjType;

	adjType = UNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	found = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			lsp = (*lspIter).second;
			assert(lsp);
			found = 1;
			os << *lsp;
			os << *session;
			os << VTY_NEWLINE;
		}
	}

	if (found) {
		vty_out(vty, "G.UNI-RSVP LSPs%s%s", VTY_NEWLINE, VTY_NEWLINE);
		vty_out(vty, "%s", os.str().c_str());
	} else {
		vty_out(vty, "No LSPs for this G.UNI-RSVP instance%s",
			VTY_NEWLINE);
	}
	return CMD_SUCCESS;
}

MYVTY_DEFUN(gennirsvp_lsps_show,
	    gennirsvp_lsps_show_cmd,
	    "show gennirsvp lsp",
	    "Show configured LSPs")
{
	RSVP::G2RSVPTE *                  proto;
	RSVP::G2RSVPTE::iterator_sessions sessionIter;
	RSVP::Session *                   session;
	RSVP::Session::iterator_lsps      lspIter;
	RSVP::LSP *                       lsp;
	std::ostringstream                os;
	int                               found;
	adj_type_t                        adjType;

	adjType = ENNI;

	proto = RSVP_MASTER.getProtoInstance(adjType);

	if (!proto) {
		vty_out(vty, "ERROR: Instance not found.%s", VTY_NEWLINE);
		return CMD_WARNING;
	}

	found = 0;
	for (sessionIter  = proto->begin_sessions();
	     sessionIter != proto->end_sessions();
	     sessionIter++) {
		session = (*sessionIter).second;
		assert(session);
		for (lspIter  = session->begin_lsps();
		     lspIter != session->end_lsps();
		     lspIter++) {
			lsp = (*lspIter).second;
			assert(lsp);
			found = 1;
			os << *lsp;
			os << *session;
			os << VTY_NEWLINE;
		}
	}

	if (found) {
		vty_out(vty, "G.ENNI-RSVP LSPs%s%s", VTY_NEWLINE, VTY_NEWLINE);
		vty_out(vty, "%s", os.str().c_str());
	} else {
		vty_out(vty, "No LSPs for this G.ENNI-RSVP instance%s",
			VTY_NEWLINE);
	}
	return CMD_SUCCESS;
}

static int g2rsvpte_config_write (struct vty * vty)
{
	vty_out(vty, "g2rsvpte%s", VTY_NEWLINE);
	vty_out(vty, "!%s",    VTY_NEWLINE);
	return 1;
}


ALIAS(gunirsvp,
      gunirsvp_compact_cmd,
      "gunirsvp",
      "Enable a G.UNI/N-RSVP instance\n")

ALIAS(g2rsvpte_show,
      g2rsvpte_show_cmd,
      "show g2rsvpte",
      "Show configured protocol instance")

ALIAS(gunirsvp_show,
      gunirsvp_show_cmd,
      "show gunirsvp",
      "Show configured protocol instance")

ALIAS(gennirsvp_show,
      gennirsvp_show_cmd,
      "show gennirsvp",
      "Show configured protocol instance")

ALIAS(g2rsvpte_lsp_create,
      g2rsvpte_lsp_create_compact_cmd,
      "lsp create tunnel-id WORD dst-rid A.B.C.D",
      "Create LSP with given input parameters")

ALIAS(g2rsvpte_lsp_update,
      g2rsvpte_lsp_update_compact_cmd,
      "lsp update data tunnel-id WORD dst-rid A.B.C.D "
      "in-te-link A.B.C.D "
      "in-data-link (WORD|A.B.C.D) "
      "in-label32-id WORD "
      "out-te-link A.B.C.D "
      "out-data-link (WORD|A.B.C.D) "
      "out-label32-id WORD "
      "sw-cap WORD "
      "enc-type WORD "
      "bw WORD "
      "gpid WORD ",
      "Update data into LSP")


ALIAS(g2rsvpte_lsp_update,
      g2rsvpte_lsp_update_compact_bis_cmd,
      "lsp update data tunnel-id WORD dst-rid A.B.C.D "
      "in-te-link A.B.C.D "
      "in-data-link (WORD|A.B.C.D) "
      "in-label32-id WORD "
      "out-te-link A.B.C.D "
      "out-data-link (WORD|A.B.C.D) "
      "out-label32-id WORD "
      "sw-cap WORD "
      "enc-type WORD "
      "bw WORD "
      "gpid WORD "
      "start-time WORD "
      "end-time WORD ",
      "Update data into LSP")

void RSVP_Master::init_vty(void)
{
	// Install g2rsvpte  top node.
	install_node (&g2rsvpte_node, g2rsvpte_config_write);

	install_element(CONFIG_NODE,   &g2rsvpte_cmd);
	install_element(CONFIG_NODE,   &no_g2rsvpte_cmd);
	install_element(CONFIG_NODE,   &gunirsvp_cmd);
	install_element(CONFIG_NODE,   &gunirsvp_compact_cmd);
	install_element(CONFIG_NODE,   &no_gunirsvp_cmd);
	install_element(CONFIG_NODE,   &gennirsvp_cmd);
	install_element(CONFIG_NODE,   &no_gennirsvp_cmd);

	install_default(G2RSVPTE_NODE);

	//protocol commands
	install_element(VIEW_NODE,     &g2rsvpte_gen_show_cmd);
	install_element(ENABLE_NODE,   &g2rsvpte_gen_show_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_gen_show_cmd);
	install_element(VIEW_NODE,     &gunirsvp_gen_show_cmd);
	install_element(ENABLE_NODE,   &gunirsvp_gen_show_cmd);
	install_element(G2RSVPTE_NODE, &gunirsvp_gen_show_cmd);
	install_element(VIEW_NODE,     &gennirsvp_gen_show_cmd);
	install_element(ENABLE_NODE,   &gennirsvp_gen_show_cmd);
	install_element(G2RSVPTE_NODE, &gennirsvp_gen_show_cmd);

	install_element(VIEW_NODE,     &g2rsvpte_show_cmd);
	install_element(ENABLE_NODE,   &g2rsvpte_show_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_show_cmd);
	install_element(VIEW_NODE,     &gunirsvp_show_cmd);
	install_element(ENABLE_NODE,   &gunirsvp_show_cmd);
	install_element(G2RSVPTE_NODE, &gunirsvp_show_cmd);
	install_element(VIEW_NODE,     &gennirsvp_show_cmd);
	install_element(ENABLE_NODE,   &gennirsvp_show_cmd);
	install_element(G2RSVPTE_NODE, &gennirsvp_show_cmd);

	//control interface commands
	install_element(VIEW_NODE,     &g2rsvpte_intf_show_cmd);
	install_element(ENABLE_NODE,   &g2rsvpte_intf_show_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_intf_show_cmd);
	install_element(VIEW_NODE,     &gunirsvp_intf_show_cmd);
	install_element(ENABLE_NODE,   &gunirsvp_intf_show_cmd);
	install_element(G2RSVPTE_NODE, &gunirsvp_intf_show_cmd);
	install_element(VIEW_NODE,     &gennirsvp_intf_show_cmd);
	install_element(ENABLE_NODE,   &gennirsvp_intf_show_cmd);
	install_element(G2RSVPTE_NODE, &gennirsvp_intf_show_cmd);

	//LSP commands
	install_element(VIEW_NODE,     &g2rsvpte_lsps_show_cmd);
	install_element(ENABLE_NODE,   &g2rsvpte_lsps_show_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsps_show_cmd);
	install_element(VIEW_NODE,     &gunirsvp_lsps_show_cmd);
	install_element(ENABLE_NODE,   &gunirsvp_lsps_show_cmd);
	install_element(G2RSVPTE_NODE, &gunirsvp_lsps_show_cmd);
	install_element(VIEW_NODE,     &gennirsvp_lsps_show_cmd);
	install_element(ENABLE_NODE,   &gennirsvp_lsps_show_cmd);
	install_element(G2RSVPTE_NODE, &gennirsvp_lsps_show_cmd);

	install_element(VIEW_NODE,     &g2rsvpte_lsp_show_cmd);
	install_element(ENABLE_NODE,   &g2rsvpte_lsp_show_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_show_cmd);

	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_create_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_create_compact_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_update_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_update_compact_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_update_compact_bis_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_behavior_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_commit_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_destroy_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_enable_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_disable_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_ipv4_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_ipv6_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_as_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_unnum_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_label_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_ero_commit_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_forceup_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_forcedown_cmd);

	// For debug
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_send_resv_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_send_confirm_cmd);
	install_element(G2RSVPTE_NODE, &g2rsvpte_lsp_post_xc_completed_cmd);
}
