//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>

#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"
#include "network.h"


#include "g2rsvpted.h"
#include "g2rsvpted/g2rsvpte_trace.h"
#include "g2rsvpted/g2rsvpte_lsp.h"


//
// CLASS RSVP_Master implementation
//
RSVP_Master::RSVP_Master(void)
{
	METIN();

	G2RSVPTE_DBG("Called RSVP_Master constructor");

	METOUT();
}

RSVP_Master::~RSVP_Master(void)
{
	METIN();

	G2RSVPTE_DBG("Called RSVP_Master destructor");

	METOUT();
}

bool
RSVP_Master::init(void)
{
	METIN();

	G2RSVPTE_DBG("Initialising RSVP_Master");

	instance_ = new RSVP_Master;
	if (!instance_) {
		zlog_err("Cannot allocate enough memory ...");
		METOUT();
		assert(0);
	}

	master_ = thread_master_create();
	if (!master_) {
		zlog_err(("Cannot create ZEBRA thread master "));
		assert(0);
	}

	g2rsvptes_.clear();

	start_time_ = quagga_time(NULL);

	METOUT();

	return true;
}

bool
RSVP_Master::destroy(void)
{
	METIN();

	G2RSVPTE_DBG("Finalising RSVP_Master");

	std::list<RSVP::G2RSVPTE *>::iterator iter;
	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		delete (*iter);
	}
	g2rsvptes_.clear();

	start_time_ = 0;

	delete instance_;
	instance_ = 0;

	thread_master_free(master_);
	master_ = 0;

	METOUT();

	return true;
}

RSVP_Master &
RSVP_Master::instance(void)
{
	METIN();

	if (instance_ == 0) {
		G2RSVPTE_DBG("RSVP_Master singleton instance is empty\n");

		instance_ = new RSVP_Master;
		if (!instance_) {
			zlog_err("Cannot allocate enough memory ...");
			METOUT();
			assert(0);
		}
	}

	METOUT();

	return * instance_;
}

struct thread_master *
RSVP_Master::getMaster(void)
{
	METIN();

	//G2RSVPTE_DBG("returning RSVP master of threads");
	METOUT();

	return master_;
}

std::list<RSVP::G2RSVPTE *>
RSVP_Master::getProtoList(void)
{
	METIN();
	std::list<RSVP::G2RSVPTE *> tmp;

	tmp.clear();

	tmp = g2rsvptes_;

	METOUT();

	return tmp;
}

int
RSVP_Master::getProtoListSize(void)
{
	METIN();
	METOUT();

	return g2rsvptes_.size();
}

RSVP::G2RSVPTE *
RSVP_Master::getProtoInstance(adj_type_t adjType)
{
	METIN();

	if (g2rsvptes_.empty()) {
		return 0;
	}

	RSVP::G2RSVPTE *                      proto;
	std::list<RSVP::G2RSVPTE *>::iterator iter;

	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		proto = *iter;
		if (proto->adjType() == adjType) {
			return proto;
		}
	}

// 	// XXX-FIXME when you have multiple instances
// 	proto = *(g2rsvptes_.begin());

// #if 0
// 	proto = rsvp_lookup();
// #endif

	METOUT();

	return 0;
}

RSVP::G2RSVPTE *
RSVP_Master::getProtoInstanceNoTun(void)
{
	METIN();

	if (g2rsvptes_.empty()) {
		return 0;
	}

	RSVP::G2RSVPTE *                      proto;
	std::list<RSVP::G2RSVPTE *>::iterator iter;

	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		proto = *iter;
		if (!proto->tunneledPackets()) {
			return proto;
		}
	}

	METOUT();

	return 0;
}

RSVP::G2RSVPTE *
RSVP_Master::getProtoFromScngwFd(int fd)
{
	METIN();

	if (g2rsvptes_.empty()) {
		return 0;
	}

	RSVP::G2RSVPTE *                      proto;
	std::list<RSVP::G2RSVPTE *>::iterator iter;

	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		proto = *iter;
		if (proto->scngwFd() == fd) {
			return proto;
		}
	}

	METOUT();

	return 0;
}

bool
RSVP_Master::attach(RSVP::G2RSVPTE * proto_instance)
{
	METIN();

	assert(proto_instance);

	std::list<RSVP::G2RSVPTE *>::iterator iter;

	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		if ((*iter) == proto_instance) {
			zlog_err("Proto instance already "
				 "attached to RSVP_Master");
			METOUT();
			return false;
		}
	}

	g2rsvptes_.push_front(proto_instance);

	G2RSVPTE_DBG("Proto instance attached to RSVP_Master");

	METOUT();

	return true;
}

bool
RSVP_Master::detach(RSVP::G2RSVPTE * proto_instance)
{
	METIN();

	assert(proto_instance);

	std::list<RSVP::G2RSVPTE *>::iterator iter;

	for (iter = g2rsvptes_.begin(); iter != g2rsvptes_.end(); iter++) {
		if ((*iter) == proto_instance) {
			g2rsvptes_.erase(iter);
			G2RSVPTE_DBG("Proto instance detached from "
				     "RSVP_Master");
			METOUT();
			return true;
		}
	}

	G2RSVPTE_DBG("Proto instance not found in RSVP_Master");

	METOUT();

	return false;
}

char *
RSVP_Master::getIorsDir(void)
{
	return iorsDir_;
}

void
RSVP_Master::setIorsDir(char * dir)
{
	iorsDir_ = dir;
}

void
RSVP_Master::manageMsg(bool tunneledPackets, g2mpls_addr_t addr, RSVP::Message * msg)
{
	if (!msg) {
		G2RSVPTE_ERR("Invalid message pointer!");
		METOUT();
		return;
	}

	RSVP::Message::Type type = (RSVP::Message::Type)msg->getType();

	RSVP::InterfaceKey_t key;
	key = RSVP::g2mpls_addr2key(addr);

	std::list<RSVP::G2RSVPTE *>           tmp;
	std::list<RSVP::G2RSVPTE *>::iterator it;
	RSVP::G2RSVPTE *                      proto;
	RSVP::Interface *                     intf;

	proto = 0;
	intf  = 0;
	tmp   = RSVP_MASTER.getProtoList();

	if (tunneledPackets) {
		// Get interface and proto instance
		for (it = tmp.begin(); it != tmp.end(); it++) {
			proto = *it;
			if (!proto) {
				G2RSVPTE_ERR("Invalid G2.RSVP-TE instance!");
				continue;
			}

			intf = proto->findInterface(key);
			if (intf) {
				G2RSVPTE_DBG("Found interface for G2.RSVP-TE "
					     "instance %p:", proto);
				intf->dump();
				break;
			}
		}

		if (!intf || !proto) {
			G2RSVPTE_ERR("No valid G2.RSVP-TE instance found for interface"
				     " key 0x%x!",
				     key);
			METOUT();
			G2RSVPTE_WRN("... deleting also Message '%s'",
				     MSGTYPE2STR(type));
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;
			return;
		}
		assert(intf->parent() == proto);
	} else {
		G2RSVPTE_DBG("Received packet for instance with packets not tunneled");
		// Get proto instance
		proto = RSVP_MASTER.getProtoInstanceNoTun();
		if (!proto) {
			G2RSVPTE_ERR("No instance with packets not tunneled is running!");
			METOUT();
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;
			return;
		}
	}

	// Get LSP (if any)
	lsp_ident_t lspInfo;
	bool        res;
	res = msg->getLSPinfo(lspInfo);
	if (!res) {
		G2RSVPTE_WRN("Possible incomplete information about lsp %d",
			     lspInfo.lsp_id);
	}

	G2RSVPTE_DBG("LSP info: ");
	G2RSVPTE_DBG("\t lspId   %d", lspInfo.lsp_id);
	G2RSVPTE_DBG("\t dst nid %s", numb_htoa(lspInfo.dst_nid));
	G2RSVPTE_DBG("\t src nid %s", numb_htoa(lspInfo.src_nid));
	G2RSVPTE_DBG("\t tunId   %d", lspInfo.tun_id);
	G2RSVPTE_DBG("\t exTunId %d", lspInfo.ext_tun_id);

	if (msg->isConsistent()) {

		if (type == RSVP::Message::NOTIFY ||
		    type == RSVP::Message::HELLO  ||
		    type == RSVP::Message::SREFRESH) {

			// TODO FIXME add code for this msg type
			G2RSVPTE_DBG("Message type %d not yet supported",
				     type);
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;

		} else if (type != RSVP::Message::ACK) { // PATH, RESV, ...

			RSVP::LSP * lsp;
			lsp = proto->findLSP(lspInfo);

			if (!lsp) {
				if (type == RSVP::Message::PATH) {
					RSVP::PathMessage * tmp;
					tmp = static_cast<RSVP::PathMessage *>(msg);

					if (!tmp->isDown()) {
						// try to install state for this new LSP
						lsp = proto->createLSP(lspInfo, intf, msg);
						if (!lsp) {
							G2RSVPTE_ERR("Impossible to create state for LSP:");
							G2RSVPTE_ERR("\t lspId   %d", lspInfo.lsp_id);
							G2RSVPTE_ERR("\t dst nid %s", numb_htoa(lspInfo.dst_nid));
							G2RSVPTE_ERR("\t src nid %s", numb_htoa(lspInfo.src_nid));
							G2RSVPTE_ERR("\t tunId   %d", lspInfo.tun_id);
							G2RSVPTE_ERR("\t exTunId %d", lspInfo.ext_tun_id);
							G2RSVPTE_WRN("... deleting also Message '%s'",
								     MSGTYPE2STR(type));
							//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
							delete msg;
							return;
						}
					} else {
						G2RSVPTE_ERR("no LSP found with LSP info:");
						G2RSVPTE_ERR("\t lspId   %d", lspInfo.lsp_id);
						G2RSVPTE_ERR("\t dst nid %s", numb_htoa(lspInfo.dst_nid));
						G2RSVPTE_ERR("\t src nid %s", numb_htoa(lspInfo.src_nid));
						G2RSVPTE_ERR("\t tunId   %d", lspInfo.tun_id);
						G2RSVPTE_ERR("\t exTunId %d", lspInfo.ext_tun_id);
						//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
						delete msg;
						return;
					}
				} else {
					G2RSVPTE_ERR("no LSP found with LSP info:");
					G2RSVPTE_ERR("\t lspId   %d", lspInfo.lsp_id);
					G2RSVPTE_ERR("\t dst nid %s", numb_htoa(lspInfo.dst_nid));
					G2RSVPTE_ERR("\t src nid %s", numb_htoa(lspInfo.src_nid));
					G2RSVPTE_ERR("\t tunId   %d", lspInfo.tun_id);
					G2RSVPTE_ERR("\t exTunId %d", lspInfo.ext_tun_id);
					//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
					delete msg;
					return;
				}
			}

			// Found/Created LSP
			if (intf) {
				G2RSVPTE_NOT("received G2.RSVP-TE %s msg for LSP id %d on interface %s",
					     MSGTYPE2STR(type), lspInfo.lsp_id, addr_ntoa(intf->localId()));
			} else {
				G2RSVPTE_NOT("received G2.RSVP-TE %s msg for LSP id %d",
					     MSGTYPE2STR(type), lspInfo.lsp_id);
			}
			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			switch (type) {
				case RSVP::Message::PATH: {
					RSVP::PathMessage * tmp;
					tmp = static_cast<RSVP::PathMessage *>(msg);
					if (tmp->isDown()) {
						ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvPathDown;
					} else {
						ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvPath;
					}
					break;
				}
				case RSVP::Message::RESV: {
					RSVP::ResvMessage * tmp;
					tmp = static_cast<RSVP::ResvMessage *>(msg);
					if (tmp->isDown()) {
						ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvResvDown;
					} else {
						ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvResv;
					}
					break;
				}
				case RSVP::Message::PATH_ERR: {
					ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvPathErr;
					break;
				}
				case RSVP::Message::RESV_ERR: {
					ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvResvErr;
					break;
				}
				case RSVP::Message::PATH_TEAR: {
					ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvPathTear;
					break;
				}
				case RSVP::Message::RESV_TEAR: {
					ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvResvTear;
				}
				case RSVP::Message::RESV_CONF: {
					ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_RecvConfirm;
					break;
				}
				//case RSVP::Message::ACK
				//case RSVP::Message::SREFRESH
				//case RSVP::Message::HELLO
				//case RSVP::Message::NOTIFY
				//case RSVP::Message::RECOVERY_PATH
				default:
					//TODO ????
					G2RSVPTE_ERR("Message type %s not supported at this level",
						     MSGTYPE2STR(type));
					//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
					delete msg;
					return; //???
			}

			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to allocate context for LSP fsm");

				// TODO: delete all useless data

				G2RSVPTE_WRN("... deleting also Message '%s'",
					     MSGTYPE2STR(type));
				//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
				delete msg;
				return;
			}

			context->lsp  = lsp;
			context->intf = intf;
			context->msg  = msg;

			assert(lsp->fsmInst_);
			lsp->fsmInst_->post(ev, context);
			// TODO check return state of FSM
			//if (!post) {
			//	//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			//	delete msg;
			//}

		} else {                                 // ACK msg

			// TODO FIXME add code for ACK msg
			G2RSVPTE_DBG("ACK message not yet supported");
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;
		}

	} else if (res) {
		if (type == RSVP::Message::PATH  ||
		    type == RSVP::Message::RESV) {

			RSVP::LSP * lsp;
			lsp = proto->findLSP(lspInfo);

			if (lsp) {
				// we look for LSP and send error msg
				// TODO FIXME
				// - get error code for msg
				// - dispatching an Err msg on this lsp
			}

			G2RSVPTE_DBG("Deleting msg type %d", type);
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;

		} else {
			G2RSVPTE_WRN("Discarding msg type %d", type);
			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			delete msg;
		}
	}

	METOUT();
}

// static
RSVP_Master *               RSVP_Master::instance_   = 0;
struct thread_master *      RSVP_Master::master_     = 0; // thread master
time_t                      RSVP_Master::start_time_ = 0; // start time
char *                      RSVP_Master::iorsDir_    = 0;
std::list<RSVP::G2RSVPTE *> RSVP_Master::g2rsvptes_;
