//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//  Copyright (C) 2006, 2007, 2008, 2009 University of Essex.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//  Eduard Escalona       (Univ. of Essex)   <eescal_at_essex.ac.uk>
//




#include <zebra.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <getopt.h>

#include "log.h"
#include "version.h"
#include "command.h"
#include "thread.h"
#include "sigevent.h"
#include "privs.h"
#include "memory.h"
#include "stream.h"
#include "sockopt.h"

#include "gunigwd.hh"

#include "BESFactoryBindingServer.hh"
#include "GridResourceRegistryBindingServer.hh"

struct thread_master * master = NULL;
static int             serverFd = -1;

// gunigwd privileges 
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

// GLOBALS
struct zebra_privs_t   gunigwd_privs;
struct quagga_signal_t gunigw_signals[4];
char *                 progname;

// Master of threads

const char *           pid_file         = PATH_GUNIGWD_PID; 
char                   config_default[] = GUNIGWD_DEFAULT_CONFIG;

GUNIGW_Master *             GUNIGW_Master::instance_  = 0;
struct thread_master *      GUNIGW_Master::master_    = 0; // thread master
time_t                      GUNIGW_Master::startTime_ = 0; // start time    
#if HAVE_OMNIORB
//CallController::Mgmt_var    GUNIGW_Master::nccMgmtProxy_ = 0;
//g2mplsTopology::Info_var    GUNIGW_Master::ospfProxy_    = 0;
#endif // HAVE_OMNIORB


struct option longopts[] = {
	{ "daemon",      no_argument,       NULL, 'd'},
	{ "config_file", required_argument, NULL, 'f'},
	{ "pid_file",    required_argument, NULL, 'i'},
	{ "iors_dir",    required_argument, NULL, 'o'},
        { "help",        no_argument,       NULL, 'h'},
	{ "vty_port",    required_argument, NULL, 'P'},
	{ "user",        required_argument, NULL, 'u'},
	{ "group",       required_argument, NULL, 'g'},
	{ "version",     no_argument,       NULL, 'v'},
	{ 0 }
};


/* SIG handlers */
static void 
sighup (void)
{
	zlog_info("SIGHUP received");
}

static int terminate_execution = 0;
static void
sigint (void)
{
	zlog_notice("Terminating on SIGINT");
	/* gunigw_terminate();*/
        terminate_execution = 1;
	exit(0);
}

/* SIGUSR1 handler. */
static void
sigusr1 (void)
{
	zlog_rotate (NULL);
}



/* Help information display. */
static void __attribute__ ((noreturn))
     usage (char *progname, int status)
{
	if (status != 0) {
		fprintf (stderr, 
			 "Try `%s --help' for more information.\n",
			 progname);
	} else {    
		printf ("Usage : %s [OPTION...]\n"
			"Daemon which manages RSVP.\n\n"
			"-d, --daemon       Runs in daemon mode\n"
			"-f, --config_file  Set configuration file name\n"
			"-i, --pid_file     Set process identifier file name\n"
		        "-o, --iors_dir     Set IORs directory\n"
                        "-P, --vty_port     Set vty's port number\n"
			"-u, --user         User to run as\n"
			"-g, --group        Group to run as\n"
			"-v, --version      Print program version\n"
			"-h, --help         Display this help and exit\n"
			"\n"
			"Report bugs to %s\n", 
			progname, ZEBRA_BUG_ADDRESS);
	}

	exit(status);
}



int main(int argc, char ** argv) 
{
	char          * p;
	int             vty_port;
	char          * vty_addr;
	int             daemon_mode;
	char          * config_file;
        char          * iors_dir;
	struct thread   thread;
	int fd_sock = 0;

	vty_port    = GUNIGWD_VTY_PORT;
	vty_addr    = NULL;
	daemon_mode = 0;
	config_file = NULL;
        iors_dir    = NULL;

	// INITIALIZATION
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	gunigwd_privs.user      = QUAGGA_USER;
	gunigwd_privs.group     = QUAGGA_GROUP;
#endif
#if defined(VTY_GROUP)
	gunigwd_privs.vty_group = VTY_GROUP;
#endif
	gunigwd_privs.caps_p    = _caps_p;
	gunigwd_privs.cap_num_p = sizeof(_caps_p)/sizeof(_caps_p[0]);
	gunigwd_privs.cap_num_i = 0;

	gunigw_signals[0].signal = SIGHUP;
	gunigw_signals[0].handler = &sighup;
	gunigw_signals[1].signal = SIGUSR1;
	gunigw_signals[1].handler = &sigusr1;
	gunigw_signals[2].signal = SIGINT;
	gunigw_signals[2].handler = &sigint;
	gunigw_signals[3].signal = SIGTERM;
	gunigw_signals[3].handler = &sigint;

	// Set umask before anything for security 
	umask(0027);

	// Get program name 
	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	// Initialize the log subsystem
	zlog_default = openzlog(progname, 
				ZLOG_GUNIGW,
				LOG_CONS | LOG_NDELAY | LOG_PID, 
				LOG_DAEMON);

	if (!GUNIGW_Master::init()) {
		zlog_notice("Cannot allocate an GUNIGW master object");
		exit (-1);
	}
	master = GUNIGW_MASTER.getMaster();

	// just reads the options from stdin
	while (1) {
		int opt;
		
		opt = getopt_long(argc, argv, "df:h:P:u:g:v:i:o:", longopts, 0);
		
		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'f':
				config_file = optarg;
				break;
                        case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
                        case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				gunigwd_privs.user = optarg;
				break;
			case 'g':
				gunigwd_privs.group = optarg;
				break;
			case 'v':
				print_version(progname);
				exit (0);
				break;
			case 'h':
				usage(progname, 0);
				break; 
			default:
				usage(progname, 1);
				break;
		}
	}

        // Library inits.
	zprivs_init(&gunigwd_privs);
	signal_init(GUNIGW_MASTER.getMaster(), 
		    Q_SIGC(gunigw_signals), 
		    gunigw_signals);
	cmd_init(1);

	// these are all from libzebra 
	vty_init(GUNIGW_MASTER.getMaster());
	memory_init();

	GUNIGW_MASTER.init_vty();

        GUNIGW_DBG("init_BESFactory()\n");
        init_BESFactory();
        GUNIGW_DBG("init_GRR()\n");
        init_GRR();

        // sorting is needed by the command subsystem to finalize init 
	// Change to the daemon program.
	if (daemon_mode) {
		daemon(0, 0);
	}

        // initialize gunigw_lock
	gunigw_lock(false, 0);
	gunigw_lock(false, 1);

        // Process id file create.
        if (!pid_output(pid_file))
            GUNIGW_ERR("ERROR in pid_output\n");

#if HAVE_OMNIORB

        try {
            corba_init(iors_dir,
                CORBA_SERVANT_GUNIGW,
                GUNIGW_MASTER.getMaster());
        } catch (std::string & e) {
            fprintf(stderr, "Cannot initialize CORBA subsystem (err=%s)\n", e.c_str());
            exit(-1);
        }

        try {
            corba_server_setup();
        } catch (std::string & e) {
            fprintf(stderr, "Cannot setup CORBA subsystem (err=%s)\n", e.c_str());
            exit(1);
        }
  
        try {
            corba_client_setup(GUNIGW_MASTER.getNccProxy());
        } catch (std::string & e) {
            fprintf(stderr, "Cannot setup client object (err=%s)\n", e.c_str());
            exit(-1);
        } catch (CORBA::BAD_PARAM & e) {
            fprintf(stderr, "Cannot setup client object " "(CORBA::BAD_PARAM)\n");
            exit(-1);
        } catch (...) {
            fprintf(stderr, "Cannot setup client object " "(exception other)\n");
            exit(-1);
        }

        try {
            corba_client_setup(GUNIGW_MASTER.getNccEWProxy());
        } catch (std::string & e) {
            fprintf(stderr, "Cannot setup client object (err=%s)\n", e.c_str());
            exit(-1);
        } catch (CORBA::BAD_PARAM & e) {
            fprintf(stderr, "Cannot setup client object " "(CORBA::BAD_PARAM)\n");
            exit(-1);
        } catch (...) {
            fprintf(stderr, "Cannot setup client object " "(exception other)\n");
            exit(-1);
        }

	try {
	    corba_client_setup(GUNIGW_MASTER.getospfProxy());
	} catch (std::string & e) {
            fprintf(stderr, "Cannot setup client object (err=%s)\n", e.c_str());
            exit(-1);
        } catch (CORBA::BAD_PARAM & e) {
            fprintf(stderr, "Cannot setup client object " "(CORBA::BAD_PARAM)\n");
            exit(-1);
        } catch (...) {
            fprintf(stderr, "Cannot setup client object " "(exception other)\n");
            exit(-1);
        }

        try {
            corba_client_setup(GUNIGW_MASTER.getLrmProxy());
        } catch (std::string & e) {
            fprintf(stderr, "Cannot setup client object (err=%s)\n", e.c_str());
            exit(-1);
        } catch (CORBA::BAD_PARAM & e) {
            fprintf(stderr, "Cannot setup client object " "(CORBA::BAD_PARAM)\n");
            exit(-1);
        } catch (...) {
            fprintf(stderr, "Cannot setup client object " "(exception other)\n");
            exit(-1);
        }

        try {
            LRM::Info_var lrmProxy;
            lrmProxy = GUNIGW_MASTER.getLrmProxy();
            if (CORBA::is_nil(lrmProxy)) {
                fprintf(stderr,
                        "Cannot get Corba LRM client object "
                        "previously set up\n");
                assert(0);
            }
            g2mplsTypes::nodeId myNid;
            myNid = lrmProxy->nodeId();
            GUNIGW_MASTER.routerId(myNid);
        } catch (...) {
            fprintf(stderr,
                    "Cannot get  Router ID from LRM\n");
            exit(-1);
        }

#endif

        vty_read_config(config_file, config_default);	// Get .conf file

	// Create VTY socket and start the TCP and unix socket listeners 
        vty_serv_sock(vty_addr, vty_port, GUNIGW_VTYSH_PATH);

	// Print banner. 
	zlog_notice("GUNIGW daemon v.%s: starting... vty@%d", GUNIGW_CODE_VERSION, vty_port);

	// Fetch next active thread.
        while (!terminate_execution) {
            if(thread_fetch(GUNIGW_MASTER.getMaster(), &thread))
            {
		// the main event loop
                thread_call(&thread);
            }
	}
#if HAVE_OMNIORB
	try {
            (void) corba_client_shutdown(CORBA_SERVANT_GUNIGW);
        } catch(...) {
            fprintf(stderr, "Somethig was wrong with corba shutdown\n");
            exit(-1);
	}

        try {
            (void) corba_server_shutdown();
        } catch(...) {
            fprintf(stderr, "Something was wrong with corba server shutdown\n");
            exit(-1);
        }
#endif
                
        return 0;
}
