//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//  Copyright (C) 2006, 2007, 2008, 2009 University of Essex.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//  Eduard Escalona       (Univ. of Essex)   <eescal_at_essex.ac.uk>
//



#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"

#include <map>
#include <list>
#include <string>
#include <sstream>

#include "hg2gwd.hh"
#include "hg2gw_trace.hh"
#include "hg2gw_utils.hh"

#define CORBAACTORDUMP2VTY(actor_, owner_)				     \
{									     \
	vty_out(vty, "%s actor%s", owner_, VTY_NEWLINE);		     \
	vty_out(vty, "\t - issuer          : %s%s",			     \
		((actor_->issuer == g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF) ? \
		 "MANAGEMENT" :					     \
		 ((actor_->issuer == g2mplsTypes::ISSUERTYPE_UNI_IF) ?	     \
		  "UNI" :						     \
		  "ENNI")),						     \
		VTY_NEWLINE);						     \
	vty_out(vty, "\t - forces command  : %s%s",			     \
		(actor_->forceCommand) ? "YES" : "NO", VTY_NEWLINE);	     \
}

#define CORBASTATESDUMP2VTY(states_, owner_)				  \
{									  \
	vty_out(vty, "%s states%s", owner_, VTY_NEWLINE);		  \
	vty_out(vty, "\t - op. state       : %s%s",			  \
		(states_->opState == g2mplsTypes::OPERSTATE_UP) ?	  \
		"UP" : "DOWN", VTY_NEWLINE);				  \
	vty_out(vty, "\t - adm. state      : %s%s",			  \
		(states_->admState == g2mplsTypes::ADMINSTATE_DISABLED) ? \
		"DISABLED" : "ENABLED", VTY_NEWLINE);			  \
}


//
//
//  CALL container
//
//

#define CALLCTRL_SCOPE "NULL, selected by CallCtrl"
#define CALLCTRL_WORK "(if " CALLCTRL_SCOPE")"

#define BAR1 "========================================"	\
             "========================================"
#define BAR2 "----------------------------------------"	\
             "----------------------------------------"
#define SHOWLIST_DEPTH 50

#define VTY_ERROR_PRINT(VTY, ERR)			\
   vty_out(VTY, "%s%s", (ERR).c_str(), VTY_NEWLINE);

#define VTY_ERROR_AND_EXIT(VTY, ERR)		\
   VTY_ERROR_PRINT(VTY, ERR);			\
   return CMD_WARNING;

//
//
//
//   VTY commands
//
//
//

struct cmd_node hg2gw_node = {
	HG2GW_NODE,
	"%s(config-hg2gw)# ",
	1
};

DEFUN(hg2gw,
      hg2gw_cmd,
      "hg2gw",
      "Enable a HG2GW process\n"
      "Start HG2GW configuration\n")
{
	std::string  err;

	vty->node  = HG2GW_NODE;
	vty->index = NULL;

	vty_out(vty,
		"Entering HG2GW Controller VTY interface%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(no_hg2gw,
      no_hg2gw_cmd,
      "no hg2gw",
      NO_STR
      "Disable a HG2GW process\n"
      "Stop HG2GW operation\n")
{
	std::string  err;

	vty_out(vty,
		"Exiting HG2GW Controller VTY interface%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

//
//
// CONFIG COMMANDS
//
//
//

DEFUN(hg2gw_call_create,
      hg2gw_call_create_cmd,
      "call create id WORD itu-country WORD itu-carrier WORD unique-AP WORD",
      "Create CALL with given input parameters")
{
	mapKey_t         key;
	uint32_t         ituCountryCode;	//  just 24 bits
	uint64_t         ituCarrierCode;	//  just 48 bits
	uint64_t         uniqueAP;		//  just 48 bits

	std::string      err;

	ituCountryCode = 0;
	ituCarrierCode = 0;
	uniqueAP       = 0;

	key = strtoull(argv[0], NULL, 16);


	if (argc > 1)
        {
		ituCountryCode  = strtoul(argv[1], NULL, 16);
		ituCountryCode &= 0x00FFFFFF;	//  just 24 bits

		ituCarrierCode  = strtoull(argv[2], NULL, 16);
		ituCarrierCode &= 0x0000FFFFFFFFFFFFull;	// 48 bits

		uniqueAP        = strtoull(argv[3], NULL, 16);
		uniqueAP       &= 0x0000FFFFFFFFFFFFull;	// 48 bits
	}

	HG2GW_DBG("Going to create a HG2GW NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key,ituCountryCode,ituCarrierCode,uniqueAP,err))
        {
		vty_out(vty,
			"Cannot update data into VTY CALL with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s successfully created in VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(hg2gw_call_create,
      hg2gw_call_create_compact_cmd,
      "call create id WORD ",
      "Create CALL with given input parameters");

DEFUN(hg2gw_call_commit,
      hg2gw_call_commit_cmd,
      "call commit id WORD",
      "Commit CALL with given input parameters")
{
	mapKey_t      key;
	std::string   err;

	key = strtoull(argv[0], NULL, 16);

	HG2GW_DBG("Going to commit a HG2GW NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallCommit(key, err)) {
		vty_out(vty,
			"Cannot commit a HG2GW NCC CALL from VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s successfully committed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_call_destroy,
      hg2gw_call_destroy_cmd,
      "call destroy id WORD src-lsr A.B.C.D",
      "Destroy CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}


	HG2GW_DBG("Going to destroy a HG2GW NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Sending callDestroy through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;
		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callDestroy(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		HG2GW_DBG("Sent callDestroy through CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s and src LSR %s "
		"successfully destroyed%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_call_enable,
      hg2gw_call_enable_cmd,
      "call enable id WORD src-lsr A.B.C.D",
      "Enable CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	HG2GW_DBG("Going to enable a HG2GW NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Sending callEnable through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);

		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callEnable(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		HG2GW_DBG("Sent callEnable through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully enabled%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_call_disable,
      hg2gw_call_disable_cmd,
      "call disable id WORD src-lsr A.B.C.D",
      "Disable CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	HG2GW_DBG("Going to disable a HG2GW NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Sending callDisable through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);

		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callDisable(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		HG2GW_DBG("Sent callDisable through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s and src LSR %s "
		"successfully disabled%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_call_setup,
      hg2gw_call_setup_cmd,
      "call set-up id WORD src-lsr A.B.C.D",
      "Set up a CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	HG2GW_DBG("Going to setup a HG2GW NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Sending callSetUp through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callSetUp(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		HG2GW_DBG("Sent callSetUp through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s and src LSR %s "
		"successfully set-up%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_call_setdown,
      hg2gw_call_setdown_cmd,
      "call set-down id WORD src-lsr A.B.C.D",
      "Set down a CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	HG2GW_DBG("Going to setdown a HG2GW NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Sending callSetDown through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set down");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callSetDown(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		HG2GW_DBG("Sent callSetDown through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"HG2GW NCC CALL with localID %s and src LSR %s "
		"successfully set-down%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}


//
//
// SHOW COMMANDS
//
//
//

DEFUN(hg2gw_show,
      hg2gw_show_cmd,
      "show hg2gw",
      "Show configured protocol instance")
{
	size_t      vtyCalls;
	size_t      nccCalls;
	std::string err;

	if (!getCalls(vtyCalls, nccCalls, err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "HG2GW CONTROLLER VTY INSTANCE%s", VTY_NEWLINE);

	vty_out(vty, "   - HG2GW NCC CALL(s) pending in VTY   : %d%s",
		vtyCalls, VTY_NEWLINE);

	vty_out(vty, "   - HG2GW NCC CALL(s) committed to NCC : %d%s",
		nccCalls, VTY_NEWLINE);

	vty_out(vty, "%s", VTY_NEWLINE);
	return CMD_SUCCESS;
}

DEFUN(hg2gw_show_calls,
      hg2gw_show_calls_cmd,
      "show hg2gw call",
      "Show all the calls on this HG2GW")
{
	int         i;
	std::string err;

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "HG2GW NCC CALL(s) installed on HG2GW%s", VTY_NEWLINE);
	if (!HG2GW_MASTER.dryrun()) {
#if HAVE_OMNIORB
		HG2GW_DBG("Requesting list of Calls");

		CallController::Mgmt_var  nccMgmtProxy;

		nccMgmtProxy = HG2GW_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			HG2GW_ERR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		CallController::Mgmt::callIdentSeq_var s;

		try {
			s = nccMgmtProxy->getCalls();
		} catch (Types::InternalProblems & e) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		if (s->length() == 0) {
			vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
			vty_out(vty, "%s", VTY_NEWLINE);
			return CMD_SUCCESS;
		}

		size_t i;

		for (i = 0; i < s->length(); i++) {
			g2mpls_addr_t lsr;
			uint64_t      id;
			try {
				g2mplsTypes::sourceId_var corbaLsr;
				corbaLsr = (s)[i].srcId;
				lsr << corbaLsr;
				id = (s)[i].localId;
			} catch (std::out_of_range & e) {
				vty_out(vty,
					"Exception occurred during "
					"types translation (%s)%s",
					e.what(), VTY_NEWLINE);
				return CMD_WARNING;
			}

			vty_out(vty,
				"   - [c] #%02d: "
				"src-LSR = %s / local-ID = %s%s",
				i,
				addr_ntoa(lsr),
				key2string(id).c_str(),
				VTY_NEWLINE);

			if (i == SHOWLIST_DEPTH) {
				break;
			}
		}

		vty_out(vty, "%s", VTY_NEWLINE);
		vty_out(vty, "   printed %d HG2GW NCC committed CALL(s)%s",
			i, VTY_NEWLINE);
#endif // HAVE_OMNIORB
	} else {
		vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
	}

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);

	vtyCallDumpVty(vty);

	vty_out(vty, "%s", VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(hg2gw_show_call,
      hg2gw_show_call_cmd,
      "show hg2gw call id WORD src-lsr A.B.C.D",
      "Show details for specified CALL")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL)
        {
		/* matched "A.B.C.D". */
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4)))
                {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}
        else
        {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "HG2GW NCC installed CALL%s", VTY_NEWLINE);

	if (HG2GW_MASTER.dryrun()) {
		vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}
#if HAVE_OMNIORB
	HG2GW_DBG("Requesting details for Call with localID %s "
		  "and src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));

	CallController::Mgmt_var  nccMgmtProxy;

	nccMgmtProxy = HG2GW_MASTER.getNccProxy();

	if (CORBA::is_nil(nccMgmtProxy)) {
		HG2GW_ERR("Cannot get Corba nccMgmtProxy client object "
			  "previously set up");
		assert(0);
	}

	g2mplsTypes::callIdent_var id;
	//	id = new g2mplsTypes::callIdent;

	try {
		id << cid;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callId translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	g2mplsTypes::callParams_var                   cinfo;
	g2mplsTypes::recoveryParams_var               rinfo;
	g2mplsTypes::lspParams_var                    linfo;
	g2mplsTypes::actorInfo_var                    actor;
	g2mplsTypes::statesBundle_var                 states;
	CallController::Mgmt::recoBundleIdentSeq_var  recoBundles;

	try {
		nccMgmtProxy->callGetDetails(id         ,
					     cinfo      ,
					     rinfo      ,
					     linfo      ,
					     actor      ,
					     states     ,
					     recoBundles);
	} catch (Types::InternalProblems & e) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for internal problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (Types::CannotFetch & e) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for fetching problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (...) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for unknown reason";
		VTY_ERROR_AND_EXIT(vty, err);
	}

	call_info_t           cdata;
	recovery_info_t       rdata;
	lsp_info_t            ldata;

	memset(&cdata, 0, sizeof(cdata));
	memset(&rdata, 0, sizeof(rdata));
	memset(&ldata, 0, sizeof(ldata));

	memset(&cid,   0, sizeof(cid));	// reset cid

	try {
		cid << id;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callId translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	try {
		cdata << cinfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!getTna(cid, cdata, 0 /* !isEgress */, err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	if (!getTna(cid, cdata, 1 /*  isEgress */, err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	vtyDump(vty, cid);
	vtyDump(vty, cdata);
	CORBAACTORDUMP2VTY(actor, "Call");
	CORBASTATESDUMP2VTY(states, "Call");

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Recovery Bundles under this CALL%s",
		VTY_NEWLINE);

	try {
		rdata << rinfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"recoveryData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	vtyDump(vty, rdata);

	size_t i;

	for (i = 0; i < recoBundles->length(); i++) {
		g2mpls_addr_t srcLsr;
		g2mpls_addr_t dstLsr;
		uint32_t      tid;

		try {
			srcLsr << (recoBundles)[i].srcAddr;
			dstLsr << (recoBundles)[i].dstAddr;
			tid = (recoBundles)[i].tunId;
		} catch (std::out_of_range & e) {
			vty_out(vty,
				"Exception occurred during "
				"types translation (%s)%s",
				e.what(), VTY_NEWLINE);
			return CMD_WARNING;
		}

		vty_out(vty,
			"   - #%02d: src-LSR = %s / "
			"dst-LSR = %s / tunnel-ID = 0x%08x%s",
			i,
			addr_ntoa(srcLsr),
			addr_ntoa(dstLsr),
			tid,
			VTY_NEWLINE);

		if (i == SHOWLIST_DEPTH) {
			break;
		}
	}

	try {
		ldata << linfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"recoveryData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Common LSP data under this CALL%s",
		VTY_NEWLINE);
	vtyDump(vty, ldata);

#endif // HAVE_OMNIORB

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	return CMD_SUCCESS;
}


static int hg2gw_config_write (struct vty * vty)
{
	vty_out(vty, "hg2gw%s", VTY_NEWLINE);
	vty_out(vty, "!%s",   VTY_NEWLINE);

	return 1;
}

void HG2GW_Master::init_vty(void)
{
	/* Install hg2gw  top node. */
	install_node (&hg2gw_node, hg2gw_config_write);

	install_element(CONFIG_NODE, &hg2gw_cmd);
	install_element(CONFIG_NODE, &no_hg2gw_cmd);

	install_default(HG2GW_NODE);

	// imperative commands
	install_element(HG2GW_NODE, &hg2gw_call_create_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_create_compact_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_callres32_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_callres32_compact_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_callres60_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_callres60_compact_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_calldnid_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_calltime_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_callnames_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_recovery_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_lspbase_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_lspbase_compact_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_lspconstr_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_lspproto_cmd);
	//install_element(HG2GW_NODE, &ncc_call_set_info_lspprototimers_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_commit_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_destroy_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_enable_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_disable_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_setup_cmd);
	install_element(HG2GW_NODE, &hg2gw_call_setdown_cmd);
        
        //protocol commands
	install_element(VIEW_NODE,   &hg2gw_show_cmd);
	install_element(ENABLE_NODE, &hg2gw_show_cmd);
	install_element(HG2GW_NODE,  &hg2gw_show_cmd);
        install_element(VIEW_NODE,   &hg2gw_show_calls_cmd);
	install_element(ENABLE_NODE, &hg2gw_show_calls_cmd);
	install_element(HG2GW_NODE,  &hg2gw_show_calls_cmd);
        install_element(VIEW_NODE,   &hg2gw_show_call_cmd);
        install_element(ENABLE_NODE, &hg2gw_show_call_cmd);
        install_element(HG2GW_NODE,  &hg2gw_show_call_cmd);
        
}
