/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *  Copyright (C) 2006, 2007, 2008, 2009 PSNC
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 *  Adam Kaliszan         (PSNC)             <kaliszan_at_man.poznan.pl>
 */



#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "if.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"
#include "zclient.h"
#include "lrmd.h"
#include "lrm_zebra.h"
#include "lrm_core.h"
#include "lrm_vty.h"


struct cmd_node lrm_node = {
	LRM_NODE,
	"%s(config-lrm)# ",
	1
};

DEFUN(lrm,
      lrm_cmd,
      "lrm",
      "Enable a LRM process\n"
      "Start LRM configuration\n")
{
	vty->node  = LRM_NODE;
	vty->index = lrm_get();

	return CMD_SUCCESS;
}

DEFUN(no_lrm,
      no_lrm_cmd,
      "no lrm",
      NO_STR
      "Disable a LRM process\n"
      "Stop LRM operation\n")
{
	lrm_t * lrm;

	lrm = lrm_lookup();
	if (lrm == NULL) {
		vty_out (vty,
			 "There isn't an active LRM instance%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	lrm_finish(lrm);

	return CMD_SUCCESS;
}


DEFUN (lrm_router_id,
       lrm_router_id_cmd,
       "(router-id|router-inni-id|router-enni-id|router-uni-id) A.B.C.D [RC_IF_PREFLEN]",
       "Router ID for the G2MPLS stack, INNI instance\n"
       "Router ID for the G2MPLS stack, INNI instance\n"
       "Router ID for the G2MPLS stack, ENNI instance\n"
       "Router ID for the G2MPLS stack, UNI instance\n"
       "A Router ID in IP address format\n"
       "Routing Controller interface pref length (required if the ENNI RC interface addres is equal to ENNI RC id)\n")
{
	ipv4_t    addr;
	u_int32_t router_id;
	lrm_t *   lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&addr, 0, sizeof(addr));

	if (!inet_aton(argv[1], &addr)) {
		vty_out(vty,"Invalid IP address (%s) for a Router ID%s",argv[1], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for a Router ID",argv[1]);
		return LRM_ERROR;
	}

	router_id = ntohl(addr.s_addr);

	if ((argc>2) && (strcmp(argv[0], "router-enni-id") == 0))
	{
		int addr_len = strtoul (argv[2], NULL, 10);
		struct interface * ifp     = if_get_by_name_len("Level1", INTERFACE_NAMSIZ);
		ifp->ifindex               = 10000; //++lrm->telink_count; TODO use LRMD index counter
		ifp->flags                 = IFF_UP|IFF_RUNNING;
		ifp->mtu                   = 65536;
		ifp->metric                = 1;
		ifp->bandwidth             = 10000;
		ifp->status                = ZEBRA_INTERFACE_ACTIVE;

		struct connected    * ifc  = connected_new();
		struct prefix_ipv4  * p    = prefix_ipv4_new ();

		p->family                  = AF_INET;
		p->prefix                  = addr;
		p->prefixlen               = addr_len;

		ifc->address               = (struct prefix *) p;
		ifc->ifp                   = ifp;

		listnode_add (ifp->connected, ifc);

		/*now we send the te link TE params to ospf*/
		ifp->adj_type              = ENNI;
		ifp->ospf_instance         = ENNI;
		ifp->te_TNA_address_type   = IPv4;
		ifp->te_TNA_prefix_length  = 0;
		ifp->te_TNA_address        = (u_int32_t *) malloc(4);
		memset(ifp->te_TNA_address, 0, 4);
		lrm_interface_add_notify(ifp);
	}
	adj_type_t adjacency = (strcmp(argv[0], "router-id") == 0) ? INNI : (strcmp(argv[0], "router-inni-id") == 0) ? INNI : (strcmp(argv[0], "router-enni-id") == 0) ? ENNI : UNI;

	if (lrm_set_rid(lrm, adjacency, router_id) == LRM_OK) {
		vty_out(vty, "Router ID %s set%s", argv[1], VTY_NEWLINE);
		zlog_info("Conf:Router ID %s set", argv[1]);
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Router ID %s cannot be set%s",
			argv[1], VTY_NEWLINE);
	}

	return CMD_WARNING;
}

DEFUN(lrm_show,
      lrm_show_cmd,
      "show lrm",
      "Show configured control interfaces")
{
	lrm_t *     lrm;
	u_int32_t   scn_ifs;
	u_int32_t   ccs;
	u_int32_t   datalinks;
	u_int32_t   telinks;
	u_int32_t   adjs;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	scn_ifs  = (lrm->scn_if_list  ) ? listcount(lrm->scn_if_list  ) : 0;
	ccs      = (lrm->cc_list      ) ? listcount(lrm->cc_list      ) : 0;
	datalinks= (lrm->datalink_list) ? listcount(lrm->datalink_list) : 0;
	telinks  = (lrm->telink_list  ) ? listcount(lrm->telink_list  ) : 0;
	adjs     = (lrm->adj_list     ) ? listcount(lrm->adj_list     ) : 0;

	vty_out (vty, "LRM INSTANCE%s", VTY_NEWLINE);
	vty_out (vty, "----------------------%s", VTY_NEWLINE);
	vty_out (vty, "   - router-id        : %s %s",
		 numb_htoa(lrm->router_id), VTY_NEWLINE);
	vty_out (vty, "   - SCN interfaces   : %d %s",
		 scn_ifs, VTY_NEWLINE);
	vty_out (vty, "   - Control Channels : %d %s",
		 ccs, VTY_NEWLINE);
	vty_out (vty, "   - Adjacencies      : %d %s",
		 adjs, VTY_NEWLINE);
	vty_out (vty, "   - TE links         : %d %s",
		 telinks, VTY_NEWLINE);
	vty_out (vty, "   - Data Links       : %d %s",
		 datalinks, VTY_NEWLINE);

	return CMD_SUCCESS;
}

/*
 *
 * handler for the scn-if command
 *
 */
DEFUN(lrm_scnif_add,
      lrm_scnif_add_cmd,
      "scn-if add ip A.B.C.D",
      "add | delete | enable | disable a control interface")
{
	if_type_t      type;
	g2mpls_addr_t  addr;
	lrm_t *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	type = BCAST;		/* XXX-FIXME: WIRED for the time being */
	memset(&addr, 0, sizeof(addr));
	addr.type    = IPv4;
	addr.preflen = 32;

	if (!inet_aton(argv[0], &(addr.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for snc scn-if%s",	argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for snc scn-if",	argv[0]);
		return LRM_ERROR;
	}

	if (scn_if_add(lrm, addr, type) == LRM_OK) {
		vty_out(vty,"Control interface %s added%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Control interface %s added",argv[0]);
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Control interface cannot be added: the IP address %s does not exist%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Control interface cannot be added: the IP address %s does not exist",argv[0]);
	}

	return CMD_WARNING;
}


DEFUN(lrm_scnif_delete,
      lrm_scnif_delete_cmd,
      "scn-if delete ip A.B.C.D",
      "Delete a control interface")
{
	g2mpls_addr_t    addr;
	lrm_t *          lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&addr, 0, sizeof(addr));
	addr.type    = IPv4;
	addr.preflen = 32;

	if (!inet_aton(argv[0], &(addr.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for snc scn-if%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	if (scn_if_del(lrm, addr) == LRM_OK) {
		vty_out(vty,
			"Control interface %s deleted%s",
			argv[0], VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Cannot delete control interface %s%s",
		argv[0], VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_scnif_delete,
      lrm_scnif_del_cmd,
      "scn-if del ip A.B.C.D",
      "Delete a control interface");

DEFUN(lrm_scnif_enable,
      lrm_scnif_enable_cmd,
      "scn-if enable ip A.B.C.D",
      "Enable a control interface")
{
	g2mpls_addr_t    addr;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&addr, 0, sizeof(addr));
	addr.type    = IPv4;
	addr.preflen = 32;

	if (!inet_aton(argv[0], &(addr.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for snc scn-if%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for snc scn-if",argv[0]);
		return LRM_ERROR;
	}

	if (scn_if_ena(lrm, addr) == LRM_OK) {
		vty_out(vty,"Control interface %s enabled%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Control interface %s enabled",argv[0]);
		return CMD_SUCCESS;
	}

	vty_out(vty,"Cannot enable control interface %s%s",argv[0], VTY_NEWLINE);
	zlog_info("Conf:Cannot enable control interface %s",argv[0]);
	return CMD_WARNING;
}

ALIAS(lrm_scnif_enable,
      lrm_scnif_en_cmd,
      "scn-if en ip A.B.C.D",
      "Enable a control interface");

DEFUN(lrm_scnif_disable,
      lrm_scnif_disable_cmd,
      "scn-if disable ip A.B.C.D",
      "Disable a control interface")
{
	g2mpls_addr_t    addr;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&addr, 0, sizeof(addr));
	addr.type    = IPv4;
	addr.preflen = 32;

	if (!inet_aton(argv[0], &(addr.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for snc scn-if%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	if (scn_if_dis(lrm, addr) == LRM_OK) {
		vty_out(vty,
			"Control interface %s disabled%s",
			argv[0], VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Cannot disable control interface %s%s",
		argv[0], VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_scnif_disable,
      lrm_scnif_dis_cmd,
      "scn-if dis ip A.B.C.D",
      "Disable a control interface");

DEFUN(lrm_scnif_show,
      lrm_scnif_show_cmd,
      "show lrm scn-if",
      "Show configured control interfaces")
{
	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node;
	ctrl_intf_t      * ctrl_if;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	anylist = lrm->scn_if_list;
	if (anylist == NULL) {
		return CMD_ERR_NOTHING_TODO;
	}

	vty_out (vty, "SCN INTERFACES%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	for (ALL_LIST_ELEMENTS_RO(anylist, node, ctrl_if)) {
		ASSERT(ctrl_if);
		vty_out (vty,
			 "----------------------%s", VTY_NEWLINE);
		vty_out (vty,
			 "   interface %s%s",
			 addr_ntoa(ctrl_if->addr), VTY_NEWLINE);
		vty_out (vty,
			 "       - type      : %s%s",
			 SHOW_IFTYPE(ctrl_if->type),VTY_NEWLINE);
		vty_out (vty,
			 "       - op state  : %s%s",
			 SHOW_OPSTATE(ctrl_if->op_state), VTY_NEWLINE);
		vty_out (vty,
			 "       - adm state : %s%s",
			 SHOW_ADMSTATE(ctrl_if->adm_state), VTY_NEWLINE);
	}

	return CMD_SUCCESS;
}


/*
 *
 * handler for the cc command
 *
 */
DEFUN(lrm_cc_add,
      lrm_cc_add_cmd,
      "cc add ccid WORD scn-ip A.B.C.D scn-nbr A.B.C.D"
      /*
	config-send ci <0-4294967295> hi <0-4294967295> hdi <0-4294967295>"
      */,
      "add | delete | enable | disable | up | down a control channel")
{
	int            cc_id;
	g2mpls_addr_t  lcl_scn_addr;
	g2mpls_addr_t  rem_scn_addr;
	lrm_t *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	memset(&lcl_scn_addr, 0, sizeof(lcl_scn_addr));
	memset(&rem_scn_addr, 0, sizeof(rem_scn_addr));

	lcl_scn_addr.type    = IPv4;
	lcl_scn_addr.preflen = 32;
	if (!inet_aton(argv[1], &(lcl_scn_addr.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for local snc scn-if%s",
			argv[1], VTY_NEWLINE);
		return LRM_ERROR;
	}

	rem_scn_addr.type    = IPv4;
	rem_scn_addr.preflen = 32;
	if (!inet_aton(argv[2], &(rem_scn_addr.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for remote snc scn-if%s",argv[2], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for remote snc scn-if",argv[2]);
		return LRM_ERROR;
	}

	if (control_channel_add(lrm, cc_id,
				lcl_scn_addr, rem_scn_addr) == LRM_OK) {
		vty_out(vty,"Control channel %s added on local scn-if %s%s",argv[0], argv[1], VTY_NEWLINE);
		zlog_info("Conf:Control channel %s added on local scn-if %s",argv[0], argv[1]);
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to create control channel %d%s",cc_id, VTY_NEWLINE);
	zlog_info("Conf:Failed to create control channel %d",cc_id);
	return CMD_WARNING;
}

DEFUN(lrm_cc_delete,
      lrm_cc_delete_cmd,
      "cc delete ccid WORD",
      "Delete a control channel")
{
	int     cc_id;
	lrm_t * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	if (control_channel_del(lrm, cc_id) == LRM_OK) {
		vty_out(vty,
			"Control channel %s deleted successfully%s",
			argv[0], VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to delete control channel %d%s",
		cc_id, VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_cc_delete,
      lrm_cc_del_cmd,
      "cc del ccid WORD",
      "Delete a control channel");

DEFUN(lrm_cc_enable,
      lrm_cc_enable_cmd,
      "cc enable ccid WORD",
      "Enable a control channel")
{
	int     cc_id;
	lrm_t * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	if (control_channel_ena(lrm, cc_id) == LRM_OK) {
		vty_out(vty,"Control channel %s enabled successfully%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Control channel %s enabled successfully",argv[0]);
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to enable up control channel %d%s",cc_id, VTY_NEWLINE);
	zlog_info("Conf:Failed to enable up control channel %d",cc_id);
	return CMD_WARNING;
}

ALIAS(lrm_cc_enable,
      lrm_cc_en_cmd,
      "cc en ccid WORD",
      "Enable a control channel");

DEFUN(lrm_cc_disable,
      lrm_cc_disable_cmd,
      "cc disable ccid WORD",
      "Disable a control channel")
{
	int     cc_id;
	lrm_t * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	if (control_channel_dis(lrm, cc_id) == LRM_OK) {
		vty_out(vty,
			"Control channel %s disabled successfully%s",
			argv[0], VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to disable control channel %d%s",
		cc_id, VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_cc_disable,
      lrm_cc_dis_cmd,
      "cc dis ccid WORD",
      "Disable a control channel");

DEFUN(lrm_cc_up,
      lrm_cc_up_cmd,
      "cc up ccid WORD",
      "Set UP a control channel")
{
	int     cc_id;
	lrm_t * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	if (control_channel_up(lrm, cc_id) == LRM_OK) {
		vty_out(vty,"Control channel %s set up successfully%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Control channel %s set up successfully",argv[0]);
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to set up control channel %d%s",cc_id, VTY_NEWLINE);
	zlog_info("Conf:Failed to set up control channel %d",cc_id);
	return CMD_WARNING;
}

DEFUN(lrm_cc_down,
      lrm_cc_down_cmd,
      "cc down ccid WORD",
      "Set DOWN a control channel")
{
	int     cc_id;
	lrm_t * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	cc_id = strtoul(argv[0], NULL, 0);

	if (control_channel_down(lrm, cc_id) == LRM_OK) {
		vty_out(vty,
			"Control channel %s set down successfully%s",
			argv[0], VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set down control channel %d%s",
		cc_id, VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_cc_show,
      lrm_cc_show_cmd,
      "show lrm cc",
      "type \"show lrm ?\" for available options")
{
	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node;
	cc_t             * cc;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	anylist = lrm->cc_list;
	if (anylist == NULL) {
		return CMD_ERR_NOTHING_TODO;
	}

	vty_out (vty, "CONTROL CHANNELS%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	for (ALL_LIST_ELEMENTS_RO(anylist, node, cc)) {
		ASSERT(cc);
		vty_out (vty,
			 "----------------------%s", VTY_NEWLINE);
		vty_out (vty,
			 "   control channel 0x%x%s",
			 cc->cc_id, VTY_NEWLINE);
		vty_out (vty,
			 "       - remote cc id  : %d%s",
			 cc->rem_cc_id, VTY_NEWLINE);
		vty_out (vty,
			 "       - local  scn if : %s%s",
			 addr_ntoa(cc->lcl_scn_addr), VTY_NEWLINE);
		vty_out (vty,
			 "       - remote scn if : %s%s",
			 addr_ntoa(cc->rem_scn_addr), VTY_NEWLINE);
		vty_out (vty,
			 "       - op state      : %s%s",
			 SHOW_OPSTATE(cc->op_state), VTY_NEWLINE);
		vty_out (vty,
			 "       - adm state     : %s%s",
			 SHOW_ADMSTATE(cc->adm_state), VTY_NEWLINE);
	}

	return CMD_SUCCESS;
}




DEFUN(lrm_datalink_add,
      lrm_datalink_add_cmd,
      "data-link add id (WORD|A.B.C.D) nbr-id (WORD|A.B.C.D)",
      "add | delete | disable | enable | up | down  a data link")
{
	ipv4_t          lcl_ip;
	ipv4_t          rem_ip;
	g2mpls_addr_t   lcl_id;
	g2mpls_addr_t   rem_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));
	memset(&rem_id, 0, sizeof(rem_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for lcl-id%s",
				argv[0], VTY_NEWLINE);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (strchr (argv[1], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[1], &(rem_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for rem-id%s",
				argv[1], VTY_NEWLINE);
			return LRM_ERROR;
		}
		rem_id.type       = IPv4;
		rem_id.preflen    = 32;
		rem_id.value.ipv4 = rem_ip;
	} else {
		/* match "<0-4294967295>". */
		rem_id.type                = UNNUMBERED;
		rem_id.preflen             = 0;
		/* rem_id.value.unnum.node_id = 0;                         */
		/* rem_id.value.unnum.addr    = strtoul(argv[1], NULL, 16);*/
		rem_id.value.unnum         = strtoul(argv[1], NULL, 16);
	}

	if (data_link_add(lrm, lcl_id, rem_id) == LRM_OK) {
		vty_out(vty,"Data link %s added successfully%s",addr_ntoa(lcl_id), VTY_NEWLINE);
		zlog_info("Conf:Data link %s added successfully",addr_ntoa(lcl_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to add data link %s%s",addr_ntoa(lcl_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to add data link %s",addr_ntoa(lcl_id));
	return CMD_WARNING;
}

DEFUN(lrm_datalink_delete,
      lrm_datalink_delete_cmd,
      "data-link delete id (WORD|A.B.C.D)",
      "Delete a data link")
{
	ipv4_t          lcl_ip;
	g2mpls_addr_t   lcl_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for lcl-id%s",
				argv[0], VTY_NEWLINE);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (data_link_del(lrm, lcl_id) == LRM_OK) {
		vty_out(vty,
			"Data link %s deleted successfully%s",
			addr_ntoa(lcl_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to delete data link %s%s",
		addr_ntoa(lcl_id), VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_datalink_delete,
      lrm_datalink_del_cmd,
      "data-link del id (WORD|A.B.C.D)",
      "Delete a data link");

DEFUN(lrm_datalink_enable,
      lrm_datalink_enable_cmd,
      "data-link enable id (WORD|A.B.C.D)",
      "Enable a datalink")
{
	ipv4_t          lcl_ip;
	g2mpls_addr_t   lcl_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,"Invalid IP address (%s) for lcl-id%s",argv[0], VTY_NEWLINE);
			zlog_info("Conf:Invalid IP address (%s) for lcl-id",argv[0]);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (data_link_ena(lrm, lcl_id) == LRM_OK) {
		vty_out(vty,"Data link %s enabled successfully%s",addr_ntoa(lcl_id), VTY_NEWLINE);
		zlog_info("Conf:Data link %s enabled successfully",addr_ntoa(lcl_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to enable data link %s%s",addr_ntoa(lcl_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to enable data link %s",addr_ntoa(lcl_id));
	return CMD_WARNING;
}

ALIAS(lrm_datalink_enable,
      lrm_datalink_en_cmd,
      "data-link en id (WORD|A.B.C.D)",
      "Enable a datalink");

DEFUN(lrm_datalink_up,
      lrm_datalink_up_cmd,
      "data-link up id (WORD|A.B.C.D)",
      "Enable a datalink")
{
	ipv4_t          lcl_ip;
	g2mpls_addr_t   lcl_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for lcl-id%s",
				argv[0], VTY_NEWLINE);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (data_link_up(lrm, lcl_id) == LRM_OK) {
		vty_out(vty,
			"Data link %s set up successfully%s",
			addr_ntoa(lcl_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set up data link %s%s",
		addr_ntoa(lcl_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_datalink_disable,
      lrm_datalink_disable_cmd,
      "data-link disable id (WORD|A.B.C.D)",
      "Disable a datalink")
{
	ipv4_t          lcl_ip;
	g2mpls_addr_t   lcl_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for lcl-id%s",
				argv[0], VTY_NEWLINE);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (data_link_dis(lrm, lcl_id) == LRM_OK) {
		vty_out(vty,
			"Data link %s disabled successfully%s",
			addr_ntoa(lcl_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to disable data link %s%s",
		addr_ntoa(lcl_id), VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_datalink_disable,
      lrm_datalink_dis_cmd,
      "data-link dis id (WORD|A.B.C.D)",
      "Enable a datalink");

DEFUN(lrm_datalink_down,
      lrm_datalink_down_cmd,
      "data-link down id (WORD|A.B.C.D)",
      "Disable a datalink")
{
	ipv4_t          lcl_ip;
	g2mpls_addr_t   lcl_id;
	lrm_t  *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&lcl_id, 0, sizeof(lcl_id));

	if (strchr (argv[0], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[0], &(lcl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for lcl-id%s",
				argv[0], VTY_NEWLINE);
			return LRM_ERROR;
		}
		lcl_id.type       = IPv4;
		lcl_id.preflen    = 32;
		lcl_id.value.ipv4 = lcl_ip;
	} else {
		/* match "<0-4294967295>". */
		lcl_id.type                = UNNUMBERED;
		lcl_id.preflen             = 0;
		/* lcl_id.value.unnum.node_id = lrm->router_id;            */
		/* lcl_id.value.unnum.addr    = strtoul(argv[0], NULL, 16);*/
		lcl_id.value.unnum         = strtoul(argv[0], NULL, 16);
	}

	if (data_link_down(lrm, lcl_id) == LRM_OK) {
		vty_out(vty,
			"Data link %s set down successfully%s",
			addr_ntoa(lcl_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set down data link %s%s",
		addr_ntoa(lcl_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_datalink_show,
      lrm_datalink_show_cmd,
      "show lrm data-link",
      "Show configured data links")
{
	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node;
	datalink_t       * dl;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	anylist = lrm->datalink_list;
	if (anylist == NULL) {
		return CMD_ERR_NOTHING_TODO;
	}

	vty_out (vty, "DATA LINKS%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	for (ALL_LIST_ELEMENTS_RO(anylist, node, dl)) {
		int i;

		ASSERT(dl);
		vty_out (vty,
			 "----------------------%s", VTY_NEWLINE);
		vty_out (vty,
			 "   data-link %s%s",
			 addr_ntoa(dl->lcl_id), VTY_NEWLINE);
		vty_out (vty,
			 "       - remote id            : %s%s",
			 addr_ntoa(dl->rem_id),VTY_NEWLINE);
		vty_out (vty,
			 "       - op state             : %s%s",
			 SHOW_OPSTATE(dl->op_state), VTY_NEWLINE);
		vty_out (vty,
			 "       - adm state            : %s%s",
			 SHOW_ADMSTATE(dl->adm_state), VTY_NEWLINE);
		vty_out (vty,
			 "    ## TE INFO ##%s", VTY_NEWLINE);
		vty_out (vty,
			 "       - swcap                : %s%s",
			 SHOW_SW_CAP(dl->swcap), VTY_NEWLINE);
		vty_out (vty,
			 "       - enctype              : %s%s",
			 SHOW_ENC_TYPE(dl->enctype), VTY_NEWLINE);
		vty_out (vty,
			 "       - max_bw               : %05.03f Mbps%s",
			 BW_HEX2MBPS(dl->max_bw), VTY_NEWLINE);
		vty_out (vty,
			 "       - max_res_bw           : %05.03f Mbps%s",
			 BW_HEX2MBPS(dl->max_res_bw), VTY_NEWLINE);
		for (i = 0; i < MAX_BW_PRIORITIES; i++) {
			vty_out (vty,
				 "       - avail_bw_per_prio[%d] : "
				 "%05.03f Mbps%s",
				 i, BW_HEX2MBPS(dl->avail_bw_per_prio[i]),
				 VTY_NEWLINE);
		}
		for (i = 0; i < MAX_BW_PRIORITIES; i++) {
			vty_out (vty,
				 "       - max_LSP_bw[%d]        : "
				 "%05.03f Mbps%s",
				 i, BW_HEX2MBPS(dl->max_LSP_bw[i]),
				 VTY_NEWLINE);
		}

		vty_out (vty,
			 "       - min_LSP_bw           : %05.03f Mbps%s",
			 BW_HEX2MBPS(dl->min_LSP_bw), VTY_NEWLINE);

		vty_out (vty,
			 "       - Adv. Rsrv events     : %d%s",
			 dl->calendar->count, VTY_NEWLINE);

		calendar_event_t * event;
		if (dl->calendar->count) {
			struct zlistnode * cal_node;
			struct zlistnode * cal_nnode;
			struct tm *        cur_time;
			const char *       time_string_format = "%a, %d %b %Y %H:%M:%S %zGMT";
			const int          time_string_length = 40;
			char               time_string[time_string_length];
			for (ALL_LIST_ELEMENTS(dl->calendar, cal_node, cal_nnode, event)) {
				cur_time = localtime((time_t *) &(event->time_stamp));
				strftime(time_string, time_string_length, time_string_format, cur_time);
				vty_out(vty,
					"                                %s", time_string);
				vty_out(vty, "%s", VTY_NEWLINE);
				vty_out(vty,
					"                                "
					"Available Bandwidth prio 0: %05.03f Mbps %s",
					BW_HEX2MBPS(event->avail_bw[0]), VTY_NEWLINE);
				vty_out(vty,
					"                               "
					"-------------------------------------------%s",
					VTY_NEWLINE);
			}
		}

		if(dl->swcap == SWCAP_LSC) {
			wdm_link_lambdas_bitmap_t bm;
			uint32_t lbl;

			bm = dl->lambdas_bitmap;

			lbl = bm.base_lambda_label;

			vty_out(vty, "       - Lambda bitmap:%s",
				VTY_NEWLINE);
			vty_out(vty, "          - Base Lambda       : 0x%08x",
				lbl);
			if (GMPLS_LABEL_TO_WDM_GRID(lbl) == WDM_GRID_ITUT_DWDM) {
				float    lambda;
				GMPLS_LABEL_TO_DWDM_LAMBDA(lbl, lambda);

				vty_out(vty, " (%.2f Thz)", lambda);
			} else if (GMPLS_LABEL_TO_WDM_GRID(lbl) == WDM_GRID_ITUT_CWDM){
				uint32_t wave;
				GMPLS_LABEL_TO_CWDM_WAVELENGTH(lbl, wave);

				vty_out(vty, " (%d nm)", wave);
			}
			vty_out(vty, "%s", VTY_NEWLINE);

			vty_out(vty, "          - Num of Lambdas    : %d%s",
				bm.num_wavelengths, VTY_NEWLINE);
			vty_out(vty, "          - Lambdas Bitmap    : 0x");
			size_t i;
			for(i = 0; i < bm.bitmap_size; i++) {
				vty_out(vty, "%.8x | ",
					*(bm.bitmap_word + i));
			}
			vty_out(vty, "%s", VTY_NEWLINE);
		}
	}

	return CMD_SUCCESS;
}


DEFUN(lrm_telink_add,
      lrm_telink_add_cmd,
      "(te-link|te-link-inni|te-link-enni| te-link-uni) \
        add ipv4 A.B.C.D pref-len <0-4294967295>\
	nbr-id A.B.C.D rem-telink-id A.B.C.D",
      /*\ls-ret <0-4294967295> ls-dead <0-4294967295>"*/
      "TE LINK (default INNI)\n"
      "TE LINK INNI\n"
      "TE LINK ENNI\n"
      "add new Te LINK\n"
      "IPv4 address type"
      "IPv4 TE LINK local ID\n"
      "Set prefix length"
      "Prefix length value <0-4294967295>\n"
      "Set neighbour ID"
      "IPv4 Te link remote neighbour node ID\n"
      "Set TE link remote ID\n"
      "Te link remote ID")
{
	g2mpls_addr_t    tel_id;
	g2mpls_addr_t    rem_tel_id;
	g2mpls_addr_t    rem_node_addr;
	u_int32_t        rem_node_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));
	memset(&rem_tel_id, 0, sizeof(rem_tel_id));
	memset(&rem_node_addr, 0, sizeof(rem_node_addr));
	memset(&rem_node_id, 0, sizeof(rem_node_id));

	tel_id.type        = IPv4;
	rem_tel_id.type    = IPv4;
	rem_node_addr.type = IPv4;

	if (!inet_aton(argv[0+1], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0+1], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0+1]);
		return LRM_ERROR;
	}

	tel_id.preflen        = strtoul(argv[1+1], NULL, 10);
	rem_tel_id.preflen    = strtoul(argv[1+1], NULL, 10);
	rem_node_addr.preflen = 32;


	if (!inet_aton(argv[2+1], &(rem_node_addr.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for router id%s",argv[2+1], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for router id",argv[2+1]);
		return LRM_ERROR;
	}

	rem_node_id = ntohl(rem_node_addr.value.ipv4.s_addr);

	if (!inet_aton(argv[3+1], &(rem_tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link id%s",argv[3], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link id",argv[3]);
		return LRM_ERROR;
	}

	/*if (te_link_add(lrm, tel_id, rem_tel_id, rem_node_id,
	  (strcmp(argv[0],"te-link-enni") == 0) ? ENNI : INNI) == LRM_OK) {
	  vty_out(vty,
	  "TE link %s added successfully%s",
	  addr_ntoa(tel_id), VTY_NEWLINE);
	  return CMD_SUCCESS;
	  }*/

	if (strcmp(argv[0],"te-link-enni") == 0) {
		if (te_link_add(lrm,
				tel_id,
				rem_tel_id,
				rem_node_id,
				ENNI) == LRM_OK) {
			vty_out(vty,"ENNI TE link %s added successfully%s",addr_ntoa(tel_id), VTY_NEWLINE);
			zlog_info("Conf:ENNI TE link %s added successfully",addr_ntoa(tel_id));
			return CMD_SUCCESS;
		}
	} else if ((strcmp(argv[0],"te-link-inni") == 0) ||
		   (strcmp(argv[0],"te-link") == 0) ) {
		if (te_link_add(lrm,
				tel_id,
				rem_tel_id,
				rem_node_id,
				INNI) == LRM_OK) {
			vty_out(vty,"INNI TE link %s added successfully%s",addr_ntoa(tel_id), VTY_NEWLINE);
			zlog_info("Conf:INNI TE link %s added successfully",addr_ntoa(tel_id));
			return CMD_SUCCESS;
		}
	} else {
		if (te_link_add(lrm,
				tel_id,
				rem_tel_id,
				rem_node_id,
				UNI) == LRM_OK) {
			vty_out(vty,"UNI TE link %s added successfully%s",addr_ntoa(tel_id), VTY_NEWLINE);
			zlog_info("Conf:UNI TE link %s added successfully",addr_ntoa(tel_id));
			return CMD_SUCCESS;
		}
	}


	vty_out(vty,"Failed to add TE link %s%s",addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to add TE link %s",addr_ntoa(tel_id));
	return CMD_WARNING;
}


DEFUN(lrm_telink_delete,
      lrm_telink_delete_cmd,
      "te-link delete ipv4 A.B.C.D pref-len <0-4294967295>",
      "Delete a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_del(lrm, tel_id) == LRM_OK) {
		vty_out(vty,
			"TE link %s deleted successfully%s",
			addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to delete TE link %s%s",
		addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

/*ALIAS(lrm_telink_delete,
  lrm_telink_del_cmd,
  "te-link del ipv4 A.B.C.D pref-len <0-4294967295>",
  "Delete a TE link");
*/
DEFUN(lrm_telink_enable,
      lrm_telink_enable_cmd,
      "te-link enable ipv4 A.B.C.D pref-len <0-4294967295>",
      "Enable a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_ena(lrm, tel_id) == LRM_OK) {
		vty_out(vty,"TE link %s enabled successfully%s",addr_ntoa(tel_id), VTY_NEWLINE);
		zlog_info("Conf:TE link %s enabled successfully",addr_ntoa(tel_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to enable TE link %s%s",addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to enable TE link %s",addr_ntoa(tel_id));
	return CMD_WARNING;
}

ALIAS(lrm_telink_enable,
      lrm_telink_en_cmd,
      "te-link en ipv4 A.B.C.D pref-len <0-4294967295>",
      "Disable a TE link");

DEFUN(lrm_telink_up,
      lrm_telink_up_cmd,
      "te-link up ipv4 A.B.C.D pref-len <0-4294967295>",
      "Enable a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_up(lrm, tel_id) == LRM_OK) {
		vty_out(vty,
			"TE link %s set up successfully%s",
			addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set up TE link %s%s",
		addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_disable,
      lrm_telink_disable_cmd,
      "te-link disable ipv4 A.B.C.D pref-len <0-4294967295>",
      "Disable a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_dis(lrm, tel_id) == LRM_OK) {
		vty_out(vty,
			"TE link %s disabled successfully%s",
			addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to disable TE link %s%s",
		addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

ALIAS(lrm_telink_disable,
      lrm_telink_dis_cmd,
      "te-link dis ipv4 A.B.C.D pref-len <0-4294967295>",
      "Disable a TE link");

DEFUN(lrm_telink_down,
      lrm_telink_down_cmd,
      "te-link down ipv4 A.B.C.D pref-len <0-4294967295>",
      "Disable a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_down(lrm, tel_id) == LRM_OK) {
		vty_out(vty,
			"TE link %s set down successfully%s",
			addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set down TE link %s%s",
		addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_run,
      lrm_telink_run_cmd,
      "te-link run ipv4 A.B.C.D pref-len <0-4294967295>",
      "Enable a TE link")
{
	g2mpls_addr_t    tel_id;
	lrm_t   *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0]);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (te_link_run(lrm, tel_id) == LRM_OK) {
		vty_out(vty,"TE link %s run successfully%s",addr_ntoa(tel_id), VTY_NEWLINE);
		zlog_info("Conf:TE link %s run successfully",addr_ntoa(tel_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to run TE link %s%s",addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to run TE link %s",addr_ntoa(tel_id));
	return CMD_WARNING;
}

DEFUN(lrm_telink_bind,
      lrm_telink_bind_cmd,
      "te-link bind ipv4 A.B.C.D pref-len <0-4294967295> ccid WORD",
      "Bind a te-link to a cc")
{
	g2mpls_addr_t      tel_id;
	u_int32_t          cc_id;
	lrm_t     *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0]);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	cc_id          = strtoul(argv[2], NULL, 0);

	if (te_link_bind_cc(lrm, tel_id, cc_id) == LRM_OK) {
		vty_out(vty,"CC 0x%x bound successfully to  TE link %s%s",cc_id, addr_ntoa(tel_id), VTY_NEWLINE);
		zlog_info("Conf:CC 0x%x bound successfully to  TE link %s",cc_id, addr_ntoa(tel_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to bind CC 0x%x to  TE link %s%s",cc_id, addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to bind CC 0x%x to  TE link %s",cc_id, addr_ntoa(tel_id));
	return CMD_WARNING;
}

DEFUN(lrm_telink_unbind,
      lrm_telink_unbind_cmd,
      "te-link unbind ipv4 A.B.C.D pref-len <0-4294967295> ccid WORD",
      "Unbind a te-link to a cc")
{
	g2mpls_addr_t      tel_id;
	u_int32_t          cc_id;
	lrm_t     *        lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	cc_id          = strtoul(argv[2], NULL, 0);

	if (te_link_unbind_cc(lrm, tel_id, cc_id) == LRM_OK) {
		vty_out(vty,
			"CC 0x%x unbound successfully to  TE link %s%s",
			cc_id, addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to unbind CC 0x%x to  TE link %s%s",
		cc_id, addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_push,
      lrm_telink_push_cmd,
      "te-link push ipv4 A.B.C.D pref-len <0-4294967295> \
        data-link-id (WORD|A.B.C.D)",
      "Push a data link in a te-link")
{
	g2mpls_addr_t   tel_id;
	ipv4_t          dl_ip;
	g2mpls_addr_t   dl_id;
	lrm_t  *        lrm;
	char *          id_dl;
	char *          id_tel;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0]);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	memset(&dl_id, 0, sizeof(dl_id));

	if (strchr (argv[2], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[2], &(dl_ip))) {
			vty_out(vty,"Invalid IP address (%s) for dl-id%s",argv[2], VTY_NEWLINE);
			zlog_info("Conf:Invalid IP address (%s) for dl-id",argv[2]);
			return LRM_ERROR;
		}
		dl_id.type       = IPv4;
		dl_id.preflen    = 32;
		dl_id.value.ipv4 = dl_ip;
	} else {
		/* match "<0-4294967295>". */
		dl_id.type                = UNNUMBERED;
		dl_id.preflen             = 0;
		/* dl_id.value.unnum.node_id = lrm->router_id;            */
		/* dl_id.value.unnum.addr    = strtoul(argv[2], NULL, 16);*/
		dl_id.value.unnum         = strtoul(argv[2], NULL, 16);
	}


	if (te_link_push_dl(lrm, tel_id, dl_id) == LRM_OK) {
		id_dl = strdup(addr_ntoa(dl_id));
		id_tel = strdup(addr_ntoa(tel_id));
		vty_out(vty,"Data link %s pushed successfully into TE link %s%s",id_dl, id_tel, VTY_NEWLINE);
		zlog_info("Conf:ConfData link %s pushed successfully into TE link %s",id_dl, id_tel);
		free(id_dl);
		free(id_tel);
		return CMD_SUCCESS;
	}
	id_dl = strdup(addr_ntoa(dl_id));
	id_tel = strdup(addr_ntoa(tel_id));
	vty_out(vty,"Failed to push Data link %s to  TE link %s%s",id_dl, id_tel, VTY_NEWLINE);
	zlog_info("Conf:Failed to push Data link %s to  TE link %s",id_dl, id_tel);
	free(id_dl);
	free(id_tel);
	return CMD_WARNING;
}

DEFUN(lrm_telink_pop,
      lrm_telink_pop_cmd,
      "te-link pop ipv4 A.B.C.D pref-len <0-4294967295> \
        data-link-id (WORD|A.B.C.D)",
      "Pop a data link in a te-link")
{
	g2mpls_addr_t   tel_id;
	ipv4_t   dl_ip;
	g2mpls_addr_t   dl_id;
	lrm_t  * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	memset(&dl_id, 0, sizeof(dl_id));

	if (strchr (argv[2], '.') != NULL) {
		/* matched "A.B.C.D". */
		if (!inet_aton(argv[2], &(dl_ip))) {
			vty_out(vty,
				"Invalid IP address (%s) for dl-id%s",
				argv[2], VTY_NEWLINE);
			return LRM_ERROR;
		}
		dl_id.type       = IPv4;
		dl_id.preflen    = 32;
		dl_id.value.ipv4 = dl_ip;
	} else {
		/* match "<0-4294967295>". */
		dl_id.type                = UNNUMBERED;
		dl_id.preflen             = 0;
		/* dl_id.value.unnum.node_id = lrm->router_id;            */
		/* dl_id.value.unnum.addr    = strtoul(argv[2], NULL, 16);*/
		dl_id.value.unnum         = strtoul(argv[2], NULL, 16);
	}


	if (te_link_pop_dl(lrm, tel_id, dl_id) == LRM_OK) {
		vty_out(vty,
			"Data link %s popped successfully into TE link %s%s",
			addr_ntoa(dl_id), addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to pop Data link %s to  TE link %s%s",
		addr_ntoa(dl_id), addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_tem,
      lrm_telink_set_tem_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> tem WORD",
      "TE link\n"
      "Set TE link parameters\n"
      "IPv4 address type\n"
      "TE link address\n"
      "Set TE link prefix length\n"
      "TE link prefix length\n"
      "Set TE link metric for a te-link\n"
      "A.B.C.D")
{
	g2mpls_addr_t      tel_id;
	u_int32_t   tem;
	lrm_t     * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0]);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	tem            = strtoul(argv[2], NULL, 10);

	if (te_link_set_te_metric(lrm, tel_id, tem) == LRM_OK) {
		vty_out(vty,"TE metric %d set successfully for TE link %s%s",tem, addr_ntoa(tel_id), VTY_NEWLINE);
		zlog_info("Conf:TE metric %d set successfully for TE link %s",tem, addr_ntoa(tel_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to set TE metric %d for TE link %s%s",tem, addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to set TE metric %d for TE link %s",tem, addr_ntoa(tel_id));
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_color,
      lrm_telink_set_color_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> color WORD",
      "TE link\n"
      "Set TE link parameters\n"
      "IPv4 address type\n"
      "TE link address\n"
      "Set TE link prefix length\n"
      "TE link prefix length\n"
      "Set TE link color mask\n"
      "TE link color mask 32-bit Hexadecimal value (ex. 0xa1b2c3d4)")
{
	g2mpls_addr_t      tel_id;
	u_int32_t   color;
	lrm_t     * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	color          = strtoul(argv[2], NULL, 16);

	if (te_link_set_link_color(lrm, tel_id, color) == LRM_OK) {
		vty_out(vty,
			"TE link color 0x%x set successfully "
			"for TE link %s%s",
			color, addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set TE link color 0x%x for TE link %s%s",
		color, addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_protection,
      lrm_telink_set_protection_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> protection WORD",
      "TE link\n"
      "Set TE link parameters\n"
      "IPv4 address type\n"
      "TE link address\n"
      "Set TE link prefix length\n"
      "TE link prefix length\n"
      "Set TE link protection type\n"
      "TE link protection type 8-bit decimal value")
{
	g2mpls_addr_t      tel_id;
	u_int8_t  protection_type;
	lrm_t     * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	protection_type= strtoul(argv[2], NULL, 0);

	if (te_link_set_link_protection(lrm, tel_id, protection_type) == LRM_OK) {
		vty_out(vty,
			"TE link protection %d set successfully "
			"for TE link %s%s",
			protection_type, addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set TE link protection %d for TE link %s%s",
		protection_type, addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_srlgid,
      lrm_telink_set_srlgid_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> srlg-id WORD",
      "TE link\n"
      "Set TE link parameters\n"
      "IPv4 address type\n"
      "TE link address\n"
      "Set TE link prefix length\n"
      "TE link prefix length\n"
      "Set SRLG id for a TE link\n"
      "SRLG set link 32-bit Hexadecimal value (ex. 0xa1b2c3d4)")
{
	g2mpls_addr_t      tel_id;
	u_int32_t   srlg_id;
	lrm_t     * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	srlg_id        = strtoul(argv[2], NULL, 16);

	if (te_link_add_srlg_id(lrm, tel_id, srlg_id) == LRM_OK) {
		vty_out(vty,
			"SRLG %d set successfully for TE link %s%s",
			srlg_id, addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to set SRLG %d for TE link %s%s",
		srlg_id, addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_unset_srlgid,
      lrm_telink_unset_srlgid_cmd,
      "te-link unset ipv4 A.B.C.D pref-len <0-4294967295> srlg-id WORD",
      "TE link\n"
      "Set TE link parameters\n"
      "IPv4 address type\n"
      "TE link address\n"
      "Set TE link prefix length\n"
      "TE link prefix length\n"
      "Unset SRLG id for a TE link\n"
      "SRLG unset link 32-bit Hexadecimal value (ex. 0xa1b2c3d4)")
{
	g2mpls_addr_t      tel_id;
	u_int32_t   srlg_id;
	lrm_t     * lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));

	tel_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);
	srlg_id        = strtoul(argv[2], NULL, 10);

	if (te_link_rem_srlg_id(lrm, tel_id, srlg_id) == LRM_OK) {
		vty_out(vty,
			"SRLG %d unset successfully for TE link %s%s",
			srlg_id, addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to unset SRLG %d for TE link %s%s",
		srlg_id, addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_tna,
      lrm_telink_set_tna_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> "
      "tna A.B.C.D pref-len <0-4294967295>",
      "Set a TNA for a te-link")
{
	g2mpls_addr_t      tel_id;
	g2mpls_addr_t      tna_id;
	lrm_t *            lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));
	memset(&tna_id, 0, sizeof(tna_id));

	tel_id.type = IPv4;
	tna_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for te-link%s",argv[0], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for te-link",argv[0]);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &(tna_id.value.ipv4))) {
		vty_out(vty,"Invalid IP address (%s) for tna%s",argv[2], VTY_NEWLINE);
		zlog_info("Conf:Invalid IP address (%s) for tna",argv[2]);
		return LRM_ERROR;
	}

	tna_id.preflen = strtoul(argv[3], NULL, 10);

	if (te_link_set_tna(lrm, tel_id, tna_id) == LRM_OK) {
		vty_out(vty,"TNA %s successfully bound to ",addr_ntoa(tna_id));
		vty_out(vty,"TE link %s%s",addr_ntoa(tel_id), VTY_NEWLINE);
		zlog_info("Conf:TNA %s successfully bound to ",addr_ntoa(tna_id));
		zlog_info("TE link %s",addr_ntoa(tel_id));
		return CMD_SUCCESS;
	}

	vty_out(vty,"Failed to bind TNA %s to ",addr_ntoa(tna_id));
	vty_out(vty,"TE link %s%s",addr_ntoa(tel_id), VTY_NEWLINE);
	zlog_info("Conf:Failed to bind TNA %s to ",addr_ntoa(tna_id));
	zlog_info("TE link %s",addr_ntoa(tel_id));
	return CMD_WARNING;
}

DEFUN(lrm_telink_set_remrcid,
      lrm_telink_set_remrcid_cmd,
      "te-link set ipv4 A.B.C.D pref-len <0-4294967295> "
      "rem-rc-id A.B.C.D",
      "Set a remote RC-ID for a te-link")
{
	g2mpls_addr_t      tel_id;
	g2mpls_addr_t      rem_rc_id;
	lrm_t *            lrm;

	lrm = vty->index;
	ASSERT(lrm);

	memset(&tel_id, 0, sizeof(tel_id));
	memset(&rem_rc_id, 0, sizeof(rem_rc_id));

	tel_id.type = IPv4;
	rem_rc_id.type = IPv4;

	if (!inet_aton(argv[0], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[0], VTY_NEWLINE);
		return LRM_ERROR;
	}

	tel_id.preflen = strtoul(argv[1], NULL, 10);

	if (!inet_aton(argv[2], &(rem_rc_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for remote RC ID%s",
			argv[2], VTY_NEWLINE);
		return LRM_ERROR;
	}

	rem_rc_id.preflen = 32;

	if (te_link_set_remrcid(lrm, tel_id, rem_rc_id) == LRM_OK) {
		vty_out(vty,
			"Remote RC ID %s successfully bound to ",
			addr_ntoa(rem_rc_id));
		vty_out(vty,
			"TE link %s%s",
			addr_ntoa(tel_id), VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	vty_out(vty,
		"Failed to bind remote RC ID %s to ",
		addr_ntoa(rem_rc_id));
	vty_out(vty,
		"TE link %s%s",
		addr_ntoa(tel_id), VTY_NEWLINE);
	return CMD_WARNING;
}

DEFUN(lrm_telink_show,
      lrm_telink_show_cmd,
      "show lrm te-link",
      "Show configured TE links")
{
	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node;
	struct zlistnode * srlg_node;
	struct zlistnode * dl_node;
	struct zlistnode * cc_node;
	te_link_t        * tel;
	datalink_t       * dl;
	cc_t             * cc;
	u_int32_t        * srlg_id;
	u_int32_t          count;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	anylist = lrm->telink_list;
	if (anylist == NULL) {
		return CMD_ERR_NOTHING_TODO;
	}

	vty_out (vty, "TE LINKS%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	for (ALL_LIST_ELEMENTS_RO(anylist, node, tel)) {
		int i;

		ASSERT(tel);

		vty_out (vty,
			 "----------------------%s", VTY_NEWLINE);
		vty_out (vty,
			 "   te-link %s%s",
			 tel->tel_name , VTY_NEWLINE);
		vty_out (vty,
			 "       - local id             : %s%s",
			 addr_ntoa(tel->lcl_id),VTY_NEWLINE);
		vty_out (vty,
			 "       - remote id            : %s%s",
			 addr_ntoa(tel->rem_id),VTY_NEWLINE);
		vty_out (vty,
			 "       - remote node id       : %s%s",
			 numb_htoa(tel->rem_node_id),VTY_NEWLINE);
		vty_out (vty,
			 "       - adj type             : %s%s",
			 SHOW_ADJTYPE(tel->adj_type),VTY_NEWLINE);
		if (tel->adj_type == UNI ||
		    tel->adj_type == ENNI) {
			vty_out (vty,
				 "       - TNA addr             : %s%s",
				 addr_ntoa(tel->tna_id), VTY_NEWLINE);
		}
		if (tel->adj_type == ENNI) {
			vty_out (vty,
				 "       - remote RC ID         : %s%s",
				 addr_ntoa(tel->rem_rc_id), VTY_NEWLINE);
		}
		vty_out (vty,
			 "       - op state             : %s%s",
			 SHOW_OPSTATE(tel->op_state), VTY_NEWLINE);
		vty_out (vty,
			 "       - adm state            : %s%s",
			 SHOW_ADMSTATE(tel->adm_state), VTY_NEWLINE);

		vty_out (vty,
			 "    ## TE INFO ##%s", VTY_NEWLINE);
		vty_out (vty,
			 "       - te-metric            : %d%s",
			 tel->te_metric, VTY_NEWLINE);
		vty_out (vty,
			 "       - link-color           : 0x%x%s",
			 tel->link_color, VTY_NEWLINE);

		count = 0;
		for (ALL_LIST_ELEMENTS_RO(tel->SRLG_ids, srlg_node, srlg_id)) {
			vty_out (vty,
				 "       - SRLG[%02d]             : 0x%x%s",
				 count, *srlg_id, VTY_NEWLINE);
			count++;
		}

		vty_out (vty,
			 "       - swcap                : %s%s",
			 SHOW_SW_CAP(tel->swcap), VTY_NEWLINE);
		vty_out (vty,
			 "       - enctype              : %s%s",
			 SHOW_ENC_TYPE(tel->enctype), VTY_NEWLINE);
		vty_out (vty,
		         "       - protection           : %s%s",
			SHOW_GMPLS_PROTTYPE(tel->protection_type), VTY_NEWLINE);
		vty_out (vty,
			 "       - max_bw               : %05.03f Mbps%s",
			 BW_HEX2MBPS(tel->max_bw), VTY_NEWLINE);
		vty_out (vty,
			 "       - max_res_bw           : %05.03f Mbps%s",
			 BW_HEX2MBPS(tel->max_res_bw), VTY_NEWLINE);
		for (i = 0; i < MAX_BW_PRIORITIES; i++) {
			vty_out (vty,
				 "       - avail_bw_per_prio[%d] : "
				 "%05.03f Mbps%s",
				 i, BW_HEX2MBPS(tel->avail_bw_per_prio[i]),
				 VTY_NEWLINE);
		}
		for (i = 0; i < MAX_BW_PRIORITIES; i++) {
			vty_out (vty,
				 "       - max_LSP_bw[%d]        : "
				 "%05.03f Mbps%s",
				 i, BW_HEX2MBPS(tel->max_LSP_bw[i]),
				 VTY_NEWLINE);
		}
		vty_out (vty,
			 "       - min_LSP_bw           : %05.03f Mbps%s",
			 BW_HEX2MBPS(tel->min_LSP_bw), VTY_NEWLINE);

		vty_out (vty,
			 "       - Adv. Rsrv events     : %d%s",
			 tel->calendar->count, VTY_NEWLINE);

		calendar_event_t * event;
		if (tel->calendar->count) {
			struct zlistnode * cal_node;
			struct zlistnode * cal_nnode;
			struct tm *        cur_time;
			const char *       time_string_format = "%a, %d %b %Y %H:%M:%S %zGMT";
			const int          time_string_length = 40;
			char               time_string[time_string_length];
			for (ALL_LIST_ELEMENTS(tel->calendar, cal_node, cal_nnode, event)) {
				cur_time = localtime((time_t *) &(event->time_stamp));
				strftime(time_string, time_string_length, time_string_format, cur_time);
				vty_out(vty,
					"                                %s", time_string);
				vty_out(vty, "%s", VTY_NEWLINE);
				vty_out(vty,
					"                                "
					"Available Bandwidth prio 0: %05.03f Mbps %s",
					BW_HEX2MBPS(event->avail_bw[0]), VTY_NEWLINE);
				vty_out(vty,
					"                               "
					"-------------------------------------------%s",
					VTY_NEWLINE);
			}
		}

		if(tel->swcap == SWCAP_LSC) {
			wdm_link_lambdas_bitmap_t bm;
			uint32_t lbl;

			bm = tel->lambdas_bitmap;

			lbl = bm.base_lambda_label;

			vty_out(vty, "       - Lambda bitmap:%s",
				VTY_NEWLINE);
			vty_out(vty, "          - Base Lambda       : 0x%08x",
				lbl);
			if (GMPLS_LABEL_TO_WDM_GRID(lbl) == WDM_GRID_ITUT_DWDM) {
				float    lambda;
				GMPLS_LABEL_TO_DWDM_LAMBDA(lbl, lambda);

				vty_out(vty, " (%.2f Thz)", lambda);
			} else if (GMPLS_LABEL_TO_WDM_GRID(lbl) == WDM_GRID_ITUT_CWDM){
				uint32_t wave;
				GMPLS_LABEL_TO_CWDM_WAVELENGTH(lbl, wave);

				vty_out(vty, " (%d nm)", wave);
			}
			vty_out(vty, "%s", VTY_NEWLINE);

			vty_out(vty, "          - Num of Lambdas    : %d%s",
				bm.num_wavelengths, VTY_NEWLINE);
			vty_out(vty, "          - Lambdas Bitmap    : 0x");
			size_t i;
			for(i = 0; i < bm.bitmap_size; i++) {
				vty_out(vty, "%.8x | ",
					*(bm.bitmap_word + i));
			}
			vty_out(vty, "%s", VTY_NEWLINE);
		}

		vty_out (vty,
			 "    ## DL INFO ##%s", VTY_NEWLINE);
		for (ALL_LIST_ELEMENTS_RO(tel->dl_list, dl_node, dl)) {
			assert(dl);
			vty_out (vty,
				 "       - local id             : %s%s",
				 addr_ntoa(dl->lcl_id),VTY_NEWLINE);
		}

		vty_out (vty,
			 "    ## CC INFO ##%s", VTY_NEWLINE);

		if (tel->assoc_cc) {
			vty_out (vty,
				 "       - associated cc id     : 0x%x%s",
				 (tel->assoc_cc)->cc_id, VTY_NEWLINE);
		} else {
			vty_out (vty,
				 "       - associated cc id     : NONE%s",
				 VTY_NEWLINE);
		}

		vty_out (vty,
			 "       - #cc in UP state      : %d%s",
			 tel->num_cc_up, VTY_NEWLINE);

		vty_out (vty,
			 "       - cc list%s",
			 VTY_NEWLINE);
		for (ALL_LIST_ELEMENTS_RO(tel->cc_list, cc_node, cc)) {
			assert(cc);
			vty_out (vty,
				 "          - cc id: 0x%x%s",
				 cc->cc_id, VTY_NEWLINE);
		}

	}

	return CMD_SUCCESS;
}


DEFUN(lrm_telink_show_byname,
      lrm_telink_show_byname_cmd,
      "show lrm te-link WORD",
      "Show configured TE links")
{
	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node;
	struct zlistnode * srlg_node;
	struct zlistnode * dl_node;
	struct zlistnode * cc_node;
	te_link_t        * tel;
	datalink_t       * dl;
	cc_t             * cc;
	u_int32_t        * srlg_id;
	u_int32_t          count;

	int tel_found = 0;

	/* Check LRM presence */
	lrm = lrm_lookup ();
	if (lrm == NULL) {
		vty_out (vty,
			 "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}

	anylist = lrm->telink_list;
	if (anylist == NULL) {
		return CMD_ERR_NOTHING_TODO;
	}

	vty_out (vty, "TE LINKS%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}

	for (ALL_LIST_ELEMENTS_RO(anylist, node, tel)) {
		int i;
		if (strncmp(tel->tel_name, argv[0],
			    strlen(tel->tel_name)) == 0) {
			ASSERT(tel);
			tel_found = 1;
			vty_out (vty,
				 "----------------------%s", VTY_NEWLINE);
			vty_out (vty,
				 "   te-link %s%s",
				 tel->tel_name , VTY_NEWLINE);
			vty_out (vty,
				 "       - local id             : %s%s",
				 addr_ntoa(tel->lcl_id),VTY_NEWLINE);
			vty_out (vty,
				 "       - remote id            : %s%s",
				 addr_ntoa(tel->rem_id),VTY_NEWLINE);
			vty_out (vty,
				 "       - remote node id       : %s%s",
				 numb_htoa(tel->rem_node_id),VTY_NEWLINE);
			vty_out (vty,
				 "       - op state             : %s%s",
				 SHOW_OPSTATE(tel->op_state), VTY_NEWLINE);
			vty_out (vty,
				 "       - adm state            : %s%s",
				 SHOW_ADMSTATE(tel->adm_state), VTY_NEWLINE);

			vty_out (vty,
				 "    ## TE INFO ##%s", VTY_NEWLINE);
			vty_out (vty,
				 "       - te-metric            : %d%s",
				 tel->te_metric, VTY_NEWLINE);
			vty_out (vty,
				 "       - link-color           : 0x%x%s",
				 tel->link_color, VTY_NEWLINE);

			count = 0;
			for (ALL_LIST_ELEMENTS_RO(tel->SRLG_ids,
						  srlg_node,
						  srlg_id)) {
				vty_out (vty,
					 "       - SRLG[%02d]             : "
					 "0x%x%s",
					 count, *srlg_id, VTY_NEWLINE);
				count++;
			}

			vty_out (vty,
				 "       - swcap                : %s%s",
				 SHOW_SW_CAP(tel->swcap), VTY_NEWLINE);
			vty_out (vty,
				 "       - enctype              : %s%s",
				 SHOW_ENC_TYPE(tel->enctype), VTY_NEWLINE);

			vty_out (vty,
				 "       - max_bw               : "
				 "%05.03f Mbps%s",
				 BW_HEX2MBPS(tel->max_bw), VTY_NEWLINE);

			vty_out(vty,
				"       - max_res_bw           : "
				"%05.03f Mbps%s",
				BW_HEX2MBPS(tel->max_res_bw), VTY_NEWLINE);
			for (i = 0; i < MAX_BW_PRIORITIES; i++) {
				vty_out (vty,
					 "       - avail_bw_per_prio[%d] : "
					 "%05.03f Mbps%s",
					 i,
					 BW_HEX2MBPS(tel->avail_bw_per_prio[i]),
					 VTY_NEWLINE);
			}
			for (i = 0; i < MAX_BW_PRIORITIES; i++) {
				vty_out (vty,
					 "       - max_LSP_bw[%d]        : "
					 "%05.03f Mbps%s",
					 i, BW_HEX2MBPS(tel->max_LSP_bw[i]),
					 VTY_NEWLINE);
			}
			vty_out (vty,
				 "       - min_LSP_bw           : "
				 "%05.03f Mbps%s",
				 BW_HEX2MBPS(tel->min_LSP_bw), VTY_NEWLINE);


			vty_out (vty,
				 "    ## DL INFO ##%s", VTY_NEWLINE);
			for (ALL_LIST_ELEMENTS_RO(tel->dl_list, dl_node, dl)) {
				assert(dl);
				vty_out (vty,
					 "       - local id             : "
					 "%s%s",
					 addr_ntoa(dl->lcl_id),VTY_NEWLINE);
			}

			vty_out (vty,
				 "    ## CC INFO ##%s", VTY_NEWLINE);

			if (tel->assoc_cc) {
				vty_out (vty,
					 "       - associated cc id     : "
					 "0x%x%s",
					 (tel->assoc_cc)->cc_id, VTY_NEWLINE);
			} else {
				vty_out (vty,
					 "       - associated cc id     : "
					 "NONE%s",
					 VTY_NEWLINE);
			}

			vty_out (vty,
				 "       - #cc in UP state      : %d%s",
				 tel->num_cc_up, VTY_NEWLINE);

			vty_out (vty,
				 "       - cc list%s",
				 VTY_NEWLINE);
			for (ALL_LIST_ELEMENTS_RO(tel->cc_list, cc_node, cc)) {
				assert(cc);
				vty_out (vty,
					 "          - cc id: 0x%x%s",
					 cc->cc_id, VTY_NEWLINE);
			}
		}

	}
	if (!tel_found) {
		vty_out(vty, "te-link %s does not exist%s",
			argv[0], VTY_NEWLINE);
	}
	return CMD_SUCCESS;
}


DEFUN(lrm_adj_add,
      lrm_adj_add_cmd,
      "adj add id <0-4294967295>",
      "add a new adjacency")
{
	u_int32_t adj_id;
	lrm_t *   lrm;

	lrm = vty->index;
	ASSERT(lrm);

	adj_id = strtoul(argv[0], NULL, 10);

	if (adj_add(lrm, adj_id)) {
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Cannot add adjacency %d%s", adj_id, VTY_NEWLINE);
		return CMD_WARNING;
	}

}

DEFUN(lrm_adj_delete,
      lrm_adj_delete_cmd,
      "adj delete id <0-4294967295>",
      "delete an adjacency")
{
	u_int32_t adj_id;
	lrm_t *   lrm;

	lrm = vty->index;
	ASSERT(lrm);

	adj_id = strtoul(argv[0], NULL, 10);

	if (adj_delete(lrm, adj_id)) {
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Cannot delete adjacency %d%s",
			adj_id, VTY_NEWLINE);
		return CMD_WARNING;
	}

}

DEFUN(lrm_adj_tel_add,
      lrm_adj_tel_add_cmd,
      "adj add-te id <0-4294967295> te-link-id ipv4 A.B.C.D",
      "add an te-link to an adjacency")
{
	u_int32_t     adj_id;
	lrm_t *       lrm;
	g2mpls_addr_t tel_id;

	lrm = vty->index;
	ASSERT(lrm);

	adj_id = strtoul(argv[0], NULL, 10);
	if (!inet_aton(argv[1], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[1], VTY_NEWLINE);
		return LRM_ERROR;
	}
	tel_id.type = IPv4;


	if (adj_add_tel(lrm, adj_id, tel_id)) {
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Cannot add te-link %s to adjacency %d%s",
			argv[1], adj_id, VTY_NEWLINE);
		return CMD_WARNING;
	}

}

DEFUN(lrm_adj_tel_rem,
      lrm_adj_tel_rem_cmd,
      "adj rem-te id <0-4294967295> te-link-id ipv4 A.B.C.D",
      "remove a te-link from an adjacency")
{
	u_int32_t     adj_id;
	lrm_t *       lrm;
	g2mpls_addr_t tel_id;

	lrm = vty->index;
	ASSERT(lrm);

	adj_id = strtoul(argv[0], NULL, 10);
	if (!inet_aton(argv[1], &(tel_id.value.ipv4))) {
		vty_out(vty,
			"Invalid IP address (%s) for te-link%s",
			argv[1], VTY_NEWLINE);
		return LRM_ERROR;
	}
	tel_id.type = IPv4;


	if (adj_rem_tel(lrm, adj_id, tel_id)) {
		return CMD_SUCCESS;
	} else {
		vty_out(vty, "Cannot remove te-link %s from adjacency %d%s",
			argv[1], adj_id, VTY_NEWLINE);
		return CMD_WARNING;
	}

}



DEFUN(lrm_adj_show,
      lrm_adj_show_cmd,
      "show lrm adjacency",
      "Show existing adjacencies")
{
	lrm_t* lrm;
	struct zlistnode* adj_node;
	struct zlistnode* tel_node;
	adj_t* adj;
	te_link_t* tel;
	struct zlist* anylist;



	lrm = lrm_lookup();
	if (lrm == NULL) {
		vty_out (vty, "LRM Process not enabled%s",
			 VTY_NEWLINE);
		return CMD_WARNING;
	}
	anylist = lrm->adj_list;
	vty_out (vty, "ADJACENCIES%s", VTY_NEWLINE);

	if (listcount(anylist) == 0) {
		vty_out (vty, "   ***EMPTY***%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}


	for (ALL_LIST_ELEMENTS_RO(anylist, adj_node, adj)) {

		vty_out(vty, " address: %s%s",
			numb_htoa(adj->adj_addr), VTY_NEWLINE);
		if (adj->link_type == INNI) {
			vty_out(vty, " type: INNI ");
		} else if (adj->link_type == ENNI) {
			vty_out(vty, " type: ENNI ");
		} else if (adj->link_type == UNI) {
			vty_out(vty, " type: UNI ");
		}

		if (adj->tel_list->count > 0) {
			vty_out(vty, "contains %d te-link(s):%s",
				adj->tel_list->count, VTY_NEWLINE);
			for (ALL_LIST_ELEMENTS_RO(adj->tel_list,
						  tel_node,
						  tel)) {

				vty_out(vty, "\t%s: lcl id: %s, ",
					tel->tel_name,
					addr_ntoa(tel->lcl_id));
				vty_out(vty, "rem id: %s, ",
					addr_ntoa(tel->rem_id));
				if (adj->link_type == UNI) {
					vty_out(vty, "TNA id: %s",
						addr_ntoa(tel->tna_id));
				} else if (adj->link_type == ENNI) {
					vty_out (vty, "remote RC ID : %s",
						 addr_ntoa(tel->rem_rc_id));
				}

				vty_out(vty, "%s", VTY_NEWLINE);
			}
		} else {
			vty_out(vty, "contains 0 te-links%s", VTY_NEWLINE);
		}
	}
	return CMD_SUCCESS;
}



static int lrm_config_write (struct vty *vty)
{

	lrm_t            * lrm;
	struct zlist     * anylist;
	struct zlistnode * node_dl;
	struct zlistnode * node_cc;
	struct zlistnode * node_srlg;
	struct zlistnode * node;


	ctrl_intf_t *scnif;
	cc_t *cc;
	te_link_t *tel;
	datalink_t *dl;

	lrm = lrm_lookup();

	vty_out (vty, "lrm%s", VTY_NEWLINE);
	vty_out (vty, "!%s",   VTY_NEWLINE);

	vty_out (vty, "router-id %s%s",
		 numb_htoa(lrm->router_id), VTY_NEWLINE);

	anylist = lrm->scn_if_list;
	for (ALL_LIST_ELEMENTS_RO(anylist, node, scnif)) {

		vty_out (vty, "scn-if add ip %s%s",
			 inet_ntoa(scnif->addr.value.ipv4), VTY_NEWLINE);
		if (scnif->adm_state == ENABLED)
			vty_out (vty, "scn-if en ip %s%s",
				 inet_ntoa(scnif->addr.value.ipv4),
				 VTY_NEWLINE);
	}
	vty_out (vty, "!%s", VTY_NEWLINE);

	anylist = lrm->cc_list;
	for (ALL_LIST_ELEMENTS_RO(anylist, node, cc)) {
		vty_out(vty, "cc add ccid 0x%x scn-ip %s scn-nbr",
			cc->cc_id, inet_ntoa(cc->lcl_scn_addr.value.ipv4));
		vty_out(vty, " %s%s",
			inet_ntoa(cc->rem_scn_addr.value.ipv4), VTY_NEWLINE);
		if (cc->adm_state == ENABLED)
			vty_out(vty, "cc en ccid 0x%x%s",
				cc->cc_id, VTY_NEWLINE);
		if (cc->op_state == UP)
			vty_out(vty, "cc up ccid 0x%x%s",
				cc->cc_id, VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
	}

	vty_out (vty, "!%s", VTY_NEWLINE);

	anylist = lrm->datalink_list;
	for (ALL_LIST_ELEMENTS_RO(anylist, node, dl)) {

		vty_out(vty, "data-link add id");
		switch(dl->lcl_id.type) {
			case IPv4:
				vty_out(vty, " %s",
					inet_ntoa(dl->lcl_id.value.ipv4));
				break;
			case UNNUMBERED:
				/* vty_out(vty, " 0x%x",
				   dl->lcl_id.value.unnum.addr);*/
				vty_out(vty, " 0x%x",
					dl->lcl_id.value.unnum);
				break;
			default:
				break;
		}
		vty_out(vty, " nbr-id");
		switch(dl->rem_id.type) {
			case IPv4:
				vty_out(vty, " %s%s",
					inet_ntoa(dl->rem_id.value.ipv4),
					VTY_NEWLINE);
				break;
			case UNNUMBERED:
				/*vty_out(vty, " 0x%x%s",
					dl->rem_id.value.unnum.addr,
					VTY_NEWLINE);*/
				vty_out(vty, " 0x%x%s",
					dl->rem_id.value.unnum,
					VTY_NEWLINE);
				break;
			default:
				break;
		}


		if (dl->adm_state == ENABLED) {
			vty_out(vty, "data-link en id");
			switch(dl->rem_id.type) {
				case IPv4:
					vty_out(vty, " %s%s",
						inet_ntoa(dl->lcl_id.value.ipv4),
						VTY_NEWLINE);
					break;
				case UNNUMBERED:
					/*vty_out(vty, " 0x%x%s",
						dl->lcl_id.value.unnum.addr,
						VTY_NEWLINE);*/
					vty_out(vty, " 0x%x%s",
						dl->lcl_id.value.unnum,
						VTY_NEWLINE);
					break;
				default:
					break;
			}
			vty_out (vty, "%s", VTY_NEWLINE);
		}
	}

	vty_out (vty, "!%s", VTY_NEWLINE);

	anylist = lrm->telink_list;
	for (ALL_LIST_ELEMENTS_RO(anylist, node, tel)) {
		vty_out(vty, "te-link add ipv4 %s pref-len %d",
			inet_ntoa(tel->lcl_id.value.ipv4),
			tel->lcl_id.preflen);
		vty_out(vty, " nbr-id %s",
			numb_htoa(tel->rem_node_id));
		vty_out(vty, " rem-telink-id %s%s",
			inet_ntoa(tel->rem_id.value.ipv4), VTY_NEWLINE);


		if (tel->assoc_cc != NULL) {
			for (ALL_LIST_ELEMENTS_RO(tel->cc_list, node_cc, cc)) {

				vty_out(vty, "te-link bind ipv4 %s "
					"pref-len %d ccid 0x%x%s",
					inet_ntoa(tel->lcl_id.value.ipv4),
					tel->lcl_id.preflen,
					cc->cc_id,
					VTY_NEWLINE);
			}
		}

		if (tel->adm_state == ENABLED)
			vty_out(vty, "te-link en ipv4 %s pref-len %d%s",
				inet_ntoa(tel->lcl_id.value.ipv4),
				tel->lcl_id.preflen,
				VTY_NEWLINE);

		if (tel->dl_list != NULL) {
			if (tel->dl_list->count > 0) {
				for (ALL_LIST_ELEMENTS_RO(tel->dl_list,
							  node_dl,
							  dl)) {
					vty_out(vty, "te-link push ipv4 %s "
						"pref-len %d data-link-id",
						inet_ntoa(tel->lcl_id.value.ipv4),
						tel->lcl_id.preflen);
					switch(dl->lcl_id.type) {
						case IPv4:
							vty_out(vty, " %s%s",
								inet_ntoa(dl->lcl_id.value.ipv4),
								VTY_NEWLINE);
							break;
						case UNNUMBERED:
							/*vty_out(vty, " 0x%x%s",
								dl->lcl_id.value.unnum.addr,
								VTY_NEWLINE);*/
							vty_out(vty, " 0x%x%s",
								dl->lcl_id.value.unnum,
								VTY_NEWLINE);
							break;
						default:
							break;
					}
				}
			}
		}
		if (tel->te_metric != 0)
			vty_out(vty, "te-link set ipv4 %s pref-len %d "
				"tem %d%s",
				inet_ntoa(tel->lcl_id.value.ipv4),
				tel->lcl_id.preflen,
				tel->te_metric,
				VTY_NEWLINE);
		if (tel->link_color != 0)
			vty_out(vty, "te-link set ipv4 %s pref-len %d "
				"color %d%s",
				inet_ntoa(tel->lcl_id.value.ipv4),
				tel->lcl_id.preflen,
				tel->link_color,
				VTY_NEWLINE);

		if (tel->SRLG_ids != NULL) {
			int *id;

			if (tel->SRLG_ids->count > 0) {
				for (ALL_LIST_ELEMENTS_RO(tel->SRLG_ids,
							  node_srlg,
							  id)) {
					vty_out(vty, "te-link set ipv4 %s "
						"pref-len %d  srlg-id %d%s",
						inet_ntoa(tel->lcl_id.value.ipv4),
						tel->lcl_id.preflen,
						*id,
						VTY_NEWLINE);
				}
			}
		}
		if (tel->op_state == UP)
			vty_out(vty, "te-link run ipv4 %s pref-len %d%s",
				inet_ntoa(tel->lcl_id.value.ipv4),
				tel->lcl_id.preflen,
				VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
	}
	return 1;
}

DEFUN (rc_interface_add,
       rc_interface_add_cmd,
       "router rc interface A.B.C.D ADDR_LEN",
       "Enable a routing process\n"
       "Routing Controller configuration\n"
       "Routing controler interface configuration\n"
       "Router Controller if IP address format\n"
       "Router Controller if address length\n"
	)
{
  struct in_addr rc_id;

  int ret;

  ret = inet_aton (argv[0], &rc_id);
  if (!ret)
  {
    if (vty)
      vty_out (vty, "Please specify Router ID by A.B.C.D%s", VTY_NEWLINE);
    else
      zlog_debug("Please specify Router ID by A.B.C.D");
    return CMD_WARNING;
  }

  int addr_len = strtoul (argv[1], NULL, 10);

  struct interface * ifp     = if_get_by_name_len("Level1", INTERFACE_NAMSIZ);
  ifp->ifindex               = 10000; //++lrm->telink_count; TODO use LRMD index counter
  ifp->flags                 = IFF_UP|IFF_RUNNING;
  ifp->mtu                   = 65536;
  ifp->metric                = 1;
  ifp->bandwidth             = 10000;
  ifp->status                = ZEBRA_INTERFACE_ACTIVE;

  struct connected    * ifc  = connected_new();
  struct prefix_ipv4  * p    = prefix_ipv4_new ();

  p->family                  = AF_INET;
  p->prefix                  = rc_id;
  p->prefixlen               = addr_len;

  ifc->address               = (struct prefix *) p;
  ifc->ifp                   = ifp;

  listnode_add (ifp->connected, ifc);

  /*now we send the te link TE params to ospf*/
  ifp->adj_type              = ENNI;
  ifp->ospf_instance         = ENNI;
  ifp->te_TNA_address_type   = IPv4;
  ifp->te_TNA_prefix_length  = 0;
  ifp->te_TNA_address        = (u_int32_t *) malloc(4);
  memset(ifp->te_TNA_address, 0, 4);
  lrm_interface_add_notify(ifp);
  return CMD_SUCCESS;
}

void lrm_vty_init (void)
{
	/* Install lrm  top node. */
	install_node (&lrm_node, lrm_config_write);

	install_element(CONFIG_NODE, &lrm_cmd);
	install_element(CONFIG_NODE, &no_lrm_cmd);

	install_default(LRM_NODE);

	//show command
	install_element(VIEW_NODE,   &lrm_show_cmd);
	install_element(ENABLE_NODE, &lrm_show_cmd);
	install_element(LRM_NODE,    &lrm_show_cmd);

	// Router ID commands
	install_element(LRM_NODE,    &lrm_router_id_cmd);

	//control interface commands
	install_element(LRM_NODE,    &rc_interface_add_cmd);
	install_element(LRM_NODE,    &lrm_scnif_add_cmd);
	install_element(LRM_NODE,    &lrm_scnif_delete_cmd);
	install_element(LRM_NODE,    &lrm_scnif_enable_cmd);
	install_element(LRM_NODE,    &lrm_scnif_disable_cmd);
	install_element(LRM_NODE,    &lrm_scnif_del_cmd);
	install_element(LRM_NODE,    &lrm_scnif_en_cmd);
	install_element(LRM_NODE,    &lrm_scnif_dis_cmd);
	//show command
	install_element(VIEW_NODE,   &lrm_scnif_show_cmd);
	install_element(ENABLE_NODE, &lrm_scnif_show_cmd);
	install_element(LRM_NODE,    &lrm_scnif_show_cmd);

	//control channel commands
	install_element(LRM_NODE,    &lrm_cc_add_cmd);
	install_element(LRM_NODE,    &lrm_cc_delete_cmd);
	install_element(LRM_NODE,    &lrm_cc_enable_cmd);
	install_element(LRM_NODE,    &lrm_cc_disable_cmd);
	install_element(LRM_NODE,    &lrm_cc_del_cmd);
	install_element(LRM_NODE,    &lrm_cc_en_cmd);
	install_element(LRM_NODE,    &lrm_cc_dis_cmd);
	install_element(LRM_NODE,    &lrm_cc_up_cmd);
	install_element(LRM_NODE,    &lrm_cc_down_cmd);
	//show commands
	install_element(VIEW_NODE,   &lrm_cc_show_cmd);
	install_element(LRM_NODE,    &lrm_cc_show_cmd);
	install_element(ENABLE_NODE, &lrm_cc_show_cmd);

	//data link commands
	install_element(LRM_NODE,    &lrm_datalink_add_cmd);
	install_element(LRM_NODE,    &lrm_datalink_delete_cmd);
	install_element(LRM_NODE,    &lrm_datalink_enable_cmd);
	install_element(LRM_NODE,    &lrm_datalink_up_cmd);
	install_element(LRM_NODE,    &lrm_datalink_disable_cmd);
	install_element(LRM_NODE,    &lrm_datalink_down_cmd);
	install_element(LRM_NODE,    &lrm_datalink_del_cmd);
	install_element(LRM_NODE,    &lrm_datalink_en_cmd);
	install_element(LRM_NODE,    &lrm_datalink_dis_cmd);
	//show command
	install_element(VIEW_NODE,   &lrm_datalink_show_cmd);
	install_element(ENABLE_NODE, &lrm_datalink_show_cmd);
	install_element(LRM_NODE,    &lrm_datalink_show_cmd);

	//te-link commands
	install_element(LRM_NODE,    &lrm_telink_add_cmd);
	install_element(LRM_NODE,    &lrm_telink_delete_cmd);
	install_element(LRM_NODE,    &lrm_telink_enable_cmd);
	install_element(LRM_NODE,    &lrm_telink_up_cmd);
	install_element(LRM_NODE,    &lrm_telink_disable_cmd);
	install_element(LRM_NODE,    &lrm_telink_down_cmd);
	//	install_element(LRM_NODE,    &lrm_telink_del_cmd);
	install_element(LRM_NODE,    &lrm_telink_en_cmd);
	install_element(LRM_NODE,    &lrm_telink_dis_cmd);
	install_element(LRM_NODE,    &lrm_telink_run_cmd);
	install_element(LRM_NODE,    &lrm_telink_bind_cmd);
	install_element(LRM_NODE,    &lrm_telink_unbind_cmd);
	install_element(LRM_NODE,    &lrm_telink_push_cmd);
	install_element(LRM_NODE,    &lrm_telink_pop_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_tem_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_color_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_protection_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_srlgid_cmd);
	install_element(LRM_NODE,    &lrm_telink_unset_srlgid_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_tna_cmd);
	install_element(LRM_NODE,    &lrm_telink_set_remrcid_cmd);
	//adjacency commands
	//install_element(LRM_NODE, &lrm_adj_add_cmd);
	//install_element(LRM_NODE, &lrm_adj_delete_cmd);
	//install_element(LRM_NODE, &lrm_adj_tel_add_cmd);
	//install_element(LRM_NODE, &lrm_adj_tel_rem_cmd);
	//show commands
	install_element(VIEW_NODE,   &lrm_adj_show_cmd);
	install_element(ENABLE_NODE, &lrm_adj_show_cmd);
	install_element(LRM_NODE,    &lrm_adj_show_cmd);



	//show command
	install_element(VIEW_NODE,   &lrm_telink_show_cmd);
	install_element(ENABLE_NODE, &lrm_telink_show_cmd);
	install_element(LRM_NODE,    &lrm_telink_show_cmd);

	install_element(VIEW_NODE,   &lrm_telink_show_byname_cmd);
	install_element(ENABLE_NODE, &lrm_telink_show_byname_cmd);
	install_element(LRM_NODE,    &lrm_telink_show_byname_cmd);

}
