/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *  Copyright (C) 2006, 2007, 2008, 2009 PSNC
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 *  Adam Kaliszan         (PSNC)             <kaliszan_at_man.poznan.pl>
 */



#include <zebra.h>

#include "prefix.h"
#include "command.h"
#include "if.h"
#include "thread.h"
#include "stream.h"
#include "memory.h"
#include "table.h"
//#include "rib.h"
#include "network.h"
#include "sockunion.h"
#include "log.h"
#include "zclient.h"
#include "privs.h"
#include "network.h"
#include "buffer.h"

//#include "zebra/zserv.h"
//#include "zebra/router-id.h"
//#include "zebra/redistribute.h"
//#include "zebra/debug.h"
//#include "zebra/ipforward.h"
#include "lrmd.h"
#include "lrm_core.h"
#include "zserv.h"

/* Event list of zebra. */
enum event { ZEBRA_SERV, ZEBRA_READ, ZEBRA_WRITE };

extern struct zebra_t zebrad;

static void zebra_event (enum event event, int sock, struct zserv *client);

//extern struct zebra_privs_t zserv_privs;

static void zebra_client_close (struct zserv *client);

static int
zserv_delayed_close(struct thread *thread)
{
  struct zserv *client = THREAD_ARG(thread);

  client->t_suicide = NULL;
  zebra_client_close(client);
  return 0;
}

static int
zserv_flush_data(struct thread *thread)
{
  struct zserv *client = THREAD_ARG(thread);

  client->t_write = NULL;
  if (client->t_suicide)
    {
      zebra_client_close(client);
      return -1;
    }
  switch (buffer_flush_available(client->wb, client->sock))
    {
    case BUFFER_ERROR:
      zlog_warn("%s: buffer_flush_available failed on zserv client fd %d, "
		"closing", __func__, client->sock);
      zebra_client_close(client);
      break;
    case BUFFER_PENDING:
      client->t_write = thread_add_write(/*zebrad.master*/master, zserv_flush_data,
					 client, client->sock);
      break;
    case BUFFER_EMPTY:
      break;
    }
  return 0;
}

static int
zebra_server_send_message(struct zserv *client)
{
  if (client->t_suicide)
    return -1;
  switch (buffer_write(client->wb, client->sock, STREAM_DATA(client->obuf),
		       stream_get_endp(client->obuf)))
    {
    case BUFFER_ERROR:
      zlog_warn("%s: buffer_write failed to zserv client fd %d, closing",
		 __func__, client->sock);
      /* Schedule a delayed close since many of the functions that call this
         one do not check the return code.  They do not allow for the
	 possibility that an I/O error may have caused the client to be
	 deleted. */
      client->t_suicide = thread_add_event(/*zebrad.master*/master, zserv_delayed_close,
					   client, 0);
      return -1;
    case BUFFER_EMPTY:
      THREAD_OFF(client->t_write);
      break;
    case BUFFER_PENDING:
      THREAD_WRITE_ON(/*zebrad.master*/ master, client->t_write,
		      zserv_flush_data, client, client->sock);
      break;
    }
  return 0;
}

static void
zserv_create_header (struct stream *s, uint16_t cmd)
{
  /* length placeholder, caller can update */
  stream_putw (s, ZEBRA_HEADER_SIZE);
  stream_putc (s, ZEBRA_HEADER_MARKER);
  stream_putc (s, ZSERV_VERSION);
  stream_putw (s, cmd);
}

/* Interface is added. Send ZEBRA_INTERFACE_ADD to client. */
/*
 * This function is called in the following situations:
 * - in response to a 3-byte ZEBRA_INTERFACE_ADD request
 *   from the client.
 * - at startup, when zebra figures out the available interfaces
 * - when an interface is added (where support for
 *   RTM_IFANNOUNCE or AF_NETLINK sockets is available), or when
 *   an interface is marked IFF_UP (i.e., an RTM_IFINFO message is
 *   received)
 */
int
zsend_interface_add (struct zserv *client, struct interface *ifp)
{
  struct stream *s;
  int i = 0;
  struct zlistnode *node, *nnode;
  calendar_event_t *event;
  int *data = NULL;

  node = NULL;
  nnode = NULL;
  /* Check this client need interface information. */
  if (! client->ifinfo)
    return 0;
  s = client->obuf;
  stream_reset (s);
  /* Message type. */
  zserv_create_header (s, ZEBRA_INTERFACE_ADD);

  /* Interface information. */
  stream_put (s, ifp->name, INTERFACE_NAMSIZ);
  stream_putl (s, ifp->ifindex);
  stream_putc (s, ifp->status);
  stream_putq (s, ifp->flags);

  stream_putl (s, ifp->metric);
  stream_putl (s, ifp->mtu);
  stream_putl (s, ifp->mtu6);
  stream_putl (s, ifp->bandwidth);
#ifdef HAVE_SOCKADDR_DL
  stream_put (s, &ifp->sdl, sizeof (ifp->sdl));
#else
  stream_putl (s, ifp->hw_addr_len);
  if (ifp->hw_addr_len)
    stream_put (s, ifp->hw_addr, ifp->hw_addr_len);
#endif /* HAVE_SOCKADDR_DL */

#ifdef GMPLS
  /*Send TE params along with the interface*/
  zlog_debug("zsend_interface_add: interface adjacency %d, ospf instance %d", ifp->adj_type, ifp->ospf_instance);
  stream_putw (s, ifp->adj_type);
  stream_putw (s, ifp->ospf_instance);
  stream_putl(s, ifp->te_local_id);
  stream_putl(s, ifp->te_remote_id);
  stream_putl(s, ifp->te_remote_node_id);
  stream_putl(s, ifp->te_metric);
  stream_putl(s, ifp->te_link_color);
  stream_putl(s, ifp->te_swcap);
  stream_putl(s, ifp->te_swcap_options);
  stream_putl(s, ifp->te_enctype);
  stream_putl(s, ifp->te_max_bw);
  stream_putl(s, ifp->te_max_res_bw);

  for (i=0; i< MAX_BW_PRIORITIES; i++)
    stream_putl(s, ifp->te_avail_bw_per_prio[i]);
  for (i=0; i< MAX_BW_PRIORITIES; i++)
    stream_putl(s, ifp->te_max_LSP_bw[i]);
  stream_putl(s, ifp->te_min_LSP_bw);
  stream_putl(s, ifp->te_protection_type);

  zlog_debug("zsend_interface_add 1");

  stream_putl(s, (uint32_t) ifp->adv_rsrv_calendar->count);
  for(ALL_LIST_ELEMENTS(ifp->adv_rsrv_calendar, node, nnode, event))
  {
    stream_putl(s, event->time_stamp);
    for (i=0; i< MAX_BW_PRIORITIES; i++)
      stream_putl(s, event->avail_bw[i]);
  }

  if (ifp->te_swcap == SWCAP_LSC) {
    stream_putl(s, ifp->lambdas_bitmap.base_lambda_label);
    stream_putw(s, ifp->lambdas_bitmap.num_wavelengths);
    stream_putw(s, ifp->lambdas_bitmap.bitmap_size);
    /*check for ifp->lambdas_bitmap_word pointer is inside stream_put function*/
    stream_put (s, ifp->lambdas_bitmap.bitmap_word,
		4 * (ifp->lambdas_bitmap.bitmap_size));
  }

  zlog_debug("zsend_interface_add 2");
  stream_putl(s, (uint32_t) ifp->te_SRLG_ids->count);
  for(ALL_LIST_ELEMENTS(ifp->te_SRLG_ids, node, nnode, data))
  {
    stream_putl(s, *data);
  }
  zlog_debug("zsend_interface_add 3");

  switch (ifp->te_TNA_address_type)
  {
    case IPv4:
      stream_putl(s, 32);
      stream_putl(s, ifp->te_TNA_prefix_length);
      stream_putl(s, *(ifp->te_TNA_address));
      break;
    case IPv6:
      stream_putl(s, 128);
      stream_putl(s, ifp->te_TNA_prefix_length);
      stream_putl(s, ifp->te_TNA_address[0]);
      stream_putl(s, ifp->te_TNA_address[1]);
      stream_putl(s, ifp->te_TNA_address[2]);
      stream_putl(s, ifp->te_TNA_address[3]);
      break;
    case NSAP:
      stream_putl(s, 160);
      stream_putl(s, ifp->te_TNA_prefix_length);
      stream_putl(s, ifp->te_TNA_address[0]);
      stream_putl(s, ifp->te_TNA_address[1]);
      stream_putl(s, ifp->te_TNA_address[2]);
      stream_putl(s, ifp->te_TNA_address[3]);
      stream_putl(s, ifp->te_TNA_address[4]);
      break;
    case UNNUMBERED:
	    break;
  }
  zlog_debug("zsend_interface_add 4");

  stream_putl(s, ifp->rem_rc_id);

#endif /* GMPLS */

  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  zlog_debug("zsend_interface_add OK");
    return zebra_server_send_message(client);
}


static int
zsend_tel_add (struct zserv *client, struct interface *ifp)
{
  struct stream *s;

  /* Check this client need interface information. */
  /*if (! client->ifinfo)
    return 0;*/

  s = client->obuf;
  stream_reset (s);

  /* Message type. */

  /* Interface information. */
  /*stream_put (s, ifp->name, INTERFACE_NAMSIZ);
  stream_putl (s, ifp->ifindex);
  stream_putc (s, ifp->status);
  stream_putq (s, ifp->flags);
  stream_putl (s, ifp->metric);
  stream_putl (s, ifp->mtu);
  stream_putl (s, ifp->mtu6);
  stream_putl (s, ifp->bandwidth);*/
#ifdef HAVE_SOCKADDR_DL
  /*stream_put (s, &ifp->sdl, sizeof (ifp->sdl));*/
#else
  /*stream_putl (s, ifp->hw_addr_len);
  if (ifp->hw_addr_len)
    stream_put (s, ifp->hw_addr, ifp->hw_addr_len);*/
#endif /* HAVE_SOCKADDR_DL */

  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}


/* Interface deletion from zebra daemon. */
int
zsend_interface_delete (struct zserv *client, struct interface *ifp)
{
  struct stream *s;

  /* Check this client need interface information. */
  if (! client->ifinfo)
    return 0;

  s = client->obuf;
  stream_reset (s);

  zserv_create_header (s, ZEBRA_INTERFACE_DELETE);

  /* Interface information. */
  stream_put (s, ifp->name, INTERFACE_NAMSIZ);
  stream_putl (s, ifp->ifindex);
  stream_putc (s, ifp->status);
  stream_putq (s, ifp->flags);
  stream_putl (s, ifp->metric);
  stream_putl (s, ifp->mtu);
  stream_putl (s, ifp->mtu6);
  stream_putl (s, ifp->bandwidth);

  /* Write packet length. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message (client);
}

/* Interface address is added/deleted. Send ZEBRA_INTERFACE_ADDRESS_ADD or
 * ZEBRA_INTERFACE_ADDRESS_DELETE to the client.
 *
 * A ZEBRA_INTERFACE_ADDRESS_ADD is sent in the following situations:
 * - in response to a 3-byte ZEBRA_INTERFACE_ADD request
 *   from the client, after the ZEBRA_INTERFACE_ADD has been
 *   sent from zebra to the client
 * - redistribute new address info to all clients in the following situations
 *    - at startup, when zebra figures out the available interfaces
 *    - when an interface is added (where support for
 *      RTM_IFANNOUNCE or AF_NETLINK sockets is available), or when
 *      an interface is marked IFF_UP (i.e., an RTM_IFINFO message is
 *      received)
 *    - for the vty commands "ip address A.B.C.D/M [<secondary>|<label LINE>]"
 *      and "no bandwidth <1-10000000>", "ipv6 address X:X::X:X/M"
 *    - when an RTM_NEWADDR message is received from the kernel,
 *
 * The call tree that triggers ZEBRA_INTERFACE_ADDRESS_DELETE:
 *
 *                   zsend_interface_address(DELETE)
 *                           ^
 *                           |
 *          zebra_interface_address_delete_update
 *             ^                        ^      ^
 *             |                        |      if_delete_update
 *             |                        |
 *         ip_address_uninstall        connected_delete_ipv4
 *         [ipv6_addresss_uninstall]   [connected_delete_ipv6]
 *             ^                        ^
 *             |                        |
 *             |                  RTM_NEWADDR on routing/netlink socket
 *             |
 *         vty commands:
 *     "no ip address A.B.C.D/M [label LINE]"
 *     "no ip address A.B.C.D/M secondary"
 *     ["no ipv6 address X:X::X:X/M"]
 *
 */
int
zsend_interface_address (int cmd, struct zserv *client,
                         struct interface *ifp, struct connected *ifc)
{
  int blen;
  struct stream *s;
  struct prefix *p;

  /* Check this client need interface information. */
  if (! client->ifinfo)
    return 0;

  s = client->obuf;
  stream_reset (s);

  zserv_create_header (s, cmd);
  stream_putl (s, ifp->ifindex);

  /* Interface address flag. */
  stream_putc (s, ifc->flags);

  /* Prefix information. */
  p = ifc->address;
  stream_putc (s, p->family);
  blen = prefix_blen (p);
  stream_put (s, &p->u.prefix, blen);

  /*
   * XXX gnu version does not send prefixlen for ZEBRA_INTERFACE_ADDRESS_DELETE
   * but zebra_interface_address_delete_read() in the gnu version
   * expects to find it
   */
  stream_putc (s, p->prefixlen);

  /* Destination. */
  p = ifc->destination;
  if (p)
    stream_put (s, &p->u.prefix, blen);
  else
    stream_put (s, NULL, blen);

  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

/*
 * The cmd passed to zsend_interface_update  may be ZEBRA_INTERFACE_UP or
 * ZEBRA_INTERFACE_DOWN.
 *
 * The ZEBRA_INTERFACE_UP message is sent from the zebra server to
 * the clients in one of 2 situations:
 *   - an if_up is detected e.g., as a result of an RTM_IFINFO message
 *   - a vty command modifying the bandwidth of an interface is received.
 * The ZEBRA_INTERFACE_DOWN message is sent when an if_down is detected.
 */
int
zsend_interface_update (int cmd, struct zserv *client, struct interface *ifp)
{
  struct stream *s;

  /* Check this client need interface information. */
  if (! client->ifinfo)
    return 0;

  s = client->obuf;
  stream_reset (s);

  zserv_create_header (s, cmd);

  /* Interface information. */
  stream_put (s, ifp->name, INTERFACE_NAMSIZ);
  stream_putl (s, ifp->ifindex);
  stream_putc (s, ifp->status);
  stream_putq (s, ifp->flags);
  stream_putl (s, ifp->metric);
  stream_putl (s, ifp->mtu);
  stream_putl (s, ifp->mtu6);
  stream_putl (s, ifp->bandwidth);

  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

/*
 * The zebra server sends the clients  a ZEBRA_IPV4_ROUTE_ADD or a
 * ZEBRA_IPV6_ROUTE_ADD via zsend_route_multipath in the following
 * situations:
 * - when the client starts up, and requests default information
 *   by sending a ZEBRA_REDISTRIBUTE_DEFAULT_ADD to the zebra server, in the
 * - case of rip, ripngd, ospfd and ospf6d, when the client sends a
 *   ZEBRA_REDISTRIBUTE_ADD as a result of the "redistribute" vty cmd,
 * - when the zebra server redistributes routes after it updates its rib
 *
 * The zebra server sends clients a ZEBRA_IPV4_ROUTE_DELETE or a
 * ZEBRA_IPV6_ROUTE_DELETE via zsend_route_multipath when:
 * - a "ip route"  or "ipv6 route" vty command is issued, a prefix is
 * - deleted from zebra's rib, and this info
 *   has to be redistributed to the clients
 *
 * XXX The ZEBRA_IPV*_ROUTE_ADD message is also sent by the client to the
 * zebra server when the client wants to tell the zebra server to add a
 * route to the kernel (zapi_ipv4_add etc. ).  Since it's essentially the
 * same message being sent back and forth, this function and
 * zapi_ipv{4,6}_{add, delete} should be re-written to avoid code
 * duplication.
 */
int
zsend_route_multipath (int cmd, struct zserv *client, struct prefix *p,
                       struct rib *rib)
{
  int psize;
  struct stream *s;
  struct nexthop *nexthop;
  unsigned long nhnummark = 0, messmark = 0;
  int nhnum = 0;
  u_char zapi_flags = 0;

  s = client->obuf;
  stream_reset (s);

  zserv_create_header (s, cmd);

  /* Put type and nexthop. */
  stream_putc (s, rib->type);
  stream_putc (s, rib->flags);

  /* marker for message flags field */
  messmark = stream_get_endp (s);
  stream_putc (s, 0);

  /* Prefix. */
  psize = PSIZE (p->prefixlen);
  stream_putc (s, p->prefixlen);
  stream_write (s, (u_char *) & p->u.prefix, psize);

  /*
   * XXX The message format sent by zebra below does not match the format
   * of the corresponding message expected by the zebra server
   * itself (e.g., see zread_ipv4_add). The nexthop_num is not set correctly,
   * (is there a bug on the client side if more than one segment is sent?)
   * nexthop ZEBRA_NEXTHOP_IPV4 is never set, ZEBRA_NEXTHOP_IFINDEX
   * is hard-coded.
   */
  /* Nexthop */

  for (nexthop = rib->nexthop; nexthop; nexthop = nexthop->next)
    {
      if (CHECK_FLAG (nexthop->flags, NEXTHOP_FLAG_FIB))
        {
          SET_FLAG (zapi_flags, ZAPI_MESSAGE_NEXTHOP);
          SET_FLAG (zapi_flags, ZAPI_MESSAGE_IFINDEX);

          if (nhnummark == 0)
            {
              nhnummark = stream_get_endp (s);
              stream_putc (s, 1); /* placeholder */
            }

          nhnum++;

          switch(nexthop->type)
            {
              case NEXTHOP_TYPE_IPV4:
              case NEXTHOP_TYPE_IPV4_IFINDEX:
                stream_put_in_addr (s, &nexthop->gate.ipv4);
                break;
#ifdef HAVE_IPV6
              case NEXTHOP_TYPE_IPV6:
              case NEXTHOP_TYPE_IPV6_IFINDEX:
              case NEXTHOP_TYPE_IPV6_IFNAME:
                stream_write (s, (u_char *) &nexthop->gate.ipv6, 16);
                break;
#endif
              default:
                if (cmd == ZEBRA_IPV4_ROUTE_ADD
                    || cmd == ZEBRA_IPV4_ROUTE_DELETE)
                  {
                    struct in_addr empty;
                    memset (&empty, 0, sizeof (struct in_addr));
                    stream_write (s, (u_char *) &empty, IPV4_MAX_BYTELEN);
                  }
                else
                  {
                    struct in6_addr empty;
                    memset (&empty, 0, sizeof (struct in6_addr));
                    stream_write (s, (u_char *) &empty, IPV6_MAX_BYTELEN);
                  }
              }

          /* Interface index. */
          stream_putc (s, 1);
          stream_putl (s, nexthop->ifindex);

          break;
        }
    }

  /* Metric */
  if (cmd == ZEBRA_IPV4_ROUTE_ADD || ZEBRA_IPV6_ROUTE_ADD)
    {
      SET_FLAG (zapi_flags, ZAPI_MESSAGE_DISTANCE);
      stream_putc (s, rib->distance);
      SET_FLAG (zapi_flags, ZAPI_MESSAGE_METRIC);
      stream_putl (s, rib->metric);
    }

  /* write real message flags value */
  stream_putc_at (s, messmark, zapi_flags);

  /* Write next-hop number */
  if (nhnummark)
    stream_putc_at (s, nhnummark, nhnum);

  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

#ifdef HAVE_IPV6
static int
zsend_ipv6_nexthop_lookup (struct zserv *client, struct in6_addr *addr)
{
  struct stream *s;
  struct rib *rib;
  unsigned long nump;
  u_char num;
  struct nexthop *nexthop;

  /* Lookup nexthop. */
  rib = rib_match_ipv6 (addr);

  /* Get output stream. */
  s = client->obuf;
  stream_reset (s);

  /* Fill in result. */
  zserv_create_header (s, ZEBRA_IPV6_NEXTHOP_LOOKUP);
  stream_put (s, &addr, 16);

  if (rib)
    {
      stream_putl (s, rib->metric);
      num = 0;
      nump = stream_get_endp(s);
      stream_putc (s, 0);
      for (nexthop = rib->nexthop; nexthop; nexthop = nexthop->next)
	if (CHECK_FLAG (nexthop->flags, NEXTHOP_FLAG_FIB))
	  {
	    stream_putc (s, nexthop->type);
	    switch (nexthop->type)
	      {
	      case ZEBRA_NEXTHOP_IPV6:
		stream_put (s, &nexthop->gate.ipv6, 16);
		break;
	      case ZEBRA_NEXTHOP_IPV6_IFINDEX:
	      case ZEBRA_NEXTHOP_IPV6_IFNAME:
		stream_put (s, &nexthop->gate.ipv6, 16);
		stream_putl (s, nexthop->ifindex);
		break;
	      case ZEBRA_NEXTHOP_IFINDEX:
	      case ZEBRA_NEXTHOP_IFNAME:
		stream_putl (s, nexthop->ifindex);
		break;
	      default:
                /* do nothing */
		break;
	      }
	    num++;
	  }
      stream_putc_at (s, nump, num);
    }
  else
    {
      stream_putl (s, 0);
      stream_putc (s, 0);
    }

  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}
#endif /* HAVE_IPV6 */

static int
zsend_ipv4_nexthop_lookup (struct zserv *client, struct in_addr addr)
{
  struct stream *s;
  struct rib *rib = NULL;
  unsigned long nump;
  u_char num;
  struct nexthop *nexthop;

  /* Lookup nexthop. */
  //rib = rib_match_ipv4 (addr);

  /* Get output stream. */
  s = client->obuf;
  stream_reset (s);

  /* Fill in result. */
  zserv_create_header (s, ZEBRA_IPV4_NEXTHOP_LOOKUP);
  stream_put_in_addr (s, &addr);

  if (rib)
    {
      stream_putl (s, rib->metric);
      num = 0;
      nump = stream_get_endp(s);
      stream_putc (s, 0);
      for (nexthop = rib->nexthop; nexthop; nexthop = nexthop->next)
	if (CHECK_FLAG (nexthop->flags, NEXTHOP_FLAG_FIB))
	  {
	    stream_putc (s, nexthop->type);
	    switch (nexthop->type)
	      {
	      case ZEBRA_NEXTHOP_IPV4:
		stream_put_in_addr (s, &nexthop->gate.ipv4);
		break;
	      case ZEBRA_NEXTHOP_IFINDEX:
	      case ZEBRA_NEXTHOP_IFNAME:
		stream_putl (s, nexthop->ifindex);
		break;
	      default:
                /* do nothing */
		break;
	      }
	    num++;
	  }
      stream_putc_at (s, nump, num);
    }
  else
    {
      stream_putl (s, 0);
      stream_putc (s, 0);
    }

  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

static int
zsend_ipv4_import_lookup (struct zserv *client, struct prefix_ipv4 *p)
{
  struct stream *s;
  struct rib *rib = NULL;
  unsigned long nump;
  u_char num;
  struct nexthop *nexthop;

  /* Lookup nexthop. */
  //rib = rib_lookup_ipv4 (p);

  /* Get output stream. */
  s = client->obuf;
  stream_reset (s);

  /* Fill in result. */
  zserv_create_header (s, ZEBRA_IPV4_IMPORT_LOOKUP);
  stream_put_in_addr (s, &p->prefix);

  if (rib)
    {
      stream_putl (s, rib->metric);
      num = 0;
      nump = stream_get_endp(s);
      stream_putc (s, 0);
      for (nexthop = rib->nexthop; nexthop; nexthop = nexthop->next)
	if (CHECK_FLAG (nexthop->flags, NEXTHOP_FLAG_FIB))
	  {
	    stream_putc (s, nexthop->type);
	    switch (nexthop->type)
	      {
	      case ZEBRA_NEXTHOP_IPV4:
		stream_put_in_addr (s, &nexthop->gate.ipv4);
		break;
	      case ZEBRA_NEXTHOP_IFINDEX:
	      case ZEBRA_NEXTHOP_IFNAME:
		stream_putl (s, nexthop->ifindex);
		break;
	      default:
                /* do nothing */
		break;
	      }
	    num++;
	  }
      stream_putc_at (s, nump, num);
    }
  else
    {
      stream_putl (s, 0);
      stream_putc (s, 0);
    }

  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

/* Router-id is updated. Send ZEBRA_ROUTER_ID_ADD to client. */
int
zsend_router_id_update (struct zserv *client, struct prefix *p, adj_type_t adj)
{
  struct stream *s;
  int blen;

  /* Check this client need interface information. */
  if (!client->ridinfo)
    return 0;

  s = client->obuf;
  stream_reset (s);

  /* Message type. */
  zserv_create_header (s, ZEBRA_ROUTER_ID_UPDATE);

  /* Prefix information. */
  stream_putc (s, p->family);
  blen = prefix_blen (p);
  stream_put (s, &p->u.prefix, blen);
  stream_putc (s, p->prefixlen);
  stream_putw (s, (uint16_t)(adj));
  /* Write packet size. */
  stream_putw_at (s, 0, stream_get_endp (s));

  return zebra_server_send_message(client);
}

/* Register zebra server interface information.  Send current all
   interface and address information. */
static int
zread_interface_add (struct zserv *client, u_short length)
{
  struct zlistnode *ifnode, *ifnnode;
  struct zlistnode *cnode, *cnnode;
  struct interface *ifp;
  struct connected *c;

  /* Interface information is needed. */
  client->ifinfo = 1;

  for (ALL_LIST_ELEMENTS (iflist, ifnode, ifnnode, ifp))
    {
      /* Skip pseudo interface. */
      if (! CHECK_FLAG (ifp->status, ZEBRA_INTERFACE_ACTIVE))
	continue;

      if (zsend_interface_add (client, ifp) < 0)
        return -1;

      for (ALL_LIST_ELEMENTS (ifp->connected, cnode, cnnode, c))
	{
	  /*if (CHECK_FLAG (c->conf, ZEBRA_IFC_REAL) &&
	      (zsend_interface_address (ZEBRA_INTERFACE_ADDRESS_ADD, client,
				        ifp, c) < 0))
	    return -1;*/
		zsend_interface_address (ZEBRA_INTERFACE_ADDRESS_ADD, client,
				        ifp, c);
	}
    }
  return 0;
}

/* Unregister zebra server interface information. */
static int
zread_interface_delete (struct zserv *client, u_short length)
{
  client->ifinfo = 0;
  return 0;
}

/* This function support multiple nexthop. */
/*
 * Parse the ZEBRA_IPV4_ROUTE_ADD sent from client. Update rib and
 * add kernel route.
 */
static int
zread_ipv4_add (struct zserv *client, u_short length)
{
  int i;
  struct rib *rib;
  struct prefix_ipv4 p;
  u_char message;
  struct in_addr nexthop;
  u_char nexthop_num;
  u_char nexthop_type;
  struct stream *s;
  unsigned int ifindex;
  u_char ifname_len;

  /* Get input stream.  */
  s = client->ibuf;

  /* Allocate new rib. */
  rib = XCALLOC (MTYPE_RIB, sizeof (struct rib));

  /* Type, flags, message. */
  rib->type = stream_getc (s);
  rib->flags = stream_getc (s);
  message = stream_getc (s);
  rib->uptime = time (NULL);

  /* IPv4 prefix. */
  memset (&p, 0, sizeof (struct prefix_ipv4));
  p.family = AF_INET;
  p.prefixlen = stream_getc (s);
  stream_get (&p.prefix, s, PSIZE (p.prefixlen));

  /* Nexthop parse. */
  if (CHECK_FLAG (message, ZAPI_MESSAGE_NEXTHOP))
    {
      nexthop_num = stream_getc (s);

      for (i = 0; i < nexthop_num; i++)
	{
	  nexthop_type = stream_getc (s);

	  switch (nexthop_type)
	    {
	    case ZEBRA_NEXTHOP_IFINDEX:
	      ifindex = stream_getl (s);
	      //nexthop_ifindex_add (rib, ifindex);
	      break;
	    case ZEBRA_NEXTHOP_IFNAME:
	      ifname_len = stream_getc (s);
	      stream_forward_getp (s, ifname_len);
	      break;
	    case ZEBRA_NEXTHOP_IPV4:
	      nexthop.s_addr = stream_get_ipv4 (s);
	      //nexthop_ipv4_add (rib, &nexthop);
	      break;
	    case ZEBRA_NEXTHOP_IPV6:
	      stream_forward_getp (s, IPV6_MAX_BYTELEN);
	      break;
      case ZEBRA_NEXTHOP_BLACKHOLE:
        //nexthop_blackhole_add (rib);
        break;
	    }
	}
    }

  /* Distance. */
  if (CHECK_FLAG (message, ZAPI_MESSAGE_DISTANCE))
    rib->distance = stream_getc (s);

  /* Metric. */
  if (CHECK_FLAG (message, ZAPI_MESSAGE_METRIC))
    rib->metric = stream_getl (s);

  /* Table */
  rib->table=zebrad.rtm_table_default;
  //rib_add_ipv4_multipath (&p, rib);
  return 0;
}

/* Zebra server IPv4 prefix delete function. */
static int
zread_ipv4_delete (struct zserv *client, u_short length)
{
  int i;
  struct stream *s;
  struct zapi_ipv4 api;
  struct in_addr nexthop;
  unsigned long ifindex;
  struct prefix_ipv4 p;
  u_char nexthop_num;
  u_char nexthop_type;
  u_char ifname_len;

  s = client->ibuf;
  ifindex = 0;
  nexthop.s_addr = 0;

  /* Type, flags, message. */
  api.type = stream_getc (s);
  api.flags = stream_getc (s);
  api.message = stream_getc (s);

  /* IPv4 prefix. */
  memset (&p, 0, sizeof (struct prefix_ipv4));
  p.family = AF_INET;
  p.prefixlen = stream_getc (s);
  stream_get (&p.prefix, s, PSIZE (p.prefixlen));

  /* Nexthop, ifindex, distance, metric. */
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_NEXTHOP))
    {
      nexthop_num = stream_getc (s);

      for (i = 0; i < nexthop_num; i++)
	{
	  nexthop_type = stream_getc (s);

	  switch (nexthop_type)
	    {
	    case ZEBRA_NEXTHOP_IFINDEX:
	      ifindex = stream_getl (s);
	      break;
	    case ZEBRA_NEXTHOP_IFNAME:
	      ifname_len = stream_getc (s);
	      stream_forward_getp (s, ifname_len);
	      break;
	    case ZEBRA_NEXTHOP_IPV4:
	      nexthop.s_addr = stream_get_ipv4 (s);
	      break;
	    case ZEBRA_NEXTHOP_IPV6:
	      stream_forward_getp (s, IPV6_MAX_BYTELEN);
	      break;
	    }
	}
    }

  /* Distance. */
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_DISTANCE))
    api.distance = stream_getc (s);
  else
    api.distance = 0;

  /* Metric. */
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_METRIC))
    api.metric = stream_getl (s);
  else
    api.metric = 0;

  /*rib_delete_ipv4 (api.type, api.flags, &p, &nexthop, ifindex,
		   client->rtm_table);*/
  return 0;
}

/* Nexthop lookup for IPv4. */
static int
zread_ipv4_nexthop_lookup (struct zserv *client, u_short length)
{
  struct in_addr addr;

  addr.s_addr = stream_get_ipv4 (client->ibuf);
  return zsend_ipv4_nexthop_lookup (client, addr);
}

/* Nexthop lookup for IPv4. */
static int
zread_ipv4_import_lookup (struct zserv *client, u_short length)
{
  struct prefix_ipv4 p;

  p.family = AF_INET;
  p.prefixlen = stream_getc (client->ibuf);
  p.prefix.s_addr = stream_get_ipv4 (client->ibuf);

  return zsend_ipv4_import_lookup (client, &p);
}

#ifdef HAVE_IPV6
/* Zebra server IPv6 prefix add function. */
static int
zread_ipv6_add (struct zserv *client, u_short length)
{
  int i;
  struct stream *s;
  struct zapi_ipv6 api;
  struct in6_addr nexthop;
  unsigned long ifindex;
  struct prefix_ipv6 p;

  s = client->ibuf;
  ifindex = 0;
  memset (&nexthop, 0, sizeof (struct in6_addr));

  /* Type, flags, message. */
  api.type = stream_getc (s);
  api.flags = stream_getc (s);
  api.message = stream_getc (s);

  /* IPv4 prefix. */
  memset (&p, 0, sizeof (struct prefix_ipv6));
  p.family = AF_INET6;
  p.prefixlen = stream_getc (s);
  stream_get (&p.prefix, s, PSIZE (p.prefixlen));

  /* Nexthop, ifindex, distance, metric. */
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_NEXTHOP))
    {
      u_char nexthop_type;

      api.nexthop_num = stream_getc (s);
      for (i = 0; i < api.nexthop_num; i++)
	{
	  nexthop_type = stream_getc (s);

	  switch (nexthop_type)
	    {
	    case ZEBRA_NEXTHOP_IPV6:
	      stream_get (&nexthop, s, 16);
	      break;
	    case ZEBRA_NEXTHOP_IFINDEX:
	      ifindex = stream_getl (s);
	      break;
	    }
	}
    }

  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_DISTANCE))
    api.distance = stream_getc (s);
  else
    api.distance = 0;

  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_METRIC))
    api.metric = stream_getl (s);
  else
    api.metric = 0;

  if (IN6_IS_ADDR_UNSPECIFIED (&nexthop))
    rib_add_ipv6 (api.type, api.flags, &p, NULL, ifindex, 0, api.metric,
		  api.distance);
  else
    rib_add_ipv6 (api.type, api.flags, &p, &nexthop, ifindex, 0, api.metric,
		  api.distance);
  return 0;
}

/* Zebra server IPv6 prefix delete function. */
static int
zread_ipv6_delete (struct zserv *client, u_short length)
{
  int i;
  struct stream *s;
  struct zapi_ipv6 api;
  struct in6_addr nexthop;
  unsigned long ifindex;
  struct prefix_ipv6 p;

  s = client->ibuf;
  ifindex = 0;
  memset (&nexthop, 0, sizeof (struct in6_addr));

  /* Type, flags, message. */
  api.type = stream_getc (s);
  api.flags = stream_getc (s);
  api.message = stream_getc (s);

  /* IPv4 prefix. */
  memset (&p, 0, sizeof (struct prefix_ipv6));
  p.family = AF_INET6;
  p.prefixlen = stream_getc (s);
  stream_get (&p.prefix, s, PSIZE (p.prefixlen));

  /* Nexthop, ifindex, distance, metric. */
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_NEXTHOP))
    {
      u_char nexthop_type;

      api.nexthop_num = stream_getc (s);
      for (i = 0; i < api.nexthop_num; i++)
	{
	  nexthop_type = stream_getc (s);

	  switch (nexthop_type)
	    {
	    case ZEBRA_NEXTHOP_IPV6:
	      stream_get (&nexthop, s, 16);
	      break;
	    case ZEBRA_NEXTHOP_IFINDEX:
	      ifindex = stream_getl (s);
	      break;
	    }
	}
    }

  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_DISTANCE))
    api.distance = stream_getc (s);
  else
    api.distance = 0;
  if (CHECK_FLAG (api.message, ZAPI_MESSAGE_METRIC))
    api.metric = stream_getl (s);
  else
    api.metric = 0;

  if (IN6_IS_ADDR_UNSPECIFIED (&nexthop))
    rib_delete_ipv6 (api.type, api.flags, &p, NULL, ifindex, 0);
  else
    rib_delete_ipv6 (api.type, api.flags, &p, &nexthop, ifindex, 0);
  return 0;
}

static int
zread_ipv6_nexthop_lookup (struct zserv *client, u_short length)
{
  struct in6_addr addr;
  char buf[BUFSIZ];

  stream_get (&addr, client->ibuf, 16);
  printf ("DEBUG %s\n", inet_ntop (AF_INET6, &addr, buf, BUFSIZ));

  return zsend_ipv6_nexthop_lookup (client, &addr);
}
#endif /* HAVE_IPV6 */

/* Register zebra server router-id information.  Send current router-id */
static int
zread_router_id_add (struct zserv *client, u_short length)
{
  struct prefix p;
  lrm_t *       lrm;

  lrm = lrm_lookup();

  p.u.prefix4.s_addr = 0;
  p.family           = AF_INET;
  p.prefixlen        = 32;

  /* Router-id information is needed. */
  client->ridinfo = 1;

  int result = 0;
  if (lrm) 
  {
    if (lrm->router_id != 0)
    {
      p.u.prefix4.s_addr = htonl(lrm->router_id);
      result = zsend_router_id_update (client,&p, INNI);
    }
    if (lrm->router_enni_id != 0)
    {
      p.u.prefix4.s_addr = htonl(lrm->router_enni_id);
      result = zsend_router_id_update (client,&p, ENNI);
    }
    if (lrm->router_uni_id != 0)
    {
      p.u.prefix4.s_addr = htonl(lrm->router_uni_id);
      result = zsend_router_id_update (client,&p, UNI);
    }
  }
  return result;
//  return zsend_router_id_update (client,&p, INNI); //TODO Adam: finish it
}

/* Unregister zebra server router-id information. */
static int
zread_router_id_delete (struct zserv *client, u_short length)
{
  client->ridinfo = 0;
  return 0;
}

/* Close zebra client. */
static void
zebra_client_close (struct zserv *client)
{
  /* Close file descriptor. */
  if (client->sock)
    {
      close (client->sock);
      client->sock = -1;
    }

  /* Free stream buffers. */
  if (client->ibuf)
    stream_free (client->ibuf);
  if (client->obuf)
    stream_free (client->obuf);
  if (client->wb)
    buffer_free(client->wb);

  /* Release threads. */
  if (client->t_read)
    thread_cancel (client->t_read);
  if (client->t_write)
    thread_cancel (client->t_write);
  if (client->t_suicide)
    thread_cancel (client->t_suicide);

#if !HAVE_OMNIORB
  if(client->client_type == SCN_CLIENT){
	zlog_debug("SCNGW connection closed");
	reset_lrm_scn();
	lrm_scn.err_counter = 0;
  }
#endif /* !HAVE_OMNIORB */

  /* Free client structure. */
  listnode_delete (/*zebrad.client_list*/lrm_m->client_list, client);
  XFREE (0, client);
}

/* Make new client. */
static void
zebra_client_create (int sock, int port)
{
  struct zserv *client;

  client = XCALLOC (0, sizeof (struct zserv));

  /* Make client input/output buffer. */
  client->sock = sock;
  client->ibuf = stream_new (ZEBRA_MAX_PACKET_SIZ);
  client->obuf = stream_new (ZEBRA_MAX_PACKET_SIZ);
  client->wb = buffer_new(0);

  /* Set table number. */
  client->rtm_table = zebrad.rtm_table_default;

#if !HAVE_OMNIORB
  if(port == SCN_PORT)
	  client->client_type = SCN_CLIENT;
#endif /* !HAVE_OMNIORB */

  /* Add this client to linked list. */
  listnode_add (/*zebrad.client_list*/ lrm_m->client_list, client);

  /* Make new read thread. */
  zebra_event (ZEBRA_READ, sock, client);
}

#if !HAVE_OMNIORB
void zread_lrm_scn_end_ack(struct zserv *client){

	zlog_debug("**End of synchronisation with SCNGW**");
	lrm_scn.status = UP;
}
#endif /* !HAVE_OMNIORB */

/* Handler of zebra service request. */
static int
zebra_client_read (struct thread *thread)
{
  int sock;
  struct zserv *client;
  size_t already;
  uint16_t length, command;
  uint8_t marker, version;

  /* Get thread data.  Reset reading thread because I'm running. */
  sock = THREAD_FD (thread);
  client = THREAD_ARG (thread);
  client->t_read = NULL;

  if (client->t_suicide)
    {
      zebra_client_close(client);
      return -1;
    }

  /* Read length and command (if we don't have it already). */
  if ((already = stream_get_endp(client->ibuf)) < ZEBRA_HEADER_SIZE)
    {
      ssize_t nbyte;
      if (((nbyte = stream_read_try (client->ibuf, sock,
				     ZEBRA_HEADER_SIZE-already)) == 0) ||
	  (nbyte == -1))
	{
	  /*if (IS_ZEBRA_DEBUG_EVENT)
	    zlog_debug ("connection closed socket [%d]", sock);*/
	  zebra_client_close (client);
	  return -1;
	}
      if (nbyte != (ssize_t)(ZEBRA_HEADER_SIZE-already))
	{
	  /* Try again later. */
	  zebra_event (ZEBRA_READ, sock, client);
	  return 0;
	}
      already = ZEBRA_HEADER_SIZE;
    }

  /* Reset to read from the beginning of the incoming packet. */
  stream_set_getp(client->ibuf, 0);

  /* Fetch header values */
  length = stream_getw (client->ibuf);
  marker = stream_getc (client->ibuf);
  version = stream_getc (client->ibuf);
  command = stream_getw (client->ibuf);

  if (marker != ZEBRA_HEADER_MARKER || version != ZSERV_VERSION)
    {
      zlog_err("%s: socket %d version mismatch, marker %d, version %d",
               __func__, sock, marker, version);
      zebra_client_close (client);
      return -1;
    }
  if (length < ZEBRA_HEADER_SIZE)
    {
      zlog_warn("%s: socket %d message length %u is less than header size %d",
	        __func__, sock, length, ZEBRA_HEADER_SIZE);
      zebra_client_close (client);
      return -1;
    }
  if (length > STREAM_SIZE(client->ibuf))
    {
      zlog_warn("%s: socket %d message length %u exceeds buffer size %lu",
	        __func__, sock, length, (u_long)STREAM_SIZE(client->ibuf));
      zebra_client_close (client);
      return -1;
    }

  /* Read rest of data. */
  if (already < length)
    {
      ssize_t nbyte;
      if (((nbyte = stream_read_try (client->ibuf, sock,
				     length-already)) == 0) ||
	  (nbyte == -1))
	{
	  /*if (IS_ZEBRA_DEBUG_EVENT)
	    zlog_debug ("connection closed [%d] when reading zebra data", sock);*/
	  zebra_client_close (client);
	  return -1;
	}
      if (nbyte != (ssize_t)(length-already))
        {
	  /* Try again later. */
	  zebra_event (ZEBRA_READ, sock, client);
	  return 0;
	}
    }

  length -= ZEBRA_HEADER_SIZE;

  /* Debug packet information. */
  /*if (IS_ZEBRA_DEBUG_EVENT)
    zlog_debug ("zebra message comes from socket [%d]", sock);*/

  /*if (IS_ZEBRA_DEBUG_PACKET && IS_ZEBRA_DEBUG_RECV)
    zlog_debug ("zebra message received [%s] %d", zserv_command_string (command), length)*/

  switch (command)
    {
    case ZEBRA_ROUTER_ID_ADD:
      zlog_debug("received ZEBRA_ROUTER_ID_ADD from client");
      zread_router_id_add (client, length);
      break;
    case ZEBRA_ROUTER_ID_DELETE:
      zread_router_id_delete (client, length);
      break;
    case ZEBRA_INTERFACE_ADD:
      zlog_debug("received ZEBRA_INTERFACE_ADD from client");
      zread_interface_add (client, length);
      break;
    case ZEBRA_INTERFACE_DELETE:
      zread_interface_delete (client, length);
      break;
    case ZEBRA_IPV4_ROUTE_ADD:
      zread_ipv4_add (client, length);
      break;
    case ZEBRA_IPV4_ROUTE_DELETE:
      zread_ipv4_delete (client, length);
      break;
#ifdef HAVE_IPV6
    case ZEBRA_IPV6_ROUTE_ADD:
      zread_ipv6_add (client, length);
      break;
    case ZEBRA_IPV6_ROUTE_DELETE:
      zread_ipv6_delete (client, length);
      break;
#endif /* HAVE_IPV6 */
    case ZEBRA_REDISTRIBUTE_ADD:
      //zebra_redistribute_add (command, client, length);
      break;
    case ZEBRA_REDISTRIBUTE_DELETE:
      //zebra_redistribute_delete (command, client, length);
      break;
    case ZEBRA_REDISTRIBUTE_DEFAULT_ADD:
      //zebra_redistribute_default_add (command, client, length);
      break;
    case ZEBRA_REDISTRIBUTE_DEFAULT_DELETE:
      //zebra_redistribute_default_delete (command, client, length);
      break;
    case ZEBRA_IPV4_NEXTHOP_LOOKUP:
      zread_ipv4_nexthop_lookup (client, length);
      break;
#ifdef HAVE_IPV6
    case ZEBRA_IPV6_NEXTHOP_LOOKUP:
      zread_ipv6_nexthop_lookup (client, length);
      break;
#endif /* HAVE_IPV6 */
    case ZEBRA_IPV4_IMPORT_LOOKUP:
      zread_ipv4_import_lookup (client, length);
      break;
    case ZEBRA_TEL_ADD:
      /*IB: should send the same message as for ZEBRA_INTERFACE_ADD
       * but with TEL adresses instead of real interface addresses*/
      zsend_tel_add(client, NULL);
      break;
#ifdef GMPLS
#if !HAVE_OMNIORB
    case LRM_SCNGW_IS_UP:
      zlog_debug("received message LRM_SCNGW_IS_UP, starting syncronisation");
      zread_lrm_scn_up(client);
      break;
    case SCNGW_ACK:
    case SCNGW_ACK_BEGIN:
      zread_lrm_scn_begin_ack(client);
      break;
    case SCNGW_SYNC_ERR:
	if(lrm_scn.err_counter < 5){
		zlog_debug("received %d error from SCNGW", lrm_scn.err_counter);
		lrm_scn.err_counter++;
		zread_lrm_scn_err(client);
        }
	else{
		zlog_debug("too many synchronisation errors");
		szend_lrm_scn_fatal_err(client);
	}
	break;
    case SCNGW_UNACK:
      break;
    case SCNGW_ACK_END:
      zread_lrm_scn_end_ack(client);
      break;
#endif /* !HAVE_OMNIORB */
#endif /* GMPLS */
    default:
      zlog_info ("Zebra received unknown command %d", command);
      break;
    }

  if (client->t_suicide)
    {
      /* No need to wait for thread callback, just kill immediately. */
      zebra_client_close(client);
      return -1;
    }

  stream_reset (client->ibuf);
  zebra_event (ZEBRA_READ, sock, client);
  return 0;
}

#if !HAVE_OMNIORB
/*begin syncro message*/
int zread_lrm_scn_up(struct zserv *client){

	struct stream *s;



	zlog_debug("**Begin of synchronisation with SCNGW**");
	s = client->obuf;
	stream_reset (s);
	zlog_debug("Sending LRM_BEGIN_SYNC");
	zserv_create_header(s, LRM_BEGIN_SYNC);

	stream_putl(s, lrm_scn.scnif_number);
	stream_putl(s, lrm_scn.cc_number);	/*send cc associations*/
	stream_putl(s, lrm_scn.tel_number);	/*send tel associations*/

	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client); /*send the LRM_BEGIN_SYNC to SCNGW*/
}

/*In case we receive a SYNC_ERROR we restart the synchronisation again*/
int zread_lrm_scn_err(struct zserv *client){

	struct stream *s;

	reset_lrm_scn();
	s = client->obuf;
	stream_reset (s);
	zlog_debug("Recived SYNC_ERROR from SCNGW");
	zserv_create_header(s, LRM_BEGIN_SYNC);
	stream_putl(s, lrm_scn.scnif_number);
	stream_putl(s, lrm_scn.cc_number);	/*send cc associations*/
	stream_putl(s, lrm_scn.tel_number);	/*send tel associations*/

	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client); /*send the LRM_BEGIN_SYNC to SCNGW*/
}


int zread_lrm_scn_begin_ack(struct zserv* client){

	lrm_t * lrm;
	struct zlist *tel_list, *cc_list, *scnif_list;
	struct zlistnode *ccnode, *ccnnode;
	struct zlistnode *telnode, *telnnode;
	struct zlistnode *scnifnode, *scnifnnode;
	te_link_t *tel;
	cc_t *cc;
	ctrl_intf_t *scnif;

	lrm = lrm_lookup();
	assert(lrm);
	tel_list = lrm->telink_list;
	cc_list  = lrm->cc_list;
	scnif_list = lrm->scn_if_list;


	zlog_debug("Received ACK FROM SCNGW");
	if(lrm_scn.status == SYNC){
		if(lrm_scn.tot_counter > 0){
			//zlog_debug("Sending one association, tot_counter = %d", lrm_scn.tot_counter);
			if(lrm_scn.scnif_counter > 0){
				for(ALL_LIST_ELEMENTS(scnif_list, scnifnode, scnifnnode, scnif)){
					if((scnif->adm_state == ENABLED) && (scnif->sync_status == NOT_SENT)){
						zsend_lrm_add_scnif(client, scnif);
						scnif->sync_status = SENT;
						lrm_scn.scnif_counter--;
						lrm_scn.tot_counter--;
						break;
					}
				}
			}
			else if(lrm_scn.cc_counter > 0){
				for(ALL_LIST_ELEMENTS(cc_list, ccnode, ccnnode, cc)){
					if((cc->op_state == UP) && (cc->sync_status == NOT_SENT)){
						zlog_debug("Sending cc assoc number %d",
							lrm_scn.cc_number - lrm_scn.cc_counter);
						zsend_lrm_add_cc(client, cc);
						cc->sync_status = SENT;
						lrm_scn.cc_counter--;
						lrm_scn.tot_counter--;
						break;
					}
				}
			}
			else if(lrm_scn.tel_counter > 0){
				for(ALL_LIST_ELEMENTS(tel_list, telnode, telnnode, tel)){
					if((tel->op_state == UP) && (tel->sync_status == NOT_SENT)){
						zlog_debug("Sending tel assoc number %d",
							lrm_scn.tel_number - lrm_scn.tel_counter);
						zsend_lrm_add_tel(client, tel);
						tel->sync_status = SENT;
						lrm_scn.tel_counter--;
						lrm_scn.tot_counter--;
						break;
					}
				}
			}
		}
		else /*lrm_scn.tot_counter == 0*/{

			zsend_lrm_scn_end_sync(client);

		}
	}
}


/*
int zread_lrm_all_assoc_add(struct zserv *client){

	struct zlist *tel_list;
	struct zlist *cc_list;
	struct zlistnode *telnode, *telnnode;
	te_link_t *tel;
	struct zlistnode *ccnode, *ccnnode;
	cc_t *cc;
	int cc_up_count = 0;
	int tel_up_count = 0;
	struct stream *s;
	lrm_t * lrm;

	lrm = lrm_lookup();

	assert(lrm);

	tel_list = lrm->telink_list;
	cc_list  = lrm->cc_list;

	s = client->obuf;
	stream_reset (s);
	zserv_create_header(s, LRM_ALL_ASSOC_ADD);
	//first we put the numbre of TEL associated to at least one control channel that we send
	for(ALL_LIST_ELEMENTS(tel_list, telnode, telnnode, tel)){
		if(tel->op_state == UP){
			tel_up_count++;
		}
	}
	zlog_debug("%s: tel_up_count = %d", __FUNCTION__, tel_up_count);
	stream_putl(s, tel_up_count);
	//loop over the TEL list
	for(ALL_LIST_ELEMENTS(tel_list, telnode, telnnode, tel)){
		if(tel->op_state == UP){
			stream_putl(s, tel->tel_key);
			zlog_debug("tel key = %d",tel->tel_key);
			stream_put_ipv4(s, tel->lcl_id.value.ipv4.s_addr);
			zlog_debug("lcl tel ip: %s", inet_ntoa(tel->lcl_id.value.ipv4));
			//the number of active control channels for this TE-Link
			stream_putl(s, tel->num_cc_up);
			zlog_debug("num cc up = %d",tel->num_cc_up);
			for(ALL_LIST_ELEMENTS(tel->cc_list, ccnode, ccnnode, cc)){
				if(cc->op_state == UP){
					stream_putl(s, cc->cc_id);
					zlog_debug("cc id= %d",cc->cc_id);
				}
			}
		}
	}

	//the cc/ctrl-if association
	//put the number of up cc to which an ctrl-if is associated that will be sent
	for(ALL_LIST_ELEMENTS(cc_list, ccnode, ccnnode, cc)){
		if(cc->op_state == UP)
			cc_up_count++;
	}
	zlog_debug("%s: cc_up_count = %d", __FUNCTION__, tel_up_count);
	stream_putl(s, cc_up_count);
	for(ALL_LIST_ELEMENTS(cc_list, ccnode, ccnnode, cc)){
		if(cc->op_state == UP){
			zlog_debug("cc_id = %d", cc->cc_id);
			stream_putl(s, cc->cc_id);
			zlog_debug("cc_ip_lcl = %s", addr_ntoa(cc->lcl_scn_addr));
			stream_put_ipv4(s, cc->lcl_scn_addr.value.ipv4.s_addr);
			zlog_debug("cc_ip_nbr = %s", addr_ntoa(cc->rem_scn_addr));
			stream_put_ipv4(s, cc->rem_scn_addr.value.ipv4.s_addr);
		}
	}

	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);

}*/

/*
 * This function writes a TEL/cc(s) associations: i.e. a TEL and
 * all the associated control channels, into the client's buffer.
 * */
void write_tel_cc_assoc(te_link_t *tel, struct stream *s){

	cc_t *cc;
	struct zlistnode *ccnode, *ccnnode;

	stream_putl(s, tel->tel_key);
	stream_putw(s, tel->adj_type);
	stream_put_ipv4(s, tel->lcl_id.value.ipv4.s_addr);
	stream_put_ipv4(s, tel->rem_id.value.ipv4.s_addr);
	stream_putl(s, tel->num_cc_up);
	for(ALL_LIST_ELEMENTS(tel->cc_list, ccnode, ccnnode, cc)){
		if((cc->op_state == UP) && (cc->adm_state == ENABLED)){
			stream_putl(s, cc->cc_id);
		}
	}
}

/*
 * this function writes a cc/(lcl ip, rem ip) into the client's buffer
 *
 */
void write_cc_ctrlif_assoc(cc_t *cc, struct stream *s){


	stream_putl(s, cc->cc_id);
	stream_put_ipv4(s, cc->lcl_scn_addr.value.ipv4.s_addr);
	stream_put_ipv4(s, cc->rem_scn_addr.value.ipv4.s_addr);
}


int zsend_lrm_scn_end_sync(struct zserv *client){

	struct stream *s;

	s = client->obuf;
	stream_reset (s);
	zlog_debug("Sending LRM_END_SYNC to SCNGW");
	zserv_create_header(s, LRM_END_SYNC);

	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}


/*
 * Deletes on single TEL. So it puts the
 * number of TEL/cc(s) to delete.
 * */
int zsend_lrm_del_tel(struct zserv *client, int tel_key){

	struct stream *s;

	s = client->obuf;
	stream_reset (s);
	zserv_create_header(s, LRM_TEL_CC_ASSOC_DELETE);
	stream_putl(s, 1);
	stream_putl(s, tel_key);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

/*
 * this function sends to SCN the addition of one single
 * TEL and its associated cc(s). the supplied TEL link must be in UP state
 * */
int zsend_lrm_add_tel(struct zserv *client, te_link_t *tel){

	struct stream *s;

	s = client->obuf;
	stream_reset(s);
	zserv_create_header(s, LRM_TEL_CC_ASSOC_ADD);
	stream_putl(s, 1);
	write_tel_cc_assoc(tel, s);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

/*
 * this function is sends to SCN a newly added cc with it associated (ip lcl, ip rem)
 * */
int zsend_lrm_add_cc(struct zserv *client, cc_t *cc){

	struct stream *s;

	s = client->obuf;
	stream_reset(s);
	zserv_create_header(s, LRM_CC_SCNIF_ASSOC_ADD);
	stream_putl(s, 1); //we send only one cc association
	zlog_debug("Sending cc to client:");
	zlog_debug("\t cc_id = %d", cc->cc_id);
	zlog_debug("\t local ip addr: %s", addr_ntoa(cc->lcl_scn_addr));
	zlog_debug("\t nbr ip addr: %s", addr_ntoa(cc->rem_scn_addr));
	write_cc_ctrlif_assoc(cc, s);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}
/*
 * This function updates one cc it sends the cc_id and the new coupe of
 * ip addresses
 * */
int zsend_lrm_update_cc(struct zserv *client, cc_t *cc){

	struct stream *s;

	s = client->obuf;
	stream_reset(s);
	zserv_create_header(s, LRM_CC_SCNIF_ASSOC_UPDATE);
	write_cc_ctrlif_assoc(cc, s);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

/*
 * this function deletes a (one) control channel.
 * its sends in the message the number of control channels to delete (1 hard coded)
 *
 * */
int zsend_lrm_del_cc(struct zserv *client, int cc_id){

	struct stream *s;

	s = client->obuf;
	stream_reset(s);
	zserv_create_header(s, LRM_CC_SCNIF_ASSOC_DELETE);
	stream_putl(s, 1);
	//writes the id of the cc we want to delete.
	stream_putl(s, cc_id);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

int zsend_lrm_add_scnif(struct zserv *client, ctrl_intf_t *scnif){

	struct stream *s;

	s = client->obuf;
	stream_reset(s);

	zserv_create_header(s, LRM_SCNIF_ADD);
	zlog_debug("sending control interface addr: %s", inet_ntoa(scnif->addr.value.ipv4));
	stream_put_ipv4(s, scnif->addr.value.ipv4.s_addr);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

int zsend_lrm_del_scnif(struct zserv *client, ctrl_intf_t *scnif){


	struct stream *s;

	s = client->obuf;
	stream_reset(s);
	zserv_create_header(s, LRM_SCNIF_DELETE);
	stream_put_ipv4(s, scnif->addr.value.ipv4.s_addr);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}



/*
 * this function updates one TE link sending the number of updates to 1
 * the cc id to add/remove
 * */
int zsend_lrm_update_tel(struct zserv *client, te_link_t *tel, int op, u_int32_t cc_id){

	struct stream *s;
	s = client->obuf;

	stream_reset(s);
	zserv_create_header(s, LRM_TEL_CC_ASSOC_UPDATE);
	//body of the message.
	stream_putl(s, tel->tel_key);
	//put the number of updates to 1 for now
	stream_putl(s, 1);
	stream_putl(s, op);
	stream_putl(s, cc_id);

	//also send the associated control channel just in case SCNGWS don't have already
	//zsend_lrm_add_cc(client, tel->assoc_cc);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

/*int zsend(struct zserv *client){

	struct stream *s;
	s = client->obuf;

	stream_reset(s);
	zserv_create_header(s, LRM_END_SYNC);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}*/

int szend_lrm_scn_fatal_err(struct zserv *client){

	struct stream *s;
	s = client->obuf;

	stream_reset(s);
	zlog_debug("sending LRM_SYNC_FATAL_ERROR to SCNGW");
	zserv_create_header(s, LRM_SYNC_FATAL_ERROR);
	stream_putw_at(s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);

}
#endif /* !HAVE_OMNIORB */

int
zsend_te_interface_update(struct zserv     *client,
			  struct interface *ifp,
			  int               up_type)
{

	struct zlistnode *node, *nnode;
	calendar_event_t * event;
	struct stream *s;
	s = client->obuf;
	int i;
	int *data;
	stream_reset(s);
	//zlog_debug("updating interface %s", ifp->name);
	zserv_create_header(s, ZEBRA_INTERFACE_UPDATE);
	stream_put (s, ifp->name, INTERFACE_NAMSIZ);
	stream_putl(s, up_type);
	switch(up_type){
		case METRIC_UPDATE:
			stream_putl(s, ifp->te_metric);
			break;
		case LINK_CLR_UPDATE:
			stream_putl(s, ifp->te_link_color);
			break;
		case SRLGid_UPDATE:
			stream_putl(s, ifp->te_SRLG_ids->count);
			for(ALL_LIST_ELEMENTS(ifp->te_SRLG_ids, node, nnode, data)){
				stream_putl(s, *data);
			}
			break;
		case BW_UPDATE:
			stream_putl(s, ifp->te_max_bw);
			stream_putl(s, ifp->te_max_res_bw);
			stream_putl(s, ifp->te_min_LSP_bw);
			for(i = 0; i < MAX_BW_PRIORITIES; i++){
				stream_putl(s, ifp->te_avail_bw_per_prio[i]);
			}
			for(i = 0; i < MAX_BW_PRIORITIES; i++){
				stream_putl(s, ifp->te_max_LSP_bw[i]);
			}
			stream_putl(s, (uint32_t) ifp->adv_rsrv_calendar->count);
			for(ALL_LIST_ELEMENTS(ifp->adv_rsrv_calendar, node, nnode, event)) {
				stream_putl(s, event->time_stamp);
				for (i=0; i< MAX_BW_PRIORITIES; i++) {
					stream_putl(s, event->avail_bw[i]);
				}
			}
			if (ifp->te_swcap == SWCAP_LSC) {
				stream_putl(s, ifp->lambdas_bitmap.base_lambda_label);
				stream_putw(s, ifp->lambdas_bitmap.num_wavelengths);
				stream_putw(s, ifp->lambdas_bitmap.bitmap_size);
				/*check for ifp->lambdas_bitmap_word pointer is inside stream_put function*/
				stream_put (s, ifp->lambdas_bitmap.bitmap_word, 4 * (ifp->lambdas_bitmap.bitmap_size));
			}
			break;
		case TNA_UPDATE:
			switch (ifp->te_TNA_address_type)
			{
				case IPv4:
					stream_putl(s, 32);
					stream_putl(s, ifp->te_TNA_prefix_length);
					stream_putl(s, *(ifp->te_TNA_address));
					break;
				case IPv6:
					stream_putl(s, 128);
					stream_putl(s, ifp->te_TNA_prefix_length);
					stream_putl(s, ifp->te_TNA_address[0]);
					stream_putl(s, ifp->te_TNA_address[1]);
					stream_putl(s, ifp->te_TNA_address[2]);
					stream_putl(s, ifp->te_TNA_address[3]);
					break;
				case NSAP:
					stream_putl(s, 160);
					stream_putl(s, ifp->te_TNA_prefix_length);
					stream_putl(s, ifp->te_TNA_address[0]);
					stream_putl(s, ifp->te_TNA_address[1]);
					stream_putl(s, ifp->te_TNA_address[2]);
					stream_putl(s, ifp->te_TNA_address[3]);
					stream_putl(s, ifp->te_TNA_address[4]);
					break;
				default:
					zlog_err("Bad TNA type %d",
						 ifp->te_TNA_address_type);
					assert(0);
			}
			break;
		case PROTECTION_UPDATE:
			stream_putl(s, ifp->te_protection_type);
			break;
		default:
			zlog_debug("unknow kind of interface update");
			break;
	}
	stream_putw_at (s, 0, stream_get_endp (s));
	return zebra_server_send_message(client);
}

/* Accept code of zebra server socket. */
static int
zebra_accept (struct thread *thread)
{
  int accept_sock;
  int client_sock;
  struct sockaddr_in client;
  socklen_t len;
  int port;

  accept_sock = THREAD_FD (thread);

  /* Reregister myself. */
  zebra_event (ZEBRA_SERV, accept_sock, NULL);

  len = sizeof (struct sockaddr_in);
  client_sock = accept (accept_sock, (struct sockaddr *) &client, &len);
  port = htons(client.sin_port);


  if (client_sock < 0)
    {
      zlog_warn ("Can't accept zebra socket: %s", safe_strerror (errno));
      return -1;
    }
  else
	  zlog_debug("connection accepted from %s:%d", inet_ntoa(client.sin_addr), ntohs(client.sin_port));

  /* Make client socket non-blocking.  */
  set_nonblocking(client_sock);

  /* Create new zebra client. */
  zebra_client_create (client_sock, port);

  return 0;
}

#ifdef HAVE_TCP_ZEBRA
/* Make zebra's server socket. */
static void
zebra_serv ()
{
  int ret;
  int accept_sock;
  struct sockaddr_in addr;

  accept_sock = socket (AF_INET, SOCK_STREAM, 0);

  if (accept_sock < 0)
    {
      zlog_warn ("Can't create zserv stream socket: %s",
                 safe_strerror (errno));
      zlog_warn ("zebra can't provice full functionality due to above error");
      return;
    }

  memset (&addr, 0, sizeof (struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_port = htons (ZEBRA_PORT);
#ifdef HAVE_SIN_LEN
  addr.sin_len = sizeof (struct sockaddr_in);
#endif /* HAVE_SIN_LEN */
  addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
  sockopt_reuseaddr (accept_sock);
  sockopt_reuseport (accept_sock);

  /*if ( zserv_privs.change(ZPRIVS_RAISE) )
    zlog (NULL, LOG_ERR, "Can't raise privileges");*/

  ret  = bind (accept_sock, (struct sockaddr *)&addr,
	       sizeof (struct sockaddr_in));
  if (ret < 0)
    {
      zlog_warn ("Can't bind to stream socket: %s",
                 safe_strerror (errno));
      zlog_warn ("zebra can't provice full functionality due to above error");
      close (accept_sock);      /* Avoid sd leak. */
      return;
    }

  /*if ( zserv_privs.change(ZPRIVS_LOWER) )
    zlog (NULL, LOG_ERR, "Can't lower privileges");*/

  ret = listen (accept_sock, 1);
  if (ret < 0)
    {
      zlog_warn ("Can't listen to stream socket: %s",
                 safe_strerror (errno));
      zlog_warn ("zebra can't provice full functionality due to above error");
      close (accept_sock);	/* Avoid sd leak. */
      return;
    }

  zebra_event (ZEBRA_SERV, accept_sock, NULL);
}
#endif /* HAVE_TCP_ZEBRA */

/* For sockaddr_un. */
#include <sys/un.h>

static void
zebra_event (enum event event, int sock, struct zserv *client)
{
  switch (event)
    {
    case ZEBRA_SERV:
      thread_add_read (/*zebrad.master*/ master, zebra_accept, client, sock);
      break;
    case ZEBRA_READ:
      client->t_read =
	thread_add_read (/*zebrad.master*/ master, zebra_client_read, client, sock);
      break;
    case ZEBRA_WRITE:
      /**/
      break;
    }
}

/* Display default rtm_table for all clients. */
/*DEFUN (show_table,
       show_table_cmd,
       "show table",
       SHOW_STR
       "default routing table to use for all clients\n")
{
  vty_out (vty, "table %d%s", zebrad.rtm_table_default,
	   VTY_NEWLINE);
  return CMD_SUCCESS;
}

DEFUN (config_table,
       config_table_cmd,
       "table TABLENO",
       "Configure target kernel routing table\n"
       "TABLE integer\n")
{
  zebrad.rtm_table_default = strtol (argv[0], (char**)0, 10);
  return CMD_SUCCESS;
}*/
/*
DEFUN (ip_forwarding,
       ip_forwarding_cmd,
       "ip forwarding",
       IP_STR
       "Turn on IP forwarding")
{
  int ret;

  ret = ipforward ();
  if (ret == 0)
    ret = ipforward_on ();

  if (ret == 0)
    {
      vty_out (vty, "Can't turn on IP forwarding%s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}*/

/*DEFUN (no_ip_forwarding,
       no_ip_forwarding_cmd,
       "no ip forwarding",
       NO_STR
       IP_STR
       "Turn off IP forwarding")
{
  int ret;

  ret = ipforward ();
  if (ret != 0)
    ret = ipforward_off ();

  if (ret != 0)
    {
      vty_out (vty, "Can't turn off IP forwarding%s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}*/

/* This command is for debugging purpose. */
/*
DEFUN (show_zebra_client,
       show_zebra_client_cmd,
       "show zebra client",
       SHOW_STR
       "Zebra information"
       "Client information")
{
  struct zlistnode *node;
  struct zserv *client;

  for (ALL_LIST_ELEMENTS_RO (zebrad.client_list, node, client))
    vty_out (vty, "Client fd %d%s", client->sock, VTY_NEWLINE);

  return CMD_SUCCESS;
}
*/

/* table node for routing tables. */
struct cmd_node table_node =
{
  TABLE_NODE,
  "",				/* This node has no interface. */
  1
};

/* Only display ip forwarding is enabled or not. */
/*DEFUN (show_ip_forwarding,
       show_ip_forwarding_cmd,
       "show ip forwarding",
       SHOW_STR
       IP_STR
       "IP forwarding status\n")
{
  int ret;

  ret = ipforward ();

  if (ret == 0)
    vty_out (vty, "IP forwarding is off%s", VTY_NEWLINE);
  else
    vty_out (vty, "IP forwarding is on%s", VTY_NEWLINE);
  return CMD_SUCCESS;
}*/

#ifdef HAVE_IPV6
/* Only display ipv6 forwarding is enabled or not. */
/*DEFUN (show_ipv6_forwarding,
       show_ipv6_forwarding_cmd,
       "show ipv6 forwarding",
       SHOW_STR
       "IPv6 information\n"
       "Forwarding status\n")
{
  int ret;

  ret = ipforward_ipv6 ();

  switch (ret)
    {
    case -1:
      vty_out (vty, "ipv6 forwarding is unknown%s", VTY_NEWLINE);
      break;
    case 0:
      vty_out (vty, "ipv6 forwarding is %s%s", "off", VTY_NEWLINE);
      break;
    case 1:
      vty_out (vty, "ipv6 forwarding is %s%s", "on", VTY_NEWLINE);
      break;
    default:
      vty_out (vty, "ipv6 forwarding is %s%s", "off", VTY_NEWLINE);
      break;
    }
  return CMD_SUCCESS;
}

DEFUN (ipv6_forwarding,
       ipv6_forwarding_cmd,
       "ipv6 forwarding",
       IPV6_STR
       "Turn on IPv6 forwarding")
{
  int ret;

  ret = ipforward_ipv6 ();
  if (ret == 0)
    ret = ipforward_ipv6_on ();

  if (ret == 0)
    {
      vty_out (vty, "Can't turn on IPv6 forwarding%s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}

DEFUN (no_ipv6_forwarding,
       no_ipv6_forwarding_cmd,
       "no ipv6 forwarding",
       NO_STR
       IPV6_STR
       "Turn off IPv6 forwarding")
{
  int ret;

  ret = ipforward_ipv6 ();
  if (ret != 0)
    ret = ipforward_ipv6_off ();

  if (ret != 0)
    {
      vty_out (vty, "Can't turn off IPv6 forwarding%s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}*/

#endif /* HAVE_IPV6 */

/* table node for routing tables. */
struct cmd_node forwarding_node =
{
  FORWARDING_NODE,
  "",				/* This node has no interface. */
  1
};


/* Initialisation of zebra and installation of commands. */
void
zebra_init (void)
{
  /* Client list init. */
  //zebrad.client_list = list_new ();
    lrm_m->client_list = list_new ();


  /* Make zebra server socket. */
#ifdef HAVE_TCP_ZEBRA
  zebra_serv ();
#else
  zebra_serv_un (ZEBRA_SERV_PATH);
#endif /* HAVE_TCP_ZEBRA */

  /* Install configuration write function. */
  //install_node (&table_node, config_write_table);
  //install_node (&forwarding_node, config_write_forwarding);

  //install_element (VIEW_NODE, &show_ip_forwarding_cmd);
  //install_element (ENABLE_NODE, &show_ip_forwarding_cmd);
  //install_element (CONFIG_NODE, &ip_forwarding_cmd);
  //install_element (CONFIG_NODE, &no_ip_forwarding_cmd);
  //install_element (ENABLE_NODE, &show_zebra_client_cmd);

#ifdef HAVE_NETLINK
  //install_element (VIEW_NODE, &show_table_cmd);
  //install_element (ENABLE_NODE, &show_table_cmd);
  //install_element (CONFIG_NODE, &config_table_cmd);
#endif /* HAVE_NETLINK */

#ifdef HAVE_IPV6
  //install_element (VIEW_NODE, &show_ipv6_forwarding_cmd);
  //install_element (ENABLE_NODE, &show_ipv6_forwarding_cmd);
  //install_element (CONFIG_NODE, &ipv6_forwarding_cmd);
  //install_element (CONFIG_NODE, &no_ipv6_forwarding_cmd);
#endif /* HAVE_IPV6 */

}
