//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//




#include <zebra.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <getopt.h>

#include "log.h"
#include "version.h"
#include "command.h"
#include "thread.h"
#include "sigevent.h"
#include "privs.h"
#include "memory.h"
#include "thread.h"
#include "stream.h"
#include "sockopt.h"

#include "nccd.hh"
#include "ncc_config.hh"
#include "ncc_corba.hh"
#include "ncc_trace.hh"
#include "lib/corba.h"


struct thread_master * master = NULL;

// nccd privileges
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

// GLOBALS
struct zebra_privs_t   nccd_privs;
struct quagga_signal_t ncc_signals[4];
char *                 progname;

// Master of threads

const char *           pid_file         = PATH_NCCD_PID;
char                   config_default[] = NCCD_DEFAULT_CONFIG;


struct option longopts[] = {
	{ "daemon",      no_argument,       NULL, 'd'},
	{ "config_file", required_argument, NULL, 'f'},
	{ "pid_file",    required_argument, NULL, 'i'},
        { "dryrun",      no_argument,       NULL, 'C'},
	{ "iors_dir",    required_argument, NULL, 'o'},
	{ "help",        no_argument,       NULL, 'h'},
	{ "vty_port",    required_argument, NULL, 'P'},
	{ "user",        required_argument, NULL, 'u'},
	{ "group",       required_argument, NULL, 'g'},
	{ "version",     no_argument,       NULL, 'v'},
	{ 0 }
};


// SIG handlers
static void
sighup (void)
{
	NCC_INFO("SIGHUP received");
}

static int terminate_execution = 0;
static void
sigint (void)
{
	NCC_NOTICE("Terminating on SIGINT");
	// ncc_terminate();
	terminate_execution = 1;

	if (NCC_MASTER.dryrun()) {
		exit(-1);
	}
}

// SIGUSR1 handler.
static void
sigusr1 (void)
{
       zlog_rotate(NULL);
}


// Help information display.
static void __attribute__ ((noreturn))
usage (char *progname, int status)
{
	if (status != 0) {
		fprintf (stderr,
			 "Try `%s --help' for more information.\n",
			 progname);
	} else {
		printf ("Usage : %s [OPTION...]\n"
			"Daemon which manages NCC.\n\n"
			"-d, --daemon       Runs in daemon mode\n"
			"-f, --config_file  Set configuration file name\n"
			"-i, --pid_file     Set process identifier file name\n"
                        "-C, --dryrun       Check configuration for validity and exit\n"
			"-o, --iors_dir     Set IORs directory\n"
			"-P, --vty_port     Set vty's port number\n"
			"-u, --user         User to run as\n"
			"-g, --group        Group to run as\n"
			"-v, --version      Print program version\n"
			"-h, --help         Display this help and exit\n"
			"\n"
			"Report bugs to %s\n",
			progname, ZEBRA_BUG_ADDRESS);
	}

	exit(status);
}

int main(int argc, char ** argv)
{
	char          * p;
	int             vty_port;
	char          * vty_addr;
	int             daemon_mode;
	char          * config_file;
	char          * iors_dir;
	struct thread   thread;

	vty_port    = NCCD_VTY_PORT;
	vty_addr    = NULL;
	daemon_mode = 0;
	config_file = NULL;
	iors_dir    = NULL;

	// INITIALIZATION
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	nccd_privs.user      = QUAGGA_USER;
	nccd_privs.group     = QUAGGA_GROUP;
#endif
#if defined(VTY_GROUP)
	nccd_privs.vty_group = VTY_GROUP;
#endif
	nccd_privs.caps_p    = _caps_p;
	nccd_privs.cap_num_p = sizeof(_caps_p)/sizeof(_caps_p[0]);
	nccd_privs.cap_num_i = 0;

	ncc_signals[0].signal = SIGHUP;
	ncc_signals[0].handler = &sighup;
	ncc_signals[1].signal = SIGUSR1;
	ncc_signals[1].handler = &sigusr1;
	ncc_signals[2].signal = SIGINT;
	ncc_signals[2].handler = &sigint;
	ncc_signals[3].signal = SIGTERM;
	ncc_signals[3].handler = &sigint;

	// Set umask before anything for security
	umask(0027);

	// Get program name
	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	// Initialize the log subsystem
	zlog_default = openzlog(progname,
				ZLOG_NCC,
				LOG_CONS | LOG_NDELAY | LOG_PID,
				LOG_DAEMON);

	//rm = new NCC_Master;
	if (!NCC_Master::init()) {
		NCC_NOTICE("Cannot allocate an NCC master object");
		exit (-1);
	}
	master = NCC_MASTER.getMaster();

	// just reads the options from stdin
	while (1) {
		int opt;

		opt = getopt_long(argc, argv, "dChvf:P:u:g:v:i:o:",
				  longopts, 0);

		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
                        case 'C':
				NCC_MASTER.dryrun(true);
                                break;
			case 'f':
				config_file = optarg;
				break;
			case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
			case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				nccd_privs.user = optarg;
				break;
			case 'g':
				nccd_privs.group = optarg;
				break;
			case 'v':
				printf("NCC daemon v.%s: vty@%d\n",
                                       NCC_CODE_VERSION, vty_port);
                                printf("This program is part of "
                                       "phosphorus-g2mpls\n");
                                //print_version(progname);
				exit (0);
				break;
			case 'h':
				usage(progname, 0);
				break;
			default:
				usage(progname, 1);
				break;
		}
	}

	// Library inits.
	zprivs_init(&nccd_privs);
	signal_init(NCC_MASTER.getMaster(),
		    Q_SIGC(ncc_signals),
		    ncc_signals);
	cmd_init(1);

	// these are all from libzebra
	vty_init(NCC_MASTER.getMaster());
	memory_init();

	NCC_MASTER.init_vty();

	// sorting is needed by the command subsystem to finalize init
	sort_node();

	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		try {
			corba_init(iors_dir,
				   CORBA_SERVANT_NETWORK_CALLCONTROLLER,
				   NCC_MASTER.getMaster());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem (err=%s)\n",
				e.c_str());
			exit(-1);
		}

		try {
			corba_client_setup(NCC_MASTER.getNccProxy());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot setup client object (err=%s)\n",
				e.c_str());
			exit(-1);
		} catch (CORBA::BAD_PARAM & e) {
			fprintf(stderr,
				"Cannot setup client object "
				"(CORBA::BAD_PARAM)\n");
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot setup client object "
				"(exception other)\n");
			exit(-1);
		}


		try {
			corba_init(iors_dir,
				   CORBA_SERVANT_RC_NBI,
				   NCC_MASTER.getMaster());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem (err=%s)\n",
				e.c_str());
			exit(-1);
		}

		try {
			corba_client_setup(NCC_MASTER.getRcNBIProxy());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot setup client object (err=%s)\n",
				e.c_str());
			exit(-1);
		} catch (CORBA::BAD_PARAM & e) {
			fprintf(stderr,
				"Cannot setup client object "
				"(CORBA::BAD_PARAM)\n");
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot setup client object "
				"(exception other)\n");
			exit(-1);
		}

		try {
			corba_init(iors_dir,
				   CORBA_SERVANT_LRM,
				   NCC_MASTER.getMaster());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem (err=%s)\n",
				e.c_str());
			exit(-1);
		}

		try {
			corba_client_setup(NCC_MASTER.getLrmProxy());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot setup client object (err=%s)\n",
				e.c_str());
			exit(-1);
		} catch (CORBA::BAD_PARAM & e) {
			fprintf(stderr,
				"Cannot setup client object "
				"(CORBA::BAD_PARAM)\n");
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot setup client object "
				"(exception other)\n");
			exit(-1);
		}


		try {
			LRM::Info_var lrmProxy;

			lrmProxy = NCC_MASTER.getLrmProxy();

			if (CORBA::is_nil(lrmProxy)) {
				fprintf(stderr,
					"Cannot get Corba LRM client object "
					"previously set up\n");
				assert(0);
			}

			g2mplsTypes::nodeId myNid;

			myNid = lrmProxy->nodeId();

			NCC_MASTER.routerId(myNid);

		} catch (...) {
			fprintf(stderr,
				"Cannot get  Router ID from LRM\n");
			exit(-1);
		}
#endif
	}

	vty_read_config(config_file, config_default);	// Get .conf file

	// Change to the daemon program.
	if (daemon_mode) {
		daemon(0, 0);
	}

	// Create VTY socket and start the TCP and unix socket listeners
	vty_serv_sock(vty_addr, vty_port, NCC_VTYSH_PATH);

	// Print banner.
	NCC_NOTICE("NCC daemon v.%s: starting... vty@%d",
		    NCC_CODE_VERSION, vty_port);

	// Process id file create.
	pid_output(pid_file);

	while (!terminate_execution) {
		if (thread_fetch(NCC_MASTER.getMaster(), &thread)) {
			thread_call(&thread);
		}
	}

	NCC_DEBUG("Termination in progress ...");

	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
	try {
		(void) corba_client_shutdown(CORBA_SERVANT_NETWORK_CALLCONTROLLER);
		(void) corba_client_shutdown(CORBA_SERVANT_RC_NBI);
		(void) corba_client_shutdown(CORBA_SERVANT_LRM);
		(void) corba_fini();
	} catch(...) {
		fprintf(stderr, "Somethig was wrong with corba shutdown\n");
		exit(-1);
	}
#endif
	}

	return 0;
}
