//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"


#include <map>
#include <list>
#include <string>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <stdexcept>

#include "nccd.hh"
#include "ncc_corba.hh"
#include "ncc_trace.hh"
#include "ncc_utils.hh"


//
//
//  CALL container
//
//
#define CALLCTRL_SCOPE "NULL, selected by CallCtrl"
#define CALLCTRL_WORK "(if " CALLCTRL_SCOPE")"

#define BAR1 "========================================"	\
             "========================================"
#define BAR2 "----------------------------------------"	\
             "----------------------------------------"
#define SHOWLIST_DEPTH 50

#define VTY_ERROR_PRINT(VTY, ERR)			\
   vty_out(VTY, "%s%s", (ERR).c_str(), VTY_NEWLINE);

#define VTY_ERROR_AND_EXIT(VTY, ERR)		\
   VTY_ERROR_PRINT(VTY, ERR);			\
   return CMD_WARNING;


//
//
//
//   VTY commands
//
//
//
struct cmd_node ncc_node = {
	NCC_NODE,
	"%s(config-ncc)# ",
	1
};

DEFUN(ncc,
      ncc_cmd,
      "ncc",
      "Enable a NCC process\n"
      "Start NCC configuration\n")
{
	std::string  err;

	vty->node  = NCC_NODE;
	vty->index = NULL;

	vty_out(vty,
		"Entering NCC Controller VTY interface%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(no_ncc,
      no_ncc_cmd,
      "no ncc",
      NO_STR
      "Disable a NCC process\n"
      "Stop NCC operation\n")
{
	std::string  err;

	vty_out(vty,
		"Exiting NCC Controller VTY interface%s",
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_create,
      ncc_call_create_cmd,
      "call create id WORD itu-country WORD itu-carrier WORD unique-AP WORD",
      "Create CALL with given input parameters")
{
	mapKey_t         key;
	uint32_t         ituCountryCode;	//  just 24 bits
	uint64_t         ituCarrierCode;	//  just 48 bits
	uint64_t         uniqueAP;		//  just 48 bits

	std::string      err;

	ituCountryCode = 0;
	ituCarrierCode = 0;
	uniqueAP       = 0;

	key = strtoull(argv[0], NULL, 16);


	if (argc > 1) {
		ituCountryCode  = strtoul(argv[1], NULL, 16);
		ituCountryCode &= 0x00FFFFFF;	//  just 24 bits

		ituCarrierCode  = strtoull(argv[2], NULL, 16);
		ituCarrierCode &= 0x0000FFFFFFFFFFFFull;	// 48 bits

		uniqueAP        = strtoull(argv[3], NULL, 16);
		uniqueAP       &= 0x0000FFFFFFFFFFFFull;	// 48 bits
	}

	NCC_DEBUG("Going to create an NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key,ituCountryCode,ituCarrierCode,uniqueAP,err)) {
		vty_out(vty,
			"Cannot update data into VTY CALL with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"NCC CALL with localID %s successfully created in VTY%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(ncc_call_create,
      ncc_call_create_compact_cmd,
      "call create id WORD ",
      "Create CALL with given input parameters");


DEFUN(ncc_call_set_info_callres60,
      ncc_call_set_info_callres60_cmd,
      "call set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "data-link (WORD|A.B.C.D) "
      "label60 dest-mac WORD vlan-id <1-4094>",
      "Update info on tributary resources into CALL")
{
	mapKey_t         key;
	call_info_t      cinfo;
	net_res_spec_t   res;
	std::string      err;

	memset(&cinfo, 0, sizeof(cinfo));
	memset(&res, 0, sizeof(res));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "in-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, iTNA_res);
		BITMASK_BITSET(cinfo.iTNA_res.mask_, net);
	} else if (strcmp(argv[1], "eg-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, eTNA_res);
		BITMASK_BITSET(cinfo.eTNA_res.mask_, net);
	} else {
		vty_out(vty,
			"Invalid TNA resource specification (%s) for a CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		res.tna.type       = IPv4;
		res.tna.preflen    = 32;
		if (!inet_aton(argv[2], &(res.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for TNA%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for TNA%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(res.mask_, tna);

	if (strchr (argv[3], '.') != NULL) {
		// matched "A.B.C.D".
		res.data_link.type       = IPv4;
		res.data_link.preflen    = 32;
		if (!inet_aton(argv[3], &(res.data_link.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for data link%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		// match "<0-4294967295>".
		res.data_link.type                = UNNUMBERED;
		res.data_link.preflen             = 0;
		res.data_link.value.unnum         = strtoul(argv[3], NULL, 16);
	}
	BITMASK_BITSET(res.mask_, data_link);

	res.label.type = LABEL_60BIT;
	std::string str;
	int   i;

	str = std::string(argv[4]);

	for (i = 0;  i< 6; i++) {
		std::string::size_type cursor;
		cursor = str.find(":");
		if (!str.empty()) {
			uint32_t num;
			num = strtoul(str.c_str(), NULL, 16);
			res.label.value.label60.mac[i] = (uint8_t) num;
		}

		str = (cursor == std::string::npos) ?
			"" :str.substr(cursor + 1);
	}

	res.label.value.label60.vlan_id = strtoul(argv[5], NULL, 10);
	BITMASK_BITSET(res.mask_, label);

	if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
		memcpy(&(cinfo.iTNA_res.net), &res, sizeof(res));
		defaultGridResSpec(cinfo.iTNA_res);
	} else {
		memcpy(&(cinfo.eTNA_res.net), &res, sizeof(res));
		defaultGridResSpec(cinfo.eTNA_res);
	}

	NCC_DEBUG("Going to update info on tributary resources "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update info on tributary resources "
			"on VTY CALL with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Info on tributary resources successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_callres32,
      ncc_call_set_info_callres32_cmd,
      "call set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "data-link (WORD|A.B.C.D) "
      "label32 id WORD",
      "Update info on tributary resources into CALL")
{
	mapKey_t         key;
	call_info_t      cinfo;
	net_res_spec_t   res;
	std::string      err;

	memset(&cinfo, 0, sizeof(cinfo));
	memset(&res, 0, sizeof(res));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "in-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, iTNA_res);
		BITMASK_BITSET(cinfo.iTNA_res.mask_, net);
	} else if (strcmp(argv[1], "eg-trib") == 0) {
		BITMASK_BITSET(cinfo.mask_, eTNA_res);
		BITMASK_BITSET(cinfo.eTNA_res.mask_, net);
	} else {
		vty_out(vty,
			"Invalid TNA resource specification (%s) for a CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		res.tna.type       = IPv4;
		res.tna.preflen    = 32;
		if (!inet_aton(argv[2], &(res.tna.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for TNA%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for TNA%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(res.mask_, tna);

	if (strchr (argv[3], '.') != NULL) {
		// matched "A.B.C.D".
		res.data_link.type       = IPv4;
		res.data_link.preflen    = 32;
		if (!inet_aton(argv[3], &(res.data_link.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for data link%s",
				argv[3], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		// match "<0-4294967295>".
		res.data_link.type                = UNNUMBERED;
		res.data_link.preflen             = 0;
		res.data_link.value.unnum         = strtoul(argv[3], NULL, 16);
	}
	BITMASK_BITSET(res.mask_, data_link);

	res.label.type = LABEL_32BIT;
	res.label.value.label32.id = strtoul(argv[4], NULL, 16);
	BITMASK_BITSET(res.mask_, label);

	if (BITMASK_BITTEST(cinfo.mask_, iTNA_res)) {
		memcpy(&(cinfo.iTNA_res.net), &res, sizeof(res));
	} else {
		memcpy(&(cinfo.eTNA_res.net), &res, sizeof(res));
	}

	NCC_DEBUG("Going to update info on tributary resources "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update info on tributary resources "
			"on VTY CALL with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Info on tributary resources successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(ncc_call_set_info_callres32,
      ncc_call_set_info_callres32_compact_cmd,
      "call set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "dl (WORD|A.B.C.D) "
      "l32 WORD",
      "Update info on tributary resources into CALL");

ALIAS(ncc_call_set_info_callres60,
      ncc_call_set_info_callres60_compact_cmd,
      "call set-info id WORD call-data (in-trib|eg-trib) net "
      "tna A.B.C.D "
      "dl (WORD|A.B.C.D) "
      "l60 dmac WORD vid <1-4094>)",
      "Update info on tributary resources into CALL");

#if 0
DEFUN(ncc_call_set_info_calldnid,
      ncc_call_set_info_calldnid_cmd,
      "call set-info id WORD call-data domain-egress-lsr A.B.C.D",
      "Update info on domain egress LSR into CALL")
{
	mapKey_t         key;
	call_info_t      cinfo;
	call_info_mask_t cmask;
	std::string      err;

	memset(&cinfo, 0, sizeof(cinfo));
	BITMASK_RESET(cmask);

	key = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cinfo.destNid.type    = IPv4;
		cinfo.destNid.preflen = 32;
		if (!inet_aton(argv[1], &(cinfo.destNid.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for "
				"domain egress LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(cmask, destNid);
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"domain egress LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to update info on domain egress LSR "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str(), VTY_NEWLINE);


	if (!vtyCallUpdate(key, cinfo, cmask, err)) {
		vty_out(vty,
			"Cannot update info on domain egress LSR "
			"on VTY CALL with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Domain egress LSR successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}
#endif

DEFUN(ncc_call_set_info_calltime,
      ncc_call_set_info_calltime_cmd,
      "call set-info id WORD gns-call-data (start-time|end-time) "
      "year <2008-2037> "
      "mon <1-12> "
      "day <1-31> "
      "hour <0-23> "
      "min <0-59> "
      "sec <0-59>",
      "Update GNS time info into CALL (in local timezone)")
{
	mapKey_t             key;
	call_info_t          cinfo;
	struct tm *          bdTime;
	time_t               unixTime;
	std::string          err;

	memset(&cinfo, 0, sizeof(cinfo));

	time(&unixTime);
	bdTime = localtime(&unixTime);

	key = strtoull(argv[0], NULL, 16);

	bdTime->tm_year  = atoi(argv[2]) - 1900;
	bdTime->tm_mon   = atoi(argv[3]) - 1;
	bdTime->tm_mday  = atoi(argv[4]);
	bdTime->tm_hour  = atoi(argv[5]);
	bdTime->tm_min   = atoi(argv[6]);
	bdTime->tm_sec   = atoi(argv[7]);
	bdTime->tm_isdst = -1;	// Actively determines whether DST is in effect

	unixTime = mktime(bdTime);

	if (strcmp(argv[1], "start-time") == 0) {
		BITMASK_BITSET(cinfo.mask_, times);
		cinfo.times.start_time = unixTime;
	} else if (strcmp(argv[1], "end-time") == 0) {
		BITMASK_BITSET(cinfo.mask_, times);
		cinfo.times.end_time = unixTime;
	} else {
		vty_out(vty,
			"Invalid timing specification (%s) for a CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to update GNS time info for "
		  "NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update GNS time info into VTY CALL "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"GNS time info successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_callnames,
      ncc_call_set_info_callnames_cmd,
      "call set-info id WORD gns-call-data "
      "job-name WORD "
      "job-project WORD",
      "Update GNS naming info into CALL")
{
	mapKey_t          key;
	call_info_t       cinfo;
	std::string       err;

	memset(&cinfo, 0, sizeof(cinfo));

	key = strtoull(argv[0], NULL, 16);

	cinfo.job_name = new std::string(argv[1]);
	BITMASK_BITSET(cinfo.mask_, job_name);
	cinfo.job_project = new std::string(argv[2]);
	BITMASK_BITSET(cinfo.mask_, job_project);

	NCC_DEBUG("Going to update GNS naming info "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key, cinfo, err)) {
		vty_out(vty,
			"Cannot update GNS naming info into VTY CALL "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"GNS naming info successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_recovery,
      ncc_call_set_info_recovery_cmd,
      "call set-info id WORD recovery-data "
      "type (none|protection|pre-planned|on-the-fly|revertive-on-the-fly) "
      "disjointness (none|link|node|srlg)",
      "Update info on recovery into CALL")
{
	mapKey_t             key;
	recovery_info_t      rinfo;
	std::string          err;

	memset(&rinfo, 0, sizeof(rinfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "none") == 0) {
		rinfo.rec_type = RECOVERY_NONE;
	} else if (strcmp(argv[1], "protection") == 0) {
		rinfo.rec_type = RECOVERY_PROTECTION;
	} else if (strcmp(argv[1], "pre-planned") == 0) {
		rinfo.rec_type = RECOVERY_PREPLANNED;
	} else if (strcmp(argv[1], "on-the-fly") == 0) {
		rinfo.rec_type = RECOVERY_ONTHEFLY;
	} else if (strcmp(argv[1], "revertive-on-the-fly") == 0) {
		rinfo.rec_type = RECOVERY_REVERTIVEONTHEFLY;
	} else {
		vty_out(vty,
			"Invalid recovery type (%s) for a CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	BITMASK_BITSET(rinfo.mask_, rec_type);

	if (strcmp(argv[2], "none") == 0) {
		rinfo.disj_type = DISJOINTNESS_NONE;
	} else if (strcmp(argv[2], "link") == 0) {
		rinfo.disj_type = DISJOINTNESS_LINK;
	} else if (strcmp(argv[2], "node") == 0) {
		rinfo.disj_type = DISJOINTNESS_NODE;
	} else if (strcmp(argv[2], "srlg") == 0) {
		rinfo.disj_type = DISJOINTNESS_SRLG;
	} else {
		vty_out(vty,
			"Invalid recovery type (%s) for a CALL%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	BITMASK_BITSET(rinfo.mask_, disj_type);

	NCC_DEBUG("Going to update info on recovery "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key, rinfo, err)) {
		vty_out(vty,
			"Cannot update info on recovery LSR "
			"on VTY CALL with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"Recovery info successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_lspbase,
      ncc_call_set_info_lspbase_cmd,
      "call set-info id WORD lsp-data "
      "sw-cap (l2sc|lsc|fsc) "
      "enc-type (ethernet|lambda|fiber) "
      "bw WORD "
      "gpid WORD "
      "action (xconnect|book) "
      "type (spc|pc)"
      "role (primary|secondary)",
      "Update LSP base info into CALL")
{
	mapKey_t             key;
	lsp_info_t           linfo;
	float                bw;
	std::string          err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp(argv[1], "l2sc") == 0) {
		linfo.sw_cap = SWCAP_L2SC;
	} else if (strcmp(argv[1], "lsc") == 0) {
		linfo.sw_cap = SWCAP_LSC;
	} else if (strcmp(argv[1], "fsc") == 0) {
		linfo.sw_cap = SWCAP_FSC;
	} else {
		vty_out(vty,
			"Invalid switching cap. (%s) for a CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, sw_cap);

	if (strcmp(argv[2], "ethernet") == 0) {
		linfo.enc_type = ENCT_ETHERNET;
	} else if (strcmp(argv[2], "lambda") == 0) {
		linfo.enc_type = ENCT_LAMBDA;
	} else if (strcmp(argv[2], "fiber") == 0) {
		linfo.enc_type = ENCT_FIBER;
	} else {
		vty_out(vty,
			"Invalid encoding type (%s) for a CALL%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, enc_type);

	bw         = strtof(argv[3], NULL);
	linfo.bw   = BW_BPS2HEX(bw);
	BITMASK_BITSET(linfo.mask_, bw);
	linfo.gpid = (gpid_t) strtoul(argv[4], NULL, 10);
	BITMASK_BITSET(linfo.mask_, gpid);

	if (argc > 5) {

		if (strcmp (argv[5], "xconnect") == 0) {
			linfo.action = RESOURCE_XCONNECT;
		} else if (strcmp (argv[5], "book") == 0) {
			linfo.action = RESOURCE_BOOK;
		} else {
			vty_out(vty,
				"Invalid resource behavior type (%s) "
				"for CALL%s",
				argv[5], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, action);

		if (strcmp (argv[6], "spc") == 0) {
			linfo.type = LSP_TYPE_SPC;
		} else if (strcmp (argv[6], "pc") == 0) {
			linfo.type = LSP_TYPE_PC;
		} else {
			vty_out(vty,
				"Invalid LSP type (%s) "
				"for CALL%s",
				argv[6], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, type);

		if (strcmp (argv[7], "primary") == 0) {
			linfo.role = LSP_ROLE_PRIMARY;
		} else if (strcmp (argv[7], "secondary") == 0) {
			linfo.role = LSP_ROLE_SECONDARY;
		} else {
			vty_out(vty,
				"Invalid LSP role (%s) "
				"for CALL%s",
				argv[7], VTY_NEWLINE);
			return CMD_WARNING;
		}
		BITMASK_BITSET(linfo.mask_, role);
	}

	NCC_DEBUG("Going to update LSP base info for "
		  "NCC CALL with localID %s",
		  key2string(key).c_str());


	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update base info into VTY CALL "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP base info successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

ALIAS(ncc_call_set_info_lspbase,
      ncc_call_set_info_lspbase_compact_cmd,
      "call set-info id WORD lsp-data "
      "sw-cap (l2sc|lsc|fsc) "
      "enc-type (ethernet|lambda|fiber) "
      "bw WORD "
      "gpid WORD",
      "Update LSP info data into CALL");

DEFUN(ncc_call_set_info_lspconstr,
      ncc_call_set_info_lspconstr_cmd,
      "call set-info id WORD lsp-constraints "
      "s-prio <0-7> "
      "h-prio <0-7> "
      "exc-any WORD "
      "inc-any WORD "
      "inc-all WORD "
      "link-prot (none|extra|unprotected|shared|ded-1:1|ded-1+1|enhanced)",
      "Update LSP constraints info into CALL")
{
	mapKey_t             key;
	lsp_info_t           linfo;
	std::string          err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	linfo.setup_prio   = strtoul(argv[1], NULL, 10);
	linfo.holding_prio = strtoul(argv[2], NULL, 10);
	linfo.exclude_any  = strtoul(argv[3], NULL, 16);
	linfo.include_any  = strtoul(argv[4], NULL, 16);
	linfo.include_all  = strtoul(argv[5], NULL, 16);

	BITMASK_BITSET(linfo.mask_, setup_prio);
	BITMASK_BITSET(linfo.mask_, holding_prio);
	BITMASK_BITSET(linfo.mask_, exclude_any);
	BITMASK_BITSET(linfo.mask_, include_any);
	BITMASK_BITSET(linfo.mask_, include_all);

	if (strcmp (argv[6], "none") == 0) {
		linfo.link_prot_mask = PROTTYPE_NONE;
	} else if (strcmp (argv[6], "extra") == 0) {
		linfo.link_prot_mask = PROTTYPE_EXTRA;
	} else if (strcmp (argv[6], "unprotected") == 0) {
		linfo.link_prot_mask = PROTTYPE_UNPROTECTED;
	} else if (strcmp (argv[6], "shared") == 0) {
		linfo.link_prot_mask = PROTTYPE_SHARED;
	} else if (strcmp (argv[6], "ded-1:1") == 0) {
		linfo.link_prot_mask = PROTTYPE_DEDICATED_1TO1;
	} else if (strcmp (argv[6], "ded-1+1") == 0) {
		linfo.link_prot_mask = PROTTYPE_DEDICATED_1PLUS1;
	} else if (strcmp (argv[6], "enhanced") == 0) {
		linfo.link_prot_mask = PROTTYPE_ENHANCED;
	} else {
		vty_out(vty,
			"Invalid link protection type (%s) "
			"for CALL%s",
			argv[6], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, link_prot_mask);

	NCC_DEBUG("Going to update LSP constraints "
		  "for NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update LSP constraints into VTY CALL "
			"with localID %s (err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP constraints successfully updated "
		"in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_lspproto,
      ncc_call_set_info_lspproto_cmd,
      "call set-info id WORD lsp-proto "
      "cback-scope (none|e2e|boundary|segment) "
      "cback-max-retr-src <0-5> "
      "cback-max-retr-intmd <0-5> "
      "record-route (all|dl-detail|tel-detail|off)",
      "Update protocol related LSP info data into CALL")
{
	mapKey_t             key;
	lsp_info_t           linfo;
	std::string          err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	if (strcmp (argv[1], "none") == 0) {
		linfo.crankback = CB_NONE;
	} else if (strcmp (argv[1], "e2e") == 0) {
		linfo.crankback = CB_E2E;
	} else if (strcmp (argv[1], "boundary") == 0) {
		linfo.crankback = CB_BOUNDARY;
	} else if (strcmp (argv[1], "segment") == 0) {
		linfo.crankback = CB_SEGMENT;
	} else {
		vty_out(vty,
			"Invalid crakback scope (%s) "
			"for CALL%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, crankback);

	linfo.max_cback_retries_src    = strtoul(argv[2], NULL, 10);
	BITMASK_BITSET(linfo.mask_, max_cback_retries_src);

	linfo.max_cback_retries_intmd  = strtoul(argv[3], NULL, 10);
	BITMASK_BITSET(linfo.mask_, max_cback_retries_intmd);

	if (strcmp (argv[4], "all") == 0) {
		linfo.rro_mode = LSP_RRO_ALL;
	} else if (strcmp (argv[4], "dl-detail") == 0) {
		linfo.rro_mode = LSP_RRO_DL_DETAIL;
	} else if (strcmp (argv[4], "tel-detail") == 0) {
		linfo.rro_mode = LSP_RRO_TEL_DETAIL;
	} else if (strcmp (argv[4], "off") == 0) {
		linfo.rro_mode = LSP_RRO_OFF;
	} else {
		vty_out(vty,
			"Invalid RRO recording mode (%s) for CALL%s",
			argv[4], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, rro_mode);

	NCC_DEBUG("Going to update LSP data for NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update data into VTY CALL with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP data successfully updated in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_set_info_lspprototimers,
      ncc_call_set_info_lspprototimers_cmd,
      "call set-info id WORD lsp-proto-timers "
      "refresh-period <0-4294967295> "
      "ack-procedure (on|off) "
      "rapid-retransmission-interval <0-4294967295> "
      "rapid-retry-limit <1-5> "
      "increment-value-delta <1-10>",
      "Update protocol related LSP info data into CALL")
{
	mapKey_t             key;
	lsp_info_t           linfo;
	std::string          err;

	memset(&linfo, 0, sizeof(linfo));

	key = strtoull(argv[0], NULL, 16);

	linfo.refresh_interval = strtoul(argv[1], NULL, 10);
	BITMASK_BITSET(linfo.mask_, refresh_interval);

	if (strcmp (argv[2], "on") == 0) {
		linfo.activate_ack = 1; //true
	} else if (strcmp (argv[2], "off") == 0) {
		linfo.activate_ack = 0; //false
	} else {
		vty_out(vty,
			"Invalid ACK mode (%s) for CALL%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}
	BITMASK_BITSET(linfo.mask_, activate_ack);

	linfo.rapid_retransm_interval = strtoul(argv[3], NULL, 10);
	linfo.rapid_retry_limit       = strtoul(argv[4], NULL, 10);
	linfo.increment_value_delta   = strtoul(argv[5], NULL, 10);

	BITMASK_BITSET(linfo.mask_, rapid_retransm_interval);
	BITMASK_BITSET(linfo.mask_, rapid_retry_limit);
	BITMASK_BITSET(linfo.mask_, increment_value_delta);

	NCC_DEBUG("Going to update LSP data for NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallUpdate(key, linfo, err)) {
		vty_out(vty,
			"Cannot update data into VTY CALL with localID %s "
			"(err reason='%s')%s",
			key2string(key).c_str(), err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"LSP data successfully updated in NCC CALL with localID %s%s",
		key2string(key).c_str(), VTY_NEWLINE);

	return CMD_SUCCESS;
}


DEFUN(ncc_call_commit,
      ncc_call_commit_cmd,
      "call commit id WORD",
      "Commit CALL with given input parameters")
{
	mapKey_t      key;
	std::string   err;

	key = strtoull(argv[0], NULL, 16);

	NCC_DEBUG("Going to commit a NCC CALL with localID %s",
		  key2string(key).c_str());

	if (!vtyCallCommit(key, err)) {
		vty_out(vty,
			"Cannot commit a NCC CALL from VTY: %s%s",
			err.c_str(),
			VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty,
		"NCC CALL with localID %s successfully committed%s",
		key2string(key).c_str(),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_destroy,
      ncc_call_destroy_cmd,
      "call destroy id WORD src-lsr A.B.C.D",
      "Destroy CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}


	NCC_DEBUG("Going to destroy a NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Sending callDestroy through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;
		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callDestroy(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callDestroy through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		NCC_DEBUG("Sent callDestroy through CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully destroyed%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_enable,
      ncc_call_enable_cmd,
      "call enable id WORD src-lsr A.B.C.D",
      "Enable CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to enable a NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Sending callEnable through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);

		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callEnable(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callEnable through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		NCC_DEBUG("Sent callEnable through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully enabled%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_disable,
      ncc_call_disable_cmd,
      "call disable id WORD src-lsr A.B.C.D",
      "Disable CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to disable a NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Sending callDisable through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);

		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callDisable(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callDisable through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		NCC_DEBUG("Sent callDisable through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully disabled%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_setup,
      ncc_call_setup_cmd,
      "call set-up id WORD src-lsr A.B.C.D",
      "Set up a CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to setup a NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Sending callSetUp through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callSetUp(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callSetUp through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		NCC_DEBUG("Sent callSetUp through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully set-up%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_call_setdown,
      ncc_call_setdown_cmd,
      "call set-down id WORD src-lsr A.B.C.D",
      "Set down a CALL with given input parameters")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	NCC_DEBUG("Going to setdown a NCC CALL with localID %s "
		  "from src LSR %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));


	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Sending callSetDown through CORBA Mgmt interface");

		CallController::Mgmt_var nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set down");
			assert(0);
		}

		g2mplsTypes::callIdent_var id;
		//		id = new g2mplsTypes::callIdent;

		try {
			id << cid;
		} catch (std::out_of_range & e) {
			err  = "Exception occurred during "
				"callId translation (";
			err += e.what();
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		g2mplsTypes::actorInfo_var actor;
		actor = new g2mplsTypes::actorInfo;

		actor->issuer = g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF;
		actor->forceCommand = true;

		try {
			nccMgmtProxy->callSetDown(id, actor);
		} catch (Types::InternalProblems & e) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "callSetDown through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		NCC_DEBUG("Sent callSetDown through "
			  "CORBA Mgmt interface");
#endif // HAVE_OMNIORB
	}

	vty_out(vty,
		"NCC CALL with localID %s and src LSR %s "
		"successfully set-down%s",
		key2string(cid.local_id).c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);

	return CMD_SUCCESS;
}

DEFUN(ncc_show,
      ncc_show_cmd,
      "show ncc",
      "Show configured protocol instance")
{
	size_t      vtyCalls;
	size_t      nccCalls;
	std::string err;

	if (!getCalls(vtyCalls, nccCalls, err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "NCC CONTROLLER VTY INSTANCE%s", VTY_NEWLINE);

	vty_out(vty, "   - NCC CALL(s) pending in VTY   : %d%s",
		vtyCalls, VTY_NEWLINE);

	vty_out(vty, "   - NCC CALL(s) committed to NCC : %d%s",
		nccCalls, VTY_NEWLINE);

	vty_out(vty, "%s", VTY_NEWLINE);
	return CMD_SUCCESS;
}


DEFUN(ncc_show_calls,
      ncc_show_calls_cmd,
      "show ncc call",
      "Show all the calls on this NCC")
{
	int         i;
	std::string err;

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "NCC CALL(s) installed on NCC%s", VTY_NEWLINE);
	if (!NCC_MASTER.dryrun()) {
#if HAVE_OMNIORB
		NCC_DEBUG("Requesting list of Calls");

		CallController::Mgmt_var  nccMgmtProxy;

		nccMgmtProxy = NCC_MASTER.getNccProxy();

		if (CORBA::is_nil(nccMgmtProxy)) {
			NCC_ERROR("Cannot get Corba nccMgmtProxy client "
				  "object previously set up");
			assert(0);
		}

		CallController::Mgmt::callIdentSeq_var s;

		try {
			s = nccMgmtProxy->getCalls();
		} catch (Types::InternalProblems & e) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for internal problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (Types::CannotFetch & e) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for fetching problems ";
			err += "(reason= ";
			err += e.what;
			err += ")";
			VTY_ERROR_AND_EXIT(vty, err);
		} catch (...) {
			err += "getCalls through CORBA Mgmt ";
			err += "interface failed for unknown reason";
			VTY_ERROR_AND_EXIT(vty, err);
		}

		if (s->length() == 0) {
			vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
			vty_out(vty, "%s", VTY_NEWLINE);
			return CMD_SUCCESS;
		}

		size_t i;

		for (i = 0; i < s->length(); i++) {
			g2mpls_addr_t lsr;
			uint64_t      id;
			try {
				g2mplsTypes::sourceId_var corbaLsr;
				corbaLsr = (s)[i].srcId;
				lsr << corbaLsr;
				id = (s)[i].localId;
			} catch (std::out_of_range & e) {
				vty_out(vty,
					"Exception occurred during "
					"types translation (%s)%s",
					e.what(), VTY_NEWLINE);
				return CMD_WARNING;
			}

			vty_out(vty,
				"   - [c] #%02d: "
				"src-LSR = %s / local-ID = %s%s",
				i,
				addr_ntoa(lsr),
				key2string(id).c_str(),
				VTY_NEWLINE);

			if (i == SHOWLIST_DEPTH) {
				break;
			}
		}

		vty_out(vty, "%s", VTY_NEWLINE);
		vty_out(vty, "   printed %d NCC committed CALL(s)%s",
			i, VTY_NEWLINE);
#endif // HAVE_OMNIORB
	} else {
		vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
	}

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);

	vtyCallDumpVty(vty);

	vty_out(vty, "%s", VTY_NEWLINE);

	return CMD_SUCCESS;
}


#define CORBAACTORDUMP2VTY(actor_, owner_)				     \
{									     \
	vty_out(vty, "%s actor%s", owner_, VTY_NEWLINE);		     \
	vty_out(vty, "\t - issuer          : %s%s",			     \
		((actor_->issuer == g2mplsTypes::ISSUERTYPE_MANAGEMENT_IF) ? \
		 "MANAGEMENT" :					     \
		 ((actor_->issuer == g2mplsTypes::ISSUERTYPE_UNI_IF) ?	     \
		  "UNI" :						     \
		  "ENNI")),						     \
		VTY_NEWLINE);						     \
	vty_out(vty, "\t - forces command  : %s%s",			     \
		(actor_->forceCommand) ? "YES" : "NO", VTY_NEWLINE);	     \
}

#define CORBASTATESDUMP2VTY(states_, owner_)				  \
{									  \
	vty_out(vty, "%s states%s", owner_, VTY_NEWLINE);		  \
	vty_out(vty, "\t - op. state       : %s%s",			  \
		(states_->opState == g2mplsTypes::OPERSTATE_UP) ?	  \
		"UP" : "DOWN", VTY_NEWLINE);				  \
	vty_out(vty, "\t - adm. state      : %s%s",			  \
		(states_->admState == g2mplsTypes::ADMINSTATE_DISABLED) ? \
		"DISABLED" : "ENABLED", VTY_NEWLINE);			  \
}


DEFUN(ncc_show_call,
      ncc_show_call_cmd,
      "show ncc call id WORD src-lsr A.B.C.D",
      "Show details for specified CALL")
{
	call_ident_t  cid;
	std::string   err;

	memset(&cid, 0, sizeof(cid));

	cid.type     = CALLID_OPSPEC;
	cid.local_id = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		cid.src_addr.type       = IPv4;
		cid.src_addr.preflen    = 32;
		if (!inet_aton(argv[1], &(cid.src_addr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	} else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "NCC installed CALL%s", VTY_NEWLINE);

	if (NCC_MASTER.dryrun()) {
		vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}
#if HAVE_OMNIORB
	NCC_DEBUG("Requesting details for Call with localID %s "
		  "and src LSR %s %s",
		  key2string(cid.local_id).c_str(),
		  addr_ntoa(cid.src_addr));

	CallController::Mgmt_var  nccMgmtProxy;

	nccMgmtProxy = NCC_MASTER.getNccProxy();

	if (CORBA::is_nil(nccMgmtProxy)) {
		NCC_ERROR("Cannot get Corba nccMgmtProxy client object "
			  "previously set up");
		assert(0);
	}

	g2mplsTypes::callIdent_var id;
	//	id = new g2mplsTypes::callIdent;

	try {
		id << cid;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callId translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	g2mplsTypes::callParams_var                   cinfo;
	g2mplsTypes::recoveryParams_var               rinfo;
	g2mplsTypes::lspParams_var                    linfo;
	g2mplsTypes::actorInfo_var                    actor;
	g2mplsTypes::statesBundle_var                 states;
	CallController::Mgmt::recoBundleIdentSeq_var  recoBundles;

	try {
		nccMgmtProxy->callGetDetails(id         ,
					     cinfo      ,
					     rinfo      ,
					     linfo      ,
					     actor      ,
					     states     ,
					     recoBundles);
	} catch (Types::InternalProblems & e) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for internal problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (Types::CannotFetch & e) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for fetching problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (...) {
		err += "getCallDetails through CORBA Mgmt ";
		err += "interface failed for unknown reason";
		VTY_ERROR_AND_EXIT(vty, err);
	}

	call_info_t           cdata;
	recovery_info_t       rdata;
	lsp_info_t            ldata;

	memset(&cdata, 0, sizeof(cdata));
	memset(&rdata, 0, sizeof(rdata));
	memset(&ldata, 0, sizeof(ldata));

	memset(&cid,   0, sizeof(cid));	// reset cid

	try {
		cid << id;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callId translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	try {
		cdata << cinfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"callData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (!getTna(cid, cdata, 0 /* !isEgress*/ , err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	if (!getTna(cid, cdata, 1 /*  isEgress*/ , err)) {
		VTY_ERROR_AND_EXIT(vty, err);
	}

	vtyDump(vty, cid);
	vtyDump(vty, cdata);
	CORBAACTORDUMP2VTY(actor, "Call");
	CORBASTATESDUMP2VTY(states, "Call");

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Recovery Bundles under this CALL%s",
		VTY_NEWLINE);

	try {
		rdata << rinfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"recoveryData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	vtyDump(vty, rdata);

	size_t i;

	for (i = 0; i < recoBundles->length(); i++) {
		g2mpls_addr_t srcLsr;
		g2mpls_addr_t dstLsr;
		uint32_t      tid;

		try {
			srcLsr << (recoBundles)[i].srcAddr;
			dstLsr << (recoBundles)[i].dstAddr;
			tid = (recoBundles)[i].tunId;
		} catch (std::out_of_range & e) {
			vty_out(vty,
				"Exception occurred during "
				"types translation (%s)%s",
				e.what(), VTY_NEWLINE);
			return CMD_WARNING;
		}

		vty_out(vty,
			"   - #%02d: src-LSR = %s / "
			"dst-LSR = %s / tunnel-ID = 0x%08x%s",
			i,
			addr_ntoa(srcLsr),
			addr_ntoa(dstLsr),
			tid,
			VTY_NEWLINE);

		if (i == SHOWLIST_DEPTH) {
			break;
		}
	}

	try {
		ldata << linfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"recoveryData translation (%s)",
			e.what(), VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Common LSP data under this CALL%s",
		VTY_NEWLINE);
	vtyDump(vty, ldata);

#endif // HAVE_OMNIORB

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	return CMD_SUCCESS;
}

DEFUN(ncc_show_callvty,
      ncc_show_callvty_cmd,
      "show ncc call id WORD",
      "Show details for specified CALL")
{
	mapKey_t      key;

	key = strtoull(argv[0], NULL, 16);

	vtyCallDumpVty(vty, key);

	return CMD_SUCCESS;
}

DEFUN(ncc_show_reco_bundle,
      ncc_show_reco_bundle_cmd,
      "show ncc recovery-bundle "
      "tunnel-id WORD source-lsr A.B.C.D destination-lsr A.B.C.D",
      "Show details for a specified recovery bundle")
{
	g2mpls_addr_t srcLsr;
	g2mpls_addr_t dstLsr;
	uint32_t      tid;
	std::string   err;

	tid = strtoull(argv[0], NULL, 16);

	if (strchr (argv[1], '.') != NULL) {
		// matched "A.B.C.D".
		srcLsr.type       = IPv4;
		srcLsr.preflen    = 32;
		if (!inet_aton(argv[1], &(srcLsr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for src LSR%s",
				argv[1], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}  else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"source LSR%s",
			argv[1], VTY_NEWLINE);
		return CMD_WARNING;
	}

	if (strchr (argv[2], '.') != NULL) {
		// matched "A.B.C.D".
		dstLsr.type       = IPv4;
		dstLsr.preflen    = 32;
		if (!inet_aton(argv[2], &(dstLsr.value.ipv4))) {
			vty_out(vty,
				"Invalid IP address (%s) for dst LSR%s",
				argv[2], VTY_NEWLINE);
			return CMD_WARNING;
		}
	}  else {
		vty_out(vty,
			"Invalid IP address (%s) for "
			"destination LSR%s",
			argv[2], VTY_NEWLINE);
		return CMD_WARNING;
	}

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "Recovery Bundle%s", VTY_NEWLINE);

	if (NCC_MASTER.dryrun()) {
		vty_out(vty, "   ***EMPTY***%s", VTY_NEWLINE);
		vty_out(vty, "%s", VTY_NEWLINE);
		return CMD_SUCCESS;
	}
#if HAVE_OMNIORB
	NCC_DEBUG("Requesting details for Recovery Bundle with "
		  "tunnelID 0x%08x, src LSR %s and dst LSR %s %s",
		  key2string(tid).c_str(),
		  addr_ntoa(srcLsr),
		  addr_ntoa(dstLsr));

	RecoveryController::NorthBound_var  rcNBIProxy;

	rcNBIProxy = NCC_MASTER.getRcNBIProxy();

	if (CORBA::is_nil(rcNBIProxy)) {
		NCC_ERROR("Cannot get Corba rcNBIProxy client object "
			  "previously set up");
		assert(0);
	}

	g2mplsTypes::recoBundleIdent_var  ident;

	try {
		ident->srcAddr << srcLsr;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"srcLsr types translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	try {
		ident->dstAddr << dstLsr;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"dstLsr types translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	try {
		ident->tunId = tid;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"tunId types translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	g2mplsTypes::callIdent_var		cident;
	g2mplsTypes::recoveryParams_var		rinfo;
	g2mplsTypes::lspParams_var		linfo;
	g2mplsTypes::adminState			rstateAdmin;
	CORBA::String_var			rstateOp;

	//g2mplsTypes::statesBundle_var           rstates;
	//CallController::Mgmt::lspIdentSeq_var   lsps;

	try {
		rcNBIProxy->rbGetDetails(ident,
					 cident,
					 rinfo,
					 linfo,
					 rstateAdmin,
					 rstateOp);
	} catch (Types::InternalProblems & e) {
		err += "getRecoBundleDetails through CORBA Mgmt ";
		err += "interface failed for internal problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (Types::CannotFetch & e) {
		err += "getRecoBundleDetails through CORBA Mgmt ";
		err += "interface failed for fetching problems ";
		err += "(reason= ";
		err += e.what;
		err += ")";
		VTY_ERROR_AND_EXIT(vty, err);
	} catch (...) {
		err += "getRecoBundleDetails through CORBA Mgmt ";
		err += "interface failed for unknown reason";
		VTY_ERROR_AND_EXIT(vty, err);
	}

	call_ident_t  cid;
	reco_bundle_ident_t  rid;
	recovery_info_t      rdata;
	lsp_info_t           ldata;

	memset(&cid,   0, sizeof(cid));
	memset(&rid,   0, sizeof(rid));
	memset(&rdata, 0, sizeof(rdata));
	memset(&ldata, 0, sizeof(ldata));

	try {
		rid.tun_id   = ident->tunId;
		srcLsr      << ident->srcAddr;
		rid.src_lsr  = ntohl(srcLsr.value.ipv4.s_addr);
		dstLsr      << ident->dstAddr;
		rid.dst_lsr  = ntohl(dstLsr.value.ipv4.s_addr);
		cid         << cident;
		rdata       << rinfo;
		ldata       << linfo;
	} catch (std::out_of_range & e) {
		vty_out(vty,
			"Exception occurred during "
			"types translation (%s)",
			e.what(), VTY_NEWLINE);;
		return CMD_WARNING;
	}

	vtyDump(vty, rid);
	vtyDump(vty, rdata);
	vty_out(vty, "Recovery Bundle states%s", VTY_NEWLINE);

	vty_out(vty, "\t - op. state       : %s%s",
		rstateOp.in(), VTY_NEWLINE);
	vty_out(vty, "\t - adm. state      : %s%s",
		(rstateAdmin == g2mplsTypes::ADMINSTATE_DISABLED) ?
		"DISABLED" : "ENABLED", VTY_NEWLINE);
	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Parent Call%s",
		VTY_NEWLINE);
	vtyDump(vty, cid);
	vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
	vty_out(vty, "Common LSP data under this CALL%s",
		VTY_NEWLINE);
	vtyDump(vty, ldata);

#if 0
	vty_out(vty, "LSPs under this Recovery Bundle%s",
		VTY_NEWLINE);

	size_t i;

	for (i = 0; i < lsps->length(); i++) {
		lsp_ident_t lid;

		memset(&lid, 0, sizeof(lid));

		try {
			lid.src_nid   << (lsps)[i].srcNodeId;
			lid.dst_nid   << (lsps)[i].dstNodeId;
			lid.tun_id     = (lsps)[i].tunId;
			lid.ext_tun_id = (lsps)[i].extTid;
			lid.lsp_id     = (lsps)[i].lspId;
		} catch (std::out_of_range & e) {
			vty_out(vty,
				"Exception occurred during "
				"types translation (%s)%s",
				e.what(), VTY_NEWLINE);
			return CMD_WARNING;
		}

		vty_out(vty, "%s%s", BAR2, VTY_NEWLINE);
		vtyDump(vty, lid);

		if (i == SHOWLIST_DEPTH) {
			break;
		}
	}
#endif // 0

#endif // HAVE_OMNIORB

	vty_out(vty, "%s%s", BAR1, VTY_NEWLINE);
	vty_out(vty, "%s", VTY_NEWLINE);
	return CMD_SUCCESS;
}

ALIAS(ncc_show_reco_bundle,
      ncc_show_reco_bundle_compact_cmd,
      "show ncc rec-bdl tid WORD src A.B.C.D dst A.B.C.D",
      "Show details for a specified recovery bundle");

static int ncc_config_write (struct vty * vty)
{
	vty_out(vty, "ncc%s", VTY_NEWLINE);
	vty_out(vty, "!%s",   VTY_NEWLINE);

	return 1;
}

void NCC_Master::init_vty(void)
{
	// Install ncc top node.
	install_node (&ncc_node, ncc_config_write);

	install_element(CONFIG_NODE, &ncc_cmd);
	install_element(CONFIG_NODE, &no_ncc_cmd);

	install_default(NCC_NODE);

	// imperative commands
	install_element(NCC_NODE,    &ncc_call_create_cmd);
	install_element(NCC_NODE,    &ncc_call_create_compact_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_callres32_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_callres32_compact_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_callres60_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_callres60_compact_cmd);
	//install_element(NCC_NODE,    &ncc_call_set_info_calldnid_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_calltime_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_callnames_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_recovery_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_lspbase_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_lspbase_compact_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_lspconstr_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_lspproto_cmd);
	install_element(NCC_NODE,    &ncc_call_set_info_lspprototimers_cmd);
	install_element(NCC_NODE,    &ncc_call_commit_cmd);
	install_element(NCC_NODE,    &ncc_call_destroy_cmd);
	install_element(NCC_NODE,    &ncc_call_enable_cmd);
	install_element(NCC_NODE,    &ncc_call_disable_cmd);
	install_element(NCC_NODE,    &ncc_call_setup_cmd);
	install_element(NCC_NODE,    &ncc_call_setdown_cmd);


	//informative commands
	install_element(VIEW_NODE,   &ncc_show_cmd);
	install_element(ENABLE_NODE, &ncc_show_cmd);
	install_element(NCC_NODE,    &ncc_show_cmd);
	install_element(VIEW_NODE,   &ncc_show_calls_cmd);
	install_element(ENABLE_NODE, &ncc_show_calls_cmd);
	install_element(NCC_NODE,    &ncc_show_calls_cmd);
	install_element(VIEW_NODE,   &ncc_show_call_cmd);
	install_element(ENABLE_NODE, &ncc_show_call_cmd);
	install_element(NCC_NODE,    &ncc_show_call_cmd);
	install_element(VIEW_NODE,   &ncc_show_callvty_cmd);
	install_element(ENABLE_NODE, &ncc_show_callvty_cmd);
	install_element(NCC_NODE,    &ncc_show_callvty_cmd);
	install_element(VIEW_NODE,   &ncc_show_reco_bundle_cmd);
	install_element(ENABLE_NODE, &ncc_show_reco_bundle_cmd);
	install_element(NCC_NODE,    &ncc_show_reco_bundle_cmd);
	install_element(VIEW_NODE,   &ncc_show_reco_bundle_compact_cmd);
	install_element(ENABLE_NODE, &ncc_show_reco_bundle_compact_cmd);
	install_element(NCC_NODE,    &ncc_show_reco_bundle_compact_cmd);
}
