/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
a *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "stream.h"
#include "if.h"
#include "log.h"
#include "memory.h"
#include "sockopt.h"
#include "sockunion.h"
#include "hash.h"
#include "checksum.h"
#include "linklist.h"
#include "network.h"
#include "scngwc.h"
#include "scngws.h"

/* Queue containing unread packets from SCNGW server */
struct zlist *clients_list;

/* This function adds a new client_data structure in the client_list when a new client registers to SCNGW server*/
static void
add_client_data (int sock, int proto, int encap, int want_ack)
{
	struct client_data *new;

	new = malloc (sizeof(struct client_data));
	if (!new) {
		zlog_debug ("add_client_data: malloc() cannot allocate memory");
		exit(1);
	}
	bzero (new, sizeof(struct client_data));

	new->queue = stream_fifo_new ();
	new->fd = sock;
	new->proto = proto;
	new->encap = encap;
	new->want_ack = want_ack;
	listnode_add (clients_list, new);
}

/* This function retrieves a client_data structure to know who is the client sending/receiving messages */
static struct client_data *
fetch_client_data (int sock)
{
	struct client_data *cd;
	struct zlistnode *cd_node, *cd_nnode;

	for (ALL_LIST_ELEMENTS (clients_list, cd_node, cd_nnode, cd)) {
		if (cd->fd == sock) {
			return cd;
		}
	}
	return NULL;
}

/* This function receives the SCNGW header in a message from SCNGW server */
static int
scngwc_recv_hdr (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_hdr;
	int ret;
	struct timeval t_out;
	long int time = TIMEOUT_ACK_DEFUALT;

	t_out.tv_sec = time / 1000;
	t_out.tv_usec = 1000*(time % 1000);

	nbytes_hdr = recv (sock, buf->data, size, flags);
	if (nbytes_hdr < 0) {
		switch (errno) {
			case EAGAIN:
				zlog_debug ("scngwc_recv_hdr: Timeout expired");
				ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&t_out, sizeof (struct timeval));
				if (ret < 0) {
					zlog_warn ("scngwc_wait_server_response: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
				}
				break;
			default:
				zlog_debug ("scngwc_recv_hdr: error in executing recv: %s", safe_strerror(errno));
				break;
		}
	}
	return nbytes_hdr;
}

/* This function receives the body of a message from SCNGW server */
static int
scngwc_recv_msg (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_msg;
	int ret;
	struct timeval t_out;
	long int time = TIMEOUT_ACK_DEFUALT;

	t_out.tv_sec = time / 1000;
	t_out.tv_usec = 1000*(time % 1000);

	nbytes_msg = recv (sock, buf->data + SCNGW_HDR_SIZE, size, flags);
	if (nbytes_msg < 0) {
		switch (errno) {
			case EAGAIN:
				zlog_debug ("scngwc_recv_msg: Timeout expired");
				ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&t_out, sizeof (struct timeval));
				if (ret < 0) {
					zlog_warn ("scngwc_wait_server_response: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
				}
				break;
			default:
				zlog_debug ("scngwc_recv_msg: error in executing recv: %s", safe_strerror(errno));
				break;
		}
	}
	return nbytes_msg;
}

/* This function sends the SCNGW header in a message to SCNGW server */
static int
scngwc_send_hdr (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_hdr;

	nbytes_hdr = send (sock, buf->data, size, flags);
	if (nbytes_hdr < 0) {
		zlog_warn ("scngwc_send_hdr: send failed, %s", safe_strerror(errno));
	}
	else if (nbytes_hdr != (int)size) {
		zlog_debug ("scngwc_send_hdr: sent wrong number of bytes, %d instead of %d", nbytes_hdr, (int)size);
	}
	return nbytes_hdr;
}

/* This function sends the body of a message to SCNGW server */
static int
scngwc_send_msg (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_msg;

	nbytes_msg = send (sock, buf->data + SCNGW_HDR_SIZE, size, flags);
	if (nbytes_msg < 0) {
		zlog_warn ("scngwc_send_msg: send failed, %s", safe_strerror(errno));
	}
	else if (nbytes_msg != (int)size) {
		zlog_debug ("scngwc_send_msg: sent wrong number of bytes, %d instead of %d", nbytes_msg, (int)size);
	}
	return nbytes_msg;
}

/* This function adds the IP header to the SDU from the client */
static void
iphdr_add (struct stream *buff, struct in_addr src, struct in_addr dst, u_int16_t sdu_len, u_int8_t proto)
{
	struct ip iph;

	memset (&iph, 0, sizeof (struct ip));
	iph.ip_hl = sizeof (struct ip) >> 2;
	iph.ip_v = IPVERSION;
	iph.ip_tos = IPTOS_PREC_INTERNETCONTROL;
	iph.ip_len = (iph.ip_hl << 2) + sdu_len;
	iph.ip_off = 0;
	iph.ip_ttl = SCNGW_IP_TTL;
	iph.ip_p = proto;
	iph.ip_src.s_addr = src.s_addr;
	iph.ip_dst.s_addr = dst.s_addr;
	sockopt_iphdrincl_swab_htosys (&iph);
	iph.ip_sum = in_cksum (&iph, iph.ip_hl*4);

	stream_write (buff, (u_char *)&iph, sizeof (struct ip));
}

/* This function extracts the IP header from a packet from SCNGW server */
static struct ip *
iphdr_extract (struct stream *buff)
{
	struct ip *hdr;
	struct ip *ptr;

	hdr = malloc (sizeof(struct ip));
	if (!hdr) {
		zlog_debug ("iphdr_extract: malloc() cannot allocate memory");
		exit(1);
	}
	bzero (hdr, sizeof(struct ip));
	ptr = (struct ip *) STREAM_PNT (buff);
	sockopt_iphdrincl_swab_systoh (ptr);
	memcpy (hdr, ptr, sizeof(struct ip));

	buff->endp += (hdr->ip_hl << 2);

	return hdr;
}

/*This function is used to connect client to SCNGW server*/
static int
scngwc_socket_connect (int sock, client_type_t client)
{
	int ret;
	struct sockaddr_in serv;
	struct sockaddr_in local_addr;
	static int         count = 1;

	/* Define server address */
	memset (&serv, 0, sizeof (struct sockaddr_in));
	serv.sin_family = AF_INET;
	serv.sin_port = htons (SCNGWS_PORT);

#ifdef HAVE_SIN_LEN
	serv.sin_len = sizeof (struct sockaddr_in);
#endif /* HAVE_SIN_LEN */

	serv.sin_addr.s_addr = htonl (INADDR_LOOPBACK);

	/* Bind local socket to a port number */
	local_addr.sin_family      = AF_INET;
	local_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	switch (client) {
		//case IPPROTO_OSPFIGP:
		case CLIENT_G2OSPFTE:
			local_addr.sin_port = htons(OSPF_BASE_PORT + count);
			break;
		case CLIENT_G2RSVPTE:
			local_addr.sin_port = htons(G2RSVPTE_BASE_PORT + count);
			break;
		case CLIENT_GENNIRSVP:
			local_addr.sin_port = htons(GENNIRSVP_BASE_PORT + count);
			break;
		case CLIENT_GUNIRSVP:
			local_addr.sin_port = htons(GUNIRSVP_BASE_PORT + count);
			break;
		default:
			zlog_err("scngwc_socket_connect: client %d not supported by SCNGW",
				 client);
			return -1;
	}
	count++;

	ret = bind(sock, (struct sockaddr *) &local_addr, sizeof(local_addr));
	if (ret < 0) {
		zlog_debug("scngwc_socket_connect: Can't bind to local address: %s\n",
			   safe_strerror(errno));
		return ret;
	}
        /* Connect to scngws*/
	ret = connect (sock, (struct sockaddr *) &serv, sizeof (serv));
	if (ret < 0) {
		zlog_debug("scngwc_socket_connect: Can't connect to SCNGWs with fd %d: %s\n",
			   sock, safe_strerror(errno));
		return ret;
	}
	/*
	while (1) {
		ret = connect (sock, (struct sockaddr *) &serv, sizeof (serv));
		if (ret < 0) {
			sleep (1);
			printf("scngwc_socket_connect: Can't connect to SCNGWs with fd %d: %s\n",
			       sock, safe_strerror(errno));
			continue;
		}

		break;
	}
	*/
	zlog_debug ("scngwc_socket_connect: connection with SCNGW server done");
	return ret;
}

/*This function si used to disconnect client from SCNGW server*/
static int
scngwc_socket_disconnect (int sock)
{
	int ret;
	struct sockaddr_in serv;

	/* Define server address */
	memset (&serv, 0, sizeof (struct sockaddr_in));
	serv.sin_family = AF_UNSPEC;
	serv.sin_port = htons (SCNGWS_PORT);
#ifdef HAVE_SIN_LEN
	serv.sin_len = sizeof (struct sockaddr_in);
#endif /* HAVE_SIN_LEN */
	serv.sin_addr.s_addr = htonl (INADDR_LOOPBACK);

        /* Connect to scngws*/
	while (1) {
		ret = connect (sock, (struct sockaddr *) &serv, sizeof (serv));
		if (ret < 0) {
			zlog_debug ("scngwc_socket_disconnect: Can't disconnect to SCNGWs with fd %d: %s", sock, safe_strerror(errno));
		} else {
			break;
		}
		sleep (1);
	}
	return ret;
}

/*This function opens the client socket. It's a kind of wrap function of the original net socket opened by client */
static int
scngwc_socket (int domain, int type, int sock_proto, int client_proto, int want_ack, int interface_type)
{
	int sock, ret;
	int on = 1;
	struct timeval timeout;
	long int t_out = TIMEOUT_ACK_DEFUALT;
	int buf_size = MAX_PACKET_SIZE;
	size_t len = sizeof (int);
	client_type_t client_type;

	sock = socket (domain, type, sock_proto);
	if (sock < 0) {
		zlog_err ("scngwc_socket: Can't open socket to scngws: %s", safe_strerror(errno));
		return -1;
	}
	if (want_ack == WANT_ACK) {
		timeout.tv_sec = t_out / 1000;
		timeout.tv_usec = 1000*(t_out % 1000);
		ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&timeout, sizeof (struct timeval));
		if (ret < 0) {
			zlog_warn ("scngwc_socket: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
		}
	}
	ret = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (void *) &on, sizeof (on));
	if (ret < 0) {
		zlog_warn ("scngwc_socket: Can't set sockopt SO_REUSEADDR to socket %d", sock);
	}
	ret = setsockopt (sock, SOL_SOCKET, SO_SNDBUF, (char *)&buf_size, len);
	if (ret < 0) {
		zlog_debug ("scngwc_socket: cannot set SO_SNDBUF %s", strerror (errno));
	}
	ret = setsockopt (sock, SOL_SOCKET, SO_RCVBUF, (char *)&buf_size, len);
	if (ret < 0) {
		zlog_debug ("scngwc_socket: cannot set SO_RCVBUF %s", strerror (errno));
	}

	switch (client_proto) {
		case IPPROTO_OSPFIGP: {
			switch (interface_type) {
				case INNI:
					client_type = CLIENT_G2OSPFTE;
					break;
				case UNI:
					client_type = CLIENT_G2OSPFTE;
					break;
				case ENNI:
					client_type = CLIENT_G2OSPFTE;
					break;
				default:
					assert(0);
					break;
			}
		}
			break;
		case IPPROTO_RSVP: {
			switch (interface_type) {
				case INNI:
					client_type = CLIENT_G2RSVPTE;
					break;
				case UNI:
					client_type = CLIENT_GUNIRSVP;
					break;
				case ENNI:
					client_type = CLIENT_GENNIRSVP;
					break;
				default:
					assert(0);
					break;
			}
		}
			break;
		default:
			assert(0);
			break;
	}

	ret = scngwc_socket_connect (sock, client_type);
	if (ret < 0) {
		return ret;
	}

	return sock;
}

/* This function sends a registration message to SCNGW server from client, specifing all parameters */
static int
send_registration (int sock, int interface_type, int protocol, int encap, int want_ack)
{
	struct stream *buff;
	u_int32_t size;
	u_int32_t resp;
	int ret;

	size = REGISTRATION_MSG_SIZE;
	buff = stream_new (size);
	memset (buff->data, 0, size);
	stream_reset (buff);

	/* Create registration message */
	stream_putl (buff, REGISTRATION_MESSAGE);
	stream_putl (buff, (u_int32_t)interface_type);
	stream_putl (buff, (u_int32_t)protocol);
	stream_putl (buff, (u_int32_t)encap);
	stream_putl (buff, (u_int32_t)want_ack);

	ret = scngwc_send_hdr (sock, buff, size, 0);
	if (ret < 0) {
		stream_free (buff);
		return ret;
	}
	stream_reset (buff);
	/* wait an ack/nack from SCNGW server */
	ret = scngwc_recv_hdr (sock, buff, 4U, MSG_WAITALL);
	if (ret < 0) {
		stream_free (buff);
		return -1;
	}
	else if (ret == 0) {
		zlog_debug ("send_registration: server shutdown");
		stream_free (buff);
		return -1;
		}
	buff->endp += ret;
	resp = stream_getl (buff);
	switch (resp) {
		case ACK_MESSAGE:
			zlog_debug ("Registration with SCNGW server success");
			break;
		case NACK_MESSAGE:
			zlog_debug ("Registration with SCNGW server failed");
			break;
	}
	stream_free (buff);
	return resp;
}

/* This function is called (only once) by the client who wants to register witn SCNGW server */
int
scngwc_init (int interface_type, int protocol, int encap, int want_ack)
{
	int sock;
	int ret;
	static int n_client = 1;

	/* Open socket with SCNGW server */
	sock = scngwc_socket (AF_INET, SOCK_STREAM, 0, protocol, want_ack, interface_type);
	if (sock < 0) {
		return sock;
	}
        ret = send_registration (sock, interface_type, protocol, encap, want_ack);
	if (ret < 0) {
		zlog_debug ("scngwc_init: Cannot registrate with SCNGW server");
		return ret;
	}
	/* if it's first client, create list of clients */
	if (n_client == 1) {
		clients_list = list_new ();
	}
	/* add the client to the list */
	add_client_data (sock, protocol, encap, want_ack);

	n_client ++;
	return sock;
}

/* This function adds the SCNGW header to the packet that client sends*/
static int
scngwc_hdr_add (struct stream *s, int flags, struct in_addr src_addr, struct in_addr dst_addr, u_int32_t sdu_length, u_int32_t msg_id)
{
	u_int32_t cc = 0; /* ctrl channel */

	stream_putl (s, PACKET_MESSAGE);
	stream_putl (s, SCNGW_HDR_SIZE);
	stream_putl (s, sdu_length);
	stream_putl (s, msg_id);

	stream_putl (s, flags);
	stream_put_ipv4 (s, src_addr.s_addr);
	stream_put_ipv4 (s, dst_addr.s_addr);

	stream_putl (s, cc);

	return SCNGW_HDR_SIZE;

}

/* This function reads the SCNGW header of the message received from SCNGWs*/
static int
scngwc_hdr_read (struct stream *buf, struct scngw_hdr *hdr)
{
	hdr->msg_type = stream_getl (buf);
	hdr->hdr_len = stream_getl (buf);
	hdr->sdu_len = stream_getl (buf);
	hdr->msg_id = stream_getl (buf);

	hdr->flags = stream_getl (buf);
	hdr->src_addr = stream_get_ipv4 (buf);
	hdr->dst_addr = stream_get_ipv4 (buf);

	hdr->cc = stream_getl (buf);

	return 0;
}

/*This function waits the server ack/nack for a client packet (only if the client specify WANT_ACK) */
static int
scngwc_wait_server_response (int sock, u_int32_t msg_id_sent, int *ack, int *unread, struct client_data *cd)
{
	struct stream *buff;
	u_int32_t val;
	u_int32_t msg_type;
	size_t size;
	struct timeval tv1;
	struct timeval tv2;
	struct timeval diff;
	struct timeval t_out;
	struct timeval new_t_out;
	long int time = TIMEOUT_ACK_DEFUALT;
	int ret;
	struct scngw_hdr hdr;

	t_out.tv_sec = time / 1000;
	t_out.tv_usec = 1000*(time % 1000);

	size = MAX_PACKET_SIZE;

	/* Block untill SCNGW server will answer with the result of sending packet in the network */
	while (1) {
		buff = stream_new (size);
		stream_reset (buff);
		memset (buff->data, 0, buff->size);
		gettimeofday (&tv1, NULL);
		ret = scngwc_recv_hdr (sock, buff, SCNGW_HDR_SIZE, MSG_WAITALL);
		if (ret < 0) {
			stream_free (buff);
			return 0;
			break;
		}
		else if (ret == 0) {
			zlog_debug ("scngwc_wait_server_response: server shutdown");
			/*
			scngwc_socket_disconnect (sock);
			zlog_debug ("Trying to connect to SCNGW server again...");
			scngwc_socket_connect (sock);
			*/
			stream_free (buff);
			return 0;
		}
		buff->endp += ret;
		memset (&hdr, 0, sizeof (struct scngw_hdr));
		scngwc_hdr_read (buff, &hdr);
		ret = scngwc_recv_msg (sock, buff, hdr.sdu_len, MSG_WAITALL);
		if (ret == 0) {
			zlog_debug ("scngwc_wait_server_response: server shutdown");
			/*
			scngwc_socket_disconnect (sock);
			zlog_debug ("Trying to connect to SCNGW server again...");
			scngwc_socket_connect (sock);
			*/
			stream_free (buff);
			return 0;
		}
		else if (ret < 0) {
			stream_free (buff);
			return -1;
		}
		buff->endp += ret;
		/* reset get pointer */
		stream_set_getp (buff, 0);
		/* Read the incoming message type */
		msg_type = stream_getl (buff);
		switch (msg_type) {
			case PACKET_MESSAGE:
				/* reset get pointer */
				stream_set_getp (buff, 0);
				if (cd->queue->count < PACKETS_QUEUE_SIZE) {
					stream_fifo_push (cd->queue, buff);
					zlog_debug ("Queued a PACKET_MESSAGE from SCNGW in packets_queue: %d packets queued.", cd->queue->count);
				}
				else {
					zlog_debug ("Try to queue a PACKET_MESSAGE from SCNGW but packets_queue is empty. Packet discarded.");
					stream_free (buff);
				}
				buff = NULL;
				*unread = 1;
				break;
			case ACK_MESSAGE:
				*ack = 1;
				if (hdr.msg_id != msg_id_sent) {
					zlog_debug ("scngwc_wait_server_response: receveid an ACK for a non coherent Mesagge ID");
					stream_free (buff);
					break;
				}
				val = stream_getl (buff);
				stream_free (buff);
				ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&t_out, sizeof (struct timeval));
				if (ret < 0) {
					zlog_warn ("scngwc_wait_server_response: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
				}
				return (int)val;
			case NACK_MESSAGE:
				*ack = 0;
				if (hdr.msg_id != msg_id_sent) {
					zlog_debug ("scngwc_wait_server_response: receveid a NACK for a non coherent Mesagge ID");
					stream_free (buff);
					break;
				}
				val = stream_getl (buff);
				stream_free (buff);
				ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&t_out, sizeof (struct timeval));
				if (ret < 0) {
					zlog_warn ("scngwc_wait_server_response: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
				}
				return -(int)val;
		}
		gettimeofday (&tv2, NULL);
		timersub (&tv2, &tv1, &diff);
		timersub (&t_out, &diff, &new_t_out);
		ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (const char *)&new_t_out, sizeof (struct timeval));
		if (ret < 0) {
			zlog_warn ("scngwc_wait_server_response: Can't set sockopt SO_RCVTIMEO to socket %d", sock);
		}
	}
}

/* This function sends the client packet (with the SCNGW header added) to SCNGWs*/
int
scngwc_sendmsg (int sock, const void *sdu, u_int16_t sdu_size, struct in_addr src_addr, struct in_addr dst_addr, int flags, int *unread_pckts)
{
	struct stream *buff;
	size_t size, pckt_size, msg_size, ret;
	static u_int32_t msg_id = 1;
	int ack = -1;
	struct client_data *cd;

	cd = fetch_client_data (sock);
	if (cd == NULL) {
		zlog_debug("scngwc_sendmsg: no client retrieved");
		return -1;
	}

	pckt_size = sdu_size + sizeof (struct ip);
	msg_size = pckt_size + SCNGW_HDR_SIZE +1;
	buff = stream_new (msg_size);
	stream_reset (buff);
	memset (buff->data, 0, buff->size);

	if (cd->encap == TUNNEL) {
		size = pckt_size;
		scngwc_hdr_add (buff, flags, src_addr, dst_addr, size, msg_id);
		iphdr_add (buff, src_addr, dst_addr, sdu_size, (u_int8_t)cd->proto);
	}
	else {
		size = sdu_size;
		scngwc_hdr_add (buff, flags, src_addr, dst_addr, size, msg_id);
	}

	memcpy (buff->data + buff->endp, sdu, sdu_size);
	buff->endp += sdu_size;

	ret = scngwc_send_hdr (sock, buff, SCNGW_HDR_SIZE, 0);
	if (ret < 0) {
		stream_free (buff);
		return ret;
	}
	ret = scngwc_send_msg (sock, buff, size, 0);
	if (ret < 0) {
		stream_free (buff);
		return ret;
	}
	if (cd->want_ack == WANT_ACK) {
		zlog_debug ("Waiting a response from SCNGW server for packet with ID = %d", msg_id);

		ret = scngwc_wait_server_response (sock, msg_id, &ack, unread_pckts, cd);
		if (ack == 1)
			zlog_debug ("SCNGW server successfully sent the packet with ID = %d", msg_id);
		else if (ack == 0)
			zlog_warn ("SCNGW server didn't send the packet with ID = %d", msg_id);
	}
	msg_id ++;
	stream_free (buff);

	return ret;
}

/* This function receives the message from SCNGWs containing a client packet (plus SCNGW header) when WANT_ACK is specified by client*/
static int
scngwc_stream_recvmsg_ack (void *sdu, int sock, struct ip **iph, size_t size, struct client_data *cd)
{
	int ret;
	size_t newsize = size + SCNGW_HDR_SIZE;
	size_t sdu_size, offset;
	struct stream *buf, *pckt;
	struct scngw_hdr scnhdr;
	u_int32_t msg_type;
	fd_set read_fds;
	struct timeval timeout;

	buf = stream_new (newsize);
	stream_reset (buf);
        memset (buf->data, 0, buf->size);
	/* First try to read from the socket */
	bzero (&timeout, sizeof (struct timeval));
	FD_SET (sock, &read_fds);
	ret = select (sock +1, &read_fds, NULL, NULL, &timeout);
	if (FD_ISSET (sock, &read_fds)) {
		ret = scngwc_recv_hdr (sock, buf, SCNGW_HDR_SIZE, MSG_WAITALL);
		if (ret == 0) {
			zlog_debug ("scngwc_stream_recvmsg: server shutdown");
			/*
			scngwc_socket_disconnect (sock);
			zlog_debug ("Trying to connect to SCNGW server again...");
			scngwc_socket_connect (sock);
			*/
			stream_free (buf);
			return 0;
		}
		else if (ret < 0) {
			stream_free (buf);
			return -1;
		}
		buf->endp += ret;
		memset (&scnhdr, 0, sizeof (struct scngw_hdr));
		scngwc_hdr_read (buf, &scnhdr);
		ret = scngwc_recv_msg (sock, buf, scnhdr.sdu_len, MSG_WAITALL);
		if (ret == 0) {
			zlog_debug ("scngwc_stream_recvmsg: server shutdown");
			/*
			scngwc_socket_disconnect (sock);
			zlog_debug ("Trying to connect to SCNGW server again...");
			scngwc_socket_connect (sock);
			*/
			stream_free (buf);
			return 0;
		}
		else if (ret < 0) {
			stream_free (buf);
			return -1;
		}
		buf->endp += ret;
		/* reset get pointer */
		stream_set_getp (buf, 0);
		/* Verify the message type */
		msg_type = stream_getl (buf);
		if (msg_type != PACKET_MESSAGE) {
			zlog_debug ("Received and deleted an outdated ACK message from SCNGW server.");
			stream_free (buf);
			return -1;
		}
		else {
			/* reset get pointer */
			stream_set_getp (buf, 0);
			if (cd->queue->count < PACKETS_QUEUE_SIZE) {
				stream_fifo_push (cd->queue, buf);
				zlog_debug ("Queued a PACKET_MESSAGE from SCNGW in packets queue: %d packets queued.", cd->queue->count);
			}
			else {
				zlog_debug ("Try to queue a PACKET_MESSAGE from SCNGW but packets queue is empty. Packet discarded.");
				stream_free (buf);
			}
		}
	}
	else {
		zlog_debug ("scngwc_stream_recvmsg: there are no packets in the socket.");
		stream_free (buf);
	}
	/* Then read from the FIFO packets_queue, only the first packet */
	zlog_debug ("Extracting first packet from unread packets queue");
	pckt = stream_fifo_pop (cd->queue);
	zlog_debug ("There are still %d packets queued.", cd->queue->count);
	memset (&scnhdr, 0, sizeof (struct scngw_hdr));
	/* read scngw header */
	scngwc_hdr_read (pckt, &scnhdr);

	if (cd->encap == TUNNEL) {
		*iph = iphdr_extract (pckt);
		offset = scnhdr.hdr_len + ((*iph)->ip_hl << 2);
		sdu_size = (*iph)->ip_len - ((*iph)->ip_hl << 2);
	}
	else {
		*iph = NULL;
		offset = scnhdr.hdr_len;
		sdu_size = scnhdr.sdu_len;
	}

	memcpy (sdu, pckt->data + offset, sdu_size);
	//s->endp += sdu_size;

	stream_free (pckt);

	return sdu_size;
}

/* This function receives the message from SCNGWs containing a client packet (plus SCNGW header) when WANT_NO_ACK is specified by client*/
static int
scngwc_stream_recvmsg_noack (void *sdu, int sock, struct ip **iph, size_t size, struct client_data *cd)
{
	int ret;
	size_t newsize = size + SCNGW_HDR_SIZE;
	size_t sdu_size, offset;
	struct stream *buf;
	struct scngw_hdr scnhdr;
	u_int32_t msg_type;

	buf = stream_new (newsize);
	stream_reset (buf);
        memset (buf->data, 0, buf->size);

	ret = scngwc_recv_hdr (sock, buf, SCNGW_HDR_SIZE, MSG_WAITALL);
	if (ret == 0) {
		zlog_debug ("scngwc_stream_recvmsg: server shutdown");
		/*
		scngwc_socket_disconnect (sock);
		zlog_debug ("Trying to connect to SCNGW server again...");
		scngwc_socket_connect (sock);
		*/
		stream_free (buf);
		return 0;
	}
	else if (ret < 0) {
		stream_free (buf);
		return -1;
	}
	buf->endp += ret;
	memset (&scnhdr, 0, sizeof (struct scngw_hdr));
	scngwc_hdr_read (buf, &scnhdr);
	ret = scngwc_recv_msg (sock, buf, scnhdr.sdu_len, MSG_WAITALL);
	if (ret == 0) {
		zlog_debug ("scngwc_stream_recvmsg: server shutdown");
		/*
		scngwc_socket_disconnect (sock);
		zlog_debug ("Trying to connect to SCNGW server again...");
		scngwc_socket_connect (sock);
		*/
		stream_free (buf);
		return 0;
	}
	else if (ret < 0) {
		stream_free (buf);
		return -1;
	}
	buf->endp += ret;
	/* reset get pointer */
	stream_set_getp (buf, 0);
	/* Verify the message type */
	msg_type = stream_getl (buf);
	if (msg_type != PACKET_MESSAGE) {
		zlog_debug ("Received wrong message type from SCNGW server.");
		zlog_debug ("Packet discarded.");
		stream_free (buf);
		return -1;
	}
	/* reset get pointer */
	stream_set_getp (buf, 0);
/* 	if (cd->encap == TUNNEL) { */
/* 		stream_forward_getp (buf, SCNGW_HDR_SIZE); */
/* 		*iph = iphdr_extract (buf); */
/* 		offset = scnhdr.hdr_len + ((*iph)->ip_hl << 2); */
/* 		sdu_size = (*iph)->ip_len - ((*iph)->ip_hl << 2); */
/* 	} */
/* 	else { */
/* 		*iph = NULL; */
/* 		offset = scnhdr.hdr_len; */
/* 		sdu_size = scnhdr.sdu_len; */
/* 	} */

	stream_forward_getp (buf, SCNGW_HDR_SIZE);
	*iph = iphdr_extract (buf);
	offset = scnhdr.hdr_len + ((*iph)->ip_hl << 2);
	sdu_size = (*iph)->ip_len - ((*iph)->ip_hl << 2);

	memcpy (sdu, buf->data + offset, sdu_size);
	//s->endp += sdu_size;

	stream_free (buf);

	return sdu_size;

}

int
scngwc_stream_recvmsg (void *sdu, int sock, struct ip **iph, size_t size)
{
	struct client_data *cd;

	cd = fetch_client_data (sock);
	if (cd == NULL) {
		zlog_debug("scngwc_stream_recvmsg: no client retrieved");
		return -1;
	}
	if (cd->want_ack == WANT_ACK) {
		return scngwc_stream_recvmsg_ack (sdu, sock, iph, size, cd);
	}
	else {
		return scngwc_stream_recvmsg_noack (sdu, sock, iph, size, cd);
	}
}

/*Function for socket closing*/
void
scngwc_close (int sock)
{
	close (sock);
	return;
}
