/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "scngwc.h"
#include "scngws.h"
#include "scngws_packet.h"
#include "scngws_lrm.h"
#include "log.h"
#include "thread.h"
#include "network.h"
#include "memory.h"
#include "zclient.h"
#include "stream.h"

/* Zebra structure to hold current status. */
struct zclient *zclient = NULL;

/* SCNGWs process wide configuration pointer. */
extern struct scn_master *scn_m;

/* SCNGWs master of the threads. */
extern struct thread_master *master;

/* SCNGWs tel_cc and cc_scnif associations lists to export */
/*struct zlist *tel_cc_list, *tmp_tel_cc_list; */
/*struct zlist *cc_scnif_list, *tmp_cc_scnif_list; */
/*struct zlist *scnif_list, *tmp_scnif_list; */
static struct zlist *tel_cc_list, *tmp_tel_cc_list;
static struct zlist *cc_scnif_list, *tmp_cc_scnif_list;
static struct zlist *scnif_list, *tmp_scnif_list;

u_int32_t n_scnif, n_assoc_tel, n_assoc_scnif;

/* This function print the correct interface type */
const char *
if_type_string(u_int16_t interface_type)
{
	switch (interface_type) {
		case SCN_INNI:
			return "INNI";
			break;
		case SCN_ENNI:
			return "ENNI";
			break;
		case SCN_UNI:
			return "UNI";
			break;
	}
	return "";
}

struct zlist *
get_tel_cc_list (void)
{
	return tel_cc_list;
}

struct zlist *
get_cc_scnif_list (void)
{
	return cc_scnif_list;
}

struct zlist *
get_scnif_list (void)
{
	return scnif_list;
}

struct zlist *
get_tmp_tel_cc_list (void)
{
	return tmp_tel_cc_list;
}

struct zlist *
get_tmp_cc_scnif_list (void)
{
	return tmp_cc_scnif_list;
}

struct zlist *
get_tmp_scnif_list (void)
{
	return tmp_scnif_list;
}

/* This function initializes the structures containing the lists of TE-link/CC/SCN-if associations */
void
tel_cc_scnif_association_init (void)
{
	scnif_list = list_new();
	tel_cc_list = list_new();
	cc_scnif_list = list_new();
}

/* This function initializes the temporary structures containing the lists of TE-link/CC/SCN-if associations */
void
tmp_tel_cc_scnif_association_init (void)
{
	tmp_scnif_list = list_new();
	tmp_tel_cc_list = list_new();
	tmp_cc_scnif_list = list_new();

}

/* This function deletes temporary association when a synchronization successes or fails */
void
delete_tmp_associations (void)
{
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct in_addr *scnif;
	struct zlistnode *scnif_node, *scnif_nnode;

	for (ALL_LIST_ELEMENTS (tmp_scnif_list, scnif_node, scnif_nnode, scnif))
		free (scnif);
	for (ALL_LIST_ELEMENTS (tmp_tel_cc_list, telcc_node, telcc_nnode, tel_cc))
		free (tel_cc);
	for (ALL_LIST_ELEMENTS (tmp_cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif))
		free (cc_scnif);

	list_delete_all_node (tmp_scnif_list);
	list_delete_all_node (tmp_tel_cc_list);
	list_delete_all_node (tmp_cc_scnif_list);
}

/* This function allocs a new scnif structure */
struct scnif *
scnif_new (void)
{
	struct scnif *new;

	new = malloc (sizeof(struct scnif));
	if (!new) {
		zlog_debug ("scnif_new: malloc(void) cannot allocate memory");
		exit(1);
	}
	bzero (new, sizeof(struct scnif));

	return new;
}

/* This function allocs a new tel_cc structure */
struct tel_cc_assoc *
tel_cc_new (void)
{
	struct tel_cc_assoc *new;

	new = malloc (sizeof(struct tel_cc_assoc));
	if (!new) {
		zlog_debug ("tel_cc_new: malloc(void) cannot allocate memory");
		exit(1);
	}
	bzero (new, sizeof(struct tel_cc_assoc));
	new->cclist = list_new();

	return new;
}

/* This function allocs a new cc_scnif structure */
struct cc_scnif_assoc *
cc_scnif_new (void)
{
	struct cc_scnif_assoc *new;

	new = malloc (sizeof(struct cc_scnif_assoc));
	if (!new) {
		zlog_debug ("cc_scnif_new: malloc(void) cannot allocate memory");
		exit(1);
	}
	bzero (new, sizeof(struct cc_scnif_assoc));

	return new;
}

#if !HAVE_OMNIORB
/* This is the callback function called when LRM-SCNGW communication goes down*/
int
scngws_lrm_connection_down (struct zclient *zclient)
{
	struct scnif *scn_if;
	struct zlistnode *scnif_node, *scnif_nnode;
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;

	scn_m->lrm_conn_status = STATUS_DOWN;

	for (ALL_LIST_ELEMENTS (scnif_list, scnif_node, scnif_nnode, scn_if))
		scn_if->status = STATUS_DOWN;
	for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, tel_cc))
		tel_cc->status = STATUS_DOWN;
	for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif))
		cc_scnif->status = STATUS_DOWN;

	n_scnif = 0;
	n_assoc_tel = 0;
	n_assoc_scnif = 0;

	return 0;
}
#endif /* !HAVE_OMNIORB */

/* This function check if a CC is already present in the correspondent list */
int
check_cc (u_int32_t cc_id)
{
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *match;
	struct zlist *assoc_list = NULL;

	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			assoc_list =  cc_scnif_list;
			break;
		case STATUS_SYNC:
			assoc_list =  tmp_cc_scnif_list;
			break;
	}
	for (ALL_LIST_ELEMENTS (assoc_list, ccscnif_node, ccscnif_nnode, match)) {
		if (match->cc_id == cc_id) {
			return 1;
		}
	}

	return 0;
}

#if !HAVE_OMNIORB
/* This function reads the list of CC associated with a TE-link in the message from LRM */
void
read_cc_list (struct tel_cc_assoc *tel_cc, struct stream *buf)
{
	u_int32_t i, n_cc, *cc_id;

	n_cc = stream_getl(buf);
	for (i = 1; i <= n_cc; i++) {
		cc_id = malloc(sizeof(u_int32_t));
		*cc_id = stream_getl(buf);
		if (check_cc (*cc_id))
			listnode_add (tel_cc->cclist, cc_id);
		else
			zlog_debug ("\t CC %u is not present in the list of CC/SCN-if associations", *cc_id);
	}
}

/* This function reads one SCN-if addition in the message from LRM */
struct scnif *
read_scnif_add (struct stream *buf)
{
	struct scnif *scn_if;
	struct zlistnode *scnif_node, *scnif_nnode;
	struct scnif *match;
	u_int32_t already = 0;

	scn_if = scnif_new ();

	scn_if->loc_addr.s_addr = stream_get_ipv4(buf);

	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			for (ALL_LIST_ELEMENTS (scnif_list, scnif_node, scnif_nnode, match)) {
				if (scn_if->loc_addr.s_addr == match->loc_addr.s_addr) {
					zlog_debug("\t SCN-if %s is already present in SCNGW", inet_ntoa(scn_if->loc_addr));
					already = 1;
					break;
				}
			}
			break;
		case STATUS_SYNC:
			n_scnif --;
			break;
	}
	if (already) {
		free (scn_if);
		return NULL;
	}
	scn_if->status = STATUS_UP;
	return scn_if;
}
#endif /* !HAVE_OMNIORB */

/* This function deletes a CC from all the associated CC's of all TE_links */
int
delete_cc_from_tel (u_int32_t id)
{
	struct zlistnode *telcc_node, *telcc_nnode;
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *cc_node, *cc_nnode;
        u_int32_t *match;
	int ret = 0;

	for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, tel_cc)) {
		for (ALL_LIST_ELEMENTS (tel_cc->cclist, cc_node, cc_nnode, match)) {
			if (*((u_int32_t *)match) == id) {
				listnode_delete (tel_cc->cclist, match);
				free (match);
				ret ++;
				break;
			}
		}
	}
	return ret;
}

/* This function deletes all CC's which have the specified SCN-if as local address */
int
delete_scnif_from_cc (struct in_addr addr)
{
	struct zlistnode *cc_scnif_node, *cc_scnif_nnode;
	struct cc_scnif_assoc *cc_scnif;
	int ret = 0;

	for (ALL_LIST_ELEMENTS (cc_scnif_list, cc_scnif_node, cc_scnif_nnode, cc_scnif)) {
		if (cc_scnif->scnif_loc.s_addr == addr.s_addr) {
			delete_cc_from_tel (cc_scnif->cc_id);
			listnode_delete (cc_scnif_list, cc_scnif);
			free (cc_scnif);
			ret ++;
		}
	}

	return ret;
}

#if !HAVE_OMNIORB
/* This function reads one SCN-if deletion in the message from LRM */
int
read_scnif_delete (struct stream *buf)
{
	u_int32_t found = 0;
	struct in_addr addr;
	int count;
	struct zlistnode *scnif_node, *scnif_nnode;
	struct scnif *match;

	addr.s_addr = stream_get_ipv4 (buf);
	for (ALL_LIST_ELEMENTS (scnif_list, scnif_node, scnif_nnode, match)) {
		if (match->loc_addr.s_addr == addr.s_addr) {
			found = 1;
			zlog_debug ("\t Deleted a SCN-if");
			zlog_debug ("\t\t address: %s", inet_ntoa(match->loc_addr));
			listnode_delete (scnif_list, match);
			free(match);
			break;
		}
	}
	if (found == 0) {
		zlog_debug("\t SCN-if %s is not present in SCNGW and cannot be deleted", inet_ntoa(addr));
		return 0;
	}
	count = delete_scnif_from_cc (addr);
	if (count) {
		zlog_debug ("\t\t This SCN-if was associated with %d CCs", count);
		zlog_debug ("\t\t Deleted these CCs from correspondent TE-links");
	}

	return 0;
}

/* This function reads one TE-link/CC association addition in the message from LRM*/
struct tel_cc_assoc *
read_tel_cc_assoc_add (struct stream *buf)
{
	u_int32_t key, already = 0;
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct tel_cc_assoc *match;

	key = stream_getl (buf);

	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, match)) {
				if (match->key == key) {
					zlog_debug("\t TE-link with key %u is already present in SCNGW", key);
					already = 1;
					break;
				}
			}
			break;
		case STATUS_SYNC:
			n_assoc_tel --;
			break;
	}
	tel_cc = tel_cc_new ();

	tel_cc->key = key;
	tel_cc->interface_type = stream_getw(buf);
	tel_cc->tel_loc.s_addr = stream_get_ipv4(buf);
	tel_cc->tel_rem.s_addr = stream_get_ipv4(buf);
	read_cc_list (tel_cc, buf);

	if (already) {
		free (tel_cc);
		return NULL;
	}

	tel_cc->status = STATUS_UP;
	return tel_cc;
}

/* This function reads one TE-link/CC association deletion in the message from LRM */
int
read_tel_cc_assoc_delete (struct stream *buf)
{
	u_int32_t key, found = 0;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct tel_cc_assoc *match;

	key = stream_getl (buf);
	for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, match)) {
		if (match->key == key) {
			found = 1;
			zlog_debug ("\t Deleted a  TE-link/CC association");
			zlog_debug ("\t\t TE-link loc: %s", inet_ntoa(match->tel_loc));
			zlog_debug ("\t\t TE-link rem: %s", inet_ntoa(match->tel_rem));
			listnode_delete (tel_cc_list, match);
			free (match);
			break;
		}
	}
	if (found == 0)
		zlog_debug("\t TE-link with key %u is not present in SCNGW and cannot be deleted", key);

	return 0;
}

/* This function reads one TE-link/CC association update in the message from LRM */
int
read_tel_cc_assoc_update (struct stream *buf, struct tel_cc_assoc *tel_cc)
{
	u_int32_t op, found = 0, *cc_id;
	struct zlistnode *cc_node, *cc_nnode;
        u_int32_t *match;

	cc_id = malloc(sizeof(u_int32_t));

	op = stream_getl (buf);
	*cc_id = stream_getl(buf);
	if (!check_cc (*cc_id)) {
		zlog_debug ("\t CC %u is not present in the list of CC/SCN-if associations", *cc_id);
		return 0;
	}
	switch (op) {
		case ADD_OP:
			for (ALL_LIST_ELEMENTS (tel_cc->cclist, cc_node, cc_nnode, match)) {
				if (*((u_int32_t *)match) == *cc_id) {
					zlog_debug ("\t CC %u is is already associated at TE-link %s", *cc_id, inet_ntoa(tel_cc->tel_loc));
					return 0;
				}
			}
			listnode_add (tel_cc->cclist, cc_id);
			zlog_debug ("\t Added a CC (id = %u) to the list of associated CC of TE-link %s", *cc_id, inet_ntoa(tel_cc->tel_loc));
			break;
		case DEL_OP:
			for (ALL_LIST_ELEMENTS (tel_cc->cclist, cc_node, cc_nnode, match)) {
				if (*((u_int32_t *)match) == *cc_id) {
					found = 1;
					zlog_debug ("\t Deleted an associated CC (id = %u) from TE-link %s", *cc_id, inet_ntoa(tel_cc->tel_loc));
					listnode_delete (tel_cc->cclist, match);
					free(match);
					break;
				}
			}
			if (found == 0)
				zlog_debug("\t CC %u is not associated at TE-link %s and cannot be deleted", *cc_id, inet_ntoa(tel_cc->tel_loc));

			free(cc_id);
			break;
	}
	return 0;
}

/* This function reads one CC/SCN-if association addition in the message from LRM */
struct cc_scnif_assoc *
read_cc_scnif_assoc_add (struct stream *buf)
{
	u_int32_t ccid, already = 0;
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *match;

	ccid = stream_getl(buf);

	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, match)) {
				if (match->cc_id == ccid) {
					zlog_debug("\t CC with id %u is already present in SCNGW", ccid);
					already = 1;
					break;
				}
			}
			break;
		case STATUS_SYNC:
			n_assoc_scnif --;
			break;
	}
	cc_scnif = cc_scnif_new ();

	cc_scnif->cc_id = ccid;
	cc_scnif->scnif_loc.s_addr = stream_get_ipv4(buf);
	cc_scnif->scnif_rem.s_addr = stream_get_ipv4(buf);

	if (already) {
		free (cc_scnif);
		return NULL;
	}

	cc_scnif->status = STATUS_UP;
	return cc_scnif;
}

/* This function reads one CC/SCN-if association deletion in the message from LRM */
int
read_cc_scnif_assoc_delete (struct stream *buf)
{
	u_int32_t cc_id, found = 0;
	int count;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *match;

	cc_id = stream_getl (buf);
	for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, match)) {
		if (match->cc_id == cc_id) {
			found = 1;
			zlog_debug ("\t Deleted a CC/SCN-if association");
			zlog_debug ("\t\t CCid: %u", match->cc_id);
			zlog_debug ("\t\t\t SCN-if loc: %s", inet_ntoa(match->scnif_loc));
			zlog_debug ("\t\t\t SCN-if rem: %s", inet_ntoa(match->scnif_rem));
			listnode_delete (cc_scnif_list, match);
			free(match);
			break;
		}
	}
	if (found == 0) {
		zlog_debug("\t CC with id %u is not present in SCNGW and cannot be deleted", cc_id);
		return 0;
	}
	count = delete_cc_from_tel (cc_id);
	if (count)
		zlog_debug ("\t\t This CC was associated with %d TE-links", count);

	return 0;
}

/* This function reads one CC/SCN-if association update in the message from LRM */
int
read_cc_scnif_assoc_update (struct stream *buf, struct cc_scnif_assoc *cc_scnif)
{
	cc_scnif->scnif_loc.s_addr = stream_get_ipv4 (buf);
	cc_scnif->scnif_rem.s_addr = stream_get_ipv4 (buf);

	zlog_debug ("\t Updated a CC/SCN-if association (CCid = %u)", cc_scnif->cc_id);
	zlog_debug ("\t\t SCN-if loc: %s", inet_ntoa(cc_scnif->scnif_loc));
	zlog_debug ("\t\t SCN-if rem: %s", inet_ntoa(cc_scnif->scnif_rem));

	return 0;
}

/* This function is called when the timeout expires before an END_SYNC message arrives from LRM */
int
scngws_timer_expired (struct thread *thread)
{
	struct zclient *zclient;

	zclient = THREAD_ARG (thread);

	zlog_debug ("Timeout expired waiting END_SYNC message from LRM. Synchronization must be restarted.");

	delete_tmp_associations ();
	scngws_lrm_connection_down (zclient);

	stream_reset (zclient->obuf);
	/* Send SCNGW_SYNC_ERR command to LRM */
	zclient_create_header (zclient->obuf, SCNGW_SYNC_ERR);

	zclient_send_message (zclient);

	return 0;
}

/* This is the callback function called when LRM sends an LRM_BEGIN_SYNC message*/
static int
scngws_begin_sync (int command, struct zclient *zclient, zebra_size_t length)
{
	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			zlog_debug ("Connection with LRM is up. Expected another kind of command.");
			return SCNGW_UNACK;
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command.");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		case STATUS_DOWN:
			zlog_debug ("The communication with LRM is again available");
			zlog_debug ("Synchronizing the TE-link/CC/SCN-if associations");
			scn_m->lrm_conn_status = STATUS_SYNC;
			break;
		case STATUS_NEVER_CONN:
			zlog_debug ("Synchronizing the TE-link/CC/SCN-if associations");
			scn_m->lrm_conn_status = STATUS_SYNC;
			tel_cc_scnif_association_init ();
			tmp_tel_cc_scnif_association_init ();
			break;
	}
	/* Set timeout waiting END_SYNC from LRM */
	scn_m->timer_thread = thread_add_timer_msec (master, scngws_timer_expired, zclient, scn_m->timeout_lrmsync);
	/* Get the number of TE-link/CC and CC/SCn-if associatons*/
	n_scnif = stream_getl (zclient->ibuf);
	n_assoc_scnif = stream_getl (zclient->ibuf);
	n_assoc_tel = stream_getl (zclient->ibuf);
	zlog_debug ("\t number of SCN-if: %u", n_scnif);
	zlog_debug ("\t number of CC: %u", n_assoc_scnif);
	zlog_debug ("\t number of TE-link: %u", n_assoc_tel);

	return SCNGW_ACK_BEGIN;
}
#endif /* !HAVE_OMNIORB */

/* This function remove old associations after synchronization phase in the LRM-SCNGW communication */
void
delete_old_associations (void)
{
	struct scnif *scn_if;
	struct zlistnode *scnif_node, *scnif_nnode;
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;

	for (ALL_LIST_ELEMENTS (scnif_list, scnif_node, scnif_nnode, scn_if)) {
		if (scn_if->status == STATUS_DOWN) {
			zlog_debug ("\t Deleted old SCN-if (local address: %s)", inet_ntoa(scn_if->loc_addr));
			listnode_delete (scnif_list, scn_if);
			free (scn_if);
		}
	}
	for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, tel_cc)) {
		if (tel_cc->status == STATUS_DOWN) {
			zlog_debug ("\t Deleted old TE-link (local address: %s)", inet_ntoa(tel_cc->tel_loc));
			listnode_delete (tel_cc_list, tel_cc);
			free (tel_cc);
		}
	}
	for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif)) {
		if (cc_scnif->status == STATUS_DOWN) {
			zlog_debug ("\t Deleted old CC (old CCid: %d)", cc_scnif->cc_id);
			listnode_delete (cc_scnif_list, cc_scnif);
			free (cc_scnif);
		}
	}
}

/* This function updates scn_if list merging the old one with the temporray list */
static void
realign_scnif (void)
{
	struct scnif *scn_if;
	struct zlistnode *scnif_node, *scnif_nnode;
	struct scnif *tmp_scnif;
	struct zlistnode *tmp_scnif_node, *tmp_scnif_nnode;
	struct scnif *new_scnif;
	u_int32_t found_scnif;

	found_scnif = 0;

	/* Realign SCN-ifs */
	for (ALL_LIST_ELEMENTS (tmp_scnif_list, tmp_scnif_node, tmp_scnif_nnode, tmp_scnif)) {
		for (ALL_LIST_ELEMENTS (scnif_list, scnif_node, scnif_nnode, scn_if)) {
			if (scn_if->loc_addr.s_addr == tmp_scnif->loc_addr.s_addr) {
				found_scnif = 1;
				scn_if->status = UP;
				zlog_debug ("\t Received unmodified an old SCN-if from LRM");
				zlog_debug ("\t\t address: %s", inet_ntoa(scn_if->loc_addr));
				break;
			}
		}
		if (!found_scnif) {
			new_scnif = scnif_new ();

			new_scnif->loc_addr.s_addr = tmp_scnif->loc_addr.s_addr;
			new_scnif->status = UP;

			listnode_add (scnif_list, new_scnif);
			zlog_debug ("\t Added a new SCN-if from LRM");
			zlog_debug ("\t\t address: %s", inet_ntoa(new_scnif->loc_addr));
			new_scnif = NULL;
		}
		found_scnif = 0;
	}
}

/* This function updates CC list merging the old one with the temporray list */
static void
realign_cc (void)
{
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *tmp_cc_scnif;
	struct zlistnode *tmp_ccscnif_node, *tmp_ccscnif_nnode;
	struct cc_scnif_assoc *new_cc_scnif;
	u_int32_t found_cc;

	found_cc = 0;

	/* Realign CCs */
	for (ALL_LIST_ELEMENTS (tmp_cc_scnif_list, tmp_ccscnif_node, tmp_ccscnif_nnode, tmp_cc_scnif)) {
		for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif)) {
			if (cc_scnif->cc_id == tmp_cc_scnif->cc_id) {
				found_cc = 1;
				cc_scnif->status = UP;
				if ((cc_scnif->scnif_loc.s_addr == tmp_cc_scnif->scnif_loc.s_addr) &&
				    (cc_scnif->scnif_rem.s_addr == tmp_cc_scnif->scnif_rem.s_addr)) {
					zlog_debug ("\t Received unmodified an old CC/SCN-if association from LRM");
				}
				else {
					cc_scnif->scnif_loc.s_addr = tmp_cc_scnif->scnif_loc.s_addr;
					cc_scnif->scnif_rem.s_addr = tmp_cc_scnif->scnif_rem.s_addr;
					zlog_debug ("\t Updated an old CC/SCN-if association from LRM");
				}
				zlog_debug ("\t\t CCid: %u", cc_scnif->cc_id);
				zlog_debug ("\t\t\t SCN-if loc: %s", inet_ntoa(cc_scnif->scnif_loc));
				zlog_debug ("\t\t\t SCN-if rem: %s", inet_ntoa(cc_scnif->scnif_rem));
				break;
			}
		}
		if (!found_cc) {
			new_cc_scnif = cc_scnif_new ();

			new_cc_scnif->cc_id = tmp_cc_scnif->cc_id;
			new_cc_scnif->scnif_loc.s_addr = tmp_cc_scnif->scnif_loc.s_addr;
			new_cc_scnif->scnif_rem.s_addr = tmp_cc_scnif->scnif_rem.s_addr;
			new_cc_scnif->status = UP;

			listnode_add (cc_scnif_list, new_cc_scnif);
			zlog_debug ("\t Added a new CC/SCN-if association from LRM");
			zlog_debug ("\t\t CCid: %u", new_cc_scnif->cc_id);
			zlog_debug ("\t\t\t SCN-if loc: %s", inet_ntoa(new_cc_scnif->scnif_loc));
			zlog_debug ("\t\t\t SCN-if rem: %s", inet_ntoa(new_cc_scnif->scnif_rem));
			new_cc_scnif = NULL;
		}
		found_cc = 0;
	}
}

/* This function updates TE-link list merging the old one with the temporray list */
static void
realign_tel (void)
{
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct tel_cc_assoc *tmp_tel_cc;
	struct zlistnode *tmp_telcc_node, *tmp_telcc_nnode;
	struct zlistnode *cc_node, *cc_nnode;
	u_int32_t *cc_id;
	u_int32_t found_tel;
	struct tel_cc_assoc *new_tel_cc;

	found_tel = 0;

	/* Realign TE-links */
	for (ALL_LIST_ELEMENTS (tmp_tel_cc_list, tmp_telcc_node, tmp_telcc_nnode, tmp_tel_cc)) {
		for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, tel_cc)) {
			if (tel_cc->tel_loc.s_addr == tmp_tel_cc->tel_loc.s_addr) {
				found_tel = 1;
				tel_cc->key = tmp_tel_cc->key;
				tel_cc->tel_rem.s_addr = tmp_tel_cc->tel_rem.s_addr;
				tel_cc->status = UP;
				list_delete_all_node (tel_cc->cclist);
				zlog_debug ("\t Updated an old TE-link/CC association (key = %u, %s) from LRM", tmp_tel_cc->key, if_type_string(tel_cc->interface_type));
				zlog_debug ("\t\t TE-link loc: %s", inet_ntoa(tel_cc->tel_loc));
				zlog_debug ("\t\t TE-link rem: %s (%u CC's associated)", inet_ntoa(tel_cc->tel_rem), tmp_tel_cc->cclist->count);
				for (ALL_LIST_ELEMENTS (tmp_tel_cc->cclist, cc_node, cc_nnode, cc_id)) {
					listnode_add (tel_cc->cclist, cc_id);
					zlog_debug ("\t\t\t CCid: %u", *cc_id);
				}
				break;
			}
		}		if (!found_tel) {
			new_tel_cc = tel_cc_new ();

			new_tel_cc->key = tmp_tel_cc->key;
			new_tel_cc->interface_type = tmp_tel_cc->interface_type;
			new_tel_cc->tel_loc.s_addr = tmp_tel_cc->tel_loc.s_addr;
			new_tel_cc->tel_rem.s_addr = tmp_tel_cc->tel_rem.s_addr;
			new_tel_cc->status = UP;

			zlog_debug ("\t Added a new TE-link/CC association (key = %u, %s) from LRM", new_tel_cc->key, if_type_string(new_tel_cc->interface_type));
			zlog_debug ("\t\t TE-link loc: %s", inet_ntoa(new_tel_cc->tel_loc));
			zlog_debug ("\t\t TE-link rem: %s (%u CC's associated)", inet_ntoa(new_tel_cc->tel_rem), tmp_tel_cc->cclist->count);
			for (ALL_LIST_ELEMENTS (tmp_tel_cc->cclist, cc_node, cc_nnode, cc_id)) {
				listnode_add (new_tel_cc->cclist, cc_id);
				zlog_debug ("\t\t\t CCid: %u", *cc_id);
			}
			listnode_add (tel_cc_list, new_tel_cc);
			new_tel_cc = NULL;
		}
		found_tel = 0;
	}

}

void
realign_associations (void)
{
	zlog_debug ("Start of associations realignment");

	realign_scnif ();
	realign_cc ();
	realign_tel ();
	delete_tmp_associations ();
}

#if !HAVE_OMNIORB
/* This is the callback function called when LRM sends an LRM_END_SYNC message*/
static int
scngws_end_sync (int command, struct zclient *zclient, zebra_size_t length)
{
	zlog_debug ("Received a message from LRM: command %d", command);

	thread_cancel (scn_m->timer_thread);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			if ((n_assoc_scnif == 0)&&(n_assoc_tel == 0)) {
				scn_m->lrm_conn_status = STATUS_UP;
				realign_associations ();
				delete_old_associations ();
				zlog_debug ("End of synchronization of the TE-link/CC/SCN-if associations");
				return SCNGW_ACK_END;
			}
			else {
				zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command.");
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				return SCNGW_SYNC_ERR;
			}
			break;
		default:
			zlog_debug ("Connection with LRM is not in synchronization phase. Expected another kind of command.");
			return SCNGW_UNACK;
	}
}

/* This is the callback function called when LRM sends an LRM_SCNIF_ADD message */
static int
scngws_scnif_add (int command, struct zclient *zclient, zebra_size_t length)
{
	struct scnif *scn_if;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			if (n_scnif == 0) {
				if ((n_assoc_scnif == 0)&&
				    (n_assoc_tel == 0)) {
					zlog_debug ("Connection with LRM is in synchronization phase. Expected command %d. Error.", LRM_END_SYNC);
				}
				else {
					zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command. Error.");
				}
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				thread_cancel (scn_m->timer_thread);
				return SCNGW_SYNC_ERR;
			}
			break;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	scn_if = read_scnif_add (zclient->ibuf);
	if (scn_if == NULL)
		return SCNGW_ACK;
	if (scn_m->lrm_conn_status == STATUS_SYNC) {
		listnode_add (tmp_scnif_list, scn_if);
		zlog_debug ("\t Added an SCN-if to the temporary list");
	} else {
		listnode_add (scnif_list, scn_if);
		zlog_debug ("\t Added a new SCN-if from LRM");
		zlog_debug ("\t\t address: %s", inet_ntoa(scn_if->loc_addr));
	}

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_SCNIF_ADD message */
static int
scngws_scnif_delete (int command, struct zclient *zclient, zebra_size_t length)
{
	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			thread_cancel (scn_m->timer_thread);
			return SCNGW_SYNC_ERR;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	read_scnif_delete (zclient->ibuf);

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_TEL_CC_ASSOC_ADD message */
static int
scngws_tel_cc_assoc_add (int command, struct zclient *zclient, zebra_size_t length)
{
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *cc_node, *cc_nnode;
	u_int32_t *match;
	u_int32_t n_assoc, i;

	zlog_debug ("Received a message from LRM: command %d", command);
	zlog_debug ("length %d", length);
	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			if ((n_scnif != 0) || (n_assoc_scnif != 0)) {
				zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command. Error.");
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				thread_cancel (scn_m->timer_thread);
				return SCNGW_SYNC_ERR;
			}
			else if (n_assoc_tel == 0) {
				zlog_debug ("Connection with LRM is in synchronization phase. Expected command %d. Error.", LRM_END_SYNC);
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				thread_cancel (scn_m->timer_thread);
				return SCNGW_SYNC_ERR;
			}
			break;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the number of TE-link/CC associatons*/
	n_assoc = stream_getl (zclient->ibuf);

	for (i = 1; i <= n_assoc; i++) {
		tel_cc = read_tel_cc_assoc_add (zclient->ibuf);
		if (tel_cc == NULL)
			continue;
		if (scn_m->lrm_conn_status == STATUS_SYNC) {
			listnode_add (tmp_tel_cc_list, tel_cc);
			zlog_debug ("\t Added a TE-link/CC association to the temporary list of associations");
		}
		else {
			listnode_add (tel_cc_list, tel_cc);
			zlog_debug ("\t Added a new TE-link/CC association (key = %u, %s) from LRM", tel_cc->key, if_type_string(tel_cc->interface_type));
			zlog_debug ("\t\t TE-link loc: %s", inet_ntoa(tel_cc->tel_loc));
			zlog_debug ("\t\t TE-link rem: %s (%u CC's associated)", inet_ntoa(tel_cc->tel_rem), tel_cc->cclist->count);
			for (ALL_LIST_ELEMENTS (tel_cc->cclist, cc_node, cc_nnode, match))
				zlog_debug ("\t\t\t CCid: %u", *match);
		}
	}

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_TEL_CC_ASSOC_DELETE message */
static int
scngws_tel_cc_assoc_delete (int command, struct zclient *zclient, zebra_size_t length)
{
	u_int32_t n_assoc, i;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			thread_cancel (scn_m->timer_thread);
			return SCNGW_SYNC_ERR;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the number of TE-link/CC associatons*/
	n_assoc = stream_getl (zclient->ibuf);

	for (i = 1; i <= n_assoc; i++)
		read_tel_cc_assoc_delete (zclient->ibuf);

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_TEL_CC_ASSOC_UPDATE message */
static int
scngws_tel_cc_assoc_update (int command, struct zclient *zclient, zebra_size_t length)
{
	struct zlistnode *telcc_node, *telcc_nnode;
	struct tel_cc_assoc *match;
	u_int32_t key, n_upd, i, found = 0;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			thread_cancel (scn_m->timer_thread);
			return SCNGW_SYNC_ERR;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the key of TE-link/CC association to update*/
	key = stream_getl (zclient->ibuf);
	/* Get the number of CC to add/delete*/
	n_upd = stream_getl (zclient->ibuf);

	for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, match)) {
		if (match->key == key) {
			found = 1;
			for (i = 1; i <= n_upd; i++) {
				read_tel_cc_assoc_update (zclient->ibuf, match);
			}
			break;
		}
	}
	if (found == 0)
		zlog_debug ("\t TE-link with key %u is not present in SCNGW and cannot be updated", key);

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_CC_SCNIF_ASSOC_ADD message */
static int
scngws_cc_scnif_assoc_add (int command, struct zclient *zclient, zebra_size_t length)
{
	struct cc_scnif_assoc *cc_scnif;
	u_int32_t n_assoc, i;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			if (n_scnif != 0) {
				zlog_debug ("Connection with LRM is in synchronization phase. Expected command %d. Error.", LRM_SCNIF_ADD);
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				thread_cancel (scn_m->timer_thread);
				return SCNGW_SYNC_ERR;
			}
			else if (n_assoc_scnif == 0) {
				if (n_assoc_tel == 0) {
					zlog_debug ("Connection with LRM is in synchronization phase. Expected command %d. Error.", LRM_END_SYNC);
				}
				else {
					zlog_debug ("Connection with LRM is in synchronization phase. Expected command %d. Error.", LRM_TEL_CC_ASSOC_ADD);
				}
				delete_tmp_associations ();
				scngws_lrm_connection_down (zclient);
				thread_cancel (scn_m->timer_thread);
				return SCNGW_SYNC_ERR;
			}
			break;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the number of CC/SCN-if associatons*/
	n_assoc = stream_getl (zclient->ibuf);

	for (i = 1; i <= n_assoc; i++) {
		cc_scnif = read_cc_scnif_assoc_add (zclient->ibuf);
		if (cc_scnif == NULL)
			continue;
		if (scn_m->lrm_conn_status == STATUS_SYNC) {
			listnode_add (tmp_cc_scnif_list, cc_scnif);
			zlog_debug ("\t Added CC/SCN-if association to the temporary list of associations");
		}
		else {
			listnode_add (cc_scnif_list, cc_scnif);
			zlog_debug ("\t Added a new CC/SCN-if association from LRM");
			zlog_debug ("\t\t CCid: %u", cc_scnif->cc_id);
			zlog_debug ("\t\t\t SCN-if loc: %s", inet_ntoa(cc_scnif->scnif_loc));
			zlog_debug ("\t\t\t SCN-if rem: %s", inet_ntoa(cc_scnif->scnif_rem));
		}
	}
	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_CC_SCNIF_ASSOC_DELETE message */
static int
scngws_cc_scnif_assoc_delete (int command, struct zclient *zclient, zebra_size_t length)
{
	u_int32_t n_assoc, i;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kind of command");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			thread_cancel (scn_m->timer_thread);
			return SCNGW_SYNC_ERR;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the number of CC/SCN-if associatons*/
	n_assoc = stream_getl (zclient->ibuf);

	for (i = 1; i <= n_assoc; i++)
		read_cc_scnif_assoc_delete (zclient->ibuf);

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_CC_SCNIF_ASSOC_UPDATE message */
static int
scngws_cc_scnif_assoc_update (int command, struct zclient *zclient, zebra_size_t length)
{
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *match;
	u_int32_t cc_id, found = 0;

	zlog_debug ("Received a message from LRM: command %d", command);

	switch (scn_m->lrm_conn_status) {
		case STATUS_SYNC:
			zlog_debug ("Connection with LRM is in synchronization phase. Expected another kin of command");
			delete_tmp_associations ();
			scngws_lrm_connection_down (zclient);
			thread_cancel (scn_m->timer_thread);
			return SCNGW_SYNC_ERR;
		case STATUS_NEVER_CONN:
		case STATUS_DOWN:
			zlog_debug ("Connection with LRM is down. Expected command %d. Error.", LRM_BEGIN_SYNC);
			scngws_lrm_connection_down (zclient);
			return SCNGW_SYNC_ERR;
		default:
			break;
	}
	/* Get the id of CC/SCN-if associaton to update*/
	cc_id = stream_getl (zclient->ibuf);

	for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, match)) {
		if (match->cc_id == cc_id) {
			found = 1;
			read_cc_scnif_assoc_update(zclient->ibuf, match);
		}
	}
	if (found == 0)
		zlog_debug ("\t CC with id %u is not present in SCNGW and cannot be updated", cc_id);

	return SCNGW_ACK;
}

/* This is the callback function called when LRM sends an LRM_FATAL_ERROR message */
static int
scngws_sync_fatal_error (int command, struct zclient *zclient, zebra_size_t length)
{
	zlog_debug ("---------------------------------------");
	zlog_debug ("Received a message from LRM: command %d", command);

	scngws_lrm_connection_down (zclient);

	close (zclient->sock);

	zlog_debug ("FATAL ERROR: loop in the LRM-SCNGW communication. Aborting..");
	zlog_debug ("---------------------------------------");

	exit(1);

	return 0;
}

/* This function initialize the communication between SCNGWs and LRM */
void
scngws_lrm_init (void)
{
	/* Allocate zebra structure. */
	zclient = zclient_new ();
	zclient_init (zclient, ZEBRA_ROUTE_SCNGWS, SCNGWS_CLIENT);

	zclient->begin_sync            = scngws_begin_sync;
	zclient->end_sync              = scngws_end_sync;
	zclient->scnif_add             = scngws_scnif_add;
	zclient->scnif_delete          = scngws_scnif_delete;
	zclient->tel_cc_assoc_add      = scngws_tel_cc_assoc_add;
	zclient->tel_cc_assoc_delete   = scngws_tel_cc_assoc_delete;
	zclient->tel_cc_assoc_update   = scngws_tel_cc_assoc_update;
	zclient->cc_scnif_assoc_add    = scngws_cc_scnif_assoc_add;
	zclient->cc_scnif_assoc_delete = scngws_cc_scnif_assoc_delete;
	zclient->cc_scnif_assoc_update = scngws_cc_scnif_assoc_update;
	zclient->lrm_connection_down   = scngws_lrm_connection_down;
	zclient->sync_fatal_error      = scngws_sync_fatal_error;
}
#endif /* !HAVE_OMNIORB */
