/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "getopt.h"
#include "scngwc.h"
#include "scngws.h"
#include "scngws_lrm.h"
#include "scngws_vty.h"
#include "thread.h"
#include "log.h"
#include "version.h"
#include "sigevent.h"
#include "privs.h"
#include "vty.h"
#include "command.h"
#include "scngw_corba.h"
#include "lib/corba.h"

/* scngwsd privileges */
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

struct zebra_privs_t scngwsd_privs = {
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	.user      = QUAGGA_USER,
	.group     = QUAGGA_GROUP,
#endif
#if defined(VTY_GROUP)
	.vty_group = VTY_GROUP,
#endif
	.caps_p    = _caps_p,
	.cap_num_p = 1,
	.cap_num_i = 0
};

/* process id. */
pid_t old_pid;
pid_t pid;

/* SCNGWs master of the threads pointer to export. */
struct thread_master *master;

/* SCNGWs process wide configuration pointer. */
extern struct scn_master *scn_m;

/* Configuration filename and directory. */
char config_default[] = SYSCONFDIR SCNGWS_DEFAULT_CONFIG;

/* Command line options. */
struct option longopts[] = {
	{ "daemon",      no_argument,       NULL, 'd'},
	{ "config_file", required_argument, NULL, 'f'},
	{ "pid_file",    required_argument, NULL, 'i'},
	{ "iors_dir",    required_argument, NULL, 'o'},
	{ "dryrun",      no_argument,       NULL, 'C'},
	{ "vty_port",    required_argument, NULL, 'P'},
	{ "user",        required_argument, NULL, 'u'},
	{ "group",       required_argument, NULL, 'g'},
	{ "version",     no_argument,       NULL, 'v'},
	{ "help",        no_argument,       NULL, 'h'},
	{ 0 }
};

/* Process ID saved for use by init system */
const char *pid_file = PATH_SCNGWSD_PID;

/* Help information display. */
static void
usage (char *progname, int status)
{
	if (status != 0) {
		fprintf (stderr,
			 "Try `%s --help' for more information.\n",
			 progname);
	} else {
		printf ("Usage : %s [OPTION...]\n\n"
			"-d, --daemon       Runs in daemon mode\n"
			"-f, --config_file  Set configuration file name\n"
			"-i, --pid_file     Set process identifier file name\n"
			"-o, --iors_dir     Set IORs directory\n"
			"-C, --dryrun       Check configuration for validity and exit\n"
			"-P, --vty_port     Set vty's port number\n"
			"-u, --user         User to run as\n"
			"-g, --group        Group to run as\n"
			"-v, --version      Print program version\n"
			"-h, --help         Display this help and exit\n"
			"\n"
			"Report bugs to %s\n",
			progname, ZEBRA_BUG_ADDRESS);
	}

	exit (status);
}

/* SIGHUP handler. */
static void
sighup (void)
{
	zlog(NULL, LOG_INFO, "SIGHUP received");
}

static int terminate_execution = 0;

/* SIGINT / SIGTERM handler. */
static void
sigint (void)
{
	zlog_notice("Terminating on signal");
	terminate_execution = 1;
}

/* SIGUSR1 handler. */
static void
sigusr1 (void)
{
	zlog_rotate(NULL);
}

struct quagga_signal_t scngws_signals[] =
{
	{
		.signal  = SIGHUP,
		.handler = &sighup,
	},
	{
		.signal  = SIGUSR1,
		.handler = &sigusr1,
	},
	{
		.signal  = SIGINT,
		.handler = &sigint,
	},
	{
		.signal  = SIGTERM,
		.handler = &sigint,
	},
};

/* Main startup routine. */
int
main (int argc, char **argv)
{
	char          * p;
	int             daemon_mode;
	int             dryrun;
	char          * progname;
	char          * rem_addr;
	int             vty_port;
	char          * vty_addr;
	char          * config_file;
	char          * iors_dir;
	struct thread   thread;

	daemon_mode = 0;
	dryrun      = 0;
	rem_addr    = NULL;
	vty_port    = SCNGWSD_VTY_PORT;
	vty_addr    = NULL;
	iors_dir    = NULL;
	config_file = NULL;

	/* Set umask before anything for security */
	umask (0027);

	/* preserve my name */
	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	/* open log file*/
	zlog_default = openzlog(progname,
				ZLOG_SCNGWS,
				LOG_CONS|LOG_NDELAY|LOG_PID,
				LOG_DAEMON);

	while (1) {

		int opt;

		opt = getopt_long (argc, argv, "dChvf:P:u:g:i:o:", longopts, 0);

		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
			case 'C':
				dryrun = 1;
				break;
			case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				scngwsd_privs.user = optarg;
				break;
			case 'g':
				scngwsd_privs.group = optarg;
				break;
			case 'v':
                                printf("SCN Gateway Server daemon v.%s: "
				       "vty@%d\n",
                                       SCNGWS_CODE_VERSION, vty_port);
                                printf("This program is part of "
                                       "phosphorus-g2mpls\n");
				print_version(progname);
				exit (0);
				break;
			case 'h':
				usage(progname, 0);
				break;
				default:
				usage(progname, 1);
				break;
		}
	}
	/* Initialize the SCNGWs master: system wide conf and variables */
	scngws_master_init();

	/* Master of the threads pointer */
	master = scn_m->master;

	/* Library inits */
	zprivs_init(&scngwsd_privs);
	signal_init(master, Q_SIGC(scngws_signals), scngws_signals);

	/* Initialize basic conf commands (only log command is used) */
	cmd_init(1);
	vty_init(master);
	//	memory_init();

#if !HAVE_OMNIORB
	/* Initialize the communications between SCNGWs and LRM */
	scngws_lrm_init();
#endif /* !HAVE_OMNIORB */

	/* Initialize vty commands */
	scngws_vty_init();

	/* Sorting is needed by the command subsystem to finalize init */
	sort_node();

	/* Configuration file read */
	vty_read_config(config_file, config_default);

	/* Don't start execution if we are in dry-run mode */
	if (dryrun)
		return(0);

	/* Change to the daemon program. */
	if (daemon_mode) {
		daemon(0, 0);
	}

	/* Create VTY socket */
	vty_serv_sock(vty_addr, vty_port, SCNGWS_VTYSH_PATH);

	/* Open the server socket */
	server_sock_init();

#if HAVE_OMNIORB
	if (!corba_init(iors_dir, CORBA_SERVANT_SCNGW, master)) {
		zlog_err("Cannot initialize CORBA subsystem");
		exit(1);
	}
	if (!corba_server_setup()) {
		zlog_err("Cannot setup CORBA subsystem");
		exit(1);
	}

	/* Process id file create. */
	pid_output(pid_file);

	while (!terminate_execution) {
		sleep(1);

		if (corba_client_setup()) {
			if (corba_signal_activity()) {
				break;
			}

			(void) corba_client_shutdown();

		}

		zlog_err("Cannot setup client object");

		// Just to catch signal handler
		if (thread_fetch(master, &thread)) {
			thread_call(&thread);
		}
	}
#endif
	/* Print banner. */
	zlog_notice("SCNGWSd %s starting: vty@%d", QUAGGA_VERSION, vty_port);

	while (!terminate_execution) {
		if (thread_fetch(master, &thread)) {
			thread_call(&thread);
		}
	}

	zlog_debug("Termination in progress ...");

#if HAVE_OMNIORB
	(void) corba_client_shutdown();
	(void) corba_server_shutdown();
	(void) corba_fini();
#endif

	exit(0);
}
