//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//


#include <zebra.h>
#include "thread.h"
#include "vty.h"
#include "log.h"
#include "thread.h"
#include "g2rsvpted.h"

#include "g2rsvpted/g2rsvpte_config.h"
#include "g2rsvpted/g2rsvpte_dm.h"
#include "g2rsvpted/g2rsvpte_common_types.h"
#include "g2rsvpted/g2rsvpte_utils.h"
#include "g2rsvpted/g2rsvpte_lrm_api.h"
#include "g2rsvpted/g2rsvpte_tnrc_api.h"
#include "g2rsvpted/g2rsvpte_rc_api.h"
#include "g2rsvpted/g2rsvpte_lsp.h"
#ifdef PERSISTENCY_ON
#include "g2rsvpted/g2rsvpte_persistency_py.h"
#endif // PERSISTENCY_ON
#include "g2rsvpted/g2rsvpte_scngw_api.h"
#include "g2rsvpted/g2rsvpte_corba.h"
#include "g2rsvpte_scngw_api.h"

#include <sstream>

namespace RSVP {

	//
	// Data Model utils
	//

	InterfaceKey_t g2mpls_addr2key(g2mpls_addr_t addr)
	{
		InterfaceKey_t key = 0;

		switch (addr.type) {
			case IPv4:
				key = addr.value.ipv4.s_addr;
				break;

			case IPv6:
				// Not supported yet
				break;

			case UNNUMBERED:
				//key = addr.value.unnum.addr;
				key = addr.value.unnum;
				break;

			default:
				break;
		}

		return key;
	}

	/////////////////////////////////////////////////////
	//         G2RSVPTE_InstanceType class             //
	/////////////////////////////////////////////////////
	G2RSVPTE_InstanceType::G2RSVPTE_InstanceType(void)
	{
		adjType_         = INNI; // default type of instance

		tunneledPackets_ = true; // default: tunneled

		isUniClient_     = false;

		isCore_          = false;

		initScngw();
	}

	G2RSVPTE_InstanceType::G2RSVPTE_InstanceType(adj_type_t adjType,
						     bool       tunneledPackets)
	{
		adjType_         = adjType;

		tunneledPackets_ = tunneledPackets;

		isUniClient_     = false;

		isCore_          = false;

		initScngw();
	}

	G2RSVPTE_InstanceType::~G2RSVPTE_InstanceType(void)
	{
	}

	void G2RSVPTE_InstanceType::initScngw(void)
	{
		int fd;

		fd = init_scngw(adjType_, tunneledPackets_);
		if (fd < 0) {
			throw("Cannot init connection with SCNgw ...");
		}
		scngwFd_ = fd;

		readThread_ = 0;
	}

	void
	G2RSVPTE_InstanceType::readThread(struct thread * thr)
	{
		readThread_ = thr;
	}

	struct thread *
	G2RSVPTE_InstanceType::readThread(void)
	{
		return readThread_;
	}

	/////////////////////////////////////////////////////
	//                   G2RSVPTE class                //
	/////////////////////////////////////////////////////
	G2RSVPTE::G2RSVPTE(void)
	{
		initConf();

		defaultRefreshInterval_      = DEF_REFRESH_INTERVAL;
		defaultRapidRetransInterval_ = DEF_RAPID_RETRANS_INTERVAL;
		defaultRapidRetryLimit_      = DEF_RAPID_RETRY_LIMIT;
		defaultExpoBackoffDelta_     = DEF_EXPO_BACKOFF_DELTA;
	}

	G2RSVPTE::G2RSVPTE(uint32_t defaultRefreshInterval,
			   uint32_t defaultRapidRetransInterval,
			   uint32_t defaultRapidRetryLimit,
			   uint32_t defaultExpoBackoffDelta) :
		defaultRefreshInterval_(defaultRefreshInterval),
		defaultRapidRetransInterval_(defaultRapidRetransInterval),
		defaultRapidRetryLimit_(defaultRapidRetryLimit),
		defaultExpoBackoffDelta_(defaultExpoBackoffDelta)
	{
		initConf();
	}

	G2RSVPTE::G2RSVPTE(adj_type_t adjType, bool tunneledPackets) :
		G2RSVPTE_InstanceType(adjType, tunneledPackets)
	{
		initConf();

		defaultRefreshInterval_      = DEF_REFRESH_INTERVAL;
		defaultRapidRetransInterval_ = DEF_RAPID_RETRANS_INTERVAL;
		defaultRapidRetryLimit_      = DEF_RAPID_RETRY_LIMIT;
		defaultExpoBackoffDelta_     = DEF_EXPO_BACKOFF_DELTA;
	}

	void G2RSVPTE::initConf()
	{
		G2RSVPTE_ERR("Error trace  ...   enabled!");
		G2RSVPTE_WRN("Warning trace ...  enabled!");
		G2RSVPTE_INF("Info trace ...     enabled!");
		G2RSVPTE_NOT("Notice trace ...   enabled!");
		G2RSVPTE_DBG("Debug trace ...    enabled!");

		nodeId_ = 0;
	}

	G2RSVPTE::~G2RSVPTE(void)
	{
		close_scngw(scngwFd_);

		if (readThread_) {
			thread_cancel(readThread_);
		}
	}

	bool G2RSVPTE::attach(SessionKey_t k, Session * e)
	{
		if (!e) {
			G2RSVPTE_ERR("Invalid Session pointer");
			return false;
		}

		Session * sess;
		sess = findSession(k);
		if (sess) {
			if (e == sess) {
				G2RSVPTE_WRN("Session already attached");
				return true;
			}

			G2RSVPTE_ERR("Key 0x%llx already exists for "
				     "different session pointer",
				     k);
			return false;
		}

		sessions_[k] = e;
		return true;
	}

	bool G2RSVPTE::detach(SessionKey_t k, Session * e)
	{
		if (!e) {
			G2RSVPTE_ERR("Invalid Session pointer");
			return false;
		}

		iterator_sessions it;
		it = sessions_.find(k);
		if (it == sessions_.end() || (*it).second != e) {
			G2RSVPTE_WRN("Wrong Session (key = 0x%llx) to detach",
				     k);
			return false;
		}

		sessions_.erase(it);
		G2RSVPTE_DBG("Session key = 0x%llx detached", k);
		return true;
	}

	int G2RSVPTE::loadInterfaces(void)
	{
		std::list<Interface_data_t>           intfs;
		std::list<Interface_data_t>::iterator it;
		iterator_interfaces                   intfIt;

		if (!getInterfaces(nodeId_, intfs, adjType_)) {
			G2RSVPTE_ERR("Impossible to get all interfaces for "
				     "router ID %s",
				     numb_htoa(nodeId_));
			return 0;
		}

		if (!interfaces_.empty()) {
			G2RSVPTE_WRN("Removing all the old interfaces for this"
				     " G2.RSVP-TE instance before loading");
			removeAllInterfaces();
		}

		int count = 0;
		for (it = intfs.begin(); it != intfs.end(); it++) {
			InterfaceKey_t   tmpKey;
			Interface_data_t tmpData;

			tmpData = *it;
			tmpKey = g2mpls_addr2key(tmpData.localId);

			intfIt = interfaces_.find(tmpKey);
			if (intfIt != end_interfaces()) {
				G2RSVPTE_ERR("Interface 0x%x already exists",
					     tmpKey);
				removeAllInterfaces();
				return 0;
			}

			Interface * intf = new Interface(this,
							 tmpData.localId,
							 tmpData.remoteId,
							 tmpData.neighbourId,
							 tmpData.op_state,
							 tmpData.adm_state);
			if (!intf) {
				G2RSVPTE_ERR("Impossible to allocate memory "
					     "for interface %s",
					     addr_ntoa(tmpData.localId));
				removeAllInterfaces();
				return 0;
			}

			intf->dump();

			interfaces_[tmpKey] = intf;
			count++;
		}

		G2RSVPTE_DBG("Got %d interfaces", count);

		return count;
	}

	bool G2RSVPTE::removeAllInterfaces(void)
	{
		iterator_interfaces it;

		int count = 0;
		for (it = interfaces_.begin(); it != interfaces_.end(); it++) {
			if ((*it).second) {
				G2RSVPTE_DBG("Removing interface 0x%x",
					     (*it).first);
				delete (*it).second;
				(*it).second = 0;    // useless!!!
				count++;
			}
		}

		interfaces_.clear();
		G2RSVPTE_DBG("Removed %d interfaces", count);

		return true;
	}

	Session * G2RSVPTE::findSession(SessionKey_t key)
	{
		iterator_sessions it;
		Session * sess = 0;

		it = sessions_.find(key);
		if (it == sessions_.end()) {
			G2RSVPTE_WRN("No session found with key 0x%llx", key);
			return 0;
		}

		sess = (*it).second;
		return sess;
	}

	Interface * G2RSVPTE::findInterface(g2mpls_addr_t addr,
					    bool          checkRemote)
	{
		iterator_interfaces it;

		for (it = interfaces_.begin(); it != interfaces_.end(); it++) {
			Interface * intf;
			intf = (*it).second;
			int res;

			if (!intf) {
				G2RSVPTE_WRN("Found invalid interface ... "
					     "skipping it");
				continue;
			}

			if (!checkRemote) {
				res = addr_equal(intf->localId(), addr);
			} else {
				res = addr_equal(intf->remoteId(), addr);
			}
			if (res) {
				G2RSVPTE_DBG("Found interface for address %s",
					     addr_ntoa(addr));
				return intf;
			}
		}

		G2RSVPTE_WRN("No interface found for address %s",
			     addr_ntoa(addr));
		return 0;
	}

	Interface * G2RSVPTE::findInterface(InterfaceKey_t key)
	{
		iterator_interfaces it;
		Interface * intf = 0;

		it = interfaces_.find(key);
		if (it == interfaces_.end()) {
			G2RSVPTE_ERR("No interface found with key %u", key);
			return 0;
		}

		intf = (*it).second;
		return intf;
	}

	LSP * G2RSVPTE::findLSP(lsp_ident_t ident)
	{
		Session * sess;
		LSP *     lsp;

		// We ignore ident.ext_tun_id
		//SHOW_LSPIDENT(ident);

		// Search into session map
		SessionKey_t sKey;
		sKey = MAKE_SESS_KEY(ident.dst_nid,
				     ident.tun_id? ident.tun_id : ident.lsp_id);

		sess = findSession(sKey);
		if (!sess) {
			G2RSVPTE_ERR("Invalid session pointer");
			return 0;
		}

		// Search into lsp map
		LSPKey_t lKey;
		lKey = MAKE_LSP_KEY(ident.src_nid ? ident.src_nid : nodeId_,
				    ident.lsp_id);
		lsp = sess->findLSP(lKey);

		return lsp;
	}

	LSP * G2RSVPTE::findLSP(uint32_t context)
	{
		Session *         sess;
		iterator_sessions sIt;
		LSP *             lsp;

		lsp = 0;

		for (sIt = sessions_.begin(); sIt != sessions_.end(); sIt++) {
			sess = (*sIt).second;

			lsp = sess->findLSP(context);

			if (lsp) {
				break;
			}
		}

		return lsp;
	}

	//VERY TMP
	bool G2RSVPTE::checkLSP(LSP * lsp) {

		Session *         sess;
		iterator_sessions sIt;

		for (sIt = sessions_.begin();
		     sIt != sessions_.end();
		     sIt++) {
			sess = (*sIt).second;
			if (sess->checkLSP(lsp)) {
				return true;
			}
		}

		return false;
	}

	LSP * G2RSVPTE::createLSP(const lsp_ident_t & ident,
				  Interface *         intf,
				  Message *           msg)
	{
		if (!msg) {
			G2RSVPTE_ERR("Invalid intf/msg pointer/s");
			return 0;
		}
		if (msg->getType() != Message::PATH) {
			G2RSVPTE_ERR("Invalid msg type, received %s, "
				     "expected PATH",
				     MSGTYPE2STR(msg->getType()));
			return 0;
		}

		LSP *            lsp;
		lsp_info_t       lspInfo;

		memset(&lspInfo, 0, sizeof(lspInfo));
		memset(&lspInfo.mask_, 1, sizeof(lspInfo.mask_));

		// TODO: check all the null parameters ???
		lsp = createLSP_(ident,
				 0,                // iRes
				 0,                // eRes
				 "",               // session name
				 0,
				 lspInfo,
				 0,
				 1);               // Called by INTERNAL functs

		if (lsp) {
			// complete LSP information
			// ... needed to send the ACK before invoking
			//     the LSP FSM transition
			assert(lsp->psb());
			if (intf) {
				lsp->psb()->interface(intf);
			}
			//TODO Get lsp_attributes from msg
			//     ... and use them into createLSP_
		}

		return lsp;
	}

	LSP * G2RSVPTE::createLSP(const lsp_ident_t &      ident,
				  const std::string &      sessionName,
				  const call_ident_t &     callIdent,
				  const net_res_spec_t &   iRes,
				  const net_res_spec_t &   eRes,
				  const lsp_info_t &       lspInfo,
				  const recovery_info_t &  recoveryInfo)
	{
		LSP * lsp;

		if (isCallIdentValid(callIdent, adjType()) > 0) {
			lsp = createLSP_(ident,
					 &iRes,
					 &eRes,
					 sessionName,
					 &callIdent,
					 lspInfo,
					 &recoveryInfo,
					 0);           // Called by EXTERNAL functs
		} else if (isCallIdentValid(callIdent, adjType()) < 0) {
			G2RSVPTE_ERR("Call Ident is not valid. Cannot create LSP");
			return 0;
		} else {
			lsp = createLSP_(ident,
					 &iRes,
					 &eRes,
					 sessionName,
					 0,
					 lspInfo,
					 &recoveryInfo,
					 0);           // Called by EXTERNAL functs
		}

		return lsp;
	}

	LSP * G2RSVPTE::createLSP(const lsp_ident_t &      ident,
				  const std::string &      sessionName,
				  const call_ident_t &     callIdent,
				  const call_info_t &      callInfo,
				  const lsp_info_t &       lspInfo,
				  const recovery_info_t &  recoveryInfo,
				  const std::string &      aaaInfo)
	{
		LSP * lsp;

		if (isCallIdentValid(callIdent, adjType()) > 0) {
			lsp = createLSP_(ident,
					 0,
					 0,
					 sessionName,
					 &callIdent,
					 lspInfo,
					 &recoveryInfo,
					 0);          // Called by EXTERNAL functs
		} else {
			G2RSVPTE_ERR("Call Ident is not valid. Cannot create LSP");
			return 0;
		}
		if (!lsp) {
			return 0;
		}

		if (!lsp->psb() || !lsp->psb()->pathOut()) {
			G2RSVPTE_ERR("Null pointer !!!");
			return 0;
		}
		//Set call type in lsp instance
		lsp->callType(callInfo.call_type);

		//Set call info paramters in PathOut and PathIn
		if (!lsp->psb()->pathOut()->setGnsCallExt(callInfo)) {
			G2RSVPTE_ERR("Impossible to set GnsCallExt object !!");
			return 0;
		}
		lsp->psb()->pathIn()->gns_call_ext_ = lsp->psb()->pathOut()->gns_call_ext_->clone();

		//Set Policy data paramters in PathOut and PathIn
		if (!aaaInfo.empty()) {
			if (!lsp->psb()->pathOut()->setPolicyData(aaaInfo)) {
				G2RSVPTE_ERR("Impossible to set PolicyData object !!");
				return 0;
			}
			lsp->psb()->pathIn()->policy_data_ = lsp->psb()->pathOut()->policy_data_->clone();
		}

		return lsp;
	}

	bool G2RSVPTE::destroyLSP(const lsp_ident_t & ident, bool internal)
	{
		SHOW_LSPIDENT(ident);

		Session * sess;
		LSP *     lsp;

		// Search into session map
		SessionKey_t sKey;
		sKey = MAKE_SESS_KEY(ident.dst_nid,
				     ident.tun_id? ident.tun_id : ident.lsp_id);

		sess = findSession(sKey);
		if (!sess) {
			G2RSVPTE_ERR("No session found!!!");
			return false;
		}

		// Search into lsp map
		LSPKey_t lKey;
		lKey = MAKE_LSP_KEY(ident.src_nid ? ident.src_nid : nodeId_,
				    ident.lsp_id);
		lsp = sess->findLSP(lKey);
		if (!lsp) {
			G2RSVPTE_ERR("No LSP found!!!");
			return false;
		}

		//if (!internal && (!isHead() && !isTail())) {
		if (!internal && !lsp->isHead() && !lsp->isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		lsp->disableLSP();

		if (!internal && lsp->isEnabled()) {
			G2RSVPTE_ERR("This LSP is not disabled!!!");
			return false;
		}
		//if (!forced)
		//if (FSM::state != G2RSVPTE_LSP_FSM_Down) {
		//	G2RSVPTE_ERR("LSP must be Down to be destroyed!!!");
		//	return false;
		//}


		//???
		//if (internal && isTail()) {
		//	if (lsp_flags havebeenactive) {
		//		lspNotifyTearDown;
		//	} else {
		//		lspNotifySignalUp;
		//	}
		//}

		// Detach this LSP from the session
		if (!sess->detach(lKey, lsp)) {
			G2RSVPTE_ERR("Impossible to detach LSP id %d",
				     lsp->id());
			return false;
		}

		// Detach session from G2RSVPTE instance if there are not LSP
		if (sess->empty()) {
			if (!detach(sKey, sess)) {
				G2RSVPTE_WRN("Impossible to detach sess nodeId "
					     "%s / tunId %d",
					     numb_htoa(sess->nodeId()),
					     sess->tunnelId());
				//TODO??? return/rollback/continue .... ???
			}
			// Deleting session
			delete sess;
		}

		// Deleting all the LSP data
		delete lsp;

		G2RSVPTE_NOT("LSP destroyed successfully");
		return true;
	}

	LSP * G2RSVPTE::createLSP_(const lsp_ident_t &     id,
				   const net_res_spec_t *  iRes,
				   const net_res_spec_t *  eRes,
				   const std::string &     sessionName,
				   const call_ident_t *    callIdent,
				   const lsp_info_t &      lspInfo,
				   const recovery_info_t * recoveryInfo,
				   bool                    internal,
				   bool                    isRebuilded)
	{
		lsp_ident_t ident = id;
		if (!isRebuilded) {
			SHOW_LSPIDENT(ident);
		}

		if (!ident.tun_id) {
			ident.tun_id = ident.lsp_id;
		}

		if (!isRebuilded && !internal) {
			if (ident.src_nid != 0 && ident.src_nid != nodeId_) {
				G2RSVPTE_ERR("No LSP originated by myself "
					     "(src_nid=%s)",
					     numb_htoa(ident.src_nid));
				return 0;
			} else {
				ident.src_nid = nodeId_;
			}

			// can't apply crankback on an adopted LSP!
			// concatenation ???
			// transparency  ???

			// iTna & eTna ARE NOT MANDATORY!!!
			//if (!iTna) {
			//	G2RSVPTE_ERR("Ingress TNA must be specified");
			//	return 0;
			//}
			//if (!eTna) {
			//	G2RSVPTE_ERR("Egress TNA must be specified");
			//	return 0;
			//}
			// check that the Ingress TNA really exists
			//if (iTna && !checkTELink(*iTna)) {
			//	G2RSVPTE_ERR("Ingress TNA (%s) is not one of "
			//		     "my TELinks",
			//		     addr_ntoa(*iTna));
			//	return 0;
			//}
		}

		Session *    sess;
		SessionKey_t sKey;

		sKey = MAKE_SESS_KEY(ident.dst_nid, ident.tun_id);
		sess = findSession(sKey);
		if (!sess) {
			sess = new Session(this,
					   ident.dst_nid,
					   ident.tun_id,
					   ident.ext_tun_id);
			if (!sess) {
				G2RSVPTE_ERR("Impossible to create session");
				return 0;
			}

			if (!attach(sKey, sess)) {
				G2RSVPTE_ERR("Impossible to attach session");
				delete sess;
				return 0;
			}
		}

		LSP * lsp;
		lsp = new LSP(sess, ident.src_nid, ident.lsp_id);
		if (!lsp) {
			G2RSVPTE_ERR("Impossible to create LSP (nid %s, "
				     "lsp_id %d)",
				     numb_htoa(ident.src_nid),
				     ident.lsp_id);
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			return 0;
		}

		// TODO:  set flags for LSP

		if (iRes) {
			lsp->iRes(*iRes);
		} else {
			net_res_spec_t nullRes;
			memset(&nullRes, 0, sizeof(nullRes));
			lsp->iRes(nullRes);
		}
		if (eRes) {
			lsp->eRes(*eRes);
		} else {
			net_res_spec_t nullRes;
			memset(&nullRes, 0, sizeof(nullRes));
			lsp->eRes(nullRes);
		}

		if (callIdent) {
			lsp->callId(*callIdent);
		} else {
			call_ident_t nullIdent;
			memset(&nullIdent, 0, sizeof(nullIdent));
			lsp->callId(nullIdent);
		}

		// Set timers/intervals for LSP
		lsp->refreshInterval(lspInfo.refresh_interval ?
				     lspInfo.refresh_interval :
				     defaultRefreshInterval_);
		lsp->rapidRetryLimit(lspInfo.rapid_retry_limit ?
				     lspInfo.rapid_retry_limit :
				     defaultRapidRetryLimit_);
		lsp->rapidRetransInterval(lspInfo.rapid_retransm_interval ?
					  lspInfo.rapid_retransm_interval :
					  defaultRapidRetransInterval_);
		lsp->expoBackoffDelta(lspInfo.increment_value_delta ?
				      lspInfo.increment_value_delta :
				      defaultExpoBackoffDelta_);
		lsp->times(lspInfo.times);

		//START - TODO move this part into LSP constructor ???
		PSB * psb;
		psb = new PSB();
		if (!psb) {
			G2RSVPTE_ERR("Impossible to create PSB for LSP id %d",
				     lsp->id());
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;
			return 0;
		}

		// Initialise PSB for this new LSP
		if (!lsp->attach(psb)) {
			G2RSVPTE_ERR("Impossible to attach PSB to LSP id %d",
				     lsp->id());
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete psb; // deleting here because it is not attached to LSP yet
			delete lsp;
			return 0;
		}

		//TODO: create also RSB here?????
		RSB * rsb;
		rsb = new RSB();
		if (!rsb) {
			G2RSVPTE_ERR("Impossible to create RSB for LSP id %d",
				     lsp->id());
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}
		if (!lsp->attach(rsb)) {
			G2RSVPTE_ERR("Impossible to attach RSB to LSP id %d",
				     lsp->id());
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete rsb;  // deleting here because it is not attached to LSP yet
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}
		//END

		if (!internal) { // I'm the iTNA - LSP is created by command
			PathMessage * pathIn;
			PathMessage * pathOut;
			uint8_t       version;
			if (adjType_ == INNI) {
				version = 1;
			} else {
				version = 4;
			}

			try {
				G2RSVPTE_DBG("Creating pathIn for LSP id %d", lsp->id());
				pathIn = new PathMessage(version,
							 ident.src_nid,
							 ident.dst_nid,
							 ident.lsp_id,
							 ident.tun_id,
							 ident.ext_tun_id,
							 nodeId_,
							 lspInfo.sw_cap,
							 lspInfo.enc_type,
							 lspInfo.bw,
							 lspInfo.gpid,
							 lspInfo.setup_prio,
							 lspInfo.holding_prio,
#ifdef CRANKBACK_ON
							 lspInfo.crankback,
#endif // CRANKBACK_ON
							 lspInfo.rro_mode,
							 lsp->refreshInterval(),
							 sessionName,
							 iRes,
							 eRes,
							 callIdent,
							 lspInfo.times);
			} catch (std::string e) {
				G2RSVPTE_DBG("Failed to create pathIn: %s", e.c_str());
				delete lsp;
				return 0;
			} catch (...) {
				G2RSVPTE_DBG("Failed to create pathIn");
				delete lsp;
				return 0;
			}

			G2RSVPTE_DBG("Cloning pathIn into pathOut for LSP id %d", lsp->id());
			if (!pathIn ||
			    !(pathOut=static_cast<RSVP::PathMessage *>(pathIn->clone()))) {
				G2RSVPTE_ERR("Impossible to create Path for LSP id %d",
					     lsp->id());
				if (sess->empty()) {
					detach(sKey, sess);
					delete sess;
				}
				delete lsp;  //TODO: it will delete all internal struct
				return 0;
			}
			G2RSVPTE_DBG("Initialise pathOut for LSP id %d", lsp->id());
			// Initialise PathOut for this new LSP
			lsp->psb()->pathOut(pathOut);

			//if (!imported) {
			// there's no incoming Path msg in the first node, but it's needed
			// for ERO processing ... let's link it to the outgoing one

			G2RSVPTE_DBG("Initialise pathIn for LSP id %d", lsp->id());
			//lsp->psb()->pathIn(lsp->psb()->pathOut());
			lsp->psb()->pathIn(pathIn);

#ifdef CRANKBACK_ON
			if (adjType_ == INNI) {
				G2RSVPTE_DBG("Initialise crankback scope (%s) for LSP id %d",
					     SHOW_CBACK_SCOPE(lspInfo.crankback), lsp->id());
				lsp->cbackScope(lspInfo.crankback);
				if (lsp->isHead()) {
					G2RSVPTE_DBG("Initialise max crankback retries (%d) "
						     "for LSP id %d",
						     lspInfo.max_cback_retries_src,
						     lsp->id());
					lsp->maxCbackRetries(lspInfo.max_cback_retries_src);
				}
			}
#endif // CRANKBACK_ON
			// TODO ....

			//if (!imported) {
			// this LSP starts on this NE, there is no PHOP
			lsp->psb()->nodeId(0);
			// lsp->psb()->interface(0); // Useless!!! new PSB has interface not initialized
		}

		UpstreamAckNack * uAN;
		uAN = new UpstreamAckNack(lsp);
		if (!uAN) {
			G2RSVPTE_ERR("Impossible to create upstream AckNack");
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}

		DownstreamAckNack * dAN;
		dAN = new DownstreamAckNack(lsp);
		if (!dAN) {
			G2RSVPTE_ERR("Impossible to create dowwnstream AckNack");
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete uAN;
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}

		// TODO: Set ACK/NACKs flags

		if (!lsp->attach(uAN) || !lsp->attach(dAN)) {
			G2RSVPTE_ERR("Impossible to attach needed struct to LSP");
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}

		G2RSVPTE_DBG("Initialise LSP-FSM for LSP id %d", lsp->id());
		// Initialize LSP's FSM
		lsp->fsmInst_ = new fsm::G2RSVPTE_LSP_FSM::virtFsm
			(fsm::base_G2RSVPTE_LSP_FSM::BaseFSM::TRACE_LOG);
		if (!lsp->fsmInst_) {
			G2RSVPTE_ERR("Impossible to allocate LSP'FSM");
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}

		G2RSVPTE_DBG("Attach LSP %d to its session", lsp->id());
		// Attach this LSP to the session
		LSPKey_t lKey;
		lKey = MAKE_LSP_KEY(ident.src_nid, ident.lsp_id);
		if (!sess->attach(lKey, lsp)) {
			G2RSVPTE_ERR("Impossible to attach LSP id %d",
				     lsp->id());
			if (sess->empty()) {
				detach(sKey, sess);
				delete sess;
			}
			delete lsp;  //TODO: it will delete all internal struct
			return 0;
		}

		//TODO Initiaize LSP's FSM ... if needed (Default params NOW!)
		//lsp->fsmInst_->traceLevel(fsm::G2RSVPTE_LSP_FSM::TRACE_DBG);

		// TODO
		// Initialize stats/counters
		// Start purging timer???

		G2RSVPTE_NOT("LSP id %d created", lsp->id());
		return lsp;
	}

#ifdef PERSISTENCY_ON
	void G2RSVPTE::databaseSync(void)
	{
		uint32_t     count;
		lsp_ident_t  ident;
		lsp_ident_t  nextIdent;
		std::string  resp;

		G2RSVPTE_DBG("Going to retrieve LSPs stored in Persistency Controller");

		count = 0;
		resp = "";
		memset(&ident, 0, sizeof(ident));
		memset(&nextIdent, 0, sizeof(nextIdent));
		while (1) {
			G2RSVPTE_DBG("Retrieving LSP n. %d", count + 1);
			if (!retrieveLsp(ident, nextIdent, resp)) {
#ifndef G2RSVPTE2PC_DEBUG
				G2RSVPTE_ERR("LSP not retrieved: %s", resp.c_str());
#endif // G2RSVPTE2PC_DEBUG
				break;
			}
			ident = nextIdent;
			memset(&nextIdent, 0, sizeof(nextIdent));
			count++;
		}
#ifndef G2RSVPTE2PC_DEBUG
		if (!count) {
			G2RSVPTE_DBG("There are no LSP stored in Persistency Controller");
		}
#endif // G2RSVPTE2PC_DEBUG
	}

	bool G2RSVPTE::retrieveLsp(const lsp_ident_t ident,
				   lsp_ident_t  &    nextIdent,
				   std::string &     resp)
	{
		lsp_info_t                lspInfo;
		std::list<lsp_ero_sobj_t> ero;
		LSP *                     lsp;
		storedLspStateBlockInfo_t psbInfo;
		storedLspStateBlockInfo_t rsbInfo;
		net_res_spec_t            iTnaRes;
		net_res_spec_t            eTnaRes;
		call_ident_t              callId;
		call_info_t               callInfo;
		storedLspDataMask_t       mask;
		grid_res_spec_t           iGnsTna;
		grid_res_spec_t           eGnsTna;
		std::string               authz;
		recovery_info_t           recoveryInfo;

		//initialise
		memset(&lspInfo, 0, sizeof(lspInfo));
		memset(&psbInfo, 0, sizeof(psbInfo));
		memset(&rsbInfo, 0, sizeof(rsbInfo));
		memset(&iTnaRes, 0, sizeof(iTnaRes));
		memset(&eTnaRes, 0, sizeof(eTnaRes));
		memset(&recoveryInfo, 0, sizeof(recoveryInfo));
		memset(&callId, 0, sizeof(callId));
		memset(&callInfo, 0, sizeof(callInfo));
		memset(&mask, 0, sizeof(mask));
		memset(&iGnsTna, 0, sizeof(iGnsTna));
		memset(&eGnsTna, 0, sizeof(eGnsTna));

		if (!persistency_.fetchLsp(ident,
					   adjType(),
					   nextIdent,
					   lspInfo,
					   psbInfo,
					   rsbInfo,
					   iTnaRes,
					   eTnaRes,
					   callId,
					   callInfo,
					   mask,
					   iGnsTna,
					   eGnsTna,
					   authz,
					   ero,
					   resp)) {
			return false;
		}

		//create LSP object
		if (isCallIdentValid(callId, adjType()) > 0) {
			lsp = createLSP_(nextIdent,
					 &iTnaRes,
					 &eTnaRes,
					 "",
					 &callId,
					 lspInfo,
					 &recoveryInfo,
					 0,
					 true);
		} else {
			resp = "Call Ident is not valid. Cannot create LSP";
			return false;
		}

		if (!lsp) {
			resp = "Failed to regenerate LSP";
			return false;
		}

		Interface * intf;
		//set PSB and RSB infos
		if ((adjType() == INNI)                  ||
		    (adjType() == UNI  && lsp->isTail()) ||
		    (adjType() == ENNI && lsp->isTail())) {
			if (!lsp->isHead()) {
				//avoided in case of INNI head
				intf = findInterface(psbInfo.locTelId, false);
				if (!intf) {
					resp = "PSB local TELink doesn't belong to this instance interfaces";
					return false;
				}
				lsp->psb()->interface(intf);
			}
			lsp->psb()->localDL(psbInfo.locDlId);
			lsp->psb()->remoteDL(psbInfo.remDlId);
			lsp->psb()->txLabel(psbInfo.labelId);
			lsp->psb()->rxLabel(psbInfo.labelId);
		}
		//
		if ((adjType() == INNI)                  ||
		    (adjType() == UNI  && lsp->isHead()) ||
		    (adjType() == ENNI && lsp->isHead())) {
			if (!lsp->isTail()) {
				//avoided in case of INNI tail
				intf = findInterface(rsbInfo.locTelId, false);
				if (!intf) {
					resp = "RSB local TELink doesn't belong to this instance interfaces";
					return false;
				}
				lsp->rsb()->interface(intf);
			}
			lsp->rsb()->localDL(rsbInfo.locDlId);
			lsp->rsb()->remoteDL(rsbInfo.remDlId);
			lsp->rsb()->txLabel(rsbInfo.labelId);
			lsp->rsb()->rxLabel(rsbInfo.labelId);
		}
		if (adjType() == UNI ||
		    adjType() == ENNI ) {
			//Set call type in lsp instance
			lsp->callType(callInfo.call_type);

			//Set call info paramters in PathOut and PathIn
			if (!lsp->psb()->pathOut()->setGnsCallExt(callInfo)) {
				G2RSVPTE_ERR("Impossible to set GnsCallExt object !!");
				return 0;
			}
			lsp->psb()->pathIn()->gns_call_ext_ = lsp->psb()->pathOut()->gns_call_ext_->clone();

			//Set GnsTnas in PathOut and PathIn
			if (mask.iGnsTna) {
				if (!lsp->psb()->pathOut()->setGnsTna(iGnsTna, true)) {
					resp = "Impossible to set ingress GNS-TNA object";
					return false;
				}
				lsp->psb()->pathIn()->gns_uni_in_ = lsp->psb()->pathOut()->gns_uni_in_->clone();
			}
			if (mask.eGnsTna) {
				if (!lsp->psb()->pathOut()->setGnsTna(eGnsTna, false)) {
					resp = "Impossible to set egress GNS-TNA object";
					return false;
				}
				lsp->psb()->pathIn()->gns_uni_eg_ = lsp->psb()->pathOut()->gns_uni_eg_->clone();
			}

			//Set Policy data paramters in PathOut and PathIn
			if (mask.authz) {
				if (!lsp->psb()->pathOut()->setPolicyData(authz)) {
					resp = "Impossible to set PolicyData object";
					delete lsp;
					return false;
				}
				lsp->psb()->pathIn()->policy_data_ = lsp->psb()->pathOut()->policy_data_->clone();
			}
		}

		//add ERO to LSP
		std::list<lsp_ero_sobj_t>::const_iterator it;
		for (it = ero.begin(); it != ero.end(); it++) {
			EroSubObject * subObj = 0;

			subObj = createEroSubObj((*it));
			if (!subObj) {
				resp = "Impossible to create ERO SUBOBJ";
				delete lsp;
				return false;
			}
			if (!lsp->attachEroSubObj(subObj)) {
				resp = "Impossible to add ERO SUBOBJ to LSP";
				delete subObj;
				delete lsp;
				return false;
			}
		}

		// send a 'LspRebuild' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			resp = "Impossible to alocate context for LSP fsm";
			return false;
		}
		context->lsp  = lsp;
		context->intf = 0;
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_LspRebuild;

		lsp->fsmInst_->post(ev, context);

		return true;
	}
#endif // PERSISTENCY_ON

	/////////////////////////////////////////////////////
	//                Interface class                  //
	/////////////////////////////////////////////////////
	Interface::Interface(G2RSVPTE *    parent,
			     g2mpls_addr_t localId,
			     g2mpls_addr_t remoteId,
			     uint32_t      nId,
			     opstate_t     op_state,
			     admstate_t    adm_state) :
		Ancestor<true, G2RSVPTE>(parent)
	{
		addr_copy(&localId_, &localId);
		addr_copy(&remoteId_, &remoteId);
		neighbourId_ = nId;
		op_state_    = op_state;
		adm_state_   = adm_state;
	}

	void Interface::dump(void) const
	{
		G2RSVPTE_DBG("Interface: ");
		G2RSVPTE_DBG("   local     %s", addr_ntoa(localId_));
		G2RSVPTE_DBG("   remote    %s", addr_ntoa(remoteId_));
		G2RSVPTE_DBG("   neighbour %s", numb_htoa(neighbourId_));
		G2RSVPTE_DBG("   op_state  %s", SHOW_OPSTATE(op_state_));
		G2RSVPTE_DBG("   adm_state %s", SHOW_ADMSTATE(adm_state_));

	}

	std::ostringstream & operator << (std::ostringstream &    os,
					  const Interface &       intf)
	{
		os << "  Interface:"
		   << "\r\n";
		os << "    local     " << addr_ntoa(intf.localId())
		   << "\r\n";
		os << "    remote    " << addr_ntoa(intf.remoteId())
		   << "\r\n";
		os << "    neighbour " << numb_htoa(intf.neighbourId())
		   << "\r\n";
		os << "    op_state  " << SHOW_OPSTATE(intf.opState())
		   << "\r\n";
		os << "    admstate  " << SHOW_ADMSTATE(intf.admState())
		   << "\r\n";
		return os;
	}

	/////////////////////////////////////////////////////
	//                  Session class                  //
	/////////////////////////////////////////////////////
	Session::Session(G2RSVPTE * parent) : Ancestor<true, G2RSVPTE>(parent)
	{
		nodeId_   = 0;
		tunnelId_ = 0;
		extTunId_ = 0;
	}

	Session::Session(G2RSVPTE * parent,
			 uint32_t   nId,
			 uint32_t   tunId,
			 uint32_t   extTunId) :
		Ancestor<true, G2RSVPTE>(parent),
		nodeId_(nId),
		tunnelId_(tunId),
		extTunId_(extTunId) { }

	Session::~Session(void)
	{
		iterator_lsps it;
		for (it = begin_lsps(); it != end_lsps(); it++) {
			G2RSVPTE_DBG("Deleting LSP id 0x%llx",
				     (*it).first);
			DELETE_OBJ((*it).second);
		}
	}

	bool Session::attach(LSPKey_t k, LSP * l)
	{
		if (!l) {
			G2RSVPTE_ERR("Invalid LSP pointer");
			return false;
		}

		LSP * lsp;
		lsp = findLSP(k);
		if (lsp) {
			if (l == lsp) {
				G2RSVPTE_WRN("LSP already attached");
				return true;
			}

			G2RSVPTE_ERR("LSP Key 0x%llx already exists for "
				     "different LSP pointer",
				     k);
			return false;
		}

		lsps_[k] = l;
		return true;
	}

	bool Session::detach(LSPKey_t k, LSP * l)
	{
		if (!l) {
			G2RSVPTE_ERR("Invalid LSP pointer");
			return false;
		}

		iterator_lsps it;
		it = lsps_.find(k);
		if (it == lsps_.end() || (*it).second != l) {
			G2RSVPTE_WRN("Wrong LSP (key = 0x%llx) to detach",
				     k);
			return false;
		}

		lsps_.erase(it);
		G2RSVPTE_DBG("LSP key = 0x%llx detached", k);
		return true;
	}

	LSP * Session::findLSP(LSPKey_t key)
	{
		LSP * lsp;

		// Search into lsp  map
		iterator_lsps lIt;
		lIt = lsps_.find(key);
		if (lIt == lsps_.end()) {
			G2RSVPTE_WRN("No LSP found with key 0x%llx",
				     key);
			return 0;
		}

		lsp = (*lIt).second;
		return lsp;
	}

	LSP * Session::findLSP(uint32_t context)
	{
		LSP *          lsp;
		iterator_lsps  lIt;

		lsp = 0;

		// Search into lsp  map
		for (lIt  = lsps_.begin(); lIt != lsps_.end(); lIt++) {
			assert((*lIt).second);
			lsp = (*lIt).second;

			if (lsp->context() == context) {
				break;;
			}
		}

		return lsp;
	}

	//VERY TMP
	bool Session::checkLSP(LSP * lsp)
	{
		iterator_lsps     lIt;

		for (lIt  = lsps_.begin();
		     lIt != lsps_.end();
		     lIt++) {
			if ((*lIt).second == lsp) {
				return true;
			}
		}

		return false;
	}

	std::ostringstream & operator << (std::ostringstream & os,
					  const Session &      session)
	{
		os << "    dst node id      : " << numb_htoa(session.nodeId())
		   << "\r\n";
		os << "    tunnel id        : " << session.tunnelId()
		   << "\r\n";
		os << "    ext tunnel id    : " << session.extTunId()
		   << "\r\n";
		return os;
	}


	/////////////////////////////////////////////////////
	//              UpstreamAckNack class              //
	/////////////////////////////////////////////////////
	bool UpstreamAckNack::sendMsg(Message * msg)
	{
		if (!msg || !parent() || !parent()->psb() ||
		    !parent()->psb()->interface()
		    //|| !parent()->psb()->nodeId()
		    ) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}
		G2RSVPTE * proto;
		proto = parent()->getInstance();
		if (!proto) {
			G2RSVPTE_ERR("Null instance pointer!!!");
			return false;
		}

		// TMP before implementing ack/nack FSM
		if (useFsm_) {
			// TODO when Ack/Nack FSM is available
			return false;
		} else {
			// Send message
			// TODO FIXME!!!
			//        INTF   = parent()->psb()->interface();
			//        NODEID = parent()->psb()->nodeId()

			return scngw_.write(msg, parent()->psb()->interface(), proto->scngwFd());
		}
	}

	bool UpstreamAckNack::errorHandle(int  flags,
					  int  errCode,
					  int  errVal,
					  bool addInfo,
					  bool xcErr)
	{
		assert(parent()->psb());
		assert(parent()->psb()->pathOut());
		assert(parent()->getInstance());

		// Check: flags   (must be  8 bits)
		//        errCode (must be  8 bits)
		//        errVal  (must be 16 bits)
		if (flags & 0xffffff00 || errCode & 0xffffff00 ||
		    errVal & 0xffff0000) {
			G2RSVPTE_ERR("some bad value(s)");
			return false;
		}

		// If we need to know last error for LSP -> write it down
		//parent()->lastErrInfo.time   = global_tod;
		//parent()->lastErrInfo.msg   = ERRORMSG_PATHERR;
		//parent()->lastErrInfo.nid = parent()->getInstance()->nodeId();
		//parent()->lastErrInfo.flags = flags;
		//parent()->lastErrInfo.code  = errCode;
		//parent()->lastErrInfo.value = errVal;

		//
		// PathErr case
		//
		if (parent()->isHead()) {
			G2RSVPTE_ERR("I am the src node, no PathErr to send??");
			return false;
		}

		if (parent()->psb()->pathErr()) {
			G2RSVPTE_WRN("PathErr already exists ... "
				     "deleting it before creating again");
			delete parent()->psb()->pathErr();
			parent()->psb()->pathErr(0);
		}

		PathErrMessage * pathErr;
		// Set NOT_ERASE flags for this pkg
		pathErr = new PathErrMessage(*parent()->psb()->pathOut());
		// TODO FIXME: the copy must involve only: check the constructor
		if (!pathErr) {
			G2RSVPTE_ERR("could not allocate PathErr msg");
			return false;
		}

		//
		// Create Error Spec
		//
		IFIDIPv4ErrorSpec_Object * tmpErrSpec;
		tmpErrSpec = new IFIDIPv4ErrorSpec_Object(parent()->getInstance()->
							  nodeId(),
							  flags,
							  errCode,
							  errVal);
		if (!tmpErrSpec) {
			G2RSVPTE_ERR("Impossible to allocate errorSpec subObj");
			delete pathErr;
			return false;
		}

#ifdef CRANKBACK_ON
		if (parent()->getInstance()->adjType() == INNI &&
		    parent()->cbackScope() != CB_NONE          &&
		    addInfo) {
			if (xcErr) {
				//add TLVs for in TElink
				IPv4_TLV * inTel;
				inTel = new IPv4_TLV();
				if (!inTel) {
					G2RSVPTE_ERR("Impossible to allocate "
						     "memory for IPv4 TLV");
					return false;
				}
				inTel->get().ipv4_addr.s_addr =
					parent()->psb()->interface()->localId().value.ipv4.s_addr;
				//Set tlv
				tmpErrSpec->get().tlvs.push_back(inTel);
				//add TLVs for out TElink
				IPv4_TLV * outTel;
				outTel = new IPv4_TLV();
				if (!outTel) {
					G2RSVPTE_ERR("Impossible to allocate "
						     "memory for IPv4 TLV");
					return false;
				}
				outTel->get().ipv4_addr.s_addr =
					parent()->rsb()->interface()->localId().value.ipv4.s_addr;
				//Set tlv
				tmpErrSpec->get().tlvs.push_back(outTel);
			} else {
				//add 1 or 2 TLVs for TElink/label errored
				if (!is_addr_null(parent()->errTELink())) {
					IPv4_TLV * errTel;
					errTel = new IPv4_TLV();
					if (!errTel) {
						G2RSVPTE_ERR("Impossible to allocate "
							     "memory for IPv4 TLV");
						return false;
					}
					errTel->get().ipv4_addr.s_addr =
						parent()->errTELink().value.ipv4.s_addr;
					//Set tlv
					tmpErrSpec->get().tlvs.push_back(errTel);
					if (parent()->errLabel() != 0 &&
					    parent()->errLabel() != 1) {
						DOWN_LABEL_TLV * lbl;
						lbl = new DOWN_LABEL_TLV();
						if (!lbl) {
							G2RSVPTE_ERR("Impossible to allocate "
								     "memory for LABEL TLV");
							return false;
						}
						lbl->get().label = parent()->errLabel();
						//Set tlv
						tmpErrSpec->get().tlvs.push_back(lbl);
					}
				}
			}
			parent()->resetErr();
		}
#endif //CRANKBACK_ON

		// RRO
		if (parent()->psb()->pathOut()->rro_) {
			pathErr->rro_ = new RecordedRoute_Object;
			assert(pathErr->rro_);
			if (pathErr->rro_->cType() != CTYPE_RECORDED_ROUTE) {
				G2RSVPTE_ERR("For now, only ctype_rro is allowed");
				return false;
			}
			//add this node to the RRO
			RecordedRoute_Object * genRRO;
			genRRO = static_cast<RecordedRoute_Object *>(pathErr->rro_);

			if (parent()->psb()->interface() &&
			    !parent()->insertIntoRRO(genRRO,
						   parent()->psb()->interface()->localId())) {
				G2RSVPTE_ERR("Problems while preparing outgoing Path RRO obj");
				return false;
			}
		}

		// Set error spec into PathErr msg
		pathErr->error_spec_ = tmpErrSpec;

		//
		// Set PathErr msg into PSB
		//
		parent()->psb()->pathErr(pathErr);

		G2RSVPTE_INF("sending(US) PathErr msg (errCode = %d, "
			     "errVal = %d)",
			     errCode, errVal);

		if (!sendMsg(parent()->psb()->pathErr())) {
			G2RSVPTE_ERR("could not deliver PathErr msg (errCode = "
				     "%d, errVal = %d) to upstream ACK/NACK",
				     errCode, errVal);
			return false;
		}
		// set isErrored flag in LSP object
		parent()->isErrored(true);

		return true;
	}


	/////////////////////////////////////////////////////
	//             DownstreamAckNack class             //
	/////////////////////////////////////////////////////
	bool DownstreamAckNack::sendMsg(Message * msg)
	{
		if (!msg || !parent() || !parent()->rsb() ||
		    !parent()->rsb()->interface()
		    //|| !parent()->rsb()->nodeId()
		    ) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}
		G2RSVPTE * proto;
		proto = parent()->getInstance();
		if (!proto) {
			G2RSVPTE_ERR("Null instance pointer!!!");
			return false;
		}

		// TMP before implementing ack/nack FSM
		if (useFsm_) {
			// TODO when Ack/Nack FSM is available
			return false;
		} else {
			// Send message
			// TODO FIXME!!!
			//  using INTF   = parent()->rsb()->interface()
			//        NODEID = parent()->rsb()->nodeId()

			return scngw_.write(msg, parent()->rsb()->interface(), proto->scngwFd());
		}
	}

	bool DownstreamAckNack::errorHandle(int  flags,
					    int  errCode,
					    int  errVal,
					    bool addInfo,
					    bool xcErr)
	{
		assert(parent()->psb());
		assert(parent()->psb()->pathOut());
		assert(parent()->getInstance());

		// Check: flags   (must be  8 bits)
		//        errCode (must be  8 bits)
		//        errVal  (must be 16 bits)
		if (flags & 0xffffff00 || errCode & 0xffffff00 ||
		    errVal & 0xffff0000) {
			G2RSVPTE_ERR("some bad value(s)");
			return false;
		}

		// If we need to know last error for LSP -> write it down
		//parent()->lastErrInfo.time   = global_tod;
		//parent()->lastErrInfo.msg   = ERRORMSG_PATHERR; //RESV????
		//parent()->lastErrInfo.nid   = parent()->getInstance()->nodeId();
		//parent()->lastErrInfo.flags = flags;
		//parent()->lastErrInfo.code  = errCode;
		//parent()->lastErrInfo.value = errVal;

		//
		// ResvErr case
		//
		int count = 0;

		// if the ResvErr PSR flag is set, then we will send a PathErr
		// upstream and a PathTear downstream instead of the ResvErr
		if (flags & G2RSVPTE_ERROR_SPEC_FLAG_PathStateRemoved) {
			// send a PathTear downstream (instead of ResvErr)
			if (!parent()->isTail()) {
				// Build the PathTear msg from the PSB PathOut
				if (!parent()->psb()->pathOut()) {
					G2RSVPTE_ERR("Invalid PathOut msg");
					return false;
				}

				if (parent()->psb()->pathTear()) {
					G2RSVPTE_WRN("Freeing path tear msg before allocating");
					delete parent()->psb()->pathTear();
					parent()->psb()->pathTear(0);
				}

				// Set NOT_ERASE flags for this pkg
				parent()->psb()->pathTear(new RSVP::PathTearMessage(*parent()->psb()->pathOut()));

				if (!parent()->psb()->pathTear()) {
					G2RSVPTE_ERR("could not allocate outgoing PathTear msg");
					return false;
				}

				G2RSVPTE_INF("sending(DS) PathTear msg to NE %s ...",
					     numb_htoa(parent()->rsb()->nodeId()));
				G2RSVPTE_INF("... through TE link %s",
					     addr_ntoa(parent()->rsb()->interface()->localId()));

				if (!sendMsg(parent()->psb()->pathTear())) {
					G2RSVPTE_ERR("could not deliver PathTear msg to the DS ACK FSM");
					//return false;
				} else {
					count++;
					// set isErrored flag in LSP object
					parent()->isErrored(true);
				}

				//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
				//delete parent()->psb()->pathTear();
				//parent()->psb()->pathTear(0);
			}

			//SAFEDEL_ON
			// don't send PathErr upstream if ds deletion failed

			// Send a PathErr with PSR upstream
			if (!parent()->isHead()) {
				// Build the PathErr msg from the PSB PathOut
				if (!parent()->psb()->pathOut()) {
					G2RSVPTE_ERR("Invalid PathOut msg");
					return (count == 0) ? false : true;  //true if at least pathTear has been sent
				}

				if (parent()->psb()->pathErr()) {
					G2RSVPTE_WRN("PathErr already exists ... deleting it before creating again");
					delete parent()->psb()->pathErr();
					parent()->psb()->pathErr(0);
				}

				PathErrMessage * pathErr;
				// Set NOT_ERASE flags for this pkg
				pathErr = new PathErrMessage(*parent()->psb()->pathOut());
				// TODO FIXME: the copy must involve only: check the constructor
				if (!pathErr) {
					G2RSVPTE_ERR("could not allocate PathErr msg");
					return (count == 0) ? false : true;  //true if at least pathTear has been sent
				}

				//
				// Create Error Spec
				//
				IFIDIPv4ErrorSpec_Object * tmpErrSpec;
				tmpErrSpec = new IFIDIPv4ErrorSpec_Object(parent()->getInstance()->
									  nodeId(),
									  flags,
									  errCode,
									  errVal);
				if (!tmpErrSpec) {
					G2RSVPTE_ERR("Impossible to allocate errorSpec subObj");
					delete pathErr;
					return false;
				}

#ifdef CRANKBACK_ON
				if (parent()->getInstance()->adjType() == INNI &&
				    parent()->cbackScope() != CB_NONE             &&
				    addInfo) {
					if (xcErr) {
						//add TLVs for in TElink
						IPv4_TLV * inTel;
						inTel = new IPv4_TLV();
						if (!inTel) {
							G2RSVPTE_ERR("Impossible to allocate "
								     "memory for IPv4 TLV");
							return false;
						}
						inTel->get().ipv4_addr.s_addr =
							parent()->psb()->interface()->localId().value.ipv4.s_addr;
						//Set tlv
						tmpErrSpec->get().tlvs.push_back(inTel);
						//add TLVs for out TElink
						IPv4_TLV * outTel;
						outTel = new IPv4_TLV();
						if (!outTel) {
							G2RSVPTE_ERR("Impossible to allocate "
								     "memory for IPv4 TLV");
							return false;
						}
						outTel->get().ipv4_addr.s_addr =
							parent()->rsb()->interface()->localId().value.ipv4.s_addr;
						//Set tlv
						tmpErrSpec->get().tlvs.push_back(outTel);
					} else {
						//add 1 or 2 TLVs for TElink/label errored
						if (!is_addr_null(parent()->errTELink())) {
							IPv4_TLV * errTel;
							errTel = new IPv4_TLV();
							if (!errTel) {
								G2RSVPTE_ERR("Impossible to allocate "
									     "memory for IPv4 TLV");
								return false;
							}
							errTel->get().ipv4_addr.s_addr =
								parent()->errTELink().value.ipv4.s_addr;
							//Set tlv
							tmpErrSpec->get().tlvs.push_back(errTel);
							if (parent()->errLabel() != 0 &&
							    parent()->errLabel() != 1) {
								DOWN_LABEL_TLV * lbl;
								lbl = new DOWN_LABEL_TLV();
								if (!lbl) {
									G2RSVPTE_ERR("Impossible to allocate "
										     "memory for LABEL TLV");
									return false;
								}
								lbl->get().label = parent()->errLabel();
								//Set tlv
								tmpErrSpec->get().tlvs.push_back(lbl);
							}
						}
					}
					parent()->resetErr();
				}
#endif //CRANKBACK_ON

				// Set error spec into PathErr msg
				pathErr->error_spec_ = tmpErrSpec;

				//
				// Set PathErr msg into PSB
				//
				parent()->psb()->pathErr(pathErr);

				G2RSVPTE_INF("sending(US) PathErr msg (errCode = %d, errVal = %d)",
					     errCode, errVal);

				if (!parent()->usAckNack() ||
				    !parent()->usAckNack()->sendMsg(parent()->psb()->pathErr())) {
					G2RSVPTE_ERR("could not deliver PathErr msg (errCode = %d, "
						     "errVal = %d) to upstream ACK/NACK",
						     errCode, errVal);
					return (count == 0) ? false : true;  //true if at least pathTear has been sent
				}
				// set isErrored flag in LSP object
				parent()->isErrored(true);

				//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
				//delete parent()->psb()->pathErr();
				//parent()->psb()->pathErr(0);
			}

		} else if (!parent()->isTail()) {

			// Build the ResvErr msg from the RSB ResvIn
			if (!parent()->rsb()->resvIn()) {
				G2RSVPTE_ERR("Invalid ResvIn msg");
				return false;
			}

			if (parent()->rsb()->resvErr()) {
				G2RSVPTE_WRN("ResvErr already exists ... deleting it before creating again");
				delete parent()->rsb()->resvErr();
				parent()->rsb()->resvErr(0);
			}

			ResvErrMessage * resvErr;
			// Set NOT_ERASE flags for this pkg
			resvErr = new ResvErrMessage(*parent()->rsb()->resvIn());

			if (!resvErr) {
				G2RSVPTE_ERR("could not allocate ResvErr msg");
				return false;
			}

			//
			// Create Error Spec, for now IPv4 but with cb IF_ID_IPv4 ?????
			//
			IPv4ErrorSpec_Object * tmpErrSpec = new IPv4ErrorSpec_Object();
			if (!tmpErrSpec) {
				G2RSVPTE_ERR("Impossible to allocate error spec subObj");
				delete resvErr;
				return false;
			}
			tmpErrSpec->get().ipv4_err_node.s_addr =
				htonl(parent()->getInstance()->nodeId());
			tmpErrSpec->get().flags            = flags;
			tmpErrSpec->get().error_code       = errCode;
			tmpErrSpec->get().error_value      = errVal;
			// Set error spec into ResvErr msg
			resvErr->error_spec_ = tmpErrSpec;

			//
			// Set ResvErr msg into RSB
			//
			parent()->rsb()->resvErr(resvErr);

			G2RSVPTE_INF("sending(DS) ResvErr msg (errCode = %d, errVal = %d)",
				     errCode, errVal);

			if (!sendMsg(parent()->rsb()->resvErr())) {
				G2RSVPTE_ERR("could not deliver ResvErr msg (errCode = %d, "
					     "errVal = %d) to downstream ACK/NACK",
					     errCode, errVal);
				return false;
			}
			// set isErrored flag in LSP object
			parent()->isErrored(true);

			//TODO: FIXME!!! cancel message (if flags NOT_ERASE is not set)
			//delete parent()->rsb()->resvErr();
			//parent()->rsb()->resvErr(0);
		}

		return true;
	}


	/////////////////////////////////////////////////////
	//                   LSP class                     //
	/////////////////////////////////////////////////////
	LSP::LSP(Session * parent) : Ancestor<true, Session>(parent)
	{
		nid_ = 0;
		id_  = 0;

		initConf();
	}

	LSP::LSP(Session * parent,
		 uint32_t  nId,
		 uint32_t  lspId) :
		Ancestor<true, Session>(parent),
		nid_(nId),
		id_(lspId)
	{
		initConf();
	}

	LSP::~LSP(void)
	{
		DELETE_OBJ(lspCtrl_);
		DELETE_OBJ(us_AckNack_);
		DELETE_OBJ(ds_AckNack_);
		DELETE_OBJ(psb_);
		DELETE_OBJ(rsb_);

		//XXX FIXME!!!!!!
		//Deleting timer
		// refreshInterval_;
		// rapidRetransInterval_;
		// rapidRetryLimit_;
		// expoBackoffDelta_;

		DELETE_OBJ(fsmInst_);

		if (timerThr_) {
			thread_cancel(timerThr_);
		}
	}

	bool LSP::initConf(void)
	{
		us_AckNack_ = 0;
		ds_AckNack_ = 0;
		psb_        = 0;
		rsb_        = 0;
		lspCtrl_    = 0;

		fsmInst_    = 0;
		//enabled_    = false;
		enabled_       = true;
		toBeDestroyed_ = false;
		isErrored_     = false;
		timerThr_      = 0;
		errLabel_      = 0;

		//memset(&ingress_tna_, 0, sizeof(ingress_tna_));
		//memset(&egress_tna_, 0, sizeof(egress_tna_));

		memset(&times_, 0, sizeof(times_));
		memset(&in_net_res_, 0, sizeof(in_net_res_));
		memset(&eg_net_res_, 0, sizeof(eg_net_res_));
		memset(&callId_, 0, sizeof(callId_));
		memset(&callId_old_, 0, sizeof(callId_old_));
		memset(&errTELink_, 0, sizeof(errTELink_));

		refreshInterval_      = 0;
		rapidRetransInterval_ = 0;
		rapidRetryLimit_      = 0;
		expoBackoffDelta_     = 0;

		tnrcCookie_           = 0;

		callType_             = CALL_TYPE_AUTO;
#ifdef CRANKBACK_ON
		cbScope_              = CB_NONE;
		currentCbackAtt_      = 0;
		maxCbackRetries_      = MAX_CBACK_RETRIES_ATTEMPTS;
		cbackWaitRelRes_      = false;
#endif // CRANKBACK_ON
		// XXX-FIXME: substitute with a UUID generation here
		struct timeval unixTime;

		if (gettimeofday(&unixTime, NULL)) {
			G2RSVPTE_ERR("ERROR on gettimeofday....");
			assert(0);
		}

		context_ = (uint32_t) unixTime.tv_sec;

		return true;
	}

	bool LSP::updateCallId(call_ident_t Id)
	{
		//update callId in this LSP instance
		callId_ = Id;

		//update callId in lsp->psb()->pathIn()
		assert(psb_->pathIn());

		if (!psb_->pathIn()->call_id_) {
			G2RSVPTE_ERR("Null CallId pointer ... strange !");
			return false;
		}
		//delete old callId
		RSVP::CallId_Class * tmp;
		tmp = psb_->pathIn()->call_id_;
		delete tmp;
		//create new callId
		char * addr     = 0;
		bit8_t addrLen  = 0;
		bit8_t addrType = 0;

		if (Id.src_addr.type == IPv4) {
			addrType = 0x01;
			addrLen  = 4;
			addr     = new char[addrLen];
			assert(addr);
			memcpy(addr, (void *) &Id.src_addr.value.ipv4.s_addr, addrLen);
		} else if (Id.src_addr.type == IPv6) {
			addrType = 0x02;
			addrLen  = 16;
			addr     = new char[addrLen];
			assert(addr);
			memcpy(addr, (void *) &Id.src_addr.value.ipv6.s6_addr32[0], addrLen);
		}

		switch (Id.type) {
			case CALLID_OPSPEC:
				psb_->pathIn()->call_id_ = new RSVP::CallIdOpSpec_Object(addrType,
											 addrLen,
											 addr,
											 Id.local_id);
				assert(psb_->pathIn()->call_id_);
				break;
			case CALLID_GLOBUNIQ:
				psb_->pathIn()->call_id_ = new RSVP::CallIdGlobal_Object(addrType,
											 0,
											 0,
											 0,
											 addrLen,
											 addr,
											 Id.local_id);
				assert(psb_->pathIn()->call_id_);
				break;
			case CALLID_NULL:
				psb_->pathIn()->call_id_ = new RSVP::CallIdNull_Object(addrType,
										       addrLen,
										       addr,
										       Id.local_id);
				assert(psb_->pathIn()->call_id_);
				break;
			default:
				break;
		}

		return true;
	}

	lsp_ident_t LSP::getIdent(void)
	{
		assert(parent());
		lsp_ident_t ident;

		ident.dst_nid    = parent()->nodeId();
		ident.src_nid    = nid_;
		ident.tun_id     = parent()->tunnelId();
		ident.ext_tun_id = parent()->extTunId();
		ident.lsp_id     = id_;

		return ident;
	}

	bool LSP::getInfo(lsp_info_t & info)
	{
		if (!psb_ || !psb_->pathIn()) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		//Get swcap, enctype and gpid
		if (!psb_->pathIn()->label_request_) {
			G2RSVPTE_ERR("Null GLabelRequest pointer!!!");
			return false;
		}
		LabelRequest_Class * lReq;
		lReq = psb_->pathIn()->label_request_;
		GLabelRequest_Object * glReq;
		glReq = static_cast<GLabelRequest_Object *> (lReq);
		assert(glReq);
		info.sw_cap   = (sw_cap_t) glReq->get().enc_type;
		info.enc_type = (enc_type_t) glReq->get().switch_type;
		info.gpid     = (gpid_t) glReq->get().g_pid;

		//Get bandwidth
		if (!psb_->pathIn()->sender_tspec_) {
			G2RSVPTE_ERR("Null SenderTSpec pointer!!!");
			return false;
		}
		SenderTSpec_Class * sTSpec;
		sTSpec = psb_->pathIn()->sender_tspec_;
		SenderTSpec_Object * sTSpecObj;
		sTSpecObj = static_cast<SenderTSpec_Object *> (sTSpec);
		assert(sTSpecObj);
		info.bw = (uint32_t) sTSpecObj->get().peak_data_rate;

		//Get Session info
		if (!psb_->pathIn()->session_attribute_) {
			G2RSVPTE_ERR("Null SessionAttribute pointer!!!");
			return false;
		}
		Session_Attr_Class * sAttr;
		sAttr = psb_->pathIn()->session_attribute_;
		LSPTunnelRASessionAttr_Object * lspTunAttr;
		lspTunAttr = static_cast<LSPTunnelRASessionAttr_Object *> (sAttr);
		assert(lspTunAttr);
		info.setup_prio   = lspTunAttr->get().setup_prio;
		info.holding_prio = lspTunAttr->get().hold_prio;
		info.include_all  = lspTunAttr->get().include_all;
		info.include_any  = lspTunAttr->get().include_any;
		info.exclude_any  = lspTunAttr->get().exclude_any;

		//Get protection info
		if (!psb_->pathIn()->protection_) {
			info.link_prot_mask = PROTTYPE_NONE;
		} else {
			Protection_Class * prot;
			prot = psb_->pathIn()->protection_;
			Protection_Object * protObj;
			protObj = static_cast<Protection_Object *> (prot);
			assert(protObj);
			info.link_prot_mask = (gmpls_prottype_t) protObj->get().link_flags;
		}
#ifdef CRANKBACK_ON
		//Get crankback info
		if (!psb_->pathIn()->lsp_attributes_) {
			info.crankback = CB_NONE;
			info.max_cback_retries_src = 0;                       ;
		} else {
			info.crankback = cbScope_;
			info.max_cback_retries_src = maxCbackRetries();
		}
#else
		info.crankback = CB_NONE;
		info.max_cback_retries_src   = 0;
#endif // CRANKBACK_ON
		info.max_cback_retries_intmd = 0;

		//Get type, role, action and rro mode
		////TMP
		info.type     = LSP_TYPE_SPC;
		info.role     = LSP_ROLE_PRIMARY;
		info.action   = RESOURCE_XCONNECT;
		info.rro_mode = LSP_RRO_TEL_DETAIL;
		////TMP

		//Get Time values
		if (!psb_->pathIn()->time_values_) {
			G2RSVPTE_ERR("Null TimeValues pointer!!!");
			return false;
		}
		TimeValues_Class * tValue;
		tValue = psb_->pathIn()->time_values_;
		RSVP::TimeValues_Object * tValueObj;
		tValueObj = static_cast<TimeValues_Object *> (tValue);
		assert(tValue);
		info.refresh_interval        = tValueObj->get().refresh_period;
		////TMP
		info.rapid_retransm_interval = 0;
		info.rapid_retry_limit       = 0;
		info.increment_value_delta   = 0;
		////

		//Get activate ack
		////TMP
		info.activate_ack = 0;
		////

		//Get Times info
		if (psb_->pathIn()->times_) {
			Times_Class * times;
			times = psb_->pathIn()->times_;
			Times_Object * timesObj;
			timesObj = static_cast<Times_Object *> (times);
			assert(timesObj);
			info.times.start_time = timesObj->get().start_time;
			info.times.end_time   = timesObj->get().end_time;
		}

		BITMASK_BITSET(info.mask_, sw_cap);
		BITMASK_BITSET(info.mask_, enc_type);
		BITMASK_BITSET(info.mask_, gpid);
		BITMASK_BITSET(info.mask_, bw);
		BITMASK_BITSET(info.mask_, setup_prio);
		BITMASK_BITSET(info.mask_, holding_prio);
		BITMASK_BITSET(info.mask_, exclude_any);
		BITMASK_BITSET(info.mask_, include_any);
		BITMASK_BITSET(info.mask_, include_all);
		BITMASK_BITSET(info.mask_, link_prot_mask);
		BITMASK_BITSET(info.mask_, crankback);
		BITMASK_BITSET(info.mask_, max_cback_retries_src);
		BITMASK_BITSET(info.mask_, max_cback_retries_intmd);
		BITMASK_BITSET(info.mask_, type);
		BITMASK_BITSET(info.mask_, role);
		BITMASK_BITSET(info.mask_, action);
		BITMASK_BITSET(info.mask_, rro_mode);
		BITMASK_BITSET(info.mask_, refresh_interval);
		BITMASK_BITSET(info.mask_, activate_ack);
		BITMASK_BITSET(info.mask_, rapid_retransm_interval);
		BITMASK_BITSET(info.mask_, rapid_retry_limit);
		BITMASK_BITSET(info.mask_, increment_value_delta);
		BITMASK_BITSET(info.mask_, times);

		return true;
	}

	bool LSP::getCallInfo(call_info_t & info)
	{
		RSVP::GnsCallExt_Class * gnsCall;

		if (isTail()) {
			if (!psb_ || !psb_->pathIn() ||
			    !psb_->pathIn()->gns_call_ext_ ||
			    !psb_->pathIn()->gen_uni_) {
				G2RSVPTE_ERR("Null pointer!!!");
				return false;
			}
			gnsCall = psb_->pathIn()->gns_call_ext_;
		} else {
			if (!psb_ || !psb_->pathOut() ||
			    !psb_->pathOut()->gns_call_ext_ ||
			    !psb_->pathOut()->gen_uni_) {
				G2RSVPTE_ERR("Null pointer!!!");
				return false;
			}
			gnsCall = psb_->pathOut()->gns_call_ext_;
		}

		RSVP::GnsCallExt_Object * gnsCallObj;
		gnsCallObj = static_cast<RSVP::GnsCallExt_Object *> (gnsCall);

		switch (gnsCallObj->get().call_type) {
			case 0x00000000:
				info.call_type = CALL_TYPE_SPC;
				break;
			case 0x00000001:
				info.call_type = CALL_TYPE_PC;
				break;
			case 0x00000002:
				info.call_type = CALL_TYPE_SC;
				break;
			case 0x00000003:
				info.call_type = CALL_TYPE_AUTO;
				break;
			case 0x00000004:
				info.call_type = CALL_TYPE_aUGWzUGW;
				break;
			case 0x00000005:
				info.call_type = CALL_TYPE_aMGTzEGW;
				break;
			case 0x00000006:
				info.call_type = CALL_TYPE_aUGWzEGW;
				break;
			case 0x00000007:
				info.call_type = CALL_TYPE_aEGWzMGT;
				break;
			case 0x00000008:
				info.call_type = CALL_TYPE_aEGWzUGW;
				break;
			case 0x00000009:
				info.call_type = CALL_TYPE_aEGWzEGW;
				break;
			default:
				G2RSVPTE_ERR("Invalid Call type parameter (0x%x)"
					     "in GnsCallExt Object", gnsCallObj->get().call_type);
				return false;
		}

		info.times.start_time = gnsCallObj->get().start_time;;
		info.times.end_time   = info.times.start_time + gnsCallObj->get().duration;
		info.job_name         = new std::string(gnsCallObj->get().job_name);
		info.job_project      = new std::string(gnsCallObj->get().job_project);
		info.call_name        = new std::string("");

		BITMASK_BITSET(info.mask_, call_type);
		BITMASK_BITSET(info.mask_, times);
		BITMASK_BITSET(info.mask_, job_name);
		BITMASK_BITSET(info.mask_, job_project);
		BITMASK_BITSET(info.mask_, call_name);

		return true;
	}

	std::string LSP::sessionName(void) const
	{
		//TODO if sess_attr obj is not present we should check sess obj
		std::string name = "NO SESSION NAME";
		if (psb_ && psb_->pathIn() &&
		    psb_->pathIn()->session_attribute_) {

			Session_Attr_Class * sess;
			sess = psb_->pathIn()->session_attribute_;

			char * pName = 0;
			size_t len   = 0;

			switch (sess->cType()) {
				case CTYPE_SESSION_ATTRIBUTE_LSP_TUNNEL: {
					pName =(static_cast
						<LSPTunnelSessionAttr_Object *>
						(sess))->get().name;
					len =(static_cast
					      <LSPTunnelSessionAttr_Object *>
					      (sess))->get().name_len;
					break;
				}
				case CTYPE_SESSION_ATTRIBUTE_LSP_TUNNEL_RA: {
					pName =(static_cast
						<LSPTunnelRASessionAttr_Object*>
						(sess))->get().name;
					len =(static_cast
					      <LSPTunnelRASessionAttr_Object*>
					      (sess))->get().name_len;
					break;
				}
				default:
					pName = 0;
					break;
			}

			if (pName && len > 0) {
				name = std::string(pName, len);
			}
		}

		return name;
	}

	bool LSP::attach(UpstreamAckNack * u)
	{
		if (!u) {
			G2RSVPTE_ERR("Invalid Upstream AckNack pointer");
			return false;
		}

		if (us_AckNack_) {
			if (us_AckNack_ == u) {
				G2RSVPTE_WRN("US AckNack already attached");
				return true;
			}

			G2RSVPTE_ERR("Different US AckNack already exists");
			return false;
		}

		us_AckNack_ = u;
		return true;
	}

	bool LSP::detach(UpstreamAckNack * u)
	{
		if (!u) {
			G2RSVPTE_ERR("Invalid Upstream AckNack pointer");
			return false;
		}

		if (us_AckNack_ != u) {
			G2RSVPTE_WRN("Wrong US AckNack to detach");
			return false;
		}

		if (us_AckNack_ == 0) {
			G2RSVPTE_ERR("US AckNack already detached");
		} else {
			us_AckNack_ = 0;
			G2RSVPTE_DBG("US AckNack detached");
		}

		return true;
	}

	bool LSP::attach(DownstreamAckNack * d)
	{
		if (!d) {
			G2RSVPTE_ERR("Invalid Downstream AckNack pointer");
			return false;
		}

		if (ds_AckNack_) {
			if (ds_AckNack_ == d) {
				G2RSVPTE_WRN("DS AckNack already attached");
				return true;
			}

			G2RSVPTE_ERR("Different DS AckNack already exists");
			return false;
		}

		ds_AckNack_ = d;
		return true;
	}


	bool LSP::detach(DownstreamAckNack * d)
	{
		if (!d) {
			G2RSVPTE_ERR("Invalid Downstream AckNack pointer");
			return false;
		}

		if (ds_AckNack_ != d) {
			G2RSVPTE_WRN("Wrong DS AckNack to detach");
			return false;
		}

		if (ds_AckNack_ == 0) {
			G2RSVPTE_ERR("DS AckNack already detached");
		} else {
			ds_AckNack_ = 0;
			G2RSVPTE_DBG("DS AckNack detached");
		}

		return true;
	}

	bool LSP::attach(PSB * psb)
	{
		if (!psb) {
			G2RSVPTE_ERR("Invalid PSB pointer");
			return false;
		}

		if (psb_) {
			if (psb_ == psb) {
				G2RSVPTE_WRN("PSB already attached");
				return true;
			}

			G2RSVPTE_ERR("Different PSB already exists");
			return false;
		}

		psb_ = psb;
		return true;
	}


	bool LSP::detach(PSB * psb)
	{
		if (!psb) {
			G2RSVPTE_ERR("Invalid PSB pointer");
			return false;
		}

		if (psb_ != psb) {
			G2RSVPTE_WRN("Wrong PSB to detach");
			return false;
		}

		if (psb_ == 0) {
			G2RSVPTE_ERR("PSB already detached");
		} else {
			psb_ = 0;
			G2RSVPTE_DBG("PSB detached");
		}

		return true;
	}

	bool LSP::attach(RSB * rsb)
	{
		if (!rsb) {
			G2RSVPTE_ERR("Invalid RSB pointer");
			return false;
		}

		if (rsb_) {
			if (rsb_ == rsb) {
				G2RSVPTE_WRN("RSB already attached");
				return true;
			}

			G2RSVPTE_ERR("Different RSB already exists");
			return false;
		}

		rsb_ = rsb;
		return true;
	}

	bool LSP::detach(RSB * rsb)
	{
		if (!rsb) {
			G2RSVPTE_ERR("Invalid RSB pointer");
			return false;
		}

		if (rsb_ != rsb) {
			G2RSVPTE_WRN("Wrong RSB to detach");
			return false;
		}

		if (rsb_ == 0) {
			G2RSVPTE_ERR("RSB already detached");
		} else {
			rsb_ = 0;
			G2RSVPTE_DBG("RSB detached");
		}

		return true;
	}

	//
	// Set   - rsb_->localDL
	// Check - rsb_->txLabel   [ -> with params from LRM ]
	//
	bool LSP::setDsDL_checkDsTxLbl_fromTELink(g2mpls_addr_t teLink,
						  uint32_t      label)
	{
		//TODO??? find the downstream DL and label
		//        from rsb_->interface() and ...

		g2mpls_addr_t dsDL;
		if (!getDL_checkLabel_on_TELink(teLink,
						label,
						dsDL)) {
			G2RSVPTE_ERR("Failed to get DL and check label from %s",
				     addr_ntoa(teLink/*rsb_->interf->locId*/));
			return false;
		}

		rsb_->localDL(dsDL);     // Downstream DL
		rsb_->txLabel(label);    // Downstream tx label

		return true;
	}

	//
	// Set - rsb_->localDL
	//     - rsb_->txLabel   [ -> with params from LRM ]
	//
	bool LSP::setDsDL_DsTxLbl_fromTELink(g2mpls_addr_t teLink)
	{
		//TODO??? find the downstream DL and label
		//        from rsb_->interface() and ...

		g2mpls_addr_t dsDL;
		uint32_t      dsTxLbl;
		if (!getDL_Label_on_TELink(teLink, //rsb_->interface()->localId,
					   //psb_->pathOut()->label_request_,
					   //psb_->pathOut()->sender_tspec_,
					   //psb_->localDL(), // Upstream DL
					   //psb_->rxLabel(), // Upstream rxlbl
					   dsDL,
					   dsTxLbl)) {
			G2RSVPTE_ERR("Failed to get DL/label from %s",
				     addr_ntoa(teLink/*rsb_->interf->locId*/));
			return false;
		}

		rsb_->localDL(dsDL);     // Downstream DL
		rsb_->txLabel(dsTxLbl);  // Downstream tx label

		return true;
	}

	//
	// Set - rsb_->interface
	//     - rsb_->localDL
	//     - rsb_->txLabel   [ -> with params from LRM ]
	bool LSP::setDsTEL_DsDL_DsTxLbl_fromTna(g2mpls_addr_t tna)
	{
		//TODO??? find the downstream DL and label
		//        from rsb_->interface() and ...

		g2mpls_addr_t teLink;

		if (!getTELinkfromTna(tna, teLink)) {
			G2RSVPTE_ERR("Cannot get TELink from TNA %s", addr_ntoa(tna));
			return false;
		}

		g2mpls_addr_t dsDL;
		uint32_t      dsTxLbl;
		if (!getDL_Label_on_TELink(teLink, //rsb_->interface()->localId,
					   //psb_->pathOut()->label_request_,
					   //psb_->pathOut()->sender_tspec_,
					   //psb_->localDL(), // Upstream DL
					   //psb_->rxLabel(), // Upstream rxlbl
					   dsDL,
					   dsTxLbl)) {
			G2RSVPTE_ERR("Failed to get DL/label from %s",
				     addr_ntoa(teLink/*rsb_->interf->locId*/));
			return false;
		}

		Interface * intf;
		intf = getInterfaceFromTELink(teLink, false);
		if (!intf) {
			G2RSVPTE_ERR("TELink from TNA %s doesn't belong to "
				     "this instance interfaces", addr_ntoa(tna));
			return false;
		}

		rsb_->interface(intf);   // Downstream TEL
		rsb_->localDL(dsDL);     // Downstream DL
		rsb_->txLabel(dsTxLbl);  // Downstream tx label

		return true;
	}

	//
	// check teLink/dataLink/label and set - rsb_->localDL
	//                                     - rsb_->txLabel
	//
	bool LSP::checkSet_DsTEL_DsDL_DsTxLbl(g2mpls_addr_t teLink,
					      g2mpls_addr_t dataLink,
					      uint32_t      lbl)
	{
		if (!checkTELink(teLink)) {
			G2RSVPTE_ERR("Invalid TELink %s", addr_ntoa(teLink));
			return false;
		}

		//TODO check Data Link!!!

		label_state_t adminStatus;
		opstate_t     operStatus;
		if (!check_label(dataLink, lbl, adminStatus, operStatus)) {
			G2RSVPTE_ERR("Label (0x%x) does not exist under DL %s",
				     lbl, addr_ntoa(dataLink));
			return false;
		}

		rsb_->localDL(dataLink); // Downstream DL
		rsb_->txLabel(lbl);      // Downstream tx label

		return true;
	}

	//
	// check teLink/dataLink/label from TNA and set - rsb_->interface
	//                                              - rsb_->localDL
	//                                              - rsb_->txLabel
	//
	bool LSP::checkSet_DsTEL_DsDL_DsTxLbl_fromTna(g2mpls_addr_t tna,
						      g2mpls_addr_t dataLink,
						      uint32_t      lbl)
	{
		g2mpls_addr_t teLink;

		if (!getTELinkfromTna(tna, teLink)) {
			G2RSVPTE_ERR("Cannot get TELink from TNA %s", addr_ntoa(tna));
			return false;
		}

		//TODO check Data Link!!!

		label_state_t adminStatus;
		opstate_t     operStatus;
		if (!check_label(dataLink, lbl, adminStatus, operStatus)) {
			G2RSVPTE_ERR("Label (0x%x) does not exist under DL %s",
				     lbl, addr_ntoa(dataLink));
			return false;
		}

		Interface * intf;
		intf = getInterfaceFromTELink(teLink, false);
		if (!intf) {
			G2RSVPTE_ERR("TELink from TNA %s doesn't belong to "
				     "this instance interfaces", addr_ntoa(tna));
			return false;
		}

		rsb_->interface(intf);   // Downstream TEL
		rsb_->localDL(dataLink); // Downstream DL
		rsb_->txLabel(lbl);      // Downstream tx label

		return true;
	}

	//
	// Set - psb_->localDL
	//     - psb_->rxLabel   [ -> with params from LRM ]
	//
	bool LSP::setUsDL_UsRxLbl_fromTELink(g2mpls_addr_t teLink)
	{
		//TODO??? find the upstream DL and label
		//        from iTNA interface and ...

		g2mpls_addr_t usDL;
		uint32_t      usRxLbl;
		if (!getDL_Label_on_TELink(teLink, //ingress_tna_,
					   //psb_->pathOut()->label_request_,
					   //psb_->pathOut()->sender_tspec_,
					   //psb_->localDL(), // Upstream DL
					   //psb_->rxLabel(), // Upstream rxlbl
					   usDL,
					   usRxLbl)) {
			G2RSVPTE_ERR("Failed to get DL/label from %s",
				     addr_ntoa(teLink)); //ingress_tna_));
			return false;
		}

		psb_->localDL(usDL);     // Upstream DL
		psb_->rxLabel(usRxLbl);  // Upstream rx label

		return true;
	}

	//
	// check teLink/dataLink/label and set - psb_->localDL
	//                                     - psb_->rxLabel
	//
	bool LSP::checkSet_UsTEL_UsDL_UsRxLbl(g2mpls_addr_t teLink,
					      g2mpls_addr_t dataLink,
					      uint32_t      lbl)
	{
		if (!checkTELink(teLink)) {
			G2RSVPTE_ERR("Invalid TELink %s", addr_ntoa(teLink));
			return false;
		}

		//TODO check Data Link!!!

		label_state_t adminStatus;
		opstate_t     operStatus;
		if (!check_label(dataLink, lbl, adminStatus, operStatus)) {
			G2RSVPTE_ERR("Label (0x%x) does not exist under DL %s",
				     lbl, addr_ntoa(dataLink));
			return false;
		}

		psb_->localDL(dataLink); // Upstream DL
		psb_->rxLabel(lbl);      // Upstream rx label

		return true;
	}

	//
	// Set  - rsb_->nodeId
	//      - rsb_->interface
	//  [optional if complete ERO enable, otherwise set to NULL/0 ]
	//      - rsb_->localDL
	//      - rsb_->txLabel
	bool LSP::eroProcess(bool recursive)
	{
		// Get g2rsvpte instance
		RSVP::G2RSVPTE * g2rsvpte = getInstance();
		assert(g2rsvpte);

		if (!rsb_) {
			G2RSVPTE_ERR("Invalid resv state block!");
			return false;
		}

		// Set default values
		g2mpls_addr_t nullAdd;
		nullAdd.type        = UNNUMBERED;
		nullAdd.preflen     = 0;
		//nullAdd.value.unnum.node_id = 0;
		//nullAdd.value.unnum.addr    = 0;
		nullAdd.value.unnum = 0;

		rsb_->nodeId(0);
		//rsb_->interface(0); // useless!!! interface = 0 by default
		rsb_->localDL(nullAdd);
		rsb_->txLabel(0);

		// Checks
		if (!getInstance()) {
			G2RSVPTE_ERR("Invalid instance pointer");
			return false;
		}
		if (!psb_ || !psb_->pathOut()) {
			G2RSVPTE_ERR("Invalid Path State Block pointer");
			return false;
		}

		bool haveEro;
		bool haveEroSobjs;
		ExplicitRoute_Object * ero;

		haveEro      = psb_->pathOut()->ero_ ? true: false;
		haveEroSobjs = true;
		ero          = 0;


		//FIX!!! Change ERO type
		//       (ExplicitRoute_Class, not ExplicitRoute_Object)
		//ExplicitRoute_Object * ero;
		//ero = psb_->pathOut()->ero_; // already checked!!!(see above)

		if (haveEro) {
			if (psb_->pathOut()->ero_->cType() != CTYPE_ERO) {
				G2RSVPTE_ERR("temporarily, "
					     "only ctype_ero is allowed");
				return false;
			}

			ero = static_cast<ExplicitRoute_Object *>
				(psb_->pathOut()->ero_);

			if (ero->get().subObjs.empty()) {
				haveEroSobjs = false;
			}
		}

		if (!haveEro || !haveEroSobjs) {
			// I am the head node
			if (isHead()) {
				G2RSVPTE_ERR("I am the head node and "
					     "the ERO is empty!");
				return false;
			}
			// I am the tail node
			else if (isTail()) {
				G2RSVPTE_DBG("Finish!!! ... I am the tail node"
					     " and the ERO is empty!");
				return true;
			}

			G2RSVPTE_WRN("ERO is empty .... no HEAD nor TAIL"
				     " complete ero ??? XXX FIXME TODO!");

#ifdef COMPLETE_ERO
			G2RSVPTE_NOT("Trying to complete this ERO");
			if (!eroComplete()) {
				G2RSVPTE_ERR("ERROR completing ERO");
				return false;
			} else {
				G2RSVPTE_INF("ERO completed successfully");
			}
#endif // COMPLETE_ERO
		}

		bool flag = false;
		int count = 0;
		std::list<EroSubObject *>::iterator it;
		it = ero->get().subObjs.begin();
		while (it != ero->get().subObjs.end()) {
			EroSubObject * sObj;
			sObj = *it;

			if (!sObj) {
				G2RSVPTE_ERR("Invalid subObject pointer");
				return false;
			}

			if (sObj->looseFlag()) {
				G2RSVPTE_ERR("Loose ERO subObjs not yet supported");
				return false;
			}

			if(sObj->type() != LABEL_ERO_SUBOBJECT        &&
			   sObj->type() != COMP_IF_ID_IPv4_SUBOBJECT  &&
			   sObj->type() != COMP_IF_ID_IPv6_SUBOBJECT  &&
			   sObj->type() != COMP_IF_ID_UNNUM_SUBOBJECT) {
				if (flag) {
#ifdef COMPLETE_ERO
					//TODO??? checknfill_ero
#endif // COMPLETE_ERO
					G2RSVPTE_WRN("check and fill ero XXX FIXME TODO!");

					if (count > 1) {      // TODO FIXME
						return true;
					}
				} else {
					G2RSVPTE_DBG("... setting flag ...");
					flag = true;
				}
			}

			g2mpls_addr_t addr;

			switch (sObj->type()) {
				case IPV4_PREF_SUBOBJECT: {
					IPv4Pref_SubObj * ipv4;
					ipv4 = static_cast<IPv4Pref_SubObj *>(sObj);

					if (count == 1 && isHead()) {
						return true;
					}

					if (ipv4->get().preflen != 32) {
						G2RSVPTE_ERR("only 32-bit prefixes "
							     "are supported now");
						return false;
					}

					G2RSVPTE_DBG("the ERO TE link is %s",
						     numb_htoa(ntohl(ipv4->get().ipv4_addr.s_addr)));

					if (!make_addr(&addr, IPv4, &ipv4->get().ipv4_addr)) {
						G2RSVPTE_ERR("Strange error in creating g2mpls address");
						return false;
					}

					Interface * intf;
					intf = getInterfaceFromTELink(addr, (count || isHead()) ? true : false);

					if (!intf) {
						// Head node - AB NODE
						if (isHead() && recursive) {
							G2RSVPTE_INF("First ERO (IPv4-prefix) subobject is not about me... "
								     "but I'm head and should try to complete ERO");
							if (eroComplete()) {
								G2RSVPTE_INF("successfully completed ERO");
								return eroProcess(false);
							} else {
								G2RSVPTE_ERR("impossible to complete ERO");
								// set error params
								errTELink(addr);
								return false;
							}
						} else {
							// this object is not about me, then I can't find
							// the way to NHOP without ERO completion

							G2RSVPTE_ERR("First ERO (IPv4-prefix) subobject "
								     "is not about me ... wrong ERO");
							// set error params
							errTELink(addr);
							return false;
						}
					}

					if (!checkTELink(intf->localId())) {
						G2RSVPTE_ERR("Invalid TELink status %s", addr_ntoa(intf->localId()));
						// set error params
						errTELink(intf->localId());
						return false;
					}

					intf->dump();

					if (count || isHead()) {
						G2RSVPTE_DBG("Setting RSB (next node parms)");

						// set interface as NHOP TE link
						rsb_->interface(intf);

						// Get neighbor node id and set it as NHOP node id
						rsb_->nodeId(getNbrId());

						G2RSVPTE_INF("Selected and checked the next node");

						count++;
						it++;
						continue;
						//return true;
					}
					count++;
					break;
				}
				case IPV6_PREF_SUBOBJECT: {
					G2RSVPTE_ERR("IPv6 prefix subobjects not yet supported");
					return false;
				}
				case AS_SUBOBJECT: {
					G2RSVPTE_ERR("AS subobjects not yet supported");
					return false;
				}
				case UNNUM_LINK_SUBOBJECT: {
					Unnum_SubObj * unn;
					unn = static_cast<Unnum_SubObj *>(sObj);

					if (count == 1 && isHead()) {
						return true;
					}

					if ((count == 0) && isTail() &&
					    (unn->get().router_id != g2rsvpte->nodeId())) {
						G2RSVPTE_ERR("First ERO (Unnumbered) subobject "
							     "is not about me ... wrong ERO");
						return false;
					}

					G2RSVPTE_DBG("the (unn) ERO interface id is %d",
						     unn->get().interface_id);
					if (getInstance()->adjType() == ENNI && !unn->get().interface_id) {
						if (count == 1) {
							if (isTail()) {
								return true;
							}
						} else {
							if (isHead()) {
								bool res = getTELinkfromRemNodeId(unn->get().router_id,
												  addr);
								if (!res) {
									G2RSVPTE_ERR("Cannot get TeLink "
										     "from Remote NodeId %s",
										     numb_htoa(unn->get().router_id));
									return false;
								}
							} else {
								if (psb_->interface()) {
									addr = psb_->interface()->localId();
								} else {
									G2RSVPTE_ERR("Cannot get TeLink "
										     "from PSB interface");
									return false;
								}
							}
						}
					} else {
						unn_t uaddr;
						//uaddr.node_id = unn->get().router_id;
						//uaddr.addr    = unn->get().interface_id;
						uaddr = unn->get().interface_id;
						if (!make_addr(&addr, UNNUMBERED, &uaddr)) {
							G2RSVPTE_ERR("Strange error in creating g2mpls address");
							return false;
						}
					}
					Interface * intf;
					intf = getInterfaceFromTELink(addr, (count || isHead()) ? true : false);

					if (!intf) {
						// Head node - AB NODE
						if (isHead() && recursive) {
							G2RSVPTE_INF("First ERO (Unnumbered) subobject is not about me... "
								     "but I'm head and should try to complete ERO");
							if (eroComplete()) {
								G2RSVPTE_INF("successfully completed ERO");
								return eroProcess(false);
							} else {
								G2RSVPTE_ERR("impossible to complete ERO");
								// set error params
								errTELink(addr);
								return false;
							}
						} else {
							// this object is not about me, then I can't find
							// the way to NHOP without ERO completion

							G2RSVPTE_ERR("First ERO (Unnumbered) subobject "
								     "is not about me ... wrong ERO");
							// set error params
							errTELink(addr);
							return false;
						}
					}

					if (!checkTELink(intf->localId())) {
						G2RSVPTE_ERR("Invalid TELink status %s", addr_ntoa(intf->localId()));
						// set error params
						errTELink(intf->localId());
						return false;
					}

					intf->dump();

					if (count || isHead()) {
						G2RSVPTE_DBG("Setting RSB (next node parms)");

						// set interface as NHOP TE link
						rsb_->interface(intf);

						// Get neighbor node id and set it as NHOP node id
						rsb_->nodeId(getNbrId());

						G2RSVPTE_INF("Selected and checked the next node");

						count++;
						it++;
						continue;
						//return true;
					}
					count++;
					break;
				}
				case LABEL_ERO_SUBOBJECT: {
					if (!flag) {
						G2RSVPTE_ERR("Label subobj should follow a 'main' subobj");
						return false;
					}

					LabelERO_SubObj * lbl;
					lbl = static_cast<LabelERO_SubObj *>(sObj);
					if (lbl->get().upstream) {
						G2RSVPTE_ERR("Only downstream Label ERO subobject supported");
						return false;
					}

					RSVP::Label_Class * label = lbl->get().label;
					if (!label ||
					    label->cType() != CTYPE_GENERALISED_LABEL) /* Useless??? */ {
						G2RSVPTE_ERR("Label ERO subobj with invalid pointer!");
						return false;
					}

					RSVP::GLabel_Object * genLabel;
					genLabel = static_cast<RSVP::GLabel_Object *>(label);

					if (isHead() || count == 2) {
						//Process next node in the ERO
						if (!setDsDL_checkDsTxLbl_fromTELink(rsb_->interface()->
										     localId(),
										     *(genLabel->get().
										       label))) {
							G2RSVPTE_ERR("could not find "
								     "DL/check label under TE link %s",
								     addr_ntoa(rsb_->interface()->localId()));
							// set error params
							errTELink(rsb_->interface()->localId());
							errLabel(*(genLabel->get().label));
							return false;
						}
						return true;
					} else {
						//Process first node in the ERO
						if (psb_->rxLabel() != *(genLabel->get().label)) {
							G2RSVPTE_ERR("In my ERO there are two Label "
								     "subobjs with different values "
								     "(0x%x, 0x%x) ... not allowed!",
								     psb_->rxLabel(),
								     *(genLabel->get().label));
							// set error params
							errTELink(rsb_->interface()->localId());
							errLabel(*(genLabel->get().label));
							return false;
						}
					}

#if 0
					if (rsb_->txLabel() &&
					    rsb_->txLabel() != *(genLabel->get().label)) {
						G2RSVPTE_ERR("In my ERO there are two Label subobjs with "
							     "different values (0x%x, 0x%x) ... not allowed!",
							     rsb_->txLabel(),
							     *(genLabel->get().label));
						//return false;
					}

					rsb_->txLabel(*(genLabel->get().label));
					G2RSVPTE_DBG("the ERO label is 0x%x", rsb_->txLabel());
#else
					//G2RSVPTE_DBG("Label subObj MUST BE CHECK AGAIN!  TODO!!!!!!!!!");
#endif

					break;

					//G2RSVPTE_WRN("ERO Labels usage is not yet supported");
					//return false;
				}
				case COMP_IF_ID_IPv4_SUBOBJECT:
				case COMP_IF_ID_IPv6_SUBOBJECT:
				case COMP_IF_ID_UNNUM_SUBOBJECT: {
					G2RSVPTE_WRN("ERO Component i/f id usage is not yet supported");
					return false;
				}
				default:
					G2RSVPTE_ERR("unknown subobj type (%d)",
						     sObj->type());
					return false;
			} // switch

			// Delete the subObj
			delete sObj;
			// Remove subObj from the list and update the iterator
			it = ero->get().subObjs.erase(it);
			G2RSVPTE_WRN("Deleted ero------------------------");

			//count++;
		} // while

		G2RSVPTE_DBG("ERO has ended");

#ifdef COMPLETE_ERO
		//TODO????
#endif // COMPLETE_ERO


		// If the ERO is empty, remove the ERO object
		if (ero->get().subObjs.empty()) {

			// TODO check and delete all memory into ero obj
			delete ero;                 // == psb_->pathOut()->ero_
			ero                   = 0;  // useless ...
			psb_->pathOut()->ero_ = 0;

		}
		return true;
	}

	//
	// Return true if it detects a loop or an error
	//
	bool LSP::loopDetect(void)
	{
		// Get g2rsvpte instance
		RSVP::G2RSVPTE * g2rsvpte = getInstance();
		assert(g2rsvpte);

		// Checks
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->rro_ ||
		    !g2rsvpte->nodeId()) {
			G2RSVPTE_ERR("Invalid RRO pointer");
			return true;
		}

		//FIX!!! Change RRO type
		//       (RecordedRoute_Class, not RecordedRoute_Object)
		if (psb_->pathIn()->rro_->cType() != CTYPE_RECORDED_ROUTE) {
			G2RSVPTE_ERR("temporarily, only ctype_rro is allowed");
			return false;
		}
		RecordedRoute_Object * rro;
		rro = static_cast<RecordedRoute_Object*>(psb_->pathIn()->rro_);

		std::list<RroSubObject *>::iterator it;
		for (it = rro->get().subObjs.begin();
		     it != rro->get().subObjs.end();
		     it++) {
			RroSubObject * sObj;
			sObj = *it;

			if (!sObj) {
				G2RSVPTE_ERR("Invalid subObject pointer");
				return true;
			}

			switch (sObj->type()) {
				case IPV4_ADDR_SUBOBJ:
				case IPV6_ADDR_SUBOBJ:
					// Skip for now
					break;

				case UNNUM_LINK_SUBOBJ: {
					Unnumbered_SubObj * unn;
					unn = static_cast<Unnumbered_SubObj *>
						(sObj);

					if (unn->get().router_id ==
					    g2rsvpte->nodeId()) {
						G2RSVPTE_ERR("loop detected!");
						return true;
					}
					break;
				}
				case LABEL_SUBOBJ:
					// Skip for now
					break;

				default:
					G2RSVPTE_ERR("unknown subobj type (%d)",
						     sObj->type());
					return true;
			}
		}

		// No loops
		return false;
	}

	bool LSP::signalUpLSP(bool enqueueEvent)
	{
		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		// Check if LSP is enable
//		if (!enabled_) {
//			G2RSVPTE_ERR("Impossible to signalUp this LSP ... "
//				     "it is not enabled!!!");
//			return false;
//		}

		// check that the DS ACK/NACK is quiet
		// TODO:
		//if (ds_acknack_->state != init) return false;

		// Checks
		if (!psb_ || !psb_->pathOut() || !psb_->pathOut()->ero_ ||
		    !getInstance()) {
			G2RSVPTE_ERR("Invalid pointers");
			return false;
		}

		//FIX!!! Change ERO type
		//       (ExplicitRoute_Class, not ExplicitRoute_Object)
		if (psb_->pathOut()->ero_->cType() != CTYPE_ERO) {
			G2RSVPTE_ERR("temporarily, only ctype_ero is allowed");
			return false;
		}
		ExplicitRoute_Object * ero;
		ero = static_cast<ExplicitRoute_Object*>(psb_->pathOut()->ero_);

		// Try to fill the ERO (if it is empty)
		if (ero->get().subObjs.empty()) {
			G2RSVPTE_DBG("ERO is empty. Impossible to start signalling.");
			return false;
			//G2RSVPTE_DBG("ERO is empty ... trying to fill in by "
			//	     "path computation");
			//
			//if (!fill_ero(nid_,                      // src node
			//	      parent()->nodeId(),        // dst node
			//	      ero->get().subObjs)) {
			//	G2RSVPTE_ERR("Fill ERO failed");
			//	return false;
			//}
		}

		// Check the exclude/include masks consistency with TE link
		// colors  [ OPTIONAL now ??? TODO ]
		//if (!checkTELinkColors()) {
		//      TODO!!!
		//	return false;
		//}

		//
		// DL and label selection in the Ingress Node/Tna
		//
		g2mpls_addr_t ingress;
		if (isNetResSpecValid(in_net_res_)) {

			if (BITMASK_BITTEST(in_net_res_.mask_, data_link) &&
			    BITMASK_BITTEST(in_net_res_.mask_, label)) {

				//!!! for now only 32bit label are supported
				uint32_t lbl = in_net_res_.label.value.
					label32.id;

				// Check usTEL, usDL and usRxLabel on Ingress Node
				if (!checkSet_UsTEL_UsDL_UsRxLbl(in_net_res_.tna,
								 in_net_res_.data_link,
								 lbl)) {
					G2RSVPTE_ERR("Invalid ingress parameters: ...");
					return false;
				}
			} else {
				// Choose upstream DL and label (rx) for I-TNA
				if (!setUsDL_UsRxLbl_fromTELink(in_net_res_.tna)) {
					G2RSVPTE_ERR("could not find any suitable Data"
						     "Link/label under Ingress TNA %s",
						     addr_ntoa(in_net_res_.tna));
					return false;
				}
			}

			G2RSVPTE_INF("Ingress-Node %s, selected:",
				     addr_ntoa(in_net_res_.tna));
			addr_copy(&ingress, &in_net_res_.tna);

		//} else if (!is_addr_null(ingress_tna_)) {
		//	// Find appropriate usDL and usRxLabel on iTna
		//	if (!setUsDL_UsRxLbl_fromTELink(ingress_tna_)) {
		//		G2RSVPTE_ERR("could not find any suitable "
		//			     "DL/label in Ingress TNA %s",
		//			     addr_ntoa(ingress_tna_));
		//		return false;
		//	}
		//
		//	G2RSVPTE_INF("I-TNA %s, selected:",
		//		     addr_ntoa(ingress_tna_));
		//	addr_copy(&ingress, &ingress_tna_);

		} else {
			G2RSVPTE_ERR("could not find any suitable "
				     "Ingress Node/TNA");
			return false;
		}

		// We assume that the US TX label is equal to the US RX
		psb_->txLabel(psb_->rxLabel());

		G2RSVPTE_INF("\t Data Link  %s",   addr_ntoa(psb_->localDL()));
		G2RSVPTE_INF("\t tx label   0x%x", psb_->txLabel());
		G2RSVPTE_INF("\t rx label   0x%x", psb_->rxLabel());

		// ERO Process
		if (!eroProcess()) {
			G2RSVPTE_ERR("ERO process failed!");
			return false;
		}

		// prepare the PHOP in the outgoing Path
		if (!preparePHOPobj()) {
			return false;
		}

		// prepare the LABEL_SET obj in the outgoing Path
		//  (using downstream tx label)
		if (!prepareLabelSetObj()) {
			return false;
		}

		// prepare the UPSTREAM_LABEL obj in the outgoing Path
		//  (using downstream rx label)
		if (!prepareUpstreamLabelObj()) {
			return false;
		}

		// Add this node to the RRO
		if (psb_->pathOut()->rro_) {
			if (psb_->pathIn()->rro_->cType() !=
			    CTYPE_RECORDED_ROUTE) {
				G2RSVPTE_ERR("temporarily, only ctype_rro "
					     "is allowed");
				return false;
			}
			RecordedRoute_Object * rro;
			rro = static_cast<RecordedRoute_Object *>
				(psb_->pathOut()->rro_);

			if (!insertIntoRRO(rro, ingress)) {
				G2RSVPTE_ERR("problems preparing outgoing "
					     "Path RRO obj");
				return false;
			}
		}

		// send a 'sendPath' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_SendPath;

		fsmInst_->post(ev, context);
		//TODO add into FSM GEN if (!post) return false;

		if (!enqueueEvent) {
			G2RSVPTE_NOT("LSP signaledUP");
		}

		return true;
	}

	bool LSP::signalDownLSP(void)
	{
		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		// if (release) SET lsp_flags RELEASED;
		// if (admin)   SET lsp_flags SIGADMINDOWN

		if (isHead()) {      // HEAD
			// check that the DS ACK/NACK is quiet
			// TODO:
			//if (ds_acknack_->state != init) return false;

			// Checks
			if (!psb_ || !psb_->pathOut()) {
				G2RSVPTE_ERR("Invalid PSB/PathOut pointers");
				return false;
			}

			if (psb_->pathDown()) {
				G2RSVPTE_WRN("PathDown is already present... de"
					     "lete it before allocating again");
				delete psb_->pathDown();
				psb_->pathDown(0);
			}

			PathMessage * pathDown;
			pathDown = new PathMessage(*psb_->pathOut(), true);
			if (!pathDown) {
				G2RSVPTE_ERR("Impossible to allocate memory for"
					     " PathDown Msg");
				return false;
			}
			psb_->pathDown(pathDown);

			// send a 'sendPathDown' basic event to the LSP FSM
			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to alocate context for"
					     " LSP fsm");
				return false;
			}
			context->lsp  = this;
			context->intf = 0;  //TODO new context? or only 1 cntx?
			context->msg  = 0;

			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
				G2RSVPTE_LSP_FSM_SendPathDown;

			fsmInst_->post(ev, context);
			//TODO add into FSM GEN if (!post) return false;

		} else {             // TAIL
			// check that the US ACK/NACK is quiet
			// TODO:
			//if (us_acknack_->state != init) return false;

			// Checks
			if (!rsb_ || !rsb_->resvOut()) {
				G2RSVPTE_ERR("Invalid RSB/ResvOut pointers");
				return false;
			}

			if (rsb_->resvDown()) {
				G2RSVPTE_WRN("ResvDown is already present... de"
					     "lete it before allocating again");
				delete rsb_->resvDown();
				rsb_->resvDown(0);
			}

			ResvMessage * resvDown;
			resvDown = new ResvMessage(*rsb_->resvOut(), true);
			if (!resvDown) {
				G2RSVPTE_ERR("Impossible to allocate memory for"
					     " ResvDown Msg");
				return false;
			}
			rsb_->resvDown(resvDown);

			// send a 'sendResvDown' basic event to the LSP FSM
			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to alocate context for"
					     " LSP fsm");
				return false;
			}
			context->lsp  = this;
			context->intf = 0;  //TODO new context? or only 1 cntx?
			context->msg  = 0;

			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
				G2RSVPTE_LSP_FSM_SendResvDown;

			fsmInst_->post(ev, context);
			//TODO add into FSM GEN if (!post) return false;
		}

		G2RSVPTE_NOT("LSP signaledDown");

		return true;
	}

	bool LSP::enableLSP(void)
	{
		if (enabled_) {
			G2RSVPTE_WRN("LSP already enabled!!!");
			return false;
		}

		//TODO check if the FSM is in 'down' state
		enabled_ = true;
		G2RSVPTE_NOT("LSP enabled");

		return true;
	}

	bool LSP::disableLSP(void)
	{
		if (!enabled_) {
			G2RSVPTE_WRN("LSP already disabled!!!");
			return false;
		}

		//TODO check if the FSM is in 'down' state
		//     otherwise signalDownLSP before
		//     return false;

		enabled_ = false;
		G2RSVPTE_NOT("LSP disabled");

		return true;
	}
#ifdef CRANKBACK_ON
	void LSP::attachExcludeEro(std::list<lsp_ero_sobj_t> exclEro)
	{
		std::list<lsp_ero_sobj_t>::iterator it;
		lsp_ero_sobj_t                      sObj;
		for (it  = exclEro.begin();
		     it != exclEro.end();
		     it++) {
			memset(&sObj, 0, sizeof(sObj));
			sObj = *it;
			excludeEro_.push_back(sObj);
		}
	}
#endif // CRANKBACK_ON
	bool LSP::attachEroSubObj(EroSubObject * eroSubObj, bool insertTail)
	{
		if (!eroSubObj) {
			G2RSVPTE_ERR("Null ERO subObj pointer!!!");
			return false;
		}

		if (!psb_ || !psb_->pathOut() || !psb_->pathOut()->ero_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		ExplicitRoute_Object * ero;
		if (psb_->pathOut()->ero_->cType() != CTYPE_ERO) {
			G2RSVPTE_ERR("temporarily, only CTYPE_ERO is allowed");
			return false;
		}
		ero = static_cast<ExplicitRoute_Object *>
			(psb_->pathOut()->ero_);

		if (!ero->insertEroSubObj(eroSubObj, insertTail)) {
			G2RSVPTE_ERR("Impossible to add EROsubObj to Path msg");
			return false;
		}

		std::ostringstream str;
		str << *eroSubObj;
		G2RSVPTE_DBG("Added ero subObj: %s", str.str().c_str());

		return true;
	}

	bool LSP::cleanEro(void)
	{
		if (!psb() || !psb()->pathOut()) {
			G2RSVPTE_ERR("Impossible to clean old Ero: PSB/PathOut pointer null");
			return false;
		}
		if (!psb()->pathOut()->ero_) {
			G2RSVPTE_ERR("Impossible to clean old Ero: Object not present");
			return false;
		}
		// delete old ero
		delete psb()->pathOut()->ero_;
		psb()->pathOut()->ero_ = 0;
		// create new ero
		psb()->pathOut()->ero_ = new ExplicitRoute_Object;
		if (!psb()->pathOut()->ero_) {
			G2RSVPTE_ERR("Impossible to create new Ero Object");
			return false;
		}

		return true;
	}

	bool LSP::cleanRsvpHop(void)
	{
		if (!psb() || !psb()->pathOut()) {
			G2RSVPTE_ERR("Impossible to clean old RsvpHop obj: PSB/PathOut pointer null");
			return false;
		}
		if (!psb()->pathOut()->rsvp_hop_) {
			G2RSVPTE_ERR("Impossible to clean old RsvpHop obj: Object not present");
			return false;
		}
		// delete old rsvp_hop_
		delete psb()->pathOut()->rsvp_hop_;
		psb()->pathOut()->rsvp_hop_ = 0;
		// create new rsvp_hop_
		psb()->pathOut()->rsvp_hop_ = new IFIDIPv4RsvpHop_Object;
		if (!psb()->pathOut()->rsvp_hop_) {
			G2RSVPTE_ERR("Impossible to create new RsvpHop Object");
			return false;
		}

		return true;
	}

	bool LSP::cleanLabelObjs(void)
	{
		if (!psb() || !psb()->pathOut()) {
			G2RSVPTE_ERR("Impossible to clean old Label objects: PSB/PathOut pointer null");
			return false;
		}
		if (!psb()->pathOut()->label_set_) {
			G2RSVPTE_ERR("Impossible to clean old LabelSet object: Object not present");
			return false;
		}
		// delete old label_set_
		delete psb()->pathOut()->label_set_;
		psb()->pathOut()->label_set_ = 0;
		// create new label_set_
		psb()->pathOut()->label_set_ = new RSVP::LabelSet_Object();
		if (!psb()->pathOut()->label_set_) {
			G2RSVPTE_ERR("Impossible to create new LabelSet Object");
			return false;
		}

		if (!psb()->pathOut()->upstream_label_) {
			G2RSVPTE_ERR("Impossible to clean old UpstreamLabel object: Object not present");
			return false;
		}
		// delete old upstream_label_
		delete psb()->pathOut()->upstream_label_;
		psb()->pathOut()->upstream_label_ = 0;
		// create new upstream_label_
		psb()->pathOut()->upstream_label_ = new RSVP::UpstreamLabel_Object();
		if (!psb()->pathOut()->upstream_label_) {
			G2RSVPTE_ERR("Impossible to create new UpstreamLabel Object");
			return false;
		}

		return true;
	}

	bool LSP::cleanSBs(void)
	{
		if (!psb() || !rsb()) {
			G2RSVPTE_ERR("Impossible to clean State Blocks: PSB/RSB pointer null");
			return false;
		}
		g2mpls_addr_t nullAddr;
		memset(&nullAddr, 0, sizeof(nullAddr));

		//clean te link
		if (psb()->interface())  {
			psb()->interface(0, false);
		}
		if (rsb()->interface()) {
			rsb()->interface(0, false);
		}
		//clean data links
		psb()->localDL(nullAddr);
		psb()->remoteDL(nullAddr);
		rsb()->localDL(nullAddr);
		rsb()->remoteDL(nullAddr);
		//clean labels
		psb()->txLabel(0);
		psb()->rxLabel(0);
		rsb()->txLabel(0);
		rsb()->rxLabel(0);

		return true;
	}

	bool LSP::setTna(const net_res_spec & tna, bool ingress)
	{
		if (!getInstance()) {
			G2RSVPTE_DBG("Null instance pointer");
			return false;
		}

		switch (getInstance()->adjType()) {
			case UNI:
				if (callType() == CALL_TYPE_PC  ||
				    callType() == CALL_TYPE_SPC) {
					G2RSVPTE_ERR("Wrong callType 0x%x for UNI "
						     "instance", callType());
				}
				break;
			case ENNI:
				if (callType() == CALL_TYPE_AUTO ||
				    callType() == CALL_TYPE_PC) {
					G2RSVPTE_ERR("Wrong callType 0x%x for ENNI "
						     "instance", callType());
				}
				break;
			case INNI:
				G2RSVPTE_ERR("Wrong instance adjType (INNI) for "
					     "setTna method");
				return false;
		}

		if (!psb() || !psb()->pathOut()) {
			G2RSVPTE_ERR("Null pointer !!!");
		}

		switch (ingress) {
			case true:
				if (!isNetResSpecNull(in_net_res_)) {
					G2RSVPTE_ERR("i-TNA already set");
					return false;
				}
				iRes(tna);
				break;
			case false:
				if (!isNetResSpecNull(eg_net_res_)) {
					G2RSVPTE_ERR("e-TNA already set");
					return false;
				}
				eRes(tna);
				break;
			default:
				return false;
		}

		if (!psb()->pathOut()->setGenUni(tna, ingress, callType())) {
			G2RSVPTE_ERR("Cannot set GeneralisedUni_Object in Path message");
			return false;
		}
		G2RSVPTE_DBG("%s-TNA set", ingress ? "i" : "e");

		return true;
	}

	bool LSP::updateTna(const net_res_spec & tna, bool ingress)
	{
		if (!getInstance()) {
			G2RSVPTE_DBG("Null instance pointer");
			return false;
		}

		switch (getInstance()->adjType()) {
			case UNI:
				if (callType() != CALL_TYPE_AUTO &&
				    callType() != CALL_TYPE_SC) {
					G2RSVPTE_ERR("Wrong callType 0x%x for UNI"
						     "instance", callType());
				}
				break;
			case ENNI:
				if (callType() != CALL_TYPE_SPC &&
				    callType() != CALL_TYPE_SC) {
					G2RSVPTE_ERR("Wrong callType 0x%x for ENNI"
						     "instance", callType());
				}
				break;
			case INNI:
				G2RSVPTE_ERR("Wrong instance adjType (INNI) for "
					     "setTna method");
				return false;
		}

		if (!psb() || !psb()->pathOut()) {
			G2RSVPTE_ERR("Null pointer !!!");
		}

		if (!psb()->pathIn()->updateGenUni(tna, ingress, callType())) {
			G2RSVPTE_ERR("Cannot update GuniSubObj in Path message");
			return false;
		}

		if (ingress) {
			iRes(tna);
		} else {
			eRes(tna);
		}

		return true;
	}

	bool LSP::getTna(bool ingress, net_res_spec & tna)
	{
		switch (ingress) {
			case true:
				if (isNetResSpecNull(in_net_res_)) {
					G2RSVPTE_ERR("i-TNA is not set");
					return false;
				} else {
					tna = iRes();
				}
				break;
			case false:
				if (isNetResSpecNull(eg_net_res_)) {
					G2RSVPTE_ERR("e-TNA is not set");
					return false;
				} else {
					tna = eRes();
				}
				break;
			default:
				return false;
		}

		return true;
	}


	bool LSP::getTnResource(net_res_spec & tnRes)
	{
		memset(&tnRes, 0, sizeof(tnRes));

		std::string side;

		if (isTail()) {
			if (!psb_ || !psb_->interface()) {
				G2RSVPTE_ERR("Get of TN Resources failed on "
					     "TAIL for missing "
					     "psb or psb interface");
				return false;
			}

			side = "TAIL (PSB)";

			// Upstream TEL
			tnRes.tna = psb_->interface()->localId();

			// Downstream datalink
			tnRes.data_link = psb_->localDL();

			// Downstream tx label
			tnRes.label.type = LABEL_32BIT;
			tnRes.label.value.label32.id = psb_->txLabel();

		} else if (isHead()) {
			if (!rsb_ || !rsb_->interface()) {
				G2RSVPTE_DBG("Get of TN Resources failed on "
					     "HEAD for missing "
					     "rsb or rsb interface");
				return false;
			}
			side = "HEAD (RSB)";

			// Downstream TEL
			tnRes.tna = rsb_->interface()->localId();

			// Downstream datalink
			tnRes.data_link = rsb_->localDL();

			// Downstream tx label
			tnRes.label.type = LABEL_32BIT;
			tnRes.label.value.label32.id = rsb_->txLabel();
		} else {
			G2RSVPTE_DBG("Get of TN Resources issued on a node "
				     "neither HEAD or TAIL");
			return false;
		}

		G2RSVPTE_DBG("TN Resources got from %s",
			     side.c_str());
		G2RSVPTE_DBG("   - TE Link    %s",
			     addr_ntoa(tnRes.tna));
		G2RSVPTE_DBG("   - Data Link  %s",
			     addr_ntoa(tnRes.data_link));
		G2RSVPTE_DBG("   - Label      %s",
			     label2string(tnRes.label).c_str());

		return true;
	}

	bool LSP::setGnsTna(const grid_res_spec_t & gnsTna,
			    bool                    ingress)
	{
		if (isHead()) {
			if (!psb() || !psb()->pathOut()) {
				G2RSVPTE_ERR("Null pointer !!!");
			}

			switch (ingress) {
				case true:
					if (psb()->pathOut()->gns_uni_in_) {
						G2RSVPTE_ERR("iGNS-TNA already set in the outgoing Path");
						return false;
					} else if (!psb()->pathOut()->gns_uni_eg_) {
						G2RSVPTE_ERR("Cannot set iGNS-TNA without eGNS-TNA !");
						return false;
					}
					break;
				case false:
					if (psb()->pathOut()->gns_uni_eg_) {
						G2RSVPTE_ERR("eGNS-TNA already set in the outgoing Path");
						return false;
					}
					break;
				default:
					return false;
			}
			if (!psb()->pathOut()->setGnsTna(gnsTna, ingress)) {
				G2RSVPTE_ERR("Cannot set GnsUni_Object in Path message");
				return false;
			}
		} else if (isTail()) {
			if (!rsb() || !psb()->pathIn()) {
				G2RSVPTE_ERR("Null pointer !!!");
			}

			if (!rsb()->resvOut()) {
				ResvMessage * resv;
				resv = new ResvMessage(*psb()->pathIn());
				if (!resv) {
					G2RSVPTE_ERR("Impossible to create Resv to be sent "
						     "for LSP id %d", id_);
					return false;
				}
				// Initialise ResvOut for this new LSP
				rsb()->resvOut(resv);
			}

			switch (ingress) {
				case true:
					if (rsb()->resvOut()->gns_uni_in_) {
						G2RSVPTE_ERR("iGNS-TNA already set in the outgoing Resv");
						return false;
					} else if (!rsb()->resvOut()->gns_uni_eg_) {
						G2RSVPTE_ERR("Cannot set iGNS-TNA without eGNS-TNA !");
						return false;
					}
					break;
				case false:
					if (rsb()->resvOut()->gns_uni_eg_) {
						G2RSVPTE_ERR("eGNS-TNA already set in the outgoing Resv");
						return false;
					}
					break;
				default:
					return false;
			}
			if (!rsb()->resvOut()->setGnsTna(gnsTna, ingress)) {
				G2RSVPTE_ERR("Cannot set GnsUni_Object in Resv message");
				return false;
			}
		}

		return true;
	}

	bool LSP::getGnsTna(bool              ingress,
			    grid_res_spec_t & gnsTna)
	{
		if (isTail()) {
			if (!psb() || !psb()->pathIn()) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!psb()->pathIn()->gns_uni_in_) {
						G2RSVPTE_ERR("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!psb()->pathIn()->gns_uni_eg_) {
						G2RSVPTE_ERR("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
			}
			if (!psb()->pathIn()->getGnsTna(gnsTna, ingress)) {
				G2RSVPTE_ERR("Impossible to retrieve GNS-TNA "
					     "parameters from incoming Path message");
				return false;
			}
		} else if (isHead()) {
			if (!rsb() || !rsb()->resvIn()) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!rsb()->resvIn()->gns_uni_in_) {
						G2RSVPTE_ERR("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!rsb()->resvIn()->gns_uni_eg_) {
						G2RSVPTE_ERR("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
			}
			if (!rsb()->resvIn()->getGnsTna(gnsTna, ingress)) {
				G2RSVPTE_ERR("Impossible to retrieve GNS-TNA"
					     "parameters from incoming Resv message");
				return false;
			}
		} else {
			if (!psb() || !psb()->pathOut()) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!psb()->pathOut()->gns_uni_in_) {
						G2RSVPTE_ERR("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!psb()->pathOut()->gns_uni_eg_) {
						G2RSVPTE_ERR("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
						}
			if (!psb()->pathOut()->getGnsTna(gnsTna, ingress)) {
				G2RSVPTE_ERR("Impossible to retrieve GNS-TNA"
					     "parameters from outgoing Path message");
				return false;
			}
		}

		return true;
	}

	bool LSP::getPolicy(std::string & authz)
	{
		Policy_Class * policy;

		if (isHead()) {
			if (!psb() || !psb()->pathOut() || !psb()->pathOut()->policy_data_) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}
			policy = psb()->pathOut()->policy_data_;
		} else {
			if (!psb() || !psb()->pathIn() || !psb()->pathIn()->policy_data_) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}
			policy = psb()->pathIn()->policy_data_;
		}

		if (policy->cType() != CTYPE_POLICY) {
			G2RSVPTE_ERR("cType %d not supported for PolicyData Obj",
				     policy->cType());
			return false;
		}

		Policy_Object * policyObj;
		policyObj = static_cast<Policy_Object *> (policy);

		if (policyObj->get().policy_elems.size() == 0) {
			G2RSVPTE_ERR("There are no Policy Element in Policy Object!");
			return false;
		}
		if (policyObj->get().policy_elems.size() > 1) {
			G2RSVPTE_ERR("One and only one Policy Element can be present in Policy Object!");
			return false;
		}

		PolicyInfo * pInfo;
		pInfo = policyObj->get().policy_elems.front();
		if (pInfo->type != PTYPE_PHOSPHORUS_AUTHZ_TOKEN) {
			G2RSVPTE_ERR("P-type %d is not supported in PolicyInfo!", pInfo->type);
				return false;
		}

		authz = pInfo->opaque_data;

		return true;
	}

	bool LSP::getErrInfo(error_info_t &            error,
			     std::list<lsp_ero_sobj_t> excludeEro,
			     bool                      lookAtPath)
	{
		memset(&error, 0, sizeof(error));
		excludeEro.clear();

		if (lookAtPath) {
			if (!psb() || !psb()->pathErr()) {
				G2RSVPTE_ERR("Null pointer !!!");
				return false;
			}
			if (!psb()->pathErr()->getErrorDetails(error, excludeEro)) {
				G2RSVPTE_ERR("Impossible to retireve error info from Path Err");
				return false;
			}
		} else {
			if (!rsb() || !rsb()->resvErr()) {
                                G2RSVPTE_ERR("Null pointer !!!");
                                return false;
                        }
			if (!rsb()->resvErr()->getErrorDetails(error)) {
				G2RSVPTE_ERR("Impossible to retireve error info from Resv Err");
				return false;
			}
		}

		return true;
	}

	bool LSP::sendResv(bool enqueueEvent)
	{
		if (!rsb_ || !psb_ || !psb_->pathIn()) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		if (!isTail()) {
			G2RSVPTE_ERR("This NE is not the tail!!!");
			return false;
		}

		// check that the US ACK/NACK is quiet
		// TODO:
		//if (us_acknack_->state != init) return false;

		if (!rsb_->resvOut()) {
			ResvMessage * resv;
			resv = new ResvMessage(*psb_->pathIn());
			if (!resv) {
				G2RSVPTE_ERR("Impossible to create Resv to be sent "
					     "for LSP id %d", id_);
				return false;
			}
			// Initialise ResvOut for this new LSP
			rsb_->resvOut(resv);
		}
		// static flags for resv msg??

		assert(rsb_->resvOut()->time_values_);
		// set time_values
		if (rsb_->resvOut()->time_values_->cType() != CTYPE_TIME_VALUES) {
			G2RSVPTE_ERR("cType %d not supported for TimeValue Obj",
				     rsb_->resvOut()->time_values_->cType());
			return false;
		}
		static_cast<RSVP::TimeValues_Object *>
			(rsb_->resvOut()->time_values_)->get().refresh_period =
			refreshInterval_;

		assert(rsb_->resvOut()->style_);
		// set style
		if (rsb_->resvOut()->style_->cType() != CTYPE_STYLE) {
			G2RSVPTE_ERR("cType %d not supported for Style Obj",
				     rsb_->resvOut()->style_->cType());
			return false;
		}
		static_cast<RSVP::Style_Object *>(rsb_->resvOut()->style_)->
			setFFstyle();
		static_cast<RSVP::Style_Object *>(rsb_->resvOut()->style_)->
			flags(0);

		// prepare NHOP
		if (!prepareNHOPobj()) {
			G2RSVPTE_ERR("Impossible to prepare Resv NHOP obj");
			return false;
		}

		assert(rsb_->resvOut()->resv_confirm_);
		// set resv_confirm
		if (rsb_->resvOut()->resv_confirm_->cType() !=
		    CTYPE_RESVCONF_IPV4) {
			G2RSVPTE_ERR("cType %d not supported for ResvConf Obj",
				     rsb_->resvOut()->resv_confirm_->cType());
			return false;
		}
		assert(getInstance());
		static_cast<RSVP::ResvConfirmIPv4_Object *>
			(rsb_->resvOut()->resv_confirm_)->get().ipv4_addr.
			s_addr = htonl(getInstance()->nodeId());

		// prepare label obj
		if (!prepareLabelObj()) {
			G2RSVPTE_ERR("Impossible to prepare Resv LABEL obj");
			return false;
		}

		if (getInstance()->adjType() == INNI) {
			// add this node to the RRO
			if (rsb_->resvOut()->flow_desc_              &&
			    rsb_->resvOut()->flow_desc_->getRROobj()) {

				g2mpls_addr_t egress;
				if (isNetResSpecValid(eg_net_res_)) {
					G2RSVPTE_INF("Egress-Node %s, selected:",
						     addr_ntoa(eg_net_res_.tna));
					addr_copy(&egress, &eg_net_res_.tna);
					//} else if (!is_addr_null(egress_tna_)) {
					//	G2RSVPTE_INF("E-TNA %s, selected:",
					//		     addr_ntoa(egress_tna_));
					//	addr_copy(&egress, &egress_tna_);
				} else {
					G2RSVPTE_ERR("could not find any suitable "
						     "Egress Node/TNA");
					return false;
				}

				if (!insertIntoRRO(rsb_->resvOut()->flow_desc_->
						   getRROobj(),
						   //rsb_->int->localId() is NULLonTail
						   egress)) { // ETNA
					G2RSVPTE_ERR("Impossible to insert Egress "
						     "Node/TNA into RRO obj");
					return false;
				}
			}
		}
		assert(rsb_->resvOut()->notify_request_);
		// set notify_request --> nodeId or interface from LRM ???
		if (rsb_->resvOut()->notify_request_->cType() !=
		    CTYPE_IPV4_NOTIFY_REQUEST) {
			G2RSVPTE_ERR("cType %d not supported for ResvConf Obj",
				     rsb_->resvOut()->notify_request_->cType());
			return false;
		}
		assert(getInstance());
		static_cast<RSVP::IPV4NotifyRequest_Object *>
			(rsb_->resvOut()->notify_request_)->get().ipv4_addr.
			s_addr = htonl(getInstance()->nodeId());

		G2RSVPTE_NOT("Posting a SendResv to LSP FSM");

		// send a 'sendResv' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_SendResv;

		fsmInst_->post(ev, context, enqueueEvent);
		//TODO add into FSM GEN if (!post) return false;

		G2RSVPTE_NOT("Posted SendResv event");

		return true;
	}

	bool LSP::sendPath(bool enqueueEvent)
	{
		////This is called only by UNI & ENNI RSVP instances
		assert(parent());
		assert(parent()->parent());

		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		// Check if LSP is enable
//		if (!enabled_) {
//			G2RSVPTE_ERR("Impossible to send Path for this LSP ... "
//				     "it is not enabled!!!");
//			return false;
//		}

		// check that the DS ACK/NACK is quiet
		// TODO:
		//if (ds_acknack_->state != init) return false;

		// Checks
		if (!psb_ || !psb_->pathOut() ||
		    !getInstance() || !rsb_) {
			G2RSVPTE_ERR("Invalid pointers");
			return false;
		}

		//FIX!!! Change ERO type
		//       (ExplicitRoute_Class, not ExplicitRoute_Object)
		if (psb_->pathOut()->ero_->cType() != CTYPE_ERO) {
			G2RSVPTE_ERR("temporarily, only ctype_ero is allowed");
			return false;
		}

		// Check the exclude/include masks consistency with TE link
		// colors  [ OPTIONAL now ??? TODO ]
		//if (!checkTELinkColors()) {
		//      TODO!!!
		//	return false;
		//}

		ExplicitRoute_Object * ero;
		ero = static_cast<ExplicitRoute_Object*>(psb_->pathOut()->ero_);

		if (getInstance()->adjType() == UNI) {
			delete ero;
			psb_->pathOut()->ero_ = 0;
			//
			// DL and label selection
			//
			net_res_spec_t net_res;
			if (getInstance()->isUniClient()) {
				net_res = in_net_res_;
			} else {
				net_res = eg_net_res_;
			}

			if (isNetResSpecValid(net_res)) {

				if (BITMASK_BITTEST(net_res.mask_, data_link) &&
				    BITMASK_BITTEST(net_res.mask_, label)) {

					//!!! for now only 32bit label are supported
					uint32_t lbl = net_res.label.value.
						label32.id;

					// Check dsTEL, dsDL and dsTxLabel on Ingress Node from TNA
					if (!checkSet_DsTEL_DsDL_DsTxLbl_fromTna(net_res.tna,
										 net_res.data_link,
										 lbl)) {
						G2RSVPTE_ERR("Invalid ingress parameters: ...");
						return false;
					}
				} else {
					// Choose downstream TEL, DL and label (rx) for TNA
					if (!setDsTEL_DsDL_DsTxLbl_fromTna(net_res.tna)) {
						G2RSVPTE_ERR("could not find any suitable TeLink/Data"
							     "Link/label under TNA %s",
							     addr_ntoa(net_res.tna));
						return false;
					}
				}

			} else {
				G2RSVPTE_ERR("could not find any suitable "
					     "downstream TeLink from TNA");
				return false;
			}

			// Get neighbor node id and set it as NHOP node id
			rsb_->nodeId(getNbrId());

		} else {
			// Try to fill the ERO (if it is empty)
			if (ero->get().subObjs.empty()) {
				G2RSVPTE_DBG("ERO is empty. Impossible to start signalling.");
				return false;
				//G2RSVPTE_DBG("ERO is empty ... trying to fill in by "
				//	     "path computation");
				//if (!fill_ero(nid_,                      // src node
				//	      parent()->nodeId(),        // dst node
				//	      ero->get().subObjs)) {
				//	G2RSVPTE_ERR("Fill ERO failed");
				//	return false;
				//}
			}
			// ERO Process
			if (!eroProcess()) {
				G2RSVPTE_ERR("ERO process failed!");
				return false;
			}
		}

		// prepare the PHOP in the outgoing Path
		if (!preparePHOPobj()) {
			return false;
		}

		// prepare the LABEL_SET obj in the outgoing Path
		//  (using downstream tx label)
		if (!prepareLabelSetObj()) {
			return false;
		}

		// prepare the UPSTREAM_LABEL obj in the outgoing Path
		//  (using downstream rx label)
		if (!prepareUpstreamLabelObj()) {
			return false;
		}

//		// Add this node to the RRO
//		if (psb_->pathOut()->rro_) {
//			if (psb_->pathIn()->rro_->cType() !=
//			    CTYPE_RECORDED_ROUTE) {
//				G2RSVPTE_ERR("temporarily, only ctype_rro "
//					     "is allowed");
//				return false;
//			}
//			RecordedRoute_Object * rro;
//			rro = static_cast<RecordedRoute_Object *>
//				(psb_->pathOut()->rro_);

//			if (!insertIntoRRO(rro, ingress)) {
//				G2RSVPTE_ERR("problems preparing outgoing "
//					     "Path RRO obj");
//				return false;
//			}
//		}

		// send a 'sendPath' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::G2RSVPTE_LSP_FSM_SendPath;

		fsmInst_->post(ev, context);
		//TODO add into FSM GEN if (!post) return false;

		G2RSVPTE_NOT("LSP signaledUP");

		return true;
	}

	bool LSP::sendDown(bool enqueueEvent)
	{
		if (!getInstance()) {
			G2RSVPTE_ERR("Null instance pointer!!!");
			return false;
		}

		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		if (isHead()) {      // HEAD
			// check that the DS ACK/NACK is quiet
			// TODO:
			//if (ds_acknack_->state != init) return false;

			// Checks
			if (!psb_ || !psb_->pathOut()) {
				G2RSVPTE_ERR("Invalid PSB/PathOut pointers");
				return false;
			}

			if (psb_->pathDown()) {
				G2RSVPTE_WRN("PathDown is already present... de"
					     "lete it before allocating again");
				delete psb_->pathDown();
				psb_->pathDown(0);
			}

			PathMessage * pathDown;
			if (getInstance()->adjType() == INNI) {
				pathDown = new PathMessage(*psb_->pathOut(), true, true);
			} else {
				pathDown = new PathMessage(*psb_->pathOut(), true, false);
			}

			if (!pathDown) {
				G2RSVPTE_ERR("Impossible to allocate memory for"
					     " PathDown Msg");
				return false;
			}
			psb_->pathDown(pathDown);

			// send a 'sendPathDown' basic event to the LSP FSM
			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to alocate context for"
					     " LSP fsm");
				return false;
			}
			context->lsp  = this;
			context->intf = 0;  //TODO new context? or only 1 cntx?
			context->msg  = 0;

			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
				G2RSVPTE_LSP_FSM_SendPathDown;

			fsmInst_->post(ev, context);

		} else {             // TAIL
			// check that the US ACK/NACK is quiet
			// TODO:
			//if (us_acknack_->state != init) return false;

			// Checks
			if (!rsb_ || !psb_ || !psb_->pathIn()) {
				G2RSVPTE_ERR("Invalid RSB/PSB/PathIn pointers");
				return false;
			}

			if (!rsb_->resvOut()) {
				ResvMessage * resv;
				resv = new ResvMessage(*psb_->pathIn());
				if (!resv) {
					G2RSVPTE_ERR("Impossible to create Resv to be sent "
						     "for LSP id %d", id_);
					return false;
				}
				// Initialise ResvOut for this new LSP
				rsb_->resvOut(resv);

				// prepare NHOP
				if (!prepareNHOPobj()) {
					G2RSVPTE_ERR("Impossible to prepare Resv Out NHOP obj");
					return false;
				}
			}

			if (rsb_->resvDown()) {
				G2RSVPTE_WRN("ResvDown is already present... de"
					     "lete it before allocating again");
				delete rsb_->resvDown();
				rsb_->resvDown(0);
			}

			ResvMessage * resvDown;
			resvDown = new ResvMessage(*rsb_->resvOut(), true);
			if (!resvDown) {
				G2RSVPTE_ERR("Impossible to allocate memory for"
					     " ResvDown Msg");
				return false;
			}
			rsb_->resvDown(resvDown);

			// send a 'sendResvDown' basic event to the LSP FSM
			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to alocate context for"
					     " LSP fsm");
				return false;
			}
			context->lsp  = this;
			context->intf = 0;  //TODO new context? or only 1 cntx?
			context->msg  = 0;

			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
				G2RSVPTE_LSP_FSM_SendResvDown;

			fsmInst_->post(ev, context, enqueueEvent);
			//TODO add into FSM GEN if (!post) return false;
		}

		G2RSVPTE_NOT("Posted Down event");

		return true;
	}

	bool LSP::sendTear(bool enqueueEvent)
	{
		if (!getInstance()) {
			G2RSVPTE_ERR("Null instance pointer!!!");
			return false;
		}

		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		if (isHead()) {      // HEAD
			// check that the DS ACK/NACK is quiet
			// TODO:
			//if (ds_acknack_->state != init) return false;

			// Checks
			if (!psb_ || !psb_->pathOut()) {
				G2RSVPTE_ERR("Invalid PSB/PathOut pointers");
				return false;
			}

			if (psb_->pathTear()) {
				G2RSVPTE_WRN("PathTear is already present... de"
					     "lete it before allocating again");
				delete psb_->pathTear();
				psb_->pathTear(0);
			}

			PathTearMessage * pathTear;
			pathTear = new PathTearMessage(*psb_->pathOut());
			if (!pathTear) {
				G2RSVPTE_ERR("Impossible to allocate memory for"
					     " PathTear Msg");
				return false;
			}
			psb_->pathTear(pathTear);

			// send a 'sendPathTear' basic event to the LSP FSM
			G2RSVPTE_LSP_FSM_context_t * context;
			context = new G2RSVPTE_LSP_FSM_context_t;
			if (!context) {
				G2RSVPTE_ERR("Impossible to alocate context for"
					     " LSP fsm");
				return false;
			}
			context->lsp  = this;
			context->intf = 0;  //TODO new context? or only 1 cntx?
			context->msg  = 0;

			fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
			ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
				G2RSVPTE_LSP_FSM_SendPathTear;

			fsmInst_->post(ev, context, enqueueEvent);
			//TODO add into FSM GEN if (!post) return false;

		} else {             // TAIL
			// check that the US ACK/NACK is quiet
			// TODO:
			//if (us_acknack_->state != init) return false;

			if (getInstance()->adjType() != INNI) {
				// Send a Path Err
				// Checks
				if (!psb_ || !psb_->pathOut() || !usAckNack()) {
					G2RSVPTE_ERR("Invalid usAckNack/PSB/PathOut pointers");
					return false;
				}

				if (psb_->pathErr()) {
					G2RSVPTE_WRN("PathErr already exists ... "
						     "deleting it before creating again");
					delete psb_->pathErr();
				        psb_->pathErr(0);
				}

				PathErrMessage * pathErr;
				// Set NOT_ERASE flags for this pkg
				pathErr = new PathErrMessage(*psb_->pathOut());
				// TODO FIXME: the copy must involve only: check the constructor
				if (!pathErr) {
					G2RSVPTE_ERR("could not allocate PathErr msg");
					return false;
				}

				//
				// Create Error Spec
				//
				IFIDIPv4ErrorSpec_Object * tmpErrSpec;
				int                        flags;
				int                        errCode;
				int                        errVal;

				flags   = G2RSVPTE_ERROR_SPEC_FLAG_PathStateRemoved;
				errCode = G2RSVPTE_ERROR_SPEC_ZERO;
				errVal  = G2RSVPTE_ERROR_SPEC_VAL_ZERO;
				tmpErrSpec = new IFIDIPv4ErrorSpec_Object(getInstance()->nodeId(),
									  flags,
									  errCode,
									  errVal);
				if (!tmpErrSpec) {
					G2RSVPTE_ERR("Impossible to allocate errorSpec subObj");
					delete pathErr;
					return false;
				}
				// Set error spec into PathErr msg
				pathErr->error_spec_ = tmpErrSpec;

				//
				// Set PathErr msg into PSB
				//
			        psb_->pathErr(pathErr);

				G2RSVPTE_INF("sending(US) PathErr msg (errCode = %d, "
					     "errVal = %d)",
					     errCode, errVal);

				if (!usAckNack()->sendMsg(psb_->pathErr())) {
					G2RSVPTE_ERR("could not deliver PathErr msg (errCode = "
						     "%d, errVal = %d) to upstream ACK/NACK",
						     errCode, errVal);
					return false;
				}
				//Unstore LSP
#ifdef PERSISTENCY_ON
				if (!getInstance()->persistency_.unstoreLsp(this)) {
					G2RSVPTE_ERR("Cannot unstore LSP data in Persistency Controller");
				}
#endif // PERSISTENCY_ON
				toBeDestroyed(true);
				postLspDeletion(getIdent(), getInstance()->adjType());
			} else {
				// Send a Resv Tear
				// Checks
				if (!rsb_ || !rsb_->resvOut()) {
					G2RSVPTE_ERR("Invalid RSB/ResvOut pointers");
					return false;
				}

				if (rsb_->resvTear()) {
					G2RSVPTE_WRN("ResvTear is already present... de"
						     "lete it before allocating again");
					delete rsb_->resvTear();
					rsb_->resvTear(0);
				}

				ResvTearMessage * resvTear;
				resvTear = new ResvTearMessage(*rsb_->resvOut());
				if (!resvTear) {
					G2RSVPTE_ERR("Impossible to allocate memory for"
						     " ResvTear Msg");
					return false;
				}
				rsb_->resvTear(resvTear);

				// send a 'sendResvTear' basic event to the LSP FSM
				G2RSVPTE_LSP_FSM_context_t * context;
				context = new G2RSVPTE_LSP_FSM_context_t;
				if (!context) {
					G2RSVPTE_ERR("Impossible to alocate context for"
						     " LSP fsm");
					return false;
				}
				context->lsp  = this;
				context->intf = 0;  //TODO new context? or only 1 cntx?
				context->msg  = 0;

				fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
				ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
					G2RSVPTE_LSP_FSM_SendResvTear;

				fsmInst_->post(ev, context, enqueueEvent);
				//TODO add into FSM GEN if (!post) return false;
			}
		}

		G2RSVPTE_NOT("Posted Tear event");

		return true;
	}

	bool LSP::sendError(const error_info_t & error,
			    bool enqueueEvent)
	{
		if (!getInstance()) {
			G2RSVPTE_ERR("Null instance pointer!!!");
			return false;
		}

		if (!isHead() && !isTail()) {
			G2RSVPTE_ERR("Neither tail nor head!!!");
			return false;
		}

		if (isHead()) {      // HEAD
			// check that the DS ACK/NACK is quiet
			// TODO:
			//if (ds_acknack_->state != init) return false;

			// Send Resv Err
			// Checks
			if (!rsb_ || !rsb_->resvIn() || !dsAckNack()) {
				G2RSVPTE_ERR("Invalid RSB/ResvIn/dsAckNack pointer!");
				return false;
			}

			if (rsb_->resvErr()) {
				G2RSVPTE_WRN("ResvErr already exists ... deleting it before creating again");
				delete rsb_->resvErr();
				rsb_->resvErr(0);
			}

			ResvErrMessage * resvErr;
			resvErr = new ResvErrMessage(*rsb_->resvIn());

			if (!resvErr) {
				G2RSVPTE_ERR("could not allocate ResvErr msg");
				return false;
			}

			//
			// Create Error Spec, for now IPv4 but with cb IF_ID_IPv4
			//
			IFIDIPv4ErrorSpec_Object * tmpResvErrSpec;
			tmpResvErrSpec = new IFIDIPv4ErrorSpec_Object(error.node_id,
								      error.flags,
								      error.err_code,
								      error.err_val);
			if (!tmpResvErrSpec) {
				G2RSVPTE_ERR("Impossible to allocate error spec subObj");
				delete resvErr;
				return false;
			}

			// Set error spec into ResvErr msg
			resvErr->error_spec_ = tmpResvErrSpec;

			//
			// Set ResvErr msg into RSB
			//
			rsb_->resvErr(resvErr);

			G2RSVPTE_INF("sending(DS) ResvErr msg (errCode = %d, errVal = %d)",
				     error.err_code, error.err_val);

			if (!dsAckNack()->sendMsg(rsb_->resvErr())) {
				G2RSVPTE_ERR("could not deliver ResvErr msg (errCode = %d, "
					     "errVal = %d) to downstream ACK/NACK",
					     error.err_code, error.err_val);
				return false;
			}
#ifdef PERSISTENCY_ON
			if (!getInstance()->persistency_.unstoreLsp(this)) {
				G2RSVPTE_ERR("Cannot unstore LSP data in Persistency Controller");
			}
#endif // PERSISTENCY_ON
			toBeDestroyed(true);
			postLspDeletion(getIdent(), getInstance()->adjType());

		} else {             // TAIL
			// check that the US ACK/NACK is quiet
			// TODO:
			//if (us_acknack_->state != init) return false;

			// Send a Path Err
			// Checks
			if (!psb_ || !psb_->pathOut() || !usAckNack()) {
				G2RSVPTE_ERR("Invalid usAckNack/PSB/PathOut pointers");
				return false;
			}

			if (psb_->pathErr()) {
				G2RSVPTE_WRN("PathErr already exists ... "
					     "deleting it before creating again");
				delete psb_->pathErr();
				psb_->pathErr(0);
			}

			PathErrMessage * pathErr;
			// Set NOT_ERASE flags for this pkg
			pathErr = new PathErrMessage(*psb_->pathOut());
			// TODO FIXME: the copy must involve only: check the constructor
			if (!pathErr) {
				G2RSVPTE_ERR("could not allocate PathErr msg");
				return false;
			}

			//
			// Create Error Spec
			//
			IFIDIPv4ErrorSpec_Object * tmpPathErrSpec;

			tmpPathErrSpec = new IFIDIPv4ErrorSpec_Object(error.node_id,
								      error.flags,
								      error.err_code,
								      error.err_val);
			if (!tmpPathErrSpec) {
				G2RSVPTE_ERR("Impossible to allocate errorSpec subObj");
				delete pathErr;
				return false;
			}
			// Set error spec into PathErr msg
			pathErr->error_spec_ = tmpPathErrSpec;

			//
			// Set PathErr msg into PSB
			//
			psb_->pathErr(pathErr);

			G2RSVPTE_INF("sending(US) PathErr msg (errCode = %d, "
				     "errVal = %d)",
				     error.err_code, error.err_val);

			if (!usAckNack()->sendMsg(psb_->pathErr())) {
				G2RSVPTE_ERR("could not deliver PathErr msg (errCode = "
					     "%d, errVal = %d) to upstream ACK/NACK",
					     error.err_code, error.err_val);
				return false;
			}
#ifdef PERSISTENCY_ON
			if (!getInstance()->persistency_.unstoreLsp(this)) {
				G2RSVPTE_ERR("Cannot unstore LSP data in Persistency Controller");
			}
#endif // PERSISTENCY_ON
			toBeDestroyed(true);
			postLspDeletion(getIdent(), getInstance()->adjType());

		}

		G2RSVPTE_NOT("Posted Error event");

		return true;
	}

	bool LSP::sendConfirm(bool enqueueEvent)
	{
		if (!psb_ || !psb_->pathOut() || !rsb_ || !rsb_->resvIn()) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		if (!isHead()) {
			G2RSVPTE_ERR("This NE is not the head!!!");
			return false;
		}

		// check that the DS ACK/NACK is quiet
		// TODO:
		//if (ds_acknack_->state != init) return false;

		ResvConfMessage * conf;
		conf = new ResvConfMessage(*psb_->pathOut(), *rsb_->resvIn());
		if (!conf) {
			G2RSVPTE_ERR("Impossible to create Confirm to be sent "
				     "for LSP id %d", id_);
			return false;
		}
		// Initialise ResvConf for this LSP
		psb_->resvConf(conf);

		// set error spec
		assert(psb_->resvConf()->error_spec_);
		if (psb_->resvConf()->error_spec_->cType() != CTYPE_ERROR_SPEC_IPV4 &&
		    psb_->resvConf()->error_spec_->cType() != CTYPE_ERROR_SPEC_IF_ID_IPV4) {
			G2RSVPTE_ERR("cType %d not supported for ErrorSpec Obj",
				     psb_->resvConf()->error_spec_->cType());
			return false;
		}
		assert(getInstance());
		static_cast<IFIDIPv4ErrorSpec_Object *>
			(psb_->resvConf()->error_spec_)->get().ipv4_err_node.
			s_addr = htonl(getInstance()->nodeId());
		static_cast<IFIDIPv4ErrorSpec_Object *>
			(psb_->resvConf()->error_spec_)->get().flags       = 0;
		static_cast<IFIDIPv4ErrorSpec_Object *>
			(psb_->resvConf()->error_spec_)->get().error_code  =
			G2RSVPTE_ERROR_SPEC_CONFIRM;
		static_cast<IFIDIPv4ErrorSpec_Object *>
			(psb_->resvConf()->error_spec_)->get().error_value = 0;

		// set resv_confirm_ (done into constructor!!!)

		G2RSVPTE_NOT("Posting a SendConfirm to LSP FSM");

		// send a 'SendConfirm' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
			G2RSVPTE_LSP_FSM_SendConfirm;

		fsmInst_->post(ev, context, enqueueEvent);
		//TODO add into FSM GEN if (!post) return false;

		G2RSVPTE_NOT("Posted SendConfirm event");

		return true;
	}

	bool LSP::xConnCompleted(bool enqueueEvent)
	{
		G2RSVPTE_NOT("Posting a XConnCompleted event to LSP FSM");

		// send a 'XConnCompleted' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
			G2RSVPTE_LSP_FSM_XConnCompleted;

		fsmInst_->post(ev, context, enqueueEvent);
		//TODO add into FSM GEN if (!post) return false;

		G2RSVPTE_NOT("Posted XConnCompleted event");

		return true;
	}

	bool LSP::xConnFailed(bool enqueueEvent)
	{
		G2RSVPTE_NOT("Posting a XConnErr event to LSP FSM");

		// send a 'XConnErr' basic event to the LSP FSM
		G2RSVPTE_LSP_FSM_context_t * context;
		context = new G2RSVPTE_LSP_FSM_context_t;
		if (!context) {
			G2RSVPTE_ERR("Impossible to alocate context for LSP "
				     "fsm");
			return false;
		}
		context->lsp  = this;
		context->intf = 0;  //TODO new context? or only 1 cntx for FSM?
		context->msg  = 0;

		fsm::G2RSVPTE_LSP_FSM::virtFsm::root_events_t ev;
		ev = fsm::G2RSVPTE_LSP_FSM::virtFsm::
			G2RSVPTE_LSP_FSM_XConnErr;

		fsmInst_->post(ev, context, enqueueEvent);
		//TODO add into FSM GEN if (!post) return false;

		G2RSVPTE_NOT("Posted XConnErr event");

		return true;
	}

	//
	// Utilities
	//
	bool LSP::isHead(void)
	{
		assert(parent());

		// if LSP node == G2RSVPTE instance node  =>  HEAD
		if (nid_ == parent()->parent()->nodeId()) {
			G2RSVPTE_DBG("I am the head node!");
			return true;
		}
		return false;
	}

	bool LSP::isTail(void)
	{
		assert(parent());
		assert(parent()->parent());

		// if session node == G2RSVPTE instance node  =>  TAIL
		if (parent()->nodeId() == parent()->parent()->nodeId()) {
			G2RSVPTE_DBG("I am the tail node!");
			return true;
		}
		return false;
	}

	bool LSP::checkGnsTna(bool ingress)
	{
		if (isTail()) {
			if (!psb() || !psb()->pathIn()) {
				G2RSVPTE_WRN("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!psb()->pathIn()->gns_uni_in_) {
						G2RSVPTE_WRN("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!psb()->pathIn()->gns_uni_eg_) {
						G2RSVPTE_WRN("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
			}
		} else if (isHead()) {
			if (!rsb() || !rsb()->resvIn()) {
				G2RSVPTE_WRN("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!rsb()->resvIn()->gns_uni_in_) {
						G2RSVPTE_WRN("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!rsb()->resvIn()->gns_uni_eg_) {
						G2RSVPTE_WRN("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
			}
		} else {
			if (!psb() || !psb()->pathOut()) {
				G2RSVPTE_WRN("Null pointer !!!");
				return false;
			}

			switch (ingress) {
				case true:
					if (!psb()->pathOut()->gns_uni_in_) {
						G2RSVPTE_WRN("iGNS-TNA pointer null");
						return false;
					}
					break;
				case false:
					if (!psb()->pathOut()->gns_uni_eg_) {
						G2RSVPTE_WRN("eGNS-TNA pointer null");
						return false;
					}
					break;
				default:
					return false;
			}
		}

		return true;
	}

	bool LSP::checkPolicy(void)
	{
		if (isHead()) {
			if (!psb() || !psb()->pathOut() || !psb()->pathOut()->policy_data_) {
				G2RSVPTE_WRN("Null pointer !!!");
				return false;
			}
		} else {
			if (!psb() || !psb()->pathIn() || !psb()->pathIn()->policy_data_) {
				G2RSVPTE_WRN("Null pointer !!!");
				return false;
			}
		}

		return true;
	}

	bool LSP::checkEro(void)
	{
		if (!psb() || !psb()->pathOut() || !psb()->pathOut()->ero_) {
			G2RSVPTE_WRN("Null pointer !!!");
			return false;
		}

		return true;
	}

	G2RSVPTE * LSP::getInstance(void)
	{
		if (parent()) {
			return parent()->parent();
		}
		return 0;
	}

#ifdef CRANKBACK_ON
	bool LSP::processLSPAttrObj(void)
	{
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->lsp_attributes_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}
		LSPAttr_Class * tmp;
		tmp = psb_->pathIn()->lsp_attributes_;

		if (tmp->cType() != CTYPE_LSP_ATTRIBUTES) {
			G2RSVPTE_ERR("Not supported LSP_ATTRIBUTES object type!");
			return false;
		}

		LSPAttr_Object * lspAttr;
	        lspAttr = static_cast<LSPAttr_Object *>(tmp);

		if (lspAttr->get().tlvs.size() > 1) {
			G2RSVPTE_ERR("Multiple TLVs for LSP Attributes Object");
			return false;
		}

		std::list<TLV *>::const_iterator it;
		for (it = lspAttr->get().tlvs.begin();
		     it != lspAttr->get().tlvs.end();
		     it++) {
			switch ((*it)->type()) {
				case TLV_LSP_ATTR:
					LSP_ATTR_TLV * tlv;
					tlv = static_cast<LSP_ATTR_TLV *> (*it);
					if ((tlv->get().e2e_cb      && tlv->get().boundary_cb) ||
					    (tlv->get().e2e_cb      && tlv->get().segment_cb) ||
					    (tlv->get().boundary_cb && tlv->get().boundary_cb)) {
						G2RSVPTE_ERR("Multiple bits set in TLV for LSP Attributes Object");
						return false;
					}
					if (tlv->get().e2e_cb) {
						cbackScope(CB_E2E);
					} else if (tlv->get().boundary_cb) {
						cbackScope(CB_BOUNDARY);
					} else if (tlv->get().segment_cb) {
						cbackScope(CB_SEGMENT);
					} else {
						G2RSVPTE_ERR("None of the crankback bit is set "
							     "in LSP Attributes TLV");
						return false;
					}
					break;
				default:
					G2RSVPTE_ERR("Unknown TLV type for LSP Attributes Object");
					return false;
			}

		}

		G2RSVPTE_DBG("Crankback scope (%s) correctly set for this LSP",
			     SHOW_CBACK_SCOPE(cbackScope()));

		return true;
	}
#endif // CRANKBACK_ON

	bool LSP::processNHOPobj(void)
	{
		if (!rsb_ || !rsb_->resvIn() || !rsb_->resvIn()->rsvp_hop_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		RSVP::Rsvp_Hop_Class * rsvpHop;
		rsvpHop = rsb_->resvIn()->rsvp_hop_;

		if (rsvpHop->cType() != CTYPE_RSVP_HOP_IF_ID_IPV4) {
			// ERR("Not supported RSVP_HOP type");
			return false;
		}
		RSVP::IFIDIPv4RsvpHop_Object * rsvpHopIfIdIPv4;
		rsvpHopIfIdIPv4 = static_cast<RSVP::IFIDIPv4RsvpHop_Object *>
			(rsvpHop);

		if (ntohl(rsvpHopIfIdIPv4->get().next_prev_ipv4_hop.s_addr) !=
		    rsb_->nodeId()) {
			G2RSVPTE_ERR("node id in NHOP (%s) is not the one that "
				     "I used to know and send to (%s)",
				     inet_ntoa(rsvpHopIfIdIPv4->get().
					       next_prev_ipv4_hop),
				     numb_htoa(rsb_->nodeId()));
			return false;
		}
		if (rsvpHopIfIdIPv4->get().logical_if_hdl != rsb_->lih()) {
			G2RSVPTE_ERR("LIH in NHOP (%d) is not the one that "
				     "I set in my previous PHOP (%d)",
				     rsvpHopIfIdIPv4->get().logical_if_hdl,
				     rsb_->lih());
			return false;
		}

		// Checking just the first TLV
		if (rsvpHopIfIdIPv4->get().tlvs.size()) {

			RSVP::TLV * hopTLV;
			hopTLV = *rsvpHopIfIdIPv4->get().tlvs.begin();
			if (!hopTLV) {
				G2RSVPTE_ERR("Invalid TLV obj");
				return false;
			}

			g2mpls_addr_t add;
			switch (hopTLV->type()) {
				case IF_ID_TLV_IPV4: {
					RSVP::IPv4_TLV * tlv;
					tlv = static_cast<RSVP::IPv4_TLV *>
						(hopTLV);

					add.type       = IPv4;
					add.preflen    = 32;
					add.value.ipv4 = tlv->get().ipv4_addr;
					rsb_->remoteDL(add);
					break;
				}
				case IF_ID_TLV_IPV6: {
					RSVP::IPv6_TLV * tlv;
					tlv = static_cast<RSVP::IPv6_TLV *>
						(hopTLV);

					add.type       = IPv6;
					add.preflen    = 128;
					add.value.ipv6 = tlv->get().ipv6_addr;
					rsb_->remoteDL(add);
					break;
				}
					//case IF_ID_TLV_COMP_IF_DOWN:
					//case IF_ID_TLV_COMP_IF_UP:

				case IF_ID_TLV_IF_INDEX:
				default:
					RSVP::IF_IDX_TLV * tlv;
					tlv = static_cast<RSVP::IF_IDX_TLV *>
						(hopTLV);

		//?? if (rsvpHopIfIdIPv4->get().next_prev_ipv4_hop.s_addr !=
		//??			tlv->get().ip_address) {
		//??	G2RSVPTE_ERR("TLV NHOP nodeId (%s) is not equal"
		//??	    " to main NHOP nodeId (%s)",
		//??	    inet_ntoa ( tlv->get().ip_address ),
		//??	    numb_htoa(rsb_->nodeId())); // == rsvpHopIfIdIPv4->
		//??			get().next_prev_ipv4_hop.s_addr
		//??	return false;
		//?? }

					add.type                = UNNUMBERED;
					add.preflen             = 0;
					//add.value.unnum.node_id = tlv->get().
					//	ip_address;
					//add.value.unnum.addr    = tlv->get().
					//	interface_id;
					add.value.unnum         =
						tlv->get().interface_id;
					rsb_->remoteDL(add);
					break;
			}

			//TODO: check?????
			if (!addr_equal(add, rsb_->localDL())) {
				G2RSVPTE_ERR("NHOP's DL %s ...",
					     addr_ntoa(add));
				G2RSVPTE_ERR("... does not match my ds DL %s",
					     addr_ntoa(rsb_->localDL()));
				return false;
			}

			//Check that the provided NHOP DL is the same ofour dsDL
			//?? TODO needed???
			//?? if (!getLocDLid_on_remDLid(rsb_->nodeId(),
			//??   //g2rsvpte needed???
			//??			   rsb_->remoteDL(),
			//??			   &add,
			//??			   0,
			//??			   0) ||
			//??     !addr_equal(add, rsb_->localDL())) {
			//??	G2RSVPTE_ERR("NHOP's DL %d does not match my "
			//??	             "downstream DL %s",
			//??	    add, rsb_->localDL());
			//??	return false;
			//?? }
			//?? DBG("NHOP's DL %d matches my DL %d\n",
			//??     rsb_->remoteDL(), rsb_->localDL());

			// check that my DL is still Up
			// .... see processPHOPobj
		}

		return true;
	}

	bool LSP::prepareNHOPobj(void)
	{
		if (!rsb_ || !rsb_->resvOut()) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		// PathIn msg MUST be present!!!
		assert(psb_);
		assert(psb_->pathIn());
		assert(psb_->pathIn()->rsvp_hop_);


		// Copy RSVP_HOP from pathIn into resvOut
		if (rsb_->resvOut()->rsvp_hop_) {
			G2RSVPTE_WRN("RSVP_HOP already exists into resvOut "
				     "msg ... overwriting!!!");
			delete rsb_->resvOut()->rsvp_hop_;
		}

		rsb_->resvOut()->rsvp_hop_ = psb_->pathIn()->rsvp_hop_->clone();
		if (!rsb_->resvOut()->rsvp_hop_) {
			G2RSVPTE_ERR("Impossible to allocate RSVP_HOP obj "
				     "for resvOut msg!");
			return false;
		}

		// Get g2rsvpte instance
		RSVP::G2RSVPTE * proto = getInstance();
		assert(proto);

		RSVP::Rsvp_Hop_Class * rsvpHop;
		rsvpHop = rsb_->resvOut()->rsvp_hop_;

		//check nedeed???
		if (rsvpHop->cType() != CTYPE_RSVP_HOP_IF_ID_IPV4) {
			// ERR("Not supported RSVP_HOP type");
			return false;
		}
		RSVP::IFIDIPv4RsvpHop_Object * rsvpHopIfIdIPv4;
		rsvpHopIfIdIPv4 = static_cast<RSVP::IFIDIPv4RsvpHop_Object *>
			(rsvpHop);

		rsvpHopIfIdIPv4->get().next_prev_ipv4_hop.s_addr =
			htonl(proto->nodeId());
		// set LIH with LIH we got from PHOP
		rsvpHopIfIdIPv4->get().logical_if_hdl = psb_->lih();

		G2RSVPTE_DBG("going to use DL %s ... ",
			     addr_ntoa(psb_->localDL()));
		G2RSVPTE_DBG(" ... under TE link %s to PHOP",
			     ((psb_->interface()) ?
			     addr_ntoa((psb_->interface())->localId()):"NULL"));

		return true;
	}

	bool LSP::processLabelObj(void)
	{
		if (!rsb_ || !rsb_->resvIn() || !rsb_->resvIn()->flow_desc_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		RSVP::Label_Class * label;
		label = rsb_->resvIn()->flow_desc_->getLabelObj();
		if (!label ||
		    label->cType() != CTYPE_GENERALISED_LABEL) {
			G2RSVPTE_ERR("Invalid pointer to label obj!");
			return false;
		}

		RSVP::GLabel_Object * genLabel;
		genLabel = static_cast<RSVP::GLabel_Object *>(label);

		if (genLabel->get().size != sizeof(uint32_t)) {
			G2RSVPTE_ERR("only 4-bytes-labels are accepted");
			return false;
		}

		// Check that this label matches the one sent in LABEL SET obj
		if (!genLabel->get().label ||
		    *genLabel->get().label != rsb_->txLabel()) {
			G2RSVPTE_ERR("the label chosen by my NHOP (0x%x) is not"
				     " the one I set in my previous LABEL SET "
				     "(0x%x)",
				     *(genLabel->get().label),
				     rsb_->txLabel());

			return false;
		}

		G2RSVPTE_DBG("The label chosen by my NHOP (0x%x) is in my "
			     "previous LABEL SET",
			     rsb_->txLabel());

		return true;
	}

	bool LSP::prepareLabelObj(void)
	{
		if (!rsb_ || !rsb_->resvOut() || !psb_ ||
		    !rsb_->resvOut()->flow_desc_) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		Label_Class * lbl;
		lbl = rsb_->resvOut()->flow_desc_->getLabelObj();
		if (!lbl) {
			lbl = new RSVP::GLabel_Object();

			if (!lbl) {
				G2RSVPTE_ERR("Impossible to allocate generic "
					     "Label obj for NHOP!");
				return false;
			}

			rsb_->resvOut()->flow_desc_->setLabelObj(lbl);
		}

		//lbl = rsb_->resvOut()->getLabelObj();
		if (lbl->cType() != CTYPE_GENERALISED_LABEL) { //useles???
			G2RSVPTE_ERR("Invalid pointer to label obj!");
			return false;
		}

		RSVP::GLabel_Object * genLabel;
		genLabel = static_cast<RSVP::GLabel_Object *>(lbl);
		assert(genLabel);

		genLabel->get().size  = sizeof(uint32_t);
		genLabel->get().label = new uint32_t;
		if (!genLabel->get().label || !genLabel->get().size) {
			G2RSVPTE_ERR("Impossible to allocate memory for "
				     "generic label");
			delete genLabel; //?????
			return false;
		}
		// the upstream label is equal to the downstream one
		*genLabel->get().label = psb_->rxLabel();

		return true;
	}

	// TODO: it could be a RSVP::RecordedRoute_Object method!!!
	bool LSP::insertIntoRRO(RSVP::RecordedRoute_Class * genRRO,    //lsp->psb()->pathOut()->rro_ or lsp->rsb()->resvOut()->getRRO()
				g2mpls_addr_t               addr)      //lsp->[psb()/rsb()]->interface()->localId()
				//Interface *              interface)  //lsp->psb()->interface()     or lsp->rsb()->interface()
	{
		if (!genRRO) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		//FIX!!! Change rro type (RecordedRoute_Class, not RecordedRoute_Object)
		if (genRRO->cType() != CTYPE_RECORDED_ROUTE) {
			G2RSVPTE_ERR("Only CTYPE_RECORDED_ROUTE is allowed");
			return false;
		}
		RSVP::RecordedRoute_Object * rro;
		rro = static_cast<RSVP::RecordedRoute_Object *>(genRRO);

		// Get g2rsvpte instance
		RSVP::G2RSVPTE * g2rsvpte = getInstance();
		assert(g2rsvpte);

		if (!g2rsvpte->nodeId()) {
			G2RSVPTE_ERR("Invalid node Id for G2RSVPTE instance!");
			return false;
		}

		// TODO: FIXME   LRM??? hub_get_lk_loclkid(interface) // ???NHOP TE link or PHOP??
		//               G2RSVPTE_ERR("can't find NHOP TE link");
		if (is_addr_null(addr)) {
			G2RSVPTE_ERR("Invalid g2mpls address(NULL)");
			return false;
		}

		switch (addr.type) {
			case IPv4: {
				RSVP::IPv4Addr_SubObj * sObj = new RSVP::IPv4Addr_SubObj();
				if (!sObj) {
					G2RSVPTE_ERR("Impossible to allocate memory for IPv4 RRO subObj");
					return false;
				}
				sObj->get().ipv4_addr.s_addr = addr.value.ipv4.s_addr;
				sObj->get().preflen          = G2RSVPTE_DEFAULT_IPv4_PREFLEN;
				sObj->get().flags            = 0;

				// Insert (head) IPv4 subObject into RRO
				rro->get().subObjs.push_front(sObj);
				break;
			}
			case IPv6: {
				RSVP::IPv6Addr_SubObj * sObj = new RSVP::IPv6Addr_SubObj();
				if (!sObj) {
					G2RSVPTE_ERR("Impossible to allocate memory for IPv6 RRO subObj");
					return false;
				}
				sObj->get().ipv6_addr = addr.value.ipv6; // ???
				sObj->get().preflen   = G2RSVPTE_DEFAULT_IPv6_PREFLEN;
				sObj->get().flags     = 0;

				// Insert (head) IPv6 subObject into RRO
				rro->get().subObjs.push_front(sObj);
				break;
			}
			case UNNUMBERED: {
				RSVP::Unnumbered_SubObj * sObj = new RSVP::Unnumbered_SubObj();
				if (!sObj) {
					G2RSVPTE_ERR("Impossible to allocate memory for unnum RRO subObj");
					return false;
				}
				//sObj->get().interface_id = addr.value.unnum.addr;
				//sObj->get().router_id    = addr.value.unnum.node_id; // XXX or overwrite with g2rsvpte->nodeId();
				sObj->get().interface_id = addr.value.unnum;
				sObj->get().router_id    = g2rsvpte->nodeId();
				sObj->get().flags        = 0;

				// Insert (head) unnum subObject into RRO
				rro->get().subObjs.push_front(sObj);
				break;
			}
			default: {
				G2RSVPTE_ERR("unknown RRO subobject type");
				return false;
			}
		}

		G2RSVPTE_DBG("added type %d RRO subobject",
			     (*rro->get().subObjs.begin())->type());

		//assert(rsb_);
		//bool ds = (interface == rsb_->interface());
		// TODO: FIXME!!! optional for now:
		// - insert component i/f id, if desired
		// - insert labels, if desired

		return true;
	}

	bool LSP::processPHOPobj(void)
	{
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->rsvp_hop_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		RSVP::Rsvp_Hop_Class * rsvpHop;
		rsvpHop = psb_->pathIn()->rsvp_hop_;

		if (rsvpHop->cType() != CTYPE_RSVP_HOP_IF_ID_IPV4) {
			G2RSVPTE_ERR("Not supported RSVP_HOP type!");
			return false;
		}
		RSVP::IFIDIPv4RsvpHop_Object * rsvpHopIfIdIPv4;
		rsvpHopIfIdIPv4 = static_cast<RSVP::IFIDIPv4RsvpHop_Object *>
			(rsvpHop);

		// Setting previous HOP node id
		psb_->nodeId(ntohl(rsvpHopIfIdIPv4->get().
				   next_prev_ipv4_hop.s_addr));

		// Setting logical interface handler in NHOP
		psb_->lih(rsvpHopIfIdIPv4->get().logical_if_hdl);

		// Processing just the first TLV
		if (rsvpHopIfIdIPv4->get().tlvs.size()) {

			RSVP::TLV * hopTLV;
			hopTLV = *rsvpHopIfIdIPv4->get().tlvs.begin();
			if (!hopTLV) {
				G2RSVPTE_ERR("Invalid TLV obj");
				return false;
			}

			g2mpls_addr_t add;
			switch (hopTLV->type()) {
				case IF_ID_TLV_IPV4: {
					RSVP::IPv4_TLV * tlv;
					tlv = static_cast<RSVP::IPv4_TLV *>
						(hopTLV);

					add.type       = IPv4;
					add.preflen    = 32;
					add.value.ipv4 = tlv->get().ipv4_addr;
					psb_->remoteDL(add);
					break;
				}
				case IF_ID_TLV_IPV6: {
					RSVP::IPv6_TLV * tlv;
					tlv = static_cast<RSVP::IPv6_TLV *>
						(hopTLV);

					add.type       = IPv6;
					add.preflen    = 128;
					add.value.ipv6 = tlv->get().ipv6_addr;
					psb_->remoteDL(add);
					break;
				}
					//case IF_ID_TLV_COMP_IF_DOWN:
					//case IF_ID_TLV_COMP_IF_UP:

				case IF_ID_TLV_IF_INDEX:
				default:
					RSVP::IF_IDX_TLV * tlv;
					tlv = static_cast<RSVP::IF_IDX_TLV *>
						(hopTLV);

					if (psb_->nodeId() !=
					    tlv->get().ip_address) {
						G2RSVPTE_ERR("TLV PHOP nodeId "
							     "%s ...", numb_htoa
							     (tlv->get().
							      ip_address));
						G2RSVPTE_ERR(" ... is not equal"
							     " to main PHOP "
							     "nodeId %s",
							     numb_htoa
							     (psb_->nodeId()));
						return false;
					}

					add.type                = UNNUMBERED;
					add.preflen             = 0;
					//add.value.unnum.node_id =
					//	tlv->get().ip_address;
					//add.value.unnum.addr    =
					//	tlv->get().interface_id;
					add.value.unnum         =
						tlv->get().interface_id;
					psb_->remoteDL(add);
					break;
			}

			// Get the local data link from the remote one and
			// check that the provided PHOP Data Link is in one
			// of our TE links
			g2mpls_addr_t localDL;
			admstate_t    dlAdminStatus;
			opstate_t     dlOperStatus;

			if (!getLocDLid_on_remDLid(psb_->nodeId(), //g2rsvpte ??
						   psb_->remoteDL(),
						   localDL,
						   dlAdminStatus,
						   dlOperStatus)) {
				G2RSVPTE_ERR("PHOP %s's DL %s does not match "
					     "any of mines",
					     numb_htoa(psb_->nodeId()),
					     addr_ntoa(psb_->remoteDL()));
				return false;
			}
			psb_->localDL(localDL);

			G2RSVPTE_DBG("PHOP's DL %s on",
				     addr_ntoa(psb_->remoteDL()));
			G2RSVPTE_DBG(" nodeId %s matches",
				     numb_htoa(psb_->nodeId()));
			G2RSVPTE_DBG(" my DL %s ... gonna check its states",
				     addr_ntoa(psb_->localDL()));

			// Check that this local data link is UP (adm and oper)
			if (dlAdminStatus != ENABLED || dlOperStatus != UP) {
				G2RSVPTE_ERR("my DL %s is not Up[Free|"
					     "Allocated] and admin Enabled; "
					     "can't work on it",
					     addr_ntoa(psb_->localDL()));
				return false;
			}

			// Get the TE-link where local Data link belongs to
			RSVP::Interface * intf;
			intf = getTELink_on_DLid(getInstance(),
						 psb_->localDL());
			if (!intf) {
				G2RSVPTE_ERR("my DL %s is not correlated with "
					     "a TE link",
					     addr_ntoa(psb_->localDL()));
				return false;
			}

			// Check that this TE data link is UP (adm and oper)
			if (intf->admState() != ENABLED ||
			    intf->opState() != UP) {
				G2RSVPTE_ERR("my TE link %s is not Up[Free|"
					     "Allocated] and admin Enabled; "
					     "can't work on it",
					     addr_ntoa(intf->localId()));
				return false;
			}

			psb_->interface(intf);

			G2RSVPTE_DBG("my DL %s belongs to my TE-LINK:",
				     addr_ntoa(psb_->localDL()));
			psb_->interface()->dump();
		}

		return true;
	}

	bool LSP::preparePHOPobj(void)
	{
		if (!psb_ || !psb_->pathOut()) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		// Get proto instance
		RSVP::G2RSVPTE * proto = getInstance();
		assert(proto);

		if (!psb_->pathOut()->rsvp_hop_) {

			RSVP::IFIDIPv4RsvpHop_Object * rsvpHopIfIdIPv4;
			rsvpHopIfIdIPv4 = new RSVP::IFIDIPv4RsvpHop_Object();

			if (!rsvpHopIfIdIPv4) {
				G2RSVPTE_ERR("Impossible to allocate "
					     "(IF_ID_IPv4) RSVP_HOP for PHOP!");
				return false;
			}

			psb_->pathOut()->rsvp_hop_ = rsvpHopIfIdIPv4;
		}

		RSVP::Rsvp_Hop_Class * rsvpHop;
		rsvpHop = psb_->pathOut()->rsvp_hop_;

		//check nedeed???
		if (rsvpHop->cType() != CTYPE_RSVP_HOP_IF_ID_IPV4) {
			G2RSVPTE_ERR("Not supported RSVP_HOP type");
			return false;
		}
		RSVP::IFIDIPv4RsvpHop_Object * rsvpHopIfIdIPv4;
		rsvpHopIfIdIPv4 = static_cast<RSVP::IFIDIPv4RsvpHop_Object *>
			(rsvpHop);

		rsvpHopIfIdIPv4->get().next_prev_ipv4_hop.s_addr =
			htonl(proto->nodeId());
		rsvpHopIfIdIPv4->get().logical_if_hdl = 0;

		if (!rsb_ || !rsb_->interface()) {
			G2RSVPTE_ERR("Invalid RSB or RSB->interface pointers!");
			return false;
		}

		G2RSVPTE_DBG("TE-Link %s, selected:",
			     addr_ntoa(rsb_->interface()->localId()));

		if (!is_addr_null(rsb_->localDL())      &&
		    //(!DLp || lsp->rsb()->localDL() != *DLp) &&
		    rsb_->txLabel()
		    //&& (!LBLp || lsp->rsb()->txLabel() != *LBLp)
		    ) {
			if (proto->adjType() != UNI) {
				G2RSVPTE_INF("(DL, label) in TE link %s imposed "
					     "through ERO",
					     addr_ntoa(rsb_->interface()->localId()));
			}
		} else {
			// If downstream DL and label (tx) are not specified
			// in ERO (std case), need to choose them under the
			// specified TE link

			if (!setDsDL_DsTxLbl_fromTELink(rsb_->interface()->
							localId())) {
				G2RSVPTE_ERR("could not find any suitable "
					     "DL/label under TE link %s",
					     addr_ntoa(rsb_->interface()->
						       localId()));
				// set error params
				errTELink(rsb_->interface()->localId());
				return false;
			}
		}
		// We assume that the DS RX label is equal to the DS TX one
		rsb_->rxLabel(rsb_->txLabel());

		G2RSVPTE_DBG("\t Data Link  %s",
			     addr_ntoa(rsb_->localDL()));
		G2RSVPTE_DBG("\t tx label   0x%x",
			     rsb_->txLabel());
		G2RSVPTE_DBG("\t rx label   0x%x",
			     rsb_->rxLabel());

		if (rsvpHopIfIdIPv4->get().tlvs.size()) { // Usefull ? Correct ?
			G2RSVPTE_ERR("TLV subObj list should be empty, "
				     "but there are %d elems",
				     rsvpHopIfIdIPv4->get().tlvs.size());
			return false;
		}

		switch (rsb_->localDL().type) {
			case IPv4: {
				RSVP::IPv4_TLV * tlv = new RSVP::IPv4_TLV();
				if (!tlv) {
					G2RSVPTE_ERR("Impossible to allocate "
						     "memory for IPv4 TLV");
					return false;
				}
				tlv->get().ipv4_addr.s_addr =
					rsb_->localDL().value.ipv4.s_addr;
				//Set dsDL in psb_->pathOut()->rsvp_hop_ subTLV
				rsvpHopIfIdIPv4->get().tlvs.push_back(tlv);
				break;
			}
			case IPv6: {
				RSVP::IPv6_TLV * tlv = new RSVP::IPv6_TLV();
				if (!tlv) {
					G2RSVPTE_ERR("Impossible to allocate "
						     "memory for IPv6 TLV");
					return false;
				}
				tlv->get().ipv6_addr=rsb_->localDL().value.ipv6;
				//Set dsDL in psb_->pathOut()->rsvp_hop_ subTLV
				rsvpHopIfIdIPv4->get().tlvs.push_back(tlv);
				break;
			}
			case UNNUMBERED:
			default: {
				RSVP::IF_IDX_TLV * tlv = new RSVP::IF_IDX_TLV();
				if (!tlv) {
					G2RSVPTE_ERR("Impossible to allocate "
						     "memory for IF_IDX TLV");
					return false;
				}
				//tlv->get().ip_address   =
				//	rsb_->localDL().value.unnum.node_id;
				//tlv->get().interface_id =
				//	rsb_->localDL().value.unnum.addr;
				tlv->get().ip_address   = proto->nodeId();
				tlv->get().interface_id =
					rsb_->localDL().value.unnum;
				//Set dsDL in psb_->pathOut()->rsvp_hop_ subTLV
				rsvpHopIfIdIPv4->get().tlvs.push_back(tlv);
				break;
			}
		}

		return true;
	}

	bool LSP::processLabelSetObj(void)
	{
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->label_set_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		if (psb_->pathIn()->label_set_->cType() != CTYPE_LABEL_SET) {
			G2RSVPTE_ERR("temporarily, only cType = %d is"
				     "allowed for Label Set obj",
				     CTYPE_LABEL_SET);
			return false;
		}
		RSVP::LabelSet_Object * labelSet;
		labelSet = static_cast<RSVP::LabelSet_Object *>
			(psb_->pathIn()->label_set_);

		// Check that there is one and only one inclusive subchannel
		if (labelSet->get().action        != 0                       ||
		    labelSet->get().label_type    != CTYPE_GENERALISED_LABEL ||
		    labelSet->get().labels.size() != 1) {
			G2RSVPTE_ERR("temporarily, only 1 inclusive generalized"
				     "-label subchannel is allowed (%d elems)",
				     labelSet->get().labels.size());
			return false;
		}

		if (labelSet->get().labels.empty()) {
			G2RSVPTE_ERR("SubChannel list is empty!");
			return false;
		}

		RSVP::Label_Class * label = labelSet->get().labels.front();
		if (!label ||
		    label->cType() != CTYPE_GENERALISED_LABEL) { // Useless???
			G2RSVPTE_ERR("SubChannel list has invalid pointer!");
			return false;
		}

		RSVP::GLabel_Object * genLabel;
		genLabel = static_cast<RSVP::GLabel_Object *>(label);

		if (genLabel->get().size != sizeof(uint32_t)) {
			G2RSVPTE_ERR("only 4-bytes-labels are accepted");
			return false;
		}

		// Check that this label is free and operationally ok
		label_state_t labelAdminStatus;
		opstate_t     labelOperStatus;

		if (!check_label(psb_->localDL(),           // Upstream DL
				 *(genLabel->get().label),  // Incoming label
				 labelAdminStatus,
				 labelOperStatus)) {
			G2RSVPTE_ERR("Label (0x%x) in LABEL SET does not exist "
				     "under DL %s",
				     *(genLabel->get().label),
				     addr_ntoa(psb_->localDL()));
			// set error params
			errTELink(psb()->interface()->localId());
			errLabel(*(genLabel->get().label));
			return false;
		}

		if (labelOperStatus != UP) {
			G2RSVPTE_ERR("Label (0x%x) in LABEL SET is "
				     "disabled/down (failure)",
				     *(genLabel->get().label));
			// set error params
			errTELink(psb()->interface()->localId());
			errLabel(*(genLabel->get().label));
			return false;
		}

		if (labelAdminStatus != LABEL_FREE) {
			G2RSVPTE_ERR("Label (%s:0x%x) in LABEL SET is "
				     "busy",
				     addr_ntoa(psb_->localDL()),
				     *(genLabel->get().label));
			// set error params
			errTELink(psb()->interface()->localId());
			errLabel(*(genLabel->get().label));
			return false;
		}

		psb_->rxLabel(*(genLabel->get().label));

		G2RSVPTE_DBG("the label (0x%x) in LABEL SET is 'enabled' and "
			     "free",
			     psb_->rxLabel());

		return true;
	}

	bool LSP::processUpstreamLabelObj(void)
	{
		if (!psb_ || !psb_->pathIn() ||
		    !psb_->pathIn()->upstream_label_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}

		RSVP::UpstreamLabel_Class * label;
		label = psb_->pathIn()->upstream_label_;

		// Check that there is one and only one inclusive subchannel
		if (!label ||
		    label->cType() != CTYPE_UPSTREAM_GENERALISED_LABEL) {
			G2RSVPTE_ERR("temporarily,only 1 inclusive upstream "
				     "generalized-label is allowed");
			return false;
		}

		RSVP::UpstreamLabel_Object * upLabel;
		upLabel = static_cast<RSVP::UpstreamLabel_Object *>(label);

		if (upLabel->get().size != sizeof(uint32_t)) {
			G2RSVPTE_ERR("only 4-bytes-labels are accepted");
			return false;
		}

		// Check that this label is free and operationally ok
		label_state_t labelAdminStatus;
		opstate_t     labelOperStatus;

		if (!check_label(psb_->localDL(),          // Upstream DL
				 *(upLabel->get().label),  // Incoming label
				 labelAdminStatus,
				 labelOperStatus)) {
			G2RSVPTE_ERR("Label (0x%x) in UPSTREAM LABEL does "
				     "not exist under DL %s",
				     *(upLabel->get().label),
				     addr_ntoa(psb_->localDL()));
			// set error params
			errTELink(psb()->interface()->localId());
		        errLabel(*(upLabel->get().label));
			return false;
		}

		if (labelOperStatus != UP) {
			G2RSVPTE_ERR("Label (0x%x) in UPSTREAM LABEL is "
				     "disabled/down (failure)",
				     *(upLabel->get().label));
			// set error params
			errTELink(psb()->interface()->localId());
			errLabel(*(upLabel->get().label));
			return false;
		}

		if (labelAdminStatus != LABEL_FREE) {
			G2RSVPTE_ERR("Label (%s:0x%x) in UPSTREAM LABEL is "
				     "already busy",
				     addr_ntoa(psb_->localDL()),
				     *(upLabel->get().label));
			// set error params
			errTELink(psb()->interface()->localId());
			errLabel(*(upLabel->get().label));
			return false;
		}

		psb_->txLabel(*(upLabel->get().label));

		G2RSVPTE_DBG("the label (0x%x) in UPSTREAM LABEL is 'enabled' "
			     "and free", psb_->txLabel());

		return true;
	}

	bool LSP::prepareLabelSetObj(void)
	{
		if (!psb_ || !psb_->pathOut() || !rsb_) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		if (!psb_->pathOut()->label_set_) {

			RSVP::LabelSet_Object * labelSet;
			labelSet = new RSVP::LabelSet_Object();

			if (!labelSet) {
				G2RSVPTE_ERR("Impossible to allocate Label "
					     "Set obj for PHOP!");
				return false;
			}

			psb_->pathOut()->label_set_ = labelSet;
		}

		if (psb_->pathOut()->label_set_->cType() != CTYPE_LABEL_SET) {
			G2RSVPTE_ERR("temporarily, only cType = %d is"
				     "allowed for Label Set obj",
				     CTYPE_LABEL_SET);
			return false;
		}
		RSVP::LabelSet_Object * labelSet;
		labelSet = static_cast<RSVP::LabelSet_Object *>
			(psb_->pathOut()->label_set_);

		// Use inclusive list
		labelSet->get().action     = 0;
		labelSet->get().label_type = CTYPE_GENERALISED_LABEL;

		if (labelSet->get().labels.size()) { // Usefull ???? Correct ???
			G2RSVPTE_ERR("Label subObj list should be empty, "
				     "but there are %d elems",
				     labelSet->get().labels.size());
			return false;
		}

		RSVP::GLabel_Object * genLabel;
		genLabel = new RSVP::GLabel_Object();
		if (!genLabel) {
			G2RSVPTE_ERR("Impossible to allocate memory for label "
				     "set subchannel");
			return false;
		}

		genLabel->get().size  = sizeof(uint32_t);
		genLabel->get().label = new uint32_t;
		if (!genLabel->get().label || !genLabel->get().size) {
			G2RSVPTE_ERR("Impossible to allocate memory for label "
				     "set subchannel label");
			delete genLabel;
			return false;
		}
		*genLabel->get().label = rsb_->txLabel();

		// Insert gen label into list
		labelSet->get().labels.push_back(genLabel);

		return true;
	}

	bool LSP::prepareUpstreamLabelObj(void)
	{
		if (!psb_ || !psb_->pathOut() || !rsb_) {
			G2RSVPTE_ERR("Invalid pointers!");
			return false;
		}

		if (!psb_->pathOut()->upstream_label_) {

			RSVP::UpstreamLabel_Object * upLabel;
			upLabel = new RSVP::UpstreamLabel_Object();

			if (!upLabel) {
				G2RSVPTE_ERR("Impossible to allocate Upstream "
					     "Label obj for PHOP!");
				return false;
			}

			psb_->pathOut()->upstream_label_ = upLabel;
		}

		if (psb_->pathOut()->upstream_label_->cType() !=
		    CTYPE_UPSTREAM_GENERALISED_LABEL) {
			G2RSVPTE_ERR("temporarily, only upstream "
				     "generalized-label is allowed");
			return false;
		}
		RSVP::UpstreamLabel_Object * upLabel;
		upLabel = static_cast<RSVP::UpstreamLabel_Object *>
			(psb_->pathOut()->upstream_label_);

		if (upLabel->get().label) {   // Usefull ???? Correct ???
			G2RSVPTE_ERR("Label subObj into upstream label obj "
				     "is not NULL");
			return false;
		}

		upLabel->get().size  = sizeof(uint32_t);
		upLabel->get().label = new uint32_t;
		if (!upLabel->get().label || !upLabel->get().size) {
			G2RSVPTE_ERR("Impossible to allocate memory for "
				     "upstream label");
			return false;
		}
		// downstream tx label = downstream rx label
		*upLabel->get().label = rsb_->rxLabel();

		return true;
	}

	bool LSP::processGenUniObj(void)
	{
		assert(parent());
		assert(parent()->parent());

		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->gen_uni_) {
			G2RSVPTE_ERR("Null PSB/PathIn/GenUNI pointer!!!");
			return false;
		}

		GeneralisedUni_Class * tmp;
		tmp = psb_->pathIn()->gen_uni_;

		if (tmp->cType() != CTYPE_GENERALIZED_UNI) {
			G2RSVPTE_ERR("Not supported Gen UNI object type!");
			return false;
		}

		GeneralisedUni_Object * genUni;
		genUni = static_cast<GeneralisedUni_Object *>(tmp);

		// Get SRC TNA
		GuniSubObj * subObjSrc;
		subObjSrc = genUni->getSrcTNA();
		if (subObjSrc) {
			g2mpls_addr_t srcAddr;
			if (!createSrcG2mplsAddr(*subObjSrc, srcAddr)) {
				G2RSVPTE_ERR("Impossible to retrieve I-TNA info");
				return false;
			}

			if (BITMASK_BITTEST(in_net_res_.mask_, tna)) {
				G2RSVPTE_ERR("Ingress TNA is already set(%s)...strange!",
					     addr_ntoa(in_net_res_.tna));
				return false;
			}
			in_net_res_.tna = srcAddr;
			BITMASK_BITSET(in_net_res_.mask_, tna);
		}

		// Get DST TNA
		GuniSubObj * subObjDst;
		subObjDst = genUni->getDestTNA();
		if (subObjDst) {
			g2mpls_addr_t dstAddr;
			if (!createDestG2mplsAddr(*subObjDst, dstAddr)) {
				G2RSVPTE_ERR("Impossible to retrieve E-TNA info");
				return false;
			}

			if (BITMASK_BITTEST(eg_net_res_.mask_, tna)) {
				G2RSVPTE_ERR("Egress TNA is already set(%s)...strange!",
					     addr_ntoa(eg_net_res_.tna));
				return false;
			}
			eg_net_res_.tna = dstAddr;
			BITMASK_BITSET(eg_net_res_.mask_, tna);

			// Check dl/label on egress node
			if (BITMASK_BITTEST(eg_net_res_.mask_, data_link) ||
			    BITMASK_BITTEST(eg_net_res_.mask_, label)) {
				G2RSVPTE_ERR("Egress DataLink/label are already "
					     "set (DL %s)...strange!",
					     addr_ntoa(eg_net_res_.data_link));
				return false;
			}

			SpcLabel_SubObj * spcLbl;
			spcLbl = static_cast<SpcLabel_SubObj *>(genUni->getSpcLabel());
			if (spcLbl) {

				g2mpls_addr_t dl;
				label_id_t    lbl;
				if (!createDlLabel(spcLbl, dl, lbl)) {
					G2RSVPTE_ERR("Impossible to retrieve egress "
						     "DataLink/Label info");
					return false;
				}

				eg_net_res_.data_link = dl;
				BITMASK_BITSET(eg_net_res_.mask_, data_link);

				eg_net_res_.label = lbl;
				BITMASK_BITSET(eg_net_res_.mask_, label);
			}
		}

		if (parent()->parent()->adjType() == INNI) {
			if (!subObjDst || !subObjSrc) {
				G2RSVPTE_ERR("Received Gen UNI object with problems");
				return false;
			}
			if (BITMASK_BITTEST(eg_net_res_.mask_, data_link) &&
			    BITMASK_BITTEST(eg_net_res_.mask_, label)) {

				//!!! for now only 32bit label are supported
				uint32_t lbl = eg_net_res_.label.value.label32.id;

				// Check dsTEL, dsDL and dsTxLabel on Egress Node
				// and set them as downstream DL and Label (tx)
				if (!checkSet_DsTEL_DsDL_DsTxLbl(eg_net_res_.tna,
								 eg_net_res_.data_link,
								 lbl)) {
					G2RSVPTE_ERR("Invalid egress parameters: ...");
					// set error params
					errTELink(eg_net_res_.tna);
					errLabel(lbl);
					return false;
				}

			} else {
				// Choose downstream DL and label (tx) for E-TNA
				if (!setDsDL_DsTxLbl_fromTELink(eg_net_res_.tna
								//egress_tna_
								)) {
					G2RSVPTE_ERR("could not find any suitable Data"
						     "Link/label under Egress TNA %s",
						     addr_ntoa(eg_net_res_.tna));
					// set error params
					errTELink(eg_net_res_.tna);
					return false;
				}
			}
		}

		return true;
	}

	bool LSP::processTimesObj(void)
	{
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->times_) {
			G2RSVPTE_ERR("Null PSB/PathIn/TIMES pointer!!!");
			return false;
		}

		Times_Class * tmp;
		tmp = psb_->pathIn()->times_;

		if (tmp->cType() != CTYPE_TIMES) {
			G2RSVPTE_ERR("Not supported TIMES object type!");
			return false;
		}

		Times_Object * times;
		times = static_cast<Times_Object *>(tmp);

		if (times->get().end_time <= times->get().start_time) {
			G2RSVPTE_ERR("Uncorrect TIMES object parameters!");
			return false;
		}

		//set times parameters in LSP instance
		times_.start_time = times->get().start_time;
		times_.end_time   = times->get().end_time;

		return true;
	}

	bool LSP::processPolicyData(void)
	{
		if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->policy_data_) {
			G2RSVPTE_ERR("Null PSB/PathIn/POLICYDATA pointer!!!");
			return false;
		}

		Policy_Class * tmp;
		tmp = psb_->pathIn()->policy_data_;

		if (tmp->cType() != CTYPE_POLICY) {
			G2RSVPTE_ERR("Not supported POLICY DATA object type!");
			return false;
		}

		Policy_Object * policy;
		policy = static_cast<Policy_Object *>(tmp);

		std::list<PolicyInfo *>::const_iterator it;
		for (it  = policy->get().policy_elems.begin();
		     it != policy->get().policy_elems.end();
		     it++) {
			if ((*it)->type != PTYPE_PHOSPHORUS_AUTHZ_TOKEN) {
				G2RSVPTE_ERR("P-type %d is not supported in PolicyInfo!", (*it)->type);
				return false;
			}
		}

		return true;
	}

	bool LSP::getCallId(call_ident_t & Id)
	{
		if (!psb_ || !psb_->pathIn() ||
		    !psb_->pathIn()->call_id_) {
			G2RSVPTE_ERR("Null pointer!!!");
			return false;
		}
		bit8_t                addrType;
		bit8_t                addrLen;
		char *                addr;
		RSVP::CallId_Class *  callObj;

		callObj = psb_->pathIn()->call_id_;

		switch (callObj->cType()) {
			case CTYPE_CALL_ID_OPERATOR_SPECIFIC: {
				CallIdOpSpec_Object * callId;
				callId = static_cast<CallIdOpSpec_Object *>(callObj);
				addrType = callId->get().addr_type;
				addrLen  = callId->get().src_lsr_addr_len;
				addr     = callId->get().src_lsr_addr;
				Id.type     = CALLID_OPSPEC;
				Id.local_id = callId->get().local_id;
				break;
			}
			case CTYPE_CALL_ID_GLOBALLY_UNIQUE: {
				CallIdGlobal_Object * callId;
				callId = static_cast<CallIdGlobal_Object *>(callObj);
				addrType = callId->get().addr_type;
				addrLen  = callId->get().src_lsr_addr_len;
				addr     = callId->get().src_lsr_addr;
				Id.type     = CALLID_GLOBUNIQ;
				Id.local_id = callId->get().local_id;
				break;
			}
			case CTYPE_CALL_ID_NULL: {
				if (getInstance()->adjType() == UNI) {
					CallIdNull_Object * callId;
					callId = static_cast<CallIdNull_Object *>(callObj);
					addrType = callId->get().addr_type;
					addrLen  = callId->get().src_lsr_addr_len;
					addr     = callId->get().src_lsr_addr;
					Id.type     = CALLID_NULL;
					Id.local_id = callId->get().local_id;
					break;
				} else {
					G2RSVPTE_ERR("Not supported CallId object type!");
					return false;
				}
			}
			default:
				G2RSVPTE_ERR("Not supported CallId object type!");
				return false;
				break;
		}

		switch (addrType) {
			case 0x01:
				memcpy(&Id.src_addr.value.ipv4.s_addr,
				       (void *) addr,
				       addrLen);
				break;
			case 0x02:
				memcpy(&Id.src_addr.value.ipv6.s6_addr32[0],
				       (void *) addr,
				       addrLen);
				break;
			case 0x03:
			case 0x04:
			case 0x7f:
				G2RSVPTE_ERR("Not supported CallId address type!");
				break;
			default:
				G2RSVPTE_ERR("Wrong CallId address type!");
				return false;
				break;
		}

		return true;
	}

	bool LSP::processCallIdObj(bool lookAtPath)
	{
		assert(getInstance());

		CallId_Class * tmp = 0;

		switch (lookAtPath) {
			case true:
				if (!psb_ || !psb_->pathIn() || !psb_->pathIn()->call_id_) {
					G2RSVPTE_ERR("Null PSB/PathIn/CallId pointer!!!");
					return false;
				}
				tmp = psb_->pathIn()->call_id_;
				break;
			case false:
				if (!rsb_ || !rsb_->resvIn() || !rsb_->resvIn()->call_id_) {
					G2RSVPTE_ERR("Null RSB/ResvIn/CallId pointer!!!");
					return false;
				}
				tmp = rsb_->resvIn()->call_id_;
		}

		if (tmp->cType() != CTYPE_CALL_ID_OPERATOR_SPECIFIC &&
		    tmp->cType() != CTYPE_CALL_ID_GLOBALLY_UNIQUE   &&
		    tmp->cType() != CTYPE_CALL_ID_NULL) {
			G2RSVPTE_ERR("Not supported CallId object type!");
			return false;
		}

		//copy old callId
		callId_old_ = callId_;

		bit8_t addrType;
		bit8_t addrLen;
		char * addr;
		switch (tmp->cType()) {
			case CTYPE_CALL_ID_OPERATOR_SPECIFIC: {
				CallIdOpSpec_Object * callId;
				callId = static_cast<CallIdOpSpec_Object *>(tmp);
				addrType = callId->get().addr_type;

				addrLen  = callId->get().src_lsr_addr_len;
				addr     = callId->get().src_lsr_addr;
				callId_.type     = CALLID_OPSPEC;
				callId_.local_id = callId->get().local_id;
				if (lookAtPath) {
					psb_->pathIn()->call_id_ = callId->clone();
				}
				break;
			}
			case CTYPE_CALL_ID_GLOBALLY_UNIQUE: {
				CallIdGlobal_Object * callId;
				callId = static_cast<CallIdGlobal_Object *>(tmp);
				addrType = callId->get().addr_type;
				addrLen  = callId->get().src_lsr_addr_len;
				addr     = callId->get().src_lsr_addr;
				callId_.type     = CALLID_GLOBUNIQ;
				callId_.local_id = callId->get().local_id;
				if (lookAtPath) {
					psb_->pathIn()->call_id_ = callId->clone();
				}
				break;
			}
			case CTYPE_CALL_ID_NULL: {
				if (getInstance()->adjType() == UNI) {
					CallIdNull_Object * callId;
					callId = static_cast<CallIdNull_Object *>(tmp);
					addrType = callId->get().addr_type;
					addrLen  = callId->get().src_lsr_addr_len;
					addr     = callId->get().src_lsr_addr;
					callId_.type     = CALLID_NULL;
					callId_.local_id = callId->get().local_id;
					if (lookAtPath) {
						psb_->pathIn()->call_id_ = callId->clone();
					}
					break;
				} else {
					G2RSVPTE_ERR("Not supported CallId object type!");
					return false;
				}
			}
			default:
				G2RSVPTE_ERR("Not supported CallId object type!");
				return false;
				break;
		}

		switch (addrType) {
			case 0x01:
				memcpy(&callId_.src_addr.value.ipv4.s_addr,
				       (void *) addr,
				       addrLen);
				break;
			case 0x02:
				memcpy(&callId_.src_addr.value.ipv6.s6_addr32[0],
				       (void *) addr,
				       addrLen);
				break;
			case 0x03:
			case 0x04:
			case 0x7f:
				G2RSVPTE_ERR("Not supported CallId address type!");
				break;
			default:
				G2RSVPTE_ERR("Wrong CallId address type!");
				return false;
				break;
		}

		return true;
	}

	std::ostringstream & operator << (std::ostringstream & os,
					  const LSP &          lsp)
	{
		os << "  LSP: "
		   << "\r\n";
		os << "    enabled          : " << (lsp.isEnabled() ? "yes" : "no")
		   << "\r\n";
		os << "    FSM state        : " << (lsp.fsmInst_->currentState())
		   << "\r\n";
		os << "    session name     : " << lsp.sessionName().c_str()
		   << "\r\n";
#ifdef CRANKBACK_ON
		os << "    crankback scope  : " << SHOW_CBACK_SCOPE(lsp.cbackScope())
		   << "\r\n";
#else
		os << "    crankback scope  : NONE"
		   << "\r\n";
#endif // CRANKBACK_ON
		os << "    context          : " << lsp.context()
		   << "\r\n";
		os << "    lsp id           : " << lsp.id()
		   << "\r\n";
		if (!is_addr_null(lsp.iRes().tna) && !is_addr_null(lsp.eRes().tna)) {
		        os << "    i-TNA            : " << addr_ntoa(lsp.iRes().tna)
		           << "\r\n";
		        os << "    e-TNA            : " << addr_ntoa(lsp.eRes().tna)
		           << "\r\n";
		}
		if (lsp.times().start_time && lsp.times().end_time) {
		        os << "    start UNIX time  : " << lsp.times().start_time
		           << "\r\n";
		        os << "    end   UNIX time  : " << lsp.times().end_time
		           << "\r\n";
		}
		os << "    src node id      : " << numb_htoa(lsp.nid())
		   << "\r\n";

		return os;
	}

	//
	// PRIVATE METHODS
	//
	Interface * LSP::getInterfaceFromTELink(g2mpls_addr_t addr,
						bool          checkRemote)
	{
	        G2RSVPTE * proto;
		proto = getInstance();

		if (!proto) {
			G2RSVPTE_ERR("Invalid protocol instance for lsp id %d",
				     id_);
			return 0;
		}

		return proto->findInterface(addr, checkRemote);
	}

	uint32_t LSP::getNbrId(void)
	{
		if (!rsb_ || !rsb_->interface()) {
			G2RSVPTE_ERR("Invalid pointers!");
			return 0;
		}

		uint32_t nodeId = 0;
		nodeId = rsb_->interface()->neighbourId();

		G2RSVPTE_DBG("Found remote router ID %s", numb_htoa(nodeId));

		return nodeId;
	}

	bool LSP::eroComplete(void)
	{
		// TODO!!!!!
		return false;
	}
}
