//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//


#include <iostream>
#include <stdio.h>
#include <string>
#include <map>
#include <list>
#include <stdlib.h>
#include <assert.h>
#include "g2rsvpte_lsp_gen.h"

#include <algorithm>

#include "g2rsvpte_lsp.h"


/*****************************************************************************/
/*                             Finite State Machine                          */
/*****************************************************************************/
namespace fsm {
	
	
	/*************************************/
	/*    Finite State Machine - Core    */
	/*************************************/
	namespace base_G2RSVPTE_LSP_FSM {

		//
		// State/s
		//

		//
		// BaseFSM
		//

		//
		// GenericFSM
		//
		bool GenericFSM::startModify(void)
		{
			if (changeInProgress_) {
				err("Changes already in progress ... call "
				    "endModify before!!!");
				return false;
			}
			dbg("### Start FSM changes");
			changeInProgress_ = true;
			return true;
		}

		bool GenericFSM::endModify(void)
		{
			if (!check()) {
				err("End FSM changes ... FAILED ###");
				return false;
			}
			dbg("End FSM changes ###");
			changeInProgress_ = false;
			return true;
		}

		bool GenericFSM::addState(std::string state)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (state.empty()) {
				err("Invalid state: string is empty");
				return false;
			}

			std::map<std::string, state_data_t>::iterator it;
			it = states_.find(state);
			if (it != states_.end()) {
				err("State " + state + " already added");
				return false;
			}

			state_data_t data;
			data.pre       = 0;
			data.post      = 0;
			data.in        = 0;
			states_[state] = data;

			dbg("Added state: " + state);
			return true;
		}

		bool GenericFSM::remState(std::string state)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (state.empty()) {
				err("Invalid state: string is empty");
				return false;
			}

			if (!states_.erase(state)) {
				err("State " + state + " not present");
				return false;
			}

			// Remove all the rows == state
			transitions_.removeRow(state);

			// Remove all the data (to-state) == state
			transitions_.remove(state);

			dbg("Removed state: " + state);

			if (startState_ == state) {
				wrn("State '" + state + "' was the start state"
				    " ... change it!!!");
				startState_ = std::string();
			}

			return true;
		}

		bool GenericFSM::addEvent(std::string event)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (event.empty()) {
				err("Invalid event: string is empty");
				return false;
			}

			std::list<std::string>::iterator it;
			it = find(events_.begin(),events_.end(),event);
			if (it != events_.end()) {
				err("Event " + event + " already added");
				return false;
			}
			events_.push_back(event);
			dbg("Added event: " + event);
			return true;
		}

		bool GenericFSM::remEvent(std::string event)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (event.empty()) {
				err("Invalid event: string is empty");
				return false;
			}

			events_.remove(event);

			// Remove all the columns == event
			transitions_.removeCol(event);

			dbg("Removed event: " + event);
			return true;
		}

		bool GenericFSM::addTransition(std::string from,
					       std::string to,
					       std::string event)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (from.empty() || to.empty() ||
			    event.empty()) {
				err("Invalid transition: some data are empty");
				return false;
			}

			Matrix<std::string,
				std::string,
				std::string>::dataIter it;
			it = transitions_.find(from, event);
			if (it != 0) {
				err("Transition from state '" + from + "' with"
				    " event '" + event + "' already added");
				return false;
			}

			transitions_[from][event] = to;
			dbg("Added transition: from state '" + from +
			    "' to state '" + to + "' with event '" +
			     event + "'");
			return true;
		}

		bool GenericFSM::remTransition(std::string from,
					       std::string to,
					       std::string event)
		{
			if (!changeInProgress_) {
				err("Call startModify before any changes!!!");
				return false;
			}

			if (from.empty() || to.empty() || event.empty()) {
				err("Invalid transition: some data are empty");
				return false;
			}

			Matrix<std::string,
				std::string,
				std::string>::dataIter it;
			it = transitions_.find(from, event);
			if (it != 0 && *it != to) {
				err("To-state for the transition from state '"
				    + from + "' event '" + event + "' is not '"
				    + to + "' but '" + *it + "'");
				return false;
			}

			if (!transitions_.remove(it)) {
				err("Transition from state '" + from +
				    "' to state '"+ to + "' with event '" +
				    event + "' is not present");
				return false;
			}

			dbg("Removed transition: from state '" + from +
			    "' to state '"+ to + "' with event '" + event +
			    "'");
			return true;
		}

		bool GenericFSM::setStartState(std::string state)
		{
			if (state.empty()) {
				err("Impossible to set start state to "
				    + state);
				return false;
			}

			dbg(std::string("Setting start state to ") + state);
			startState_ = state;
			return true;
		}

		bool GenericFSM::check(void)
		{
			dbg("Checking FSM integrity ...");

			if (transitions_.empty() && states_.empty()) {
				dbg("The FSM is completely empty!!!");
				return true;
			}

			if (states_.find(startState_) ==
			    states_.end()) {
				err("The start state is not a valid state");
				return false;
			}

			/*
			 *  Check if any transition state/event belong
			 *  to FSM states and events
			 */
			Matrix<std::string,
				std::string,
				std::string>::rowIter rowIt;
			for (rowIt = transitions_.begin();
			     rowIt != transitions_.end();
			     rowIt = transitions_.next(rowIt)) {

				// Check from-state
				if (states_.find(*rowIt) == states_.end()) {
					err("From-state " + *rowIt +
					    " is not a valid state");
					return false;
				}
				dbg("From-state '" + *rowIt +
				    "' is a valid state");

				Matrix<std::string,
					std::string,
					std::string>::colIter colIt;
				for (colIt = transitions_.begin(rowIt);
				     colIt != transitions_.end(rowIt);
				     colIt = transitions_.next(rowIt, colIt)){

					// Check event
					if (find(events_.begin(),
						 events_.end(),
						 *colIt) == events_.end()) {
						err("Event " + *colIt +
						    " is not a valid event");
						return false;
					}
					dbg("Event '" + *colIt +
					    "' is a valid event");

					// Check to-state
					std::string tmp;
					tmp = transitions_[*rowIt][*colIt];
					if (states_.find(tmp) ==
					    states_.end()) {
						err("To-state " + tmp +
						    " is not a valid state");
						return false;
					}
					dbg("To-state '" + tmp +
					    "' is a valid state");
				}
			}

			//TODO Graph check (partitioned/discontinued..)

			return true;
		}


		//
		// Fsm
		//
		Fsm::Fsm(traceLevel_t level)
			throw(std::string) :
			GenericFSM(level)
		{
			if (!startModify()) {
				throw(std::string("Impossible to init FSM"));
			}

			if (!setStartState("Down")) {
				throw(std::string("Impossible to set start "
						  "state for th FSM"));
			}
			currentState_ = G2RSVPTE_LSP_FSM_Down;
			prevState_    = currentState_;

			
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvResv] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvResvTimer] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResv",
					   "evRecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvTimer] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "SendConfirm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_SendConfirm] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Installed",
					   "evSendConfirm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evSendConfirm] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvConfirm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvConfirm] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Installed",
					   "evRecvConfirmOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvConfirmOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "XConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_XConnPreempt] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evXConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evXConnPreempt] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvConfirmKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvConfirmKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evRecvPathAlarm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathAlarm] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvResvErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvResvErr] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvErrKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvErrOk] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "RecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_RecvPathTimer] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResv",
					   "evRecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathTimer] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResvConf",
					   "WaitResvConf",
					   "evRecvResvOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvOk] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResvConf",
					   "virt_WaitResvConf",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResvConf][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_WaitResvConf; 
			if (!addTransition("virt_WaitResvConf",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_WaitResvConf",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResvConf][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPathTimeout] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResv] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "XConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_XConnCompleted] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitResvConf",
					   "evXConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evXConnCompleted] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResvTimer] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvTimer] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "XConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_XConnErr] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "PathReceived",
					   "evXConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evXConnErr] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResvTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResvTimeout] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvResvTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvResvVeryKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvVeryKo] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvResvErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvResvErr] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvErrKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvErrOk] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "RecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_RecvPathTimer] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathTimer] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitEqptConf",
					   "WaitEqptConf",
					   "evRecvResvOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvOk] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("WaitEqptConf",
					   "virt_WaitEqptConf",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitEqptConf][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_WaitEqptConf; 
			if (!addTransition("virt_WaitEqptConf",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_WaitEqptConf",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "XConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_XConnErr] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evXConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evXConnErr] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvConfirm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvConfirm] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvConfirmOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvConfirmOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "XConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_XConnPreempt] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evXConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evXConnPreempt] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "XConnDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_XConnDown] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evXConnDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evXConnDown] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathAlarm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathAlarm] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvConfirmKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvConfirmKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvPathTimer] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathTimer] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvResv] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvNotify")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvNotify] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvNotifyOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvNotifyOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Down",
					   "evRecvNotifyDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvNotifyDown] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "XConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_XConnCompleted] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evXConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evXConnCompleted] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvActivateErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvActivateErr] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvResvTimer] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvTimer] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvActivate")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvActivate] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvActivateOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvActivateOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_Installed",
					   "Down",
					   "evRecvResvVeryKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvVeryKo] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvNotifyKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvNotifyKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "RecvResvErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_RecvResvErr] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvErrKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvActivateKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvActivateKo] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("virt_Installed",
					   "Installed",
					   "evRecvResvErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evRecvResvErrOk] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Installed",
					   "virt_Installed",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Installed][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_Installed; 
			if (!addTransition("virt_Installed",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Installed][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("Down",
					   "virt_Down",
					   "LspRebuild")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Down][G2RSVPTE_LSP_FSM_LspRebuild] = G2RSVPTE_LSP_FSM_virt_Down; 
			if (!addTransition("virt_Down",
					   "Rebuild",
					   "evLspRebuild")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Down][G2RSVPTE_LSP_FSM_evLspRebuild] = G2RSVPTE_LSP_FSM_Rebuild; 
			if (!addTransition("Down",
					   "virt_Down",
					   "SendPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Down][G2RSVPTE_LSP_FSM_SendPath] = G2RSVPTE_LSP_FSM_virt_Down; 
			if (!addTransition("virt_Down",
					   "PathReceived",
					   "evSendPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Down][G2RSVPTE_LSP_FSM_evSendPath] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("Down",
					   "virt_Down",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Down][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_Down; 
			if (!addTransition("virt_Down",
					   "PathReceived",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Down][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("virt_Down",
					   "Down",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Down][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPathTimeout] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evRecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvResv] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvResvKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvResvKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "XConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_XConnCompleted] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "WaitResv",
					   "evXConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evXConnCompleted] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "XConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_XConnErr] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evXConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evXConnErr] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathErrCrankback")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathErrCrankback] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "RecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_RecvPathTimer] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "PathReceived",
					   "evRecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathTimer] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_PathReceived",
					   "WaitEqptConf",
					   "evRecvResvOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvResvOk] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_PathReceived",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("PathReceived",
					   "virt_PathReceived",
					   "SendResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_PathReceived][G2RSVPTE_LSP_FSM_SendResv] = G2RSVPTE_LSP_FSM_virt_PathReceived; 
			if (!addTransition("virt_PathReceived",
					   "WaitEqptConf",
					   "evSendResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_PathReceived][G2RSVPTE_LSP_FSM_evSendResv] = G2RSVPTE_LSP_FSM_WaitEqptConf; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvPathTimeout] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evRecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvResvTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvResvTimeout] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evRecvResvTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvPathAlarm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathAlarm] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "RecvResvErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_RecvResvErr] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvResvErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvErrKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvResvErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvErrOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_TearDown",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("TearDown",
					   "virt_TearDown",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_TearDown][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_TearDown; 
			if (!addTransition("virt_TearDown",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_TearDown][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "XConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_XConnCompleted] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evXConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evXConnCompleted] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "XConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_XConnErr] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evXConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evXConnErr] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "RecvResvErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_RecvResvErr] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvResvErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvErrKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvResvErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvErrOk] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("RebuildTearDown",
					   "virt_RebuildTearDown",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_RebuildTearDown][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_RebuildTearDown; 
			if (!addTransition("virt_RebuildTearDown",
					   "RebuildTearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPathTimeout] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvPathTimeout")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathTimeout] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "TearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvResv] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvResvKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPath")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPath] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathOk] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "SendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_SendPathTear] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evSendPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evSendPathTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "TearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "XConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_XConnPreempt] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evXConnPreempt")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evXConnPreempt] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvResvVeryKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvVeryKo] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvResvTear] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvResvTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvTearKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "PathReceived",
					   "evRecvPathErrCrankback")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathErrCrankback] = G2RSVPTE_LSP_FSM_PathReceived; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvResvTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvPathAlarm")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathAlarm] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "SendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_SendResvTear] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evSendResvTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evSendResvTear] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "RecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_RecvPathTimer] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResv",
					   "evRecvPathTimer")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathTimer] = G2RSVPTE_LSP_FSM_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("virt_WaitResv",
					   "WaitResvConf",
					   "evRecvResvOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvOk] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "TearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("virt_WaitResv",
					   "TearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_TearDown; 
			if (!addTransition("WaitResv",
					   "virt_WaitResv",
					   "SendResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_WaitResv][G2RSVPTE_LSP_FSM_SendResv] = G2RSVPTE_LSP_FSM_virt_WaitResv; 
			if (!addTransition("virt_WaitResv",
					   "WaitResvConf",
					   "evSendResv")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_WaitResv][G2RSVPTE_LSP_FSM_evSendResv] = G2RSVPTE_LSP_FSM_WaitResvConf; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "RecvPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_RecvPathDown] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Rebuild",
					   "evRecvPathDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathDownKo] = G2RSVPTE_LSP_FSM_Rebuild; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "SendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_SendPathDown] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "RebuildTearDown",
					   "evSendPathDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evSendPathDown] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "XConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_XConnCompleted] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Installed",
					   "evXConnCompleted")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evXConnCompleted] = G2RSVPTE_LSP_FSM_Installed; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "RecvPathErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_RecvPathErr] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Rebuild",
					   "evRecvPathErrKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathErrKo] = G2RSVPTE_LSP_FSM_Rebuild; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "XConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_XConnErr] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Down",
					   "evXConnErr")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evXConnErr] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "RecvPathTear")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_RecvPathTear] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Down",
					   "evRecvPathTearOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathTearOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "RecvResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_RecvResvDown] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Rebuild",
					   "evRecvResvDownKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvResvDownKo] = G2RSVPTE_LSP_FSM_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Rebuild",
					   "evRecvPathTearKo")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathTearKo] = G2RSVPTE_LSP_FSM_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "Down",
					   "evRecvPathErrOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathErrOk] = G2RSVPTE_LSP_FSM_Down; 
			if (!addTransition("Rebuild",
					   "virt_Rebuild",
					   "SendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_Rebuild][G2RSVPTE_LSP_FSM_SendResvDown] = G2RSVPTE_LSP_FSM_virt_Rebuild; 
			if (!addTransition("virt_Rebuild",
					   "RebuildTearDown",
					   "evSendResvDown")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evSendResvDown] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_Rebuild",
					   "RebuildTearDown",
					   "evRecvPathDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvPathDownOk] = G2RSVPTE_LSP_FSM_RebuildTearDown; 
			if (!addTransition("virt_Rebuild",
					   "RebuildTearDown",
					   "evRecvResvDownOk")) {
				throw(std::string("Impossible to add "
						  "transition .... "));
			}

			nextState_[G2RSVPTE_LSP_FSM_virt_Rebuild][G2RSVPTE_LSP_FSM_evRecvResvDownOk] = G2RSVPTE_LSP_FSM_RebuildTearDown; 

			
			if (!addState("WaitResvConf")) {
				throw(std::string("Impossible to add state "
						  "WaitResvConf"));
			}
			states_[G2RSVPTE_LSP_FSM_WaitResvConf] = new ::WaitResvConf_i; 
			if (!addState("WaitEqptConf")) {
				throw(std::string("Impossible to add state "
						  "WaitEqptConf"));
			}
			states_[G2RSVPTE_LSP_FSM_WaitEqptConf] = new ::WaitEqptConf_i; 
			if (!addState("Installed")) {
				throw(std::string("Impossible to add state "
						  "Installed"));
			}
			states_[G2RSVPTE_LSP_FSM_Installed] = new ::Installed_i; 
			if (!addState("Down")) {
				throw(std::string("Impossible to add state "
						  "Down"));
			}
			states_[G2RSVPTE_LSP_FSM_Down] = new ::Down_i; 
			if (!addState("PathReceived")) {
				throw(std::string("Impossible to add state "
						  "PathReceived"));
			}
			states_[G2RSVPTE_LSP_FSM_PathReceived] = new ::PathReceived_i; 
			if (!addState("TearDown")) {
				throw(std::string("Impossible to add state "
						  "TearDown"));
			}
			states_[G2RSVPTE_LSP_FSM_TearDown] = new ::TearDown_i; 
			if (!addState("RebuildTearDown")) {
				throw(std::string("Impossible to add state "
						  "RebuildTearDown"));
			}
			states_[G2RSVPTE_LSP_FSM_RebuildTearDown] = new ::RebuildTearDown_i; 
			if (!addState("WaitResv")) {
				throw(std::string("Impossible to add state "
						  "WaitResv"));
			}
			states_[G2RSVPTE_LSP_FSM_WaitResv] = new ::WaitResv_i; 
			if (!addState("Rebuild")) {
				throw(std::string("Impossible to add state "
						  "Rebuild"));
			}
			states_[G2RSVPTE_LSP_FSM_Rebuild] = new ::Rebuild_i; 
			
			if (!addState("virt_WaitResvConf")) {
				throw(std::string("Impossible to add state "
						  "virt_WaitResvConf"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_WaitResvConf] = new ::virt_WaitResvConf_i; 
			if (!addState("virt_WaitEqptConf")) {
				throw(std::string("Impossible to add state "
						  "virt_WaitEqptConf"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_WaitEqptConf] = new ::virt_WaitEqptConf_i; 
			if (!addState("virt_Installed")) {
				throw(std::string("Impossible to add state "
						  "virt_Installed"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_Installed] = new ::virt_Installed_i; 
			if (!addState("virt_Down")) {
				throw(std::string("Impossible to add state "
						  "virt_Down"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_Down] = new ::virt_Down_i; 
			if (!addState("virt_PathReceived")) {
				throw(std::string("Impossible to add state "
						  "virt_PathReceived"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_PathReceived] = new ::virt_PathReceived_i; 
			if (!addState("virt_TearDown")) {
				throw(std::string("Impossible to add state "
						  "virt_TearDown"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_TearDown] = new ::virt_TearDown_i; 
			if (!addState("virt_RebuildTearDown")) {
				throw(std::string("Impossible to add state "
						  "virt_RebuildTearDown"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_RebuildTearDown] = new ::virt_RebuildTearDown_i; 
			if (!addState("virt_WaitResv")) {
				throw(std::string("Impossible to add state "
						  "virt_WaitResv"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_WaitResv] = new ::virt_WaitResv_i; 
			if (!addState("virt_Rebuild")) {
				throw(std::string("Impossible to add state "
						  "virt_Rebuild"));
			}
			states_[G2RSVPTE_LSP_FSM_virt_Rebuild] = new ::virt_Rebuild_i; 

			//Only to check FSM integrity
			
			if (!addEvent("evRecvResvTearOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvTearOk"));
			}
			if (!addEvent("evRecvResvTearKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvTearKo"));
			}
			if (!addEvent("evRecvPathTimer")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathTimer"));
			}
			if (!addEvent("evXConnPreempt")) {
				throw(std::string("Impossible to add event "
						  "evXConnPreempt"));
			}
			if (!addEvent("evXConnErr")) {
				throw(std::string("Impossible to add event "
						  "evXConnErr"));
			}
			if (!addEvent("evRecvPathDownOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathDownOk"));
			}
			if (!addEvent("evRecvPathDownKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathDownKo"));
			}
			if (!addEvent("evXConnCompleted")) {
				throw(std::string("Impossible to add event "
						  "evXConnCompleted"));
			}
			if (!addEvent("evRecvResvOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvOk"));
			}
			if (!addEvent("evRecvResvKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvKo"));
			}
			if (!addEvent("evRecvResvVeryKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvVeryKo"));
			}
			if (!addEvent("evRecvPathTimeout")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathTimeout"));
			}
			if (!addEvent("evRecvResvErrOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvErrOk"));
			}
			if (!addEvent("evRecvResvErrKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvErrKo"));
			}
			if (!addEvent("evRecvResvDownOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvDownOk"));
			}
			if (!addEvent("evRecvResvDownKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvDownKo"));
			}
			if (!addEvent("evRecvActivateOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvActivateOk"));
			}
			if (!addEvent("evRecvActivateKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvActivateKo"));
			}
			if (!addEvent("evSendResvDown")) {
				throw(std::string("Impossible to add event "
						  "evSendResvDown"));
			}
			if (!addEvent("evSendPathDown")) {
				throw(std::string("Impossible to add event "
						  "evSendPathDown"));
			}
			if (!addEvent("evSendResv")) {
				throw(std::string("Impossible to add event "
						  "evSendResv"));
			}
			if (!addEvent("evRecvConfirmOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvConfirmOk"));
			}
			if (!addEvent("evRecvConfirmKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvConfirmKo"));
			}
			if (!addEvent("evRecvPathTearOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathTearOk"));
			}
			if (!addEvent("evRecvPathTearKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathTearKo"));
			}
			if (!addEvent("evSendConfirm")) {
				throw(std::string("Impossible to add event "
						  "evSendConfirm"));
			}
			if (!addEvent("evSendPath")) {
				throw(std::string("Impossible to add event "
						  "evSendPath"));
			}
			if (!addEvent("evLspRebuild")) {
				throw(std::string("Impossible to add event "
						  "evLspRebuild"));
			}
			if (!addEvent("evSendPathTear")) {
				throw(std::string("Impossible to add event "
						  "evSendPathTear"));
			}
			if (!addEvent("evRecvResvTimeout")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvTimeout"));
			}
			if (!addEvent("evSendResvTear")) {
				throw(std::string("Impossible to add event "
						  "evSendResvTear"));
			}
			if (!addEvent("evXConnDown")) {
				throw(std::string("Impossible to add event "
						  "evXConnDown"));
			}
			if (!addEvent("evRecvPathOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathOk"));
			}
			if (!addEvent("evRecvPathKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathKo"));
			}
			if (!addEvent("evRecvNotifyOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvNotifyOk"));
			}
			if (!addEvent("evRecvNotifyDown")) {
				throw(std::string("Impossible to add event "
						  "evRecvNotifyDown"));
			}
			if (!addEvent("evRecvNotifyKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvNotifyKo"));
			}
			if (!addEvent("evRecvPathErrOk")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathErrOk"));
			}
			if (!addEvent("evRecvPathAlarm")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathAlarm"));
			}
			if (!addEvent("evRecvPathErrCrankback")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathErrCrankback"));
			}
			if (!addEvent("evRecvActivateErr")) {
				throw(std::string("Impossible to add event "
						  "evRecvActivateErr"));
			}
			if (!addEvent("evRecvPathErrKo")) {
				throw(std::string("Impossible to add event "
						  "evRecvPathErrKo"));
			}
			if (!addEvent("evRecvResvTimer")) {
				throw(std::string("Impossible to add event "
						  "evRecvResvTimer"));
			}
			
			if (!addEvent("RecvResvTear")) {
				throw(std::string("Impossible to add event "
						  "RecvResvTear"));
			}
			if (!addEvent("RecvPathTimer")) {
				throw(std::string("Impossible to add event "
						  "RecvPathTimer"));
			}
			if (!addEvent("XConnPreempt")) {
				throw(std::string("Impossible to add event "
						  "XConnPreempt"));
			}
			if (!addEvent("XConnErr")) {
				throw(std::string("Impossible to add event "
						  "XConnErr"));
			}
			if (!addEvent("RecvPathDown")) {
				throw(std::string("Impossible to add event "
						  "RecvPathDown"));
			}
			if (!addEvent("XConnCompleted")) {
				throw(std::string("Impossible to add event "
						  "XConnCompleted"));
			}
			if (!addEvent("RecvResv")) {
				throw(std::string("Impossible to add event "
						  "RecvResv"));
			}
			if (!addEvent("RecvPathTimeout")) {
				throw(std::string("Impossible to add event "
						  "RecvPathTimeout"));
			}
			if (!addEvent("RecvResvErr")) {
				throw(std::string("Impossible to add event "
						  "RecvResvErr"));
			}
			if (!addEvent("RecvResvDown")) {
				throw(std::string("Impossible to add event "
						  "RecvResvDown"));
			}
			if (!addEvent("RecvActivate")) {
				throw(std::string("Impossible to add event "
						  "RecvActivate"));
			}
			if (!addEvent("SendResvDown")) {
				throw(std::string("Impossible to add event "
						  "SendResvDown"));
			}
			if (!addEvent("SendPathDown")) {
				throw(std::string("Impossible to add event "
						  "SendPathDown"));
			}
			if (!addEvent("SendResv")) {
				throw(std::string("Impossible to add event "
						  "SendResv"));
			}
			if (!addEvent("RecvConfirm")) {
				throw(std::string("Impossible to add event "
						  "RecvConfirm"));
			}
			if (!addEvent("RecvPathTear")) {
				throw(std::string("Impossible to add event "
						  "RecvPathTear"));
			}
			if (!addEvent("SendConfirm")) {
				throw(std::string("Impossible to add event "
						  "SendConfirm"));
			}
			if (!addEvent("SendPath")) {
				throw(std::string("Impossible to add event "
						  "SendPath"));
			}
			if (!addEvent("LspRebuild")) {
				throw(std::string("Impossible to add event "
						  "LspRebuild"));
			}
			if (!addEvent("SendPathTear")) {
				throw(std::string("Impossible to add event "
						  "SendPathTear"));
			}
			if (!addEvent("RecvResvTimeout")) {
				throw(std::string("Impossible to add event "
						  "RecvResvTimeout"));
			}
			if (!addEvent("SendResvTear")) {
				throw(std::string("Impossible to add event "
						  "SendResvTear"));
			}
			if (!addEvent("XConnDown")) {
				throw(std::string("Impossible to add event "
						  "XConnDown"));
			}
			if (!addEvent("RecvPath")) {
				throw(std::string("Impossible to add event "
						  "RecvPath"));
			}
			if (!addEvent("RecvNotify")) {
				throw(std::string("Impossible to add event "
						  "RecvNotify"));
			}
			if (!addEvent("RecvPathErr")) {
				throw(std::string("Impossible to add event "
						  "RecvPathErr"));
			}
			if (!addEvent("RecvResvTimer")) {
				throw(std::string("Impossible to add event "
						  "RecvResvTimer"));
			}

			if (!endModify()) {
				throw(std::string("Impossible to fini FSM"));
			}
		}

		Fsm::~Fsm(void)
		{
			startModify(); //useless ... just for check
			std::map<states_t, State *>::iterator it;
			for (it = states_.begin(); it != states_.end(); it++) {
				State * tmp;
				tmp = (*it).second;
				if (tmp) {
					// check and print error if needed
					remState(tmp->name());
					dbg(std::string("Deleting state: ") +
					    tmp->name());
					delete tmp;
				}
			}
			endModify(); //useless ... just for check
		}

		std::ostream& operator<<(std::ostream& s, const Fsm& f)
		{
			s << std::endl << "###### FSM [" << f.name_
			  << "] DUMP ### START ######" << std::endl;
			s << f.nextState_;
			s << "###### FSM [" << f.name_
			  << "] DUMP ###  END  ######" << std::endl
			  << std::endl;
			return s;
		}

		
		bool Fsm::evRecvResvTearOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvTearOk");

			if (!tmp->evRecvResvTearOk(context)) {
				err("Error in event 'evRecvResvTearOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvTearOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvTearOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvTearOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvTearOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvTearOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvTearOk_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvTearOk_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvTearOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvTearKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvTearKo");

			if (!tmp->evRecvResvTearKo(context)) {
				err("Error in event 'evRecvResvTearKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvTearKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvTearKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvTearKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvTearKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvTearKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvTearKo_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvTearKo_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvTearKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathTimer(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathTimer");

			if (!tmp->evRecvPathTimer(context)) {
				err("Error in event 'evRecvPathTimer' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathTimer];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathTimer_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathTimer_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathTimer_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathTimer_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathTimer_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathTimer_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evXConnPreempt(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evXConnPreempt");

			if (!tmp->evXConnPreempt(context)) {
				err("Error in event 'evXConnPreempt' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evXConnPreempt];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evXConnPreempt_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evXConnPreempt_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evXConnPreempt_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evXConnPreempt_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evXConnErr(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evXConnErr");

			if (!tmp->evXConnErr(context)) {
				err("Error in event 'evXConnErr' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evXConnErr];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evXConnErr_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evXConnErr_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evXConnErr_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evXConnErr_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evXConnErr_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evXConnErr_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathDownOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathDownOk");

			if (!tmp->evRecvPathDownOk(context)) {
				err("Error in event 'evRecvPathDownOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathDownOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathDownOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathDownOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathDownOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathDownOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathDownOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathDownOk_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathDownOk_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathDownOk_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathDownOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathDownKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathDownKo");

			if (!tmp->evRecvPathDownKo(context)) {
				err("Error in event 'evRecvPathDownKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathDownKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathDownKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathDownKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathDownKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathDownKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathDownKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathDownKo_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathDownKo_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathDownKo_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathDownKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evXConnCompleted(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evXConnCompleted");

			if (!tmp->evXConnCompleted(context)) {
				err("Error in event 'evXConnCompleted' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evXConnCompleted];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evXConnCompleted_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evXConnCompleted_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evXConnCompleted_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evXConnCompleted_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evXConnCompleted_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evXConnCompleted_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvOk");

			if (!tmp->evRecvResvOk(context)) {
				err("Error in event 'evRecvResvOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvResvOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvOk_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvKo");

			if (!tmp->evRecvResvKo(context)) {
				err("Error in event 'evRecvResvKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvResvKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvKo_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvVeryKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvVeryKo");

			if (!tmp->evRecvResvVeryKo(context)) {
				err("Error in event 'evRecvResvVeryKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvVeryKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvVeryKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvVeryKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvVeryKo_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvVeryKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathTimeout(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathTimeout");

			if (!tmp->evRecvPathTimeout(context)) {
				err("Error in event 'evRecvPathTimeout' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathTimeout];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathTimeout_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathTimeout_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathTimeout_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathTimeout_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathTimeout_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvErrOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvErrOk");

			if (!tmp->evRecvResvErrOk(context)) {
				err("Error in event 'evRecvResvErrOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvErrOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvErrOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvErrOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvErrOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvErrOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvErrOk_from_virt_RebuildTearDown(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvErrOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvErrKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvErrKo");

			if (!tmp->evRecvResvErrKo(context)) {
				err("Error in event 'evRecvResvErrKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvErrKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvErrKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvErrKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvErrKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvErrKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvErrKo_from_virt_RebuildTearDown(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvErrKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvDownOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvDownOk");

			if (!tmp->evRecvResvDownOk(context)) {
				err("Error in event 'evRecvResvDownOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvDownOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvDownOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvDownOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvDownOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvResvDownOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvDownOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvDownOk_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvDownOk_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvResvDownOk_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvDownOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvDownKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvDownKo");

			if (!tmp->evRecvResvDownKo(context)) {
				err("Error in event 'evRecvResvDownKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvDownKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvDownKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvDownKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvDownKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvResvDownKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvDownKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvResvDownKo_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvResvDownKo_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvResvDownKo_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvDownKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvActivateOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvActivateOk");

			if (!tmp->evRecvActivateOk(context)) {
				err("Error in event 'evRecvActivateOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvActivateOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvActivateOk_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvActivateOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvActivateKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvActivateKo");

			if (!tmp->evRecvActivateKo(context)) {
				err("Error in event 'evRecvActivateKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvActivateKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvActivateKo_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvActivateKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendResvDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendResvDown");

			if (!tmp->evSendResvDown(context)) {
				err("Error in event 'evSendResvDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendResvDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evSendResvDown_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evSendResvDown_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evSendResvDown_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evSendResvDown_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evSendResvDown_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evSendResvDown_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evSendResvDown_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evSendResvDown_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendResvDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendPathDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendPathDown");

			if (!tmp->evSendPathDown(context)) {
				err("Error in event 'evSendPathDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendPathDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evSendPathDown_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evSendPathDown_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evSendPathDown_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evSendPathDown_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evSendPathDown_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evSendPathDown_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evSendPathDown_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evSendPathDown_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendPathDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendResv(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendResv");

			if (!tmp->evSendResv(context)) {
				err("Error in event 'evSendResv' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendResv];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evSendResv_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evSendResv_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendResv_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvConfirmOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvConfirmOk");

			if (!tmp->evRecvConfirmOk(context)) {
				err("Error in event 'evRecvConfirmOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvConfirmOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvConfirmOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvConfirmOk_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvConfirmOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvConfirmKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvConfirmKo");

			if (!tmp->evRecvConfirmKo(context)) {
				err("Error in event 'evRecvConfirmKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvConfirmKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvConfirmKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvConfirmKo_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvConfirmKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathTearOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathTearOk");

			if (!tmp->evRecvPathTearOk(context)) {
				err("Error in event 'evRecvPathTearOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathTearOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathTearOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathTearOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathTearOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathTearOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathTearOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathTearOk_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathTearOk_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathTearOk_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathTearOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathTearKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathTearKo");

			if (!tmp->evRecvPathTearKo(context)) {
				err("Error in event 'evRecvPathTearKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathTearKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathTearKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathTearKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathTearKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathTearKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathTearKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathTearKo_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathTearKo_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathTearKo_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathTearKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendConfirm(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendConfirm");

			if (!tmp->evSendConfirm(context)) {
				err("Error in event 'evSendConfirm' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendConfirm];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evSendConfirm_from_virt_WaitResvConf(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendConfirm_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendPath(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendPath");

			if (!tmp->evSendPath(context)) {
				err("Error in event 'evSendPath' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendPath];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Down) : {
					tmp->after_evSendPath_from_virt_Down(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendPath_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evLspRebuild(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evLspRebuild");

			if (!tmp->evLspRebuild(context)) {
				err("Error in event 'evLspRebuild' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evLspRebuild];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Down) : {
					tmp->after_evLspRebuild_from_virt_Down(context);
					break;
				}
				default: {
					dbg(std::string("No after_evLspRebuild_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendPathTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendPathTear");

			if (!tmp->evSendPathTear(context)) {
				err("Error in event 'evSendPathTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendPathTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evSendPathTear_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evSendPathTear_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evSendPathTear_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evSendPathTear_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evSendPathTear_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evSendPathTear_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendPathTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvTimeout(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvTimeout");

			if (!tmp->evRecvResvTimeout(context)) {
				err("Error in event 'evRecvResvTimeout' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvTimeout];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvTimeout_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvResvTimeout_from_virt_TearDown(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvTimeout_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evSendResvTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evSendResvTear");

			if (!tmp->evSendResvTear(context)) {
				err("Error in event 'evSendResvTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evSendResvTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evSendResvTear_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evSendResvTear_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evSendResvTear_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evSendResvTear_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evSendResvTear_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evSendResvTear_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evSendResvTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evXConnDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evXConnDown");

			if (!tmp->evXConnDown(context)) {
				err("Error in event 'evXConnDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evXConnDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evXConnDown_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evXConnDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathOk");

			if (!tmp->evRecvPathOk(context)) {
				err("Error in event 'evRecvPathOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Down) : {
					tmp->after_evRecvPathOk_from_virt_Down(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathOk_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathKo");

			if (!tmp->evRecvPathKo(context)) {
				err("Error in event 'evRecvPathKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Down) : {
					tmp->after_evRecvPathKo_from_virt_Down(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathKo_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvNotifyOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvNotifyOk");

			if (!tmp->evRecvNotifyOk(context)) {
				err("Error in event 'evRecvNotifyOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvNotifyOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvNotifyOk_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvNotifyOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvNotifyDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvNotifyDown");

			if (!tmp->evRecvNotifyDown(context)) {
				err("Error in event 'evRecvNotifyDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvNotifyDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvNotifyDown_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvNotifyDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvNotifyKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvNotifyKo");

			if (!tmp->evRecvNotifyKo(context)) {
				err("Error in event 'evRecvNotifyKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvNotifyKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvNotifyKo_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvNotifyKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathErrOk(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathErrOk");

			if (!tmp->evRecvPathErrOk(context)) {
				err("Error in event 'evRecvPathErrOk' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathErrOk];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathErrOk_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathErrOk_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathErrOk_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathErrOk_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathErrOk_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathErrOk_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathErrOk_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathErrOk_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathErrOk_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathAlarm(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathAlarm");

			if (!tmp->evRecvPathAlarm(context)) {
				err("Error in event 'evRecvPathAlarm' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathAlarm];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathAlarm_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathAlarm_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathAlarm_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathAlarm_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathAlarm_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathErrCrankback(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathErrCrankback");

			if (!tmp->evRecvPathErrCrankback(context)) {
				err("Error in event 'evRecvPathErrCrankback' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathErrCrankback];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathErrCrankback_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathErrCrankback_from_virt_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathErrCrankback_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvActivateErr(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvActivateErr");

			if (!tmp->evRecvActivateErr(context)) {
				err("Error in event 'evRecvActivateErr' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvActivateErr];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvActivateErr_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvActivateErr_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvPathErrKo(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvPathErrKo");

			if (!tmp->evRecvPathErrKo(context)) {
				err("Error in event 'evRecvPathErrKo' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvPathErrKo];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvPathErrKo_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvPathErrKo_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvPathErrKo_from_virt_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_PathReceived) : {
					tmp->after_evRecvPathErrKo_from_virt_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_TearDown) : {
					tmp->after_evRecvPathErrKo_from_virt_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_RebuildTearDown) : {
					tmp->after_evRecvPathErrKo_from_virt_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitResv) : {
					tmp->after_evRecvPathErrKo_from_virt_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Rebuild) : {
					tmp->after_evRecvPathErrKo_from_virt_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvPathErrKo_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		
		bool Fsm::evRecvResvTimer(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: evRecvResvTimer");

			if (!tmp->evRecvResvTimer(context)) {
				err("Error in event 'evRecvResvTimer' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return false;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_evRecvResvTimer];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_virt_WaitResvConf) : {
					tmp->after_evRecvResvTimer_from_virt_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_WaitEqptConf) : {
					tmp->after_evRecvResvTimer_from_virt_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_virt_Installed) : {
					tmp->after_evRecvResvTimer_from_virt_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_evRecvResvTimer_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return true;
		}
		

		
		nextEvFor_RecvResvTear_t Fsm::RecvResvTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResvTear");

			nextEvFor_RecvResvTear_t next_ev;
			next_ev = tmp->RecvResvTear(context);

			if (!next_ev) {
				err("Error in event 'RecvResvTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResvTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvResvTear_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResvTear_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvResvTear_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvResvTear_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvResvTear_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvResvTear_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResvTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPathTimer_t Fsm::RecvPathTimer(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPathTimer");

			nextEvFor_RecvPathTimer_t next_ev;
			next_ev = tmp->RecvPathTimer(context);

			if (!next_ev) {
				err("Error in event 'RecvPathTimer' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPathTimer];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvPathTimer_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPathTimer_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvPathTimer_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPathTimer_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPathTimer_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPathTimer_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_XConnPreempt_t Fsm::XConnPreempt(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: XConnPreempt");

			nextEvFor_XConnPreempt_t next_ev;
			next_ev = tmp->XConnPreempt(context);

			if (!next_ev) {
				err("Error in event 'XConnPreempt' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_XConnPreempt];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_XConnPreempt_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_XConnPreempt_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_XConnPreempt_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_XConnPreempt_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_XConnErr_t Fsm::XConnErr(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: XConnErr");

			nextEvFor_XConnErr_t next_ev;
			next_ev = tmp->XConnErr(context);

			if (!next_ev) {
				err("Error in event 'XConnErr' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_XConnErr];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_XConnErr_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_XConnErr_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_XConnErr_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_XConnErr_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_XConnErr_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_XConnErr_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPathDown_t Fsm::RecvPathDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPathDown");

			nextEvFor_RecvPathDown_t next_ev;
			next_ev = tmp->RecvPathDown(context);

			if (!next_ev) {
				err("Error in event 'RecvPathDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPathDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvPathDown_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPathDown_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvPathDown_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPathDown_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvPathDown_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvPathDown_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPathDown_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_RecvPathDown_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPathDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_XConnCompleted_t Fsm::XConnCompleted(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: XConnCompleted");

			nextEvFor_XConnCompleted_t next_ev;
			next_ev = tmp->XConnCompleted(context);

			if (!next_ev) {
				err("Error in event 'XConnCompleted' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_XConnCompleted];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_XConnCompleted_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_XConnCompleted_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_XConnCompleted_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_XConnCompleted_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_XConnCompleted_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_XConnCompleted_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvResv_t Fsm::RecvResv(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResv");

			nextEvFor_RecvResv_t next_ev;
			next_ev = tmp->RecvResv(context);

			if (!next_ev) {
				err("Error in event 'RecvResv' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResv];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvResv_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResv_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvResv_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvResv_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvResv_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResv_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPathTimeout_t Fsm::RecvPathTimeout(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPathTimeout");

			nextEvFor_RecvPathTimeout_t next_ev;
			next_ev = tmp->RecvPathTimeout(context);

			if (!next_ev) {
				err("Error in event 'RecvPathTimeout' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPathTimeout];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPathTimeout_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPathTimeout_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvPathTimeout_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPathTimeout_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPathTimeout_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvResvErr_t Fsm::RecvResvErr(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResvErr");

			nextEvFor_RecvResvErr_t next_ev;
			next_ev = tmp->RecvResvErr(context);

			if (!next_ev) {
				err("Error in event 'RecvResvErr' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResvErr];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvResvErr_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResvErr_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvResvErr_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvResvErr_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvResvErr_from_RebuildTearDown(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResvErr_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvResvDown_t Fsm::RecvResvDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResvDown");

			nextEvFor_RecvResvDown_t next_ev;
			next_ev = tmp->RecvResvDown(context);

			if (!next_ev) {
				err("Error in event 'RecvResvDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResvDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvResvDown_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResvDown_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvResvDown_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvResvDown_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvResvDown_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvResvDown_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvResvDown_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_RecvResvDown_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResvDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvActivate_t Fsm::RecvActivate(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvActivate");

			nextEvFor_RecvActivate_t next_ev;
			next_ev = tmp->RecvActivate(context);

			if (!next_ev) {
				err("Error in event 'RecvActivate' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvActivate];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvActivate_from_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvActivate_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendResvDown_t Fsm::SendResvDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendResvDown");

			nextEvFor_SendResvDown_t next_ev;
			next_ev = tmp->SendResvDown(context);

			if (!next_ev) {
				err("Error in event 'SendResvDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendResvDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_SendResvDown_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_SendResvDown_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_SendResvDown_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_SendResvDown_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_SendResvDown_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_SendResvDown_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_SendResvDown_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_SendResvDown_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendResvDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendPathDown_t Fsm::SendPathDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendPathDown");

			nextEvFor_SendPathDown_t next_ev;
			next_ev = tmp->SendPathDown(context);

			if (!next_ev) {
				err("Error in event 'SendPathDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendPathDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_SendPathDown_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_SendPathDown_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_SendPathDown_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_SendPathDown_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_SendPathDown_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_SendPathDown_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_SendPathDown_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_SendPathDown_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendPathDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendResv_t Fsm::SendResv(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendResv");

			nextEvFor_SendResv_t next_ev;
			next_ev = tmp->SendResv(context);

			if (!next_ev) {
				err("Error in event 'SendResv' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendResv];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_SendResv_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_SendResv_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendResv_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvConfirm_t Fsm::RecvConfirm(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvConfirm");

			nextEvFor_RecvConfirm_t next_ev;
			next_ev = tmp->RecvConfirm(context);

			if (!next_ev) {
				err("Error in event 'RecvConfirm' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvConfirm];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvConfirm_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvConfirm_from_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvConfirm_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPathTear_t Fsm::RecvPathTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPathTear");

			nextEvFor_RecvPathTear_t next_ev;
			next_ev = tmp->RecvPathTear(context);

			if (!next_ev) {
				err("Error in event 'RecvPathTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPathTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvPathTear_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPathTear_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvPathTear_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPathTear_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvPathTear_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvPathTear_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPathTear_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_RecvPathTear_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPathTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendConfirm_t Fsm::SendConfirm(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendConfirm");

			nextEvFor_SendConfirm_t next_ev;
			next_ev = tmp->SendConfirm(context);

			if (!next_ev) {
				err("Error in event 'SendConfirm' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendConfirm];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_SendConfirm_from_WaitResvConf(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendConfirm_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendPath_t Fsm::SendPath(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendPath");

			nextEvFor_SendPath_t next_ev;
			next_ev = tmp->SendPath(context);

			if (!next_ev) {
				err("Error in event 'SendPath' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendPath];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_Down) : {
					tmp->after_SendPath_from_Down(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendPath_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_LspRebuild_t Fsm::LspRebuild(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: LspRebuild");

			nextEvFor_LspRebuild_t next_ev;
			next_ev = tmp->LspRebuild(context);

			if (!next_ev) {
				err("Error in event 'LspRebuild' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_LspRebuild];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_Down) : {
					tmp->after_LspRebuild_from_Down(context);
					break;
				}
				default: {
					dbg(std::string("No after_LspRebuild_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendPathTear_t Fsm::SendPathTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendPathTear");

			nextEvFor_SendPathTear_t next_ev;
			next_ev = tmp->SendPathTear(context);

			if (!next_ev) {
				err("Error in event 'SendPathTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendPathTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_SendPathTear_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_SendPathTear_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_SendPathTear_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_SendPathTear_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_SendPathTear_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_SendPathTear_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendPathTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvResvTimeout_t Fsm::RecvResvTimeout(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResvTimeout");

			nextEvFor_RecvResvTimeout_t next_ev;
			next_ev = tmp->RecvResvTimeout(context);

			if (!next_ev) {
				err("Error in event 'RecvResvTimeout' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResvTimeout];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResvTimeout_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvResvTimeout_from_TearDown(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResvTimeout_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_SendResvTear_t Fsm::SendResvTear(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: SendResvTear");

			nextEvFor_SendResvTear_t next_ev;
			next_ev = tmp->SendResvTear(context);

			if (!next_ev) {
				err("Error in event 'SendResvTear' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_SendResvTear];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_SendResvTear_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_SendResvTear_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_SendResvTear_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_SendResvTear_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_SendResvTear_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_SendResvTear_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_SendResvTear_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_XConnDown_t Fsm::XConnDown(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: XConnDown");

			nextEvFor_XConnDown_t next_ev;
			next_ev = tmp->XConnDown(context);

			if (!next_ev) {
				err("Error in event 'XConnDown' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_XConnDown];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_XConnDown_from_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_XConnDown_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPath_t Fsm::RecvPath(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPath");

			nextEvFor_RecvPath_t next_ev;
			next_ev = tmp->RecvPath(context);

			if (!next_ev) {
				err("Error in event 'RecvPath' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPath];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvPath_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPath_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvPath_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Down) : {
					tmp->after_RecvPath_from_Down(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPath_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPath_from_WaitResv(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPath_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvNotify_t Fsm::RecvNotify(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvNotify");

			nextEvFor_RecvNotify_t next_ev;
			next_ev = tmp->RecvNotify(context);

			if (!next_ev) {
				err("Error in event 'RecvNotify' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvNotify];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvNotify_from_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvNotify_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvPathErr_t Fsm::RecvPathErr(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvPathErr");

			nextEvFor_RecvPathErr_t next_ev;
			next_ev = tmp->RecvPathErr(context);

			if (!next_ev) {
				err("Error in event 'RecvPathErr' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvPathErr];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvPathErr_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvPathErr_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvPathErr_from_Installed(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_PathReceived) : {
					tmp->after_RecvPathErr_from_PathReceived(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_TearDown) : {
					tmp->after_RecvPathErr_from_TearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_RebuildTearDown) : {
					tmp->after_RecvPathErr_from_RebuildTearDown(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitResv) : {
					tmp->after_RecvPathErr_from_WaitResv(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Rebuild) : {
					tmp->after_RecvPathErr_from_Rebuild(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvPathErr_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		
		nextEvFor_RecvResvTimer_t Fsm::RecvResvTimer(void * context)
		{
			State * tmp;
			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid current state!"));
				//return false;
			}
			dbg("Event: RecvResvTimer");

			nextEvFor_RecvResvTimer_t next_ev;
			next_ev = tmp->RecvResvTimer(context);

			if (!next_ev) {
				err("Error in event 'RecvResvTimer' ... "
				    "maintaining the old state '"
				    + tmp->name() + "'");
				return next_ev;
			}

			prevState_    = currentState_;
			currentState_ = nextState_[currentState_][G2RSVPTE_LSP_FSM_RecvResvTimer];
			dbg(std::string("New state: ") +
			    STATE_NAME(currentState_));

			tmp = currentState();
			if (!tmp) {
				throw(std::string("Invalid next state!"));
			}

			switch (prevState_) { 
				case (G2RSVPTE_LSP_FSM_WaitResvConf) : {
					tmp->after_RecvResvTimer_from_WaitResvConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_WaitEqptConf) : {
					tmp->after_RecvResvTimer_from_WaitEqptConf(context);
					break;
				}
				case (G2RSVPTE_LSP_FSM_Installed) : {
					tmp->after_RecvResvTimer_from_Installed(context);
					break;
				}
				default: {
					dbg(std::string("No after_RecvResvTimer_from_") +
					    STATE_NAME(prevState_));
					break;
				}
			}

			return next_ev;
		}
		

		State * Fsm::currentState(void)
		{
			std::map<states_t, State *>::iterator it;
			it = states_.find(currentState_);
			if (it == states_.end()) {
				err(std::string("Invalid pointer for "
						"current state!"));
				return 0;
			}
			std::string name = (*it).second ?
				(*it).second->name() : "Invalid pointer";
			dbg("Current state: " + name);
			return (*it).second;
		}

		bool Fsm::go2prevState(void)
		{
			if (prevState_ == currentState_) {
				err("Previous and current state are equal!"
				    " ... only 1 step back is allowed");
				return false;
			}

			currentState_ = prevState_;

			std::string name = STATE_NAME(currentState_);
			dbg("Setting FSM state to previous one: " + name);

			return true;
		}

		std::ostream& operator<<(std::ostream& s,
					 const fsm::base_G2RSVPTE_LSP_FSM::
					 Fsm::states_t& st)
		{
			s << STATE_NAME(st);
			return s;
		}

		std::ostream& operator<<(std::ostream& s,
					 const fsm::base_G2RSVPTE_LSP_FSM::
					 Fsm::events_t& ev)
		{
			s << EVENT_NAME(ev);
			return s;
		}
	}
	


	/*************************************/
	/*   Finite State Machine - Wrapper  */
	/*************************************/
	
	
	namespace G2RSVPTE_LSP_FSM {

		//
		// virtFsm
		//
		virtFsm::virtFsm(base_G2RSVPTE_LSP_FSM::BaseFSM::traceLevel_t level)
			throw(std::string)
		{
			try {
				fsm_ = new base_G2RSVPTE_LSP_FSM::Fsm(level);
				assert(fsm_);
			} catch (std::string e) {
				throw(e);
			} catch (...) {
				throw(std::string("Impossible to start virtFSM"));
			}
		}

		virtFsm::~virtFsm(void)
		{
			if (fsm_) delete fsm_;
		}

		std::ostream& operator<<(std::ostream& s, const virtFsm& f)
		{
			s << std::endl << "###### virtFSM [" //<< f.name_
			  << "] DUMP ### START ######" << std::endl;
			//s << f.nextState_;
			s << "###### FSM [" //<< f.name_
			  << "] DUMP ###  END  ######" << std::endl
			  << std::endl;
			return s;
		}

		void virtFsm::post(root_events_t ev, void * context, bool enqueue)
		{
			data_event_t * data;
			data = new data_event_t;
			if (!data) {
				std::cout << "ERR: Cannot allocate memory for "
					"a new data event struct" << std::endl;
				return;
			}
			data->ev      = ev;
			data->context = context;

			events_.push_back(data);

			if (!enqueue) {
				runPendingWork();
			}
		}

		void virtFsm::runPendingWork(void)
		{
			if (events_.empty()) {
				std::cout << "[DBG]: Empty event list ... "
					"nothing else to do" << std::endl;
				return;
			}

			data_event_t * data;
			data = events_.front();
			events_.pop_front();

			if (!data) {
				std::cout << "ERR: Invalid data event from "
					"event list" << std::endl;
				return;
			}

			switch (data->ev) {
				
				case G2RSVPTE_LSP_FSM_RecvResvTear:
					RecvResvTear(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPathTimer:
					RecvPathTimer(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_XConnPreempt:
					XConnPreempt(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_XConnErr:
					XConnErr(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPathDown:
					RecvPathDown(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_XConnCompleted:
					XConnCompleted(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvResv:
					RecvResv(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPathTimeout:
					RecvPathTimeout(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvResvErr:
					RecvResvErr(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvResvDown:
					RecvResvDown(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvActivate:
					RecvActivate(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendResvDown:
					SendResvDown(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendPathDown:
					SendPathDown(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendResv:
					SendResv(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvConfirm:
					RecvConfirm(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPathTear:
					RecvPathTear(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendConfirm:
					SendConfirm(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendPath:
					SendPath(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_LspRebuild:
					LspRebuild(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendPathTear:
					SendPathTear(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvResvTimeout:
					RecvResvTimeout(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_SendResvTear:
					SendResvTear(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_XConnDown:
					XConnDown(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPath:
					RecvPath(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvNotify:
					RecvNotify(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvPathErr:
					RecvPathErr(data->context);
					break;
				
				case G2RSVPTE_LSP_FSM_RecvResvTimer:
					RecvResvTimer(data->context);
					break;
				
				//case G2RSVPTE_LSP_FSM_invalid_event:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
			}

			delete data;

			runPendingWork();
		}

		
		void virtFsm::RecvResvTear(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t next_ev;
			try {
				next_ev = fsm_->RecvResvTear(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTear_to_evRecvResvTearOk:
					try {
						res = fsm_->evRecvResvTearOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTear_to_evRecvResvTearKo:
					try {
						res = fsm_->evRecvResvTearKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTear_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPathTimer(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t next_ev;
			try {
				next_ev = fsm_->RecvPathTimer(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTimer_to_evRecvPathTimer:
					try {
						res = fsm_->evRecvPathTimer(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTimer_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::XConnPreempt(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnPreempt_t next_ev;
			try {
				next_ev = fsm_->XConnPreempt(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnPreempt_to_evXConnPreempt:
					try {
						res = fsm_->evXConnPreempt(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnPreempt_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::XConnErr(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t next_ev;
			try {
				next_ev = fsm_->XConnErr(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnErr_to_evXConnErr:
					try {
						res = fsm_->evXConnErr(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnErr_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPathDown(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t next_ev;
			try {
				next_ev = fsm_->RecvPathDown(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathDown_to_evRecvPathDownOk:
					try {
						res = fsm_->evRecvPathDownOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathDown_to_evRecvPathDownKo:
					try {
						res = fsm_->evRecvPathDownKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathDown_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::XConnCompleted(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t next_ev;
			try {
				next_ev = fsm_->XConnCompleted(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnCompleted_to_evXConnCompleted:
					try {
						res = fsm_->evXConnCompleted(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnCompleted_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvResv(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t next_ev;
			try {
				next_ev = fsm_->RecvResv(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvOk:
					try {
						res = fsm_->evRecvResvOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvKo:
					try {
						res = fsm_->evRecvResvKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvVeryKo:
					try {
						res = fsm_->evRecvResvVeryKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResv_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPathTimeout(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimeout_t next_ev;
			try {
				next_ev = fsm_->RecvPathTimeout(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTimeout_to_evRecvPathTimeout:
					try {
						res = fsm_->evRecvPathTimeout(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTimeout_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvResvErr(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t next_ev;
			try {
				next_ev = fsm_->RecvResvErr(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvErr_to_evRecvResvErrOk:
					try {
						res = fsm_->evRecvResvErrOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvErr_to_evRecvResvErrKo:
					try {
						res = fsm_->evRecvResvErrKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvErr_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvResvDown(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t next_ev;
			try {
				next_ev = fsm_->RecvResvDown(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvDown_to_evRecvResvDownOk:
					try {
						res = fsm_->evRecvResvDownOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvDown_to_evRecvResvDownKo:
					try {
						res = fsm_->evRecvResvDownKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvDown_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvActivate(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvActivate_t next_ev;
			try {
				next_ev = fsm_->RecvActivate(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvActivate_to_evRecvActivateOk:
					try {
						res = fsm_->evRecvActivateOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvActivate_to_evRecvActivateKo:
					try {
						res = fsm_->evRecvActivateKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvActivate_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendResvDown(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t next_ev;
			try {
				next_ev = fsm_->SendResvDown(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResvDown_to_evSendResvDown:
					try {
						res = fsm_->evSendResvDown(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResvDown_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendPathDown(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t next_ev;
			try {
				next_ev = fsm_->SendPathDown(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPathDown_to_evSendPathDown:
					try {
						res = fsm_->evSendPathDown(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPathDown_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendResv(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResv_t next_ev;
			try {
				next_ev = fsm_->SendResv(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResv_to_evSendResv:
					try {
						res = fsm_->evSendResv(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResv_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvConfirm(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvConfirm_t next_ev;
			try {
				next_ev = fsm_->RecvConfirm(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvConfirm_to_evRecvConfirmOk:
					try {
						res = fsm_->evRecvConfirmOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvConfirm_to_evRecvConfirmKo:
					try {
						res = fsm_->evRecvConfirmKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvConfirm_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPathTear(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t next_ev;
			try {
				next_ev = fsm_->RecvPathTear(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTear_to_evRecvPathTearOk:
					try {
						res = fsm_->evRecvPathTearOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTear_to_evRecvPathTearKo:
					try {
						res = fsm_->evRecvPathTearKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathTear_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendConfirm(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendConfirm_t next_ev;
			try {
				next_ev = fsm_->SendConfirm(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendConfirm_to_evSendConfirm:
					try {
						res = fsm_->evSendConfirm(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendConfirm_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendPath(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPath_t next_ev;
			try {
				next_ev = fsm_->SendPath(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPath_to_evSendPath:
					try {
						res = fsm_->evSendPath(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPath_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::LspRebuild(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_LspRebuild_t next_ev;
			try {
				next_ev = fsm_->LspRebuild(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_LspRebuild_to_evLspRebuild:
					try {
						res = fsm_->evLspRebuild(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_LspRebuild_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendPathTear(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t next_ev;
			try {
				next_ev = fsm_->SendPathTear(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPathTear_to_evSendPathTear:
					try {
						res = fsm_->evSendPathTear(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendPathTear_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvResvTimeout(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimeout_t next_ev;
			try {
				next_ev = fsm_->RecvResvTimeout(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTimeout_to_evRecvResvTimeout:
					try {
						res = fsm_->evRecvResvTimeout(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTimeout_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::SendResvTear(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t next_ev;
			try {
				next_ev = fsm_->SendResvTear(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResvTear_to_evSendResvTear:
					try {
						res = fsm_->evSendResvTear(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_SendResvTear_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::XConnDown(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnDown_t next_ev;
			try {
				next_ev = fsm_->XConnDown(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnDown_to_evXConnDown:
					try {
						res = fsm_->evXConnDown(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_XConnDown_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPath(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t next_ev;
			try {
				next_ev = fsm_->RecvPath(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPath_to_evRecvPathOk:
					try {
						res = fsm_->evRecvPathOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPath_to_evRecvPathKo:
					try {
						res = fsm_->evRecvPathKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPath_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvNotify(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvNotify_t next_ev;
			try {
				next_ev = fsm_->RecvNotify(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyOk:
					try {
						res = fsm_->evRecvNotifyOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyDown:
					try {
						res = fsm_->evRecvNotifyDown(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyKo:
					try {
						res = fsm_->evRecvNotifyKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvNotify_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvPathErr(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t next_ev;
			try {
				next_ev = fsm_->RecvPathErr(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrOk:
					try {
						res = fsm_->evRecvPathErrOk(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathAlarm:
					try {
						res = fsm_->evRecvPathAlarm(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrCrankback:
					try {
						res = fsm_->evRecvPathErrCrankback(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvActivateErr:
					try {
						res = fsm_->evRecvActivateErr(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrKo:
					try {
						res = fsm_->evRecvPathErrKo(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvPathErr_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		
		void virtFsm::RecvResvTimer(void * context)
		{
			// case sul tipo di next_event e se errore
			//  torno stato precedente

			fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimer_t next_ev;
			try {
				next_ev = fsm_->RecvResvTimer(context);
			} catch (std::string e) {
				std::cout << "EXCEPTION: " << e << std::endl;
				return;
			} catch (...) {
				std::cout << "UNKNOWN EXCEPTION" << std::endl;
				return;
			}

			bool res;
			switch (next_ev) { 
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTimer_to_evRecvResvTimer:
					try {
						res = fsm_->evRecvResvTimer(context);
					} catch (std::string e) {
						std::cout << "EXCEPTION: " << e << std::endl;
						fsm_->go2prevState();
						return;
					} catch (...) {
						std::cout << "UNKNOWN EXCEPTION" << std::endl;
						fsm_->go2prevState();
						return;
					}
					break;
				
				case fsm::base_G2RSVPTE_LSP_FSM::G2RSVPTE_LSP_FSM_from_RecvResvTimer_to_InvalidEvent:
				default:
					std::cout << "ERR: Invalid event type"
						  << std::endl;
					fsm_->go2prevState();
					return;
			}
			if (!res) {
				std::cout << "ERR: First step for "
					  << __FUNCTION__ <<" failed"
					  << std::endl;
				fsm_->go2prevState();
			}
		}
		

		std::string virtFsm::currentState(void) {
			return fsm_->currentState() ?
				fsm_->currentState()->name(): "Unknown";
		}

		//std::ostream& operator<<(std::ostream& s,
		//			 const fsm::virtual_RSVP::
		//			 virtFsm::states_t& st)
		//{
		//	s << STATE_NAME(st);
		//	return s;
		//}
		//
		//std::ostream& operator<<(std::ostream& s,
		//			 const fsm::virtual_RSVP::
		//			 virtFsm::events_t& ev)
		//{
		//	s << EVENT_NAME(ev);
		//	return s;
		//}
	}
	
}

