/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */

#ifndef FSMGEN_H
#define FSMGEN_H

#ifndef FSMGEN_UTILITY
#define FSMGEN_UTILITY

#include <iostream>
#include <stdio.h>
#include <string>
#include <map>
#include <list>
#include <stdlib.h>

/*****************************************************************************/
/*                          Utility - Matrix class                           */
/*****************************************************************************/

template <class ROW, class COLUMN, class DATA>
class Matrix {
public:
	Matrix() { }
	~Matrix() { }

	//friend class Fsm;

	// Copy operator
	Matrix(const Matrix<ROW, COLUMN, DATA>& m) { matrix_ = m.matrix_; }

	//
	// Type definitions
	//
	typedef ROW *          rowIter;
	typedef const ROW *    const_rowIter;
	typedef COLUMN *       colIter;
	typedef const COLUMN * const_colIter;
	typedef DATA *         dataIter;
	typedef const DATA *   const_dataIter;

	//
	// Iterator support
	//
	rowIter begin(void);
	const_rowIter begin(void) const;
	rowIter end(void);
	const_rowIter end(void) const;
	rowIter next(rowIter rowIt);
	const_rowIter next(const_rowIter rowIt) const;
	colIter begin(rowIter rowIt);
	const_colIter begin(const_rowIter rowIt) const;
	colIter end(rowIter rowIt);
	const_colIter end(const_rowIter rowIt) const;
	colIter next(rowIter rowIt, colIter colIt);
	const_colIter next(const_rowIter rowIt, const_colIter colIt) const;


	// Return the number of deleted cells within the row
	size_t removeRow(const ROW& row);

	// Return the number of deleted cells within the column
	size_t removeCol(const COLUMN& column);

	// Return the number (1 or 0) of deleted data for this pair row/column
	size_t remove(const ROW& row, const COLUMN& column);

	// Return the number of deleted data
	size_t remove(const DATA& data);

	bool remove(dataIter dIt);

	void insert(const ROW& row, const COLUMN& column, const DATA& data);

	dataIter find(const ROW& row, const COLUMN& column);

	// Return the number of data
	size_t size(void) const;

	bool empty(void) const;

	// Assignment operator
	Matrix<ROW, COLUMN, DATA>& operator=(const Matrix<ROW,COLUMN,DATA>& o);

	std::map<COLUMN, DATA>& operator[](const ROW s);

	friend std::ostream& operator<<(std::ostream& s, const Matrix& m) {
		typename std::map<ROW,
			std::map<COLUMN, DATA> >::const_iterator rIt;
		for (rIt = m.matrix_.begin(); rIt != m.matrix_.end(); rIt++) {
			s << "[" << (*rIt).first << "] ";

			typename std::map<COLUMN, DATA>::const_iterator cIt;
			for (cIt = (*rIt).second.begin();
			     cIt != (*rIt).second.end();
			     cIt++) {
				if (cIt != (*rIt).second.begin()) {
					s << " - ";
				}
				s << (*cIt).first << "/" << (*cIt).second;
			}

			s << std::endl;
		}
		return s;
	}

private:
	std::map< ROW, std::map<COLUMN, DATA> > matrix_;
};


template <class ROW, class COLUMN, class DATA>
size_t
Matrix<ROW, COLUMN, DATA>::removeRow(const ROW& row)
{
	size_t count;
	count = 0;
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	rIt = matrix_.find(row);
	if (rIt != matrix_.end()) {
		//std::cout << "Removed row: " << (*rIt).first
		//	  << std::endl;
		count += (*rIt).second.size();
		matrix_.erase(rIt);
	}

	return count;
}

template <class ROW, class COLUMN, class DATA>
size_t
Matrix<ROW, COLUMN, DATA>::removeCol(const COLUMN& column)
{
	size_t count;
	count = 0;
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	for (rIt = matrix_.begin(); rIt != matrix_.end(); rIt++) {
		typename std::map<COLUMN, DATA>::iterator cIt;

		cIt = (*rIt).second.find(column);
		if (cIt != (*rIt).second.end()) {
			//std::cout << "Removed col: " << (*cIt).first
			//	  << std::endl;
			(*rIt).second.erase(cIt);
			count++;
		}
	}

	return count;
}

template <class ROW, class COLUMN, class DATA>
size_t
Matrix<ROW, COLUMN, DATA>::remove(const ROW& row, const COLUMN& column)
{
	size_t count;
	count = 0;
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	rIt = matrix_.find(row);
	if (rIt != matrix_.end()) {
		//std::cout << "Row: " << (*rIt).first << std::endl;
		typename std::map<COLUMN, DATA>::iterator cIt;

		cIt = (*rIt).second.find(column);
		if (cIt != (*rIt).second.end()) {
			//std::cout << "  Col: " << (*cIt).first
			//	  << std::endl
			//	  << "    Removed data: "
			//	  << (*cIt).second << std::endl;
			(*rIt).second.erase(cIt);
			count++;
		}
	}

	return count;
}

template <class ROW, class COLUMN, class DATA>
size_t
Matrix<ROW, COLUMN, DATA>::remove(const DATA& data)
{
	size_t count;
	count = 0;
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	for (rIt = matrix_.begin(); rIt != matrix_.end(); rIt++) {
		typename std::map<COLUMN, DATA>::iterator cIt;
		cIt = (*rIt).second.begin();
		while (cIt != (*rIt).second.end()) {
			if ((*cIt).second == data) {
				(*rIt).second.erase(cIt++);
				count++;
				continue;
			}
			cIt++;
		}
	}

	return count;
}

template <class ROW, class COLUMN, class DATA>
bool
Matrix<ROW, COLUMN, DATA>::remove(dataIter dIt)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	for (rIt = matrix_.begin(); rIt != matrix_.end(); rIt++) {
		typename std::map<COLUMN, DATA>::iterator cIt;
		for (cIt = (*rIt).second.begin();
		     cIt != (*rIt).second.end();
		     cIt++) {
			if (dIt == &(*cIt).second) {
				(*rIt).second.erase(cIt);
				return true;
			}
		}
	}

	return false;
}

template <class ROW, class COLUMN, class DATA>
void
Matrix<ROW, COLUMN, DATA>::insert(const ROW& row,
				  const COLUMN& column,
				  const DATA& data)
{
	matrix_[row][column] = data;
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::dataIter
Matrix<ROW, COLUMN, DATA>::find(const ROW& row, const COLUMN& column)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;

	rIt = matrix_.find(row);
	if (rIt != matrix_.end()) {
		//std::cout << "Row: " << (*rIt).first << std::endl;
		typename std::map<COLUMN, DATA>::iterator cIt;

		cIt = (*rIt).second.find(column);
		if (cIt != (*rIt).second.end()) {
			//std::cout << "  Col: " << (*cIt).first
			//	  << std::endl
			//	  << "    Removed data: "
			//	  << (*cIt).second << std::endl;
			return (dataIter)&((*cIt).second);
		}
	}

	return (dataIter)0;
}

template <class ROW, class COLUMN, class DATA>
size_t
Matrix<ROW, COLUMN, DATA>::size(void) const
{
	size_t count;
	count = 0;
	typename std::map< ROW, std::map<COLUMN, DATA> >::const_iterator rIt;

	for (rIt = matrix_.begin(); rIt != matrix_.end(); rIt++) {
		count += (*rIt).second.size();
	}

	return count;
}

template <class ROW, class COLUMN, class DATA>
bool
Matrix<ROW, COLUMN, DATA>::empty(void) const
{
	return size() ? false : true;
}

template <class ROW, class COLUMN, class DATA>
Matrix<ROW, COLUMN, DATA>&
Matrix<ROW, COLUMN, DATA>::operator=(const Matrix<ROW, COLUMN, DATA>& orig)
{
	if(this != &orig) {
		typename std::map< ROW, std::map<COLUMN,DATA> >::iterator rIt;
		for (rIt=matrix_.begin(); rIt!=matrix_.end(); rIt++) {
			(*rIt).second.clear();
		}
		matrix_.clear();

		matrix_ = orig.matrix_;
	}
	return *this;
}

template <class ROW, class COLUMN, class DATA>
std::map<COLUMN, DATA>&
Matrix<ROW, COLUMN, DATA>::operator[](const ROW s)
{
	return matrix_[s];
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::rowIter
Matrix<ROW, COLUMN, DATA>::begin(void)
{
	return (rowIter)&((*(matrix_.begin())).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_rowIter
Matrix<ROW, COLUMN, DATA>::begin(void) const
{
	return (const_rowIter)&((*(matrix_.begin())).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::rowIter
Matrix<ROW, COLUMN, DATA>::end(void)
{
	return (rowIter)&((*(matrix_.end())).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_rowIter
Matrix<ROW, COLUMN, DATA>::end(void) const
{
	return (const_rowIter)&((*(matrix_.end())).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::rowIter
Matrix<ROW, COLUMN, DATA>::next(rowIter rowIt)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;
	rIt = matrix_.find(*rowIt);
	rIt++;
	return (rowIter)&((*rIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_rowIter
Matrix<ROW, COLUMN, DATA>::next(const_rowIter rowIt) const
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::const_iterator rIt;
	rIt = matrix_.find(*rowIt);
	rIt++;
	return (const_rowIter)&((*rIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::colIter
Matrix<ROW, COLUMN, DATA>::begin(rowIter rowIt)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;
	typename std::map<COLUMN, DATA>::iterator cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.begin();
	return (colIter)&((*cIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_colIter
Matrix<ROW, COLUMN, DATA>::begin(const_rowIter rowIt) const
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::const_iterator rIt;
	typename std::map<COLUMN, DATA>::const_iterator  cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.begin();
	return (const_colIter)&((*cIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::colIter
Matrix<ROW, COLUMN, DATA>::end(rowIter rowIt)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;
	typename std::map<COLUMN, DATA>::iterator cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.end();
	return (colIter)&((*cIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_colIter
Matrix<ROW, COLUMN, DATA>::end(const_rowIter rowIt) const
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::const_iterator rIt;
	typename std::map<COLUMN, DATA>::const_iterator  cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.end();
	return (const_colIter)&((*cIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::colIter
Matrix<ROW, COLUMN, DATA>::next(rowIter rowIt, colIter colIt)
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::iterator rIt;
	typename std::map<COLUMN, DATA>::iterator cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.find(*colIt);
	cIt++;
	return (colIter)&((*cIt).first);
}

template <class ROW, class COLUMN, class DATA>
typename Matrix<ROW, COLUMN, DATA>::const_colIter
Matrix<ROW, COLUMN, DATA>::next(const_rowIter rowIt,
				const_colIter colIt) const
{
	typename std::map< ROW, std::map<COLUMN, DATA> >::const_iterator rIt;
	typename std::map<COLUMN, DATA>::const_iterator  cIt;
	rIt = matrix_.find(*rowIt);
	cIt = (*rIt).second.find(*colIt);
	cIt++;
	return (const_colIter)&((*cIt).first);
}
#endif // FSMGEN_UTILITY

namespace fsm {
	
	
#ifndef NAMESPACE_BASE_G2RSVPTE_LSP_FSM
#define NAMESPACE_BASE_G2RSVPTE_LSP_FSM

#include <iostream>
#include <stdio.h>
#include <string>
#include <map>
#include <list>
#include <stdlib.h>

/*****************************************************************************/
/*                             Finite State Machine                          */
/*****************************************************************************/
	/*************************************/
	/*    Finite State Machine - Core    */
	/*************************************/
	namespace base_G2RSVPTE_LSP_FSM {

#define STATE_NAME(x)              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_WaitResvConf ? "WaitResvConf" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_WaitEqptConf ? "WaitEqptConf" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_Installed ? "Installed" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_Down ? "Down" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_PathReceived ? "PathReceived" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_TearDown ? "TearDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RebuildTearDown ? "RebuildTearDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_WaitResv ? "WaitResv" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_Rebuild ? "Rebuild" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_WaitResvConf ? "virt_WaitResvConf" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_WaitEqptConf ? "virt_WaitEqptConf" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_Installed ? "virt_Installed" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_Down ? "virt_Down" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_PathReceived ? "virt_PathReceived" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_TearDown ? "virt_TearDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_RebuildTearDown ? "virt_RebuildTearDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_WaitResv ? "virt_WaitResv" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_virt_Rebuild ? "virt_Rebuild" :     \
		       "Unknown"))))))))))))))))))))))))))))))))))))

#define EVENT_NAME(x)              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvTearOk ? "evRecvResvTearOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvTearKo ? "evRecvResvTearKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathTimer ? "evRecvPathTimer" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evXConnPreempt ? "evXConnPreempt" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evXConnErr ? "evXConnErr" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathDownOk ? "evRecvPathDownOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathDownKo ? "evRecvPathDownKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evXConnCompleted ? "evXConnCompleted" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvOk ? "evRecvResvOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvKo ? "evRecvResvKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvVeryKo ? "evRecvResvVeryKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathTimeout ? "evRecvPathTimeout" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvErrOk ? "evRecvResvErrOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvErrKo ? "evRecvResvErrKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvDownOk ? "evRecvResvDownOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvDownKo ? "evRecvResvDownKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvActivateOk ? "evRecvActivateOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvActivateKo ? "evRecvActivateKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendResvDown ? "evSendResvDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendPathDown ? "evSendPathDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendResv ? "evSendResv" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvConfirmOk ? "evRecvConfirmOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvConfirmKo ? "evRecvConfirmKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathTearOk ? "evRecvPathTearOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathTearKo ? "evRecvPathTearKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendConfirm ? "evSendConfirm" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendPath ? "evSendPath" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evLspRebuild ? "evLspRebuild" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendPathTear ? "evSendPathTear" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvTimeout ? "evRecvResvTimeout" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evSendResvTear ? "evSendResvTear" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evXConnDown ? "evXConnDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathOk ? "evRecvPathOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathKo ? "evRecvPathKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvNotifyOk ? "evRecvNotifyOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvNotifyDown ? "evRecvNotifyDown" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvNotifyKo ? "evRecvNotifyKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathErrOk ? "evRecvPathErrOk" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathAlarm ? "evRecvPathAlarm" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathErrCrankback ? "evRecvPathErrCrankback" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvActivateErr ? "evRecvActivateErr" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvPathErrKo ? "evRecvPathErrKo" :              \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_evRecvResvTimer ? "evRecvResvTimer" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResvTear ? "RecvResvTear" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPathTimer ? "RecvPathTimer" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_XConnPreempt ? "XConnPreempt" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_XConnErr ? "XConnErr" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPathDown ? "RecvPathDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_XConnCompleted ? "XConnCompleted" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResv ? "RecvResv" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPathTimeout ? "RecvPathTimeout" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResvErr ? "RecvResvErr" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResvDown ? "RecvResvDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvActivate ? "RecvActivate" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendResvDown ? "SendResvDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendPathDown ? "SendPathDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendResv ? "SendResv" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvConfirm ? "RecvConfirm" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPathTear ? "RecvPathTear" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendConfirm ? "SendConfirm" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendPath ? "SendPath" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_LspRebuild ? "LspRebuild" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendPathTear ? "SendPathTear" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResvTimeout ? "RecvResvTimeout" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_SendResvTear ? "SendResvTear" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_XConnDown ? "XConnDown" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPath ? "RecvPath" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvNotify ? "RecvNotify" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvPathErr ? "RecvPathErr" : \
  ((x == fsm::base_G2RSVPTE_LSP_FSM::Fsm::G2RSVPTE_LSP_FSM_RecvResvTimer ? "RecvResvTimer" :      \
		       "Unknown"))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))

#define DUMP(level, threshold, tag, str)				\
		if ((level) <= threshold) {				\
			std::cout << "[" << getpid() << "]"		\
				  << "[" << __FUNCTION__ << "]: "	\
				  << "[" << tag << "] "			\
				  << str				\
				  << std::endl;				\
		}

		
		enum nextEvFor_RecvResvTear_t {
			G2RSVPTE_LSP_FSM_from_RecvResvTear_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResvTear_to_evRecvResvTearOk,
			G2RSVPTE_LSP_FSM_from_RecvResvTear_to_evRecvResvTearKo,
			
		};
		
		enum nextEvFor_RecvPathTimer_t {
			G2RSVPTE_LSP_FSM_from_RecvPathTimer_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPathTimer_to_evRecvPathTimer,
			
		};
		
		enum nextEvFor_XConnPreempt_t {
			G2RSVPTE_LSP_FSM_from_XConnPreempt_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_XConnPreempt_to_evXConnPreempt,
			
		};
		
		enum nextEvFor_XConnErr_t {
			G2RSVPTE_LSP_FSM_from_XConnErr_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_XConnErr_to_evXConnErr,
			
		};
		
		enum nextEvFor_RecvPathDown_t {
			G2RSVPTE_LSP_FSM_from_RecvPathDown_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPathDown_to_evRecvPathDownOk,
			G2RSVPTE_LSP_FSM_from_RecvPathDown_to_evRecvPathDownKo,
			
		};
		
		enum nextEvFor_XConnCompleted_t {
			G2RSVPTE_LSP_FSM_from_XConnCompleted_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_XConnCompleted_to_evXConnCompleted,
			
		};
		
		enum nextEvFor_RecvResv_t {
			G2RSVPTE_LSP_FSM_from_RecvResv_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvOk,
			G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvKo,
			G2RSVPTE_LSP_FSM_from_RecvResv_to_evRecvResvVeryKo,
			
		};
		
		enum nextEvFor_RecvPathTimeout_t {
			G2RSVPTE_LSP_FSM_from_RecvPathTimeout_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPathTimeout_to_evRecvPathTimeout,
			
		};
		
		enum nextEvFor_RecvResvErr_t {
			G2RSVPTE_LSP_FSM_from_RecvResvErr_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResvErr_to_evRecvResvErrOk,
			G2RSVPTE_LSP_FSM_from_RecvResvErr_to_evRecvResvErrKo,
			
		};
		
		enum nextEvFor_RecvResvDown_t {
			G2RSVPTE_LSP_FSM_from_RecvResvDown_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResvDown_to_evRecvResvDownOk,
			G2RSVPTE_LSP_FSM_from_RecvResvDown_to_evRecvResvDownKo,
			
		};
		
		enum nextEvFor_RecvActivate_t {
			G2RSVPTE_LSP_FSM_from_RecvActivate_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvActivate_to_evRecvActivateOk,
			G2RSVPTE_LSP_FSM_from_RecvActivate_to_evRecvActivateKo,
			
		};
		
		enum nextEvFor_SendResvDown_t {
			G2RSVPTE_LSP_FSM_from_SendResvDown_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendResvDown_to_evSendResvDown,
			
		};
		
		enum nextEvFor_SendPathDown_t {
			G2RSVPTE_LSP_FSM_from_SendPathDown_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendPathDown_to_evSendPathDown,
			
		};
		
		enum nextEvFor_SendResv_t {
			G2RSVPTE_LSP_FSM_from_SendResv_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendResv_to_evSendResv,
			
		};
		
		enum nextEvFor_RecvConfirm_t {
			G2RSVPTE_LSP_FSM_from_RecvConfirm_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvConfirm_to_evRecvConfirmOk,
			G2RSVPTE_LSP_FSM_from_RecvConfirm_to_evRecvConfirmKo,
			
		};
		
		enum nextEvFor_RecvPathTear_t {
			G2RSVPTE_LSP_FSM_from_RecvPathTear_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPathTear_to_evRecvPathTearOk,
			G2RSVPTE_LSP_FSM_from_RecvPathTear_to_evRecvPathTearKo,
			
		};
		
		enum nextEvFor_SendConfirm_t {
			G2RSVPTE_LSP_FSM_from_SendConfirm_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendConfirm_to_evSendConfirm,
			
		};
		
		enum nextEvFor_SendPath_t {
			G2RSVPTE_LSP_FSM_from_SendPath_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendPath_to_evSendPath,
			
		};
		
		enum nextEvFor_LspRebuild_t {
			G2RSVPTE_LSP_FSM_from_LspRebuild_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_LspRebuild_to_evLspRebuild,
			
		};
		
		enum nextEvFor_SendPathTear_t {
			G2RSVPTE_LSP_FSM_from_SendPathTear_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendPathTear_to_evSendPathTear,
			
		};
		
		enum nextEvFor_RecvResvTimeout_t {
			G2RSVPTE_LSP_FSM_from_RecvResvTimeout_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResvTimeout_to_evRecvResvTimeout,
			
		};
		
		enum nextEvFor_SendResvTear_t {
			G2RSVPTE_LSP_FSM_from_SendResvTear_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_SendResvTear_to_evSendResvTear,
			
		};
		
		enum nextEvFor_XConnDown_t {
			G2RSVPTE_LSP_FSM_from_XConnDown_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_XConnDown_to_evXConnDown,
			
		};
		
		enum nextEvFor_RecvPath_t {
			G2RSVPTE_LSP_FSM_from_RecvPath_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPath_to_evRecvPathOk,
			G2RSVPTE_LSP_FSM_from_RecvPath_to_evRecvPathKo,
			
		};
		
		enum nextEvFor_RecvNotify_t {
			G2RSVPTE_LSP_FSM_from_RecvNotify_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyOk,
			G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyDown,
			G2RSVPTE_LSP_FSM_from_RecvNotify_to_evRecvNotifyKo,
			
		};
		
		enum nextEvFor_RecvPathErr_t {
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrOk,
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathAlarm,
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrCrankback,
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvActivateErr,
			G2RSVPTE_LSP_FSM_from_RecvPathErr_to_evRecvPathErrKo,
			
		};
		
		enum nextEvFor_RecvResvTimer_t {
			G2RSVPTE_LSP_FSM_from_RecvResvTimer_to_InvalidEvent = 0,
			G2RSVPTE_LSP_FSM_from_RecvResvTimer_to_evRecvResvTimer,
			
		};
		

		class State {
		public:
			State(std::string name = "Base state") :
				name_(name) { }
			virtual ~State(void) { }

			std::string name(void) { return name_; }

			// On event
			
			virtual bool evRecvResvTearOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvTearOk'"));
			}
			
			virtual bool evRecvResvTearKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvTearKo'"));
			}
			
			virtual bool evRecvPathTimer(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathTimer'"));
			}
			
			virtual bool evXConnPreempt(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evXConnPreempt'"));
			}
			
			virtual bool evXConnErr(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evXConnErr'"));
			}
			
			virtual bool evRecvPathDownOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathDownOk'"));
			}
			
			virtual bool evRecvPathDownKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathDownKo'"));
			}
			
			virtual bool evXConnCompleted(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evXConnCompleted'"));
			}
			
			virtual bool evRecvResvOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvOk'"));
			}
			
			virtual bool evRecvResvKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvKo'"));
			}
			
			virtual bool evRecvResvVeryKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvVeryKo'"));
			}
			
			virtual bool evRecvPathTimeout(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathTimeout'"));
			}
			
			virtual bool evRecvResvErrOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvErrOk'"));
			}
			
			virtual bool evRecvResvErrKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvErrKo'"));
			}
			
			virtual bool evRecvResvDownOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvDownOk'"));
			}
			
			virtual bool evRecvResvDownKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvDownKo'"));
			}
			
			virtual bool evRecvActivateOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvActivateOk'"));
			}
			
			virtual bool evRecvActivateKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvActivateKo'"));
			}
			
			virtual bool evSendResvDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendResvDown'"));
			}
			
			virtual bool evSendPathDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendPathDown'"));
			}
			
			virtual bool evSendResv(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendResv'"));
			}
			
			virtual bool evRecvConfirmOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvConfirmOk'"));
			}
			
			virtual bool evRecvConfirmKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvConfirmKo'"));
			}
			
			virtual bool evRecvPathTearOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathTearOk'"));
			}
			
			virtual bool evRecvPathTearKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathTearKo'"));
			}
			
			virtual bool evSendConfirm(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendConfirm'"));
			}
			
			virtual bool evSendPath(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendPath'"));
			}
			
			virtual bool evLspRebuild(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evLspRebuild'"));
			}
			
			virtual bool evSendPathTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendPathTear'"));
			}
			
			virtual bool evRecvResvTimeout(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvTimeout'"));
			}
			
			virtual bool evSendResvTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evSendResvTear'"));
			}
			
			virtual bool evXConnDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evXConnDown'"));
			}
			
			virtual bool evRecvPathOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathOk'"));
			}
			
			virtual bool evRecvPathKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathKo'"));
			}
			
			virtual bool evRecvNotifyOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvNotifyOk'"));
			}
			
			virtual bool evRecvNotifyDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvNotifyDown'"));
			}
			
			virtual bool evRecvNotifyKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvNotifyKo'"));
			}
			
			virtual bool evRecvPathErrOk(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathErrOk'"));
			}
			
			virtual bool evRecvPathAlarm(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathAlarm'"));
			}
			
			virtual bool evRecvPathErrCrankback(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathErrCrankback'"));
			}
			
			virtual bool evRecvActivateErr(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvActivateErr'"));
			}
			
			virtual bool evRecvPathErrKo(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvPathErrKo'"));
			}
			
			virtual bool evRecvResvTimer(void * context) {
				throw(std::string("Undefined transition on event "
				      "'evRecvResvTimer'"));
			}
			
			virtual nextEvFor_RecvResvTear_t RecvResvTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResvTear'"));
			}
			
			virtual nextEvFor_RecvPathTimer_t RecvPathTimer(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPathTimer'"));
			}
			
			virtual nextEvFor_XConnPreempt_t XConnPreempt(void * context) {
				throw(std::string("Undefined transition on event "
				      "'XConnPreempt'"));
			}
			
			virtual nextEvFor_XConnErr_t XConnErr(void * context) {
				throw(std::string("Undefined transition on event "
				      "'XConnErr'"));
			}
			
			virtual nextEvFor_RecvPathDown_t RecvPathDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPathDown'"));
			}
			
			virtual nextEvFor_XConnCompleted_t XConnCompleted(void * context) {
				throw(std::string("Undefined transition on event "
				      "'XConnCompleted'"));
			}
			
			virtual nextEvFor_RecvResv_t RecvResv(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResv'"));
			}
			
			virtual nextEvFor_RecvPathTimeout_t RecvPathTimeout(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPathTimeout'"));
			}
			
			virtual nextEvFor_RecvResvErr_t RecvResvErr(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResvErr'"));
			}
			
			virtual nextEvFor_RecvResvDown_t RecvResvDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResvDown'"));
			}
			
			virtual nextEvFor_RecvActivate_t RecvActivate(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvActivate'"));
			}
			
			virtual nextEvFor_SendResvDown_t SendResvDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendResvDown'"));
			}
			
			virtual nextEvFor_SendPathDown_t SendPathDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendPathDown'"));
			}
			
			virtual nextEvFor_SendResv_t SendResv(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendResv'"));
			}
			
			virtual nextEvFor_RecvConfirm_t RecvConfirm(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvConfirm'"));
			}
			
			virtual nextEvFor_RecvPathTear_t RecvPathTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPathTear'"));
			}
			
			virtual nextEvFor_SendConfirm_t SendConfirm(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendConfirm'"));
			}
			
			virtual nextEvFor_SendPath_t SendPath(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendPath'"));
			}
			
			virtual nextEvFor_LspRebuild_t LspRebuild(void * context) {
				throw(std::string("Undefined transition on event "
				      "'LspRebuild'"));
			}
			
			virtual nextEvFor_SendPathTear_t SendPathTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendPathTear'"));
			}
			
			virtual nextEvFor_RecvResvTimeout_t RecvResvTimeout(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResvTimeout'"));
			}
			
			virtual nextEvFor_SendResvTear_t SendResvTear(void * context) {
				throw(std::string("Undefined transition on event "
				      "'SendResvTear'"));
			}
			
			virtual nextEvFor_XConnDown_t XConnDown(void * context) {
				throw(std::string("Undefined transition on event "
				      "'XConnDown'"));
			}
			
			virtual nextEvFor_RecvPath_t RecvPath(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPath'"));
			}
			
			virtual nextEvFor_RecvNotify_t RecvNotify(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvNotify'"));
			}
			
			virtual nextEvFor_RecvPathErr_t RecvPathErr(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvPathErr'"));
			}
			
			virtual nextEvFor_RecvResvTimer_t RecvResvTimer(void * context) {
				throw(std::string("Undefined transition on event "
				      "'RecvResvTimer'"));
			}
			

			// After event from state
			 
			virtual void after_SendPathDown_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evSendPathDown_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvResv_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResv' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvResvKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPathErr_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPath_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvResvTimer_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTimer' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvResvTimer_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTimer' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_SendPathTear_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evSendPathTear_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvResvDown_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_SendConfirm_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendConfirm' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evSendConfirm_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendConfirm' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvConfirm_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvConfirm' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvConfirmOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvConfirmOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_XConnPreempt_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnPreempt' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evXConnPreempt_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnPreempt' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPathDown_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvResvTear_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvConfirmKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvConfirmKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvPathAlarm_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathAlarm' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_SendResvTear_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evSendResvTear_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvResvErr_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvErr' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvResvErrKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPathTear_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvResvErrOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPathTimer_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimer' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evRecvPathTimer_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimer' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvResvOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_SendResvDown_from_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'WaitResvConf'"));
			}
			
			virtual void after_evSendResvDown_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_WaitResvConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_WaitResvConf'"));
			}
			
			virtual void after_RecvPathTimeout_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimeout' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathTimeout_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimeout' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_SendPathDown_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evSendPathDown_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResv_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResv' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_XConnCompleted_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnCompleted' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evXConnCompleted_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnCompleted' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvPathErr_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvPath_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResvTimer_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTimer' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvTimer_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTimer' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_SendPathTear_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evSendPathTear_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_XConnErr_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnErr' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evXConnErr_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnErr' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResvDown_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvPathDown_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResvTimeout_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTimeout' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvTimeout_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTimeout' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvVeryKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvVeryKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResvTear_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_SendResvTear_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evSendResvTear_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvPathTear_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvResvErr_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvErr' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvErrKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvErrOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_RecvPathTimer_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimer' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathTimer_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimer' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_SendResvDown_from_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'WaitEqptConf'"));
			}
			
			virtual void after_evSendResvDown_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_WaitEqptConf(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_WaitEqptConf'"));
			}
			
			virtual void after_SendPathDown_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'Installed'"));
			}
			
			virtual void after_evSendPathDown_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_XConnErr_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnErr' from state "
				      "'Installed'"));
			}
			
			virtual void after_evXConnErr_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnErr' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvConfirm_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvConfirm' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvConfirmOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvConfirmOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_XConnPreempt_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnPreempt' from state "
				      "'Installed'"));
			}
			
			virtual void after_evXConnPreempt_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnPreempt' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvPathDown_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_XConnDown_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnDown' from state "
				      "'Installed'"));
			}
			
			virtual void after_evXConnDown_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnDown' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvPath_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvResvTear_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvPathErr_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvPathAlarm_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathAlarm' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvConfirmKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvConfirmKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvPathTear_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvPathTimer_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimer' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvPathTimer_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimer' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvResv_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResv' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvResvOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvResvDown_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvNotify_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvNotify' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvNotifyOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvNotifyOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvResvKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvNotifyDown_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvNotifyDown' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_XConnCompleted_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnCompleted' from state "
				      "'Installed'"));
			}
			
			virtual void after_evXConnCompleted_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnCompleted' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvActivateErr_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvActivateErr' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvResvTimer_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTimer' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvResvTimer_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTimer' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvActivate_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvActivate' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvActivateOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvActivateOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvResvVeryKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvVeryKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvNotifyKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvNotifyKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_RecvResvErr_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvErr' from state "
				      "'Installed'"));
			}
			
			virtual void after_evRecvResvErrKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvActivateKo_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvActivateKo' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_evRecvResvErrOk_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrOk' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_SendResvDown_from_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'Installed'"));
			}
			
			virtual void after_evSendResvDown_from_virt_Installed(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_Installed'"));
			}
			
			virtual void after_LspRebuild_from_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'LspRebuild' from state "
				      "'Down'"));
			}
			
			virtual void after_evLspRebuild_from_virt_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evLspRebuild' from state "
				      "'virt_Down'"));
			}
			
			virtual void after_SendPath_from_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPath' from state "
				      "'Down'"));
			}
			
			virtual void after_evSendPath_from_virt_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPath' from state "
				      "'virt_Down'"));
			}
			
			virtual void after_RecvPath_from_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'Down'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_Down'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_Down(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_Down'"));
			}
			
			virtual void after_RecvPathTimeout_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimeout' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathTimeout_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimeout' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_SendPathDown_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evSendPathDown_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvResv_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResv' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvResvKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_XConnCompleted_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnCompleted' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evXConnCompleted_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnCompleted' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPathErr_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPath_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_SendPathTear_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evSendPathTear_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_XConnErr_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnErr' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evXConnErr_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnErr' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvResvDown_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPathDown_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvPathErrCrankback_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrCrankback' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_SendResvTear_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evSendResvTear_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPathTear_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPathTimer_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimer' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evRecvPathTimer_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimer' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvResvOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_SendResvDown_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evSendResvDown_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_SendResv_from_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResv' from state "
				      "'PathReceived'"));
			}
			
			virtual void after_evSendResv_from_virt_PathReceived(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResv' from state "
				      "'virt_PathReceived'"));
			}
			
			virtual void after_RecvPathTimeout_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimeout' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvPathTimeout_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimeout' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_SendPathDown_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evSendPathDown_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvPathErr_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_SendPathTear_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evSendPathTear_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvResvDown_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvPathDown_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvResvTimeout_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTimeout' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvResvTimeout_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTimeout' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvResvTear_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvPathAlarm_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathAlarm' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_SendResvTear_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evSendResvTear_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvPathTear_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_RecvResvErr_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvErr' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evRecvResvErrKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvResvErrOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_SendResvDown_from_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'TearDown'"));
			}
			
			virtual void after_evSendResvDown_from_virt_TearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_TearDown'"));
			}
			
			virtual void after_SendPathDown_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evSendPathDown_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_XConnCompleted_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnCompleted' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evXConnCompleted_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnCompleted' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvPathErr_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_SendPathTear_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evSendPathTear_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_XConnErr_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnErr' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evXConnErr_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnErr' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvResvDown_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvPathDown_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvResvTear_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_SendResvTear_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evSendResvTear_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvPathTear_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvResvErr_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvErr' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvErrKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvErrOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvErrOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_SendResvDown_from_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'RebuildTearDown'"));
			}
			
			virtual void after_evSendResvDown_from_virt_RebuildTearDown(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_RebuildTearDown'"));
			}
			
			virtual void after_RecvPathTimeout_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimeout' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathTimeout_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimeout' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_SendPathDown_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evSendPathDown_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvResv_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResv' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvResvKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPathErr_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPath_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPath' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_SendPathTear_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathTear' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evSendPathTear_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathTear' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvResvDown_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPathDown_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_XConnPreempt_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnPreempt' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evXConnPreempt_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnPreempt' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvResvVeryKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvVeryKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvResvTear_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvTear' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvResvTearKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathErrCrankback_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrCrankback' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvResvTearOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvTearOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathAlarm_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathAlarm' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_SendResvTear_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvTear' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evSendResvTear_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvTear' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPathTear_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPathTimer_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTimer' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evRecvPathTimer_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTimer' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvResvOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_SendResvDown_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evSendResvDown_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_SendResv_from_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResv' from state "
				      "'WaitResv'"));
			}
			
			virtual void after_evSendResv_from_virt_WaitResv(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResv' from state "
				      "'virt_WaitResv'"));
			}
			
			virtual void after_RecvPathDown_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathDown' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evRecvPathDownKo_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownKo' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_SendPathDown_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendPathDown' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evSendPathDown_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendPathDown' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_XConnCompleted_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnCompleted' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evXConnCompleted_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnCompleted' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_RecvPathErr_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathErr' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evRecvPathErrKo_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrKo' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_XConnErr_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'XConnErr' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evXConnErr_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evXConnErr' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_RecvPathTear_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvPathTear' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evRecvPathTearOk_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearOk' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_RecvResvDown_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'RecvResvDown' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evRecvResvDownKo_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownKo' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_evRecvPathTearKo_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathTearKo' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_evRecvPathErrOk_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathErrOk' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_SendResvDown_from_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'SendResvDown' from state "
				      "'Rebuild'"));
			}
			
			virtual void after_evSendResvDown_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evSendResvDown' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_evRecvPathDownOk_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvPathDownOk' from state "
				      "'virt_Rebuild'"));
			}
			
			virtual void after_evRecvResvDownOk_from_virt_Rebuild(void * context) {
				throw(std::string("Undefined transition after event "
				      "'evRecvResvDownOk' from state "
				      "'virt_Rebuild'"));
			}
			

		private:
			std::string name_;
		};

		/*
		 *  Classes that MUST be derived!!! - START
		 */
		class WaitResvConf_i;
		class WaitEqptConf_i;
		class Installed_i;
		class Down_i;
		class PathReceived_i;
		class TearDown_i;
		class RebuildTearDown_i;
		class WaitResv_i;
		class Rebuild_i;
		
		class virt_WaitResvConf_i;
		class virt_WaitEqptConf_i;
		class virt_Installed_i;
		class virt_Down_i;
		class virt_PathReceived_i;
		class virt_TearDown_i;
		class virt_RebuildTearDown_i;
		class virt_WaitResv_i;
		class virt_Rebuild_i;
		

		
		class WaitResvConf : public State {
		public:
			WaitResvConf() :
				State(std::string("WaitResvConf")) { }
			virtual ~WaitResvConf() { }
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t RecvResv(void* context) = 0;
			  
			virtual void after_evRecvResvKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			  
			virtual void after_evRecvPathOk_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimer_t RecvResvTimer(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendConfirm_t SendConfirm(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvConfirm_t RecvConfirm(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnPreempt_t XConnPreempt(void* context) = 0;
			  
			virtual void after_evXConnPreempt_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			    
			virtual void after_evRecvPathDownKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			  
			virtual void after_evRecvResvTearKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual void after_evRecvPathKo_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual void after_evRecvConfirmKo_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t RecvResvErr(void* context) = 0;
			  
			virtual void after_evRecvResvErrKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual void after_evRecvPathTearKo_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual void after_evRecvResvErrOk_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t RecvPathTimer(void* context) = 0;
			    
			virtual void after_evRecvResvOk_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			                    
			virtual void after_evXConnCompleted_from_virt_WaitEqptConf(void * context) = 0;
			                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
			virtual void after_evRecvResvOk_from_virt_WaitResv(void * context) = 0;
			          
			virtual void after_evSendResv_from_virt_WaitResv(void * context) = 0;
			                                         
		};
		
		class WaitEqptConf : public State {
		public:
			WaitEqptConf() :
				State(std::string("WaitEqptConf")) { }
			virtual ~WaitEqptConf() { }
			                                                                                          
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimeout_t RecvPathTimeout(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t RecvResv(void* context) = 0;
			  
			virtual void after_evRecvResvKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t XConnCompleted(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			  
			virtual void after_evRecvPathOk_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimer_t RecvResvTimer(void* context) = 0;
			  
			virtual void after_evRecvResvTimer_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t XConnErr(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			    
			virtual void after_evRecvPathDownKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimeout_t RecvResvTimeout(void* context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			  
			virtual void after_evRecvResvTearKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual void after_evRecvPathKo_from_virt_WaitEqptConf(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t RecvResvErr(void* context) = 0;
			  
			virtual void after_evRecvResvErrKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual void after_evRecvPathTearKo_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual void after_evRecvResvErrOk_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t RecvPathTimer(void* context) = 0;
			  
			virtual void after_evRecvPathTimer_from_virt_WaitEqptConf(void * context) = 0;
			    
			virtual void after_evRecvResvOk_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			                                                                                                                                                                                            
			virtual void after_evRecvResvOk_from_virt_PathReceived(void * context) = 0;
			          
			virtual void after_evSendResv_from_virt_PathReceived(void * context) = 0;
			                                                                                                                                                                                                                                                   
		};
		
		class Installed : public State {
		public:
			Installed() :
				State(std::string("Installed")) { }
			virtual ~Installed() { }
			                                
			virtual void after_evSendConfirm_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual void after_evRecvConfirmOk_from_virt_WaitResvConf(void * context) = 0;
			                                                                                                                                                
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t XConnErr(void* context) = 0;
			  
			virtual void after_evXConnErr_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvConfirm_t RecvConfirm(void* context) = 0;
			  
			virtual void after_evRecvConfirmOk_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnPreempt_t XConnPreempt(void* context) = 0;
			  
			virtual void after_evXConnPreempt_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			  
			virtual void after_evRecvPathDownKo_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnDown_t XConnDown(void* context) = 0;
			  
			virtual void after_evXConnDown_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			  
			virtual void after_evRecvPathKo_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathAlarm_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvConfirmKo_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			  
			virtual void after_evRecvPathTearKo_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t RecvPathTimer(void* context) = 0;
			  
			virtual void after_evRecvPathTimer_from_virt_Installed(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t RecvResv(void* context) = 0;
			  
			virtual void after_evRecvResvOk_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvNotify_t RecvNotify(void* context) = 0;
			  
			virtual void after_evRecvNotifyOk_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvResvKo_from_virt_Installed(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t XConnCompleted(void* context) = 0;
			  
			virtual void after_evXConnCompleted_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvPathOk_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvActivateErr_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimer_t RecvResvTimer(void* context) = 0;
			  
			virtual void after_evRecvResvTimer_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvActivate_t RecvActivate(void* context) = 0;
			  
			virtual void after_evRecvActivateOk_from_virt_Installed(void * context) = 0;
			      
			virtual void after_evRecvResvTearKo_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvNotifyKo_from_virt_Installed(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t RecvResvErr(void* context) = 0;
			  
			virtual void after_evRecvResvErrKo_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvActivateKo_from_virt_Installed(void * context) = 0;
			  
			virtual void after_evRecvResvErrOk_from_virt_Installed(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			                                                                                                                                                                                                                                                                                                                
			virtual void after_evXConnCompleted_from_virt_Rebuild(void * context) = 0;
			                             
		};
		
		class Down : public State {
		public:
			Down() :
				State(std::string("Down")) { }
			virtual ~Down() { }
			                        
			virtual void after_evSendPathTear_from_virt_WaitResvConf(void * context) = 0;
			                              
			virtual void after_evRecvResvTearOk_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual void after_evRecvPathAlarm_from_virt_WaitResvConf(void * context) = 0;
			    
			virtual void after_evSendResvTear_from_virt_WaitResvConf(void * context) = 0;
			        
			virtual void after_evRecvPathTearOk_from_virt_WaitResvConf(void * context) = 0;
			            
			virtual void after_evRecvPathErrOk_from_virt_WaitResvConf(void * context) = 0;
			          
			virtual void after_evRecvPathTimeout_from_virt_WaitEqptConf(void * context) = 0;
			                            
			virtual void after_evSendPathTear_from_virt_WaitEqptConf(void * context) = 0;
			                  
			virtual void after_evRecvResvTimeout_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual void after_evRecvResvVeryKo_from_virt_WaitEqptConf(void * context) = 0;
			        
			virtual void after_evRecvResvTearOk_from_virt_WaitEqptConf(void * context) = 0;
			    
			virtual void after_evSendResvTear_from_virt_WaitEqptConf(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_WaitEqptConf(void * context) = 0;
			              
			virtual void after_evRecvPathErrOk_from_virt_WaitEqptConf(void * context) = 0;
			                                        
			virtual void after_evRecvResvTearOk_from_virt_Installed(void * context) = 0;
			                
			virtual void after_evRecvPathErrOk_from_virt_Installed(void * context) = 0;
			                
			virtual void after_evRecvNotifyDown_from_virt_Installed(void * context) = 0;
			                        
			virtual void after_evRecvResvVeryKo_from_virt_Installed(void * context) = 0;
			      
			virtual void after_evRecvPathTearOk_from_virt_Installed(void * context) = 0;
			              
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_LspRebuild_t LspRebuild(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPath_t SendPath(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			    
			virtual void after_evRecvPathKo_from_virt_Down(void * context) = 0;
			    
			virtual void after_evRecvPathTimeout_from_virt_PathReceived(void * context) = 0;
			                        
			virtual void after_evSendPathTear_from_virt_PathReceived(void * context) = 0;
			    
			virtual void after_evXConnErr_from_virt_PathReceived(void * context) = 0;
			                  
			virtual void after_evSendResvTear_from_virt_PathReceived(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_PathReceived(void * context) = 0;
			        
			virtual void after_evRecvPathErrOk_from_virt_PathReceived(void * context) = 0;
			                
			virtual void after_evRecvPathTimeout_from_virt_TearDown(void * context) = 0;
			            
			virtual void after_evSendPathTear_from_virt_TearDown(void * context) = 0;
			            
			virtual void after_evRecvResvTimeout_from_virt_TearDown(void * context) = 0;
			        
			virtual void after_evRecvResvTearOk_from_virt_TearDown(void * context) = 0;
			      
			virtual void after_evSendResvTear_from_virt_TearDown(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_TearDown(void * context) = 0;
			          
			virtual void after_evRecvPathErrOk_from_virt_TearDown(void * context) = 0;
			                      
			virtual void after_evSendPathTear_from_virt_RebuildTearDown(void * context) = 0;
			                    
			virtual void after_evRecvResvTearOk_from_virt_RebuildTearDown(void * context) = 0;
			    
			virtual void after_evSendResvTear_from_virt_RebuildTearDown(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_RebuildTearDown(void * context) = 0;
			          
			virtual void after_evRecvPathErrOk_from_virt_RebuildTearDown(void * context) = 0;
			          
			virtual void after_evRecvPathTimeout_from_virt_WaitResv(void * context) = 0;
			                    
			virtual void after_evSendPathTear_from_virt_WaitResv(void * context) = 0;
			                
			virtual void after_evRecvResvVeryKo_from_virt_WaitResv(void * context) = 0;
			          
			virtual void after_evRecvResvTearOk_from_virt_WaitResv(void * context) = 0;
			  
			virtual void after_evRecvPathAlarm_from_virt_WaitResv(void * context) = 0;
			    
			virtual void after_evSendResvTear_from_virt_WaitResv(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_WaitResv(void * context) = 0;
			        
			virtual void after_evRecvPathErrOk_from_virt_WaitResv(void * context) = 0;
			                                
			virtual void after_evXConnErr_from_virt_Rebuild(void * context) = 0;
			    
			virtual void after_evRecvPathTearOk_from_virt_Rebuild(void * context) = 0;
			        
			virtual void after_evRecvPathErrOk_from_virt_Rebuild(void * context) = 0;
			         
		};
		
		class PathReceived : public State {
		public:
			PathReceived() :
				State(std::string("PathReceived")) { }
			virtual ~PathReceived() { }
			                                                                                                                            
			virtual void after_evXConnErr_from_virt_WaitEqptConf(void * context) = 0;
			                                                                                                                                                                        
			virtual void after_evSendPath_from_virt_Down(void * context) = 0;
			    
			virtual void after_evRecvPathOk_from_virt_Down(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimeout_t RecvPathTimeout(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t RecvResv(void* context) = 0;
			  
			virtual void after_evRecvResvKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t XConnCompleted(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			  
			virtual void after_evRecvPathOk_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t XConnErr(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			    
			virtual void after_evRecvPathDownKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual void after_evRecvPathErrCrankback_from_virt_PathReceived(void * context) = 0;
			  
			virtual void after_evRecvPathKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual void after_evRecvPathTearKo_from_virt_PathReceived(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t RecvPathTimer(void* context) = 0;
			  
			virtual void after_evRecvPathTimer_from_virt_PathReceived(void * context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResv_t SendResv(void* context) = 0;
			                                                                                                                                                                          
			virtual void after_evRecvPathErrCrankback_from_virt_WaitResv(void * context) = 0;
			                                                                           
		};
		
		class TearDown : public State {
		public:
			TearDown() :
				State(std::string("TearDown")) { }
			virtual ~TearDown() { }
			    
			virtual void after_evSendPathDown_from_virt_WaitResvConf(void * context) = 0;
			                                        
			virtual void after_evRecvPathDownOk_from_virt_WaitResvConf(void * context) = 0;
			                                          
			virtual void after_evSendResvDown_from_virt_WaitResvConf(void * context) = 0;
			  
			virtual void after_evRecvResvDownOk_from_virt_WaitResvConf(void * context) = 0;
			        
			virtual void after_evSendPathDown_from_virt_WaitEqptConf(void * context) = 0;
			                                    
			virtual void after_evRecvPathDownOk_from_virt_WaitEqptConf(void * context) = 0;
			                                            
			virtual void after_evSendResvDown_from_virt_WaitEqptConf(void * context) = 0;
			  
			virtual void after_evRecvResvDownOk_from_virt_WaitEqptConf(void * context) = 0;
			    
			virtual void after_evSendPathDown_from_virt_Installed(void * context) = 0;
			                                                    
			virtual void after_evRecvResvDownOk_from_virt_Installed(void * context) = 0;
			                              
			virtual void after_evRecvPathDownOk_from_virt_Installed(void * context) = 0;
			                    
			virtual void after_evSendResvDown_from_virt_Installed(void * context) = 0;
			                      
			virtual void after_evSendPathDown_from_virt_PathReceived(void * context) = 0;
			                                
			virtual void after_evRecvPathDownOk_from_virt_PathReceived(void * context) = 0;
			                            
			virtual void after_evSendResvDown_from_virt_PathReceived(void * context) = 0;
			  
			virtual void after_evRecvResvDownOk_from_virt_PathReceived(void * context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimeout_t RecvPathTimeout(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			  
			virtual void after_evSendPathDown_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			  
			virtual void after_evRecvPathDownOk_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTimeout_t RecvResvTimeout(void* context) = 0;
			    
			virtual void after_evRecvPathDownKo_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			  
			virtual void after_evRecvResvTearKo_from_virt_TearDown(void * context) = 0;
			    
			virtual void after_evRecvPathAlarm_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t RecvResvErr(void* context) = 0;
			  
			virtual void after_evRecvResvErrKo_from_virt_TearDown(void * context) = 0;
			  
			virtual void after_evRecvPathTearKo_from_virt_TearDown(void * context) = 0;
			  
			virtual void after_evRecvResvErrOk_from_virt_TearDown(void * context) = 0;
			    
			virtual void after_evRecvResvDownOk_from_virt_TearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			  
			virtual void after_evSendResvDown_from_virt_TearDown(void * context) = 0;
			                                                                    
			virtual void after_evSendPathDown_from_virt_WaitResv(void * context) = 0;
			                        
			virtual void after_evRecvPathDownOk_from_virt_WaitResv(void * context) = 0;
			                                          
			virtual void after_evSendResvDown_from_virt_WaitResv(void * context) = 0;
			  
			virtual void after_evRecvResvDownOk_from_virt_WaitResv(void * context) = 0;
			                                             
		};
		
		class RebuildTearDown : public State {
		public:
			RebuildTearDown() :
				State(std::string("RebuildTearDown")) { }
			virtual ~RebuildTearDown() { }
			                                                                                                                                                                                                                                                                                                                                                                                                                                                    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			  
			virtual void after_evSendPathDown_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t XConnCompleted(void* context) = 0;
			  
			virtual void after_evXConnCompleted_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t XConnErr(void* context) = 0;
			  
			virtual void after_evXConnErr_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			  
			virtual void after_evRecvPathDownOk_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual void after_evRecvPathDownKo_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			  
			virtual void after_evRecvResvTearKo_from_virt_RebuildTearDown(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvErr_t RecvResvErr(void* context) = 0;
			  
			virtual void after_evRecvResvErrKo_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual void after_evRecvPathTearKo_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual void after_evRecvResvErrOk_from_virt_RebuildTearDown(void * context) = 0;
			    
			virtual void after_evRecvResvDownOk_from_virt_RebuildTearDown(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			  
			virtual void after_evSendResvDown_from_virt_RebuildTearDown(void * context) = 0;
			                                                                                        
			virtual void after_evSendPathDown_from_virt_Rebuild(void * context) = 0;
			                            
			virtual void after_evSendResvDown_from_virt_Rebuild(void * context) = 0;
			  
			virtual void after_evRecvPathDownOk_from_virt_Rebuild(void * context) = 0;
			  
			virtual void after_evRecvResvDownOk_from_virt_Rebuild(void * context) = 0;
			 
		};
		
		class WaitResv : public State {
		public:
			WaitResv() :
				State(std::string("WaitResv")) { }
			virtual ~WaitResv() { }
			                    
			virtual void after_evRecvResvTimer_from_virt_WaitResvConf(void * context) = 0;
			                                                          
			virtual void after_evRecvPathTimer_from_virt_WaitResvConf(void * context) = 0;
			                                                                                                                                                                                                                                            
			virtual void after_evXConnCompleted_from_virt_PathReceived(void * context) = 0;
			                                                                                                                                                                                      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimeout_t RecvPathTimeout(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResv_t RecvResv(void* context) = 0;
			  
			virtual void after_evRecvResvKo_from_virt_WaitResv(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_WaitResv(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPath_t RecvPath(void* context) = 0;
			  
			virtual void after_evRecvPathOk_from_virt_WaitResv(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathTear_t SendPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_WaitResv(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnPreempt_t XConnPreempt(void* context) = 0;
			  
			virtual void after_evXConnPreempt_from_virt_WaitResv(void * context) = 0;
			  
			virtual void after_evRecvPathDownKo_from_virt_WaitResv(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvTear_t RecvResvTear(void* context) = 0;
			  
			virtual void after_evRecvResvTearKo_from_virt_WaitResv(void * context) = 0;
			    
			virtual void after_evRecvPathKo_from_virt_WaitResv(void * context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvTear_t SendResvTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual void after_evRecvPathTearKo_from_virt_WaitResv(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTimer_t RecvPathTimer(void* context) = 0;
			  
			virtual void after_evRecvPathTimer_from_virt_WaitResv(void * context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			      
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResv_t SendResv(void* context) = 0;
			                                           
		};
		
		class Rebuild : public State {
		public:
			Rebuild() :
				State(std::string("Rebuild")) { }
			virtual ~Rebuild() { }
			                                                                                                                                                                                                                                                                                                
			virtual void after_evLspRebuild_from_virt_Down(void * context) = 0;
			                                                                                                                                                                                                                                                                                                
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathDown_t RecvPathDown(void* context) = 0;
			  
			virtual void after_evRecvPathDownKo_from_virt_Rebuild(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendPathDown_t SendPathDown(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnCompleted_t XConnCompleted(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathErr_t RecvPathErr(void* context) = 0;
			  
			virtual void after_evRecvPathErrKo_from_virt_Rebuild(void * context) = 0;
			  
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_XConnErr_t XConnErr(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvPathTear_t RecvPathTear(void* context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_RecvResvDown_t RecvResvDown(void* context) = 0;
			  
			virtual void after_evRecvResvDownKo_from_virt_Rebuild(void * context) = 0;
			  
			virtual void after_evRecvPathTearKo_from_virt_Rebuild(void * context) = 0;
			    
			virtual fsm::base_G2RSVPTE_LSP_FSM::nextEvFor_SendResvDown_t SendResvDown(void* context) = 0;
			       
		};
		
		class virt_WaitResvConf : public State {
		public:
			virt_WaitResvConf() :
				State(std::string("virt_WaitResvConf")) { }
			virtual ~virt_WaitResvConf() { }
			  
			virtual void after_SendPathDown_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_RecvResv_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvResvKo(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_RecvPath_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual void after_RecvResvTimer_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvResvTimer(void* context) = 0;
			  
			virtual void after_SendPathTear_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_SendConfirm_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evSendConfirm(void* context) = 0;
			  
			virtual void after_RecvConfirm_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvConfirmOk(void* context) = 0;
			  
			virtual void after_XConnPreempt_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evXConnPreempt(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual bool evRecvConfirmKo(void* context) = 0;
			  
			virtual bool evRecvPathAlarm(void* context) = 0;
			  
			virtual void after_SendResvTear_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvResvErr_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvResvErrKo(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual bool evRecvResvErrOk(void* context) = 0;
			  
			virtual void after_RecvPathTimer_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evRecvPathTimer(void* context) = 0;
			  
			virtual bool evRecvResvOk(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_WaitResvConf(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               
		};
		
		class virt_WaitEqptConf : public State {
		public:
			virt_WaitEqptConf() :
				State(std::string("virt_WaitEqptConf")) { }
			virtual ~virt_WaitEqptConf() { }
			                                                                                          
			virtual void after_RecvPathTimeout_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathTimeout(void* context) = 0;
			  
			virtual void after_SendPathDown_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_RecvResv_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvKo(void* context) = 0;
			  
			virtual void after_XConnCompleted_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evXConnCompleted(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_RecvPath_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual void after_RecvResvTimer_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvTimer(void* context) = 0;
			  
			virtual void after_SendPathTear_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_XConnErr_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evXConnErr(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_RecvResvTimeout_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvTimeout(void* context) = 0;
			  
			virtual bool evRecvResvVeryKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual void after_SendResvTear_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual void after_RecvResvErr_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvResvErrKo(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual bool evRecvResvErrOk(void* context) = 0;
			  
			virtual void after_RecvPathTimer_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evRecvPathTimer(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual bool evRecvResvOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_WaitEqptConf(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			                                                                                                                                                                                                                                                                                                                                                                                                                                                     
		};
		
		class virt_Installed : public State {
		public:
			virt_Installed() :
				State(std::string("virt_Installed")) { }
			virtual ~virt_Installed() { }
			                                                                                                                                                                                    
			virtual void after_SendPathDown_from_Installed(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_XConnErr_from_Installed(void * context) = 0;
			  
			virtual bool evXConnErr(void* context) = 0;
			  
			virtual void after_RecvConfirm_from_Installed(void * context) = 0;
			  
			virtual bool evRecvConfirmOk(void* context) = 0;
			  
			virtual void after_XConnPreempt_from_Installed(void * context) = 0;
			  
			virtual bool evXConnPreempt(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_Installed(void * context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_XConnDown_from_Installed(void * context) = 0;
			  
			virtual bool evXConnDown(void* context) = 0;
			  
			virtual void after_RecvPath_from_Installed(void * context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_Installed(void * context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_Installed(void * context) = 0;
			  
			virtual bool evRecvPathAlarm(void* context) = 0;
			  
			virtual bool evRecvConfirmKo(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_Installed(void * context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual void after_RecvPathTimer_from_Installed(void * context) = 0;
			  
			virtual bool evRecvPathTimer(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual void after_RecvResv_from_Installed(void * context) = 0;
			  
			virtual bool evRecvResvOk(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_Installed(void * context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			  
			virtual void after_RecvNotify_from_Installed(void * context) = 0;
			  
			virtual bool evRecvNotifyOk(void* context) = 0;
			  
			virtual bool evRecvResvKo(void* context) = 0;
			  
			virtual bool evRecvNotifyDown(void* context) = 0;
			  
			virtual void after_XConnCompleted_from_Installed(void * context) = 0;
			  
			virtual bool evXConnCompleted(void* context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual bool evRecvActivateErr(void* context) = 0;
			  
			virtual void after_RecvResvTimer_from_Installed(void * context) = 0;
			  
			virtual bool evRecvResvTimer(void* context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvActivate_from_Installed(void * context) = 0;
			  
			virtual bool evRecvActivateOk(void* context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvResvVeryKo(void* context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvNotifyKo(void* context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual void after_RecvResvErr_from_Installed(void * context) = 0;
			  
			virtual bool evRecvResvErrKo(void* context) = 0;
			  
			virtual bool evRecvActivateKo(void* context) = 0;
			  
			virtual bool evRecvResvErrOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_Installed(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			                                                                                                                                                                                                                                                                                                                                           
		};
		
		class virt_Down : public State {
		public:
			virt_Down() :
				State(std::string("virt_Down")) { }
			virtual ~virt_Down() { }
			                                                                                                                                                                                                                                                                                              
			virtual void after_LspRebuild_from_Down(void * context) = 0;
			  
			virtual bool evLspRebuild(void* context) = 0;
			  
			virtual void after_SendPath_from_Down(void * context) = 0;
			  
			virtual bool evSendPath(void* context) = 0;
			  
			virtual void after_RecvPath_from_Down(void * context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			                                                                                                                                                                                                                                                                                                                             
		};
		
		class virt_PathReceived : public State {
		public:
			virt_PathReceived() :
				State(std::string("virt_PathReceived")) { }
			virtual ~virt_PathReceived() { }
			                                                                                                                                                                                                                                                                                                            
			virtual void after_RecvPathTimeout_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathTimeout(void* context) = 0;
			  
			virtual void after_SendPathDown_from_PathReceived(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_RecvResv_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvResvKo(void* context) = 0;
			  
			virtual void after_XConnCompleted_from_PathReceived(void * context) = 0;
			  
			virtual bool evXConnCompleted(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_RecvPath_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual void after_SendPathTear_from_PathReceived(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_XConnErr_from_PathReceived(void * context) = 0;
			  
			virtual bool evXConnErr(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual bool evRecvPathErrCrankback(void* context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			  
			virtual void after_SendResvTear_from_PathReceived(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual void after_RecvPathTimer_from_PathReceived(void * context) = 0;
			  
			virtual bool evRecvPathTimer(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual bool evRecvResvOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_PathReceived(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			  
			virtual void after_SendResv_from_PathReceived(void * context) = 0;
			  
			virtual bool evSendResv(void* context) = 0;
			                                                                                                                                                                                                                                                   
		};
		
		class virt_TearDown : public State {
		public:
			virt_TearDown() :
				State(std::string("virt_TearDown")) { }
			virtual ~virt_TearDown() { }
			                                                                                                                                                                                                                                                                                                                                                                                      
			virtual void after_RecvPathTimeout_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvPathTimeout(void* context) = 0;
			  
			virtual void after_SendPathDown_from_TearDown(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_SendPathTear_from_TearDown(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual void after_RecvResvTimeout_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvResvTimeout(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual bool evRecvPathAlarm(void* context) = 0;
			  
			virtual void after_SendResvTear_from_TearDown(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual void after_RecvResvErr_from_TearDown(void * context) = 0;
			  
			virtual bool evRecvResvErrKo(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual bool evRecvResvErrOk(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_TearDown(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			                                                                                                                                                                                     
		};
		
		class virt_RebuildTearDown : public State {
		public:
			virt_RebuildTearDown() :
				State(std::string("virt_RebuildTearDown")) { }
			virtual ~virt_RebuildTearDown() { }
			                                                                                                                                                                                                                                                                                                                                                                                                                                                    
			virtual void after_SendPathDown_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_XConnCompleted_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evXConnCompleted(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_SendPathTear_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_XConnErr_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evXConnErr(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual void after_SendResvTear_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual void after_RecvResvErr_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evRecvResvErrKo(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual bool evRecvResvErrOk(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_RebuildTearDown(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			                                                                                                                         
		};
		
		class virt_WaitResv : public State {
		public:
			virt_WaitResv() :
				State(std::string("virt_WaitResv")) { }
			virtual ~virt_WaitResv() { }
			                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
			virtual void after_RecvPathTimeout_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathTimeout(void* context) = 0;
			  
			virtual void after_SendPathDown_from_WaitResv(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_RecvResv_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvResvKo(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_RecvPath_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathOk(void* context) = 0;
			  
			virtual void after_SendPathTear_from_WaitResv(void * context) = 0;
			  
			virtual bool evSendPathTear(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual void after_RecvPathDown_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual void after_XConnPreempt_from_WaitResv(void * context) = 0;
			  
			virtual bool evXConnPreempt(void* context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual bool evRecvResvVeryKo(void* context) = 0;
			  
			virtual void after_RecvResvTear_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvResvTearKo(void* context) = 0;
			  
			virtual bool evRecvPathErrCrankback(void* context) = 0;
			  
			virtual bool evRecvPathKo(void* context) = 0;
			  
			virtual bool evRecvResvTearOk(void* context) = 0;
			  
			virtual bool evRecvPathAlarm(void* context) = 0;
			  
			virtual void after_SendResvTear_from_WaitResv(void * context) = 0;
			  
			virtual bool evSendResvTear(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual void after_RecvPathTimer_from_WaitResv(void * context) = 0;
			  
			virtual bool evRecvPathTimer(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual bool evRecvResvOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_WaitResv(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			  
			virtual void after_SendResv_from_WaitResv(void * context) = 0;
			  
			virtual bool evSendResv(void* context) = 0;
			                                         
		};
		
		class virt_Rebuild : public State {
		public:
			virt_Rebuild() :
				State(std::string("virt_Rebuild")) { }
			virtual ~virt_Rebuild() { }
			                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
			virtual void after_RecvPathDown_from_Rebuild(void * context) = 0;
			  
			virtual bool evRecvPathDownKo(void* context) = 0;
			  
			virtual void after_SendPathDown_from_Rebuild(void * context) = 0;
			  
			virtual bool evSendPathDown(void* context) = 0;
			  
			virtual void after_XConnCompleted_from_Rebuild(void * context) = 0;
			  
			virtual bool evXConnCompleted(void* context) = 0;
			  
			virtual void after_RecvPathErr_from_Rebuild(void * context) = 0;
			  
			virtual bool evRecvPathErrKo(void* context) = 0;
			  
			virtual void after_XConnErr_from_Rebuild(void * context) = 0;
			  
			virtual bool evXConnErr(void* context) = 0;
			  
			virtual void after_RecvPathTear_from_Rebuild(void * context) = 0;
			  
			virtual bool evRecvPathTearOk(void* context) = 0;
			  
			virtual void after_RecvResvDown_from_Rebuild(void * context) = 0;
			  
			virtual bool evRecvResvDownKo(void* context) = 0;
			  
			virtual bool evRecvPathTearKo(void* context) = 0;
			  
			virtual bool evRecvPathErrOk(void* context) = 0;
			  
			virtual void after_SendResvDown_from_Rebuild(void * context) = 0;
			  
			virtual bool evSendResvDown(void* context) = 0;
			  
			virtual bool evRecvPathDownOk(void* context) = 0;
			  
			virtual bool evRecvResvDownOk(void* context) = 0;
			 
		};
		
		/*
		 *  Classes that MUST be derived!!! - END
		 */

		class BaseFSM {
		public:
			enum traceLevel_t {
				TRACE_DBG = 0,
				TRACE_LOG,
				TRACE_INF,
				TRACE_WRN,
				TRACE_ERR
			};

			BaseFSM(traceLevel_t level = TRACE_DBG) :
				level_(level),
				name_(std::string("G2RSVPTE_LSP_FSM")) { }
			virtual ~BaseFSM(void) { }

			std::string name(void) { return name_; }

			traceLevel_t traceLevel(void) {	return level_; }

			void traceLevel(traceLevel_t level) { level_ = level; }

			void dbg(std::string text) {
				DUMP(level_, TRACE_DBG, name_, text);
			}

			void log(std::string text) {
				DUMP(level_, TRACE_LOG, name_, text);
			}

			void inf(std::string text) {
				DUMP(level_, TRACE_INF, name_, text);
			}

			void wrn(std::string text) {
				DUMP(level_, TRACE_WRN, name_, text);
			}

			void err(std::string text) {
				DUMP(level_, TRACE_ERR, name_, text);
			}

		private:
			traceLevel_t level_;
		protected:
			std::string  name_;
		};

		/*
		 *  Class for checking FSM integrity ... not FULL IMPLEMENTED
		 */
		class GenericFSM : public BaseFSM {
		public:
			GenericFSM(traceLevel_t level = TRACE_DBG) :
				BaseFSM(level),
				changeInProgress_(false) { }
			virtual ~GenericFSM(void) { }

			bool startModify(void);
			bool endModify(void);

			//TODO: callback????
			typedef void (* callback_t) (std::string from_state,
						     std::string to_state,
						     std::string on_event,
						     void *      context);

			//// CallBack
			//bool addPreCallBack(callback_t cb,
			//		   void *     context);
			//bool addPostCallback(callback_t cb,
			//		    void *     context);
			//bool addInCallback(callback_t cb,
			//		  void * context);
			//
			//bool remPreCallBack(callback_t cb);
			//bool remPostCallback(callback_t cb);
			//bool remInCallback(callback_t cb);


			// States
			bool addState(std::string state);
			bool remState(std::string state);

			// Events
			bool addEvent(std::string event);
			bool remEvent(std::string event);

			// Transitions
			bool addTransition(std::string from,
					   std::string to,
					   std::string event);
			bool remTransition(std::string from,
					   std::string to,
					   std::string event);

			// General
			bool setStartState(std::string state);

			//bool run(std::string event);
			//std::string getState(void);

			//operator<<
			//operator>>

		private:
			bool check(void);

			struct state_data_t {
				callback_t  pre;
				callback_t  post;
				callback_t  in;
			};

			Matrix<std::string,
			       std::string,
			       std::string>                 transitions_;
			std::map<callback_t, void *>        contexts_;
			std::map<std::string, state_data_t> states_;
			std::list<std::string>              events_;
			bool                                changeInProgress_;
			std::string                         startState_;
		};

		class Fsm : public GenericFSM {
		public:
			/*  add/rem of states/events/callback for
			 *  checking FSM consistency
			 */
			Fsm(traceLevel_t level = TRACE_DBG)
				throw(std::string);

			virtual ~Fsm(void);

			friend std::ostream& operator<<(std::ostream&  s,
							const Fsm& f);

			
			bool evRecvResvTearOk(void * context);
			
			bool evRecvResvTearKo(void * context);
			
			bool evRecvPathTimer(void * context);
			
			bool evXConnPreempt(void * context);
			
			bool evXConnErr(void * context);
			
			bool evRecvPathDownOk(void * context);
			
			bool evRecvPathDownKo(void * context);
			
			bool evXConnCompleted(void * context);
			
			bool evRecvResvOk(void * context);
			
			bool evRecvResvKo(void * context);
			
			bool evRecvResvVeryKo(void * context);
			
			bool evRecvPathTimeout(void * context);
			
			bool evRecvResvErrOk(void * context);
			
			bool evRecvResvErrKo(void * context);
			
			bool evRecvResvDownOk(void * context);
			
			bool evRecvResvDownKo(void * context);
			
			bool evRecvActivateOk(void * context);
			
			bool evRecvActivateKo(void * context);
			
			bool evSendResvDown(void * context);
			
			bool evSendPathDown(void * context);
			
			bool evSendResv(void * context);
			
			bool evRecvConfirmOk(void * context);
			
			bool evRecvConfirmKo(void * context);
			
			bool evRecvPathTearOk(void * context);
			
			bool evRecvPathTearKo(void * context);
			
			bool evSendConfirm(void * context);
			
			bool evSendPath(void * context);
			
			bool evLspRebuild(void * context);
			
			bool evSendPathTear(void * context);
			
			bool evRecvResvTimeout(void * context);
			
			bool evSendResvTear(void * context);
			
			bool evXConnDown(void * context);
			
			bool evRecvPathOk(void * context);
			
			bool evRecvPathKo(void * context);
			
			bool evRecvNotifyOk(void * context);
			
			bool evRecvNotifyDown(void * context);
			
			bool evRecvNotifyKo(void * context);
			
			bool evRecvPathErrOk(void * context);
			
			bool evRecvPathAlarm(void * context);
			
			bool evRecvPathErrCrankback(void * context);
			
			bool evRecvActivateErr(void * context);
			
			bool evRecvPathErrKo(void * context);
			
			bool evRecvResvTimer(void * context);
			
			nextEvFor_RecvResvTear_t RecvResvTear(void * context);
			
			nextEvFor_RecvPathTimer_t RecvPathTimer(void * context);
			
			nextEvFor_XConnPreempt_t XConnPreempt(void * context);
			
			nextEvFor_XConnErr_t XConnErr(void * context);
			
			nextEvFor_RecvPathDown_t RecvPathDown(void * context);
			
			nextEvFor_XConnCompleted_t XConnCompleted(void * context);
			
			nextEvFor_RecvResv_t RecvResv(void * context);
			
			nextEvFor_RecvPathTimeout_t RecvPathTimeout(void * context);
			
			nextEvFor_RecvResvErr_t RecvResvErr(void * context);
			
			nextEvFor_RecvResvDown_t RecvResvDown(void * context);
			
			nextEvFor_RecvActivate_t RecvActivate(void * context);
			
			nextEvFor_SendResvDown_t SendResvDown(void * context);
			
			nextEvFor_SendPathDown_t SendPathDown(void * context);
			
			nextEvFor_SendResv_t SendResv(void * context);
			
			nextEvFor_RecvConfirm_t RecvConfirm(void * context);
			
			nextEvFor_RecvPathTear_t RecvPathTear(void * context);
			
			nextEvFor_SendConfirm_t SendConfirm(void * context);
			
			nextEvFor_SendPath_t SendPath(void * context);
			
			nextEvFor_LspRebuild_t LspRebuild(void * context);
			
			nextEvFor_SendPathTear_t SendPathTear(void * context);
			
			nextEvFor_RecvResvTimeout_t RecvResvTimeout(void * context);
			
			nextEvFor_SendResvTear_t SendResvTear(void * context);
			
			nextEvFor_XConnDown_t XConnDown(void * context);
			
			nextEvFor_RecvPath_t RecvPath(void * context);
			
			nextEvFor_RecvNotify_t RecvNotify(void * context);
			
			nextEvFor_RecvPathErr_t RecvPathErr(void * context);
			
			nextEvFor_RecvResvTimer_t RecvResvTimer(void * context);
			

			//operator<<
			//operator>>

			State * currentState(void);
			bool    go2prevState(void);

		private:
			enum states_t {
				G2RSVPTE_LSP_FSM_WaitResvConf,
				G2RSVPTE_LSP_FSM_WaitEqptConf,
				G2RSVPTE_LSP_FSM_Installed,
				G2RSVPTE_LSP_FSM_Down,
				G2RSVPTE_LSP_FSM_PathReceived,
				G2RSVPTE_LSP_FSM_TearDown,
				G2RSVPTE_LSP_FSM_RebuildTearDown,
				G2RSVPTE_LSP_FSM_WaitResv,
				G2RSVPTE_LSP_FSM_Rebuild,
				G2RSVPTE_LSP_FSM_virt_WaitResvConf,
				G2RSVPTE_LSP_FSM_virt_WaitEqptConf,
				G2RSVPTE_LSP_FSM_virt_Installed,
				G2RSVPTE_LSP_FSM_virt_Down,
				G2RSVPTE_LSP_FSM_virt_PathReceived,
				G2RSVPTE_LSP_FSM_virt_TearDown,
				G2RSVPTE_LSP_FSM_virt_RebuildTearDown,
				G2RSVPTE_LSP_FSM_virt_WaitResv,
				G2RSVPTE_LSP_FSM_virt_Rebuild,
				
			};

			enum events_t {
				G2RSVPTE_LSP_FSM_evRecvResvTearOk,
				G2RSVPTE_LSP_FSM_evRecvResvTearKo,
				G2RSVPTE_LSP_FSM_evRecvPathTimer,
				G2RSVPTE_LSP_FSM_evXConnPreempt,
				G2RSVPTE_LSP_FSM_evXConnErr,
				G2RSVPTE_LSP_FSM_evRecvPathDownOk,
				G2RSVPTE_LSP_FSM_evRecvPathDownKo,
				G2RSVPTE_LSP_FSM_evXConnCompleted,
				G2RSVPTE_LSP_FSM_evRecvResvOk,
				G2RSVPTE_LSP_FSM_evRecvResvKo,
				G2RSVPTE_LSP_FSM_evRecvResvVeryKo,
				G2RSVPTE_LSP_FSM_evRecvPathTimeout,
				G2RSVPTE_LSP_FSM_evRecvResvErrOk,
				G2RSVPTE_LSP_FSM_evRecvResvErrKo,
				G2RSVPTE_LSP_FSM_evRecvResvDownOk,
				G2RSVPTE_LSP_FSM_evRecvResvDownKo,
				G2RSVPTE_LSP_FSM_evRecvActivateOk,
				G2RSVPTE_LSP_FSM_evRecvActivateKo,
				G2RSVPTE_LSP_FSM_evSendResvDown,
				G2RSVPTE_LSP_FSM_evSendPathDown,
				G2RSVPTE_LSP_FSM_evSendResv,
				G2RSVPTE_LSP_FSM_evRecvConfirmOk,
				G2RSVPTE_LSP_FSM_evRecvConfirmKo,
				G2RSVPTE_LSP_FSM_evRecvPathTearOk,
				G2RSVPTE_LSP_FSM_evRecvPathTearKo,
				G2RSVPTE_LSP_FSM_evSendConfirm,
				G2RSVPTE_LSP_FSM_evSendPath,
				G2RSVPTE_LSP_FSM_evLspRebuild,
				G2RSVPTE_LSP_FSM_evSendPathTear,
				G2RSVPTE_LSP_FSM_evRecvResvTimeout,
				G2RSVPTE_LSP_FSM_evSendResvTear,
				G2RSVPTE_LSP_FSM_evXConnDown,
				G2RSVPTE_LSP_FSM_evRecvPathOk,
				G2RSVPTE_LSP_FSM_evRecvPathKo,
				G2RSVPTE_LSP_FSM_evRecvNotifyOk,
				G2RSVPTE_LSP_FSM_evRecvNotifyDown,
				G2RSVPTE_LSP_FSM_evRecvNotifyKo,
				G2RSVPTE_LSP_FSM_evRecvPathErrOk,
				G2RSVPTE_LSP_FSM_evRecvPathAlarm,
				G2RSVPTE_LSP_FSM_evRecvPathErrCrankback,
				G2RSVPTE_LSP_FSM_evRecvActivateErr,
				G2RSVPTE_LSP_FSM_evRecvPathErrKo,
				G2RSVPTE_LSP_FSM_evRecvResvTimer,
				G2RSVPTE_LSP_FSM_RecvResvTear,
				G2RSVPTE_LSP_FSM_RecvPathTimer,
				G2RSVPTE_LSP_FSM_XConnPreempt,
				G2RSVPTE_LSP_FSM_XConnErr,
				G2RSVPTE_LSP_FSM_RecvPathDown,
				G2RSVPTE_LSP_FSM_XConnCompleted,
				G2RSVPTE_LSP_FSM_RecvResv,
				G2RSVPTE_LSP_FSM_RecvPathTimeout,
				G2RSVPTE_LSP_FSM_RecvResvErr,
				G2RSVPTE_LSP_FSM_RecvResvDown,
				G2RSVPTE_LSP_FSM_RecvActivate,
				G2RSVPTE_LSP_FSM_SendResvDown,
				G2RSVPTE_LSP_FSM_SendPathDown,
				G2RSVPTE_LSP_FSM_SendResv,
				G2RSVPTE_LSP_FSM_RecvConfirm,
				G2RSVPTE_LSP_FSM_RecvPathTear,
				G2RSVPTE_LSP_FSM_SendConfirm,
				G2RSVPTE_LSP_FSM_SendPath,
				G2RSVPTE_LSP_FSM_LspRebuild,
				G2RSVPTE_LSP_FSM_SendPathTear,
				G2RSVPTE_LSP_FSM_RecvResvTimeout,
				G2RSVPTE_LSP_FSM_SendResvTear,
				G2RSVPTE_LSP_FSM_XConnDown,
				G2RSVPTE_LSP_FSM_RecvPath,
				G2RSVPTE_LSP_FSM_RecvNotify,
				G2RSVPTE_LSP_FSM_RecvPathErr,
				G2RSVPTE_LSP_FSM_RecvResvTimer,
				
			};

			friend std::ostream& operator<<(std::ostream&   s,
							const states_t& st);

			friend std::ostream& operator<<(std::ostream&   s,
							const events_t& ev);

			states_t                             currentState_;
			states_t                             prevState_;
			std::map<states_t, State *>          states_;
			Matrix<states_t, events_t, states_t> nextState_;
		};
	}
#endif // NAMESPACE_G2RSVPTE_LSP_FSM
	



	/*************************************/
	/*   Finite State Machine - Wrapper  */
	/*************************************/
	
	

#ifndef NAMESPACE_G2RSVPTE_LSP_FSM
#define NAMESPACE_G2RSVPTE_LSP_FSM

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string>
#include <map>
#include <list>
#include <iostream>

	namespace G2RSVPTE_LSP_FSM {

		class virtFsm {
		public:
			virtFsm(base_G2RSVPTE_LSP_FSM::BaseFSM::traceLevel_t
				level =	base_G2RSVPTE_LSP_FSM::BaseFSM::TRACE_DBG)
				throw(std::string);

			virtual ~virtFsm(void);

			friend std::ostream& operator<<(std::ostream&  s,
							const virtFsm& f);

			enum root_events_t {
				//G2RSVPTE_LSP_FSM_invalid_event,
				G2RSVPTE_LSP_FSM_RecvResvTear,
				G2RSVPTE_LSP_FSM_RecvPathTimer,
				G2RSVPTE_LSP_FSM_XConnPreempt,
				G2RSVPTE_LSP_FSM_XConnErr,
				G2RSVPTE_LSP_FSM_RecvPathDown,
				G2RSVPTE_LSP_FSM_XConnCompleted,
				G2RSVPTE_LSP_FSM_RecvResv,
				G2RSVPTE_LSP_FSM_RecvPathTimeout,
				G2RSVPTE_LSP_FSM_RecvResvErr,
				G2RSVPTE_LSP_FSM_RecvResvDown,
				G2RSVPTE_LSP_FSM_RecvActivate,
				G2RSVPTE_LSP_FSM_SendResvDown,
				G2RSVPTE_LSP_FSM_SendPathDown,
				G2RSVPTE_LSP_FSM_SendResv,
				G2RSVPTE_LSP_FSM_RecvConfirm,
				G2RSVPTE_LSP_FSM_RecvPathTear,
				G2RSVPTE_LSP_FSM_SendConfirm,
				G2RSVPTE_LSP_FSM_SendPath,
				G2RSVPTE_LSP_FSM_LspRebuild,
				G2RSVPTE_LSP_FSM_SendPathTear,
				G2RSVPTE_LSP_FSM_RecvResvTimeout,
				G2RSVPTE_LSP_FSM_SendResvTear,
				G2RSVPTE_LSP_FSM_XConnDown,
				G2RSVPTE_LSP_FSM_RecvPath,
				G2RSVPTE_LSP_FSM_RecvNotify,
				G2RSVPTE_LSP_FSM_RecvPathErr,
				G2RSVPTE_LSP_FSM_RecvResvTimer,
				
			};

			void post(root_events_t ev, void * context, bool enqueue = false);

			std::string currentState(void);

		private:
			void runPendingWork(void);

			
			void RecvResvTear(void * context);
			
			void RecvPathTimer(void * context);
			
			void XConnPreempt(void * context);
			
			void XConnErr(void * context);
			
			void RecvPathDown(void * context);
			
			void XConnCompleted(void * context);
			
			void RecvResv(void * context);
			
			void RecvPathTimeout(void * context);
			
			void RecvResvErr(void * context);
			
			void RecvResvDown(void * context);
			
			void RecvActivate(void * context);
			
			void SendResvDown(void * context);
			
			void SendPathDown(void * context);
			
			void SendResv(void * context);
			
			void RecvConfirm(void * context);
			
			void RecvPathTear(void * context);
			
			void SendConfirm(void * context);
			
			void SendPath(void * context);
			
			void LspRebuild(void * context);
			
			void SendPathTear(void * context);
			
			void RecvResvTimeout(void * context);
			
			void SendResvTear(void * context);
			
			void XConnDown(void * context);
			
			void RecvPath(void * context);
			
			void RecvNotify(void * context);
			
			void RecvPathErr(void * context);
			
			void RecvResvTimer(void * context);
			

			typedef struct {
				root_events_t ev;
				void *        context;
			} data_event_t;

			base_G2RSVPTE_LSP_FSM::Fsm * fsm_;
			std::list<data_event_t *>  events_;
		};

	}
#endif // NAMESPACE_G2RSVPTE_LSP_FSM
	
}

#endif // FSMGEN_H

