//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include "g2rsvpte_config.h"
#include "g2rsvpte_utils.h"
#include "g2rsvpte_messages.h"
#include "g2rsvpte_message_objects.h"

#define RANGE_SET(DST, SRC)	\
	DST.lower_bound = SRC.lower_bound;	\
	DST.upper_bound = SRC.upper_bound;      \
	DST.lb_included = (SRC.SF & 0x2) >> 1;	\
	DST.ub_included = (SRC.SF & 0x1);

#define VERSION_SET(DST, SRC)                           \
	DST.major_revision = (SRC & 0xF000) >> 12;	\
	DST.minor_revision = (SRC & 0x0FC0) >>  6;      \
	DST.build_fix      =  SRC & 0x003F;

namespace RSVP {

	/////////////////////////////////
	//        Message class        //
	/////////////////////////////////

	MemoryCursor & operator >> (MemoryCursor & mc, Message * & msg)
	{
		MemoryCursor mcHdr(mc);
		mcHdr.resize(mc.start(),
			     mc.start() + (Message::msgHdrLength - 1));
		assert(mcHdr.remainingSize() >= Message::msgHdrLength);

		// Reads Message Header
		uint16_t buffLen;
		uint16_t checkSum;
		uint8_t  reserved;

		uint8_t  versFlags;
		uint8_t  ttl;
		uint8_t  type;

		mcHdr >> versFlags;
		mcHdr >> type;
		mcHdr >> checkSum;
		mcHdr >> ttl;
		mcHdr >> reserved;
		mcHdr >> buffLen;

		assert(mcHdr.remainingSize() == 0);

		if (buffLen != mc.remainingSize()) {
			G2RSVPTE_ERR("Remainig size (%d) and read buffer "
				     "length (%d) are different ... exiting",
				     mc.remainingSize(), buffLen);
			throw(std::string("Problem with input buffer length"));
		}

		int len = (mc.stop_ - mc.start_) - (Message::msgHdrLength - 1);
		if (len <= 0) {
			throw(std::string("Message empty ... no objects!"));
		}
		MemoryCursor mcMsg = mcHdr.neighbor(len);

		// Reads Message
		CAST_STREAM2MSG(mcMsg, msg, type);

		assert(msg);

		msg->versFlags_ = versFlags;
		msg->type_      = type;
		msg->ttl_       = ttl;

		return mc;
	}


	MemoryCursor & operator << (MemoryCursor & mc, const Message & msg)
	{
		// Writes Message Header
		MemoryCursor mcHdr(mc);
		mcHdr.resize(mc.start(), mc.start() + (msg.msgHdrLength - 1));

		mcHdr << msg.versFlags_ << msg.type_  << (uint16_t)0 //Checksum
		      << msg.ttl_       << (uint8_t)0 << msg.getLength(); //Res

		assert(mcHdr.remainingSize() == 0);

		int len = (mc.stop_ - mc.start_) - (Message::msgHdrLength - 1);
		if (len <= 0) {
			throw(std::string("Message empty ... no objects!"));
		}
		MemoryCursor mcMsg = mcHdr.neighbor(len);

		// Writes Message
		CAST_MSG2STREAM(mcMsg, msg);

		return mc;
	}

	std::ostream & operator << (std::ostream & os, const Message & msg)
	{
		// Writes Message Header
		os << "Message "   << MSGTYPE2STR(msg.getType()) << std::endl
		   << " version: " << msg.getVersion()           << std::endl
		   << " flags:   " << msg.getFlags()             << std::endl
		   << " sendTTL: " << msg.getTTL()               << std::endl
		   << " length:  " << msg.getLength()            << std::endl;

		CAST_MSG2STREAM(os, msg);
		return os;
	}


	/////////////////////////////////
	//      PathMessage class      //
	/////////////////////////////////
	PathMessage::PathMessage(void) : Message(PATH)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		ero_                 = 0;
		session_attribute_   = 0;
#ifdef CRANKBACK_ON
		lsp_attributes_      = 0;
#endif // CRANKBACK_ON
		admin_status_        = 0;
		protection_          = 0;
		notify_request_      = 0;
		label_set_           = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		label_request_       = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gen_uni_             = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		times_               = 0;

		sender_template_     = 0;
		sender_tspec_        = 0;
		adspec_              = 0;
		rro_                 = 0;
		suggested_label_     = 0;
		recovery_label_      = 0;
		policy_data_         = 0;
		upstream_label_      = 0;
	}

	PathMessage::PathMessage(uint8_t version) : Message(version, PATH)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		ero_                 = 0;
		session_attribute_   = 0;
#ifdef CRANKBACK_ON
		lsp_attributes_      = 0;
#endif // CRANKBACK_ON
		admin_status_        = 0;
		protection_          = 0;
		notify_request_      = 0;
		label_set_           = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		label_request_       = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gen_uni_             = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		times_               = 0;

		sender_template_     = 0;
		sender_tspec_        = 0;
		adspec_              = 0;
		rro_                 = 0;
		suggested_label_     = 0;
		recovery_label_      = 0;
		policy_data_         = 0;
		upstream_label_      = 0;
	}

	PathMessage::PathMessage(uint8_t            version,
				 uint32_t           src_nid,
				 uint32_t           dst_nid,
				 uint32_t           lsp_id,
				 uint32_t           tun_id,
				 uint32_t           ext_tun_id,
				 uint32_t           nodeId,
				 sw_cap_t           swcap,
				 enc_type_t         enctype,
				 uint32_t           bw,
				 gpid_t             gpid,
				 uint32_t           setupPrio,
				 uint32_t           holdingPrio,
#ifdef CRANKBACK_ON
				 crankback_scope_t  cbScope,
#endif // CRANKBACK_ON
				 lsp_rro_mode_t     rroMode,
				 uint32_t           refreshTime,
				 std::string        sessionName,
				 const net_res_spec_t * iRes,
				 const net_res_spec_t * eRes,
				 const call_ident_t   * callId,
				 const times_t          times) : Message(version, PATH)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
#ifdef CRANKBACK_ON
		lsp_attributes_      = 0;
#endif // CRANKBACK_ON
		admin_status_        = 0;
		protection_          = 0;

		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gen_uni_             = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		times_               = 0;

		sender_tspec_        = 0;
		adspec_              = 0;
		rro_                 = 0;
		suggested_label_     = 0;
		recovery_label_      = 0;
		policy_data_         = 0;


		//TODO: VERY TEMPORARY add code below for this obj!!!
		// BIT_SET  GRSVPTE_PPKT_FLAGS_STATIC
		rsvp_hop_ = new IFIDIPv4RsvpHop_Object;
		assert(rsvp_hop_);
		ero_ = new ExplicitRoute_Object;
		assert(ero_);
		label_set_ = new LabelSet_Object;
		assert(label_set_);
		upstream_label_ = new UpstreamLabel_Object;
		assert(upstream_label_);

		if (rroMode != LSP_RRO_OFF && version == 1) {
			rro_ = new RecordedRoute_Object;
			assert(rro_);
		}

		//assert((!iTna && !eTna) || (iTna && eTna));
		assert((!iRes && !eRes) || (iRes && eRes));
		if (iRes && eRes) {
			gen_uni_ = new GeneralisedUni_Object(*iRes, *eRes);
			assert(gen_uni_);
		}


#ifndef G2RSVPTE2CC_DEBUG
		if (callId) {
			char * addr;
			bit8_t addrLen;
			bit8_t addrType;

			if (callId->src_addr.type == IPv4) {
				addrType = 0x01;
				addrLen  = 4;
				addr     = new char[addrLen];
				assert(addr);
				memcpy(addr, (void *) &callId->src_addr.value.ipv4.s_addr, addrLen);
			} else if (callId->src_addr.type == IPv6) {
				addrType = 0x02;
				addrLen  = 16;
				addr     = new char[addrLen];
				assert(addr);
				memcpy(addr, (void *) &callId->src_addr.value.ipv6.s6_addr32[0], addrLen);
			}

			switch (callId->type) {
				case CALLID_OPSPEC:
					call_id_ = new CallIdOpSpec_Object(addrType,
									   addrLen,
									   addr,
									   callId->local_id);
					assert(call_id_);
					break;
				case CALLID_GLOBUNIQ:
					call_id_ = new CallIdGlobal_Object(addrType,
									   0,
									   0,
									   0,
									   addrLen,
									   addr,
									   callId->local_id);
					assert(call_id_);
					break;
				case CALLID_NULL:
					call_id_ = new CallIdNull_Object(addrType,
									 addrLen,
									 addr,
									 callId->local_id);
					assert(call_id_);
					break;
				default:
					break;
			}
		}
#endif // G2RSVPTE2CC_DEBUG

		if (times.start_time && times.end_time) {
			times_ = new Times_Object(times.start_time,
						  times.end_time);
			assert(times_);
		}

		//
		// Preparation of PATH msg - START
		//

		//TODO  flags

		// Sender TSpec
		SenderTSpec_Object * sTSpecTemp;
		sTSpecTemp = new SenderTSpec_Object;
		assert(sTSpecTemp);
		sTSpecTemp->get().msg_version = 0x0;
		sTSpecTemp->get().length = 0x7;
		sTSpecTemp->get().service_header = 0x1;
		sTSpecTemp->get().zero_bit = 0x0;
		sTSpecTemp->get().service_data_length = 0x6;
		sTSpecTemp->get().param_id_127 = 0x7F; // 127
		sTSpecTemp->get().param127_flags = 0x0;
		sTSpecTemp->get().param127_length = 0x5;
		sTSpecTemp->get().token_bucket_rate = 0x0;
		sTSpecTemp->get().token_bucket_size = 0x0;
		sTSpecTemp->get().peak_data_rate = bw;
		sTSpecTemp->get().min_policed_unit = 0x14; // 20
		sTSpecTemp->get().max_policed_unit = 0x5DC; // 1500
		sender_tspec_ = sTSpecTemp;

		// Sender Template
		IPV4LSPSenderTemplate_Object * sTemp;
		sTemp = new IPV4LSPSenderTemplate_Object;
		assert(sTemp);
		sTemp->get().ipv4_addr.s_addr = htonl(src_nid);
		sTemp->get().lsp_id           = lsp_id;
		sender_template_ = sTemp;

		// Session
		LSPTunnelIPv4Session_Object * sess;
		sess = new LSPTunnelIPv4Session_Object;
		assert(sess);
		sess->get().ipv4_addr.s_addr = htonl(dst_nid);
		sess->get().tunnelID         = tun_id;
		sess->get().extendedTunnelID = ext_tun_id;
		session_ = sess;

		// LSP_EROP(lsp)->subobjects = NULL;

		// Time Values
		TimeValues_Object * tValue;
		tValue = new TimeValues_Object;
		assert(tValue);
		tValue->get().refresh_period = refreshTime; // = lsp->refresh_intval;
		time_values_ = tValue;

		// Label Request
		GLabelRequest_Object * lReq;
		lReq = new GLabelRequest_Object;
		assert(lReq);
		lReq->get().enc_type    = swcap;
		lReq->get().switch_type = enctype;
		lReq->get().g_pid       = gpid;
		label_request_ = lReq;

		// Session Attribute
		LSPTunnelRASessionAttr_Object * sAttr;
		sAttr = new LSPTunnelRASessionAttr_Object;
		assert(sAttr);

		sAttr->get().setup_prio = setupPrio;
		sAttr->get().hold_prio  = holdingPrio;

		sAttr->get().include_all = 0; //include_all
		sAttr->get().include_any = 0; //include_any
		sAttr->get().exclude_any = 0; //exclude_any

		std::string sessName = sessionName;
		if (sessName.empty()) {
			sessName = std::string("SESS-NAME: LSP XXX");
		}
		int    len  = strlen(sessName.c_str());
		char * name = new char[len];
		assert(name);
		strncpy(name, sessName.c_str(), len);

		sAttr->get().name     = name;
		sAttr->get().name_len = len;

		session_attribute_ = sAttr;
#ifdef CRANKBACK_ON
		// LSP Attributes
		if (cbScope != CB_NONE &&
		    version == 1) {
			LSPAttr_Object * lspAttr;
			lspAttr = new LSPAttr_Object;
			assert(lspAttr);

			LSP_ATTR_TLV * tlv;
			tlv = new LSP_ATTR_TLV();
			assert(tlv);

			tlv->get().flags = 0;

			switch (cbScope) {
				case CB_E2E:
					tlv->get().e2e_cb = 1;
					tlv->get().boundary_cb = 0;
					tlv->get().segment_cb = 0;
					break;
				case CB_BOUNDARY:
					tlv->get().e2e_cb = 0;
					tlv->get().boundary_cb = 1;
					tlv->get().segment_cb = 0;
					break;
				case CB_SEGMENT:
					tlv->get().e2e_cb = 0;
					tlv->get().boundary_cb = 0;
					tlv->get().segment_cb = 1;
					break;
				default:
					G2RSVPTE_ERR("Unknown crankback scope parameter");
					throw(std::string("Unknown crankback scope parameter!"));
			}

			lspAttr->get().tlvs.push_back(tlv);

			lsp_attributes_ = lspAttr;
		}
#endif // CRANKBACK_ON
		// Notify Request //if needed
		IPV4NotifyRequest_Object * nReq;
		nReq = new IPV4NotifyRequest_Object;
		assert(nReq);
		nReq->get().ipv4_addr.s_addr = htonl(nodeId);
		notify_request_ = nReq;

		// ... TODO

		//if (lspinfo->loc_prot_flg)
		//if (BIT_TEST(lspinfo->recroute_flgs, GRSVPTE_RECROUTE_FLAG_LCREC))
		//if (BIT_TEST(lspinfo->recroute_flgs, GRSVPTE_RECROUTE_FLAG_LBLREC))
		//if (lspinfo->se_sty_flg)

		//if (lspinfo->link_prot_mask != 0) {

		//if (GRSVPTE_LSP_ISADOPTED(lsp)) { //set admin_status_

		//
		// Preparation of PATH msg - END
		//
	}

	PathMessage::PathMessage(const PathMessage & path) :
		Message(path.getVersion(), PATH)
	{
		createPathMessage(path, false, false);
	}

	PathMessage::PathMessage(const PathMessage & path, bool down, bool reflect) :
		Message(path.getVersion(), PATH)
	{
		createPathMessage(path, down, reflect);
	}

	void PathMessage::createPathMessage(const PathMessage & path, bool down, bool reflect)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		ero_                 = 0;
		session_attribute_   = 0;
#ifdef CRANKBACK_ON
		lsp_attributes_      = 0;
#endif // CRANKBACK_ON
		admin_status_        = 0;
		protection_          = 0;
		notify_request_      = 0;
		label_set_           = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		label_request_       = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gen_uni_             = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		times_               = 0;

		sender_template_     = 0;
		sender_tspec_        = 0;
		adspec_              = 0;
		rro_                 = 0;
		suggested_label_     = 0;
		recovery_label_      = 0;
		policy_data_         = 0;
		upstream_label_      = 0;

		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!

		//Todo  set flags STATIC

		if (down) {
			admin_status_ = new AdminStatus_Object(reflect,//REFL
							       false,  //TEST
							       true,   //ADM
							       true);  //DEL
		} else if (path.admin_status_) {
			admin_status_ = path.admin_status_->clone();
		} else {
			admin_status_ = new AdminStatus_Object();
		}
		if (!admin_status_) {
			throw(std::string("Impossible to allocate admin"
					  "Status obj for Path msg!"));
		}

		if (path.session_) {
			session_ = path.session_->clone();
			if (!session_) {
				throw(std::string("Impossible to allocate "
						  "session obj for Path msg!"));
			}
		}

		if (path.rsvp_hop_) {
			rsvp_hop_ = path.rsvp_hop_->clone();
			if (!rsvp_hop_) {
				throw(std::string("Impossible to allocate "
						  "rsvpHop obj for Path msg!"));
			}
		}

		if (path.time_values_) {
			time_values_ = path.time_values_->clone();
			if (!time_values_) {
				throw(std::string("Impossible to allocate time "
						  "Values obj for Path msg!"));
			}
		}

		if (path.label_request_) {
			label_request_ = path.label_request_->clone();
			if (!label_request_) {
				throw(std::string("Impossible to allocate label"
						  "Request obj for Path Msg!"));
			}
		}

		if (path.sender_template_) {
			sender_template_ = path.sender_template_->clone();
			if (!sender_template_) {
				throw(std::string("Impossible to allocate send"
						  "erTemp obj for Path msg!"));
			}
		}

		if (path.sender_tspec_) {
			sender_tspec_ = path.sender_tspec_->clone();
			if (!sender_tspec_) {
				throw(std::string("Impossible to allocate send"
						  "erTSpec obj for Path msg!"));
			}
		}

		if (path.call_id_) {
			call_id_ = path.call_id_->clone();
			if (!call_id_) {
				throw(std::string("Impossible to allocate send"
						  "CallId obj for Path msg!"));
			}
		}

		if (path.call_ops_) {
			call_ops_ = path.call_ops_->clone();
			if (!call_ops_) {
				throw(std::string("Impossible to allocate send"
						  "CallOps obj for Path msg!"));
			}
		}

		if (path.gns_call_ext_) {
			gns_call_ext_ = path.gns_call_ext_->clone();
			if (!gns_call_ext_) {
				throw(std::string("Impossible to allocate send"
						  "GnsCallExt obj for Path msg!"));
			}
		}

		if (path.gns_uni_eg_) {
			gns_uni_eg_ = path.gns_uni_eg_->clone();
			if (!gns_uni_eg_) {
				throw(std::string("Impossible to allocate send"
						  "egress GnsUni obj for Path msg!"));
			}
		}

		if (path.gns_uni_in_) {
			gns_uni_in_ = path.gns_uni_in_->clone();
			if (!gns_uni_in_) {
				throw(std::string("Impossible to allocate send"
						  "ingress GnsUni obj for Path msg!"));
			}
		}
	}

	PathMessage::~PathMessage(void)
	{
		G2RSVPTE_DBG("Destroying PathMessage %p ...", this);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(message_id_ack_nack_);
		DELETE_OBJ(message_id_);
		DELETE_OBJ(ero_);
		DELETE_OBJ(session_attribute_);
#ifdef CRANKBACK_ON
		DELETE_OBJ(lsp_attributes_);
#endif // CRANKBACK_ON
		DELETE_OBJ(admin_status_);
		DELETE_OBJ(protection_);
		DELETE_OBJ(notify_request_);
		DELETE_OBJ(label_set_);

		DELETE_OBJ(session_);
		DELETE_OBJ(rsvp_hop_);
		DELETE_OBJ(time_values_);
		DELETE_OBJ(label_request_);
		DELETE_OBJ(call_ops_);
		DELETE_OBJ(call_id_);
		DELETE_OBJ(gns_call_ext_);
		DELETE_OBJ(gen_uni_);
		DELETE_OBJ(gns_uni_eg_);
		DELETE_OBJ(gns_uni_in_);
		DELETE_OBJ(times_);
		DELETE_OBJ(sender_template_);
		DELETE_OBJ(sender_tspec_);
		DELETE_OBJ(adspec_);
		DELETE_OBJ(rro_);
		DELETE_OBJ(suggested_label_);
		DELETE_OBJ(recovery_label_);
		DELETE_OBJ(policy_data_);
		DELETE_OBJ(upstream_label_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	PathMessage * PathMessage::partialClone(void) const
	{
		PathMessage * path = new PathMessage(getVersion());
		if (!path) return 0;

		// TODO flags, error code, actinfo ...

		// Session
		CLONE_OBJ_INTO_MSG(session_, path);

		// Time Values
		CLONE_OBJ_INTO_MSG(time_values_, path);

		// ERO
		CLONE_OBJ_INTO_MSG(ero_, path);

		// Label Request
		CLONE_OBJ_INTO_MSG(label_request_, path);

		// Protection
		CLONE_OBJ_INTO_MSG(protection_, path);

		// Session Attribute
		CLONE_OBJ_INTO_MSG(session_attribute_, path);

#ifdef CRANKBACK_ON
		// LSP Attributes
		CLONE_OBJ_INTO_MSG(lsp_attributes_, path);
#endif // CRANKBACK_ON

		// NOTIFY REQUEST
		CLONE_OBJ_INTO_MSG(notify_request_, path);

		// ADMIN_STATUS
		CLONE_OBJ_INTO_MSG(admin_status_, path);

		// POLICY_DATA
		CLONE_OBJ_INTO_MSG(policy_data_, path);

		// SENDER_TEMPLATE
		CLONE_OBJ_INTO_MSG(sender_template_, path);

		// SENDER_TSPEC
		CLONE_OBJ_INTO_MSG(sender_tspec_, path);

		// ADSPEC
		CLONE_OBJ_INTO_MSG(adspec_, path);

		// RRO
		CLONE_OBJ_INTO_MSG(rro_, path);

		// RESV_CONFIRM
		//CLONE_OBJ_INTO_MSG(resv_confirm_, path);

		// SCOPE
		//CLONE_OBJ_INTO_MSG(scope_, path);

		// STYLE
		//CLONE_OBJ_INTO_MSG(style_, path);

		// FLOW_SPEC & FILTER_SPEC
		//CLONE_OBJ_INTO_MSG(flow_desc_, path);

		// ERROR SPEC
		//CLONE_OBJ_INTO_MSG(error_spec_, path);

		// LSP_ATTRIBUTES ???

		// Generalised UNI -> TODO: check???
		CLONE_OBJ_INTO_MSG(gen_uni_, path);

		// TIMES Object
		CLONE_OBJ_INTO_MSG(times_, path);

		// CALL ID Object
		CLONE_OBJ_INTO_MSG(call_id_, path);

		// CALL OPS Object
		CLONE_OBJ_INTO_MSG(call_ops_, path);

		// GNS CALL EXT Object
		CLONE_OBJ_INTO_MSG(gns_call_ext_, path);

		// egress GNS UNI Object
		CLONE_OBJ_INTO_MSG(gns_uni_eg_, path);

		// ingress GNS UNI Object
		CLONE_OBJ_INTO_MSG(gns_uni_in_, path);

		return path;
	}

	bool PathMessage::setGnsCallExt(const call_info_t & callInfo)
	{
		if (gns_call_ext_) {
			G2RSVPTE_ERR("GnsCallExt_Object already set ... strange !!");
			return false;
		}
		GnsCallExt_Object * gnsCall;
		gnsCall = new GnsCallExt_Object(callInfo);
		if (!gnsCall) {
			G2RSVPTE_ERR("Impossible to allocate new GnsCallExt_Object");
			return false;
		}

		gns_call_ext_ = gnsCall;

		return true;
	}

	bool PathMessage::updateGenUni(const net_res_spec & tna,
				       bool                 source,
				       call_type_t          type)
	{
		if (!gen_uni_) {
			G2RSVPTE_ERR("GeneneralisedUni_Object is not set");
			return false;
		}

		GeneralisedUni_Class * tmp;
		tmp = gen_uni_;

		if (tmp->cType() != CTYPE_GENERALIZED_UNI) {
			G2RSVPTE_ERR("Unsupported Gen UNI object type!");
			return false;
		}

		GeneralisedUni_Object * genUni;
		genUni = static_cast<GeneralisedUni_Object *>(tmp);

		if (source) {
			// Get SRC TNA
			GuniSubObj * subObj;
			subObj = genUni->getSrcTNA();
			if (subObj) {
				delete subObj;
			}
			subObj = createGuniSrcSubObj(tna.tna);
			if (!genUni->insertGuniSubObj(subObj)) {
				G2RSVPTE_ERR("Impossible to insert SRC_TNA");
				return false;
			}
		} else {
			// Get DST TNA
			GuniSubObj * subObj;
			subObj = genUni->getDestTNA();
			if (subObj) {
				delete subObj;
			}
			subObj = createGuniDstSubObj(tna.tna);
			if (!genUni->insertGuniSubObj(subObj)) {
				G2RSVPTE_ERR("Impossible to insert DST_TNA");
				return false;
			}
			if (BITMASK_BITTEST(tna.mask_, data_link) &&
			    BITMASK_BITTEST(tna.mask_, label)) {
				if (type == CALL_TYPE_SPC) {
					subObj = genUni->getSpcLabel();
					if (subObj) {
						delete subObj;
					}
					subObj = createGuniSpcLblSubObj(tna.data_link,
									tna.label);
					if (!genUni->insertGuniSubObj(subObj)) {
						throw(std::string("Impossible to ins "
								  "SPC_LABEL"));
					}
				} else if (type == CALL_TYPE_SC || type == CALL_TYPE_AUTO) {
					subObj = genUni->getEgressLabel();
					if (subObj) {
						delete subObj;
					}
					subObj = createGuniEgrLblSubObj(tna.data_link,
									tna.label);
					if (!genUni->insertGuniSubObj(subObj)) {
						throw(std::string("Impossible to ins "
								  "EGRESS_LABEL"));
					}
				} else {
					G2RSVPTE_ERR("Call type 0x%x not supported", type);
					return false;
				}
			}
		}

		return true;
	}

	bool PathMessage::setGenUni(const net_res_spec & tna,
				    bool                 ingress,
				    call_type_t          type)
	{
		if (!gen_uni_) {
			gen_uni_ = new GeneralisedUni_Object;
			if (!gen_uni_) {
				G2RSVPTE_ERR("Impossible to allocate new GeneralisedUni_Object");
				return false;
			}
		}
		assert(isNetResSpecValid(tna));

		GeneralisedUni_Object * genUni;
		GuniSubObj *            subObj;

		genUni = static_cast<GeneralisedUni_Object *>(gen_uni_);

		switch (ingress) {
			case true:
				subObj = createGuniSrcSubObj(tna.tna);
				if (!genUni->insertGuniSubObj(subObj)) {
					G2RSVPTE_ERR("Impossible to ins SRC_TNA");
					delete gen_uni_;
					gen_uni_ = 0;
					return false;
				}
				break;
			case false:
				subObj = createGuniDstSubObj(tna.tna);
				if (!genUni->insertGuniSubObj(subObj)) {
					G2RSVPTE_ERR("Impossible to ins DST_TNA");
					delete gen_uni_;
					gen_uni_ = 0;
					return false;
				}
				if (BITMASK_BITTEST(tna.mask_, data_link) &&
				    BITMASK_BITTEST(tna.mask_, label)) {
					if (type == CALL_TYPE_SPC) {
						subObj = createGuniSpcLblSubObj(tna.data_link,
										tna.label);
						if (!genUni->insertGuniSubObj(subObj)) {
							G2RSVPTE_ERR("Impossible to ins SPC_LABEL");
							delete gen_uni_;
							gen_uni_ = 0;
							return false;
						}
					} else {
						subObj = createGuniEgrLblSubObj(tna.data_link,
										tna.label);
						if (!genUni->insertGuniSubObj(subObj)) {
							G2RSVPTE_ERR("Impossible to ins EGRESS_LABEL");;
						}
					}
				}
				break;
			default:
				return false;
		}

		return true;
	}

	bool PathMessage::setPolicyData(const std::string aaaInfo)
	{
		if (policy_data_) {
			G2RSVPTE_ERR("PolicyData_Object already set ... strange !!");
			return false;
		}

		Policy_Object * policyData = new Policy_Object();
		if (!policyData) {
			G2RSVPTE_ERR("Impossible to allocate new PolicyData_Object");
			return false;
		}

		char * data = new char[aaaInfo.length()];
		strncpy(data, aaaInfo.c_str(), aaaInfo.length());

		PolicyInfo * pInfo = new PolicyInfo(PTYPE_PHOSPHORUS_AUTHZ_TOKEN, strlen(data), data);
		if (!pInfo) {
			G2RSVPTE_ERR("Impossible to allocate new PolicyInfo");
			delete policyData;
			delete [] data;
			return false;
		}
		delete [] data;

		policyData->get().policy_elems.push_back(pInfo);
		policy_data_ = policyData;

		return true;
	}

	bool PathMessage::getEro(std::list<lsp_ero_sobj_t> & ero)
	{
		if (!ero_) {
			G2RSVPTE_ERR("Ero Object pointer null !!");
			return false;
		}

		ExplicitRoute_Object * eroObj;
		std::list<EroSubObject *>::iterator it;
		lsp_ero_sobj_t eroSubObj;

		eroObj = static_cast<ExplicitRoute_Object *>(ero_);

		it = eroObj->get().subObjs.begin();
		for (it =  eroObj->get().subObjs.begin();
		     it != eroObj->get().subObjs.end();
		     it++) {
			EroSubObject * sObj;
			sObj = *it;

			if (!sObj) {
				G2RSVPTE_ERR("Invalid subObject pointer");
				return false;
			}

			memset(&eroSubObj, 0, sizeof(eroSubObj));

			switch (sObj->type()) {
				case IPV4_PREF_SUBOBJECT: {
					IPv4Pref_SubObj * ipv4;
					ipv4 = static_cast<IPv4Pref_SubObj *>(sObj);

					eroSubObj.type            = ERO_SUBOBJ_IPV4;
					eroSubObj.hop.ipv4.addr   = ipv4->get().ipv4_addr;
					eroSubObj.hop.ipv4.prefix = ipv4->get().preflen;
					eroSubObj.loose           = sObj->looseFlag() ?
						ERO_SUBOBJ_LOOSE : ERO_SUBOBJ_STRICT;

					break;
				}
				case IPV6_PREF_SUBOBJECT: {
					G2RSVPTE_ERR("IPv6 prefix subobjects not yet supported");
					return false;
				}
				case AS_SUBOBJECT: {
					G2RSVPTE_ERR("AS subobjects not yet supported");
					return false;
				}
				case UNNUM_LINK_SUBOBJECT: {
					Unnum_SubObj * unn;
					unn = static_cast<Unnum_SubObj *>(sObj);

					if (unn->get().interface_id == 0) {
						eroSubObj.type             = ERO_SUBOBJ_NODE;
						eroSubObj.hop.node.node_id = unn->get().router_id;
					} else {
						eroSubObj.type                = ERO_SUBOBJ_UNNUM;
						eroSubObj.hop.unnum.router_id = unn->get().router_id;
						eroSubObj.hop.unnum.intf_id   = unn->get().interface_id;
					}
					eroSubObj.loose           = sObj->looseFlag() ?
						ERO_SUBOBJ_LOOSE : ERO_SUBOBJ_STRICT;
					break;
				}
				case LABEL_ERO_SUBOBJECT: {
					LabelERO_SubObj * lbl;
					lbl = static_cast<LabelERO_SubObj *>(sObj);
					if (lbl->get().upstream) {
						G2RSVPTE_ERR("Only downstream Label ERO subobject supported");
						return false;
					}

					RSVP::Label_Class * label = lbl->get().label;
					if (!label ||
					    label->cType() != CTYPE_GENERALISED_LABEL) /* Useless??? */ {
						G2RSVPTE_ERR("Label ERO subobj with invalid pointer!");
						return false;
					}

					RSVP::GLabel_Object * genLabel;
					genLabel = static_cast<RSVP::GLabel_Object *>(label);

					eroSubObj.type            = ERO_SUBOBJ_LABEL;
					eroSubObj.hop.label.value = *(genLabel->get().label);
					eroSubObj.hop.label.upstream = lbl->get().upstream ?
						LABEL_UPSTREAM : LABEL_DOWNSTREAM;
					break;
				}
				case COMP_IF_ID_IPv4_SUBOBJECT:
				case COMP_IF_ID_IPv6_SUBOBJECT:
				case COMP_IF_ID_UNNUM_SUBOBJECT: {
					G2RSVPTE_WRN("ERO Component i/f id usage is not yet supported");
					return false;
				}
				default:
					G2RSVPTE_ERR("unknown subobj type (%d)",
						     sObj->type());
					return false;
			}
			ero.push_back(eroSubObj);
		}
		return true;
	}

	bool PathMessage::setGnsTna(const grid_res_spec_t & gnsTna,
				    bool                    ingress)
	{
		GnsUni_Object * gnsUni;
		gnsUni = new GnsUni_Object;
		if (!gnsUni) {
			G2RSVPTE_ERR("Impossible to allocate new GnsUni_Object");
			return false;
		}

		// Application_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, application)) {
			Application_SubObj * app;
			app = new Application_SubObj(gnsTna.application);
			assert(app);
			gnsUni->get().subObjs.push_back(app);
		}
		// Candidate host SuObj
		if (BITMASK_BITTEST(gnsTna.mask_, cand_host)) {
			switch (gnsTna.cand_host.type) {
				case IPv4:
					// IPv4CandidateHost_SubObj
					IPv4CandidateHost_SubObj * ipv4Host;
					ipv4Host = new IPv4CandidateHost_SubObj(gnsTna.cand_host.value.ipv4);
					assert(ipv4Host);
					gnsUni->get().subObjs.push_back(ipv4Host);
					break;
				case IPv6:
					// IPv4CandidateHost_SubObj
					IPv6CandidateHost_SubObj * ipv6Host;
					ipv6Host = new IPv6CandidateHost_SubObj(gnsTna.cand_host.value.ipv6);
					assert(ipv6Host);
					gnsUni->get().subObjs.push_back(ipv6Host);
					break;
				case NSAP:
					// NsapCandidateHost_SubObj
					NsapCandidateHost_SubObj * nsapHost;
					nsapHost = new NsapCandidateHost_SubObj(gnsTna.cand_host.value.nsap);
					assert(nsapHost);
					gnsUni->get().subObjs.push_back(nsapHost);
					break;
				default:
					G2RSVPTE_ERR("Candidate host address type not supported");
					delete gnsUni;
					return false;
					break;
			}
		}
		// FsResources_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, fs_resources)) {
			FsResources_SubObj * fsRes;
			fsRes = new FsResources_SubObj(gnsTna.fs_resources);
			assert(fsRes);
			gnsUni->get().subObjs.push_back(fsRes);
		}
		// SystemCaps_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, sys_caps)) {
			SystemCaps_SubObj * sysCap;
			sysCap = new SystemCaps_SubObj(gnsTna.sys_caps);
			assert(sysCap);
			gnsUni->get().subObjs.push_back(sysCap);
		}
		// IndvCpuSpeed_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_speed)) {
			IndvCpuSpeed_SubObj * ics;
			ics = new IndvCpuSpeed_SubObj;
			assert(ics);
			ics->get().lower_bound   = gnsTna.ind_cpu_speed.lower_bound;
			ics->get().upper_bound   = gnsTna.ind_cpu_speed.upper_bound;
			ics->get().SF            = (gnsTna.ind_cpu_speed.ub_included << 1) |
				                   (gnsTna.ind_cpu_speed.lb_included);
			gnsUni->get().subObjs.push_back(ics);
		}
		// IndvCpuTime_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_time)) {
			IndvCpuTime_SubObj * ict;
			ict = new IndvCpuTime_SubObj(gnsTna.ind_cpu_time);
			assert(ict);
			gnsUni->get().subObjs.push_back(ict);
		}
		// IndvCpuCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_count)) {
			IndvCpuCount_SubObj * icc;
			icc = new IndvCpuCount_SubObj(gnsTna.ind_cpu_count);
			assert(icc);
			gnsUni->get().subObjs.push_back(icc);
		}
		// IndvNetBw_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_net_bw)) {
			IndvNetBw_SubObj * nbw;
			nbw = new IndvNetBw_SubObj(gnsTna.ind_net_bw);
			assert(nbw);
			gnsUni->get().subObjs.push_back(nbw);
		}
		// IndvPhyMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_phy_mem)) {
			IndvPhyMem_SubObj * ipm;
			ipm = new IndvPhyMem_SubObj(gnsTna.ind_phy_mem);
			assert(ipm);
			gnsUni->get().subObjs.push_back(ipm);
		}
		// IndvVirMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_vir_mem)) {
			IndvVirMem_SubObj * ivm;
			ivm = new IndvVirMem_SubObj(gnsTna.ind_vir_mem);
			assert(ivm);
			gnsUni->get().subObjs.push_back(ivm);
		}
		// IndvDiskSpace_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_disk_space)) {
			IndvDiskSpace_SubObj * ids;
			ids = new IndvDiskSpace_SubObj(gnsTna.ind_disk_space);
			assert(ids);
			gnsUni->get().subObjs.push_back(ids);
		}
		// TotCpuTime_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_cpu_time)) {
			TotCpuTime_SubObj * tct;
			tct = new TotCpuTime_SubObj(gnsTna.tot_cpu_time);
			assert(tct);
			gnsUni->get().subObjs.push_back(tct);
		}
		// TotCpuCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_cpu_count)) {
			TotCpuCount_SubObj * tcc;
			tcc = new TotCpuCount_SubObj(gnsTna.tot_cpu_count);
			assert(tcc);
			gnsUni->get().subObjs.push_back(tcc);
		}
		// TotPhyMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_phy_mem)) {
			TotPhyMem_SubObj * tpm;
			tpm = new TotPhyMem_SubObj(gnsTna.tot_phy_mem);
			assert(tpm);
			gnsUni->get().subObjs.push_back(tpm);
		}
		// TotVirMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_vir_mem)) {
			TotVirMem_SubObj * tvm;
			tvm = new TotVirMem_SubObj(gnsTna.tot_vir_mem);
			assert(tvm);
			gnsUni->get().subObjs.push_back(tvm);
		}
		// TotDiskSpace_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_disk_space)) {
			TotDiskSpace_SubObj * tds;
			tds = new TotDiskSpace_SubObj(gnsTna.tot_disk_space);
			assert(tds);
			gnsUni->get().subObjs.push_back(tds);
		}
		// TotResCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_res_count)) {
			TotResCount_SubObj * trc;
			trc = new TotResCount_SubObj(gnsTna.tot_res_count);
			assert(trc);
			gnsUni->get().subObjs.push_back(trc);
		}
		// DataStaging_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, data_staging)) {
			DataStaging_SubObj * ds;
			ds = new DataStaging_SubObj(gnsTna.data_staging);
			assert(ds);
			gnsUni->get().subObjs.push_back(ds);
		}

		if (ingress) {
			gns_uni_in_ = gnsUni;
		} else {
			gns_uni_eg_ = gnsUni;
		}

		return true;
	}

	bool PathMessage::getGnsTna(grid_res_spec_t & gnsTna,
				    bool              ingress)
	{
		GnsUni_Class *                         gnsUniClass;
		GnsUni_Object *                        gnsUni;
		GnsSubObj *                            subObj;
		std::list<GnsSubObj *>::const_iterator it;

		if (ingress) {
		       gnsUniClass = gns_uni_in_;
		} else {
		       gnsUniClass = gns_uni_eg_;
		}

		if (!gnsUniClass) {
			G2RSVPTE_ERR("Null pointer !");
			return false;
		}
		gnsUni = static_cast<GnsUni_Object *> (gnsUniClass);

		for (it  = gnsUni->get().subObjs.begin();
		     it != gnsUni->get().subObjs.end();
		     it++) {
			subObj = *it;
			switch (subObj->type()) {
				case TYPE_APPLICATION: {
					switch (subObj->subType()) {
						case SUBTYPE_APPLICATION: {
							Application_SubObj * app;
							app = static_cast<Application_SubObj *> (subObj);
							gnsTna.application.type = (grid_appname_t) app->get().app_type;
							VERSION_SET(gnsTna.application.version,
								    app->get().app_version);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, application);
					break;
				}
				case TYPE_CANDIDATE_HOSTS: {
					switch (subObj->subType()) {
						case SUBTYPE_IPv4_CANDIDATE_HOSTS: {
							IPv4CandidateHost_SubObj * ipv4Host;
							ipv4Host = static_cast<IPv4CandidateHost_SubObj *> (subObj);
							gnsTna.cand_host.value.ipv4 = ipv4Host->get().ipv4_addr;
							break;
						}
						case SUBTYPE_IPv6_CANDIDATE_HOSTS: {
							IPv6CandidateHost_SubObj * ipv6Host;
							ipv6Host = static_cast<IPv6CandidateHost_SubObj *> (subObj);
							gnsTna.cand_host.value.ipv6 = ipv6Host->get().ipv6_addr;
							break;
						}
						case SUBTYPE_NSAP_CANDIDATE_HOSTS: {
							NsapCandidateHost_SubObj * nsapHost;
							nsapHost = static_cast<NsapCandidateHost_SubObj *> (subObj);
							memcpy(&gnsTna.cand_host.value.nsap,
							       nsapHost->get().nsap_addr,
							       nsapHost->get().nsap_len);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, cand_host);
					break;
				}
				case TYPE_FILE_SYSTEM_RES: {
					switch (subObj->subType()) {
						case SUBTYPE_FILE_SYSTEM_RES: {
							FsResources_SubObj * fsRes;
							fsRes = static_cast<FsResources_SubObj *> (subObj);
							gnsTna.fs_resources.fs_name = (grid_fs_name_t)fsRes->get().fs_name;
							gnsTna.fs_resources.fs_type = (grid_fs_type_t)fsRes->get().fs_type;
							gnsTna.fs_resources.disk_space.lower_bound =
								fsRes->get().disk_space_lb;
							gnsTna.fs_resources.disk_space.upper_bound =
								fsRes->get().disk_space_ub;
							gnsTna.fs_resources.disk_space.lb_included =
								(fsRes->get().range_flags & 0x2) >> 1;
							gnsTna.fs_resources.disk_space.ub_included =
								fsRes->get().range_flags & 0x1;
							gnsTna.fs_resources.mount_point =
								new std::string(fsRes->get().mount_point);
							gnsTna.fs_resources.mount_source =
								new std::string(fsRes->get().mount_source);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, fs_resources);
					break;
				}
				case TYPE_SYSTEM_CAPS: {
					switch (subObj->subType()) {
						case SUBTYPE_SYSTEM_CAPS: {
							SystemCaps_SubObj * sysCap;
							sysCap = static_cast<SystemCaps_SubObj *> (subObj);
							VERSION_SET(gnsTna.sys_caps.os_version,
								    sysCap->get().os_version);
							gnsTna.sys_caps.os_type  =
								(grid_os_type_t) sysCap->get().os_type;
							gnsTna.sys_caps.cpu_arch =
								(grid_cpu_arch_t) sysCap->get().cpu_arch;
							gnsTna.sys_caps.exclusive_access = sysCap->get().E;
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, sys_caps);
					break;
				}
				case TYPE_INDIV_CPU_SPEED: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_SPEED: {
							IndvCpuSpeed_SubObj * ics;
							ics = static_cast<IndvCpuSpeed_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_speed,
								  ics->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_speed);
					break;
				}
				case TYPE_INDIV_CPU_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_COUNT: {
							IndvCpuCount_SubObj * icc;
							icc = static_cast<IndvCpuCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_count,
								  icc->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_count);
					break;
				}
				case TYPE_INDIV_CPU_TIME: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_TIME: {
							IndvCpuTime_SubObj * ict;
							ict = static_cast<IndvCpuTime_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_time,
								  ict->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_time);
					break;
				}
				case TYPE_INDIV_NET_BW: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_NET_BW: {
							IndvNetBw_SubObj * nbw;
							nbw = static_cast<IndvNetBw_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_net_bw,
								  nbw->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_net_bw);
					break;
				}
				case TYPE_INDIV_PHY_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_PHY_MEM: {
							IndvPhyMem_SubObj * ipm;
							ipm = static_cast<IndvPhyMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_phy_mem,
								  ipm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_phy_mem);
					break;
				}
				case TYPE_INDIV_VIR_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_VIR_MEM: {
							IndvVirMem_SubObj * ivm;
							ivm = static_cast<IndvVirMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_vir_mem,
								  ivm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_vir_mem);
					break;
				}
				case TYPE_INDIV_DISK_SPACE: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_DISK_SPACE: {
							IndvDiskSpace_SubObj * ids;
							ids = static_cast<IndvDiskSpace_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_disk_space,
								  ids->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_disk_space);
					break;
				}
				case TYPE_TOTAL_CPU_TIME: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_CPU_TIME: {
							TotCpuTime_SubObj * tct;
							tct = static_cast<TotCpuTime_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_cpu_time,
								  tct->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_cpu_time);
					break;
				}
				case TYPE_TOTAL_CPU_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_CPU_COUNT: {
							TotCpuCount_SubObj * tcc;
							tcc = static_cast<TotCpuCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_cpu_count,
								  tcc->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_cpu_count);
					break;
				}
				case TYPE_TOTAL_PHY_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_PHY_MEM: {
							TotPhyMem_SubObj * tpm;
							tpm = static_cast<TotPhyMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_phy_mem,
								  tpm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_phy_mem);
					break;
				}
				case TYPE_TOTAL_VIR_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_VIR_MEM: {
							TotVirMem_SubObj * tvm;
							tvm = static_cast<TotVirMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_vir_mem,
								  tvm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_vir_mem);
					break;
				}
				case TYPE_TOTAL_DISK_SPACE: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_DISK_SPACE: {
							TotDiskSpace_SubObj * tds;
							tds = static_cast<TotDiskSpace_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_disk_space,
								  tds->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_disk_space);
					break;
				}
				case TYPE_TOTAL_RES_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_RES_COUNT: {
							TotResCount_SubObj * trs;
							trs = static_cast<TotResCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_res_count,
								  trs->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_res_count);
					break;
				}
				case TYPE_DATA_STAGING: {
					switch (subObj->subType()) {
						case SUBTYPE_DATA_STAGING: {
							DataStaging_SubObj * ds;
							ds = static_cast<DataStaging_SubObj *> (subObj);
							gnsTna.data_staging.fs_name            =
								(grid_fs_name_t) ds->get().FSname;
						        gnsTna.data_staging.creation_flag      =
								(grid_creation_flag_t) ds->get().creation_flag;
							gnsTna.data_staging.del_on_termination =
								ds->get().delete_on_termination;
							gnsTna.data_staging.filename           =
								new std::string(ds->get().file_name);
							gnsTna.data_staging.source             =
								new std::string(ds->get().source);
							gnsTna.data_staging.target             =
								new std::string(ds->get().target);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, data_staging);
					break;
				}
				default:
					return false;
			}

		}

		return true;
	}

	Message * PathMessage::clone(void) const
	{
		PathMessage * msg = new PathMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(message_id_ack_nack_, msg);
		CLONE_OBJ_INTO_MSG(message_id_, msg);
		CLONE_OBJ_INTO_MSG(ero_, msg);
		CLONE_OBJ_INTO_MSG(session_attribute_, msg);
#ifdef CRANKBACK_ON
		CLONE_OBJ_INTO_MSG(lsp_attributes_, msg);
#endif // CRANKBACK_ON
		CLONE_OBJ_INTO_MSG(admin_status_, msg);
		CLONE_OBJ_INTO_MSG(protection_, msg);
		CLONE_OBJ_INTO_MSG(notify_request_, msg);
		CLONE_OBJ_INTO_MSG(label_set_, msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(rsvp_hop_, msg);
		CLONE_OBJ_INTO_MSG(time_values_, msg);
		CLONE_OBJ_INTO_MSG(label_request_, msg);
		CLONE_OBJ_INTO_MSG(call_ops_, msg);
		CLONE_OBJ_INTO_MSG(call_id_, msg);
		CLONE_OBJ_INTO_MSG(gns_call_ext_, msg);
		CLONE_OBJ_INTO_MSG(gen_uni_, msg);
		CLONE_OBJ_INTO_MSG(gns_uni_eg_, msg);
		CLONE_OBJ_INTO_MSG(gns_uni_in_, msg);
		CLONE_OBJ_INTO_MSG(times_, msg);
		CLONE_OBJ_INTO_MSG(sender_template_, msg);
		CLONE_OBJ_INTO_MSG(sender_tspec_, msg);
		CLONE_OBJ_INTO_MSG(adspec_, msg);
		CLONE_OBJ_INTO_MSG(rro_, msg);
		CLONE_OBJ_INTO_MSG(suggested_label_, msg);
		CLONE_OBJ_INTO_MSG(recovery_label_, msg);
		CLONE_OBJ_INTO_MSG(policy_data_, msg);
		CLONE_OBJ_INTO_MSG(upstream_label_, msg);

		return msg;
	}

	bool PathMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(rsvp_hop_);
		CHECK_MANDATORY_OBJ(time_values_);
		CHECK_MANDATORY_OBJ(label_request_);
		CHECK_MANDATORY_OBJ(sender_template_);
		CHECK_MANDATORY_OBJ(sender_tspec_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(message_id_ack_nack_);
		CHECK_OPTIONAL_OBJ(message_id_);
		CHECK_OPTIONAL_OBJ(ero_);
		CHECK_OPTIONAL_OBJ(session_attribute_);
#ifdef CRANKBACK_ON
		CHECK_OPTIONAL_OBJ(lsp_attributes_);
#endif // CRANKBACK_ON
		CHECK_OPTIONAL_OBJ(admin_status_);
		CHECK_OPTIONAL_OBJ(protection_);
		CHECK_OPTIONAL_OBJ(notify_request_);
		CHECK_OPTIONAL_OBJ(label_set_);
		CHECK_OPTIONAL_OBJ(call_ops_);
		CHECK_OPTIONAL_OBJ(call_id_);
		CHECK_OPTIONAL_OBJ(gns_call_ext_);
		CHECK_OPTIONAL_OBJ(times_);
		CHECK_OPTIONAL_OBJ(gen_uni_);
		CHECK_OPTIONAL_OBJ(gns_uni_eg_);
		CHECK_OPTIONAL_OBJ(gns_uni_in_);
		CHECK_OPTIONAL_OBJ(adspec_);
		CHECK_OPTIONAL_OBJ(rro_);
		CHECK_OPTIONAL_OBJ(suggested_label_);
		CHECK_OPTIONAL_OBJ(recovery_label_);
		CHECK_OPTIONAL_OBJ(policy_data_);
		CHECK_OPTIONAL_OBJ(upstream_label_);

		return true;
	}

	bool PathMessage::isDown(void) const
	{
		if (admin_status_) {
			assert(admin_status_->cType() == CTYPE_ADMIN_STATUS);
			return (static_cast<AdminStatus_Object *>
				(admin_status_)->get().D == 1);
		}

		return false;
	}

	bool PathMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (sender_template_) {
			GETLSPINFO_FROM_SENDTEMP(sender_template_,
						 &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t PathMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(message_id_ack_nack_, len);
		ADD_LENGTH(message_id_, len);
		ADD_LENGTH(ero_, len);
		ADD_LENGTH(session_attribute_, len);
#ifdef CRANKBACK_ON
		ADD_LENGTH(lsp_attributes_, len);
#endif // CRANKBACK_ON
		ADD_LENGTH(admin_status_, len);
		ADD_LENGTH(protection_, len);
		ADD_LENGTH(notify_request_, len);
		ADD_LENGTH(label_set_, len);

		ADD_LENGTH(session_, len);
		ADD_LENGTH(rsvp_hop_, len);
		ADD_LENGTH(time_values_, len);
		ADD_LENGTH(label_request_, len);
		ADD_LENGTH(call_ops_, len);
		ADD_LENGTH(call_id_, len);
		ADD_LENGTH(gns_call_ext_, len);
		ADD_LENGTH(times_, len);
		ADD_LENGTH(gen_uni_, len);
		ADD_LENGTH(gns_uni_eg_, len);
		ADD_LENGTH(gns_uni_in_, len);
		ADD_LENGTH(sender_template_, len);
		ADD_LENGTH(sender_tspec_, len);
		ADD_LENGTH(adspec_, len);
		ADD_LENGTH(rro_, len);
		ADD_LENGTH(suggested_label_, len);
		ADD_LENGTH(recovery_label_, len);
		ADD_LENGTH(policy_data_, len);
		ADD_LENGTH(upstream_label_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor & mc, PathMessage & path)
	{
		// TODO parse all PathMsg objs
		int length = mc.remainingSize();
		bool gnsUniEgGot = false;

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!path.session_);
					mcObj >> path.session_;
					assert(path.session_);
					break;
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
					assert(!path.session_attribute_);
					mcObj >> path.session_attribute_;
					assert(path.session_attribute_);
					break;
#ifdef CRANKBACK_ON
				case G2RSVP_CLASS_LSP_ATTRIBUTES:
					assert(!path.lsp_attributes_);
					mcObj >> path.lsp_attributes_;
					assert(path.lsp_attributes_);
					break;
#endif // CRANKBACK_ON
				case G2RSVP_CLASS_SUGGESTED_LABEL:
					assert(!path.suggested_label_);
					mcObj >> path.suggested_label_;
					assert(path.suggested_label_);
					break;
				case G2RSVP_CLASS_UPSTREAM_LABEL:
					assert(!path.upstream_label_);
					mcObj >> path.upstream_label_;
					assert(path.upstream_label_);
					break;
				case G2RSVP_CLASS_RECOVERY_LABEL:
					assert(!path.recovery_label_);
					mcObj >> path.recovery_label_;
					assert(path.recovery_label_);
					break;
				case G2RSVP_CLASS_LABEL_REQ:
					assert(!path.label_request_);
					mcObj >> path.label_request_;
					assert(path.label_request_);
					break;
				case G2RSVP_CLASS_LABEL_SET:
					assert(!path.label_set_);
					mcObj >> path.label_set_;
					assert(path.label_set_);
					break;
				case G2RSVP_CLASS_TIME_VALUES:
					assert(!path.time_values_);
					mcObj >> path.time_values_;
					assert(path.time_values_);
					break;
				case G2RSVP_CLASS_RSVP_HOP:
					assert(!path.rsvp_hop_);
					mcObj >> path.rsvp_hop_;
					assert(path.rsvp_hop_);
					break;
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
					assert(!path.ero_);
					mcObj >> path.ero_;
					assert(path.ero_);
					break;
				case G2RSVP_CLASS_RECORDED_ROUTE:
					assert(!path.rro_);
					mcObj >> path.rro_;
					assert(path.rro_);
					break;
				case G2RSVP_CLASS_SENDER_TEMPLATE:
					assert(!path.sender_template_);
					mcObj >> path.sender_template_;
					assert(path.sender_template_);
					break;
				case G2RSVP_CLASS_SENDER_TSPEC:
					assert(!path.sender_tspec_);
					mcObj >> path.sender_tspec_;
					assert(path.sender_tspec_);
					break;
				case G2RSVP_CLASS_PROTECTION:
					assert(!path.protection_);
					mcObj >> path.protection_;
					assert(path.protection_);
					break;
				case G2RSVP_CLASS_ADMIN_STATUS:
					assert(!path.admin_status_);
					mcObj >> path.admin_status_;
					assert(path.admin_status_);
					break;
				case G2RSVP_CLASS_GNS_CALL_EXT:
					assert(!path.gns_call_ext_);
					mcObj >> path.gns_call_ext_;
					assert(path.gns_call_ext_);
					break;
				case G2RSVP_CLASS_TIMES:
					assert(!path.times_);
					mcObj >> path.times_;
					assert(path.times_);
					break;
				case G2RSVP_CLASS_INTEGRITY:
					assert(!path.integrity_);
					mcObj >> path.integrity_;
					assert(path.integrity_);
					break;
				case G2RSVP_CLASS_GENERALIZED_UNI:
					assert(!path.gen_uni_);
					mcObj >> path.gen_uni_;
					assert(path.gen_uni_);
					break;
				case G2RSVP_CLASS_GNS_UNI:
					if (!gnsUniEgGot) {
						assert(!path.gns_uni_eg_);
						mcObj >> path.gns_uni_eg_;
						assert(path.gns_uni_eg_);
						gnsUniEgGot = true;
					} else {
						assert(!path.gns_uni_in_);
						mcObj >> path.gns_uni_in_;
						assert(path.gns_uni_in_);
					}
					break;
				case G2RSVP_CLASS_CALL_OPS:
					assert(!path.call_ops_);
					mcObj >> path.call_ops_;
					assert(path.call_ops_);
					break;
				case G2RSVP_CLASS_CALL_ID:
					assert(!path.call_id_);
					mcObj >> path.call_id_;
					assert(path.call_id_);
					break;
				case G2RSVP_CLASS_NOTIFY_REQUEST:
					assert(!path.notify_request_);
					mcObj >> path.notify_request_;
					assert(path.notify_request_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID:
					assert(!path.message_id_);
					mcObj >> path.message_id_;
					assert(path.message_id_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
					assert(!path.message_id_ack_nack_);
					mcObj >> path.message_id_ack_nack_;
					assert(path.message_id_ack_nack_);
					break;
				case G2RSVP_CLASS_POLICY:
					assert(!path.policy_data_);
					mcObj >> path.policy_data_;
					assert(path.policy_data_);
					break;
				case G2RSVP_CLASS_ADSPEC:
					// to be filled up
					throw(std::string("ADSPEC !supported"));
				case G2RSVP_CLASS_SCOPE:
					assert(!path.policy_data_);
					mcObj >> path.policy_data_;
					assert(path.policy_data_);
					break;
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_ERROR_SPEC:
				case G2RSVP_CLASS_STYLE:
				case G2RSVP_CLASS_FLOWSPEC:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_FILTER_SPEC:
				case G2RSVP_CLASS_LABEL:
				case G2RSVP_CLASS_RESV_CONFIRM:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in PathMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor & mc, const PathMessage & path)
	{
		// All PathMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(path.session_, cur); //it's the first mandatory
		assert(path.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (path.session_->length() - 1));
		cur << *(path.session_);

		OBJ2CURSOR(path.integrity_, cur);
		OBJ2CURSOR(path.message_id_ack_nack_, cur);
		OBJ2CURSOR(path.message_id_, cur);
		//session obj should be here
		OBJ2CURSOR(path.rsvp_hop_, cur);
		OBJ2CURSOR(path.time_values_, cur);
		OBJ2CURSOR(path.ero_, cur);
		OBJ2CURSOR(path.label_request_, cur);
		OBJ2CURSOR(path.call_ops_, cur);
		OBJ2CURSOR(path.call_id_, cur);
		OBJ2CURSOR(path.protection_, cur);
		OBJ2CURSOR(path.label_set_, cur);
		OBJ2CURSOR(path.session_attribute_, cur);
#ifdef CRANKBACK_ON
		OBJ2CURSOR(path.lsp_attributes_, cur);
#endif // CRANKBACK_ON
		OBJ2CURSOR(path.notify_request_, cur);
		OBJ2CURSOR(path.admin_status_, cur);
		OBJ2CURSOR(path.gns_call_ext_, cur);
		OBJ2CURSOR(path.times_, cur);
		OBJ2CURSOR(path.gen_uni_, cur);
		OBJ2CURSOR(path.gns_uni_eg_, cur);
		OBJ2CURSOR(path.gns_uni_in_, cur);

		OBJ2CURSOR(path.policy_data_, cur);
		//sender description
		OBJ2CURSOR(path.sender_template_, cur);
		OBJ2CURSOR(path.sender_tspec_, cur);
//		OBJ2CURSOR(path.adspec_, cur);
		OBJ2CURSOR(path.rro_, cur);
		OBJ2CURSOR(path.suggested_label_, cur);
		OBJ2CURSOR(path.recovery_label_, cur);
		OBJ2CURSOR(path.upstream_label_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream & os, const PathMessage & path)
	{
		// All PathMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(path.integrity_, os);
		OBJ2STREAM(path.message_id_ack_nack_, os);
		OBJ2STREAM(path.message_id_, os);
		OBJ2STREAM(path.session_, os);
		OBJ2STREAM(path.rsvp_hop_, os);
		OBJ2STREAM(path.time_values_, os);
		OBJ2STREAM(path.ero_, os);
		OBJ2STREAM(path.label_request_, os);
		OBJ2STREAM(path.call_ops_, os);
		OBJ2STREAM(path.call_id_, os);
		OBJ2STREAM(path.protection_, os);
		OBJ2STREAM(path.label_set_, os);
		OBJ2STREAM(path.session_attribute_, os);
#ifdef CRANKBACK_ON
		OBJ2STREAM(path.lsp_attributes_, os);
#endif // CRANKBACK_ON
		OBJ2STREAM(path.notify_request_, os);
		OBJ2STREAM(path.admin_status_, os);

		OBJ2STREAM(path.gns_call_ext_, os);
		OBJ2STREAM(path.times_, os);
		OBJ2STREAM(path.gen_uni_, os);
		OBJ2STREAM(path.gns_uni_eg_, os);
		OBJ2STREAM(path.gns_uni_in_, os);

		OBJ2STREAM(path.policy_data_, os);
		//sender description
		OBJ2STREAM(path.sender_template_, os);
		OBJ2STREAM(path.sender_tspec_, os);
//		OBJ2STREAM(path.adspec_, os);
		OBJ2STREAM(path.rro_, os);
		OBJ2STREAM(path.suggested_label_, os);
		OBJ2STREAM(path.recovery_label_, os);
		OBJ2STREAM(path.upstream_label_, os);

		return os;
	}


	/////////////////////////////////
	//      ResvMessage class      //
	/////////////////////////////////
	ResvMessage::ResvMessage(void) : Message(RESV)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		resv_confirm_        = 0;
		scope_               = 0;
		notify_request_      = 0;
		admin_status_        = 0;
		policy_data_         = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		style_               = 0;
		flow_desc_           = 0;
	}

	ResvMessage::ResvMessage(uint8_t version) : Message(version, RESV)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		resv_confirm_        = 0;
		scope_               = 0;
		notify_request_      = 0;
		admin_status_        = 0;
		policy_data_         = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		style_               = 0;
		flow_desc_           = 0;
	}

	ResvMessage::ResvMessage(const PathMessage & path) : Message(path.getVersion(), RESV)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		scope_               = 0;
		admin_status_        = 0;
		policy_data_         = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;

		// integrity_ = new ();
		// message_id_ack_nack_ = new ();
		// message_id_ = new ();

		resv_confirm_ = new ResvConfirmIPv4_Object();
		if (!resv_confirm_) {
			throw(std::string("Impossible to allocate "
					  "resv confirm obj for Resv msg!"));
		}

		// scope_ = new ();

		notify_request_ = new IPV4NotifyRequest_Object();
		if (!notify_request_) {
			throw(std::string("Impossible to allocate notify "
					  "request obj for Resv msg!"));
		}

		// admin_status_ = new AdminStatus_Object();
		// policy_data_ = new ();

		if (path.session_) {
			session_ = path.session_->clone();
			if (!session_) {
				throw(std::string("Impossible to allocate "
						  "session obj for Resv msg!"));
			}
		}

		//This will be done on prepareNHOP
		//if (path.rsvp_hop_) {
		//	rsvp_hop_ = path.rsvp_hop_->clone();
		//	if (!rsvp_hop_) {
		//		throw(std::string("Impossible to allocate "
		//				  "rsvpHop obj for Resv msg!"));
		//	}
		//}

		if (path.time_values_) {
			time_values_ = path.time_values_->clone();
			if (!time_values_) {
				throw(std::string("Impossible to allocate time "
						  "Values obj for Resv msg!"));
			}
		}

		// call_ops_ = new ();
		// call_id_ = new ();
		// gns_call_ext_ = new ();

		style_ = new Style_Object();
		if (!style_) {
			throw(std::string("Impossible to allocate "
					  "style obj for Resv msg!"));
		}

		// Copy Filter/Flow Spec obj from Sender Template/TSpec obj in
		// the received PathMsg, (the cType are the same ...)
		// Before that create the Flow Desc that has Filter/Flow Spec

		//assert(style_);
		//if (style_->cType() == CTYPE_STYLE
		//    && static_cast<Style_Object *>(style_)->isFF()) {
		FF_FlowDesc_t * ff;
		ff = new FF_FlowDesc_t();
		assert(ff);
		flow_desc_ = new Flow_Desc_List(ff);
		assert(flow_desc_);
		//} else {
		//	throw(std::string("Only FF STYLE obj is supported!!!"));
		//}
		if (path.sender_template_) {
			FilterSpec_Class * filter = 0;
			SenderTemplate2FilterSpec_Class(path.sender_template_,
							filter);
			if (!filter) {
				throw(std::string("Impossible to allocate fil"
						  "terSpec obj for ResvMsg"));
			}
			if (!flow_desc_->setFilterSpecObj(filter)) {
				throw(std::string("Impossible to add "
						  "FilterSpec Obj"));
			}
		}
		if (path.sender_tspec_) {
			FlowSpec_Class * flow = 0;
			SenderTSpec_Object * sTSpec;
			IntServFlowSpec_Object * fTemp;
			sTSpec = static_cast<SenderTSpec_Object *> (path.sender_tspec_);
			assert(sTSpec);
			fTemp = new IntServFlowSpec_Object();
			if (!fTemp) {
				throw(std::string("Impossible to allocate "
						  "flowSpec obj for ResvMsg!"));
			}
			//set Guaranteed load params
			fTemp->get().data.msg_version = sTSpec->get().msg_version;
			fTemp->get().data.length = 0xA; // 10
			fTemp->get().data.service_header = GUARANTEED_SERVICE;
			fTemp->get().data.service_data_length = 0x9;
			fTemp->get().data.param_id_127 = sTSpec->get().param_id_127;
			fTemp->get().data.param127_flags = sTSpec->get().param127_flags;
			fTemp->get().data.param127_length = sTSpec->get().param127_length;
			fTemp->get().data.token_bucket_rate = sTSpec->get().token_bucket_rate;
			fTemp->get().data.token_bucket_size = sTSpec->get().token_bucket_size;
			fTemp->get().data.peak_data_rate = sTSpec->get().peak_data_rate;
			fTemp->get().data.min_policed_unit = sTSpec->get().min_policed_unit;
			fTemp->get().data.max_policed_unit = sTSpec->get().max_policed_unit;
			fTemp->get().param_id_130 = new uint8_t;
			*fTemp->get().param_id_130 = 0x82; // 130
			fTemp->get().param130_flags = new uint8_t;
			*fTemp->get().param130_flags = 0x0;
			fTemp->get().param130_length = new uint16_t;
			*fTemp->get().param130_length = 0x2;
			fTemp->get().rate = new uint32_t;
			*fTemp->get().rate = fTemp->get().data.peak_data_rate;
			fTemp->get().slack_term = new uint32_t;
			*fTemp->get().slack_term = 0x0;
			flow = fTemp;
			if (!flow_desc_->setFlowSpecObj(flow)) {
				throw(std::string("Impossible to add "
						  "FlowSpec Obj"));
			}
		}
		if (path.rro_) {
			RecordedRoute_Class * rro;
			//rro = path.rro_->clone();
			rro = new RecordedRoute_Object();
			if (!rro) {
				throw(std::string("Impossible to allocate "
						  "RRO obj for ResvMsg!"));
			}
			if (!flow_desc_->setRROobj(rro)) {
				throw(std::string("Impossible to add RRO Obj"));
			}
		}
		if (path.call_id_) {
			call_id_ = path.call_id_->clone();
			if (!call_id_) {
				throw(std::string("Impossible to allocate "
						  "CallId obj for ResvMsg!"));
			}
		}

		if (path.call_ops_) {
			call_ops_ = path.call_ops_->clone();
			if (!call_ops_) {
				throw(std::string("Impossible to allocate label"
						  "CallOps obj for Resv Msg!"));
			}
		}

		if (path.gns_call_ext_) {
			gns_call_ext_ = path.gns_call_ext_->clone();
			if (!gns_call_ext_) {
				throw(std::string("Impossible to allocate label"
						  "GnsCallExt obj for Resv Msg!"));
			}
		}


		Label_Class * label;
		label = new GLabel_Object();
		if (!label) {
			throw(std::string("Impossible to allocate "
					  "label obj for ResvMsg!"));
		}
		if (!flow_desc_->setLabelObj(label)) {
			throw(std::string("Impossible to add label Obj"));
		}
	}

	ResvMessage::ResvMessage(const ResvMessage & resv) :
		Message(resv.getVersion(), RESV)
	{
		createResvMessage(resv, false);
	}

	ResvMessage::ResvMessage(const ResvMessage & resv, bool down) :
		Message(resv.getVersion(), RESV)
	{
		createResvMessage(resv, down);
	}

	void ResvMessage::createResvMessage(const ResvMessage & resv, bool down)
	{
		integrity_           = 0;
		message_id_ack_nack_ = 0;
		message_id_          = 0;
		resv_confirm_        = 0;
		scope_               = 0;
		notify_request_      = 0;
		//admin_status_        = 0;
		policy_data_         = 0;

		session_             = 0;
		rsvp_hop_            = 0;
		time_values_         = 0;
		call_ops_            = 0;
		call_id_             = 0;
		gns_call_ext_        = 0;
		gns_uni_eg_          = 0;
		gns_uni_in_          = 0;
		style_               = 0;
		flow_desc_           = 0;

		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!

		//TODO  set flags STATIC

		if (down) {
			admin_status_ = new AdminStatus_Object(false, //REFL
							       false, //TEST
							       true,  //ADM
							       true); //DEL
		} else if (resv.admin_status_) {
			admin_status_ = resv.admin_status_->clone();
		} else {
			admin_status_ = new AdminStatus_Object();
		}
		if (!admin_status_) {
			throw(std::string("Impossible to allocate admin"
					  "Status obj for Resv msg!"));
		}

		if (resv.session_) {
			session_ = resv.session_->clone();
			if (!session_) {
				throw(std::string("Impossible to allocate "
						  "session obj for Resv msg!"));
			}
		}

		//FIX Mandatory???
		assert(resv.rsvp_hop_);
		rsvp_hop_ = resv.rsvp_hop_->clone();
		if (!rsvp_hop_) {
			throw(std::string("Impossible to allocate "
					  "rsvpHop obj for Resv msg!"));
		}

		if (resv.time_values_) {
			time_values_ = resv.time_values_->clone();
			if (!time_values_) {
				throw(std::string("Impossible to allocate time "
						  "Values obj for Resv msg!"));
			}
		}

		if (resv.style_) {
			style_ = resv.style_->clone();
			if (!style_) {
				throw(std::string("Impossible to allocate label"
						  "Request obj for Resv Msg!"));
			}
		}

		if (resv.call_id_) {
			call_id_ = resv.call_id_->clone();
			if (!call_id_) {
				throw(std::string("Impossible to allocate label"
						  "CallId obj for Resv Msg!"));
			}
		}

		if (resv.call_ops_) {
			call_ops_ = resv.call_ops_->clone();
			if (!call_ops_) {
				throw(std::string("Impossible to allocate label"
						  "CallOps obj for Resv Msg!"));
			}
		}

		if (resv.gns_call_ext_) {
			gns_call_ext_ = resv.gns_call_ext_->clone();
			if (!gns_call_ext_) {
				throw(std::string("Impossible to allocate label"
						  "GnsCallExt obj for Resv Msg!"));
			}
		}

		if (resv.gns_uni_eg_) {
			gns_uni_eg_ = resv.gns_uni_eg_->clone();
			if (!gns_uni_eg_) {
				throw(std::string("Impossible to allocate send"
						  "egress GnsUni obj for Resv msg!"));
			}
		}

		if (resv.gns_uni_in_) {
			gns_uni_in_ = resv.gns_uni_in_->clone();
			if (!gns_uni_in_) {
				throw(std::string("Impossible to allocate send"
						  "ingress GnsUni obj for Resv msg!"));
			}
		}

		//assert(style_);
		//if (style_->cType() == CTYPE_STYLE
		//    && static_cast<Style_Object *>(style_)->isFF()) {
		FF_FlowDesc_t * ff;
		ff = new FF_FlowDesc_t();
		assert(ff);
		flow_desc_ = new Flow_Desc_List(ff);
		assert(flow_desc_);
		//} else {
		//	throw(std::string("Only FF STYLE obj is supported!!!"));
		//}
		if (resv.flow_desc_ && resv.flow_desc_->getFilterSpecObj()) {
			FilterSpec_Class * filter;
			filter = resv.flow_desc_->getFilterSpecObj()->clone();
			if (!filter) {
				throw(std::string("Impossible to allocate fil"
						  "terSpec obj for ResvMsg"));
			}
			if (!flow_desc_->setFilterSpecObj(filter)) {
				throw(std::string("Impossible to add FilterSpec"
						  " obj for ResvMsg"));
			}
		}
		//FIX Mandatory???
		//assert(resv.flow_desc_ && resv.flow_desc_->getLabelObj());
		if (resv.flow_desc_ && resv.flow_desc_->getLabelObj()) {
			Label_Class * label;
			label = resv.flow_desc_->getLabelObj()->clone();
			if (!label) {
				throw(std::string("Impossible to allocate Label"
						  " obj for ResvMsg"));
			}
			if (!flow_desc_->setLabelObj(label)) {
				throw(std::string("Impossible to add Label "
						  "obj for ResvMsg"));
			}
		}
	}

	ResvMessage::~ResvMessage(void)
	{
		G2RSVPTE_DBG("Destroying ResvMessage %p ...", this);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(message_id_ack_nack_);
		DELETE_OBJ(message_id_);
		DELETE_OBJ(resv_confirm_);
		DELETE_OBJ(scope_);
		DELETE_OBJ(notify_request_);
		DELETE_OBJ(admin_status_);
		DELETE_OBJ(policy_data_);

		DELETE_OBJ(session_);
		DELETE_OBJ(rsvp_hop_);
		DELETE_OBJ(time_values_);
		DELETE_OBJ(call_ops_);
		DELETE_OBJ(call_id_);
		DELETE_OBJ(gns_call_ext_);
		DELETE_OBJ(gns_uni_in_);
		DELETE_OBJ(gns_uni_eg_);
		DELETE_OBJ(style_);
		DELETE_OBJ(flow_desc_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	ResvMessage * ResvMessage::partialClone(void) const
	{
		ResvMessage * resv = new ResvMessage(getVersion());
		if (!resv) return 0;

		// TODO flags, error code, actinfo ...

		// Session
		CLONE_OBJ_INTO_MSG(session_, resv);

		// Time Values
		CLONE_OBJ_INTO_MSG(time_values_, resv);

		// ERO
		//CLONE_OBJ_INTO_MSG(ero_, resv);

		// Label Request
		//CLONE_OBJ_INTO_MSG(label_request_, resv);

		// Protection
		//CLONE_OBJ_INTO_MSG(protection_, resv);

		// Session Attribute
		//CLONE_OBJ_INTO_MSG(session_attribute_, resv);

		// NOTIFY REQUEST
		CLONE_OBJ_INTO_MSG(notify_request_, resv);

		// ADMIN_STATUS
		CLONE_OBJ_INTO_MSG(admin_status_, resv);

		// POLICY_DATA
		CLONE_OBJ_INTO_MSG(policy_data_, resv);

		// SENDER_TEMPLATE
		//CLONE_OBJ_INTO_MSG(sender_template_, resv);

		// SENDER_TSPEC
		//CLONE_OBJ_INTO_MSG(sender_tspec_, resv);

		// ADSPEC
		//CLONE_OBJ_INTO_MSG(adspec_, resv);

		// RRO
		//CLONE_OBJ_INTO_MSG(rro_, resv);

		// RESV_CONFIRM
		CLONE_OBJ_INTO_MSG(resv_confirm_, resv);

		// SCOPE
		CLONE_OBJ_INTO_MSG(scope_, resv);

		// STYLE
		CLONE_OBJ_INTO_MSG(style_, resv);

		// FLOW_SPEC & FILTER_SPEC
		CLONE_OBJ_INTO_MSG(flow_desc_, resv);

		// ERROR SPEC
		//CLONE_OBJ_INTO_MSG(error_spec_, resv);

		// CALL ID
		CLONE_OBJ_INTO_MSG(call_id_, resv);

		// CALL OPS
		CLONE_OBJ_INTO_MSG(call_ops_, resv);

		// GNS CALL EXT
		CLONE_OBJ_INTO_MSG(gns_call_ext_, resv);

		// ingress GNS UNI Object
		CLONE_OBJ_INTO_MSG(gns_uni_in_, resv);

		// egress GNS UNI Object
		CLONE_OBJ_INTO_MSG(gns_uni_eg_, resv);

		return resv;
	}

	Message * ResvMessage::clone(void) const
	{
		ResvMessage * msg = new ResvMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(message_id_ack_nack_, msg);
		CLONE_OBJ_INTO_MSG(message_id_, msg);
		CLONE_OBJ_INTO_MSG(resv_confirm_, msg);
		CLONE_OBJ_INTO_MSG(scope_, msg);
		CLONE_OBJ_INTO_MSG(notify_request_, msg);
		CLONE_OBJ_INTO_MSG(admin_status_, msg);
		CLONE_OBJ_INTO_MSG(policy_data_, msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(rsvp_hop_, msg);
		CLONE_OBJ_INTO_MSG(time_values_, msg);
		CLONE_OBJ_INTO_MSG(call_ops_, msg);
		CLONE_OBJ_INTO_MSG(call_id_, msg);
		CLONE_OBJ_INTO_MSG(gns_call_ext_, msg);
		CLONE_OBJ_INTO_MSG(gns_uni_in_, msg);
		CLONE_OBJ_INTO_MSG(gns_uni_eg_, msg);
		CLONE_OBJ_INTO_MSG(style_, msg);
		CLONE_OBJ_INTO_MSG(flow_desc_, msg);

		return msg;
	}

	bool ResvMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(rsvp_hop_);
		CHECK_MANDATORY_OBJ(time_values_);
		CHECK_MANDATORY_OBJ(style_);
		CHECK_MANDATORY_OBJ(flow_desc_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(message_id_ack_nack_);
		CHECK_OPTIONAL_OBJ(message_id_);
		CHECK_OPTIONAL_OBJ(resv_confirm_);
		CHECK_OPTIONAL_OBJ(scope_);
		CHECK_OPTIONAL_OBJ(notify_request_);
		CHECK_OPTIONAL_OBJ(admin_status_);
		CHECK_OPTIONAL_OBJ(policy_data_);
		CHECK_OPTIONAL_OBJ(call_ops_);
		CHECK_OPTIONAL_OBJ(call_id_);
		CHECK_OPTIONAL_OBJ(gns_call_ext_);
		CHECK_OPTIONAL_OBJ(gns_uni_eg_);
		CHECK_OPTIONAL_OBJ(gns_uni_in_);

		return true;
	}

	bool ResvMessage::isDown(void) const
	{
		if (admin_status_) {
			assert(admin_status_->cType() == CTYPE_ADMIN_STATUS);
			return (static_cast<AdminStatus_Object *>
				(admin_status_)->get().D == 1);
		}

		return false;
	}

	bool ResvMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (flow_desc_ && flow_desc_->getFilterSpecObj()) {
			GETLSPINFO_FROM_FILTERSPEC(flow_desc_->
						   getFilterSpecObj(),
						   &info);
		} else {
			return false;
		}

		return true;
	}

	bool ResvMessage::setGnsTna(const grid_res_spec_t & gnsTna,
				    bool                    ingress)
	{
		GnsUni_Object * gnsUni;
		gnsUni = new GnsUni_Object;
		if (!gnsUni) {
			G2RSVPTE_ERR("Impossible to allocate new GnsUni_Object");
			return false;
		}

		// Application_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, application)) {
			Application_SubObj * app;
			app = new Application_SubObj(gnsTna.application);
			assert(app);
			gnsUni->get().subObjs.push_back(app);
		}
		// Candidate host SuObj
		if (BITMASK_BITTEST(gnsTna.mask_, cand_host)) {
			switch (gnsTna.cand_host.type) {
				case IPv4:
					// IPv4CandidateHost_SubObj
					IPv4CandidateHost_SubObj * ipv4Host;
					ipv4Host = new IPv4CandidateHost_SubObj(gnsTna.cand_host.value.ipv4);
					assert(ipv4Host);
					gnsUni->get().subObjs.push_back(ipv4Host);
					break;
				case IPv6:
					// IPv4CandidateHost_SubObj
					IPv6CandidateHost_SubObj * ipv6Host;
					ipv6Host = new IPv6CandidateHost_SubObj(gnsTna.cand_host.value.ipv6);
					assert(ipv6Host);
					gnsUni->get().subObjs.push_back(ipv6Host);
					break;
				case NSAP:
					// NsapCandidateHost_SubObj
					NsapCandidateHost_SubObj * nsapHost;
					nsapHost = new NsapCandidateHost_SubObj(gnsTna.cand_host.value.nsap);
					assert(nsapHost);
					gnsUni->get().subObjs.push_back(nsapHost);
					break;
				default:
					G2RSVPTE_ERR("Candidate host address type not supported");
					delete gnsUni;
					return false;
					break;
			}
		}
		// FsResources_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, fs_resources)) {
			FsResources_SubObj * fsRes;
			fsRes = new FsResources_SubObj(gnsTna.fs_resources);
			assert(fsRes);
			gnsUni->get().subObjs.push_back(fsRes);
		}
		// SystemCaps_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, sys_caps)) {
			SystemCaps_SubObj * sysCap;
			sysCap = new SystemCaps_SubObj(gnsTna.sys_caps);
			assert(sysCap);
			gnsUni->get().subObjs.push_back(sysCap);
		}
		// IndvCpuSpeed_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_speed)) {
			IndvCpuSpeed_SubObj * ics;
			ics = new IndvCpuSpeed_SubObj;
			assert(ics);
			ics->get().lower_bound   = gnsTna.ind_cpu_speed.lower_bound;
			ics->get().upper_bound   = gnsTna.ind_cpu_speed.upper_bound;
			ics->get().SF            = (gnsTna.ind_cpu_speed.ub_included << 1) |
				                   (gnsTna.ind_cpu_speed.lb_included);
			gnsUni->get().subObjs.push_back(ics);
		}
		// IndvCpuTime_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_time)) {
			IndvCpuTime_SubObj * ict;
			ict = new IndvCpuTime_SubObj(gnsTna.ind_cpu_time);
			assert(ict);
			gnsUni->get().subObjs.push_back(ict);
		}
		// IndvCpuCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_cpu_count)) {
			IndvCpuCount_SubObj * icc;
			icc = new IndvCpuCount_SubObj(gnsTna.ind_cpu_count);
			assert(icc);
			gnsUni->get().subObjs.push_back(icc);
		}
		// IndvNetBw_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_net_bw)) {
			IndvNetBw_SubObj * nbw;
			nbw = new IndvNetBw_SubObj(gnsTna.ind_net_bw);
			assert(nbw);
			gnsUni->get().subObjs.push_back(nbw);
		}
		// IndvPhyMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_phy_mem)) {
			IndvPhyMem_SubObj * ipm;
			ipm = new IndvPhyMem_SubObj(gnsTna.ind_phy_mem);
			assert(ipm);
			gnsUni->get().subObjs.push_back(ipm);
		}
		// IndvVirMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_vir_mem)) {
			IndvVirMem_SubObj * ivm;
			ivm = new IndvVirMem_SubObj(gnsTna.ind_vir_mem);
			assert(ivm);
			gnsUni->get().subObjs.push_back(ivm);
		}
		// IndvDiskSpace_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, ind_disk_space)) {
			IndvDiskSpace_SubObj * ids;
			ids = new IndvDiskSpace_SubObj(gnsTna.ind_disk_space);
			assert(ids);
			gnsUni->get().subObjs.push_back(ids);
		}
		// TotCpuTime_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_cpu_time)) {
			TotCpuTime_SubObj * tct;
			tct = new TotCpuTime_SubObj(gnsTna.tot_cpu_time);
			assert(tct);
			gnsUni->get().subObjs.push_back(tct);
		}
		// TotCpuCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_cpu_count)) {
			TotCpuCount_SubObj * tcc;
			tcc = new TotCpuCount_SubObj(gnsTna.tot_cpu_count);
			assert(tcc);
			gnsUni->get().subObjs.push_back(tcc);
		}
		// TotPhyMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_phy_mem)) {
			TotPhyMem_SubObj * tpm;
			tpm = new TotPhyMem_SubObj(gnsTna.tot_phy_mem);
			assert(tpm);
			gnsUni->get().subObjs.push_back(tpm);
		}
		// TotVirMem_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_vir_mem)) {
			TotVirMem_SubObj * tvm;
			tvm = new TotVirMem_SubObj(gnsTna.tot_vir_mem);
			assert(tvm);
			gnsUni->get().subObjs.push_back(tvm);
		}
		// TotDiskSpace_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_disk_space)) {
			TotDiskSpace_SubObj * tds;
			tds = new TotDiskSpace_SubObj(gnsTna.tot_disk_space);
			assert(tds);
			gnsUni->get().subObjs.push_back(tds);
		}
		// TotResCount_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, tot_res_count)) {
			TotResCount_SubObj * trc;
			trc = new TotResCount_SubObj(gnsTna.tot_res_count);
			assert(trc);
			gnsUni->get().subObjs.push_back(trc);
		}
		// DataStaging_SubObj
		if (BITMASK_BITTEST(gnsTna.mask_, data_staging)) {
			DataStaging_SubObj * ds;
			ds = new DataStaging_SubObj(gnsTna.data_staging);
			assert(ds);
			gnsUni->get().subObjs.push_back(ds);
		}

		if (ingress) {
			gns_uni_in_ = gnsUni;
		} else {
			gns_uni_eg_ = gnsUni;
		}

		return true;
	}

	bool ResvMessage::getGnsTna(grid_res_spec_t & gnsTna,
				    bool              ingress)
	{
		GnsUni_Class *                         gnsUniClass;
		GnsUni_Object *                        gnsUni;
		GnsSubObj *                            subObj;
		std::list<GnsSubObj *>::const_iterator it;

		if (ingress) {
		       gnsUniClass = gns_uni_in_;
		} else {
		       gnsUniClass = gns_uni_eg_;
		}

		if (!gnsUniClass) {
			G2RSVPTE_ERR("Null pointer !");
			return false;
		}
		gnsUni = static_cast<GnsUni_Object *> (gnsUniClass);

		for (it  = gnsUni->get().subObjs.begin();
		     it != gnsUni->get().subObjs.end();
		     it++) {
			subObj = *it;
			switch (subObj->type()) {
				case TYPE_APPLICATION: {
					switch (subObj->subType()) {
						case SUBTYPE_APPLICATION: {
							Application_SubObj * app;
							app = static_cast<Application_SubObj *> (subObj);
							gnsTna.application.type = (grid_appname_t) app->get().app_type;
							VERSION_SET(gnsTna.application.version,
								    app->get().app_version);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, application);
					break;
				}
				case TYPE_CANDIDATE_HOSTS: {
					switch (subObj->subType()) {
						case SUBTYPE_IPv4_CANDIDATE_HOSTS: {
							IPv4CandidateHost_SubObj * ipv4Host;
							ipv4Host = static_cast<IPv4CandidateHost_SubObj *> (subObj);
							gnsTna.cand_host.value.ipv4 = ipv4Host->get().ipv4_addr;
							break;
						}
						case SUBTYPE_IPv6_CANDIDATE_HOSTS: {
							IPv6CandidateHost_SubObj * ipv6Host;
							ipv6Host = static_cast<IPv6CandidateHost_SubObj *> (subObj);
							gnsTna.cand_host.value.ipv6 = ipv6Host->get().ipv6_addr;
							break;
						}
						case SUBTYPE_NSAP_CANDIDATE_HOSTS: {
							NsapCandidateHost_SubObj * nsapHost;
							nsapHost = static_cast<NsapCandidateHost_SubObj *> (subObj);
							memcpy(&gnsTna.cand_host.value.nsap,
							       nsapHost->get().nsap_addr,
							       nsapHost->get().nsap_len);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, cand_host);
					break;
				}
				case TYPE_FILE_SYSTEM_RES: {
					switch (subObj->subType()) {
						case SUBTYPE_FILE_SYSTEM_RES: {
							FsResources_SubObj * fsRes;
							fsRes = static_cast<FsResources_SubObj *> (subObj);
							gnsTna.fs_resources.fs_name = (grid_fs_name_t)fsRes->get().fs_name;
							gnsTna.fs_resources.fs_type = (grid_fs_type_t)fsRes->get().fs_type;
							gnsTna.fs_resources.disk_space.lower_bound =
								fsRes->get().disk_space_lb;
							gnsTna.fs_resources.disk_space.upper_bound =
								fsRes->get().disk_space_ub;
							gnsTna.fs_resources.disk_space.lb_included =
								(fsRes->get().range_flags & 0x2) >> 1;
							gnsTna.fs_resources.disk_space.ub_included =
								fsRes->get().range_flags & 0x1;
							gnsTna.fs_resources.mount_point =
								new std::string(fsRes->get().mount_point);
							gnsTna.fs_resources.mount_source =
								new std::string(fsRes->get().mount_source);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, fs_resources);
					break;
				}
				case TYPE_SYSTEM_CAPS: {
					switch (subObj->subType()) {
						case SUBTYPE_SYSTEM_CAPS: {
							SystemCaps_SubObj * sysCap;
							sysCap = static_cast<SystemCaps_SubObj *> (subObj);
							VERSION_SET(gnsTna.sys_caps.os_version,
								    sysCap->get().os_version);
							gnsTna.sys_caps.os_type  =
								(grid_os_type_t) sysCap->get().os_type;
							gnsTna.sys_caps.cpu_arch =
								(grid_cpu_arch_t) sysCap->get().cpu_arch;
							gnsTna.sys_caps.exclusive_access = sysCap->get().E;
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, sys_caps);
					break;
				}
				case TYPE_INDIV_CPU_SPEED: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_SPEED: {
							IndvCpuSpeed_SubObj * ics;
							ics = static_cast<IndvCpuSpeed_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_speed,
								  ics->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_speed);
					break;
				}
				case TYPE_INDIV_CPU_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_COUNT: {
							IndvCpuCount_SubObj * icc;
							icc = static_cast<IndvCpuCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_count,
								  icc->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_count);
					break;
				}
				case TYPE_INDIV_CPU_TIME: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_CPU_TIME: {
							IndvCpuTime_SubObj * ict;
							ict = static_cast<IndvCpuTime_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_cpu_time,
								  ict->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_cpu_time);
					break;
				}
				case TYPE_INDIV_NET_BW: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_NET_BW: {
							IndvNetBw_SubObj * nbw;
							nbw = static_cast<IndvNetBw_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_net_bw,
								  nbw->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_net_bw);
					break;
				}
				case TYPE_INDIV_PHY_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_PHY_MEM: {
							IndvPhyMem_SubObj * ipm;
							ipm = static_cast<IndvPhyMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_phy_mem,
								  ipm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_phy_mem);
					break;
				}
				case TYPE_INDIV_VIR_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_VIR_MEM: {
							IndvVirMem_SubObj * ivm;
							ivm = static_cast<IndvVirMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_vir_mem,
								  ivm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_vir_mem);
					break;
				}
				case TYPE_INDIV_DISK_SPACE: {
					switch (subObj->subType()) {
						case SUBTYPE_INDIV_DISK_SPACE: {
							IndvDiskSpace_SubObj * ids;
							ids = static_cast<IndvDiskSpace_SubObj *> (subObj);
							RANGE_SET(gnsTna.ind_disk_space,
								  ids->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, ind_disk_space);
					break;
				}
				case TYPE_TOTAL_CPU_TIME: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_CPU_TIME: {
							TotCpuTime_SubObj * tct;
							tct = static_cast<TotCpuTime_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_cpu_time,
								  tct->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_cpu_time);
					break;
				}
				case TYPE_TOTAL_CPU_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_CPU_COUNT: {
							TotCpuCount_SubObj * tcc;
							tcc = static_cast<TotCpuCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_cpu_count,
								  tcc->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_cpu_count);
					break;
				}
				case TYPE_TOTAL_PHY_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_PHY_MEM: {
							TotPhyMem_SubObj * tpm;
							tpm = static_cast<TotPhyMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_phy_mem,
								  tpm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_phy_mem);
					break;
				}
				case TYPE_TOTAL_VIR_MEM: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_VIR_MEM: {
							TotVirMem_SubObj * tvm;
							tvm = static_cast<TotVirMem_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_vir_mem,
								  tvm->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_vir_mem);
					break;
				}
				case TYPE_TOTAL_DISK_SPACE: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_DISK_SPACE: {
							TotDiskSpace_SubObj * tds;
							tds = static_cast<TotDiskSpace_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_disk_space,
								  tds->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_disk_space);
					break;
				}
				case TYPE_TOTAL_RES_COUNT: {
					switch (subObj->subType()) {
						case SUBTYPE_TOTAL_RES_COUNT: {
							TotResCount_SubObj * trs;
							trs = static_cast<TotResCount_SubObj *> (subObj);
							RANGE_SET(gnsTna.tot_res_count,
								  trs->get());
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, tot_res_count);
					break;
				}
				case TYPE_DATA_STAGING: {
					switch (subObj->subType()) {
						case SUBTYPE_DATA_STAGING: {
							DataStaging_SubObj * ds;
							ds = static_cast<DataStaging_SubObj *> (subObj);
							gnsTna.data_staging.fs_name            =
								(grid_fs_name_t) ds->get().FSname;
						        gnsTna.data_staging.creation_flag      =
								(grid_creation_flag_t) ds->get().creation_flag;
							gnsTna.data_staging.del_on_termination =
								ds->get().delete_on_termination;
							gnsTna.data_staging.filename           =
								new std::string(ds->get().file_name);
							gnsTna.data_staging.source             =
								new std::string(ds->get().source);
							gnsTna.data_staging.target             =
								new std::string(ds->get().target);
							break;
						}
						default:
							G2RSVPTE_ERR("Invalid subObj type");
							return false;
					}
					BITMASK_BITSET(gnsTna.mask_, data_staging);
					break;
				}
				default:
					return false;
			}

		}

		return true;
	}

	bool ResvMessage::getRro(std::list<lsp_ero_sobj_t> & rro)
	{
		if (!flow_desc_ || !flow_desc_->getRROobj()) {
			G2RSVPTE_ERR("Rro Object pointer null !!");
			return false;
		}

		RecordedRoute_Object * rroObj;
		std::list<RroSubObject *>::iterator it;
		lsp_ero_sobj_t rroSubObj;

		rroObj = static_cast<RecordedRoute_Object *>(flow_desc_->getRROobj());

		for (it =  rroObj->get().subObjs.begin();
		     it != rroObj->get().subObjs.end();
		     it++) {
			RroSubObject * sObj;
			sObj = *it;

			if (!sObj) {
				G2RSVPTE_ERR("Invalid subObject pointer");
				return false;
			}

			memset(&rroSubObj, 0, sizeof(rroSubObj));

			switch (sObj->type()) {
				case IPV4_ADDR_SUBOBJ: {
					IPv4Addr_SubObj * ipv4;
					ipv4 = static_cast<IPv4Addr_SubObj *>(sObj);

					rroSubObj.type            = ERO_SUBOBJ_IPV4;
					rroSubObj.hop.ipv4.addr   = ipv4->get().ipv4_addr;
					rroSubObj.hop.ipv4.prefix = ipv4->get().preflen;
					rroSubObj.loose           = ERO_SUBOBJ_STRICT;

					break;
				}
				case IPV6_ADDR_SUBOBJ: {
					G2RSVPTE_ERR("IPv6 prefix subobjects not yet supported");
					return false;
				}
				case UNNUM_LINK_SUBOBJ: {
					Unnumbered_SubObj * unn;
					unn = static_cast<Unnumbered_SubObj *>(sObj);

					if (unn->get().interface_id == 0) {
						rroSubObj.type             = ERO_SUBOBJ_NODE;
						rroSubObj.hop.node.node_id = unn->get().router_id;
					} else {
						rroSubObj.type                = ERO_SUBOBJ_UNNUM;
						rroSubObj.hop.unnum.router_id = unn->get().router_id;
						rroSubObj.hop.unnum.intf_id   = unn->get().interface_id;
					}
					rroSubObj.loose = ERO_SUBOBJ_STRICT;
					break;
				}
				case LABEL_SUBOBJ: {
					Label_SubObj * lbl;
					lbl = static_cast<Label_SubObj *>(sObj);

					RSVP::Label_Class * label = lbl->get().label;
					if (!label ||
					    label->cType() != CTYPE_GENERALISED_LABEL) /* Useless??? */ {
						G2RSVPTE_ERR("Label RRO subobj with invalid pointer!");
						return false;
					}

					RSVP::GLabel_Object * genLabel;
					genLabel = static_cast<RSVP::GLabel_Object *>(label);

					rroSubObj.type               = ERO_SUBOBJ_LABEL;
					rroSubObj.hop.label.value    = *(genLabel->get().label);
					rroSubObj.hop.label.upstream = LABEL_DOWNSTREAM;
					break;
				}
				default:
					G2RSVPTE_ERR("unknown RRO subobj type (%d)",
						     sObj->type());
					return false;
			}
			rro.push_back(rroSubObj);
		}
		return true;
	}

	uint16_t ResvMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(message_id_ack_nack_, len);
		ADD_LENGTH(message_id_, len);
		ADD_LENGTH(resv_confirm_, len);
		ADD_LENGTH(scope_, len);
		ADD_LENGTH(notify_request_, len);
		ADD_LENGTH(admin_status_, len);
		ADD_LENGTH(policy_data_, len);

		ADD_LENGTH(session_, len);
		ADD_LENGTH(rsvp_hop_, len);
		ADD_LENGTH(time_values_, len);
		ADD_LENGTH(call_ops_, len);
		ADD_LENGTH(call_id_, len);
		ADD_LENGTH(gns_call_ext_, len);
		ADD_LENGTH(gns_uni_eg_, len);
		ADD_LENGTH(gns_uni_in_, len);
		ADD_LENGTH(style_, len);
		ADD_LENGTH(flow_desc_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor & mc, ResvMessage & resv)
	{
		// TODO parse all ResvMsg objs
		int length = mc.remainingSize();
		bool gnsUniEgGot = false;

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!resv.session_);
					mcObj >> resv.session_;
					assert(resv.session_);
					break;
				case G2RSVP_CLASS_TIME_VALUES:
					assert(!resv.time_values_);
					mcObj >> resv.time_values_;
					assert(resv.time_values_);
					break;
				case G2RSVP_CLASS_RSVP_HOP:
					assert(!resv.rsvp_hop_);
					mcObj >> resv.rsvp_hop_;
					assert(resv.rsvp_hop_);
					break;
				case G2RSVP_CLASS_ADMIN_STATUS:
					assert(!resv.admin_status_);
					mcObj >> resv.admin_status_;
					assert(resv.admin_status_);
					break;
				case G2RSVP_CLASS_GNS_CALL_EXT:
					assert(!resv.gns_call_ext_);
					mcObj >> resv.gns_call_ext_;
					assert(resv.gns_call_ext_);
					break;
				case G2RSVP_CLASS_INTEGRITY:
					assert(!resv.integrity_);
					mcObj >> resv.integrity_;
					assert(resv.integrity_);
					break;
				case G2RSVP_CLASS_GNS_UNI:
					if (!gnsUniEgGot) {
						assert(!resv.gns_uni_eg_);
						mcObj >> resv.gns_uni_eg_;
						assert(resv.gns_uni_eg_);
						gnsUniEgGot = true;
					} else {
						assert(!resv.gns_uni_in_);
						mcObj >> resv.gns_uni_in_;
						assert(resv.gns_uni_in_);
					}
					break;
				case G2RSVP_CLASS_CALL_OPS:
					assert(!resv.call_ops_);
					mcObj >> resv.call_ops_;
					assert(resv.call_ops_);
					break;
				case G2RSVP_CLASS_CALL_ID:
					assert(!resv.call_id_);
					mcObj >> resv.call_id_;
					assert(resv.call_id_);
					break;
				case G2RSVP_CLASS_NOTIFY_REQUEST:
					assert(!resv.notify_request_);
					mcObj >> resv.notify_request_;
					assert(resv.notify_request_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID:
					assert(!resv.message_id_);
					mcObj >> resv.message_id_;
					assert(resv.message_id_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
					assert(!resv.message_id_ack_nack_);
					mcObj >> resv.message_id_ack_nack_;
					assert(resv.message_id_ack_nack_);
					break;
				case G2RSVP_CLASS_RESV_CONFIRM:
					assert(!resv.resv_confirm_);
					mcObj >> resv.resv_confirm_;
					assert(resv.resv_confirm_);
					break;
				case G2RSVP_CLASS_SCOPE:
					assert(!resv.scope_);
					mcObj >> resv.scope_;
					assert(resv.scope_);
					break;
				case G2RSVP_CLASS_POLICY:
					assert(!resv.policy_data_);
					mcObj >> resv.policy_data_;
					assert(resv.policy_data_);
					break;
				case G2RSVP_CLASS_STYLE:
					assert(!resv.style_);
					mcObj >> resv.style_;
					assert(resv.style_);

					if (resv.style_->cType() == CTYPE_STYLE
					    && static_cast<Style_Object *>
					    (resv.style_)->isFF()) {
						assert(!resv.flow_desc_);
						FF_FlowDesc_t * ff;
						ff = new FF_FlowDesc_t();
						assert(ff);
						resv.flow_desc_ = new
							Flow_Desc_List(ff);
						assert(resv.flow_desc_);
					} else {
						throw(std::string
						      ("Only FF STYLE obj is "
						       "supported now"));
					}
					break;

				// START - objs related to FLOW DESCRIPTOR
				case G2RSVP_CLASS_FILTER_SPEC: {
					CHECK_FLOW_DESC(resv.flow_desc_);

					FilterSpec_Class * filter = 0;
					mcObj >> filter;
					assert(filter);

					if (!resv.flow_desc_->
					    setFilterSpecObj(filter)) {
						throw(std::string
						      ("Impossible to add "
						       "FilterSpec Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_LABEL: {
					CHECK_FLOW_DESC(resv.flow_desc_);

					Label_Class * label = 0;
					mcObj >> label;
					assert(label);

					if (!resv.flow_desc_->
					    setLabelObj(label)) {
						throw(std::string
						      ("Impossible to add "
						       "Label Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_RECORDED_ROUTE: {
					CHECK_FLOW_DESC(resv.flow_desc_);

					RecordedRoute_Class * rro = 0;
					mcObj >> rro;
					assert(rro);

					if (!resv.flow_desc_->
					    setRROobj(rro)) {
						throw(std::string
						      ("Impossible to add "
						       "RRO Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_FLOWSPEC: {
					CHECK_FLOW_DESC(resv.flow_desc_);

					FlowSpec_Class * flow = 0;
					mcObj >> flow;
					assert(flow);

					if (!resv.flow_desc_->
					    setFlowSpecObj(flow)) {
						throw(std::string
						      ("Impossible to add "
						       "FlowSpec Obj"));
					}
					break;
				}
				// END - objs related to FLOW DESCRIPTOR

				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_SUGGESTED_LABEL:
				case G2RSVP_CLASS_UPSTREAM_LABEL:
				case G2RSVP_CLASS_RECOVERY_LABEL:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_SENDER_TEMPLATE:
				case G2RSVP_CLASS_SENDER_TSPEC:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_ADSPEC:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_ERROR_SPEC:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in ResvMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor & mc, const ResvMessage & resv)
	{
		// All ResvMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(resv.session_, cur); //it's the first mandatory
		assert(resv.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (resv.session_->length() - 1));
		cur << *(resv.session_);

		OBJ2CURSOR(resv.integrity_, cur);
		OBJ2CURSOR(resv.message_id_ack_nack_, cur);
		OBJ2CURSOR(resv.message_id_, cur);
		//session obj should be here
		OBJ2CURSOR(resv.rsvp_hop_, cur);
		OBJ2CURSOR(resv.time_values_, cur);
		OBJ2CURSOR(resv.call_ops_, cur);
		OBJ2CURSOR(resv.call_id_, cur);
		OBJ2CURSOR(resv.gns_call_ext_, cur);
		OBJ2CURSOR(resv.gns_uni_eg_, cur);
		OBJ2CURSOR(resv.gns_uni_in_, cur);
		OBJ2CURSOR(resv.resv_confirm_, cur);
		OBJ2CURSOR(resv.scope_, cur);
		OBJ2CURSOR(resv.notify_request_, cur);
		OBJ2CURSOR(resv.admin_status_, cur);
		OBJ2CURSOR(resv.policy_data_, cur);
		OBJ2CURSOR(resv.style_, cur);
		OBJ2CURSOR(resv.flow_desc_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream & os, const ResvMessage & resv)
	{
		// All ResvMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(resv.session_, os); //it's the first mandatory
		OBJ2STREAM(resv.integrity_, os);
		OBJ2STREAM(resv.message_id_ack_nack_, os);
		OBJ2STREAM(resv.message_id_, os);
		//session obj should be here
		OBJ2STREAM(resv.rsvp_hop_, os);
		OBJ2STREAM(resv.time_values_, os);
		OBJ2STREAM(resv.call_ops_, os);
		OBJ2STREAM(resv.call_id_, os);
		OBJ2STREAM(resv.gns_call_ext_, os);
		OBJ2STREAM(resv.gns_uni_eg_, os);
		OBJ2STREAM(resv.gns_uni_in_, os);
		OBJ2STREAM(resv.resv_confirm_, os);
		OBJ2STREAM(resv.scope_, os);
		OBJ2STREAM(resv.notify_request_, os);
		OBJ2STREAM(resv.admin_status_, os);
		OBJ2STREAM(resv.policy_data_, os);
		OBJ2STREAM(resv.style_, os);
		OBJ2STREAM(resv.flow_desc_, os);

		return os;
	}


	/////////////////////////////////
	//     PathErrMessage class    //
	/////////////////////////////////
	PathErrMessage::PathErrMessage(void) : Message(PATH_ERR)
	{
		session_           = 0;
		error_spec_        = 0;
		call_ops_          = 0;
		call_id_           = 0;
		gns_call_ext_      = 0;

		integrity_         = 0;
		policy_data_       = 0;
		sender_template_   = 0;
		sender_tspec_      = 0;
		adspec_            = 0;
		rro_               = 0;
		suggested_label_   = 0;
		recovery_label_    = 0;
		upstream_label_    = 0;
	}

	PathErrMessage::PathErrMessage(uint8_t version) : Message(version, PATH_ERR)
	{
		session_           = 0;
		error_spec_        = 0;
		call_ops_          = 0;
		call_id_           = 0;
		gns_call_ext_      = 0;

		integrity_         = 0;
		policy_data_       = 0;
		sender_template_   = 0;
		sender_tspec_      = 0;
		adspec_            = 0;
		rro_               = 0;
		suggested_label_   = 0;
		recovery_label_    = 0;
		upstream_label_    = 0;
	}

	PathErrMessage::PathErrMessage(const PathMessage & path) :
		Message(path.getVersion(), PATH_ERR)
	{
		// TODO FIXME copy constructor or copy operator ???
		session_           = 0;
		error_spec_        = 0;
		call_ops_          = 0;
		call_id_           = 0;
		gns_call_ext_      = 0;

		integrity_         = 0;
		policy_data_       = 0;
		sender_template_   = 0;
		sender_tspec_      = 0;
		adspec_            = 0;
		rro_               = 0;
		suggested_label_   = 0;
		recovery_label_    = 0;
		upstream_label_    = 0;

		//TODO - FIXME
		//assert(path.gns_call_ext_);

		if (path.session_) {
			session_ = path.session_->clone();
			if (!session_) {
				throw(std::string("Impossible to allocate sess"
						  "ion obj for PathErr msg!"));
			}
		}

		if (path.sender_template_) {
			sender_template_ = path.sender_template_->clone();
			if (!sender_template_) {
				//delete session_;
				throw(std::string("Impossible to allocate "
						  "sender template obj for "
						  "PathErr msg!"));
			}
		}

		if (path.sender_tspec_) {
			sender_tspec_ = path.sender_tspec_->clone();
			if (!sender_tspec_) {
				//delete sender_template_;
				//delete session_;
				throw(std::string("Impossible to allocate "
						  "sender tspec obj for "
						  "PathErr msg!"));
			}
		}

		if (path.call_ops_) {
			call_ops_ = path.call_ops_->clone();
			if (!call_ops_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				throw(std::string("Impossible to allocate "
						  "Call Ops obj for PathErr "
						  "msg!"));
			}
		}

		if (path.call_id_) {
			call_id_ = path.call_id_->clone();
			if (!call_id_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				//delete call_ops_;
				throw(std::string("Impossible to allocate "
						  "Call Id obj for PathErr "
						  "msg!"));
			}
		}

		if (path.gns_call_ext_) {
			gns_call_ext_ = path.gns_call_ext_->clone();
			if (!gns_call_ext_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				//delete call_ops_;
				//delete call_id_;
				throw(std::string("Impossible to allocate GNS "
						  "Call Ext obj for PathErr "
						  "msg!"));
			}
		}

		//TODO add ERROR_SPEC (cb ? ERRSPEC_IFIDIPV4 : ERRSPEC_CT_IPV4)
	}

	PathErrMessage::PathErrMessage(const PathErrMessage & pathErr) :
		Message(PATH_ERR)
	{
		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!
	}

	PathErrMessage::~PathErrMessage(void)
	{
		G2RSVPTE_DBG("Destroying PethErrMessage %p ...", this);

		DELETE_OBJ(session_);
		DELETE_OBJ(error_spec_);
		DELETE_OBJ(call_ops_);
		DELETE_OBJ(call_id_);
		DELETE_OBJ(gns_call_ext_);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(policy_data_);
		DELETE_OBJ(sender_template_);
		DELETE_OBJ(sender_tspec_);
		DELETE_OBJ(adspec_);
		DELETE_OBJ(rro_);
		DELETE_OBJ(suggested_label_);
		DELETE_OBJ(recovery_label_);
		DELETE_OBJ(upstream_label_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	//PathErrMessage * PathErrMessage::partialClone(void) const

	bool PathErrMessage::getErrorDetails(error_info_t &              error,
					     std::list<lsp_ero_sobj_t> & excludeEro)
	{
		if (!error_spec_) {
			G2RSVPTE_ERR("Invalid error_spec obj");
			return false;
		}

		// TODO insert method getErrVal, getErrCode, ...
		//       into errorSpec base class
		switch (error_spec_->cType()) {
			case CTYPE_ERROR_SPEC_IPV4: {
				RSVP::IPv4ErrorSpec_Object * tmp;
				tmp = static_cast<RSVP::IPv4ErrorSpec_Object *>
					(error_spec_);
				error.node_id  = ntohl(tmp->get().ipv4_err_node.s_addr);
				error.flags    = tmp->get().flags;
				error.err_code = tmp->get().error_code;
				error.err_val  = tmp->get().error_value;
				break;
			}
			case CTYPE_ERROR_SPEC_IPV6: {
				G2RSVPTE_WRN("for now CTYPE_ERROR_SPEC_IPV6 "
					     "is not supported!");
				return false;
			}
			case CTYPE_ERROR_SPEC_IF_ID_IPV4: {
				RSVP::IFIDIPv4ErrorSpec_Object * tmp;
				tmp = static_cast<RSVP::
					IFIDIPv4ErrorSpec_Object*>(error_spec_);
				error.node_id  = ntohl(tmp->get().ipv4_err_node.s_addr);
				error.flags    = tmp->get().flags;
				error.err_code = tmp->get().error_code;
				error.err_val  = tmp->get().error_value;
#ifdef CRANKBACK_ON
				//process TLVs
				std::list<TLV *>::const_iterator it;
				lsp_ero_sobj_t                   sObj;
				//add on top of exclude ERO an ERO_SUBOBJ_NODE
				//subobj to specify errored node
				memset(&sObj, 0, sizeof(sObj));
				sObj.type             = ERO_SUBOBJ_NODE;
				sObj.hop.node.node_id = error.node_id;
				sObj.loose            = ERO_SUBOBJ_STRICT;
				excludeEro.push_back(sObj);
				for (it  = tmp->get().tlvs.begin();
				     it != tmp->get().tlvs.end();
				     it++) {
					memset(&sObj, 0, sizeof(sObj));
					switch ((*it)->type()) {
						case IF_ID_TLV_IPV4: {
							IPv4_TLV * tlv;
							tlv = static_cast<IPv4_TLV *> (*it);
							sObj.type                 = ERO_SUBOBJ_IPV4;
							sObj.hop.ipv4.addr.s_addr = ntohl(tlv->get().ipv4_addr.s_addr);
							sObj.hop.ipv4.prefix      = 32;
							sObj.loose                = ERO_SUBOBJ_STRICT;
							//push into exclude Ero
							excludeEro.push_back(sObj);
							break;
						}
						case IF_ID_TLV_DOWN_LABEL: {
							DOWN_LABEL_TLV * tlv;
							tlv = static_cast<DOWN_LABEL_TLV *> (*it);
							sObj.type = ERO_SUBOBJ_LABEL;
							sObj.hop.label.value    = tlv->get().label;
							sObj.hop.label.upstream = LABEL_DOWNSTREAM;
							sObj.loose              = ERO_SUBOBJ_STRICT;
							//push into exclude Ero
							excludeEro.push_back(sObj);
							break;
						}
						case IF_ID_TLV_UP_LABEL: {
							UP_LABEL_TLV * tlv;
							tlv = static_cast<UP_LABEL_TLV *> (*it);
							sObj.type = ERO_SUBOBJ_LABEL;
							sObj.hop.label.value    = tlv->get().label;
							sObj.hop.label.upstream = LABEL_UPSTREAM;
							sObj.loose              = ERO_SUBOBJ_STRICT;
							//push into exclude Ero
							excludeEro.push_back(sObj);
							break;
						}
						default:
							G2RSVPTE_ERR("Unknown TLV type (%d) for IF ID "
								     "IPv4 Error Spec obj", (*it)->type());
							return false;
					}
				}
#endif // CRANKBACK_ON
				break;
			}
			case CTYPE_ERROR_SPEC_IF_ID_IPV6: {
				G2RSVPTE_WRN("for now CTYPE_IF_ID_ERROR_"
					     "SPEC_IPV6 is not supported!");
				return false;
			}
			default:
				G2RSVPTE_ERR("unknown error spec obj type (%d)",
					     error_spec_->cType());
				return false;
		}

		return true;
	}

	bool PathErrMessage::getRro(std::list<lsp_ero_sobj_t> & rro)
	{
		if (!rro_) {
			G2RSVPTE_ERR("Rro Object pointer null !!");
			return false;
		}

		RecordedRoute_Object * rroObj;
		std::list<RroSubObject *>::iterator it;
		lsp_ero_sobj_t rroSubObj;

		rroObj = static_cast<RecordedRoute_Object *>(rro_);

		for (it =  rroObj->get().subObjs.begin();
		     it != rroObj->get().subObjs.end();
		     it++) {
			RroSubObject * sObj;
			sObj = *it;

			if (!sObj) {
				G2RSVPTE_ERR("Invalid subObject pointer");
				return false;
			}

			memset(&rroSubObj, 0, sizeof(rroSubObj));

			switch (sObj->type()) {
				case IPV4_ADDR_SUBOBJ: {
					IPv4Addr_SubObj * ipv4;
					ipv4 = static_cast<IPv4Addr_SubObj *>(sObj);

					rroSubObj.type            = ERO_SUBOBJ_IPV4;
					rroSubObj.hop.ipv4.addr   = ipv4->get().ipv4_addr;
					rroSubObj.hop.ipv4.prefix = ipv4->get().preflen;
					rroSubObj.loose           = ERO_SUBOBJ_STRICT;

					break;
				}
				case IPV6_ADDR_SUBOBJ: {
					G2RSVPTE_ERR("IPv6 prefix subobjects not yet supported");
					return false;
				}
				case UNNUM_LINK_SUBOBJ: {
					Unnumbered_SubObj * unn;
					unn = static_cast<Unnumbered_SubObj *>(sObj);

					if (unn->get().interface_id == 0) {
						rroSubObj.type             = ERO_SUBOBJ_NODE;
						rroSubObj.hop.node.node_id = unn->get().router_id;
					} else {
						rroSubObj.type                = ERO_SUBOBJ_UNNUM;
						rroSubObj.hop.unnum.router_id = unn->get().router_id;
						rroSubObj.hop.unnum.intf_id   = unn->get().interface_id;
					}
					rroSubObj.loose = ERO_SUBOBJ_STRICT;
					break;
				}
				case LABEL_SUBOBJ: {
					Label_SubObj * lbl;
					lbl = static_cast<Label_SubObj *>(sObj);

					RSVP::Label_Class * label = lbl->get().label;
					if (!label ||
					    label->cType() != CTYPE_GENERALISED_LABEL) /* Useless??? */ {
						G2RSVPTE_ERR("Label RRO subobj with invalid pointer!");
						return false;
					}

					RSVP::GLabel_Object * genLabel;
					genLabel = static_cast<RSVP::GLabel_Object *>(label);

					rroSubObj.type               = ERO_SUBOBJ_LABEL;
					rroSubObj.hop.label.value    = *(genLabel->get().label);
					rroSubObj.hop.label.upstream = LABEL_DOWNSTREAM;
					break;
				}
				default:
					G2RSVPTE_ERR("unknown RRO subobj type (%d)",
						     sObj->type());
					return false;
			}
			rro.push_back(rroSubObj);
		}
		return true;
	}

	Message * PathErrMessage::clone(void) const
	{
		PathErrMessage * msg = new PathErrMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(error_spec_, msg);
		CLONE_OBJ_INTO_MSG(call_ops_, msg);
		CLONE_OBJ_INTO_MSG(call_id_, msg);
		CLONE_OBJ_INTO_MSG(gns_call_ext_, msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(policy_data_, msg);
		CLONE_OBJ_INTO_MSG(sender_template_, msg);
		CLONE_OBJ_INTO_MSG(sender_tspec_, msg);
		CLONE_OBJ_INTO_MSG(adspec_, msg);
		CLONE_OBJ_INTO_MSG(rro_, msg);
		CLONE_OBJ_INTO_MSG(suggested_label_, msg);
		CLONE_OBJ_INTO_MSG(recovery_label_, msg);
		CLONE_OBJ_INTO_MSG(upstream_label_, msg);

		return msg;
	}

	bool PathErrMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(error_spec_);
		// TODO add check for call relative objects
		//CHECK_MANDATORY_OBJ(call_ops_);
		//CHECK_MANDATORY_OBJ(call_id_);
		//CHECK_MANDATORY_OBJ(gns_call_ext_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(policy_data_);
		CHECK_OPTIONAL_OBJ(sender_template_);
		CHECK_OPTIONAL_OBJ(sender_tspec_);
		CHECK_OPTIONAL_OBJ(adspec_);
		CHECK_OPTIONAL_OBJ(rro_);
		CHECK_OPTIONAL_OBJ(suggested_label_);
		CHECK_OPTIONAL_OBJ(recovery_label_);
		CHECK_OPTIONAL_OBJ(upstream_label_);

		return true;
	}

	bool PathErrMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (sender_template_) {
			GETLSPINFO_FROM_SENDTEMP(sender_template_, &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t PathErrMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(session_, len);
		ADD_LENGTH(error_spec_, len);
		ADD_LENGTH(call_ops_, len);
		ADD_LENGTH(call_id_, len);
		ADD_LENGTH(gns_call_ext_, len);

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(policy_data_, len);
		ADD_LENGTH(sender_template_, len);
		ADD_LENGTH(sender_tspec_, len);
		ADD_LENGTH(adspec_, len);
		ADD_LENGTH(rro_, len);
		ADD_LENGTH(suggested_label_, len);
		ADD_LENGTH(recovery_label_, len);
		ADD_LENGTH(upstream_label_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor & mc, PathErrMessage & pathErr)
	{
		// TODO parse all PathErrMsg objs
		int length = mc.remainingSize();

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!pathErr.session_);
					mcObj >> pathErr.session_;
					assert(pathErr.session_);
					break;
				case G2RSVP_CLASS_GNS_CALL_EXT:
					assert(!pathErr.gns_call_ext_);
					mcObj >> pathErr.gns_call_ext_;
					assert(pathErr.gns_call_ext_);
					break;
				case G2RSVP_CLASS_CALL_OPS:
					assert(!pathErr.call_ops_);
					mcObj >> pathErr.call_ops_;
					assert(pathErr.call_ops_);
					break;
				case G2RSVP_CLASS_CALL_ID:
					assert(!pathErr.call_id_);
					mcObj >> pathErr.call_id_;
					assert(pathErr.call_id_);
					break;
				case G2RSVP_CLASS_ERROR_SPEC:
					assert(!pathErr.error_spec_);
					mcObj >> pathErr.error_spec_;
					assert(pathErr.error_spec_);
					break;

				case G2RSVP_CLASS_INTEGRITY:
					assert(!pathErr.integrity_);
					mcObj >> pathErr.integrity_;
					assert(pathErr.integrity_);
					break;
				case G2RSVP_CLASS_POLICY:
					assert(!pathErr.policy_data_);
					mcObj >> pathErr.policy_data_;
					assert(pathErr.policy_data_);
					break;
				case G2RSVP_CLASS_SENDER_TEMPLATE:
					assert(!pathErr.sender_template_);
					mcObj >> pathErr.sender_template_;
					assert(pathErr.sender_template_);
					break;
				case G2RSVP_CLASS_SENDER_TSPEC:
					assert(!pathErr.sender_tspec_);
					mcObj >> pathErr.sender_tspec_;
					assert(pathErr.sender_tspec_);
					break;
				case G2RSVP_CLASS_ADSPEC:
					// to be filled up
					throw(std::string("ADSPEC !supported"));
				case G2RSVP_CLASS_RECORDED_ROUTE:
					assert(!pathErr.rro_);
					mcObj >> pathErr.rro_;
					assert(pathErr.rro_);
					break;
				case G2RSVP_CLASS_SUGGESTED_LABEL:
					assert(!pathErr.suggested_label_);
					mcObj >> pathErr.suggested_label_;
					assert(pathErr.suggested_label_);
					break;
				case G2RSVP_CLASS_UPSTREAM_LABEL:
					assert(!pathErr.upstream_label_);
					mcObj >> pathErr.upstream_label_;
					assert(pathErr.upstream_label_);
					break;
				case G2RSVP_CLASS_RECOVERY_LABEL:
					assert(!pathErr.recovery_label_);
					mcObj >> pathErr.recovery_label_;
					assert(pathErr.recovery_label_);
					break;

				case G2RSVP_CLASS_FILTER_SPEC:
				case G2RSVP_CLASS_LABEL:
				case G2RSVP_CLASS_FLOWSPEC:
				case G2RSVP_CLASS_TIME_VALUES:
				case G2RSVP_CLASS_RSVP_HOP:
				case G2RSVP_CLASS_ADMIN_STATUS:
				case G2RSVP_CLASS_NOTIFY_REQUEST:
				case G2RSVP_CLASS_MESSAGE_ID:
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
				case G2RSVP_CLASS_RESV_CONFIRM:
				case G2RSVP_CLASS_SCOPE:
				case G2RSVP_CLASS_STYLE:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_GNS_UNI:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in PathErrMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor &         mc,
				    const PathErrMessage & pathErr)
	{
		// All PathErrMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(pathErr.session_, cur); //it's the first mandatory
		assert(pathErr.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (pathErr.session_->length() - 1));
		cur << *(pathErr.session_);

		OBJ2CURSOR(pathErr.integrity_, cur);
		//OBJ2CURSOR(pathErr.session_, cur);
		OBJ2CURSOR(pathErr.error_spec_, cur);
		OBJ2CURSOR(pathErr.policy_data_, cur);
		OBJ2CURSOR(pathErr.call_ops_, cur);
		OBJ2CURSOR(pathErr.call_id_, cur);
		OBJ2CURSOR(pathErr.gns_call_ext_, cur);
		OBJ2CURSOR(pathErr.sender_template_, cur);
		OBJ2CURSOR(pathErr.sender_tspec_, cur);
		OBJ2CURSOR(pathErr.adspec_, cur);
		OBJ2CURSOR(pathErr.rro_, cur);
		OBJ2CURSOR(pathErr.suggested_label_, cur);
		OBJ2CURSOR(pathErr.recovery_label_, cur);
		OBJ2CURSOR(pathErr.upstream_label_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream &         os,
				    const PathErrMessage & pathErr)
	{
		// All PathErrMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(pathErr.integrity_, os);
		OBJ2STREAM(pathErr.session_, os);
		OBJ2STREAM(pathErr.error_spec_, os);
		OBJ2STREAM(pathErr.policy_data_, os);
		OBJ2STREAM(pathErr.call_ops_, os);
		OBJ2STREAM(pathErr.call_id_, os);
		OBJ2STREAM(pathErr.gns_call_ext_, os);
		OBJ2STREAM(pathErr.sender_template_, os);
		OBJ2STREAM(pathErr.sender_tspec_, os);
		OBJ2STREAM(pathErr.adspec_, os);
		OBJ2STREAM(pathErr.rro_, os);
		OBJ2STREAM(pathErr.suggested_label_, os);
		OBJ2STREAM(pathErr.recovery_label_, os);
		OBJ2STREAM(pathErr.upstream_label_, os);

		return os;
	}


	/////////////////////////////////
	//     ResvErrMessage class    //
	/////////////////////////////////
	ResvErrMessage::ResvErrMessage(void) : Message(RESV_ERR)
	{
		session_              = 0;
		error_spec_           = 0;
		style_                = 0;
		rsvp_hop_             = 0;
		flow_desc_            = 0;
		error_spec_           = 0;

		integrity_            = 0;
		policy_data_          = 0;
		message_id_ack_nack_  = 0;
		message_id_           = 0;
		policy_data_          = 0;
		scope_                = 0;
		acceptable_label_set_ = 0;
	}

	ResvErrMessage::ResvErrMessage(uint8_t version) : Message(version, RESV_ERR)
	{
		session_              = 0;
		error_spec_           = 0;
		style_                = 0;
		rsvp_hop_             = 0;
		flow_desc_            = 0;
		error_spec_           = 0;

		integrity_            = 0;
		policy_data_          = 0;
		message_id_ack_nack_  = 0;
		message_id_           = 0;
		policy_data_          = 0;
		scope_                = 0;
		acceptable_label_set_ = 0;
	}

	ResvErrMessage::ResvErrMessage(const ResvErrMessage & resvErr) :
		Message(RESV_ERR)
	{
		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!
	}

	ResvErrMessage::ResvErrMessage(const ResvMessage & resv) :
		Message(resv.getVersion(), RESV_ERR)
	{
		// TODO FIXME copy constructor or copy operator ???
		// ... at least mandatory objects

		session_              = 0;
		style_                = 0;
		rsvp_hop_             = 0;
		flow_desc_            = 0;
		error_spec_           = 0;

		integrity_            = 0;
		policy_data_          = 0;
		message_id_ack_nack_  = 0;
		message_id_           = 0;
		policy_data_          = 0;
		scope_                = 0;
		acceptable_label_set_ = 0;

		// TODO FIX add // = new IPv4ErrorSpec_Object(); ???
		error_spec_ = 0;

		//FIX Mandatory???
		assert(resv.rsvp_hop_);
		rsvp_hop_ = resv.rsvp_hop_->clone();
		if (!rsvp_hop_) {
			throw(std::string("Impossible to allocate "
					  "rsvp_hop obj for ResvErr msg!"));
		}

		if (resv.session_) {
			session_ = resv.session_->clone();
			if (!session_) {
				//delete rsvp_hop_;
				throw(std::string("Impossible to allocate sessi"
						  "on obj for ResvErr msg!"));
			}
		}

		if (resv.style_) {
			style_ = resv.style_->clone();
			if (!style_) {
				//delete rsvp_hop_;
				//delete session_;
				throw(std::string("Impossible to allocate style"
						  " obj for ResvErr msg!"));
			}
		}

		//FIX label/filterSpec together, label Mandatory???
		if (resv.flow_desc_) {
			flow_desc_ = resv.flow_desc_->clone();
			if (!flow_desc_) {
				//delete rsvp_hop_;
				//delete session_;
				//delete style_;
				throw(std::string("Impossible to allocate Flow "
						  "Desc obj for ResvErr msg!"));
			}
		}
	}

	ResvErrMessage::~ResvErrMessage(void)
	{
		G2RSVPTE_DBG("Destroying ResvErrMessage %p ...", this);

		DELETE_OBJ(session_);
		DELETE_OBJ(error_spec_);
		DELETE_OBJ(style_);
		DELETE_OBJ(rsvp_hop_);
		DELETE_OBJ(flow_desc_);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(message_id_ack_nack_);
		DELETE_OBJ(message_id_);
		DELETE_OBJ(policy_data_);
		DELETE_OBJ(scope_);
		DELETE_OBJ(acceptable_label_set_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	Message * ResvErrMessage::clone(void) const
	{
		ResvErrMessage * msg = new ResvErrMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(error_spec_, msg);
		CLONE_OBJ_INTO_MSG(style_, msg);
		CLONE_OBJ_INTO_MSG(rsvp_hop_, msg);
		CLONE_OBJ_INTO_MSG(flow_desc_, msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(message_id_ack_nack_, msg);
		CLONE_OBJ_INTO_MSG(message_id_, msg);
		CLONE_OBJ_INTO_MSG(policy_data_, msg);
		CLONE_OBJ_INTO_MSG(scope_, msg);
		CLONE_OBJ_INTO_MSG(acceptable_label_set_, msg);

		return msg;
	}

	bool ResvErrMessage::getErrorDetails(error_info_t & error)
	{
		if (!error_spec_) {
			G2RSVPTE_ERR("Invalid error_spec obj");
			return false;
		}

		// TODO insert method getErrVal, getErrCode, ...
		//       into errorSpec base class
		switch (error_spec_->cType()) {
			case CTYPE_ERROR_SPEC_IPV4: {
				RSVP::IPv4ErrorSpec_Object * tmp;
				tmp = static_cast<RSVP::IPv4ErrorSpec_Object *>
					(error_spec_);
				error.node_id  = ntohl(tmp->get().ipv4_err_node.s_addr);
				error.flags    = tmp->get().flags;
				error.err_code = tmp->get().error_code;
				error.err_val  = tmp->get().error_value;
				break;
			}
			case CTYPE_ERROR_SPEC_IPV6: {
				G2RSVPTE_WRN("for now CTYPE_ERROR_SPEC_IPV6 "
					     "is not supported!");
				return false;
			}
			case CTYPE_ERROR_SPEC_IF_ID_IPV4: {
				RSVP::IFIDIPv4ErrorSpec_Object * tmp;
				tmp = static_cast<RSVP::
					IFIDIPv4ErrorSpec_Object*>(error_spec_);
				error.node_id  = ntohl(tmp->get().ipv4_err_node.s_addr);
				error.flags    = tmp->get().flags;
				error.err_code = tmp->get().error_code;
				error.err_val  = tmp->get().error_value;
				break;
			}
			case CTYPE_ERROR_SPEC_IF_ID_IPV6: {
				G2RSVPTE_WRN("for now CTYPE_IF_ID_ERROR_"
					     "SPEC_IPV6 is not supported!");
				return false;
			}
			default:
				G2RSVPTE_ERR("unknown error spec obj type (%d)",
					     error_spec_->cType());
				return false;
		}

		return true;
	}

	bool ResvErrMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(error_spec_);
		CHECK_MANDATORY_OBJ(style_);
		CHECK_MANDATORY_OBJ(rsvp_hop_);
		CHECK_MANDATORY_OBJ(flow_desc_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(message_id_ack_nack_);
		CHECK_OPTIONAL_OBJ(message_id_);
		CHECK_OPTIONAL_OBJ(policy_data_);
		CHECK_OPTIONAL_OBJ(scope_);
		CHECK_OPTIONAL_OBJ(acceptable_label_set_);

		return true;
	}

	bool ResvErrMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (flow_desc_ && flow_desc_->getFilterSpecObj()) {
			GETLSPINFO_FROM_FILTERSPEC(flow_desc_->
						   getFilterSpecObj(),
						   &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t ResvErrMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(session_, len);
		ADD_LENGTH(error_spec_, len);
		ADD_LENGTH(style_, len);
		ADD_LENGTH(rsvp_hop_, len);
		ADD_LENGTH(flow_desc_, len);

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(message_id_ack_nack_, len);
		ADD_LENGTH(message_id_, len);
		ADD_LENGTH(policy_data_, len);
		ADD_LENGTH(scope_, len);
		ADD_LENGTH(acceptable_label_set_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor & mc, ResvErrMessage & resvErr)
	{
		// TODO parse all ResvErrMsg objs
		int length = mc.remainingSize();

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!resvErr.session_);
					mcObj >> resvErr.session_;
					assert(resvErr.session_);
					break;
				case G2RSVP_CLASS_RSVP_HOP:
					assert(!resvErr.rsvp_hop_);
					mcObj >> resvErr.rsvp_hop_;
					assert(resvErr.rsvp_hop_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID:
					assert(!resvErr.message_id_);
					mcObj >> resvErr.message_id_;
					assert(resvErr.message_id_);
					break;
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
					assert(!resvErr.message_id_ack_nack_);
					mcObj >> resvErr.message_id_ack_nack_;
					assert(resvErr.message_id_ack_nack_);
					break;
				case G2RSVP_CLASS_POLICY:
					assert(!resvErr.policy_data_);
					mcObj >> resvErr.policy_data_;
					assert(resvErr.policy_data_);
					break;
				case G2RSVP_CLASS_ERROR_SPEC:
					assert(!resvErr.error_spec_);
					mcObj >> resvErr.error_spec_;
					assert(resvErr.error_spec_);
					break;
				case G2RSVP_CLASS_INTEGRITY:
					assert(!resvErr.integrity_);
					mcObj >> resvErr.integrity_;
					assert(resvErr.integrity_);
					break;
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
					assert(!resvErr.acceptable_label_set_);
					mcObj >> resvErr.acceptable_label_set_;
					assert(resvErr.acceptable_label_set_);
					break;
				case G2RSVP_CLASS_SCOPE:
					assert(!resvErr.scope_);
					mcObj >> resvErr.scope_;
					assert(resvErr.scope_);
					break;
				case G2RSVP_CLASS_STYLE:
					assert(!resvErr.style_);
					mcObj >> resvErr.style_;
					assert(resvErr.style_);

					if (resvErr.style_->cType() ==
					    CTYPE_STYLE                 &&
					    static_cast<Style_Object *>
					    (resvErr.style_)->isFF()) {
						assert(!resvErr.flow_desc_);
						FF_FlowDesc_t * ff;
						ff = new FF_FlowDesc_t();
						assert(ff);
						resvErr.flow_desc_ = new
							Flow_Desc_List(ff);
						assert(resvErr.flow_desc_);
					} else {
						throw(std::string
						      ("Only FF STYLE obj is "
						       "supported now"));
					}
					break;

				// START - objs related to FLOW DESCRIPTOR
				case G2RSVP_CLASS_FILTER_SPEC: {
					CHECK_FLOW_DESC(resvErr.flow_desc_);

					FilterSpec_Class * filter = 0;
					mcObj >> filter;
					assert(filter);

					if (!resvErr.flow_desc_->
					    setFilterSpecObj(filter)) {
						throw(std::string
						      ("Impossible to add "
						       "FilterSpec Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_LABEL: {
					CHECK_FLOW_DESC(resvErr.flow_desc_);

					Label_Class * label = 0;
					mcObj >> label;
					assert(label);

					if (!resvErr.flow_desc_->
					    setLabelObj(label)) {
						throw(std::string
						      ("Impossible to add "
						       "Label Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_RECORDED_ROUTE: {
					CHECK_FLOW_DESC(resvErr.flow_desc_);

					RecordedRoute_Class * rro = 0;
					mcObj >> rro;
					assert(rro);

					if (!resvErr.flow_desc_->
					    setRROobj(rro)) {
						throw(std::string
						      ("Impossible to add "
						       "RRO Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_FLOWSPEC: {
					CHECK_FLOW_DESC(resvErr.flow_desc_);

					FlowSpec_Class * flow = 0;
					mcObj >> flow;
					assert(flow);

					if (!resvErr.flow_desc_->
					    setFlowSpecObj(flow)) {
						throw(std::string
						      ("Impossible to add "
						       "FlowSpec Obj"));
					}
					break;
				}
				// END - objs related to FLOW DESCRIPTOR

				case G2RSVP_CLASS_TIME_VALUES:
				case G2RSVP_CLASS_ADMIN_STATUS:
				case G2RSVP_CLASS_GNS_CALL_EXT:
				case G2RSVP_CLASS_CALL_OPS:
				case G2RSVP_CLASS_CALL_ID:
				case G2RSVP_CLASS_NOTIFY_REQUEST:
				case G2RSVP_CLASS_RESV_CONFIRM:
				case G2RSVP_CLASS_SUGGESTED_LABEL:
				case G2RSVP_CLASS_UPSTREAM_LABEL:
				case G2RSVP_CLASS_RECOVERY_LABEL:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_SENDER_TEMPLATE:
				case G2RSVP_CLASS_SENDER_TSPEC:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_GNS_UNI:
				case G2RSVP_CLASS_ADSPEC:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in ResvErrMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor &         mc,
				    const ResvErrMessage & resvErr)
	{
		// All ResvErrMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(resvErr.session_, cur); //it's the first mandatory
		assert(resvErr.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (resvErr.session_->length() - 1));
		cur << *(resvErr.session_);

		OBJ2CURSOR(resvErr.integrity_, cur);
		OBJ2CURSOR(resvErr.message_id_ack_nack_, cur);
		OBJ2CURSOR(resvErr.message_id_, cur);
		//OBJ2CURSOR(resvErr.session_, cur);
		OBJ2CURSOR(resvErr.rsvp_hop_, cur);
		OBJ2CURSOR(resvErr.error_spec_, cur);
		OBJ2CURSOR(resvErr.scope_, cur);
		OBJ2CURSOR(resvErr.acceptable_label_set_, cur);
		OBJ2CURSOR(resvErr.policy_data_, cur);
		OBJ2CURSOR(resvErr.style_, cur);
		OBJ2CURSOR(resvErr.flow_desc_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream &         os,
				    const ResvErrMessage & resvErr)
	{
		// All ResvErrMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(resvErr.session_, os); //it's the first mandatory
		OBJ2STREAM(resvErr.integrity_, os);
		OBJ2STREAM(resvErr.message_id_ack_nack_, os);
		OBJ2STREAM(resvErr.message_id_, os);
		//OBJ2STREAM(resvErr.session_, os);
		OBJ2STREAM(resvErr.rsvp_hop_, os);
		OBJ2STREAM(resvErr.error_spec_, os);
		OBJ2STREAM(resvErr.scope_, os);
		OBJ2STREAM(resvErr.acceptable_label_set_, os);
		OBJ2STREAM(resvErr.policy_data_, os);
		OBJ2STREAM(resvErr.style_, os);
		OBJ2STREAM(resvErr.flow_desc_, os);

		return os;
	}


	/////////////////////////////////
	//    PathTearMessage class    //
	/////////////////////////////////
	PathTearMessage::PathTearMessage(void) : Message(PATH_TEAR)
	{
			session_           = 0;
			rsvp_hop_          = 0;
			call_ops_          = 0;
			call_id_           = 0;
			gns_call_ext_      = 0;

			integrity_         = 0;
			sender_template_   = 0;
			sender_tspec_      = 0;
			adspec_            = 0;
			rro_               = 0;
			suggested_label_   = 0;
			recovery_label_    = 0;
			upstream_label_    = 0;
	}

	PathTearMessage::PathTearMessage(uint8_t version) : Message(version, PATH_TEAR)
	{
			session_           = 0;
			rsvp_hop_          = 0;
			call_ops_          = 0;
			call_id_           = 0;
			gns_call_ext_      = 0;

			integrity_         = 0;
			sender_template_   = 0;
			sender_tspec_      = 0;
			adspec_            = 0;
			rro_               = 0;
			suggested_label_   = 0;
			recovery_label_    = 0;
			upstream_label_    = 0;
	}

	PathTearMessage::PathTearMessage(const PathMessage & path) :
		Message(path.getVersion(), PATH_TEAR)
	{
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields from a path msg!!!

		session_           = 0;
		call_ops_          = 0; // = new ???
		call_id_           = 0; // = new ???
		gns_call_ext_      = 0;

		integrity_         = 0;
		sender_template_   = 0;
		sender_tspec_      = 0;
		adspec_            = 0;
		rro_               = 0;
		suggested_label_   = 0;
		recovery_label_    = 0;
		upstream_label_    = 0;

		//FIX Mandatory???
		assert(path.rsvp_hop_);
		rsvp_hop_ = path.rsvp_hop_->clone();
		if (!rsvp_hop_) {
			throw(std::string("Impossible to allocate rsvp_hop obj "
					  " for PathTear msg!"));
		}

		if (path.session_) {
			session_ = path.session_->clone();
			if (!session_) {
				//delete rsvp_hop_;
				throw(std::string("Impossible to allocate sess"
						  "ion obj for PathTear msg!"));
			}
		}

		if (path.sender_template_) {
			sender_template_ = path.sender_template_->clone();
			if (!sender_template_) {
				//delete rsvp_hop_;
				//delete session_;
				throw(std::string("Impossible to allocate "
						  "sender template obj for "
						  "PathTear msg!"));
			}
		}

		if (path.sender_tspec_) {
			sender_tspec_ = path.sender_tspec_->clone();
			if (!sender_tspec_) {
				//delete rsvp_hop_;
				//delete sender_template_;
				//delete session_;
				throw(std::string("Impossible to allocate "
						  "sender tspec obj for "
						  "PathTear msg!"));
			}
		}

		if (path.call_ops_) {
			call_ops_ = path.call_ops_->clone();
			if (!call_ops_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				throw(std::string("Impossible to allocate "
						  "Call Ops obj for PathTear "
						  "msg!"));
			}
		}

		if (path.call_id_) {
			call_id_ = path.call_id_->clone();
			if (!call_id_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				//delete call_ops_;
				throw(std::string("Impossible to allocate "
						  "Call Id obj for PathTear "
						  "msg!"));
			}
		}

		if (path.gns_call_ext_) {
			gns_call_ext_ = path.gns_call_ext_->clone();
			if (!gns_call_ext_) {
				//delete sender_template_;
				//delete session_;
				//delete sender_tspec_;
				//delete call_ops_;
				//delete call_id_;
				throw(std::string("Impossible to allocate GNS "
						  "Call Ext obj for PathTear "
						  "msg!"));
			}
		}

		// TODO: error_spec???  if yes -> ERROR_SPEC_IPV4 ???
	}

	PathTearMessage::PathTearMessage(const PathTearMessage & pathTear) :
		Message(PATH_TEAR)
	{
		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!
	}

	PathTearMessage::~PathTearMessage(void)
	{
		G2RSVPTE_DBG("Destroying PathTearMessage %p ...", this);

		DELETE_OBJ(session_);
		DELETE_OBJ(rsvp_hop_);
		DELETE_OBJ(call_ops_);
		DELETE_OBJ(call_id_);
		DELETE_OBJ(gns_call_ext_);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(sender_template_);
		DELETE_OBJ(sender_tspec_);
		DELETE_OBJ(adspec_);
		DELETE_OBJ(rro_);
		DELETE_OBJ(suggested_label_);
		DELETE_OBJ(recovery_label_);
		DELETE_OBJ(upstream_label_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	//PathTearMessage * PathTearMessage::partialClone(void) const

	Message * PathTearMessage::clone(void) const
	{
		PathTearMessage * msg = new PathTearMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(rsvp_hop_, msg);
		CLONE_OBJ_INTO_MSG(call_ops_, msg);
		CLONE_OBJ_INTO_MSG(call_id_, msg);
		CLONE_OBJ_INTO_MSG(gns_call_ext_, msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(sender_template_, msg);
		CLONE_OBJ_INTO_MSG(sender_tspec_, msg);
		CLONE_OBJ_INTO_MSG(adspec_, msg);
		CLONE_OBJ_INTO_MSG(rro_, msg);
		CLONE_OBJ_INTO_MSG(suggested_label_, msg);
		CLONE_OBJ_INTO_MSG(recovery_label_, msg);
		CLONE_OBJ_INTO_MSG(upstream_label_, msg);

		return msg;
	}

	bool PathTearMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(rsvp_hop_);
		// TODO add check for call relative objects
		//CHECK_MANDATORY_OBJ(call_ops_);
		//CHECK_MANDATORY_OBJ(call_id_);
		//CHECK_MANDATORY_OBJ(gns_call_ext_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(sender_template_);
		CHECK_OPTIONAL_OBJ(sender_tspec_);
		CHECK_OPTIONAL_OBJ(adspec_);
		CHECK_OPTIONAL_OBJ(rro_);
		CHECK_OPTIONAL_OBJ(suggested_label_);
		CHECK_OPTIONAL_OBJ(recovery_label_);
		CHECK_OPTIONAL_OBJ(upstream_label_);

		return true;
	}

	bool PathTearMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (sender_template_) {
			GETLSPINFO_FROM_SENDTEMP(sender_template_, &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t PathTearMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(session_, len);
		ADD_LENGTH(rsvp_hop_, len);
		ADD_LENGTH(call_ops_, len);
		ADD_LENGTH(call_id_, len);
		ADD_LENGTH(gns_call_ext_, len);

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(sender_template_, len);
		ADD_LENGTH(sender_tspec_, len);
		ADD_LENGTH(adspec_, len);
		ADD_LENGTH(rro_, len);
		ADD_LENGTH(suggested_label_, len);
		ADD_LENGTH(recovery_label_, len);
		ADD_LENGTH(upstream_label_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor &    mc,
				    PathTearMessage & pathTear)
	{
		// TODO parse all PathTearMsg objs
		int length = mc.remainingSize();

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!pathTear.session_);
					mcObj >> pathTear.session_;
					assert(pathTear.session_);
					break;
				case G2RSVP_CLASS_CALL_OPS:
					assert(!pathTear.call_ops_);
					mcObj >> pathTear.call_ops_;
					assert(pathTear.call_ops_);
					break;
				case G2RSVP_CLASS_CALL_ID:
					assert(!pathTear.call_id_);
					mcObj >> pathTear.call_id_;
					assert(pathTear.call_id_);
					break;
				case G2RSVP_CLASS_RSVP_HOP:
					assert(!pathTear.rsvp_hop_);
					mcObj >> pathTear.rsvp_hop_;
					assert(pathTear.rsvp_hop_);
					break;
				case G2RSVP_CLASS_GNS_CALL_EXT:
					assert(!pathTear.gns_call_ext_);
					mcObj >> pathTear.gns_call_ext_;
					assert(pathTear.gns_call_ext_);
					break;
				case G2RSVP_CLASS_INTEGRITY:
					assert(!pathTear.integrity_);
					mcObj >> pathTear.integrity_;
					assert(pathTear.integrity_);
					break;
				case G2RSVP_CLASS_SENDER_TEMPLATE:
					assert(!pathTear.sender_template_);
					mcObj >> pathTear.sender_template_;
					assert(pathTear.sender_template_);
					break;
				case G2RSVP_CLASS_SENDER_TSPEC:
					assert(!pathTear.sender_tspec_);
					mcObj >> pathTear.sender_tspec_;
					assert(pathTear.sender_tspec_);
					break;
				case G2RSVP_CLASS_ADSPEC:
					// to be filled up
					throw(std::string("ADSPEC !supported"));
				case G2RSVP_CLASS_RECORDED_ROUTE:
					assert(!pathTear.rro_);
					mcObj >> pathTear.rro_;
					assert(pathTear.rro_);
					break;
				case G2RSVP_CLASS_SUGGESTED_LABEL:
					assert(!pathTear.suggested_label_);
					mcObj >> pathTear.suggested_label_;
					assert(pathTear.suggested_label_);
					break;
				case G2RSVP_CLASS_UPSTREAM_LABEL:
					assert(!pathTear.upstream_label_);
					mcObj >> pathTear.upstream_label_;
					assert(pathTear.upstream_label_);
					break;
				case G2RSVP_CLASS_RECOVERY_LABEL:
					assert(!pathTear.recovery_label_);
					mcObj >> pathTear.recovery_label_;
					assert(pathTear.recovery_label_);
					break;

				case G2RSVP_CLASS_POLICY:
				case G2RSVP_CLASS_ERROR_SPEC:
				case G2RSVP_CLASS_FILTER_SPEC:
				case G2RSVP_CLASS_LABEL:
				case G2RSVP_CLASS_FLOWSPEC:
				case G2RSVP_CLASS_TIME_VALUES:
				case G2RSVP_CLASS_ADMIN_STATUS:
				case G2RSVP_CLASS_NOTIFY_REQUEST:
				case G2RSVP_CLASS_MESSAGE_ID:
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
				case G2RSVP_CLASS_RESV_CONFIRM:
				case G2RSVP_CLASS_SCOPE:
				case G2RSVP_CLASS_STYLE:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_GNS_UNI:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in PathTearMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor &          mc,
				    const PathTearMessage & pathTear)
	{
		// All PathTearMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(pathTear.session_, cur); //it's the first mandatory
		assert(pathTear.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (pathTear.session_->length() - 1));
		cur << *(pathTear.session_);

		OBJ2CURSOR(pathTear.integrity_, cur);
		//OBJ2CURSOR(pathTear.session_, cur);
		OBJ2CURSOR(pathTear.rsvp_hop_, cur);
		OBJ2CURSOR(pathTear.call_ops_, cur);
		OBJ2CURSOR(pathTear.call_id_, cur);
		OBJ2CURSOR(pathTear.gns_call_ext_, cur);

		OBJ2CURSOR(pathTear.sender_template_, cur);
		OBJ2CURSOR(pathTear.sender_tspec_, cur);
		OBJ2CURSOR(pathTear.adspec_, cur);
		OBJ2CURSOR(pathTear.rro_, cur);
		OBJ2CURSOR(pathTear.suggested_label_, cur);
		OBJ2CURSOR(pathTear.recovery_label_, cur);
		OBJ2CURSOR(pathTear.upstream_label_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream &          os,
				    const PathTearMessage & pathTear)
	{
		// All PathTearMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(pathTear.integrity_, os);
		OBJ2STREAM(pathTear.session_, os);
		OBJ2STREAM(pathTear.rsvp_hop_, os);
		OBJ2STREAM(pathTear.call_ops_, os);
		OBJ2STREAM(pathTear.call_id_, os);
		OBJ2STREAM(pathTear.gns_call_ext_, os);

		OBJ2STREAM(pathTear.sender_template_, os);
		OBJ2STREAM(pathTear.sender_tspec_, os);
		OBJ2STREAM(pathTear.adspec_, os);
		OBJ2STREAM(pathTear.rro_, os);
		OBJ2STREAM(pathTear.suggested_label_, os);
		OBJ2STREAM(pathTear.recovery_label_, os);
		OBJ2STREAM(pathTear.upstream_label_, os);

		return os;
	}


	/////////////////////////////////
	//    ResvTearMessage class    //
	/////////////////////////////////
	ResvTearMessage::ResvTearMessage(void) : Message(RESV_TEAR)
	{
		session_   = 0;
		rsvp_hop_  = 0;
		style_     = 0;
		flow_desc_ = 0;

		integrity_ = 0;
		scope_     = 0;
	}

	ResvTearMessage::ResvTearMessage(uint8_t version) : Message(version, RESV_TEAR)
	{
		session_   = 0;
		rsvp_hop_  = 0;
		style_     = 0;
		flow_desc_ = 0;

		integrity_ = 0;
		scope_     = 0;
	}

	ResvTearMessage::ResvTearMessage(const ResvTearMessage & resvTear) :
		Message(RESV_TEAR)
	{
		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!
	}

	ResvTearMessage::ResvTearMessage(const ResvMessage & resv) :
		Message(resv.getVersion(), RESV_TEAR)
	{
		// TODO FIXME copy constructor or copy operator ???
		// ... at least mandatory objects

		session_   = 0;
		rsvp_hop_  = 0;
		style_     = 0;
		flow_desc_ = 0;

		integrity_ = 0;
		scope_     = 0;

		//FIX Mandatory???
		assert(resv.rsvp_hop_);
		rsvp_hop_ = resv.rsvp_hop_->clone();
		if (!rsvp_hop_) {
			throw(std::string("Impossible to allocate "
					  "rsvp_hop obj for ResvTear msg!"));
		}

		if (resv.integrity_) {
			integrity_ = resv.integrity_->clone();
			if (!integrity_) {
				//delete rsvp_hop_;
				throw(std::string("Impossible to allocate "
						  "integrity obj for ResvTear "
						  "msg!"));
			}
		}

		if (resv.scope_) {
			scope_ = resv.scope_->clone();
			if (!scope_) {
				//delete rsvp_hop_;
				//delete integrity_;
				throw(std::string("Impossible to allocate scope"
						  " obj for ResvTear msg!"));
			}
		}

		if (resv.session_) {
			session_ = resv.session_->clone();
			if (!session_) {
				//delete rsvp_hop_;
				//delete integrity_;
				//delete scope_;
				throw(std::string("Impossible to allocate sess"
						  "ion obj for ResvTear msg!"));
			}
		}

		if (resv.style_) {
			style_ = resv.style_->clone();
			if (!style_) {
				//delete rsvp_hop_;
				//delete integrity_;
				//delete scope_;
				//delete session_;
				throw(std::string("Impossible to allocate style"
						  " obj for ResvTear msg!"));
			}
		}

		//FIX label/filterSpec together, label Mandatory???
		if (resv.flow_desc_) {
			flow_desc_ = resv.flow_desc_->clone();
			if (!flow_desc_) {
				//delete rsvp_hop_;
				//delete integrity_;
				//delete scope_;
				//delete session_;
				//delete style_;
				throw(std::string("Impossible to allocate FlowD"
						  "esc Obj for ResvTear msg!"));
			}
		}
	}

	ResvTearMessage::~ResvTearMessage(void)
	{
		G2RSVPTE_DBG("Destroying ResvTearMessage %p ...", this);

		DELETE_OBJ(session_);
		DELETE_OBJ(style_);
		DELETE_OBJ(rsvp_hop_);
		DELETE_OBJ(flow_desc_);

		DELETE_OBJ(integrity_);
		DELETE_OBJ(scope_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	Message * ResvTearMessage::clone(void) const
	{
		ResvTearMessage * msg = new ResvTearMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(style_, msg);
		CLONE_OBJ_INTO_MSG(rsvp_hop_, msg);
		CLONE_OBJ_INTO_MSG(flow_desc_, msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);
		CLONE_OBJ_INTO_MSG(scope_, msg);

		return msg;
	}

	bool ResvTearMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(style_);
		CHECK_MANDATORY_OBJ(rsvp_hop_);
		CHECK_MANDATORY_OBJ(flow_desc_);

		CHECK_OPTIONAL_OBJ(integrity_);
		CHECK_OPTIONAL_OBJ(scope_);

		return true;
	}

	bool ResvTearMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (flow_desc_ && flow_desc_->getFilterSpecObj()) {
			GETLSPINFO_FROM_FILTERSPEC(flow_desc_->
						   getFilterSpecObj(),
						   &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t ResvTearMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(session_, len);
		ADD_LENGTH(style_, len);
		ADD_LENGTH(rsvp_hop_, len);
		ADD_LENGTH(flow_desc_, len);

		ADD_LENGTH(integrity_, len);
		ADD_LENGTH(scope_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor &    mc,
				    ResvTearMessage & resvTear)
	{
		// TODO parse all ResvTearMsg objs
		int length = mc.remainingSize();

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!resvTear.session_);
					mcObj >> resvTear.session_;
					assert(resvTear.session_);
					break;
				case G2RSVP_CLASS_RSVP_HOP:
					assert(!resvTear.rsvp_hop_);
					mcObj >> resvTear.rsvp_hop_;
					assert(resvTear.rsvp_hop_);
					break;

				case G2RSVP_CLASS_INTEGRITY:
					assert(!resvTear.integrity_);
					mcObj >> resvTear.integrity_;
					assert(resvTear.integrity_);
					break;
				case G2RSVP_CLASS_SCOPE:
					assert(!resvTear.scope_);
					mcObj >> resvTear.scope_;
					assert(resvTear.scope_);
					break;
				case G2RSVP_CLASS_STYLE:
					assert(!resvTear.style_);
					mcObj >> resvTear.style_;
					assert(resvTear.style_);

					if (resvTear.style_->cType() ==
					    CTYPE_STYLE                 &&
					    static_cast<Style_Object *>
					    (resvTear.style_)->isFF()) {
						assert(!resvTear.flow_desc_);
						FF_FlowDesc_t * ff;
						ff = new FF_FlowDesc_t();
						assert(ff);
						resvTear.flow_desc_ = new
							Flow_Desc_List(ff);
						assert(resvTear.flow_desc_);
					} else {
						throw(std::string
						      ("Only FF STYLE obj is "
						       "supported now"));
					}
					break;

				// START - objs related to FLOW DESCRIPTOR
				case G2RSVP_CLASS_FILTER_SPEC: {
					CHECK_FLOW_DESC(resvTear.flow_desc_);

					FilterSpec_Class * filter = 0;
					mcObj >> filter;
					assert(filter);

					if (!resvTear.flow_desc_->
					    setFilterSpecObj(filter)) {
						throw(std::string
						      ("Impossible to add "
						       "FilterSpec Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_LABEL: {
					CHECK_FLOW_DESC(resvTear.flow_desc_);

					Label_Class * label = 0;
					mcObj >> label;
					assert(label);

					if (!resvTear.flow_desc_->
					    setLabelObj(label)) {
						throw(std::string
						      ("Impossible to add "
						       "Label Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_RECORDED_ROUTE: {
					CHECK_FLOW_DESC(resvTear.flow_desc_);

					RecordedRoute_Class * rro = 0;
					mcObj >> rro;
					assert(rro);

					if (!resvTear.flow_desc_->
					    setRROobj(rro)) {
						throw(std::string
						      ("Impossible to add "
						       "RRO Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_FLOWSPEC: {
					CHECK_FLOW_DESC(resvTear.flow_desc_);

					FlowSpec_Class * flow = 0;
					mcObj >> flow;
					assert(flow);

					if (!resvTear.flow_desc_->
					    setFlowSpecObj(flow)) {
						throw(std::string
						      ("Impossible to add "
						       "FlowSpec Obj"));
					}
					break;
				}
				// END - objs related to FLOW DESCRIPTOR

				case G2RSVP_CLASS_MESSAGE_ID:
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
				case G2RSVP_CLASS_POLICY:
				case G2RSVP_CLASS_ERROR_SPEC:
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_TIME_VALUES:
				case G2RSVP_CLASS_ADMIN_STATUS:
				case G2RSVP_CLASS_GNS_CALL_EXT:
				case G2RSVP_CLASS_CALL_OPS:
				case G2RSVP_CLASS_CALL_ID:
				case G2RSVP_CLASS_NOTIFY_REQUEST:
				case G2RSVP_CLASS_RESV_CONFIRM:
				case G2RSVP_CLASS_SUGGESTED_LABEL:
				case G2RSVP_CLASS_UPSTREAM_LABEL:
				case G2RSVP_CLASS_RECOVERY_LABEL:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_SENDER_TEMPLATE:
				case G2RSVP_CLASS_SENDER_TSPEC:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_GNS_UNI:
				case G2RSVP_CLASS_ADSPEC:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in ResvTearMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor &          mc,
				    const ResvTearMessage & resvTear)
	{
		// All ResvTearMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(resvTear.session_, cur); //it's the first mandatory
		assert(resvTear.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (resvTear.session_->length() - 1));
		cur << *(resvTear.session_);

		OBJ2CURSOR(resvTear.integrity_, cur);
		//OBJ2CURSOR(resvTear.session_, cur);
		OBJ2CURSOR(resvTear.rsvp_hop_, cur);
		OBJ2CURSOR(resvTear.scope_, cur);
		OBJ2CURSOR(resvTear.style_, cur);
		OBJ2CURSOR(resvTear.flow_desc_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream &          os,
				    const ResvTearMessage & resvTear)
	{
		// All ResvTearMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(resvTear.integrity_, os);
		OBJ2STREAM(resvTear.session_, os); //it's the first mandatory
		OBJ2STREAM(resvTear.rsvp_hop_, os);
		OBJ2STREAM(resvTear.scope_, os);
		OBJ2STREAM(resvTear.style_, os);
		OBJ2STREAM(resvTear.flow_desc_, os);

		return os;
	}


	/////////////////////////////////
	//     ResvConfMessage class   //
	/////////////////////////////////
	ResvConfMessage::ResvConfMessage(void) : Message(RESV_CONF)
	{
		session_      = 0;
		error_spec_   = 0;
		style_        = 0;
		resv_confirm_ = 0;
		flow_desc_    = 0;

		integrity_    = 0;
	}

	ResvConfMessage::ResvConfMessage(uint8_t version) : Message(version, RESV_CONF)
	{
		session_      = 0;
		error_spec_   = 0;
		style_        = 0;
		resv_confirm_ = 0;
		flow_desc_    = 0;

		integrity_    = 0;
	}

	ResvConfMessage::ResvConfMessage(const PathMessage & path,
					 const ResvMessage & resv) :
		Message(path.getVersion(), RESV_CONF)
	{
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields from a path msg!!!
		integrity_  = 0;

		assert(path.session_);
		assert(resv.flow_desc_);
		assert(resv.style_);

		session_ = path.session_->clone();
		if (!session_) {
			throw(std::string("Impossible to allocate session"
					  " obj for ResvConf msg"));
		}

		//assert(style_);
		//if (style_->cType() == CTYPE_STYLE
		//    && static_cast<Style_Object *>(style_)->isFF()) {
		FF_FlowDesc_t * ff;
		ff = new FF_FlowDesc_t();
		assert(ff);
		flow_desc_ = new Flow_Desc_List(ff);
		assert(flow_desc_);

		FlowSpec_Class * flowTmp;
		flowTmp = resv.flow_desc_->getFlowSpecObj();
		if (flowTmp) {
			FlowSpec_Class * flow = flowTmp->clone();
			if (!flow) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to allocate flow"
						  "Spec obj for ResvConf msg"));
			}
			if (!flow_desc_->setFlowSpecObj(flow)) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to add FlowSpec "
						  "Obj"));
			}
		}

		FilterSpec_Class * filterTmp;
		filterTmp = resv.flow_desc_->getFilterSpecObj();
		if (filterTmp) {
			FilterSpec_Class * filter = filterTmp->clone();
			if (!filter) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to allocate filtr"
						  "Spec obj for ResvConf msg"));
			}
			if (!flow_desc_->setFilterSpecObj(filter)) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to add FilterSpec"
						  " Obj"));
			}
		}

		Label_Class * labelTmp;
		labelTmp = resv.flow_desc_->getLabelObj();
		if (labelTmp) {
			Label_Class * label = labelTmp->clone();
			if (!label) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to allocate label"
						  " obj for ResvConf msg"));
			}
			if (!flow_desc_->setLabelObj(label)) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to add Label"
						  " Obj"));
			}
		}

		if (resv.resv_confirm_) {
			resv_confirm_ = resv.resv_confirm_->clone();
			if (!resv_confirm_) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to allocate resv"
						  "Conf obj for ResvConf msg"));
			}
		} else {
			if (path.session_->cType() !=
			    CTYPE_SESSION_LSP_TUNNEL_IPV4) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to create resv"
						  "Conf obj for ResvConf msg"));
			}
			resv_confirm_ = new ResvConfirmIPv4_Object
				(static_cast<LSPTunnelIPv4Session_Object *>
				 (path.session_)->get().ipv4_addr);
			if (!resv_confirm_) {
				//delete session_;
				//delete flow_desc__;
				throw(std::string("Impossible to allocate resv"
						  "Conf obj for ResvConf msg"));
			}
		}

		style_ = resv.style_->clone();
		if (!style_) {
			//delete session_;
			//delete flow_desc__;
			//delete resv_confirm_;
			throw(std::string("Impossible to allocate style"
					  " obj for ResvConf msg!"));
		}

		error_spec_ = new IFIDIPv4ErrorSpec_Object();
		if (!error_spec_) {
			//delete session_;
			//delete flow_desc__;
			//delete resv_confirm_;
			//delete style_;
			throw(std::string("Impossible to allocate erroSpec"
					  " obj for ResvConf msg!"));
		}
	}

	ResvConfMessage::ResvConfMessage(const ResvConfMessage & resvConf) :
		Message(RESV_CONF)
	{
		// Copy constructor
		//TODO FIXME: copy constructor MUST copy at least
		//            the mandatory fields!!!
	}

	ResvConfMessage::~ResvConfMessage(void)
	{
		G2RSVPTE_DBG("Destroying ResvConfMessage %p ...", this);

		DELETE_OBJ(session_);
		DELETE_OBJ(error_spec_);
		DELETE_OBJ(style_);
		DELETE_OBJ(resv_confirm_);
		DELETE_OBJ(flow_desc_);

		DELETE_OBJ(integrity_);

		G2RSVPTE_DBG("Message %p destroyed successfully", this);
	}

	Message * ResvConfMessage::clone(void) const
	{
		ResvConfMessage * msg = new ResvConfMessage(getVersion());
		assert(msg);

		CLONE_OBJ_INTO_MSG(session_, msg);
		CLONE_OBJ_INTO_MSG(error_spec_, msg);
		CLONE_OBJ_INTO_MSG(resv_confirm_, msg);
		CLONE_OBJ_INTO_MSG(style_, msg);
		CLONE_OBJ_INTO_MSG(flow_desc_, msg);

		CLONE_OBJ_INTO_MSG(integrity_, msg);

		return msg;
	}

	bool ResvConfMessage::isConsistent(void) const
	{
		CHECK_MANDATORY_OBJ(session_);
		CHECK_MANDATORY_OBJ(error_spec_);
		CHECK_MANDATORY_OBJ(style_);
		CHECK_MANDATORY_OBJ(resv_confirm_);
		CHECK_MANDATORY_OBJ(flow_desc_);

		CHECK_OPTIONAL_OBJ(integrity_);

		return true;
	}

	bool ResvConfMessage::getLSPinfo(lsp_ident_t & info)
	{
		if (session_) {
			GETLSPINFO_FROM_SESSION(session_, &info);
		} else {
			return false;
		}

		if (flow_desc_ && flow_desc_->getFilterSpecObj()) {
			GETLSPINFO_FROM_FILTERSPEC(flow_desc_->
						   getFilterSpecObj(),
						   &info);
		} else {
			return false;
		}

		return true;
	}

	uint16_t ResvConfMessage::getLength(void) const
	{
		uint16_t len = msgHdrLength;

		ADD_LENGTH(session_, len);
		ADD_LENGTH(error_spec_, len);
		ADD_LENGTH(resv_confirm_, len);
		ADD_LENGTH(style_, len);
		ADD_LENGTH(flow_desc_, len);

		ADD_LENGTH(integrity_, len);

		return len;
	}

	MemoryCursor & operator >> (MemoryCursor &    mc,
				    ResvConfMessage & resvConf)
	{
		// TODO parse all ResvConfMsg objs
		int length = mc.remainingSize();

		uint8_t  cNum;
		uint8_t  cType;
		uint16_t len;

		MemoryCursor mcObjHdr(mc);
		mcObjHdr.resize(mc.start(),
				mc.start() + (objHeader::objHdrLength - 1));

		while (length > 0) {

			mcObjHdr >> len >> cNum >> cType;
			assert(mcObjHdr.remainingSize() == 0);

			MemoryCursor mcObj(mc);
			mcObj.resize(mcObjHdr.start(),
				     mcObjHdr.start() + (len - 1));

			switch (cNum) {
				case G2RSVP_CLASS_SESSION:
					assert(!resvConf.session_);
					mcObj >> resvConf.session_;
					assert(resvConf.session_);
					break;
				case G2RSVP_CLASS_ERROR_SPEC:
					assert(!resvConf.error_spec_);
					mcObj >> resvConf.error_spec_;
					assert(resvConf.error_spec_);
					break;
				case G2RSVP_CLASS_RESV_CONFIRM:
					assert(!resvConf.resv_confirm_);
					mcObj >> resvConf.resv_confirm_;
					assert(resvConf.resv_confirm_);
					break;
				case G2RSVP_CLASS_INTEGRITY:
					assert(!resvConf.integrity_);
					mcObj >> resvConf.integrity_;
					assert(resvConf.integrity_);
					break;
				case G2RSVP_CLASS_STYLE:
					assert(!resvConf.style_);
					mcObj >> resvConf.style_;
					assert(resvConf.style_);

					if (resvConf.style_->cType() ==
					    CTYPE_STYLE                 &&
					    static_cast<Style_Object *>
					    (resvConf.style_)->isFF()) {
						assert(!resvConf.flow_desc_);
						FF_FlowDesc_t * ff;
						ff = new FF_FlowDesc_t();
						assert(ff);
						resvConf.flow_desc_ = new
							Flow_Desc_List(ff);
						assert(resvConf.flow_desc_);
					} else {
						throw(std::string
						      ("Only FF STYLE obj is "
						       "supported now"));
					}
					break;

				// START - objs related to FLOW DESCRIPTOR
				case G2RSVP_CLASS_FILTER_SPEC: {
					CHECK_FLOW_DESC(resvConf.flow_desc_);

					FilterSpec_Class * filter = 0;
					mcObj >> filter;
					assert(filter);

					if (!resvConf.flow_desc_->
					    setFilterSpecObj(filter)) {
						throw(std::string
						      ("Impossible to add "
						       "FilterSpec Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_LABEL: {
					CHECK_FLOW_DESC(resvConf.flow_desc_);

					Label_Class * label = 0;
					mcObj >> label;
					assert(label);

					if (!resvConf.flow_desc_->
					    setLabelObj(label)) {
						throw(std::string
						      ("Impossible to add "
						       "Label Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_RECORDED_ROUTE: {
					CHECK_FLOW_DESC(resvConf.flow_desc_);

					RecordedRoute_Class * rro = 0;
					mcObj >> rro;
					assert(rro);

					if (!resvConf.flow_desc_->
					    setRROobj(rro)) {
						throw(std::string
						      ("Impossible to add "
						       "RRO Obj"));
					}
					break;
				}
				case G2RSVP_CLASS_FLOWSPEC: {
					CHECK_FLOW_DESC(resvConf.flow_desc_);

					FlowSpec_Class * flow = 0;
					mcObj >> flow;
					assert(flow);

					if (!resvConf.flow_desc_->
					    setFlowSpecObj(flow)) {
						throw(std::string
						      ("Impossible to add "
						       "FlowSpec Obj"));
					}
					break;
				}
				// END - objs related to FLOW DESCRIPTOR


				case G2RSVP_CLASS_RSVP_HOP:
				case G2RSVP_CLASS_MESSAGE_ID:
				case G2RSVP_CLASS_MESSAGE_ID_ACK_NACK:
				case G2RSVP_CLASS_POLICY:
				case G2RSVP_CLASS_ACCEPTABLE_LABEL_SET:
				case G2RSVP_CLASS_SCOPE:
				case G2RSVP_CLASS_TIME_VALUES:
				case G2RSVP_CLASS_ADMIN_STATUS:
				case G2RSVP_CLASS_GNS_CALL_EXT:
				case G2RSVP_CLASS_CALL_OPS:
				case G2RSVP_CLASS_CALL_ID:
				case G2RSVP_CLASS_NOTIFY_REQUEST:
				case G2RSVP_CLASS_SUGGESTED_LABEL:
				case G2RSVP_CLASS_UPSTREAM_LABEL:
				case G2RSVP_CLASS_RECOVERY_LABEL:
				case G2RSVP_CLASS_LABEL_REQ:
				case G2RSVP_CLASS_LABEL_SET:
				case G2RSVP_CLASS_EXPLICIT_ROUTE:
				case G2RSVP_CLASS_SENDER_TEMPLATE:
				case G2RSVP_CLASS_SENDER_TSPEC:
				case G2RSVP_CLASS_PROTECTION:
				case G2RSVP_CLASS_GENERALIZED_UNI:
				case G2RSVP_CLASS_GNS_UNI:
				case G2RSVP_CLASS_ADSPEC:
				case G2RSVP_CLASS_SESSION_ATTRIBUTE:
				case G2RSVP_CLASS_MESSAGE_ID_LIST:
				case G2RSVP_CLASS_HELLO:
				case G2RSVP_CLASS_RESTARTCAP:
				default:
					throw(std::string("Obj type not support"
							  "ed in ResvConfMsg"));
			}

			length -= len;

			if (length > 0) {
				mcObjHdr = mcObj.
					neighbor(objHeader::objHdrLength);
			}
		}

		return mc;
	}

	MemoryCursor & operator << (MemoryCursor &          mc,
				    const ResvConfMessage & resvConf)
	{
		// All ResvConfMsg objs [ACCORDING TO CORRECT ORDER]

		//OBJ2CURSOR(resvConf.session_, cur); //it's the first mandatory
		assert(resvConf.session_); // Mandatory!!!
		MemoryCursor cur(mc);
		cur.resize(mc.start(),
			   mc.start() + (resvConf.session_->length() - 1));
		cur << *(resvConf.session_);

		OBJ2CURSOR(resvConf.integrity_, cur);
		//OBJ2CURSOR(resvConf.session_, cur);
		OBJ2CURSOR(resvConf.error_spec_, cur);
		OBJ2CURSOR(resvConf.resv_confirm_, cur);
		OBJ2CURSOR(resvConf.style_, cur);
		OBJ2CURSOR(resvConf.flow_desc_, cur);

		return mc;
	}

	std::ostream & operator << (std::ostream &          os,
				    const ResvConfMessage & resvConf)
	{
		// All ResvConfMsg objs [ACCORDING TO CORRECT ORDER]

		OBJ2STREAM(resvConf.integrity_, os);
		OBJ2STREAM(resvConf.session_, os); //it's the first mandatory
		OBJ2STREAM(resvConf.error_spec_, os);
		OBJ2STREAM(resvConf.resv_confirm_, os);
		OBJ2STREAM(resvConf.style_, os);
		OBJ2STREAM(resvConf.flow_desc_, os);

		return os;
	}
}
